#include "MUD.h"
#include "types.h"
#include "MUDInternal.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * byteIO.c - routines for managing byte stream stuff.
 */

/*
 * expandWriteBuffer - called to double the size of the write buffer.
 */

void
expandWriteBuffer(MUDState_t *ms, UINT_T wantLen)
{
    BYTE_T *newBuffer;
    UINT_T newLength;
    BOOL_T (*reclaim)(BOOL_T panicing);

    /* The call to memAlloc can make the allocater need more memory than
       it is supposed to try to limit itself to. This would cause it to
       call the memReclaim function, which would try to flush some
       from memory. Unfortunately, this can require writing out the procs,
       which use the byte output stuff here. We do not support that kind
       of recursion, so we must, unfortunately, disable the reclaiming
       around this call. */
    reclaim = ms->ms_ReclaimMem;
    ms->ms_ReclaimMem = P_NULL;
    newLength = ms->WriteBufferSize * 2;
    if (newLength < wantLen) {
	newLength = wantLen;
    }
    newBuffer = MEM_ALLOC(ms, newLength, "WBUF");
    ms->ms_ReclaimMem = reclaim;
    blockCopy(newBuffer, ms->WriteBuffer, ms->WritePos);
    memFree(ms, ms->WriteBuffer, ms->WriteBufferSize);
    ms->WriteBuffer = newBuffer;
    ms->WriteBufferSize = newLength;
}

/*
 * writeResetE - entry to allow ressetting of write stream.
 */

void
writeResetE(MUDState_t *ms)
{

    ms->WritePos = 0;
}

/*
 * writeByte - add a byte to the byte stream.
 */

void
writeByte(MUDState_t *ms, BYTE_T b)
{
    UINT_T writePos;

    writePos = ms->WritePos;
    if (writePos >= ms->WriteBufferSize) {
	expandWriteBuffer(ms, 0);
    }
    *(ms->WriteBuffer + writePos) = b;
    ms->WritePos = writePos + 1;
}

/*
 * writeByteE - external interface.
 */

void
writeByteE(MUDState_t *ms, BYTE_T b)
{

    writeByte(ms, b);
}

/*
 * writeWord - add a 16 bit word to the byte stream.
 */

void
writeWord(MUDState_t *ms, UINT_T w)
{
    BYTE_T *p, *q;
    UINT_T writePos;

    /* Have to be a bit careful here. Cannot use WriteBuffer until after
       the possible expansion, but want the WritePos from before the
       incrementing as the pointer value. */
    writePos = ms->WritePos;
    if (writePos + sizeof(UINT_T) > ms->WriteBufferSize) {
	expandWriteBuffer(ms, 0);
    }
    p = (BYTE_T *) &w;
    q = ms->WriteBuffer + writePos;
#if MUD_ENDIAN == MUD_BIG_ENDIAN
    *q++ = *p++;
    *q = *p;
#else
    ++p;
    *q++ = *p;
    *q = *--p;
#endif
    writePos += sizeof(UINT_T);
    ms->WritePos = writePos;
}

/*
 * writeWordE - external interface.
 */

void
writeWordE(MUDState_t *ms, UINT_T w)
{

    writeWord(ms, w);
}

/*
 * writeLong - add a 32 bit word to the byte stream.
 */

void
writeLong(MUDState_t *ms, ULONG_T l)
{
    BYTE_T *p, *q;
    UINT_T writePos;

    writePos = ms->WritePos;
    if (writePos + sizeof(ULONG_T) > ms->WriteBufferSize) {
	expandWriteBuffer(ms, 0);
    }
    p = (BYTE_T *) &l;
    q = ms->WriteBuffer + writePos;
#if MUD_ENDIAN == MUD_BIG_ENDIAN
    *q++ = *p++;
    *q++ = *p++;
    *q++ = *p++;
    *q = *p;
#else
    p += 3;
    *q++ = *p;
    *q++ = *--p;
    *q++ = *--p;
    *q = *--p;
#endif
    writePos += sizeof(ULONG_T);
    ms->WritePos = writePos;
}

/*
 * writeLongE - external interface.
 */

void
writeLongE(MUDState_t *ms, ULONG_T l)
{

    writeLong(ms, l);
}

/*
 * writeBytes - add a bunch of bytes to the byte stream.
 */

void
writeBytes(MUDState_t *ms, void *b, UINT_T len)
{
    BYTE_T *p, *q;
    UINT_T writePos;

    p = (BYTE_T *) b;
    writePos = ms->WritePos;
    if (writePos + len > ms->WriteBufferSize) {
	expandWriteBuffer(ms, writePos + len);
    }
    q = ms->WriteBuffer + writePos;
    writePos += len;
    while (len != 0) {
	len -= 1;
	*q++ = *p++;
    }
    ms->WritePos = writePos;
}

/*
 * writeBytesE - external interface.
 */

void
writeBytesE(MUDState_t *ms, void *b, UINT_T l)
{

    writeBytes(ms, b, l);
}

/*
 * writeString - add a NUL-terminated string to the BYTE_T stream.
 */

void
writeString(MUDState_t *ms, CHAR_T *st)
{

    while (B_TRUE) {
	writeByte(ms, *st);
	if (*st == '\0') {
	    break;
	}
	++st;
    }
}

/*
 * writeStringE - external interface.
 */

void
writeStringE(MUDState_t *ms, CHAR_T *st)
{

    writeString(ms, st);
}

/*
 * writeResult - return the address and length of current bufferfull.
 */

void
writeResult(MUDState_t *ms, BYTE_T **pBytes, UINT_T *pLen)
{

    *pBytes = ms->WriteBuffer;
    *pLen = ms->WritePos;
}

/*
 * writeResultE - external version of writeResult.
 */

void
writeResultE(MUDState_t *ms, BYTE_T **pBytes, UINT_T *pLen)
{

    *pBytes = ms->WriteBuffer;
    *pLen = ms->WritePos;
}

/*
 * readWord - assemble a word from bytes (avoid alignment problems).
 */

UINT_T
readWord(void **pb)
{
    BYTE_T *p, *q;
    UINT_T w;

    p = *(BYTE_T **) pb;
    q = (BYTE_T *) &w;
#if MUD_ENDIAN == MUD_BIG_ENDIAN
    *q++ = *p++;
    *q = *p++;
    *(BYTE_T **) pb = p;
#else
    ++p;
    *q++ = *p;
    *q = *--p;
    *(BYTE_T **) pb = p + 2;
#endif
    return w;
}

/*
 * readLong - assemble a long from bytes (avoid alignment problems).
 */

ULONG_T
readLong(void **pb)
{
    BYTE_T *p, *q;
    ULONG_T l;

    p = *(BYTE_T **) pb;
    q = (BYTE_T *) &l;
#if MUD_ENDIAN == MUD_BIG_ENDIAN
    *q++ = *p++;
    *q++ = *p++;
    *q++ = *p++;
    *q = *p++;
    *(BYTE_T **) pb = p;
#else
    p += 3;
    *q++ = *p;
    *q++ = *--p;
    *q++ = *--p;
    *q = *--p;
    *(BYTE_T **) pb = p + 4;
#endif
    return l;
}

/*
 * readLongE - external interface.
 */

ULONG_T
readLongE(void **pb)
{

    return readLong(pb);
}
