#include "MUD.h"
#include "types.h"
#include "exec.h"
#include "lex.h"
#include "MUDInternal.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * decl.c - code to handle declarations, proc headers, etc.
 */

/*
 * parseType - parse a type name.
 */

UINT_T
parseType(MUDState_t *ms)
{
    UINT_T t;

    switch (ms->tk_kind) {
    case tk_void:
	t = TYPE_VOID;
	break;
    case tk_status:
	t = TYPE_STATUS;
	break;
    case tk_player:
	t = TYPE_PLAYER;
	break;
    case tk_bool:
	t = TYPE_BOOLEAN;
	break;
    case tk_int:
	t = TYPE_INTEGER;
	break;
    case tk_fixed:
	t = TYPE_FIXED;
	break;
    case tk_string:
	t = TYPE_STRING;
	break;
    case tk_thing:
	t = TYPE_THING;
	break;
    case tk_action:
	t = TYPE_ACTION;
	break;
    case tk_table:
	t = TYPE_TABLE;
	break;
    case tk_grammar:
	t = TYPE_GRAMMAR;
	break;
    case tk_list:
	getToken(ms);
	switch (ms->tk_kind) {
	case tk_int:
	    t = TYPE_LIST_INTEGER;
	    break;
	case tk_fixed:
	    t = TYPE_LIST_FIXED;
	    break;
	case tk_thing:
	    t = TYPE_LIST_THING;
	    break;
	case tk_action:
	    t = TYPE_LIST_ACTION;
	    break;
	default:
	    parseError(ms, "expecting token for list type");
	    t = TYPE_ERROR;
	    break;
	}
	break;
    case tk_property:
	getToken(ms);
	switch (ms->tk_kind) {
	case tk_bool:
	    t = TYPE_PROP_BOOLEAN;
	    break;
	case tk_int:
	    t = TYPE_PROP_INTEGER;
	    break;
	case tk_fixed:
	    t = TYPE_PROP_FIXED;
	    break;
	case tk_string:
	    t = TYPE_PROP_STRING;
	    break;
	case tk_thing:
	    t = TYPE_PROP_THING;
	    break;
	case tk_action:
	    t = TYPE_PROP_ACTION;
	    break;
	case tk_table:
	    t = TYPE_PROP_TABLE;
	    break;
	case tk_grammar:
	    t = TYPE_PROP_GRAMMAR;
	    break;
	case tk_list:
	    getToken(ms);
	    switch (ms->tk_kind) {
	    case tk_int:
		t = TYPE_PROP_LIST_INTEGER;
		break;
	    case tk_fixed:
		t = TYPE_PROP_LIST_FIXED;
		break;
	    case tk_thing:
		t = TYPE_PROP_LIST_THING;
		break;
	    case tk_action:
		t = TYPE_PROP_LIST_ACTION;
		break;
	    default:
		parseError(ms, "expecting token for property list type");
		t = TYPE_ERROR;
		break;
	    }
	    break;
	default:
	    parseError(ms, "expecting token for property type");
	    t = TYPE_ERROR;
	    break;
	}
	break;
    default:
	parseError(ms, "expecting type token");
	t = TYPE_ERROR;
	break;
    }
    if (t != TYPE_ERROR) {
	getToken(ms);
    }
    return t;
}

/*
 * findLocal - local routine to see if a given local name exists.
 */

static BOOL_T
findLocal(LocalList_t *ll, CHAR_T *name)
{

    while (ll != P_NULL) {
	if (charsEqual(ll->ll_name, name)) {
	    return B_TRUE;
	}
	ll = ll->ll_next;
    }
    return B_FALSE;
}

/*
 * addLocal - local routine to append a new local (parm or var).
 */

static LocalList_t *
addLocal(MUDState_t *ms, LocalList_t ***ppTail)
{
    LocalList_t *ll;

    ll = MEM_ALLOC(ms, sizeof(LocalList_t), "LL_T");
    ll->ll_name = MEM_ALLOC(ms, ms->tk_stringLength, "LNAM");
    charsCopy(ll->ll_name, ms->tk_stringValue);
    **ppTail = ll;
    *ppTail = &ll->ll_next;
    ll->ll_next = P_NULL;
    return ll;
}

/*
 * newProcHeader - local utility to build and setup a Proc_t
 */

static Proc_t *
newProcHeader(MUDState_t *ms)
{
    Proc_t *pr;

    pr = MEM_ALLOC(ms, sizeof(Proc_t), "NPRC");
    pr->pr_locals = P_NULL;
    pr->pr_parameters = P_NULL;
    pr->pr_refs = P_NULL;
    pr->pr_body = P_NULL;
    pr->pr_compiled = P_NULL;
    pr->pr_byteCode = P_NULL;
    pr->pr_owner = ms->PlayerKey;
    pr->pr_resultType = TYPE_STATUS;
    pr->pr_useCount = 0;	/* 0 - incremented when symbol defined */
    pr->pr_byteCodeLen = 0;
    pr->pr_localCount = BC_STACK_EXTRA;		/* fp, pc for byte-code */
    pr->pr_utility = B_FALSE;
    pr->pr_wizard = B_FALSE;
    pr->pr_public = B_FALSE;
    pr->pr_isCompiled = B_FALSE;
    pr->pr_stripped = B_FALSE;
    pr->pr_locked = B_FALSE;
    pr->pr_byteSwapped = B_FALSE;
    ms->ParsingProc = pr;
    ms->ParsingProcKey = NIL_ACTION;
    return pr;
}

/*
 * statementStart - return 'true' if token starts a "statement"
 */

static BOOL_T
statementStart(UINT_T kind)
{

    return (BOOL_T) (kind == tk_if || kind == tk_while || kind == tk_for ||
		     kind == tk_case || kind == tk_assign || kind == tk_proc ||
		     kind == tk_corp || kind == tk_eof);
}

/*
 * getParameters - get a set of parameters for a proc header.
 */

static LocalList_t *
getParameters(MUDState_t *ms, UINT_T *pLocalOffset)
{
    LocalList_t **lTail;
    LocalList_t *ll, *localList;
    UINT_T kind, localOffset, parType;

    if (ms->tk_kind == tk_leftParen) {
	getToken(ms);
    } else {
	parseError(ms, "expecting '(' for proc parameter list");
    }
    localOffset = 0;
    lTail = &localList;
    *lTail = P_NULL;
    while (B_TRUE) {
	kind = ms->tk_kind;
	if (kind == tk_rightParen || kind == tk_colon || statementStart(kind)){
	    break;
	}
	parType = parseType(ms);
	if (parType == TYPE_VOID) {
	    parseError(ms, "void is invalid type for variable");
	    parType = TYPE_ERROR;
	}
	while (B_TRUE) {
	    kind = ms->tk_kind;
	    if (kind == tk_semicolon || kind == tk_rightParen ||
		kind == tk_colon || statementStart(kind))
	    {
		break;
	    }
	    if (kind == tk_identifier) {
		if (findLocal(localList, ms->tk_stringValue)) {
		    libSymbolError(ms, ms->tk_stringValue,
				"proc parameter already defined");
		} else {
		    ll = addLocal(ms, &lTail);
		    ll->ll_offset = localOffset;
		    ll->ll_type = parType;
		    localOffset += sizeof(QValue_t);
		}
		getToken(ms);
	    } else {
		parseError(ms, "expecting identifier for proc parameter name");
	    }
	    if (ms->tk_kind == tk_comma) {
		getToken(ms);
	    } else if (ms->tk_kind != tk_semicolon &&
		       ms->tk_kind != tk_rightParen)
	    {
		parseError(ms,
		    "expecting ',', ';' or ')' after proc parameter");
		while (B_TRUE) {
		    kind = ms->tk_kind;
		    if (kind == tk_semicolon || kind == tk_colon ||
			kind == tk_rightParen || kind == tk_identifier ||
			statementStart(kind))
		    {
			break;
		    }
		    getToken(ms);
		}
		if (kind == tk_comma) {
		    getToken(ms);
		}
	    }
	}
	if (ms->tk_kind == tk_semicolon) {
	    getToken(ms);
	} else if (ms->tk_kind != tk_rightParen) {
	    parseError(ms, "expecting ';' or ')' after proc parameter set");
	}
    }
    if (ms->tk_kind == tk_rightParen) {
	getToken(ms);
    } else {
	parseError(ms, "expecting ')' after proc parameters");
    }
    *pLocalOffset = localOffset;
    return localList;
}

/*
 * checkProcEdits - check that the type and parameters of the new proc
 *	match those of the old version.
 */

static void
checkProcEdits(MUDState_t *ms, ULONG_T oldKey, CHAR_T *oldName, Proc_t *pr)
{
    LocalList_t *ll, *oldLL;
    Proc_t *oldPr;
    UINT_T *pLockCount;

    oldPr = findProcOnList(ms, oldKey, &pLockCount);
    if (oldPr->pr_resultType != pr->pr_resultType) {
	libSymbolError(ms, oldName, "cannot change result type");
    }
    oldLL = oldPr->pr_parameters;
    ll = pr->pr_parameters;
    while (ll != P_NULL && oldLL != P_NULL) {
	if (oldLL->ll_type != ll->ll_type) {
	    libSymbolError(ms, ll->ll_name, "cannot change parameter type");
	}
	ll = ll->ll_next;
	oldLL = oldLL->ll_next;
    }
    if (oldLL != ll) {
	libSymbolError(ms, oldName, "cannot change parameter count");
    }
    *pLockCount -= 1;
}

/*
 * getLocals - get a set of local variables.
 */

static void
getLocals(MUDState_t *ms, Proc_t *pr, UINT_T localOffset)
{
    LocalList_t **lTail;
    LocalList_t *ll;
    UINT_T kind, localType;

    /* byte-code interpreter stores the caller's fp and pc in between the
       parameters and the locals */
    localOffset += BC_STACK_EXTRA;
    lTail = &pr->pr_locals;
    *lTail = P_NULL;
    while (B_TRUE) {
	kind = ms->tk_kind;
	if (kind < tk_void || kind > tk_property) {
	    break;
	}
	localType = parseType(ms);
	while (B_TRUE) {
	    kind = ms->tk_kind;
	    if (kind == tk_semicolon || statementStart(kind)) {
		break;
	    }
	    if (kind == tk_identifier) {
		if (findLocal(pr->pr_parameters, ms->tk_stringValue)) {
		    libSymbolError(ms, ms->tk_stringValue,
				"proc local already defined as parameter");
		} else if (findLocal(pr->pr_locals, ms->tk_stringValue)) {
		    libSymbolError(ms, ms->tk_stringValue,
				"proc local already defined");
		} else {
		    ll = addLocal(ms, &lTail);
		    ll->ll_offset = localOffset;
		    ll->ll_type = localType;
		    localOffset += sizeof(QValue_t);
		}
		getToken(ms);
	    } else {
		parseError(ms, "expecting identifier for proc local name");
	    }
	    if (ms->tk_kind == tk_comma) {
		getToken(ms);
	    } else if (ms->tk_kind != tk_semicolon &&
		ms->tk_kind != tk_rightParen)
	    {
		parseError(ms, "expecting ',', ';' or ')' after proc local");
		while (B_TRUE) {
		    kind = ms->tk_kind;
		    if (kind == tk_semicolon || kind == tk_identifier ||
			statementStart(kind))
		    {
			break;
		    }
		    getToken(ms);
		}
		if (kind == tk_comma) {
		    getToken(ms);
		}
	    }
	}
	if (ms->tk_kind == tk_semicolon) {
	    getToken(ms);
	} else {
	    parseError(ms, "expecting ';' or ')' after proc local set");
	}
    }
    pr->pr_localCount = localOffset;
}

/*
 * quoteName - utility, part of parseProc.
 */

static void
quoteName(MUDState_t *ms, CHAR_T *name)
{

    if (name != P_NULL) {
	outputString(ms, " \"");
	outputString(ms, name);
	ms->ms_outputChar('\"');
    }
}

/*
 * parseProc - top level of parsing a proc definition.
 */

void
parseProc(MUDState_t *ms, ULONG_T table, ULONG_T oldKey, CHAR_T *oldName,
	  BOOL_T errorDefine, BOOL_T verbose)
{
    Proc_t *pr;
    CHAR_T *procName;
    ULONG_T key;
    UINT_T procNameLen, resType, localOffset, kind;
    BOOL_T isUtility, isWizard, isPublic, isCompiled, isStripped, noSymbol;

    if (ms->tk_kind == tk_proc) {
	getToken(ms);
    }
    isUtility = B_FALSE;
    isWizard = B_FALSE;
    isPublic = B_FALSE;
    isCompiled = B_FALSE;
    isStripped = B_FALSE;
    while (B_TRUE) {
	kind = ms->tk_kind;
	if (kind != tk_utility && kind != tk_wizard &&
	    kind != tk_public && kind != tk_compiled && kind != tk_stripped)
	{
	    break;
	}
	if (kind == tk_utility) {
	    isUtility = B_TRUE;
	} else if (kind == tk_wizard) {
	    isWizard = B_TRUE;
	} else if (kind == tk_public) {
	    isPublic = B_TRUE;
	} else if (kind == tk_compiled) {
	    isCompiled = B_TRUE;
	} else {
	    isStripped = B_TRUE;
	}
	getToken(ms);
    }
    if (isStripped && ! isCompiled) {
	parseError(ms, "cannot have stripped but not compiled proc");
	isStripped = B_FALSE;
    }
    noSymbol = B_TRUE;
    procName = P_NULL;	/* shut gcc up */
    procNameLen = 0;	/* shut gcc up */
    if (ms->tk_kind != tk_identifier) {
	if (oldKey == NIL_ACTION || oldName != P_NULL) {
	    parseError(ms, "expecting identifier for proc name");
	}
    } else {
	if (oldKey != NIL_ACTION) {
	    if (oldName != P_NULL) {
		if (! charsEqual(oldName, ms->tk_stringValue)) {
		    libSymbolError(ms, oldName, "cannot change proc name");
		}
	    } else {
		parseError(ms, "cannot give name by editing a proc");
	    }
	} else if (libLookup(ms, ms->tk_stringValue,ms->tk_stringLength) == 0){
	    procNameLen = ms->tk_stringLength;
	    procName = MEM_ALLOC(ms, procNameLen, "PRNM");
	    charsCopy(procName, ms->tk_stringValue);
	    noSymbol = B_FALSE;
	} else {
	    libSymbolError(ms, ms->tk_stringValue, "symbol already defined");
	}
	getToken(ms);
    }
    pr = newProcHeader(ms);
    pr->pr_utility = isUtility;
    pr->pr_wizard = isWizard;
    pr->pr_public = isPublic;
    pr->pr_isCompiled = isCompiled;
    pr->pr_stripped = isStripped;

    /* go parse and set up the parameters */

    pr->pr_parameters = getParameters(ms, &localOffset);

    pr->pr_resultType = parseType(ms);

    if (ms->tk_kind == tk_colon) {
	getToken(ms);
    } else {
	parseError(ms, "expecting ':' after proc header");
    }

    /* check against old header if we are editing a proc */

    if (oldKey != NIL_ACTION) {
	checkProcEdits(ms, oldKey, oldName, pr);
    }

    /* do this here so that we do not bother sending over the locals */

    ms->ParsingProc = pr;
    if (oldKey != NIL_ACTION) {
	key = oldKey;
    } else if (! noSymbol) {
	key = writeProcNew(ms, pr);
	defineSymbol(ms, procName, procNameLen, key, table);
    } else {
	key = NIL_ACTION;
    }
    ms->ParsingProcKey = key;

    /* go parse and set up the local variables */

    getLocals(ms, pr, localOffset);

    /* go parse and set up the proc body */

    parseBody(ms, &pr->pr_body, &resType, B_TRUE);
    if (resType == TYPE_NIL && libNilType(pr->pr_resultType)) {
	resType = pr->pr_resultType;
    }
    if (resType != pr->pr_resultType) {
	parseError(ms, "proc result type does not match declared type");
    }

    /* ending stuff */

    if (ms->tk_kind != tk_corp) {
	parseError(ms, "expecting 'corp' to end proc definition");
    }
    ms->ParsingProc = P_NULL;
    ms->ParsingProcKey = NIL_ACTION;

    /* all done, now define or replace it if all ok */

    if (oldKey != NIL_ACTION) {
	if (ms->ErrorCount != 0) {
	    errorHeader(ms);
	    errorMessage(ms, "Errors detected, proc");
	    /* can't use 'quoteName' - want to go through 'errorMessage' */
	    if (oldName != P_NULL) {
		errorMessage(ms, " \"");
		errorMessage(ms, oldName);
		errorMessage(ms, "\"");
	    }
	    errorMessage(ms, " not updated.\n");
	    errorTrailer(ms);
	    /* errors, so free any refs here */
	    freeProc(ms, pr, key, B_TRUE);
	} else {
	    replaceProc(ms, key, pr);
	    /* refs are now owned by the server */
	    freeProc(ms, pr, key, B_FALSE);
	    if (verbose) {
		outputString(ms, "Proc");
		quoteName(ms, oldName);
		outputString(ms, " updated.\n");
	    }
	}
    } else {
	if ((ms->ErrorCount != 0 && ! errorDefine) || noSymbol) {
	    outputString(ms, "Errors detected, proc");
	    if (! noSymbol) {
		quoteName(ms, procName);
		deleteSymbol(ms, procName, procNameLen, key, table);
	    }
	    outputString(ms, " not created.\n");
	    /* free refs - the proc will no longer exist */
	    freeProc(ms, pr, key, B_TRUE);
	} else {
	    pr->pr_useCount = 1;   /* the symbol table reference */
	    if (ms->ErrorCount != 0) {
		/* had errors, but we want to keep it around */
		clearProc(ms, pr, key);
		pr->pr_isCompiled = B_FALSE;
		pr->pr_stripped = B_FALSE;
	    }
	    replaceProc(ms, key, pr);
	    /* note the 'false' - the refs are over in the server, so we leave
	       them alone here. */
	    freeProc(ms, pr, key, B_FALSE);
	}
	if (! noSymbol) {
	    memFree(ms, procName, procNameLen);
	}
    }
}

/*
 * parseAction - parse an explicit (non-proc) action.
 */

void
parseAction(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    Exec_t *ex;
    Proc_t *pr;
    Proc_t *saveCurrent;
    ULONG_T saveCurrentKey;
    UINT_T oldErrorCount;

    getToken(ms);
    saveCurrent = ms->ParsingProc;
    saveCurrentKey = ms->ParsingProcKey;
    oldErrorCount = ms->ErrorCount;
    pr = newProcHeader(ms);
    parseBody(ms, &pr->pr_body, &pr->pr_resultType, B_FALSE);
    if (pr->pr_resultType == TYPE_NIL) {
	parseError(ms, "direct proc cannot return nil (type unknown)");
    }
    if (ms->tk_kind == tk_corp) {
	getToken(ms);
    } else {
	parseError(ms, "expecting 'corp' to end proc definition");
    }
    ms->ParsingProc = saveCurrent;
    ms->ParsingProcKey = saveCurrentKey;
    ex = MEM_ALLOC(ms, sizeof(Exec_t), "BODY");
    if (ms->ErrorCount != oldErrorCount) {
	ex->ex_kind = ex_error;
	/* note the 'true' - here we do not pass the proc onto the server,
	   so we want to decrement the refs. */
	freeProc(ms, pr, NIL_ACTION, B_TRUE);
	*pType = TYPE_ERROR;
    } else {
	ex->ex_kind = ex_procRef;
	ex->ex_v.ex_refPtr = makeRef(ms, writeProcNew(ms, pr));
	/* note the 'false' - we do not want to decrement refs here, since
	   we have passed the proc over to the server who will have refs
	   there */
	freeProc(ms, pr, NIL_ACTION, B_FALSE);
	*pType = TYPE_ACTION;
    }
    *pEx = ex;
}

/*
 * parseImmediateAction - parse a statement sequence, bundling it into a proc
 *	for use by the server.
 */

Proc_t *
parseImmediateAction(MUDState_t *ms, BOOL_T allowDir)
{
    Proc_t *pr;

    pr = newProcHeader(ms);
    ms->ParsingProc = pr;
    ms->ErrorCount = 0;
    parseBody(ms, &pr->pr_body, &pr->pr_resultType, allowDir);
    if (pr->pr_resultType == TYPE_NIL) {
	parseError(ms, "direct action cannot return nil (type unknown)");
    }
    ms->ParsingProc = P_NULL;
    return pr;
}

/*
 * parseImmediateProc - parse all the pieces of a new unnamed proc.
 */

Proc_t *
parseImmediateProc(MUDState_t *ms)
{
    Proc_t *pr;
    UINT_T localOffset, resType;

    ms->ErrorCount = 0;
    if (ms->tk_kind == tk_proc) {
	getToken(ms);
    } else {
	parseError(ms, "expecting 'proc' to start proc");
    }
    pr = newProcHeader(ms);

    pr->pr_parameters = getParameters(ms, &localOffset);

    pr->pr_resultType = parseType(ms);

    if (ms->tk_kind == tk_colon) {
	getToken(ms);
    } else {
	parseError(ms, "expecting ':' after proc header");
    }

    getLocals(ms, pr, localOffset);

    parseBody(ms, &pr->pr_body, &resType, B_TRUE);

    if (resType == TYPE_NIL && libNilType(pr->pr_resultType)) {
	resType = pr->pr_resultType;
    }
    if (resType != pr->pr_resultType) {
	parseError(ms, "proc result type does not match declared type");
    }

    if (ms->tk_kind != tk_corp) {
	parseError(ms, "expecting 'corp' to end proc definition");
    }

    ms->ParsingProc = P_NULL;

    return pr;
}

/*
 * makeRef - add a reference to a key to the current procs ref list.
 */

RefList_t *
makeRef(MUDState_t *ms, ULONG_T key)
{
    RefList_t **prl;
    RefList_t *rl;

    prl = &ms->ParsingProc->pr_refs;
    rl = *prl;
    while (rl != P_NULL) {
	if (rl->rl_key == key) {
	    return rl;
	}
	rl = rl->rl_next;
    }
    rl = MEM_ALLOC(ms, sizeof(RefList_t), "RL_T");
    rl->rl_next = *prl;
    rl->rl_name = P_NULL;
    rl->rl_key = key;
    *prl = rl;
    if ((key >> TYPE_SHIFT) != TYPE_ACTION ||
	((key & INDEX_MASK) < SMALLEST_BUILTIN && key != ms->ParsingProcKey))
    {
	libIncRef(ms, key);
    }
    return rl;
}

/*
 * localLookup - look up a name in the current proc's parms and vars.
 */

LocalList_t *
localLookup(MUDState_t *ms, CHAR_T *name)
{
    LocalList_t *ll;

    if (ms->ParsingProc != P_NULL) {
	ll = ms->ParsingProc->pr_parameters;
	while (ll != P_NULL) {
	    if (charsEqual(ll->ll_name, name)) {
		return ll;
	    }
	    ll = ll->ll_next;
	}
	ll = ms->ParsingProc->pr_locals;
	while (ll != P_NULL) {
	    if (charsEqual(ll->ll_name, name)) {
		return ll;
	    }
	    ll = ll->ll_next;
	}
    }
    return P_NULL;
}
