#include <ctype.h>

#include "MUD.h"
#include "types.h"
#include "MUDInternal.h"
#include "lex.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * lex.c - the lexical scanner.
 */

#define DEFAULT_BUFFER_SIZE	256

#define Eof_none	0
#define Eof_one 	(Eof_none + 1)
#define Eof_two 	(Eof_one + 1)

#define NOT_DIGIT	17

/*
 * reservedWord - internal routine to add a reserved word to the table.
 */

static void
reservedWord(MUDState_t *ms, CHAR_T *name, UINT_T tokenCode)
{

    symEnter(ms, &ms->ReservedWordTable, name, charsLen1(name))->en_key =
	tokenCode - tk_eof;
}

/*
 * lexInit - initial startup of the lexical scanner.
 */

void
lexInit(MUDState_t *ms)
{
    register void(*r_reservedWord)(MUDState_t *ms, CHAR_T *name,
				   UINT_T tokenCode);

    ms->Buffer = MEM_ALLOC(ms, DEFAULT_BUFFER_SIZE, "LXBF");
    ms->BufferSize = DEFAULT_BUFFER_SIZE;
    ms->NewLineSignificant = B_TRUE;

    symCreate(ms, &ms->ReservedWordTable, 6);
    r_reservedWord = reservedWord;
    r_reservedWord(ms, "and"	  , tk_and     );
    r_reservedWord(ms, "or"	  , tk_or      );
    r_reservedWord(ms, "not"	  , tk_not     );
    r_reservedWord(ms, "if"	  , tk_if      );
    r_reservedWord(ms, "then"	  , tk_then    );
    r_reservedWord(ms, "elif"	  , tk_elif    );
    r_reservedWord(ms, "else"	  , tk_else    );
    r_reservedWord(ms, "fi"	  , tk_fi      );
    r_reservedWord(ms, "while"	  , tk_while   );
    r_reservedWord(ms, "do"	  , tk_do      );
    r_reservedWord(ms, "od"	  , tk_od      );
    r_reservedWord(ms, "for"	  , tk_for     );
    r_reservedWord(ms, "from"	  , tk_from    );
    r_reservedWord(ms, "upto"	  , tk_upto    );
    r_reservedWord(ms, "case"	  , tk_case    );
    r_reservedWord(ms, "incase"   , tk_incase  );
    r_reservedWord(ms, "default"  , tk_default );
    r_reservedWord(ms, "esac"	  , tk_esac    );
    r_reservedWord(ms, "ignore"   , tk_ignore  );
    r_reservedWord(ms, "call"	  , tk_call    );
    r_reservedWord(ms, "note"	  , tk_note    );
    r_reservedWord(ms, "proc"	  , tk_proc    );
    r_reservedWord(ms, "utility"  , tk_utility );
    r_reservedWord(ms, "wizard"   , tk_wizard  );
    r_reservedWord(ms, "public"   , tk_public  );
    r_reservedWord(ms, "compiled" , tk_compiled);
    r_reservedWord(ms, "stripped" , tk_stripped);
    r_reservedWord(ms, "corp"	  , tk_corp    );
    r_reservedWord(ms, "void"	  , tk_void    );
    r_reservedWord(ms, "bool"	  , tk_bool    );
    r_reservedWord(ms, "int"	  , tk_int     );
    r_reservedWord(ms, "fixed"	  , tk_fixed   );
    r_reservedWord(ms, "string"   , tk_string  );
    r_reservedWord(ms, "thing"	  , tk_thing   );
    r_reservedWord(ms, "status"   , tk_status  );
    r_reservedWord(ms, "grammar"  , tk_grammar );
    r_reservedWord(ms, "character", tk_player  );
    r_reservedWord(ms, "table"	  , tk_table   );
    r_reservedWord(ms, "action"   , tk_action  );
    r_reservedWord(ms, "list"	  , tk_list    );
    r_reservedWord(ms, "property" , tk_property);
    r_reservedWord(ms, "true"	  , tk_true    );
    r_reservedWord(ms, "false"	  , tk_false   );
    r_reservedWord(ms, "succeed"  , tk_succeed );
    r_reservedWord(ms, "fail"	  , tk_fail    );
    r_reservedWord(ms, "continue" , tk_continue);
    r_reservedWord(ms, "nil"	  , tk_nil     );
    r_reservedWord(ms, "ts_private" , tk_ts_private );
    r_reservedWord(ms, "ts_readonly", tk_ts_readonly);
    r_reservedWord(ms, "ts_wizard"  , tk_ts_wizard  );
    r_reservedWord(ms, "ts_public"  , tk_ts_public  );
}

/*
 * lexStart - start the scanner actually operating (can be used more than
 *	once).
 */

void
lexStart(MUDState_t *ms, CHAR_T current, CHAR_T next)
{

    ms->EofState = Eof_none;
    ms->CurrentChar = current;
    ms->NextChar = next;
}

/*
 * lexTerm - free up any resources used by the lexical scanner.
 */

void
lexTerm(MUDState_t *ms)
{

    symDestroy(ms, &ms->ReservedWordTable);
    memFree(ms, ms->Buffer, ms->BufferSize);
}

/*
 * addChar - routine to add a character to the string/id buffer.
 */

static void
addChar(MUDState_t *ms, CHAR_T ch)
{
    ULONG_T newSize;
    CHAR_T *newBuffer;

    if (ms->BufferPos == ms->BufferSize) {
	newSize = ms->BufferSize * 2;
	newBuffer = MEM_ALLOC(ms, newSize, "LXBF");
	blockCopy(newBuffer, ms->Buffer, ms->BufferSize);
	memFree(ms, ms->Buffer, ms->BufferSize);
	ms->BufferSize = newSize;
	ms->Buffer = newBuffer;
    }
    *(ms->Buffer + ms->BufferPos) = ch;
    ms->BufferPos += 1;
}

/*
 * getNextChar - internal routine to get the next input character.
 */

static CHAR_T
getNextChar(MUDState_t *ms)
{

    ms->CurrentChar = ms->NextChar;
    if (ms->EofState == Eof_one) {
	ms->EofState = Eof_two;
	ms->NextChar = ' ';
    } else if (! getChar(ms, &ms->NextChar)) {
	ms->EofState = Eof_one;
	ms->NextChar = ' ';
    }
    return ms->CurrentChar;
}

/*
 * whiteSpace - routine to skip over whitespace between tokens.
 */

void
whiteSpace(MUDState_t *ms)
{
    UINT_T level, startLine, startCol;
    CHAR_T ch, nextCh;
    CHAR_T startFile[TEXT_COLUMNS + 1];

    level = 0;
    ch = ms->CurrentChar;
    nextCh = ms->NextChar;
    while (ms->EofState != Eof_two) {
	if (ch != ' ' &&
	    ch != '\t' &&
	    (ch != '\n' || ms->NewLineSignificant) &&
	    (ch != '/' || nextCh != '*') &&
	    (ch != '/' || nextCh != '/') &&
	    level == 0)
	{
	    break;
	}
	if (level == 0 && ch == '/' && nextCh == '/') {
	    while (ms->EofState != Eof_two && ch != '\n') {
		ch = getNextChar(ms);
	    }
	    nextCh = ms->NextChar;
	} else {
	    if (ch == '/' && nextCh == '*') {
		if (level == 0) {
		    getCurrentPos(ms, &startFile[0], &startLine, &startCol);
		}
		level += 1;
		(void) getNextChar(ms);
	    } else if (ch == '*' && nextCh == '/') {
		if (level == 0) {
		    parseError(ms, "unmatched closing comment - ignored");
		} else {
		    level -= 1;
		}
		(void) getNextChar(ms);
	    }
	    ch = getNextChar(ms);
	    nextCh = ms->NextChar;
	}
    }
    if (ms->EofState == Eof_two && level != 0) {
	outputString(ms, &startFile[0]);
	ms->ms_outputChar('(');
	outputDecimal(ms, startLine);
	ms->ms_outputChar(',');
	outputDecimal(ms, startCol);
	outputString(ms, ") *** unclosed comment\n");
    }
}

/*
 * getDigit - part of getting numbers.
 */

static UINT_T
getDigit(MUDState_t *ms, ULONG_T base, CHAR_T currentChar)
{
    ULONG_T digit;

    if (ms->EofState != Eof_two &&
	(isdigit(currentChar) || (base == 16 && isxdigit(currentChar))))
    {
	if (isdigit(currentChar)) {
	    digit = currentChar - '0';
	    if (digit >= base) {
		parseError(ms, "invalid digit for given base");
	    }
	} else if (islower(currentChar)) {
	    digit = currentChar - ('a' - 10);
	} else {
	    digit = currentChar - ('A' - 10);
	}
	return (UINT_T) digit;
    } else {
	return NOT_DIGIT;
    }
}

/*
 * getToken - the real thing - get the next input token.
 */

void
getToken(MUDState_t *ms)
{
    Entry_t *en;
    register CHAR_T (*r_getNextChar)(MUDState_t *ms);
    ULONG_T uns, base, digit, denom;
    UINT_T kind, bitCount, digitCount;
    CHAR_T currentChar, nextChar;
    BOOL_T hadOverflow;

    whiteSpace(ms);
    r_getNextChar = getNextChar;

    if (ms->EofState == Eof_two) {
	kind = tk_eof;
    } else if (ms->NewLineSignificant && ms->CurrentChar == '\n') {
	kind = tk_newline;
	(void) r_getNextChar(ms);
    } else {

	kind = tk_illegal;
	currentChar = ms->CurrentChar;
	nextChar = ms->NextChar;
	switch (currentChar) {

	/* Gack! Why don't C switch statements allow ranges? */

	case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
	case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
	case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
	case 's': case 't': case 'u': case 'v': case 'w': case 'x':
	case 'y': case 'z':
	case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
	case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
	case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
	case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
	case 'Y': case 'Z':
	case '_':
	case 0xc0: case 0xc1: case 0xc2: case 0xc3: case 0xc4: case 0xc5:
	case 0xc6: case 0xc7: case 0xc8: case 0xc9: case 0xca: case 0xcb:
	case 0xcc: case 0xcd: case 0xce: case 0xcf: case 0xd0: case 0xd1:
	case 0xd2: case 0xd3: case 0xd4: case 0xd5: case 0xd6:
	case 0xd8: case 0xd9: case 0xda: case 0xdb: case 0xdc: case 0xdd:
	case 0xde: case 0xdf: case 0xe0: case 0xe1: case 0xe2: case 0xe3:
	case 0xe4: case 0xe5: case 0xe6: case 0xe7: case 0xe8: case 0xe9:
	case 0xea: case 0xeb: case 0xec: case 0xed: case 0xee: case 0xef:
	case 0xf0: case 0xf1: case 0xf2: case 0xf3: case 0xf4: case 0xf5:
	case 0xf6:
	case 0xf8: case 0xf9: case 0xfa: case 0xfb: case 0xfc: case 0xfd:
	case 0xfe: case 0xff:
	    ms->BufferPos = 0;
	    while (
		(isalnum(currentChar) || currentChar == '_' ||
		 (currentChar >= 0xc0 + '\0' && currentChar <= 0xd6 + '\0') ||
		 (currentChar >= 0xd8 + '\0' && currentChar <= 0xf6 + '\0') ||
		 (currentChar >= 0xf8 + '\0' && currentChar <= 0xff + '\0')) &&
		ms->EofState != Eof_two)
	    {
		addChar(ms, currentChar);
		currentChar = r_getNextChar(ms);
	    }
	    addChar(ms, '\0');
	    en = symLookup(&ms->ReservedWordTable, ms->Buffer, ms->BufferPos);
	    if (en != P_NULL) {
		kind = en->en_key + tk_eof;
	    } else {
		kind = tk_identifier;
		ms->tk_stringValue = ms->Buffer;
		ms->tk_stringLength = ms->BufferPos;
	    }
	    break;

	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
	case '.':
	    base = 10;
	    if (currentChar == '0') {
		currentChar = r_getNextChar(ms);
		if (currentChar == 'x' || currentChar == 'X') {
		    base = 16;
		    currentChar = r_getNextChar(ms);
		} else if (currentChar == 'o' || currentChar == 'O') {
		    base = 8;
		    currentChar = r_getNextChar(ms);
		} else if (currentChar == 'b' || currentChar == 'B') {
		    base = 2;
		    currentChar = r_getNextChar(ms);
		}
	    }

	    hadOverflow = B_FALSE;
	    uns = 0;
	    while (B_TRUE) {
		digit = getDigit(ms, base, currentChar);
		if (digit == NOT_DIGIT) {
		    break;
		}
		if ((~0 - digit) / base >= uns) {
		    uns = uns * base + digit;
		} else {
		    hadOverflow = B_TRUE;
		    uns = ~0;
		}
		currentChar = r_getNextChar(ms);
	    }

	    if (currentChar == '.') {
		currentChar = r_getNextChar(ms);
		if (uns > (1 << FRAC_BITS) - 1) {
		    parseError(ms, "overflow in fixed point constant");
		    uns = ~0;
		} else {
		    uns <<= FRAC_BITS;
		}
		ms->tk_numericValue = uns;

		uns = 0;
		digitCount = 0;
		denom = 1;
		while (B_TRUE) {
		    digit = getDigit(ms, base, currentChar);
		    if (digit == NOT_DIGIT) {
			break;
		    }
		    /* What is going on here? The value 1000000 is large
		       enough that we get enough digits to round correctly,
		       but not large enough to risk any overflow. */
		    if (denom < 1000000) {
			denom *= base;
			uns = uns * base + digit;
			digitCount += 1;
		    }
		    currentChar = r_getNextChar(ms);
		}

		/* The following statement essentially adds
			1/2 * base ** digitCount
		   to 'uns'. This, followed by the loops below, does the
		   needed rounding of the 'uns'. */
		uns += (denom >> (FRAC_BITS + 1));

		/* Value is base ** digitCount too large, but also needs
		   to be shifted left FRAC_BITS, all without losing bits. */
		for (bitCount = 0; bitCount < FRAC_BITS; bitCount += 1) {
		    if (uns < (1 << (ULONG_BITS - 1))) {
			uns <<= 1;
		    } else {
			/* The bases are always even, so we divide and
			   left-shift in one step. (digitCount can't be
			   zero here.) */
			uns /= (base / 2);
			digitCount -= 1;
		    }
		}
		while (digitCount > 0) {
		    digitCount -= 1;
		    uns /= base;
		}

		ms->tk_numericValue = ms->tk_numericValue + uns;
		if (uns > (ULONG_T) ms->tk_numericValue) {
		    /* A rounding overflow has occurred - we ignore it. */
		    ms->tk_numericValue = ~0;
		}

		kind = tk_fixedConst;
	    } else {
		ms->tk_numericValue = uns;
		if (hadOverflow) {
		    parseError(ms, "overflow in integer constant");
		}
		kind = tk_intConst;
	    }
	    break;

	case '\"':
	    kind = tk_stringConst;
	    ms->BufferPos = 0;
	    while (B_TRUE) {
		currentChar = r_getNextChar(ms);
		while (currentChar != '\"' && currentChar != '\n' &&
		    ms->EofState != Eof_two)
		{
		    if (currentChar == '\\') {
			currentChar = r_getNextChar(ms);
			if (currentChar == 'n') {
			    currentChar = '\n';
			} else if (currentChar == 't') {
			    currentChar = '\t';
			}
		    }
		    addChar(ms, currentChar);
		    currentChar = r_getNextChar(ms);
		}
		if (currentChar == '\"') {
		    (void) r_getNextChar(ms);
		    whiteSpace(ms);
		    if (ms->CurrentChar != '\"') {
			break;
		    }
		} else {
		    parseError(ms, "missing closing \" in string constant\n");
		    break;
		}
	    }
	    addChar(ms, '\0');
	    ms->tk_stringValue = ms->Buffer;
	    ms->tk_stringLength = ms->BufferPos;
	    break;

	case '$':
	    kind = tk_dollar;
	    (void) r_getNextChar(ms);
	    break;
	case '=':
	    if (nextChar == '=') {
		kind = tk_equalEqual;
		(void) r_getNextChar(ms);
	    } else {
		kind = tk_equal;
	    }
	    (void) r_getNextChar(ms);
	    break;
	case '~':
	    if (nextChar == '=') {
		kind = tk_notEqual;
		(void) r_getNextChar(ms);
	    } else {
		kind = tk_tilda;
	    }
	    (void) r_getNextChar(ms);
	    break;
	case '<':
	    if (nextChar == '=') {
		kind = tk_lessOrEqual;
		(void) r_getNextChar(ms);
	    } else if (nextChar == '<') {
		kind = tk_shiftLeft;
		(void) r_getNextChar(ms);
	    } else {
		kind = tk_less;
	    }
	    (void) r_getNextChar(ms);
	    break;
	case '>':
	    if (nextChar == '=') {
		kind = tk_greaterOrEqual;
		(void) r_getNextChar(ms);
	    } else if (nextChar == '<') {
		kind = tk_xor;
		(void) r_getNextChar(ms);
	    } else if (nextChar == '>') {
		kind = tk_shiftRight;
		(void) r_getNextChar(ms);
	    } else {
		kind = tk_greater;
	    }
	    (void) r_getNextChar(ms);
	    break;
	case ':':
	    if (nextChar == '=') {
		kind = tk_assign;
		(void) r_getNextChar(ms);
	    } else {
		kind = tk_colon;
	    }
	    (void) r_getNextChar(ms);
	    break;
	case '+':
	    kind = tk_plus;
	    (void) r_getNextChar(ms);
	    break;
	case '-':
	    if (nextChar == '-') {
		kind = tk_minusMinus;
		(void) r_getNextChar(ms);
	    } else {
		kind = tk_minus;
	    }
	    (void) r_getNextChar(ms);
	    break;
	case '*':
	    kind = tk_star;
	    (void) r_getNextChar(ms);
	    break;
	case '/':
	    kind = tk_slash;
	    (void) r_getNextChar(ms);
	    break;
	case '%':
	    kind = tk_percent;
	    (void) r_getNextChar(ms);
	    break;
	case '&':
	    kind = tk_ampersand;
	    (void) r_getNextChar(ms);
	    break;
	case '|':
	    kind = tk_stick;
	    (void) r_getNextChar(ms);
	    break;
	case '(':
	    kind = tk_leftParen;
	    (void) r_getNextChar(ms);
	    break;
	case ')':
	    kind = tk_rightParen;
	    (void) r_getNextChar(ms);
	    break;
	case ',':
	    kind = tk_comma;
	    (void) r_getNextChar(ms);
	    break;
	case ';':
	    kind = tk_semicolon;
	    (void) r_getNextChar(ms);
	    break;
	case '@':
	    kind = tk_at;
	    (void) r_getNextChar(ms);
	    break;
	case '[':
	    kind = tk_leftSquare;
	    (void) r_getNextChar(ms);
	    break;
	case ']':
	    kind = tk_rightSquare;
	    (void) r_getNextChar(ms);
	    break;
	default:
	    (void) r_getNextChar(ms);
	    break;
	}
    }
    ms->tk_kind = kind;
}

/*
 * peekChar - allow looking at the next input character. Note that we make
 *	newlines significant here.
 */

CHAR_T
peekChar(MUDState_t *ms)
{
    BOOL_T saveNLS;

    saveNLS = ms->NewLineSignificant;
    ms->NewLineSignificant = B_TRUE;
    whiteSpace(ms);
    ms->NewLineSignificant = saveNLS;
    return ms->CurrentChar;
}

/*
 * setNewLineSignificant - make newlines be individual tokens.
 */

void
setNewLineSignificant(MUDState_t *ms)
{

    ms->NewLineSignificant = B_TRUE;
}

/*
 * clearNewLineSignificant - make newlines be whitespace.
 */

void
clearNewLineSignificant(MUDState_t *ms)
{

    ms->NewLineSignificant = B_FALSE;
}

/*
 * getNote - buffer characters up to the end of the line and leave them
 *	in the string buffer.
 */

void
getNote(MUDState_t *ms)
{
    CHAR_T ch;

    ms->BufferPos = 0;
    ch = ms->CurrentChar;
    while (ch != '\n' && ms->EofState != Eof_two) {
	addChar(ms, ch);
	ch = getNextChar(ms);
    }
    addChar(ms, '\0');
    ms->tk_stringValue = ms->Buffer;
    ms->tk_stringLength = ms->BufferPos;
}
