#include <fcntl.h>
#include <string.h>

#include "MUD.h"
#include "Request.h"
#include "Effects.h"
#include "types.h"
#include "lex.h"
#include "MUDInternal.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * main.c - startup and mainline code for any MUD client.
 */

#define INTERNAL_REQUESTS	B_TRUE

#define PROC_MAP_SIZE		256

static void crypt(CHAR_T *p, BOOL_T decrypt, ULONG_T sessionKey);

/* NOTE: change libInitE if this stops being the first function. */

/*
 * errorHeader - print a header for an error message.
 */

void
errorHeader(MUDState_t *ms)
{
    FileList_t *fl;
    UINT_T col;

    if (ms->EditReading) {
	col = ms->InputPos;
	/* after lexing a token, the current character is the one after the
	   token, the next character is the one after that, and InputPos will
	   point after that one, so subtract 3 to get to the last character
	   of the token just yielded */
	if (col >= 3) {
	    col -= 3;
	} else {
	    col = 0;
	}
	if (ms->ms_editErrorStart != P_NULL) {
	    ms->ms_editErrorStart(ms->EditLineNumber, col);
	} else {
	    ms->ms_outputChar('(');
	    outputDecimal(ms, ms->EditLineNumber);
	    ms->ms_outputChar(',');
	    outputDecimal(ms, col);
	    outputString(ms, ") *** ");
	}
    } else {
	fl = ms->FileList;
	if (fl != P_NULL) {
	    outputString(ms, &fl->fl_fileName[0]);
	    ms->ms_outputChar('(');
	    outputDecimal(ms, fl->fl_prevPrevLine);
	    ms->ms_outputChar(',');
	    outputDecimal(ms, fl->fl_prevPrevColumn);
	    outputString(ms, ") ");
	}
	outputString(ms, "*** ");
    }
}

/*
 * getCurrentPos - return the current line/column.
 */

void
getCurrentPos(MUDState_t *ms, CHAR_T *pFName, UINT_T *pLine, UINT_T *pCol)
{
    FileList_t *fl;
    UINT_T col;

    if (ms->EditReading) {
	*pFName = '\0';
	*pLine = ms->EditLineNumber;
	col = ms->InputPos;
	*pCol = (col >= 3) ? col - 3 : col;
    } else {
	fl = ms->FileList;
	if (fl != P_NULL) {
	    charsCopy(pFName, &fl->fl_fileName[0]);
	    *pLine = fl->fl_prevPrevLine;
	    *pCol = fl->fl_prevColumn;
	} else {
	    *pFName = '\0';
	    *pLine = 0;
	    *pCol = 0;
	}
    }
}

/*
 * errorMessage - body part of an error message.
 */

void
errorMessage(MUDState_t *ms, CHAR_T *m)
{

    if (ms->EditReading && ms->ms_editErrorStart != P_NULL) {
	ms->ms_editErrorString(m);
    } else {
	outputString(ms, m);
    }
}

/*
 * errorTrailer - end of an error message.
 */

void
errorTrailer(MUDState_t *ms)
{

    ms->ErrorCount += 1;
    if (ms->EditReading && ms->ms_editErrorStart != P_NULL) {
	ms->ms_editErrorEnd();
    }
}

/*
 * parseError - do the sequence for a simple string error message.
 */

void
parseError(MUDState_t *ms, CHAR_T *message)
{

    errorHeader(ms);
    errorMessage(ms, message);
    errorMessage(ms, "\n");
    errorTrailer(ms);
}

/*
 * libSymbolError - a parse-time error message involving a symbol.
 */

void
libSymbolError(MUDState_t *ms, CHAR_T *symbol, CHAR_T *message)
{

    errorHeader(ms);
    if (symbol != P_NULL) {
	errorMessage(ms, symbol);
	errorMessage(ms, ": ");
    }
    errorMessage(ms, message);
    errorMessage(ms, "\n");
    errorTrailer(ms);
}

/*
 * parseTerminate - shut down the parsing-related stuff.
 */

static void
parseTerminate(MUDState_t *ms, BOOL_T doMem)
{
    FileList_t *fl;
    ProcList_t *prl, *prlTemp;

    if (ms->ParseStarted) {
	ms->ParseStarted = B_FALSE;
	while (B_TRUE) {
	    fl = ms->FileList;
	    if (fl == P_NULL) {
		break;
	    }
	    ms->FileList = fl->fl_next;
	    ms->ms_Close(fl->fl_fd);
	    if (doMem) {
		memFree(ms, fl, sizeof(FileList_t));
	    }
	}
	if (doMem) {
	    memFree(ms, ms->ProcMap, sizeof(ProcList_t *) * PROC_MAP_SIZE);
	    prl = ms->ProcList;
	    while (prl != P_NULL) {
		freeProc(ms, prl->prl_proc, prl->prl_key, B_FALSE);
		prlTemp = prl;
		prl = prl->prl_next;
		memFree(ms, prlTemp, sizeof(ProcList_t));
	    }
	    lexTerm(ms);
	    memFree(ms, ms->WriteBuffer, ms->WriteBufferSize);
	}
    }
}

/*
 * freeRequests - free all of the requests we have
 */

void
freeRequests(MUDState_t *ms)
{
    Request_t *rq;

    while (B_TRUE) {
	rq = ms->FreeRequests;
	if (rq == P_NULL) {
	    break;
	}
	ms->FreeRequests = rq->rq_next;
#if INTERNAL_REQUESTS
	memFree(ms, rq, sizeof(Request_t) - TEXT_LENGTH + rq->rq_availLen);
#else
	ms->ms_FreeMem(rq, sizeof(Request_t) - TEXT_LENGTH + rq->rq_availLen);
#endif
    }
}

/*
 * memAbort - something wrong in memory routines - shutdown, but do not
 *	try to free internal memory, until we are done and then we (void)
 *	everything except our list of allocated large blocks.
 */

void
memAbort(MUDState_t *ms, BOOL_T isServerNuke)
{

    parseTerminate(ms, B_FALSE);
#if ! INTERNAL_REQUESTS
    freeRequests(ms);
#endif
#if 0
    /* Cannot do memFreeBlocks - that results in all of our memory being
       gone at this point if the system actually frees it. This has been
       seen on Linux. There is also no real point on Linux. Later parts
       of abort do things like 'messageFlush', scanning the proc cache
       looking for locked-in-memory ones, etc. Cannot do that without
       our memory present! Also, if we do this, and the memory goes
       away, it makes debugging a core file kind of tough! */
    memFreeBlocks(ms);
#endif
    ms->ms_abort(isServerNuke);
}

/*
 * libAbort - something wrong - do full shutdown and abort.
 */

void
libAbort(MUDState_t *ms, BOOL_T isServerNuke)
{

    if (ms->CacheSymbols) {
	symDestroy(ms, &ms->SymbolCache);
    }
    memAbort(ms, isServerNuke);
}

/*
 * allocRequest - allocate a request of the given size.
 */

static Request_t *
allocRequest(MUDState_t *ms, UINT_T originalLen)
{
    Request_t *rq;
    UINT_T len;

    len = originalLen;
    if (len < TEXT_COLUMNS) {
	len = TEXT_COLUMNS;
    } else {
	len = (len + 3) & (-4);
    }
    rq = ms->FreeRequests;
    while (rq != P_NULL && rq->rq_availLen < len) {
	rq = rq->rq_next;
    }
    if (rq == P_NULL) {
#if INTERNAL_REQUESTS
	rq = MEM_ALLOC(ms, sizeof(Request_t) - TEXT_LENGTH + len, "RQST");
#else
	rq = ms->ms_AllocMem(sizeof(Request_t) - TEXT_LENGTH + len);
	if (rq == P_NULL) {
	    errorString(ms, "no memory for server request\n");
	    libAbort(ms, B_FALSE);
	}
#endif
	rq->rq_availLen = len;
    } else {
	if (rq->rq_prev == P_NULL) {
	    ms->FreeRequests = rq->rq_next;
	} else {
	    rq->rq_prev->rq_next = rq->rq_next;
	}
	if (rq->rq_next != P_NULL) {
	    rq->rq_next->rq_prev = rq->rq_prev;
	}
	ms->RequestCount -= 1;
    }

    /* All of this has to be done whether we had to allocate a new request or
       not, since one we freed may have come from the server, and as such it
       will not have these valid for us. */

    rq->rq_usedLen = originalLen;
    return rq;
}

/*
 * allocRequestE - external interface to allocRequest.
 */

Request_t *
allocRequestE(MUDState_t *ms, ULONG_T len)
{

    return allocRequest(ms, len);
}

/*
 * freeRequest - free the given request.
 */

static void
freeRequest(MUDState_t *ms, Request_t *rq)
{
    Request_t *prev, *next;
    UINT_T minLen, len;

    rq->rq_next = ms->FreeRequests;
    if (ms->FreeRequests != P_NULL) {
	ms->FreeRequests->rq_prev = rq;
    }
    rq->rq_prev = P_NULL;
    ms->FreeRequests = rq;
    ms->RequestCount += 1;

    /* this is a while loop because of the cheat we do in libFindName */
    while (ms->RequestCount > ms->RequestLimit) {
	/* find the smallest request and free it */
	ms->RequestCount -= 1;
	minLen = 0xffff;
	prev = ms->FreeRequests;
	while (prev != P_NULL) {
	    len = prev->rq_availLen;
	    if (len < minLen) {
		minLen = len;
		rq = prev;
	    }
	    prev = prev->rq_next;
	}
	prev = rq->rq_prev;
	next = rq->rq_next;
	if (prev == P_NULL) {
	    ms->FreeRequests = next;
	} else {
	    prev->rq_next = next;
	}
	if (next != P_NULL) {
	    next->rq_prev = prev;
	}
#if INTERNAL_REQUESTS
	memFree(ms, rq, sizeof(Request_t) - TEXT_LENGTH + minLen);
#else
	ms->ms_FreeMem(rq, sizeof(Request_t) - TEXT_LENGTH + minLen);
#endif
    }
}

/*
 * freeRequestE - external interface to freeRequest.
 */

void
freeRequestE(MUDState_t *ms, Request_t *rq)
{

    freeRequest(ms, rq);
}

/*
 * requestInitE - initialize for just request maintenance.
 */

void
requestInitE(MUDState_t *ms, ULONG_T rqMax)
{

    ms->FreeRequests = P_NULL;
    ms->RequestLimit = rqMax;
    ms->RequestCount = 0;
    /* have to do enough setup so that 'libAbort' will work OK */
    ms->CacheSymbols = B_FALSE;
    ms->ParseStarted = B_FALSE;
    ms->MemBlocks = P_NULL;
}

/*
 * requestTermE - shut down for a request-only run.
 */

void
requestTermE(MUDState_t *ms)
{

    freeRequests(ms);
}

/*
 * deleteProcFromList - delete the given proc from the local list, if there.
 */

void
deleteProcFromList(MUDState_t *ms, ULONG_T key)
{
    ProcList_t **pprl;
    ProcList_t *prl;

    pprl = &ms->ProcList;
    while (B_TRUE) {
	prl = *pprl;
	if (prl == P_NULL) {
	    break;
	}
	if (prl->prl_key == key) {
	    freeProc(ms, prl->prl_proc, key, B_FALSE);
	    *pprl = prl->prl_next;
	    memFree(ms, prl, sizeof(ProcList_t));
	    return;
	} else {
	    pprl = &prl->prl_next;
	}
    }
}

/*
 * resetState - handy utility for serverMessage and inputLine.
 */

static void resetState(MUDState_t *ms)
{

    ms->State = ms->SaveState;
    ms->ms_setEcho(B_TRUE);
    /* We may get the proper prompt from the server, but do this anyway. */
    if (ms->State == st_normal) {
	ms->ms_setPrompt("");
    } else {
	ms->ms_setPrompt("> ");
    }
}

/*
 * serverMessage - handle an asynchronous message from the server.
 */

void
serverMessage(MUDState_t *ms, Request_t *rq)
{
    RequestType_t rtGot;
    CHAR_T *stringResult;
    int fd;
    UINT_T promptLen;

    rtGot = rq->rq_type;
    switch (rtGot) {
    case rt_endClient:
	ms->State = st_exiting;
	ms->ms_setPrompt("Type a return to exit: ");
	break;
    case rt_killClient:
	ms->WantQuit = B_TRUE;
	break;
    case rt_nukeClient:
	errorString(ms, "Client being nuked by server!\n");
	libAbort(ms, B_TRUE);
	break;
    case rt_runClient:
	ms->GoAhead = B_TRUE;
	break;
    case rt_message:
	/* an asynchronous line of output text */
	ms->ms_putText(&rq->rq_u.ru_text[0], rq->rq_usedLen);
	break;
    case rt_enterWizardMode:
	ms->State = st_parsing;
	ms->PlayerKey = rq->rq_key;
	ms->ms_setPrompt("> ");
	/* we wait until we get back to the main loop before we actually
	   start lexing/parsing - this saves some stack space */
	break;
    case rt_leaveWizardMode:
	ms->State = st_normal;
	ms->ms_setPrompt("");
	break;
    case rt_setContinue:
	if (rq->rq_key == BOOLEAN_TRUE) {
	    ms->DefineOnError = B_TRUE;
	} else {
	    ms->DefineOnError = B_FALSE;
	}
	break;
    case rt_changePassword:
	switch (rq->rq_key) {
	case 0:
	    /* Initial request - want old password. */
	    ms->SaveState = ms->State;
	    ms->State = st_changePassword1;
	    ms->ms_setEcho(B_FALSE);
	    if (rq->rq_flag) {
		ms->ms_setPrompt("Enter old character password: ");
	    } else {
		ms->ms_setPrompt("Enter old creation password: ");
	    }
	    break;
	case 1:
	    /* Old password was correct - enter new one. */
	    ms->State = st_changePassword2;
	    ms->ms_setEcho(B_FALSE);
	    if (rq->rq_flag) {
		ms->ms_setPrompt("Enter new character password: ");
	    } else {
		ms->ms_setPrompt("Enter new creation password: ");
	    }
	    break;
	case 2:
	    /* Old password was incorrect - go back to normal mode. */
	    outputString(ms, "Password incorrect. No change done.\n");
	    resetState(ms);
	    break;
	}
	break;
    case rt_setPrompt:
	ms->ms_setPrompt(&rq->rq_u.ru_text[0]);
	break;
    case rt_deleteProc:
	deleteProcFromList(ms, rq->rq_key);
	if (rq->rq_key == ms->EditProc) {
	    ms->EditProc = NIL_ACTION;
	}
	break;
    case rt_effects:
	ms->ms_doEffects(&rq->rq_u.ru_bytes[0], rq->rq_usedLen);
	break;
    case rt_editString:
	promptLen = charsLen1(&rq->rq_u.ru_text[0]);
	if (! ms->ms_doEdit(&rq->rq_u.ru_text[promptLen],
			    rq->rq_usedLen - promptLen - 1, rq->rq_key,
			    B_FALSE, B_TRUE, B_TRUE, &rq->rq_u.ru_text[0]))
	{
	    /* something wrong in editing - just send back a fail */
	    rq->rq_usedLen = 1;
	    rq->rq_u.ru_text[0] = '\0';
	    rq->rq_type = rt_editStringDone;
	    rq->rq_flag = B_FALSE;
	    ms->ms_serverSend(rq);
	    return;
	}
	break;
    case rt_editProc:
	ms->EditProc = rq->rq_key;
	if (! ms->ms_doEdit(&rq->rq_u.ru_text[0], rq->rq_usedLen, EDIT_CODE,
			    B_TRUE, rq->rq_flag, B_TRUE, P_NULL))
	{
	    ms->EditProc = NIL_ACTION;
	    rq->rq_type = rt_editProcDone;
	    ms->ms_serverSend(rq);
	    return;
	}
	break;
    case rt_getString:
	promptLen = charsLen1(&rq->rq_u.ru_text[0]);
	if (! ms->ms_getString(&rq->rq_u.ru_text[0],
			       &rq->rq_u.ru_text[promptLen], &stringResult))
	{
	    /* user cancelled or other problem */
	    rq->rq_usedLen = 1;
	    rq->rq_u.ru_text[0] = '\0';
	    rq->rq_flag = B_FALSE;
	} else {
	    /* all is well */
	    freeRequest(ms, rq);
	    rq = allocRequest(ms, charsLen1(stringResult));
	    charsCopy(&rq->rq_u.ru_text[0], stringResult);
	    rq->rq_flag = B_TRUE;
	}
	rq->rq_type = rt_getStringDone;
	ms->ms_serverSend(rq);
	return;
    case rt_flushSymbol:
	if (ms->CacheSymbols) {
	    symKeyDelete(ms, &ms->SymbolCache, rq->rq_key);
	}
	break;
    case rt_defineEffect:
	if (ms->ms_defineEffect != P_NULL) {
	    ms->ms_defineEffect(rq->rq_key, &rq->rq_u.ru_bytes[0],
				rq->rq_usedLen);
	}
	break;
    case rt_queryFile:
	fd = ms->ms_Open(&rq->rq_u.ru_text[0], O_RDONLY, 0);
	rq->rq_type = rt_queryFileDone;
	rq->rq_usedLen = 0;
	if (fd == -1) {
	    rq->rq_flag = B_FALSE;
	} else {
	    ms->ms_Close(fd);
	    rq->rq_flag = B_TRUE;
	}
	ms->ms_serverSend(rq);
	return;
    case rt_flushEffect:
	if (ms->ms_flushEffect != P_NULL) {
	    ms->ms_flushEffect(rq->rq_key);
	}
	break;
    default:
	/* shouldn't get anything else */
	errorString(ms, "serverMessage: got unexpected type ");
	errorDecimal(ms, rtGot - rt_first);
	errorString(ms, " from server\n");
	libAbort(ms, B_FALSE);
	break;
    }
    freeRequest(ms, rq);
}

static void
zapPW(CHAR_T *p)
{
    UINT_T i;

    for (i = 0; i < PASSWORD_LEN; i += 1) {
	*p++ = i + 'Q';
    }
}

/*
 * genericRequest - common code for a simple, short request that does not
 *	need a reply.
 */

static void
genericRequest(MUDState_t *ms, RequestType_t rt, ULONG_T key)
{
    Request_t *rq;

    rq = allocRequest(ms, 0);
    rq->rq_type = rt;
    rq->rq_key = key;
    ms->ms_serverSend(rq);
}

/*
 * log - send a string to the server to log.
 */

static void
log(MUDState_t *ms, CHAR_T *message)
{
    Request_t *rq;
    UINT_T len;

    len = charsLen1(message);
    rq = allocRequest(ms, len);
    charsCopy(&rq->rq_u.ru_text[0], message);
    rq->rq_type = rt_log;
    ms->ms_serverSend(rq);
}

/*
 * logPlayer - handy little routine.
 */

static void
logPlayer(MUDState_t *ms, CHAR_T *m1, CHAR_T *m2)
{
    CHAR_T buffer[PLAYER_NAME_LEN + 80];
    UINT_T len;

    charsCopy(&buffer[0], m1);
    len = charsLen1(m1) - 1;
    charsCopy(&buffer[len], &ms->PlayerName[0]);
    len += charsLen1(&ms->PlayerName[0]) - 1;
    charsCopy(&buffer[len], m2);
    log(ms, &buffer[0]);
}

/*
 * outputPlayer - another handy one.
 */

static void
outputPlayer(MUDState_t *ms, CHAR_T *m1, CHAR_T *m2)
{

    if (m1 != P_NULL) {
	outputString(ms, m1);
    }
    ms->ms_outputChar('\"');
    outputString(ms, &ms->PlayerName[0]);
    ms->ms_outputChar('\"');
    outputString(ms, m2);
}

/*
 * startPlayer - all OK, start the player running.
 */

static void
startPlayer(MUDState_t *ms)
{

    zapPW(&ms->Password[0]);
    logPlayer(ms, "Character ", " enters the game\n");
    ms->State = st_normal;
    ms->ms_setPrompt("");
    ms->ms_setEcho(B_TRUE);
}

/*
 * newPlayer - we are creating a new player.
 */

static void
newPlayer(MUDState_t *ms)
{

    ms->ms_setEcho(B_FALSE);
    ms->State = st_npPassword1;
    ms->ms_setPrompt("Enter password for new character: ");
}

/*
 * setExiting - common code used hither and yon.
 */

static void
setExiting(MUDState_t *ms)
{
    ms->State = st_exiting;
    ms->ms_setPrompt("Type a return to exit: ");
}

/*
 * requestPlayer - send a request for the connection of the given player.
 */

static void
requestPlayer(MUDState_t *ms, CHAR_T *name, UINT_T len)
{
    Request_t *rq;

    rq = allocRequest(ms, len + 1);
    rq->rq_type = rt_askPlayer;
    blockCopy(&rq->rq_u.ru_text[0], name, len);
    rq->rq_u.ru_text[len] = '\0';
    charsCopy(&ms->PlayerName[0], &rq->rq_u.ru_text[0]);
    rq->rq_key = ms->IsRemote ? 1 : 0;
    rq = ms->ms_serverRequest(rq);
    switch (rq->rq_key) {
    case 0:
	/* no such player and automatic creation disallowed */
	outputPlayer(ms, "Character ", " does not exist.\n");
	if (ms->nameCount == NAME_TRIES) {
	    setExiting(ms);
	    logPlayer(ms, "Character ", " creation disallowed\n");
	} else {
	    ms->nameCount += 1;
	}
	break;
    case 1:
	/* no such player, but it is ok to create a new one, and there
	   is no creation password */
	ms->State = st_newPlayerNCP;
	ms->ynCount = 1;
	outputPlayer(ms, "Character ", " does not exist.\n");
	ms->ms_setPrompt("Do you wish to create it (N/y)? ");
	break;
    case 2:
	/* no such player, but it is ok to create a new one, but there
	   is a creation password */
	ms->State = st_newPlayerWCP;
	ms->ynCount = 1;
	outputPlayer(ms, "Character ", " does not exist.\n");
	ms->ms_setPrompt("Do you wish to create it (N/y)? ");
	break;
    case 3:
	/* the given name is not a player in the character table */
	outputPlayer(ms, P_NULL, " is not a character.\n");
	logPlayer(ms, "Non-character ", " used as name\n");
	if (ms->nameCount == NAME_TRIES) {
	    setExiting(ms);
	} else {
	    ms->nameCount += 1;
	}
	break;
    case 4:
	/* remote login of SysAdmin denied */
	outputString(ms, "Remote login of SysAdmin denied.\n");
	setExiting(ms);
	log(ms, "Remote login of SysAdmin denied.\n");
	break;
    case 5:
	/* it is a player, now get the password */
	ms->State = st_plPassword;
	ms->ms_setPrompt("Enter password: ");
	ms->ms_setEcho(B_FALSE);
	break;
    case 6:
	/* player is already connected */
	outputPlayer(ms, P_NULL, " is already connected.\n");
	logPlayer(ms, "Attempt to reconnect connected character ", "\n");
	if (ms->nameCount == NAME_TRIES) {
	    setExiting(ms);
	} else {
	    ms->nameCount += 1;
	}
	break;
    }
    freeRequest(ms, rq);
}

/*
 * nameRetry - ask for player name again?
 */

static void
nameRetry(MUDState_t *ms)
{

    if (ms->nameCount == NAME_TRIES) {
	setExiting(ms);
    } else {
	ms->nameCount += 1;
    }
}

/*
 * nameAgain - check number of name entry tries, and perhaps allow retry.
 */

static void
nameAgain(MUDState_t *ms, BOOL_T needPrompt)
{

    if (ms->nameCount == NAME_TRIES) {
	setExiting(ms);
    } else {
	ms->State = st_name;
	ms->nameCount += 1;
	if (needPrompt) {
	    ms->ms_setPrompt("Enter character name: ");
	}
    }
}

/*
 * inputLine - process an input line from the player. The line is not
 *	terminated with anything in particular.
 */

void
inputLine(MUDState_t *ms, CHAR_T *p, ULONG_T len)
{
    Request_t *rq;
    CHAR_T *q;
    CHAR_T ch;

    if (ms->WantQuit) {
	/* we are trying to exit - ignore any input line */
	return;
    }

    switch (ms->State) {
    case st_name:
	if (len >= PLAYER_NAME_LEN) {
	    len = PLAYER_NAME_LEN - 1;
	}
	blockCopy(&ms->PlayerName[0], p, len);
	ms->PlayerName[len] = '\0';
	q = &ms->PlayerName[0];
	do {
	    ch = *q++;
	} while (ch != '\0' && ch != ' ' && ch != '\t' &&
		 ch != '.' && ch != ';' && ch != ',' && ch != '\"');
	if (len == 0) {
	    ms->WantQuit = B_TRUE;
	} else if (ch != '\0') {
	    outputString(ms,
		"Character names cannot contain spaces or punctuation.\n");
	    logPlayer(ms, "Character name '",
		      "' contained space or punctuation\n");
	    nameRetry(ms);
	} else {
	    requestPlayer(ms, p, len);
	}
	break;
    case st_plPassword:
	rq = allocRequest(ms, PASSWORD_LEN);
	rq->rq_type = rt_loginPlayer;
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	blockCopy(&rq->rq_u.ru_text[0], p, len);
	rq->rq_u.ru_text[len] = '\0';
	crypt(&rq->rq_u.ru_text[0], B_FALSE, ms->SessionKey);
	rq = ms->ms_serverRequest(rq);
	switch (rq->rq_key) {
	case 0:
	    /* Password OK. Start playing. */
	    startPlayer(ms);
	    break;
	case 1:
	    /* pw wrong, try again. */
	    outputString(ms, "Password incorrect. Try again.\n");
	    break;
	case 2:
	    /* pw wrong, run denied */
	    outputString(ms, "Password incorrect.\n");
	    ms->ms_setEcho(B_TRUE);
	    nameAgain(ms, B_TRUE);
	    break;
	}
	freeRequest(ms, rq);
	break;
    case st_newPlayerNCP:
    case st_newPlayerWCP:
	while (len != 0 && *p == ' ') {
	    len -= 1;
	    ++p;
	}
	if (len != 0 && (*p == 'y' || *p == 'Y')) {
	    if (ms->State == st_newPlayerNCP) {
		/* no creation password needed */
		newPlayer(ms);
	    } else {
		ms->State = st_crPassword;
		ms->ms_setPrompt("Enter character creation password: ");
		ms->ms_setEcho(B_FALSE);
	    }
	} else if (len == 0 || *p == 'n' || *p == 'N') {
	    logPlayer(ms, "New character ", " creation declined\n");
	    nameAgain(ms, B_TRUE);
	} else {
	    if (ms->ynCount == YN_TRIES) {
		setExiting(ms);
	    } else {
		/* bump up by one */
		ms->ynCount += 1;
	    }
	}
	break;
    case st_crPassword:
	rq = allocRequest(ms, PASSWORD_LEN);
	rq->rq_type = rt_creationCheck;
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	blockCopy(&rq->rq_u.ru_text[0], p, len);
	rq->rq_u.ru_text[len] = '\0';
	crypt(&rq->rq_u.ru_text[0], B_FALSE, ms->SessionKey);
	rq = ms->ms_serverRequest(rq);
	switch (rq->rq_key) {
	case 0:
	    /* Password OK. Start playing. */
	    newPlayer(ms);
	    break;
	case 1:
	    /* pw wrong, try again. */
	    outputString(ms, "Creation password incorrect. Try again.\n");
	    break;
	case 2:
	    /* pw wrong, run denied */
	    outputString(ms, "Creation password incorrect.\n");
	    ms->ms_setEcho(B_TRUE);
	    nameAgain(ms, B_TRUE);
	    break;
	}
	freeRequest(ms, rq);
	break;
    case st_npPassword1:
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	} else {
	    ms->Password[len] = '\0';
	}
	blockCopy(&ms->Password[0], p, len);
	crypt(&ms->Password[0], B_FALSE, ms->SessionKey);
	ms->ms_setPrompt("Re-enter password for verification: ");
	ms->State = st_npPassword2;
	break;
    case st_npPassword2:
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	*(p + len) = '\0';
	q = &ms->Password[0];
	crypt(q, B_TRUE, ms->SessionKey);
	if (! charsEqual(p, q)) {
	    memset(&ms->Password[0], 0, PASSWORD_LEN);
	    outputString(ms, "Password not verified.\n");
	    logPlayer(ms, "New character ", " password not verified\n");
	    ms->ms_setEcho(B_TRUE);
	    nameAgain(ms, B_TRUE);
	} else {
	    rq = allocRequest(ms, PASSWORD_LEN + PLAYER_NAME_LEN);
	    crypt(&ms->Password[0], B_FALSE, ms->SessionKey);
	    blockCopy(&rq->rq_u.ru_text[0], &ms->Password[0], PASSWORD_LEN);
	    memset(&ms->Password[0], 0, PASSWORD_LEN);
	    charsCopy(&rq->rq_u.ru_text[PASSWORD_LEN], &ms->PlayerName[0]);
	    rq->rq_type = rt_createPlayer;
	    rq = ms->ms_serverRequest(rq);
	    if (rq->rq_key == 0) {
		/* hmmm. Someone got there before us! */
		setExiting(ms);
		outputString(ms, "Sorry! Someone just created it!\n");
		logPlayer(ms,
		    "New character ", " creation failed - duplicate\n");
	    } else {
		outputPlayer(ms, "New character ", " created.\n");
		logPlayer(ms, "New character ", " created\n");
		startPlayer(ms);
	    }
	    freeRequest(ms, rq);
	}
	break;
    case st_parsing:
	if (ms->FileList == P_NULL) {
	    ms->EditLineNumber += 1;
	    /* when reading from files, this is done in 'getChar' */
	    if (len != 0) {
		if (len >= INPUT_BUFFER_SIZE) {
		    len = INPUT_BUFFER_SIZE - 1;
		}
		blockCopy(&ms->InputLine[0], p, len);
		ms->InputLine[len] = '\n';
		len += 1;
		ms->InputTotal = len;
		ms->InputPos = 0;
	    }
	}
	break;
    case st_normal:
	/* no other check on 'len' here - should be OK, even in server */
	if (len == 0xffff) {
	    len = 0xfffe;
	}
	rq = allocRequest(ms, len + 1);
	rq->rq_type = rt_inputLine;
	blockCopy(&rq->rq_u.ru_text[0], p, len);
	rq->rq_u.ru_text[len] = '\0';
	if (ms->InputSync) {
	    freeRequest(ms, ms->ms_serverRequest(rq));
	} else {
	    ms->ms_serverSend(rq);
	}
	break;
    case st_exiting:
	ms->WantQuit = B_TRUE;
	break;
    case st_changePassword1:
	/* The old password */
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	rq = allocRequest(ms, PASSWORD_LEN);
	rq->rq_type = rt_changePassword;
	blockCopy(&rq->rq_u.ru_text[0], p, len);
	rq->rq_u.ru_text[len] = '\0';
	crypt(&rq->rq_u.ru_text[0], B_FALSE, ms->SessionKey);
	ms->ms_serverSend(rq);
	resetState(ms);
	break;
    case st_changePassword2:
	/* The first copy of the new password */
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	blockCopy(&ms->Password[0], p, len);
	ms->Password[len] = '\0';
	crypt(&ms->Password[0], B_FALSE, ms->SessionKey);
	ms->ms_setPrompt("Re-enter password for verification: ");
	ms->State = st_changePassword3;
	break;
    case st_changePassword3:
	/* The second copy of the new password */
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	*(p + len) = '\0';
	resetState(ms);
	crypt(&ms->Password[0], B_TRUE, ms->SessionKey);
	if (charsEqual(p, &ms->Password[0])) {
	    rq = allocRequest(ms, PASSWORD_LEN);
	    crypt(&ms->Password[0], B_FALSE, ms->SessionKey);
	    blockCopy(&rq->rq_u.ru_text[0], &ms->Password[0], PASSWORD_LEN);
	    outputString(ms, "Password changed.\n");
	} else {
	    outputString(ms, "Verification failed. Password not changed.\n");
	    rq = allocRequest(ms, 0);
	}
	rq->rq_type = rt_changePassword;
	ms->ms_serverSend(rq);
	memset(&ms->Password[0], 0, PASSWORD_LEN);
	break;
    }
}

/*
 * syncControl - control whether there are replies to rt_inputLine messages,
 *	thus making them synchronous instead of the usual asynchronous.
 */

void
syncControl(MUDState_t *ms, ULONG_T onOff)
{

    ms->InputSync = (BOOL_T) (onOff != 0);
    genericRequest(ms, rt_setSync, onOff);
}

/*
 * rawKeyEvent - send a raw key event to the server.
 */

void
rawKeyEvent(MUDState_t *ms, ULONG_T key)
{

    if (ms->State == st_normal || ms->State == st_parsing) {
	genericRequest(ms, rt_rawKey, key);
    }
}

/*
 * mouseUpEvent - send a mouse up event to the server.
 */

void
mouseUpEvent(MUDState_t *ms, ULONG_T n, ULONG_T position)
{
    Request_t *rq;

    if (ms->State == st_normal || ms->State == st_parsing) {
	rq = allocRequest(ms, sizeof(ULONG_T));
	rq->rq_type = rt_regionSelect;
	rq->rq_key = n;
	rq->rq_u.ru_otherKey = position;
	ms->ms_serverSend(rq);
    }
}

/*
 * buttonEvent - send a button event to the server.
 */

void
buttonEvent(MUDState_t *ms, ULONG_T whichButton)
{

    if (ms->State == st_normal || ms->State == st_parsing) {
	genericRequest(ms, rt_buttonHit, whichButton);
    }
}

/*
 * releaseEffect - send a flushEffect request to the server.
 */

void
releaseEffect(MUDState_t *ms, ULONG_T key)
{

    if (ms->State == st_normal || ms->State == st_parsing) {
	genericRequest(ms, rt_flushEffect, key);
    }
}

/*
 * textResize - sends a textResize request to the server.
 */

void
textResize(MUDState_t *ms, ULONG_T newRows, ULONG_T newCols)
{

    genericRequest(ms, rt_textResize, (newRows << 16) | newCols);
}

/*
 * parseActionFromString - library entry point to allow the server to parse
 *	a string into an anonymous action.
 */

ULONG_T
parseActionFromString(MUDState_t *ms, BYTE_T *p, ULONG_T owner)
{
    Proc_t *pr;
    UINT_T len;

    ms->InputPtr = p;
    ms->NewLineState = nl_clear;
    ms->FileList = P_NULL;
    ms->HadEof = B_FALSE;
    ms->ProcList = P_NULL;
    ms->PlayerKey = owner;
    lexStart(ms, ' ', ' ');
    getToken(ms);
    pr = parseImmediateAction(ms, B_TRUE);
    if (ms->ErrorCount == 0) {
	writeProc(ms, pr);
	freeProc(ms, pr, NIL_ACTION, B_FALSE);
	writeResult(ms, &p, &len);
	return ms->ms_writeProcNew(p, len);
    } else {
	freeProc(ms, pr, NIL_ACTION, B_TRUE);
	return NIL_ACTION;
    }
}

/*
 * parseProcFromString - library entry point to allow the server to parse
 *	a string into an anonymous action. This differs from the previous
 *	routine in that the action parsed must start with 'proc', end
 *	with 'corp', include a parameter list, and may have local variables.
 */

ULONG_T
parseProcFromString(MUDState_t *ms, BYTE_T *p, ULONG_T owner)
{
    Proc_t *pr;
    UINT_T len;

    ms->InputPtr = p;
    ms->NewLineState = nl_clear;
    ms->FileList = P_NULL;
    ms->HadEof = B_FALSE;
    ms->ProcList = P_NULL;
    ms->PlayerKey = owner;
    lexStart(ms, ' ', ' ');
    getToken(ms);
    pr = parseImmediateProc(ms);
    if (ms->ErrorCount == 0) {
	writeProc(ms, pr);
	freeProc(ms, pr, NIL_ACTION, B_FALSE);
	writeResult(ms, &p, &len);
	return ms->ms_writeProcNew(p, len);
    } else {
	freeProc(ms, pr, NIL_ACTION, B_TRUE);
	return NIL_ACTION;
    }
}

/*
 * library entry points to allow toggling of effects states.
 */

void
graphicsFlip(MUDState_t *ms, ULONG_T onOff)
{

    genericRequest(ms, rt_graphicsFlip, onOff);
}

void
voiceFlip(MUDState_t *ms, ULONG_T onOff)
{

    genericRequest(ms, rt_voiceFlip, onOff);
}

void
soundFlip(MUDState_t *ms, ULONG_T onOff)
{

    genericRequest(ms, rt_soundFlip, onOff);
}

void
musicFlip(MUDState_t *ms, ULONG_T onOff)
{

    genericRequest(ms, rt_musicFlip, onOff);
}

void
effectDone(MUDState_t *ms, ULONG_T kind, ULONG_T id)
{
    Request_t *rq;

    rq = allocRequest(ms, 0);
    rq->rq_type = rt_effectDone;
    rq->rq_key = kind;
    rq->rq_u.ru_otherKey = id;
    ms->ms_serverSend(rq);
}

void
editStringDone(MUDState_t *ms, CHAR_T *buf, ULONG_T len, ULONG_T ok)
{
    Request_t *rq;

    rq = allocRequest(ms, len + 1);
    rq->rq_type = rt_editStringDone;
    rq->rq_flag = (BOOL_T) (ok != BOOLEAN_FALSE);
    blockCopy(&rq->rq_u.ru_text[0], buf, len);
    rq->rq_u.ru_text[len] = '\0';
    ms->ms_serverSend(rq);
}

void
editProcStart(MUDState_t *ms)
{
    CHAR_T *name;

    name = libFindName(ms, ms->EditProc);
    if (name != P_NULL) {
	if (*name == '\0') {
	    name = P_NULL;
	} else {
	    name = charsDup(ms, name);
	}
    }
    ms->EditReadEof = B_FALSE;
    ms->EditReading = B_TRUE;
    ms->ErrorCount = 0;
    ms->SaveState = ms->State;
    ms->State = st_parsing;
    ms->EditLineNumber = 0;
    lexStart(ms, ' ', ' ');
    clearNewLineSignificant(ms);
    getToken(ms);
    if (ms->tk_kind != tk_proc) {
	parseError(ms, "expecting 'proc' in replacement text");
    }
    parseProc(ms, NIL_TABLE, ms->EditProc, name, B_FALSE, B_TRUE);
    if (name != P_NULL) {
	charsFree(ms, name);
    }
    ms->State = ms->SaveState;
    ms->EditReading = B_FALSE;
    lexStart(ms, '\n', ' ');
    setNewLineSignificant(ms);
    /* If there were no errors, the client will go out of edit mode, but if
       there were errors, it will stay in edit mode. The player can get out
       of edit mode with ^C, which will call 'editProcCancel', below. In the
       case of an external editor, editProcCancel is called unconditionally
       after the last line is parsed. */
    if (ms->ErrorCount == 0) {
	genericRequest(ms, rt_editProcDone, 0);
    }
}

void
editProcDone(MUDState_t *ms)
{

    ms->EditReadEof = B_TRUE;
}

void
editProcCancel(MUDState_t *ms)
{

    genericRequest(ms, rt_editProcDone, 0);
}

void
getStringDone(MUDState_t *ms, CHAR_T *buf, ULONG_T len, ULONG_T ok)
{
    Request_t *rq;

    rq = allocRequest(ms, len + 1);
    rq->rq_type = rt_getStringDone;
    rq->rq_flag = (BOOL_T) (ok != BOOLEAN_FALSE);
    blockCopy(&rq->rq_u.ru_text[0], buf, len);
    rq->rq_u.ru_text[len] = '\0';
    ms->ms_serverSend(rq);
}

/*
 * getChar - bottom level of lexical scanner - return the next input character.
 *	Return 'false' if we have hit end-of-file.
 */

BOOL_T
getChar(MUDState_t *ms, CHAR_T *pCh)
{
    FileList_t *fl;
    CHAR_T *p;
    CHAR_T ch;
    LONG_T len;

    if (ms->HadEof) {
	*pCh = ' ';
	return B_FALSE;
    }

    p = ms->InputPtr;
    if (p != P_NULL) {
	ch = *p;
	if (ch != '\0') {
	    *pCh = ch;
	    ++p;
	    ms->InputPtr = p;
	} else {
	    *pCh = ' ';
	    if (ms->NewLineState == nl_clear) {
		ms->NewLineState = nl_faked;
	    } else {
		ms->HadEof = B_TRUE;
	    }
	}
	return B_TRUE;
    }

    if (ms->EditReading) {
	if (ms->InputPos == ms->InputTotal) {
	    if (ms->EditReadEof) {
		*pCh = ' ';
		return B_FALSE;
	    }
	    /* this and below are the only places that 'wantInput' is true */
	    ms->ms_handleAll(B_TRUE);
	    if (ms->EditReadEof) {
		ms->LastChar = ' ';
		*pCh = ' ';
		return B_FALSE;
	    }
	    if (ms->InputPos == ms->InputTotal) {
		ch = ' ';
	    } else {
		ch = ms->InputLine[0];
		ms->InputPos = 1;
	    }
	} else {
	    ch = ms->InputLine[ms->InputPos];
	    ms->InputPos += 1;
	}
	*pCh = ch;
	ms->LastChar = ch;
	return B_TRUE;
    }

    fl = ms->FileList;
    if (fl != P_NULL) {
	fl->fl_prevPrevLine = fl->fl_prevLine;
	fl->fl_prevLine = fl->fl_lineNumber;
	fl->fl_prevPrevColumn = fl->fl_prevColumn;
	fl->fl_prevColumn = fl->fl_column;
	while (fl != P_NULL && fl->fl_inputPos == fl->fl_inputTotal) {
	    /* Need more input data from the file.
	       Note: the following call to 'checkMessages' was once a cause
	       of a bug - this call could read more input from the mainline
	       of the client, resulting in a call to 'inputLine', which
	       would overwrite any previous input line so given. This is
	       now prevented by the 'wantInput' parameter to handleAll. */
	    ms->ms_checkMessages();
	    len = ms->ms_Read(fl->fl_fd, &fl->fl_inputLine[0],
			      INPUT_BUFFER_SIZE);
	    if (len <= 0) {
		ms->ms_Close(fl->fl_fd);
		ms->FileList = fl->fl_next;
		memFree(ms, fl, sizeof(FileList_t));
		fl = ms->FileList;
		if (fl == P_NULL) {
		    lexStart(ms, '\n', ' ');
		    ms->ms_setPrompt("> ");
		} else {
		    lexStart(ms, fl->fl_lastChar, ' ');
		}
		ms->ms_checkMessages();
	    } else {
		fl->fl_inputTotal = len;
		fl->fl_inputPos = 0;
	    }
	}
	if (fl == P_NULL) {
	    ms->HadEof = B_TRUE;
	    *pCh = ' ';
	    ms->LastChar = '\n';
	    return B_FALSE;
	}
	ch = fl->fl_inputLine[fl->fl_inputPos];
	fl->fl_lastChar = ch;
	fl->fl_inputPos += 1;
	if (ch == '\n') {
	    fl->fl_lineNumber += 1;
	    fl->fl_column = 0;
	} else if (ch == '\t') {
	    fl->fl_column = (fl->fl_column + 8) & (-8);
	} else {
	    fl->fl_column += 1;
	}
    } else {
	if (ms->InputPos == ms->InputTotal) {
	    ms->LastChar = '\n';
	    switch (ms->NewLineState) {
	    case nl_clear:
		ms->NewLineState = nl_faked;
		ch = '\n';
		break;
	    case nl_faked:
		ms->NewLineState = nl_blanked1;
		ch = ' ';
		break;
	    case nl_blanked1:
		ms->NewLineState = nl_blanked2;
		ch = ' ';
		break;
	    case nl_blanked2:
		/* wait for user to complete an input line */
		while (! ms->WantQuit && ms->InputPos == ms->InputTotal &&
		    ms->FileList == P_NULL)
		{
		    /* this and above only places that 'wantInput' is true */
		    ms->ms_handleAll(B_TRUE);
		}
		ms->NewLineState = nl_clear;
		if (ms->InputTotal == 0) {
		    ch = '\n';
		    ms->NewLineState = nl_faked;
		    ms->InputPos = 0;
		} else {
		    ch = ms->InputLine[0];
		    ms->InputPos = 1;
		}
		break;
	    default:
		ch = ' ';	/* shut gcc up about 'ch' */
	    }
	} else {
	    ch = ms->InputLine[ms->InputPos];
	    ms->LastChar = ch;
	    ms->InputPos += 1;
	}
    }
    *pCh = ch;
    return (BOOL_T) (! ms->WantQuit);
}

/*
 * libIncRef - ask server to increment reference count on key.
 */

void
libIncRef(MUDState_t *ms, ULONG_T key)
{

    if (ms->ms_incRef != P_NULL) {
	/* we must be the server */
	ms->ms_incRef(key);
    } else {
	/* we must be the client */
	genericRequest(ms, rt_incRef, key);
    }
}

/*
 * libDecRef - ask server to decrement reference count on key.
 */

void
libDecRef(MUDState_t *ms, ULONG_T key)
{

    if (ms->ms_decRef != P_NULL) {
	ms->ms_decRef(key);
    } else {
	genericRequest(ms, rt_decRef, key);
    }
}

/*
 * libLookup - ask server to look up a symbol.
 */

ULONG_T
libLookup(MUDState_t *ms, CHAR_T *name, UINT_T nameLen)
{
    Entry_t *en;
    Request_t *rq;
    ULONG_T key;

    if (ms->ms_lookup != P_NULL) {
	return ms->ms_lookup(name, nameLen);
    } else {
	if (ms->CacheSymbols) {
	    en = symLookup(&ms->SymbolCache, name, nameLen);
	} else {
	    en = P_NULL;
	}
	if (en != P_NULL) {
	    return en->en_key;
	} else {
	    rq = allocRequest(ms, nameLen);
	    rq->rq_type = rt_lookup;
	    blockCopy(&rq->rq_u.ru_text[0], name, nameLen);
	    rq = ms->ms_serverRequest(rq);
	    key = rq->rq_key;
	    freeRequest(ms, rq);
	    if (ms->CacheSymbols && key != 0) {
		symEnter(ms, &ms->SymbolCache, name, nameLen)->en_key = key;
	    }
	    return key;
	}
    }
}

/*
 * libFindName - ask server to find a name for a given key.
 *	NOTE: we are lying and cheating here. We return a pointer to the body
 *	part of a Request_t that we have already freed. This works if the
 *	caller of 'libFindName' does nothing that could allocate a transaction.
 *	Currently, there is only one caller, and he just prints the string
 *	to the user's output.
 */

CHAR_T *
libFindName(MUDState_t *ms, ULONG_T key)
{
    Request_t *rq;

    if (ms->ms_findAnyName != P_NULL) {
	return ms->ms_findAnyName(key);
    } else {
	rq = allocRequest(ms, 0);
	rq->rq_type = rt_findName;
	rq->rq_key = key;
	rq = ms->ms_serverRequest(rq);
	/* we do NOT call freeRequest, since it COULD decide to actually
	   give this request back to the system! */
	rq->rq_next = ms->FreeRequests;
	if (ms->FreeRequests != P_NULL) {
	    ms->FreeRequests->rq_prev = rq;
	}
	rq->rq_prev = P_NULL;
	ms->FreeRequests = rq;
	ms->RequestCount += 1;
	return &rq->rq_u.ru_text[0];
    }
}

/*
 * findProcOnList - find a proc by key on the proc list. If it is not there,
 *	then get it from the server and add it to the list.
 *	Also return a pointer to the lock count for the entry, so that
 *	we can arrange to not flush one that is still referenced. This
 *	is needed for when 'StringToAction' is called, since it will likely
 *	allocate memory, which may try to flush something out of the server's
 *	proc cache, which may be still pointed at by our caller.
 */

Proc_t *
findProcOnList(MUDState_t *ms, ULONG_T key, UINT_T **pLockCount)
{
    ProcList_t *prl;
    ProcList_t **pprl;
    Request_t *rq;

    if (ms->ms_findProc != P_NULL) {
	/* note: this will only be true when we are called by the server */
	return ms->ms_findProc(key, pLockCount);
    } else {
	if (! ms->CacheProcs) {
	    /* There isn't any other time when we delete things from this
	       list, so, since we are supposed to be saving memory, do it
	       now, but don't do the very one we are trying to find! */
	    pprl = &ms->ProcList;
	    while (B_TRUE) {
		prl = *pprl;
		if (prl == P_NULL) {
		    break;
		}
		if (prl->prl_lockCount == 0 && prl->prl_key != key) {
		    *pprl = prl->prl_next;
		    freeProc(ms, prl->prl_proc, prl->prl_key, B_FALSE);
		    memFree(ms, prl, sizeof(ProcList_t));
		} else {
		    pprl = &prl->prl_next;
		}
	    }
	}
	if (ms->ProcMap != P_NULL) {
	    prl = ms->ProcMap[key % PROC_MAP_SIZE];
	    if (prl != P_NULL && prl->prl_key == key) {
		prl->prl_lockCount += 1;
		*pLockCount = &prl->prl_lockCount;
		return prl->prl_proc;
	    }
	}
	prl = ms->ProcList;
	while (prl != P_NULL && prl->prl_key != key) {
	    prl = prl->prl_next;
	}
	if (prl == P_NULL) {
	    prl = MEM_ALLOC(ms, sizeof(ProcList_t), "PRCL");
	    prl->prl_key = key;
	    prl->prl_lockCount = 1;
	    rq = allocRequest(ms, 0);
	    rq->rq_type = rt_readProc;
	    rq->rq_key = key;
	    rq = ms->ms_serverRequest(rq);
	    prl->prl_proc = readProc(ms, &rq->rq_u.ru_bytes[0]);
	    freeRequest(ms, rq);
	    /* don't put it on the list until we are all done setting it up */
	    prl->prl_next = ms->ProcList;
	    ms->ProcList = prl;
	} else {
	    prl->prl_lockCount += 1;
	}
	if (ms->ProcMap != P_NULL) {
	    ms->ProcMap[key % PROC_MAP_SIZE] = prl;
	}
	*pLockCount = &prl->prl_lockCount;
	return prl->prl_proc;
    }
}

/*
 * writeProcNew - a new proc - write it to the server and get a key.
 */

ULONG_T
writeProcNew(MUDState_t *ms, Proc_t *pr)
{
    Request_t *rq;
    BYTE_T *p;
    ULONG_T key;
    UINT_T len;

    writeProc(ms, pr);
    writeResult(ms, &p, &len);
    if (ms->ms_writeProcNew != P_NULL) {
	return ms->ms_writeProcNew(p, len);
    } else {
	rq = allocRequest(ms, len);
	rq->rq_type = rt_writeProcNew;
	blockCopy(&rq->rq_u.ru_bytes[0], p, len);
	rq = ms->ms_serverRequest(rq);
	key = rq->rq_key;
	freeRequest(ms, rq);
	return key;
    }
}

/*
 * replaceProc - replace the indicated proc with the new definition.
 */

void
replaceProc(MUDState_t *ms, ULONG_T key, Proc_t *pr)
{
    Request_t *rq;
    BYTE_T *p;
    UINT_T len;

    writeProc(ms, pr);
    writeResult(ms, &p, &len);
    rq = allocRequest(ms, len);
    rq->rq_type = rt_replaceProc;
    blockCopy(&rq->rq_u.ru_bytes[0], p, len);
    rq->rq_key = key;
    ms->ms_serverSend(rq);
}

/*
 * symbolRequest - common code for defining/deleting a symbol.
 */

static void
symbolRequest(MUDState_t *ms, CHAR_T *name, UINT_T nameLen,
	      ULONG_T valu, ULONG_T table, RequestType_t rt)
{
    Request_t *rq;

    rq = allocRequest(ms, nameLen + sizeof(ULONG_T));
    rq->rq_type = rt;
    blockCopy(&rq->rq_u.ru_text[sizeof(ULONG_T)], name, nameLen);
    rq->rq_key = valu;
    rq->rq_u.ru_otherKey = table;
    ms->ms_serverSend(rq);
}

/*
 * defineSymbol - call the server to define a symbol.
 */

void
defineSymbol(MUDState_t *ms, CHAR_T *name, UINT_T nameLen,
	     ULONG_T key, ULONG_T table)
{

    symbolRequest(ms, name, nameLen, key, table, rt_symbolEnter);
    if (ms->CacheSymbols) {
	symEnter(ms, &ms->SymbolCache, name, nameLen)->en_key = key;
    }
}

/*
 * deleteSymbol - call the server to delete a symbol.
 */

void
deleteSymbol(MUDState_t *ms, CHAR_T *name,
	     UINT_T nameLen, ULONG_T key, ULONG_T table)
{

    symbolRequest(ms, name, nameLen, key, table, rt_symbolDelete);
    if (ms->CacheSymbols) {
	symDelete(ms, &ms->SymbolCache, name, nameLen);
    }
}

/*
 * parseAndRun - parse and run an expression.
 *	Return 'true' if all goes well.
 *	Variants: 0 - send answer back for defining
 *		  1 - print the result
 *		  2 - describe the result
 *		  3 - show the resulting key
 *		  4 - edit the resulting proc
 */

static ULONG_T
parseAndRun(MUDState_t *ms, UINT_T variant)
{
    Proc_t *pr;
    Request_t *rq;
    BYTE_T *p;
    ULONG_T key;
    UINT_T len;

    ms->ErrorCount = 0;
    if (ms->FileList == P_NULL) {
	ms->ms_setPrompt(": ");
    }
    clearNewLineSignificant(ms);
    pr = parseImmediateAction(ms, (BOOL_T) (variant != 0));
    setNewLineSignificant(ms);
    if (ms->FileList == P_NULL) {
	ms->ms_setPrompt("> ");
    }
    if (ms->ErrorCount == 0) {
	writeProc(ms, pr);
	writeResult(ms, &p, &len);
	/* refs will be freed in server */
	freeProc(ms, pr, NIL_ACTION, B_FALSE);
	rq = allocRequest(ms, len);
	blockCopy(&rq->rq_u.ru_bytes[0], p, len);
	rq->rq_type = rt_runProc;
	rq->rq_key = variant;
	rq = ms->ms_serverRequest(rq);
	key = rq->rq_key;
	freeRequest(ms, rq);
	if (ms->State == st_parsing) {
	    /* might have gone out of wizard mode */
	    if (ms->FileList == P_NULL) {
		ms->ms_setPrompt("> ");
	    }
	}
	return key;
    } else {
	freeProc(ms, pr, NIL_ACTION, B_TRUE);
	return 0;
    }
}

/*
 * defineName - handle the 'public' and 'private' define commands.
 */

static void
defineName(MUDState_t *ms, ULONG_T table)
{
    CHAR_T *name;
    ULONG_T key;
    UINT_T nameLen, resultType;

    if (ms->tk_kind == tk_proc) {
	ms->ErrorCount = 0;
	if (ms->FileList == P_NULL) {
	    ms->ms_setPrompt(": ");
	}
	clearNewLineSignificant(ms);
	parseProc(ms, table, NIL_ACTION, P_NULL, ms->DefineOnError, B_FALSE);
	setNewLineSignificant(ms);
	if (ms->FileList == P_NULL) {
	    ms->ms_setPrompt("> ");
	}
    } else if (ms->tk_kind == tk_identifier) {
	if (libLookup(ms, ms->tk_stringValue, ms->tk_stringLength) != 0) {
	    libSymbolError(ms, ms->tk_stringValue, "already defined");
	} else {
	    nameLen = ms->tk_stringLength;
	    name = MEM_ALLOC(ms, nameLen, "NAME");
	    charsCopy(name, ms->tk_stringValue);
	    getToken(ms);
	    key = parseAndRun(ms, 0);
	    if (key != 0) {
		resultType = key >> TYPE_SHIFT;
		if (resultType != TYPE_THING && resultType != TYPE_GRAMMAR &&
		    resultType != TYPE_TABLE && resultType != TYPE_INTEGER &&
		    resultType != TYPE_FIXED &&
		    ! (resultType >= TYPE_PROP_BOOLEAN &&
		       resultType <= TYPE_PROP_LIST_ACTION))
		{
		    if (resultType != TYPE_ERROR) {
			parseError(ms, "cannot define a symbol of that type");
			libIncRef(ms, key);
			libDecRef(ms, key);
		    }
		} else {
		    defineSymbol(ms, name, nameLen, key, table);
		}
	    }
	    memFree(ms, name, nameLen);
	}
    } else {
	parseError(ms, "expecting 'proc' or name to define");
    }
}

/*
 * deleteName - command request to delete a symbol from a table.
 */

static void
deleteName(MUDState_t *ms, ULONG_T table)
{

    if (ms->tk_kind == tk_identifier) {
	deleteSymbol(ms, ms->tk_stringValue, ms->tk_stringLength, 0, table);
    } else {
	parseError(ms, "expecting name to delete");
    }
}

/*
 * sourceFile - common code for doSource and MUDSource.
 */

static void
sourceFile(MUDState_t *ms, FileList_t *fl, CHAR_T ch)
{

    fl->fl_fd = ms->ms_Open(&fl->fl_fileName[0], O_RDONLY, 0);
    if (fl->fl_fd == -1) {
	outputString(ms, "*** can't open file \"");
	outputString(ms, &fl->fl_fileName[0]);
	outputString(ms, "\" ***\n");
	memFree(ms, fl, sizeof(FileList_t));
	lexStart(ms, ch, ' ');
    } else {
	if (ms->FileList == P_NULL) {
	    ms->ms_setPrompt("+ ");
	}
	fl->fl_lineNumber = 1;
	fl->fl_column = 0;
	fl->fl_inputPos = 0;
	fl->fl_inputTotal = 0;
	fl->fl_next = ms->FileList;
	ms->FileList = fl;
	outputString(ms, "Sourcing file \"");
	outputString(ms, &fl->fl_fileName[0]);
	outputString(ms, "\".\n");
	lexStart(ms, ' ', ' ');
    }
}

/*
 * doSource - handle the 'source' command.
 */

static void
doSource(MUDState_t *ms)
{
    FileList_t *fl;
    CHAR_T *p;
    UINT_T len;
    CHAR_T ch;

    fl = ms->FileList;
    if (fl == P_NULL) {
	ch = ms->LastChar;
    } else {
	ch = fl->fl_lastChar;
    }
    while ((ch == ' ' || ch == '\t') && getChar(ms, &ch)) {
    }
    if (ms->HadEof || ch == '\n') {
	parseError(ms, "missing file name for 'source'");
	lexStart(ms, '\n', ' ');
    } else {
	fl = MEM_ALLOC(ms, sizeof(FileList_t), "FLLS");
	p = &fl->fl_fileName[0];
	len = 1;
	*p++ = ch;
	while (getChar(ms, &ch) && ch != '\n') {
	    if (len != TEXT_COLUMNS) {
		*p++ = ch;
		len += 1;
	    }
	}
	*p = '\0';
	sourceFile(ms, fl, ch);
    }
}

/*
 * parseInit - setup for parsing, storage allocation, etc.
 */

static void
parseInit(MUDState_t *ms)
{
    UINT_T i;

    if (! ms->ParseStarted) {
	ms->ParseStarted = B_TRUE;
	ms->InputPtr = P_NULL;
	ms->FileList = P_NULL;
	ms->DefineOnError = B_FALSE;
	memInit(ms);
	lexInit(ms);
	ms->ProcList = P_NULL;
	ms->ProcMap =
	    MEM_ALLOC(ms, sizeof(ProcList_t *) * PROC_MAP_SIZE, "pmap");
	for (i = 0; i != PROC_MAP_SIZE; i += 1) {
	    ms->ProcMap[i] = P_NULL;
	}
    }
}

/*
 * doParsing - top level entry point for wizard mode stuff.
 */

static void
doParsing(MUDState_t *ms)
{
    Request_t *rq;
    CHAR_T *id;
    ULONG_T key;

    parseInit(ms);
    setNewLineSignificant(ms);
    lexStart(ms, ' ', ' ');
    getToken(ms);
    while (! ms->WantQuit && ms->tk_kind != tk_eof) {
	if (ms->tk_kind == tk_newline) {
	    getToken(ms);
	} else {
	    if (ms->tk_kind == tk_public) {
		getToken(ms);
		defineName(ms, TABLE_PUBLIC);
	    } else if (ms->tk_kind == tk_identifier) {
		id = ms->tk_stringValue;
		if (charsEqual(id, "private")) {
		    getToken(ms);
		    defineName(ms, TABLE_PRIVATE);
		} else if (charsEqual(id, "define") || charsEqual(id, "def")) {
		    getToken(ms);
		    if (ms->tk_kind == tk_identifier) {
			id = ms->tk_stringValue;
			key = libLookup(ms, id, ms->tk_stringLength);
			if ((key >> TYPE_SHIFT) != TYPE_TABLE ||
			    key == NIL_TABLE)
			{
			    libSymbolError(ms, id, "is not a table");
			} else {
			    getToken(ms);
			    defineName(ms, key);
			}
		    } else {
			parseError(ms, "missing table for 'define'");
		    }
		} else if (charsEqual(id, "delete") || charsEqual(id, "del")) {
		    getToken(ms);
		    if (ms->tk_kind == tk_public) {
			getToken(ms);
			deleteName(ms, TABLE_PUBLIC);
		    } else if (ms->tk_kind == tk_identifier) {
			id = ms->tk_stringValue;
			if (charsEqual(id, "private")) {
			    getToken(ms);
			    deleteName(ms, TABLE_PRIVATE);
			} else {
			    key = libLookup(ms, id, ms->tk_stringLength);
			    if ((key >> TYPE_SHIFT) != TYPE_TABLE ||
				key == NIL_TABLE)
			    {
				libSymbolError(ms, id, "is not a table");
			    } else {
				getToken(ms);
				deleteName(ms, key);
			    }
			}
		    } else {
			parseError(ms, "missing table for 'delete'");
		    }
		} else if (charsEqual(id, "bye")) {
		    ms->WantQuit = B_TRUE;
		} else if (charsEqual(id, "source")) {
		    doSource(ms);
		    ms->tk_kind = tk_newline;
		} else if (charsEqual(id, "d") || charsEqual(id, "desc") ||
		    charsEqual(id, "describe"))
		{
		    getToken(ms);
		    if (ms->tk_kind == tk_identifier &&
			(peekChar(ms) == '$' || peekChar(ms) == '\n'))
		    {
			id = ms->tk_stringValue;
			key = libLookup(ms, id, ms->tk_stringLength);
			if ((key & INDEX_MASK) >= SMALLEST_BUILTIN) {
			    rq = allocRequest(ms, ms->tk_stringLength);
			    rq->rq_type = rt_describeBuiltin;
			    rq->rq_key = key;
			    blockCopy(&rq->rq_u.ru_text[0],
				      id, ms->tk_stringLength);
			    ms->ms_serverSend(rq);
			} else {
			    (void) parseAndRun(ms, 2);
			}
		    } else {
			(void) parseAndRun(ms, 2);
		    }
		} else if (charsEqual(id, "edit") || charsEqual(id, "ed") ||
		    charsEqual(id, "e"))
		{
		    getToken(ms);
		    if (ms->tk_kind == tk_newline) {
			if (ms->EditProc == NIL_ACTION) {
			    parseError(ms, "no proc selected to edit yet");
			} else {
			    genericRequest(ms, rt_doEditProc, NIL_ACTION);
			    if (! ms->ms_doEdit(P_NULL, 0, EDIT_CODE,
				 B_TRUE, B_TRUE, B_FALSE, P_NULL))
			    {
				genericRequest(ms, rt_editProcDone, 0);
			    }
			}
		    } else {
			(void) parseAndRun(ms, 4);
		    }
		} else if (charsEqual(id, "replace")) {
		    getToken(ms);
		    if (ms->tk_kind == tk_identifier) {
			id = ms->tk_stringValue;
			key = libLookup(ms, id, ms->tk_stringLength);
			if ((key >> TYPE_SHIFT) != TYPE_ACTION) {
			    libSymbolError(ms, id, "is not an action");
			} else {
			    id = charsDup(ms, id);
			    if (ms->FileList == P_NULL) {
				ms->ms_setPrompt(": ");
			    }
			    clearNewLineSignificant(ms);
			    ms->ErrorCount = 0;
			    parseProc(ms, NIL_TABLE, key, id, B_FALSE,
				      (BOOL_T) (ms->FileList == P_NULL));
			    setNewLineSignificant(ms);
			    if (ms->FileList == P_NULL) {
				ms->ms_setPrompt("> ");
			    }
			    charsFree(ms, id);
			}
		    } else {
			parseError(ms, "expecting proc name for 'replace'");
		    }
		} else if (charsEqual(id, "key")) {
		    getToken(ms);
		    (void) parseAndRun(ms, 3);
		} else if (charsEqual(id, "dk")) {
		    getToken(ms);
		    if (ms->tk_kind == tk_intConst) {
			genericRequest(ms, rt_describe, ms->tk_numericValue);
		    } else {
			parseError(ms, "expecting key");
		    }
		} else if (charsEqual(id, "use")) {
		    getToken(ms);
		    if (ms->tk_kind == tk_identifier) {
			id = ms->tk_stringValue;
			key = libLookup(ms, id, ms->tk_stringLength);
			if ((key >> TYPE_SHIFT) != TYPE_TABLE) {
			    libSymbolError(ms, id, "is not a table");
			} else {
			    getToken(ms);
			    genericRequest(ms, rt_useTable, key);
			}
		    } else {
			parseError(ms, "missing table for 'use'");
		    }
		} else if (charsEqual(id, "unuse")) {
		    getToken(ms);
		    if (ms->tk_kind == tk_identifier) {
			id = ms->tk_stringValue;
			key = libLookup(ms, id, ms->tk_stringLength);
			if ((key >> TYPE_SHIFT) != TYPE_TABLE) {
			    libSymbolError(ms, id, "is not a table");
			} else {
			    getToken(ms);
			    genericRequest(ms, rt_unuseTable, key);
			}
		    } else {
			parseError(ms, "missing table for 'unuse'");
		    }
		} else {
		    (void) parseAndRun(ms, 1);
		}
	    } else {
		(void) parseAndRun(ms, 1);
	    }
	    while (! ms->WantQuit &&
		ms->tk_kind != tk_newline && ms->tk_kind != tk_eof)
	    {
		getToken(ms);
	    }
	}
	if (ms->HadEof) {
	    ms->HadEof = B_FALSE;
	    ms->NewLineState = nl_blanked2;
	    lexStart(ms, ms->LastChar, ' ');
	}
    }
}

/*
 * libInitE - external server interface to initialize non-client stuff.
 */

ULONG_T
libInitE(MUDState_t *ms, ULONG_T memLimit, ULONG_T rqMax)
{

    ms->ParseStarted = B_FALSE;
    ms->CacheProcs = B_FALSE;
    ms->CacheSymbols = B_FALSE;
    ms->ProcList = P_NULL;
    ms->FreeRequests = P_NULL;
    ms->RequestLimit = rqMax;
    ms->RequestCount = 0;

    ms->MemoryLimit = memLimit;
    memInit(ms);
    lexInit(ms);
    clearNewLineSignificant(ms);
    return (ULONG_T) errorHeader;
}

/*
 * libTermE - external server interface to terminate non-client stuff.
 */

void
libTermE(MUDState_t *ms)
{

    freeRequests(ms);
    /* no need for these, since memTerm takes care of them, but since
       we are counting allocs/frees, and showing blocks on memTerm,
       it is helpful to do it */
    lexTerm(ms);
    memFree(ms, ms->WriteBuffer, ms->WriteBufferSize);
    memTerm(ms);
}

/*
 * MUDInitialize - startup of one or more client runs.
 */

BOOL_T
MUDInitialize(MUDState_t *ms, ULONG_T flags, ULONG_T rqMax)
{
    ULONG_T len;

    /* This concheck is used since the internal structure of MUDState_t is
       not exported - we just say we need 1000 bytes. Well, we had better
       not actually need more than that! */
    len = sizeof(MUDState_t) -
	((ULONG_T) &ms->MemBlocks - (ULONG_T) &ms->ms_AllocMem);
    if (len > 1000) {
	errorString(ms, "MUDState ");
	errorDecimal(ms, len);
	errorString(ms, " > 1000\n");
	return B_FALSE;
    }
    ms->InputSync = B_FALSE;
    ms->GoAhead = B_FALSE;
    ms->ParseStarted = B_FALSE;
    ms->EditReading = B_FALSE;
    ms->CacheProcs = (BOOL_T) ((flags & FLAG_CACHE_PROCS) != 0);
    ms->CacheSymbols = (BOOL_T) ((flags & FLAG_CACHE_SYMBOLS) != 0);
    ms->IsRemote = (BOOL_T) ((flags & FLAG_IS_REMOTE) != 0);
    ms->ProcList = P_NULL;
    ms->FreeRequests = P_NULL;
    ms->RequestLimit = rqMax;
    ms->RequestCount = 0;
    ms->MemoryLimit = 1000000000;
    ms->PlayerName[0] = '\0';
    ms->Password[0] = '\0';
    ms->SuppliedPassword = P_NULL;
    if (ms->CacheSymbols) {
	parseInit(ms);
	symCreate(ms, &ms->SymbolCache, 8);
    }
    return B_TRUE;
}

/*
 * MUDMain - main line for running a MUD client.
 */

void
MUDMain(MUDState_t *ms, CHAR_T *playerName, CHAR_T *password)
{
    Request_t *rq;
    CHAR_T message[100];
    UINT_T res, len;

    ms->WantQuit = B_FALSE;

    rq = allocRequest(ms, sizeof(struct EffectsInfo));
    rq->rq_type = rt_beginClient;
    for (res = 0; res != BEGIN_KEY_LEN; res += 1) {
	rq->rq_u.ru_bytes[res] = (BYTE_T) (res * res);
    }
    ms->ms_getEffects((struct EffectsInfo *) &rq->rq_u.ru_bytes[0]);
    rq = ms->ms_serverRequest(rq);
    if (rq != P_NULL) {
	ms->SessionKey = rq->rq_key;
	res = rq->rq_uint;
	freeRequest(ms, rq);
    } else {
	ms->WantQuit = B_TRUE;
	res = 0;	/* shut gcc up */
    }

    if (! ms->WantQuit) {
#if 1
	/* In the language this was originally written in, you can use
	   a character escape in a string constant that refers to a
	   defined constant. Nothing like that in C, unfortunately. This
	   stuff not being written originally in C, it makes no use of
	   things like printf/sprintf, etc. so do things manually here
	   to avoid sucking in half a meg of stuff just for that. */
	charsCopy(&message[0], "MUD client V");
	len = charsLen1("MUD client V");
	message[len - 1] = VERSION1;
	message[len] = '.';
	message[len + 1] = VERSION2;
	charsCopy(&message[len + 2], " Copyright 19");
	len = charsLen1(&message[0]);
	message[len - 1] = YEAR1;
	message[len] = YEAR2;
	charsCopy(&message[len + 1], " by Chris Gray started\n");
	log(ms, &message[0]);
#else
	log(ms, "MUD client V\(VERSION1).\(VERSION2) "
	    "Copyright 19\(YEAR1)\(YEAR2) by Chris Gray started\n");
#endif

	if (res == 2) {
	    /* multi-user - get a character name */
	    ms->SuppliedPassword = password;
	    ms->State = st_name;
	    ms->nameCount = 1;
	    if (playerName == P_NULL) {
		ms->ms_setPrompt("Enter character name: ");
	    } else {
		len = charsLen1(playerName) - 1;
		if (len > PLAYER_NAME_LEN - 1) {
		    len = PLAYER_NAME_LEN - 1;
		    *(playerName + len) = '\0';
		}
		requestPlayer(ms, playerName, len);
	    }
	} else if (res == 1) {
	    /* single-user - dive right in */
	    charsCopy(&ms->PlayerName[0], "<SINGLE>");
	    startPlayer(ms);
	} else {
	    /* single-user, but someone else already in */
	    outputString(ms,"This single player dungeon is already in use.\n");
	    setExiting(ms);
	    log(ms, "subsequent client when single-player\n");
	}

	/* I don't know why the Draco/Amiga version was happy having
	   these (with nl_blanked2 instead) inside of doParsing, since
	   that throws away an input line gotten from the first call
	   to handleAll. Something about the order things are done in,
	   I guess. */
	ms->HadEof = B_FALSE;
	ms->NewLineState = nl_clear;
	ms->InputPos = 0;
	ms->InputTotal = 0;
	ms->EditProc = NIL_ACTION;
	if (ms->State == st_parsing) {
	    /* handle wizard-mode stuff when char and pw pre-supplied */
	    doParsing(ms);
	}
	while (! ms->WantQuit) {
	    if (ms->State == st_parsing) {
		ms->ms_handleAll(ms->InputPos == ms->InputTotal);
		/* wizard mode enabled by the server */
		doParsing(ms);
	    } else if (ms->State == st_normal) {
		ms->ms_handleAll(ms->GoAhead);
	    } else {
		ms->ms_handleAll(B_TRUE);
	    }
	}

	logPlayer(ms, "Character ", " leaves the game\n");

#if 1
	charsCopy(&message[0], "MUD client V");
	len = charsLen1("MUD client V");
	message[len - 1] = VERSION1;
	message[len] = '.';
	message[len + 1] = VERSION2;
	charsCopy(&message[len + 2], " Copyright 19");
	len = charsLen1(&message[0]);
	message[len - 1] = YEAR1;
	message[len] = YEAR2;
	charsCopy(&message[len + 1], " by Chris Gray terminated\n");
	log(ms, &message[0]);
#else
	log(ms, "MUD client V\(VERSION1).\(VERSION2) "
	    "Copyright 19\(YEAR1)\(YEAR2) by Chris Gray terminated\n");
#endif

	rq = allocRequest(ms, 0);
	rq->rq_type = rt_endClient;
	freeRequest(ms, ms->ms_serverRequest(rq));
    }
    parseTerminate(ms, B_TRUE);
}

/*
 * MUDAbort - caller has trouble - wants us to clean up. Caller will NOT
 *	be calling MUDTerminate.
 */

void
MUDAbort(MUDState_t *ms)
{

    if (ms->CacheSymbols) {
	symDestroy(ms, &ms->SymbolCache);
    }
    parseTerminate(ms, B_TRUE);
    freeRequests(ms);
    memTerm(ms);
}

/*
 * MUDShutDown - caller wants us to shut down soon (e.g. close box hit)
 */

void
MUDShutDown(MUDState_t *ms)
{

    ms->WantQuit = B_TRUE;
}

/*
 * MUDTerminate - termination from one or more MUD client runs.
 */

void
MUDTerminate(MUDState_t *ms)
{

    /* cannot do this in parseTerminate, since we would want a symEmpty to
       do it there - it must be valid for any subsequent run */
    if (ms->CacheSymbols) {
	symDestroy(ms, &ms->SymbolCache);
    }
    freeRequests(ms);
    memTerm(ms);
}

/*
 * MUDSource - allow sourcing a file while in wizard mode.
 */

BOOL_T
MUDSource(MUDState_t *ms, CHAR_T *fileName)
{
    FileList_t *fl;
    CHAR_T *p;
    UINT_T len;

    if (ms->State == st_parsing) {
	fl = MEM_ALLOC(ms, sizeof(FileList_t), "FLLS");
	p = &fl->fl_fileName[0];
	len = 0;
	while (*fileName != '\0') {
	    if (len != TEXT_COLUMNS) {
		*p++ = *fileName;
		len += 1;
	    }
	    ++fileName;
	}
	*p = '\0';
	sourceFile(ms, fl, ' ');
	return B_TRUE;
    } else {
	/* not in wizard mode */
	return B_FALSE;
    }
}

/*
 * MUDCrypt - simple password encrypt/decrypt.
 */

static void
swapBits(BYTE_T *p, UINT_T index1, UINT_T index2)
{
    BYTE_T byte1, byte2, mask1, mask2, new1, new2;
    UINT_T mod1, mod2;
#define BITS_PER_BYTE	8

    mod1 = index1 % BITS_PER_BYTE;
    mod2 = index2 % BITS_PER_BYTE;
    index1 /= BITS_PER_BYTE;
    index2 /= BITS_PER_BYTE;
    if (index1 != index2) {
	mask1 = 1 << mod1;
	mask2 = 1 << mod2;
	byte1 = *(p + index1);
	byte2 = *(p + index2);
	new1 = byte1;
	new2 = byte2;
	new1 &= ~mask1;
	new2 &= ~mask2;
	byte1 >>= mod1;
	byte2 >>= mod2;
	byte1 &= 1;
	byte2 &= 1;
	byte1 <<= mod2;
	byte2 <<= mod1;
	new1 |= byte2;
	new2 |= byte1;
	*(p + index1) = new1;
	*(p + index2) = new2;
    }
#undef BITS_PER_BYTE
}

#if 0
djr version
> static void swapBits(BYTE_T *p, UINT_T index1, UINT_T index2) {
>     BYTE_T mask1, mask2;
>     UINT_T mod1, mod2, zero1, zero2;
>
>     mod1 = index1 % BITS_PER_BYTE;
>     mod2 = index2 % BITS_PER_BYTE;
>     index1 /= BITS_PER_BYTE;
>     index2 /= BITS_PER_BYTE;
>     if (index1 != index2) {
> 	mask1 = 1 << mod1;
> 	mask2 = 1 << mod2;
> 	zero1 = (p[index1] & mask1) == 0;
> 	zero2 = (p[index2] & mask2) == 0;
> 	if (zero1 != zero2) {
> 	    p[index1] ^= mask1;
> 	    p[index2] ^= mask2;
> 	}
>     }
> }
#endif

static void
crypt(CHAR_T *pp, BOOL_T decrypt, ULONG_T sessionKey)
{
#define SWAP_COUNT	50
    BYTE_T key[4];
    BYTE_T *p;
    int i;
    static BYTE_T swaps[SWAP_COUNT * 2] = {
	119,  55,
	 78,  45,
	149,  48,
	 35,  90,
	 77, 133,
	 10,  14,
	 80, 136,
	 17,  36,
	111,  46,
	 74,  51,
	 29,  12,
	102,  19,
	156, 138,
	 99, 112,
	118,  26,
	  5, 116,
	128, 109,
	 62, 104,
	 52, 134,
	  2,  33,
	 72, 150,
	106,  93,
	 16,  60,
	132,  92,
	 47,  63,
	  1,  97,
	113,  38,
	 85,  32,
	 11,   0,
	 98, 139,
	  7,  58,
	  4, 142,
	 56,  64,
	125,  34,
	 54, 103,
	130,  76,
	127,  79,
	 50,  24,
	 70,  89,
	 94,  18,
	 71, 135,
	140, 131,
	 44, 153,
	  6,  13,
	115,  87,
	107,  95,
	114, 141,
	 49, 151,
	 27,  83,
	 65, 126
    };

    /* session key is to be applied in a big-endian nature */
    key[0] = sessionKey >> 24;
    key[1] = sessionKey >> 16;
    key[2] = sessionKey >> 8;
    key[3] = sessionKey;
    p = (BYTE_T *) pp;
    if (! decrypt) {
	for (i = 0; i != PASSWORD_LEN; i += 1) {
	    *p++ ^= key[i % 4];
	}
    }
    p = (BYTE_T *) pp;
    for (i = 0; i != SWAP_COUNT; i += 1) {
	swapBits(p, swaps[i * 2], swaps[i * 2 + 1]);
    }
    if (decrypt) {
	for (i = 0; i != PASSWORD_LEN; i += 1) {
	    *p++ ^= key[i % 4];
	}
    }
}

void
MUDCrypt(CHAR_T *p, ULONG_T decrypt, ULONG_T sessionKey)
{

    crypt(p, (BOOL_T) (decrypt != 0), sessionKey);
}
