#include "MUD.h"
#include "types.h"
#include "MUDInternal.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * memory.c - memory allocator.
 */

#define DEBUG_SOFT	B_FALSE
#define DEBUG		B_FALSE
#define DEBUG_HARD	B_FALSE
#define CHECK		B_FALSE
#define CHECK_BLOCKS	B_FALSE
#define CHECK_HARD	B_FALSE
#define CLEAR_ON_ALLOC	B_FALSE
#define ENABLE_SHOW	B_FALSE
#define SHOW_ON_EXIT	B_FALSE
#define ENABLE_VALIDATE	B_FALSE

#define DEFAULT_BUFFER_SIZE	1024
#define MAX_BLOCK		(1024 * 1024)
#define TAG_COUNT		1000

#define FIXED_STRIDE	4	 /* size increment - power of 2 */
#define FIXED_MINIMUM	sizeof(MemPiece_t)
#define FIXED_TOTAL	(FIXED_MINIMUM + (FIXED_COUNT - 1) * FIXED_STRIDE)

#define LENGTH_MASK	0x00ffffff

#define BLOCK_MINIMUM	8192
#define BLOCK_STRIDE	4096	/* power of two */

#define TAG_FREE	0x5a
#define TAG_ALLOCATED	0xa5
#define TAG_END 	0xf3
#define TAG_OTHER	0x11

#if CHECK_HARD || CHECK_BLOCKS
#define MEM_VALIDATE(ms, where)	memValidate(ms, where)
#else
#define MEM_VALIDATE(ms, where)
#endif

/*
 * memFreeBlocks - free the large blocks that we had allocated.
 */

void
memFreeBlocks(MUDState_t *ms)
{
    MemBlock_t *mb;

    while (B_TRUE) {
	mb = ms->MemBlocks;
	if (mb == P_NULL) {
	    break;
	}
	ms->MemBlocks = mb->mb_next;
	ms->ms_FreeMem(mb, mb->mb_length);
    }
    /* note that we leave ms->MemBlocks == P_NULL, so this routine can be
       called again with no harm (the out of memory abort) */
}

/*
 * memAbort1 - local routine to print errors and abort.
 */

static void
#if TAG_ALLOCS
memAbort1(MUDState_t *ms, CHAR_T *st, CHAR_T *what)
#else
memAbort1(MUDState_t *ms, CHAR_T *st)
#endif
{

    errorString(ms, "*** ");
#if TAG_ALLOCS
    if (what != P_NULL) {
	errorString(ms, what);
	ms->ms_errorChar(' ');
    }
#endif
    errorString(ms, st);
    errorString(ms, " ***\n");
    memAbort(ms, B_FALSE);
}

#if TAG_ALLOCS
#define MEM_ABORT1(ms, st, what)	memAbort1(ms, st, what)
#else
#define MEM_ABORT1(ms, st, what)	memAbort1(ms, st)
#endif

#if ENABLE_VALIDATE || CHECK_HARD || CHECK_BLOCKS

/*
 * memAbort2 - another one, this says where it is aborting from.
 */

static void
#if TAG_ALLOCS
memAbort2(MUDState_t *ms, CHAR_T *where, CHAR_T *kind, CHAR_T *what)
#else
memAbort2(MUDState_t *ms, CHAR_T *where, CHAR_T *kind)
#endif
{

    errorString(ms, "*** ");
    errorString(ms, where);
    errorString(ms, ": ");
    errorString(ms, kind);
#if TAG_ALLOCS
    if (what != P_NULL) {
	ms->ms_errorChar(' ');
	errorString(ms, what);
    }
#endif
    errorString(ms, " ***\n");
    memAbort(ms, B_FALSE);
}

#if TAG_ALLOCS
#define MEM_ABORT2(ms, where, kind, what)	memAbort2(ms, where, kind,what)
#else
#define MEM_ABORT2(ms, where, kind, what)	memAbort2(ms, where, kind)
#endif

#endif

#if ENABLE_SHOW

/*
 * dumpBlocks - dump out what we have for memory blocks.
 */

static void
dumpBlocks(MUDState_t *ms, CHAR_T *where)
{
    MemPiece_t *mp;
    MemBlock_t *mb;
    ULONG_T blockLen, pieceLen;

    errorString(ms, "dumpBlocks @ ");
    errorString(ms, where);
    errorString(ms, ":\n");
    mb = ms->MemBlocks;
    while (mb != P_NULL) {
	blockLen = mb->mb_length;
	errorString(ms, "block @ ");
	errorHex(ms, (ULONG_T) mb);
	errorString(ms, " blockLen ");
	errorDecimal(ms, blockLen);
	errorString(ms, "\n");
	blockLen -= (sizeof(MemBlock_t) + sizeof(MemPiece_t));
	mp = (MemPiece_t *) (mb + 1);
	while (blockLen != 0) {
	    pieceLen = mp->mp_.mp_length & LENGTH_MASK;
#if TAG_ALLOCS
	    ms->ms_errorChar(' ');
	    errorString(ms, mp->mp_what);
#endif
	    errorString(ms, " piece @ ");
	    errorHex(ms, (ULONG_T) mp);
	    errorString(ms, " len ");
	    errorDecimal(ms, pieceLen);
	    errorString(ms,
		(mp->mp_.mp_tag[TAG_INDEX] == TAG_ALLOCATED) ? " allocated\n" :
		" free\n");
	    blockLen -= pieceLen;
	    mp = (MemPiece_t *) ((BYTE_T *) mp + pieceLen);
	}
	mb = mb->mb_next;
    }
}

#endif

void
dumpBlocksE(MUDState_t *ms, CHAR_T *where)
{

#if ENABLE_SHOW
    dumpBlocks(ms, where);
#endif
}

#if ENABLE_SHOW

static void
showBlocks(MUDState_t *ms, CHAR_T *where)
{
    typedef struct {
	CHAR_T *tc_tag;
	ULONG_T tc_len;
	ULONG_T tc_count;
    } tagCount_t;
    tagCount_t tagCounts[TAG_COUNT];
    MemPiece_t *mp;
    CHAR_T *tag;
    MemBlock_t *mb;
    tagCount_t *tc;
    ULONG_T blockLen, count, i, pieceLen, blocksAlloced;

    blocksAlloced = 0;
    count = 0;
    mb = ms->MemBlocks;
    while (mb != P_NULL) {
	blockLen = mb->mb_length;
	blockLen -= (sizeof(MemBlock_t) + sizeof(MemPiece_t));
	mp = (MemPiece_t *) (mb + 1);
	while (blockLen != 0) {
	    pieceLen = mp->mp_.mp_length & LENGTH_MASK;
	    if (mp->mp_.mp_tag[TAG_INDEX] == TAG_ALLOCATED) {
		blocksAlloced += 1;
		tag = mp->mp_what;
		tc = &tagCounts[0];
		i = 0;
		while (i != count &&
		    (tc->tc_tag != tag || tc->tc_len != pieceLen))
		{
		    i += 1;
		    ++tc;
		}
		if (i == count) {
		    if (count != TAG_COUNT) {
			count += 1;
			tc->tc_tag = tag;
			tc->tc_len = pieceLen;
			tc->tc_count = 1;
		    }
		} else {
		    tc->tc_count += 1;
		}
	    }
	    blockLen -= pieceLen;
	    mp = (MemPiece_t *) ((BYTE_T *) mp + pieceLen);
	}
	mb = mb->mb_next;
    }

    if (count != 0) {
	errorString(ms, "showBlocks @ ");
	errorString(ms, where);
	errorString(ms, ", ");
	errorDecimal(ms, blocksAlloced);
	errorString(ms, " blocks alloced:\n");
	tc = &tagCounts[0];
	for (i = 0; i < count; i += 1) {
	    ms->ms_errorChar(' ');
	    errorString(ms, tc->tc_tag);
	    ms->ms_errorChar(' ');
	    errorDecimal(ms, tc->tc_len);
	    ms->ms_errorChar(' ');
	    errorDecimal(ms, tc->tc_count);
	    ms->ms_errorChar('\n');
	    ++tc;
	}
    } else {
	errorString(ms, where);
	errorString(ms, ": no memory allocated\n");
    }
}

#endif

void
showBlocksE(MUDState_t *ms, CHAR_T *where)
{
#if TAG_ALLOCS
#if ENABLE_SHOW
    showBlocks(ms, where);
#endif
#endif
}

#if ENABLE_VALIDATE || CHECK_HARD || CHECK_BLOCKS

/*
 * memValidate - run through and validate memory structures.
 */

static void
memValidate(MUDState_t *ms, CHAR_T *where)
{
    MemBlock_t *mb;
    ULONG_T wantLen;
#if ENABLE_VALIDATE || CHECK_HARD
    MemPiece_t *mp, *mpPrev;
    ULONG_T len;
    UINT_T i;
#endif

#if ENABLE_VALIDATE || CHECK_HARD
    wantLen = FIXED_MINIMUM;
    for (i = 0; i < FIXED_COUNT; i += 1) {
	mp = ms->ShortList[i];
	while (mp != P_NULL) {
	    if (mp->mp_.mp_tag[TAG_INDEX] != TAG_FREE) {
		MEM_ABORT2(ms, where, "tag on short list", mp->mp_what);
	    }
	    len = mp->mp_.mp_length & LENGTH_MASK;
	    if (len != wantLen) {
		MEM_ABORT2(ms, where, "len on short list", mp->mp_what);
	    }
	    mp = mp->mp_next;
	}
	wantLen += FIXED_STRIDE;
    }
    mpPrev = P_NULL;
    mp = ms->LongList;
    while (mp != P_NULL) {
	if (mp->mp_prev != mpPrev) {
	    MEM_ABORT2(ms, where, "prev on long list", mp->mp_what);
	}
	if (mp->mp_.mp_tag[TAG_INDEX] != TAG_FREE) {
	    MEM_ABORT2(ms, where, "tag on long list", mp->mp_what);
	}
	len = mp->mp_.mp_length & LENGTH_MASK;
	if (len <= FIXED_TOTAL) {
	    MEM_ABORT2(ms, where, "short len on long list", mp->mp_what);
	}
	if (len > MAX_BLOCK) {
	    MEM_ABORT2(ms,where,"len > 1 Meg on long list", mp->mp_what);
	}
	if ((len & (FIXED_STRIDE - 1)) != 0) {
	    MEM_ABORT2(ms, where, "len on long list", mp->mp_what);
	}
	mpPrev = mp;
	mp = mp->mp_next;
    }
#endif
    mb = ms->MemBlocks;
    while (mb != P_NULL) {
	wantLen = mb->mb_length;
	if (wantLen > MAX_BLOCK + sizeof(MemBlock_t) + sizeof(MemPiece_t)) {
	    MEM_ABORT2(ms, where, "length of block", P_NULL);
	}
#if ENABLE_VALIDATE || CHECK_HARD
	wantLen -= (sizeof(MemBlock_t) + sizeof(MemPiece_t));
	mp = (MemPiece_t *) (mb + 1);
	while (wantLen >= sizeof(MemPiece_t)) {
	    if (mp->mp_.mp_tag[TAG_INDEX] != TAG_FREE &&
		mp->mp_.mp_tag[TAG_INDEX] != TAG_ALLOCATED)
	    {
		MEM_ABORT2(ms, where, "tag in block", mp->mp_what);
	    }
	    len = mp->mp_.mp_length & LENGTH_MASK;
	    if ((len & (FIXED_STRIDE - 1)) != 0) {
		MEM_ABORT2(ms, where, "len in block", mp->mp_what);
	    }
	    if (len > wantLen) {
		MEM_ABORT2(ms, where, "long len in block", mp->mp_what);
	    }
	    if (len > MAX_BLOCK) {
		MEM_ABORT2(ms, where, "len > 1 Meg in block", mp->mp_what);
	    }
	    wantLen -= len;
	    mp = (MemPiece_t *) ((BYTE_T *) mp + len);
	}
#endif
	mb = mb->mb_next;
    }
}

#endif

void
memValidateE(MUDState_t *ms, CHAR_T *where)
{

#if ENABLE_VALIDATE
    memValidate(ms, where);
#endif
}

/*
 * memMergePiece - merge this piece into whatever it can be merged into.
 *	Return the new length of the merged piece.
 *	Assumption: we are only called on a free piece.
 */

static ULONG_T
memMergePiece(MUDState_t *ms, MemPiece_t *mp)
{
    MemPiece_t *mp2;
    MemPiece_t **mpp;
    ULONG_T pieceLen, nextLen;

    pieceLen = mp->mp_.mp_length & LENGTH_MASK;
    while (B_TRUE) {
	mp2 = (MemPiece_t *) ((BYTE_T *) mp + pieceLen);
	/* Hmm. Note that this can be referencing memory that we have never
	   allocated from the system! Oops. How to fix? */
	if (mp2->mp_.mp_tag[TAG_INDEX] != TAG_FREE) {
	    break;
	}
	/* can merge this piece with the next one up */
#if DEBUG_HARD
	errorString(ms, "memMerge merging\n");
#endif
	nextLen = mp2->mp_.mp_length & LENGTH_MASK;
	pieceLen += nextLen;
	if (nextLen <= FIXED_TOTAL) {
	    /* the tail piece is on a short list */
#if DEBUG_HARD
	    errorString(ms, "tail piece on short list\n");
#endif
	    mpp = &ms->ShortList[(nextLen - FIXED_MINIMUM) / FIXED_STRIDE];
	    while (*mpp != mp2) {
		if (*mpp == P_NULL) {
		    MEM_ABORT1(ms, "short piece not found in memMergePiece",
			      P_NULL);
		}
		mpp = &(*mpp)->mp_next;
	    }
	    *mpp = mp2->mp_next;
	} else {
	    /* the tail piece is on the long list */
#if DEBUG_HARD
	    errorString(ms, "tail piece on long list\n");
#endif
	    if (mp2->mp_prev != P_NULL) {
		mp2->mp_prev->mp_next = mp2->mp_next;
	    } else {
		ms->LongList = mp2->mp_next;
	    }
	    if (mp2->mp_next != P_NULL) {
		mp2->mp_next->mp_prev = mp2->mp_prev;
	    }
	}
    }
    mp->mp_.mp_length = pieceLen;
    mp->mp_.mp_tag[TAG_INDEX] = TAG_FREE;
    return pieceLen;
}

/*
 * memFree - free a chunk of memory.
 */

void
memFree(MUDState_t *ms, void *p, ULONG_T len)
{
    MemPiece_t *mp, *mp2;
    MemPiece_t **mpp;
    ULONG_T actualLen;

#if DEBUG
    debugX(ms, "memFree(", (ULONG_T) p, ", ");
    debugX(ms, "", len, ")\n");
#endif
#if CHECK
    if (p == P_NULL) {
	MEM_ABORT1(ms, "nil ptr to memFree", P_NULL);
    }
#endif
    MEM_VALIDATE(ms, "memFree start");

    ms->MemoryFrees += 1;
#if CHECK
    if (len < FIXED_MINIMUM - MEM_EXTRA) {
	len = FIXED_MINIMUM;
    } else {
	len += (MEM_EXTRA + FIXED_STRIDE - 1);
	len &= (-FIXED_STRIDE);
    }
#if DEBUG_HARD
    errorString(ms, "memFree fixed len\n");
#endif
#endif
    mp = (MemPiece_t *) ((BYTE_T *) p - MEM_EXTRA);
#if TAG_ALLOCS
    mp->mp_what = "FREE";
#endif

#if CHECK
    if (len > MAX_BLOCK) {
	MEM_ABORT1(ms, "len > 1 Meg on memFree", mp->mp_what);
    } else if (mp->mp_.mp_tag[TAG_INDEX] == TAG_FREE) {
	MEM_ABORT1(ms, "freeing already free memory in memFree", mp->mp_what);
    } else if (mp->mp_.mp_tag[TAG_INDEX] != TAG_ALLOCATED) {
	MEM_ABORT1(ms, "invalid tag on memFree", mp->mp_what);
    }
#endif
    actualLen = mp->mp_.mp_length & LENGTH_MASK;
#if CHECK
    if (actualLen < len) {
	MEM_ABORT1(ms, "len too big on memFree", mp->mp_what);
    }
    if ((actualLen & (FIXED_STRIDE - 1)) != 0) {
	MEM_ABORT1(ms, "stored length bad on memFree", mp->mp_what);
    }
    if (actualLen >= len + FIXED_MINIMUM) {
	MEM_ABORT1(ms, "len too small on memFree", mp->mp_what);
    }
#endif
#if DEBUG_HARD
    errorString(ms, "memFree past all checks\n");
#endif
    ms->MemoryInUse -= actualLen;

    /* Merge this new free piece with any just after it that are already
       free. This hopefully avoids some fragmentation. Note that it is
       'memMergePiece' that sets the merged length, AND TAG_FREE into the
       piece. */
    actualLen = memMergePiece(ms, mp);

    if (actualLen <= FIXED_TOTAL) {
#if DEBUG_HARD
	errorString(ms, "final piece on short list\n");
#endif
	mpp = &ms->ShortList[(actualLen - FIXED_MINIMUM) / FIXED_STRIDE];
	mp->mp_next = *mpp;
	*mpp = mp;
    } else {
#if DEBUG_HARD
	errorString(ms, "final piece on long list\n");
#endif
	mp2 = ms->LongList;
	mp->mp_next = mp2;
	if (mp2 != P_NULL) {
	    mp2->mp_prev = mp;
	}
	mp->mp_prev = P_NULL;
	ms->LongList = mp;
    }
    MEM_VALIDATE(ms, "memFree end");
}

/*
 * memFreeE - external interface to memFree
 */

void
memFreeE(MUDState_t *ms, void *p, ULONG_T len)
{

    memFree(ms, p, len);
}

/*
 * memMerge - called by allocMem when it does not have a piece big enough,
 *	but it is possible that merging can build one.
 */

static void
memMerge(MUDState_t *ms)
{
    MemBlock_t *mb;
    MemPiece_t *mp, *mp2;
    MemPiece_t **mpp;
    ULONG_T blockLen, pieceLen, oldLen;

    mb = ms->MemBlocks;
    while (mb != P_NULL) {
	blockLen = mb->mb_length;
	blockLen -= (sizeof(MemBlock_t) + sizeof(MemPiece_t));
	mp = (MemPiece_t *) (mb + 1);
	while (blockLen != 0) {
	    pieceLen = mp->mp_.mp_length & LENGTH_MASK;
	    if (mp->mp_.mp_tag[TAG_INDEX] == TAG_FREE) {
		oldLen = pieceLen;
		pieceLen = memMergePiece(ms, mp);
		if (oldLen <= FIXED_TOTAL && pieceLen != oldLen) {
		    /* have to move the newly grown piece from the short
		       list it was on to either another short list or to
		       the long list */
		    mpp = &ms->ShortList[(oldLen-FIXED_MINIMUM)/FIXED_STRIDE];
		    while (*mpp != mp) {
			if (*mpp == P_NULL) {
			    MEM_ABORT1(ms, "short piece not found in memMerge",
				       mp->mp_what);
			}
			mpp = &(*mpp)->mp_next;
		    }
		    *mpp = mp->mp_next;
		    if (pieceLen <= FIXED_TOTAL) {
			oldLen = (pieceLen - FIXED_MINIMUM) / FIXED_STRIDE;
			mpp = &ms->ShortList[oldLen];
			mp->mp_next = *mpp;
			*mpp = mp;
		    } else {
			mp2 = ms->LongList;
			mp->mp_next = mp2;
			if (mp2 != P_NULL) {
			    mp2->mp_prev = mp;
			}
			mp->mp_prev = P_NULL;
			ms->LongList = mp;
		    }
		}
	    }
	    blockLen -= pieceLen;
	    mp = (MemPiece_t *) ((BYTE_T *) mp + pieceLen);
	}
	mb = mb->mb_next;
    }
}

/*
 * memAlloc - try to allocate some memory. Lots can happen.
 */

void *
#if TAG_ALLOCS
memAllocT(MUDState_t *ms, ULONG_T len, char *what)
#else
memAlloc(MUDState_t *ms, ULONG_T len)
#endif
{
    MemPiece_t *mp, *mp2;
    MemPiece_t **mpp;
    MemBlock_t *mb;
#if CLEAR_ON_ALLOC
    ULONG_T *lp;
#endif
    ULONG_T pieceLen;
    BOOL_T (*reclaim)(BOOL_T panicing);

#if DEBUG
#if TAG_ALLOCS
    errorString(ms, "memAlloc(");
    errorDecimal(ms, len);
    errorString(ms, ", ['");
    errorString(ms, what);
    errorString(ms, "'])\n");
#else
    debugX(ms, "memAlloc(", len, ")\n");
#endif
#endif
    MEM_VALIDATE(ms, "memAlloc start");
    ms->MemoryAllocs += 1;

    /* normalize the requested length */

    if (len < FIXED_MINIMUM - MEM_EXTRA) {
	len = FIXED_MINIMUM;
    } else {
	len += (MEM_EXTRA + FIXED_STRIDE - 1);
	len &= (-FIXED_STRIDE);
    }

    /* check for a quick allocation succeeding */

    if (len <= FIXED_TOTAL) {
	/* size is within the quick-allocation range */
#if DEBUG_HARD
	errorString(ms, "within small size\n");
#endif
	mpp = &ms->ShortList[(len - FIXED_MINIMUM) / FIXED_STRIDE];
	mp = *mpp;
	if (mp != P_NULL) {
	    *mpp = mp->mp_next;
	    mp->mp_.mp_tag[TAG_INDEX] = TAG_ALLOCATED;
	    ms->MemoryInUse += len;
#if CLEAR_ON_ALLOC
	    lp = (ULONG_T *) &mp->mp_next;
	    len /= sizeof(ULONG_T);
	    while (B_TRUE) {
		len -= 1;
		if (len == 0) {
		    break;
		}
		*lp++ = 0;
	    }
#endif
#if DEBUG
	    debugX(ms, " => ", (ULONG_T) &mp->mp_next, "\n");
#endif
	    MEM_VALIDATE(ms, "memAlloc first return");
#if TAG_ALLOCS
	    mp->mp_what = what;
#endif
	    return &mp->mp_next;
	}
    }

    /* no quick allocation - check on the long list */

#if DEBUG_HARD
    errorString(ms, "checking long list\n");
#endif
    mp = ms->LongList;
    while (mp != P_NULL) {

	/* Before we even check if this piece is big enough, merge it with
	   any just after it that are free. */
	pieceLen = memMergePiece(ms, mp);

	if (pieceLen >= len) {
	    /* found one big enough - use it */
	    mp2 = mp->mp_next;
	    if (mp2 != P_NULL) {
		mp2->mp_prev = mp->mp_prev;
	    }
	    if (mp->mp_prev != P_NULL) {
		mp->mp_prev->mp_next = mp2;
	    } else {
		ms->LongList = mp2;
	    }
	    pieceLen -= len;
	    if (pieceLen >= FIXED_MINIMUM) {
#if DEBUG_HARD
		errorString(ms, "chopping off tail of long list piece\n");
#endif
		/* chop off the tail and free it */
		mp->mp_.mp_length = len;
#if ENABLE_VALIDATE || CHECK_HARD
		/* needed for the block conchecks in memValidate */
		mp->mp_.mp_tag[TAG_INDEX] = TAG_ALLOCATED;
#endif
		mp2 = (MemPiece_t *) ((BYTE_T *) mp + len);
		mp2->mp_.mp_length = pieceLen;
#if ENABLE_VALIDATE || CHECK
		/* We can omit this since we know that the call to
		   'memFree' will not try to merge into this block
		   from below, so the tag is irrelevant. */
		mp2->mp_.mp_tag[TAG_INDEX] = TAG_ALLOCATED;
#endif
#if TAG_ALLOCS
		mp2->mp_what = "FREE";
#endif
		ms->MemoryInUse += pieceLen;
		pieceLen -= MEM_EXTRA;
		mp2 = (MemPiece_t *) ((BYTE_T *) mp2 + MEM_EXTRA);
		/* compensate for add in memFree */
		ms->MemoryFrees -= 1;
		memFree(ms, mp2, pieceLen);
	    }
	    /* Similarly, this does not have to be done before the call
	       to 'memFree' above. */
	    mp->mp_.mp_tag[TAG_INDEX] = TAG_ALLOCATED;
	    ms->MemoryInUse += len;
#if CLEAR_ON_ALLOC
	    lp = (ULONG_T *) &mp->mp_next;
	    len /= sizeof(ULONG_T);
	    while (B_TRUE) {
		len -= 1;
		if (len == 0) {
		    break;
		}
		*lp++ = 0;
	    }
#endif
#if DEBUG
	    debugX(ms, " => ", (ULONG_T) &mp->mp_next, "\n");
#endif
	    MEM_VALIDATE(ms, "memAlloc second return");
#if TAG_ALLOCS
	    mp->mp_what = what;
#endif
	    return &mp->mp_next;
	}
	mp = mp->mp_next;
    }

    len -= MEM_EXTRA;			/* recursive calls here to end */

    /* found nothing on the long list - try a general merge */

    if (! ms->Merging && ms->MemoryAllocated - ms->MemoryInUse >= len) {
	ms->Merging = B_TRUE;
	memMerge(ms);
	ms->MemoryAllocs -= 1;
	mp = MEM_ALLOC(ms, len, what);
	ms->Merging = B_FALSE;
	return mp;
    }

    /* found nothing on the merged lists - need more from elsewhere */

    if (ms->MemoryAllocated > ms->MemoryLimit && ms->ms_ReclaimMem != P_NULL) {
	/* we have gone past the limit requested - try to reclaim some from
	   the rest of the program. */
#if DEBUG_SOFT
	errorString(ms, "memAlloc wants to reclaim, allocated ");
	errorDecimal(ms, ms->MemoryAllocated);
	errorString(ms, ", in use ");
	errorDecimal(ms, ms->MemoryInUse);
	errorString(ms, ", limit ");
	errorDecimal(ms, ms->MemoryLimit);
	errorString(ms, "\n");
#endif
	/* we save/restore this pointer so that we will not just recurse
	   forever here! Note that there are cases in the server where even
	   reclaimMem can USE memory (it calls flushProcCache, which tries
	   to write out a proc which needs a larger byte buffer, which it
	   will try to allocate...) */
	reclaim = ms->ms_ReclaimMem;
	ms->ms_ReclaimMem = P_NULL;
	while (reclaim(B_FALSE) && ms->MemoryInUse > ms->MemoryLimit) {
	}
	ms->MemoryReclaims += 1;
	ms->Merging = B_TRUE;
	memMerge(ms);
	ms->MemoryAllocs -= 1;
	mp = MEM_ALLOC(ms, len, what);
	ms->Merging = B_FALSE;
	ms->ms_ReclaimMem = reclaim;
	/* The above could have cleared 'Merging' prematurely, if this call
	   to memAlloc was actually from the above normal merge case, but
	   it will not matter, since both this case and the above case will
	   just return right away. */
	return mp;
    }

    /* no luck - have to get more from the system. */

    pieceLen = len + (sizeof(MemBlock_t) + sizeof(MemPiece_t));
    if (pieceLen < BLOCK_MINIMUM) {
	pieceLen = BLOCK_MINIMUM;
    } else {
	pieceLen += (BLOCK_STRIDE - 1);
	pieceLen &= (-BLOCK_STRIDE);
    }

#if DEBUG_SOFT || DEBUG_HARD
    debugX(ms, "Requesting ", pieceLen, " from system\n");
#endif
    mb = ms->ms_AllocMem(pieceLen);
    if (mb == P_NULL) {
	if (ms->Panicing) {
	    /* note that this is the only time that we can return a nil
	       pointer, and it is to the recursive calls just below */
	    return mb;
	}
	/* We could be in trouble folks. Try what we can. */
	if (ms->ms_ReclaimMem != P_NULL) {
	    reclaim = ms->ms_ReclaimMem;
	    while (ms->ms_ReclaimMem(B_TRUE)) {
	    }
	    ms->ms_ReclaimMem = reclaim;
	    ms->MemoryReclaims += 1;
	}
	memMerge(ms);
	/* we now should have back just about everything that we can free
	   up without doing disk I/O */
	ms->Panicing = B_TRUE;
	ms->MemoryPanics += 1;
	ms->MemoryAllocs -= 1;
	mp = MEM_ALLOC(ms, len, what);
	ms->Panicing = B_FALSE;
	if (mp != P_NULL) {
	    return mp;
	}
	if (ms->ms_RetryMem != P_NULL) {
	    /* RetryMem will ask the user to free up some external memory */
	    while (mb == P_NULL && ms->ms_RetryMem()) {
		mb = ms->ms_AllocMem(pieceLen);
	    }
	}
	if (mb == P_NULL) {
	    /* do this right away, so we have better chances of actually
	       getting the error message out! */
	    memFreeBlocks(ms);
	    MEM_ABORT1(ms, "out of memory", what);
	}
    }

    ms->MemoryAllocated += pieceLen;
    mb->mb_next = ms->MemBlocks;
    mb->mb_length = pieceLen;
    ms->MemBlocks = mb;

    mp2 = (MemPiece_t *) (mb + 1);
    mp = (MemPiece_t *) ((BYTE_T *) mb + pieceLen - sizeof(MemPiece_t));
    mp->mp_.mp_length = 0;
    mp->mp_.mp_tag[TAG_INDEX] = TAG_END;
#if TAG_ALLOCS
    mp->mp_what = "FREE";
    mp2->mp_what = "NEW ";
#endif
    pieceLen -= (sizeof(MemBlock_t) + sizeof(MemPiece_t));
    mp2->mp_.mp_length = pieceLen;
#if ENABLE_VALIDATE || CHECK
    mp2->mp_.mp_tag[TAG_INDEX] = TAG_ALLOCATED;
#endif
#if DEBUG_HARD
    errorString(ms, "freeing all of requested block\n");
#endif
    mp2 = (MemPiece_t *) ((BYTE_T *) mp2 + MEM_EXTRA);
    ms->MemoryInUse += pieceLen;
    pieceLen -= MEM_EXTRA;
    ms->MemoryFrees -= 1;
    memFree(ms, mp2, pieceLen);
#if DEBUG_HARD
    errorString(ms, "recursing in memAlloc after getting block\n");
#endif
    ms->MemoryAllocs -= 1;
    return MEM_ALLOC(ms, len, what);
}

#if TAG_ALLOCS

/*
 * memAllocET - external interface to memAlloc
 */

void *
memAllocET(MUDState_t *ms, ULONG_T len, char *what)
{

    return memAllocT(ms, len, what);
}

#else

/*
 * memAllocE - external interface to memAlloc
 */

void *
memAllocE(MUDState_t *ms, ULONG_T len)
{

    return memAlloc(ms, len);
}

#endif

/*
 * memRecoverE - attempt to get our total memory usage under what we
 *	were told to try to stay under.
 */

void
memRecoverE(MUDState_t *ms)
{

    if (ms->ms_RecoverMem != P_NULL && ms->MemoryAllocated > ms->MemoryLimit) {
	while (ms->ms_RecoverMem() && ms->MemoryInUse > ms->MemoryLimit) {
	}
	ms->MemoryReclaims += 1;
	memMerge(ms);
    }
}

/*
 * memInit - initialization code for memory allocation structures.
 */

void
memInit(MUDState_t *ms)
{
    ULONG_T firstGrab;
    UINT_T i;

    for (i = 0; i < FIXED_COUNT; i += 1) {
	ms->ShortList[i] = P_NULL;
    }
    ms->LongList = P_NULL;
    ms->MemBlocks = P_NULL;
    ms->MemoryAllocated = 0;
    ms->MemoryInUse = 0;
    ms->MemoryReclaims = 0;
    ms->MemoryPanics = 0;
    ms->MemoryAllocs = 0;
    ms->MemoryFrees = 0;
    ms->Merging = B_FALSE;
    ms->Panicing = B_FALSE;

    ms->WriteBufferSize = DEFAULT_BUFFER_SIZE;
    ms->WriteBuffer = MEM_ALLOC(ms, DEFAULT_BUFFER_SIZE, "WBUF");

    if (ms->ms_incRef != P_NULL) {
	/* we are running in the server */
	firstGrab = ms->MemoryLimit;
	if (firstGrab > 100 * 1024) {
	    firstGrab = 100 * 1024;
	}
	firstGrab /= 2;
	memFree(ms, MEM_ALLOC(ms, firstGrab, "FRST"), firstGrab);
    }
}

/*
 * memTerm - terminate the memory allocation structures.
 */

void
memTerm(MUDState_t *ms)
{

#if ENABLE_VALIDATE
    memValidate(ms, "memTerm");
#endif
#if TAG_ALLOCS
#if ENABLE_SHOW && SHOW_ON_EXIT
    showBlocks(ms, "termination");
#endif
#endif
    memFreeBlocks(ms);

    if (ms->ms_incRef != P_NULL) {
	/* running in server - errorString goes to log file */
	errorString(ms, "Allocator:  used: ");
	errorDecimal(ms, ms->MemoryAllocated);
	errorString(ms, "  limit: ");
	errorDecimal(ms, ms->MemoryLimit);
	ms->ms_errorChar('\n');
	errorString(ms, "allocs: ");
	errorDecimal(ms, ms->MemoryAllocs);
	errorString(ms, "  frees: ");
	errorDecimal(ms, ms->MemoryFrees);
	errorString(ms, "  reclaims: ");
	errorDecimal(ms, ms->MemoryReclaims);
	errorString(ms, "  panics: ");
	errorDecimal(ms, ms->MemoryPanics);
	errorString(ms, "\n");
    }
}
