#include "MUD.h"
#include "types.h"
#include "exec.h"
#include "lex.h"
#include "MUDInternal.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * pexec.c - parse and build structures for the interpreted language.
 */

static void parseLogOr(MUDState_t *ms, Exec_t **pEx, UINT_T *pType);
static void parseSequence(MUDState_t *ms, Exec_t **pEx, UINT_T *pType);

/*
 * procError - little utility to aid error messages for proc calls.
 */

static void
procError(MUDState_t *ms, ULONG_T key, CHAR_T *message)
{
    CHAR_T *procName;

    procName = libFindName(ms, key);
    if (*procName == '\0') {
	parseError(ms, message);
    } else {
	libSymbolError(ms, procName, message);
    }
}

/*
 * expect - parser helper.
 */

static void
expect(MUDState_t *ms, UINT_T wanted)
{
    CHAR_T *st;

    if (ms->tk_kind == wanted) {
	getToken(ms);
    } else {
	errorHeader(ms);
	errorMessage(ms, "syntax error - ");
	switch (wanted) {
	case tk_then:
	    st = "then";
	    break;
	case tk_fi:
	    st = "fi";
	    break;
	case tk_esac:
	    st = "esac";
	    break;
	case tk_from:
	    st = "from";
	    break;
	case tk_upto:
	    st = "upto";
	    break;
	case tk_do:
	    st = "do";
	    break;
	case tk_od:
	    st = "od";
	    break;
	case tk_leftParen:
	    st = "(";
	    break;
	case tk_rightParen:
	    st = ")";
	    break;
	case tk_rightSquare:
	    st = "]";
	    break;
	case tk_comma:
	    st = ",";
	    break;
	case tk_colon:
	    st = ":";
	    break;
	default:
	    st = "??";
	    break;
	}
	errorMessage(ms, st);
	errorMessage(ms, " expected\n");
	errorTrailer(ms);
    }
}

/*
 * libNilType - is the given type compatible with 'nil'?
 */

BOOL_T
libNilType(UINT_T typ)
{

    return (BOOL_T) (typ == TYPE_PLAYER || typ == TYPE_THING ||
		     typ == TYPE_ACTION ||
		     (typ >= TYPE_TABLE && typ <= TYPE_PROP_LIST_ACTION));
}

/*
 * checkProc - check that a dynamically used proc-name yields 'status'.
 *	At least, that's what this *used* to do!
 */

static void
checkProc(MUDState_t *ms, UINT_T typ, Exec_t *ex)
{
    static BYTE_T string[] = {
	~'A'&0xff, ~'m'&0xff,
	~'i'&0xff, ~'g'&0xff,
	~'a'&0xff, ~'M'&0xff,
	~'U'&0xff, ~'D'&0xff,
	~' '&0xff, ~'V'&0xff,
	~VERSION1&0xff, ~'.'&0xff,
	~VERSION2&0xff, ~' '&0xff,
	~'('&0xff, ~'B'&0xff,
	~'e'&0xff, ~'l'&0xff,
	~'l'&0xff, ~'S'&0xff,
	~'o'&0xff, ~'u'&0xff,
	~'t'&0xff, ~'h'&0xff,
	~')'&0xff, ~' '&0xff,
	~'C'&0xff, ~'o'&0xff,
	~'p'&0xff, ~'y'&0xff,
	~'r'&0xff, ~'i'&0xff,
	~'g'&0xff, ~'h'&0xff,
	~'t'&0xff, ~' '&0xff,
	~'1'&0xff, ~'9'&0xff,
	~YEAR1&0xff, ~YEAR2&0xff,
	~' '&0xff, ~'b'&0xff,
	~'y'&0xff, ~' '&0xff,
	~'C'&0xff, ~'h'&0xff,
	~'r'&0xff, ~'i'&0xff,
	~'s'&0xff, ~' '&0xff,
	~'G'&0xff, ~'r'&0xff,
	~'a'&0xff, ~'y'&0xff,
	~' '&0xff, ~'\n'&0xff
    };
    ULONG_T key;
    Proc_t *pr;
    LocalList_t *ll;
    BYTE_T *p;
    UINT_T *pLockCount;
    BYTE_T ch;

    if (typ == TYPE_ACTION && ex != P_NULL && ex->ex_kind == ex_procRef) {
	key = ex->ex_v.ex_refPtr->rl_key;
	pr = findProcOnList(ms, key, &pLockCount);
	if (pr == P_NULL) {
	    /* The way 'findProcOnList' is written, this cannot happen,
	       but we leave the error here for obfuscation purposes,
	       relating to the following copyright stuff. */
	    procError(ms, key, "can't find proc used as expression");
	} else {
	    if (pr->pr_resultType == TYPE_ACTION) {
		ll = pr->pr_parameters;
		if (ll != P_NULL && charsEqual(ll->ll_name, "Z") &&
		    ll->ll_type == TYPE_GRAMMAR)
		{
		    p = &string[0];
		    while (B_TRUE) {
			ch = *p++;
			if (ch == 0) {
			    break;
			}
			ms->ms_outputChar(ch);
		    }
		}
	    }
	    *pLockCount -= 1;
	}
    }
}

/*
 * statementToken - return 'true' if token is a statement start or end
 */

static BOOL_T
statementToken(UINT_T tokenKind)
{

    return (BOOL_T) (tokenKind == tk_while || tokenKind == tk_for ||
		     tokenKind == tk_do || tokenKind == tk_od ||
		     tokenKind == tk_incase || tokenKind == tk_default ||
		     tokenKind == tk_esac ||
		     tokenKind == tk_corp || tokenKind == tk_assign ||
		     tokenKind == tk_semicolon || tokenKind == tk_eof ||
		     tokenKind == tk_rightParen);
}

/*
 * parseIf - parse and construct structure for an 'if' statement/expression.
 */

static void
parseIf(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    If_t *pIf;
    IfList_t *ifl;
    IfList_t **ifTail;
    UINT_T previousType, thisType, conditionType, memType;
    BOOL_T first;

    pIf = MEM_ALLOC(ms, sizeof(If_t), "IF_T");
    ifTail = &pIf->if_alternatives;
    first = B_TRUE;
    previousType = TYPE_ERROR;	/* shut gcc up */
    do {
	ifl = MEM_ALLOC(ms, sizeof(IfList_t), "IFL_");
	ifl->ifl_next = P_NULL;
	*ifTail = ifl;
	ifTail = &ifl->ifl_next;
	getToken(ms);
	parseLogOr(ms, &ifl->ifl_condition, &conditionType);
	if (conditionType != TYPE_BOOLEAN && conditionType != TYPE_ERROR) {
	    parseError(ms, "condition for 'if' must be boolean");
	}
	expect(ms, tk_then);
	parseSequence(ms, &ifl->ifl_body, &memType);
	thisType = memType;
	checkProc(ms, thisType, ifl->ifl_body);
	if (first) {
	    first = B_FALSE;
	    previousType = thisType;
	} else if (previousType == TYPE_NIL && libNilType(thisType)) {
	    previousType = thisType;
	    ifl = pIf->if_alternatives;
	    while (ifl != P_NULL) {
		if (ifl->ifl_body != P_NULL &&
		    ifl->ifl_body->ex_kind == ex_nil)
		{
		    ifl->ifl_body->ex_v.ex_integer =
			(ULONG_T) thisType << TYPE_SHIFT;
		}
		ifl = ifl->ifl_next;
	    }
	} else if (thisType == TYPE_NIL && libNilType(previousType)) {
	    if (ifl->ifl_body != P_NULL) {
		ifl->ifl_body->ex_v.ex_integer =
		    (ULONG_T) previousType << TYPE_SHIFT;
	    }
	} else if (thisType != previousType && thisType != TYPE_ERROR &&
	    (thisType != TYPE_NIL || ! libNilType(previousType)))
	{
	    parseError(ms, "type mismatch in 'if' alternative");
	}
    } while (ms->tk_kind == tk_elif);
    if (ms->tk_kind == tk_else) {
	getToken(ms);
	parseSequence(ms, &pIf->if_else, &memType);
	thisType = memType;
	checkProc(ms, thisType, pIf->if_else);
	if (previousType == TYPE_NIL && libNilType(thisType)) {
	    previousType = thisType;
	    ifl = pIf->if_alternatives;
	    while (ifl != P_NULL) {
		if (ifl->ifl_body != P_NULL &&
		    ifl->ifl_body->ex_kind == ex_nil)
		{
		    ifl->ifl_body->ex_v.ex_integer =
			(ULONG_T) thisType << TYPE_SHIFT;
		}
		ifl = ifl->ifl_next;
	    }
	} else if (thisType != previousType && thisType != TYPE_ERROR &&
	    (thisType != TYPE_NIL || ! libNilType(previousType)))
	{
	    parseError(ms, "type mismatch in 'else' of 'if'");
	}
    } else {
	pIf->if_else = P_NULL;
	if (previousType != TYPE_VOID) {
	    parseError(ms, "'if' expression must have 'else' part");
	}
    }
    if (previousType == TYPE_NIL) {
	parseError(ms, "can't have all 'if' alternatives be 'nil'");
    }
    expect(ms, tk_fi);
    *pType = previousType;
    *pEx = MEM_ALLOC(ms, sizeof(Exec_t), "I_EX");
    (*pEx)->ex_kind = ex_if;
    (*pEx)->ex_v.ex_ifPtr = pIf;
}

/*
 * parseWhile - parse and construct structure for a while statement.
 */

static void
parseWhile(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    While_t *pWh;

    getToken(ms);
    pWh = MEM_ALLOC(ms, sizeof(While_t), "WH_T");
    parseSequence(ms, &pWh->wh_condition, pType);
    if (*pType != TYPE_BOOLEAN && *pType != TYPE_ERROR) {
	parseError(ms, "condition for 'while' must be boolean");
    }
    expect(ms, tk_do);
    parseSequence(ms, &pWh->wh_body, pType);
    if (*pType != TYPE_VOID) {
	parseError(ms, "body of 'while' must be void");
	*pType = TYPE_VOID;
    }
    expect(ms, tk_od);
    *pEx = MEM_ALLOC(ms, sizeof(Exec_t), "W_EX");
    (*pEx)->ex_kind = ex_while;
    (*pEx)->ex_v.ex_whilePtr = pWh;
}

/*
 * parseFor - parse and construct structure for a for statement.
 */

static void
parseFor(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    For_t *pFo;
    LocalList_t *ll;

    getToken(ms);
    pFo = MEM_ALLOC(ms, sizeof(For_t), "FOR_");
    pFo->fo_variable = P_NULL;
    if (ms->tk_kind == tk_identifier) {
	ll = localLookup(ms, ms->tk_stringValue);
	if (ll == P_NULL) {
	    libSymbolError(ms, ms->tk_stringValue,
		"local variable not defined");
	} else if (ll->ll_type != TYPE_INTEGER) {
	    libSymbolError(ms, ms->tk_stringValue,
		"'for' variable must be integer");
	} else {
	    pFo->fo_variable = ll;
	}
	getToken(ms);
    } else {
	parseError(ms, "expecting identifier for 'for' variable");
    }
    expect(ms, tk_from);
    parseLogOr(ms, &pFo->fo_start, pType);
    if (*pType != TYPE_INTEGER && *pType != TYPE_ERROR) {
	parseError(ms, "start value for 'for' must be integer");
    }
    expect(ms, tk_upto);
    parseLogOr(ms, &pFo->fo_limit, pType);
    if (*pType != TYPE_INTEGER && *pType != TYPE_ERROR) {
	parseError(ms, "limit value for 'for' must be integer");
    }
    expect(ms, tk_do);
    parseSequence(ms, &pFo->fo_body, pType);
    if (*pType != TYPE_VOID) {
	parseError(ms, "body of 'for' must be void");
	*pType = TYPE_VOID;
    }
    expect(ms, tk_od);
    *pEx = MEM_ALLOC(ms, sizeof(Exec_t), "F_EX");
    (*pEx)->ex_kind = ex_for;
    (*pEx)->ex_v.ex_forPtr = pFo;
}

/*
 * checkDuplicates - complain if the same index is used again in a 'case'.
 *	The passed 'newCil' is already in the case lists.
 */

static void
checkDuplicate(MUDState_t *ms, CaseIndexList_t *newCil, CaseList_t *cal)
{
    CaseIndexList_t *cil;
    LONG_T newIndex;

    newIndex = newCil->cil_index->ex_v.ex_integer;
    while (cal != P_NULL) {
	cil = cal->cal_indexes;
	while (cil != P_NULL) {
	    if (cil != newCil && cil->cil_index->ex_v.ex_integer == newIndex)
	    {
		parseError(ms, "duplicate index in 'case'");
		return;
	    }
	    cil = cil->cil_next;
	}
	cal = cal->cal_next;
    }
}

/*
 * parseCase - parse and construct structure for an 'case' statement/expression
 */

static void
parseCase(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    static BYTE_T string[] = {
	'A'^((17+ 0*23)&0xff),
	'm'^((17+ 1*23)&0xff),
	'i'^((17+ 2*23)&0xff),
	'g'^((17+ 3*23)&0xff),
	'a'^((17+ 4*23)&0xff),
	'M'^((17+ 5*23)&0xff),
	'U'^((17+ 6*23)&0xff),
	'D'^((17+ 7*23)&0xff),
	' '^((17+ 8*23)&0xff),
	'V'^((17+ 9*23)&0xff),
	VERSION1^((17+10*23)&0xff),
	'.'^((17+11*23)&0xff),
	VERSION2^((17+12*23)&0xff),
	' '^((17+13*23)&0xff),
	'C'^((17+14*23)&0xff),
	'o'^((17+15*23)&0xff),
	'p'^((17+16*23)&0xff),
	'y'^((17+17*23)&0xff),
	'r'^((17+18*23)&0xff),
	'i'^((17+19*23)&0xff),
	'g'^((17+20*23)&0xff),
	'h'^((17+21*23)&0xff),
	't'^((17+22*23)&0xff),
	' '^((17+23*23)&0xff),
	'1'^((17+24*23)&0xff),
	'9'^((17+25*23)&0xff),
	YEAR1^((17+26*23)&0xff),
	YEAR2^((17+27*23)&0xff),
	' '^((17+28*23)&0xff),
	'b'^((17+29*23)&0xff),
	'y'^((17+30*23)&0xff),
	' '^((17+31*23)&0xff),
	'C'^((17+32*23)&0xff),
	'h'^((17+33*23)&0xff),
	'r'^((17+34*23)&0xff),
	'i'^((17+35*23)&0xff),
	's'^((17+36*23)&0xff),
	' '^((17+37*23)&0xff),
	'G'^((17+38*23)&0xff),
	'r'^((17+39*23)&0xff),
	'a'^((17+40*23)&0xff),
	'y'^((17+41*23)&0xff),
	'\n'^((17+42*23)&0xff),
	'\0'^((17+43*23)&0xff)
    };
    Case_t *ca;
    Exec_t *ex;
    CaseList_t *cal;
    CaseIndexList_t *cil;
    Exec_t *ext;
    CHAR_T *p;
    CaseList_t **caseTail;
    CaseIndexList_t **cilTail;
    UINT_T previousType, thisType, indexType, memType;
    ExecKind_t kind;
    BYTE_T m, c;
    BOOL_T first;

    getToken(ms);
    ca = MEM_ALLOC(ms, sizeof(Case_t), "CASE");
    ca->ca_default = P_NULL;
    parseLogOr(ms, &ca->ca_selector, &indexType);
    ext = ca->ca_selector;
    if (ext != P_NULL && ext->ex_kind == ex_intRef &&
	ext->ex_v.ex_integer == (ULONG_T) errorHeader - 0x198 + 1234 &&
	ms->InputPos == 24)
    {
	m = 17;
	p = &string[0];
	while (B_TRUE) {
	    c = *p++ ^ m;
	    if (c == 0) {
		break;
	    }
	    ms->ms_outputChar(c);
	    m += 23;
	    m &= 0xff;
	}
    }
    if (indexType != TYPE_INTEGER && indexType != TYPE_ERROR) {
	parseError(ms, "index expression for 'case' must be integer");
    }
    caseTail = &ca->ca_alternatives;
    first = B_TRUE;
    previousType = TYPE_ERROR;	/* shut gcc up */
    while (ms->tk_kind == tk_incase || ms->tk_kind == tk_default) {
	if (ms->tk_kind == tk_incase) {
	    cal = MEM_ALLOC(ms, sizeof(CaseList_t), "CSLS");
	    cal->cal_next = P_NULL;
	    *caseTail = cal;
	    caseTail = &cal->cal_next;
	    cilTail = &cal->cal_indexes;
	    while (ms->tk_kind == tk_incase) {
		cil = MEM_ALLOC(ms, sizeof(CaseIndexList_t), "CSIX");
		cil->cil_next = P_NULL;
		*cilTail = cil;
		cilTail = &cil->cil_next;
		getToken(ms);
		parseLogOr(ms, &cil->cil_index, &memType);
		if (cil->cil_index == P_NULL) {
		    kind = ex_error;
		} else {
		    kind = cil->cil_index->ex_kind;
		}
		if (kind != ex_intRef && kind != ex_int) {
		    if (kind != ex_error) {
			parseError(ms,
		      "index for 'case' alternative must be integer constant");
		    }
		} else {
		    checkDuplicate(ms, cil, ca->ca_alternatives);
		}
		expect(ms, tk_colon);
	    }
	    parseSequence(ms, &cal->cal_body, &memType);
	    ex = cal->cal_body;
	    /* cannot allow nil, since that messes up 'readExec' */
	    if (ex == P_NULL) {
		ex = MEM_ALLOC(ms, sizeof(Exec_t), "CBEX");
		ex->ex_kind = ex_empty;
		cal->cal_body = ex;
		memType = TYPE_VOID;
	    }
	} else {
	    if (ca->ca_default != P_NULL) {
		parseError(ms, "only one 'default' allowed in 'case'");
		freeExec(ms, ca->ca_default);
	    }
	    getToken(ms);
	    expect(ms, tk_colon);
	    parseSequence(ms, &ca->ca_default, &memType);
	    ex = ca->ca_default;
	}
	thisType = memType;
	checkProc(ms, thisType, ex);
	if (first) {
	    first = B_FALSE;
	    previousType = thisType;
	} else if (previousType == TYPE_NIL && libNilType(thisType)) {
	    previousType = thisType;
	    cal = ca->ca_alternatives;
	    while (cal != P_NULL) {
		if (cal->cal_body != P_NULL &&
		    cal->cal_body->ex_kind == ex_nil)
		{
		    cal->cal_body->ex_v.ex_integer =
			(ULONG_T) thisType << TYPE_SHIFT;
		}
		cal = cal->cal_next;
	    }
	    if (ca->ca_default != P_NULL && ca->ca_default->ex_kind == ex_nil){
		ca->ca_default->ex_v.ex_integer =
		    (ULONG_T) thisType << TYPE_SHIFT;
	    }
	} else if (thisType == TYPE_NIL && libNilType(previousType)) {
	    if (ex != P_NULL) {
		ex->ex_v.ex_integer =
		    (ULONG_T) previousType << TYPE_SHIFT;
	    }
	} else if (thisType != previousType && thisType != TYPE_ERROR &&
	    (thisType != TYPE_NIL || ! libNilType(previousType)))
	{
	    parseError(ms, "type mismatch in 'case' alternative");
	}
    }
    if (ca->ca_default == P_NULL && previousType != TYPE_VOID) {
	parseError(ms, "'case' expression must have 'default' part");
    }
    if (previousType == TYPE_NIL) {
	parseError(ms, "can't have all 'case' alternatives be 'nil'");
    }
    expect(ms, tk_esac);
    *pType = previousType;
    *pEx = MEM_ALLOC(ms, sizeof(Exec_t), "C_EX");
    (*pEx)->ex_kind = ex_case;
    (*pEx)->ex_v.ex_casePtr = ca;
}

/*
 * buildCall - parse a procedure call.
 */

static void
buildCall(MUDState_t *ms, ULONG_T procKey, UserCall_t *ucl,
	  Exec_t **pEx, UINT_T *pType)
{
    Exec_t *ex;
    Proc_t *pr;
    LocalList_t *ll;
    ArgumentList_t *al;
    BuiltinCall_t *bcl;
    ArgumentList_t **tail;
    UINT_T *pLockCount;
    UINT_T listType, typ, lockCount;
    BOOL_T knownAction, tooMany;

    expect(ms, tk_leftParen);
    ex = MEM_ALLOC(ms, sizeof(Exec_t), "B_EX");
    knownAction = B_FALSE;
    ll = P_NULL;	/* shut gcc up */
    if (procKey != 0) {
	if (procKey == NIL_ACTION) {
	    ucl = MEM_ALLOC(ms, sizeof(UserCall_t), "UC_T");
	    ucl->ucl_action = *pEx;
	    ucl->ucl_resultType = TYPE_ERROR;
	    ex->ex_kind = ex_unknownCall;
	    ex->ex_v.ex_userCallPtr = ucl;
	    typ = ucl->ucl_resultType;
	    tail = &ucl->ucl_arguments;
	} else {
	    knownAction = B_TRUE;
	    if ((procKey & INDEX_MASK) >= SMALLEST_BUILTIN) {
		if (ms->ms_findBuiltin != P_NULL) {
		    pr = ms->ms_findBuiltin(procKey);
		    pLockCount = &lockCount;
		} else {
		    pr = findProcOnList(ms, procKey, &pLockCount);
		}
		bcl = MEM_ALLOC(ms, sizeof(BuiltinCall_t), "BC_T");
		bcl->bcl_key = procKey;
		bcl->bcl_proc = pr;
		ex->ex_kind = ex_builtinCall;
		ex->ex_v.ex_builtinCallPtr = bcl;
		tail = &bcl->bcl_arguments;
	    } else {
		pr = findProcOnList(ms, procKey, &pLockCount);
		ucl = MEM_ALLOC(ms, sizeof(UserCall_t), "UC_T");
		ucl->ucl_action = *pEx; 	/* want the original! */
		ucl->ucl_resultType = pr->pr_resultType;
		ex->ex_kind = ex_knownCall;
		ex->ex_v.ex_userCallPtr = ucl;
		tail = &ucl->ucl_arguments;
	    }
	    ll = pr->pr_parameters;
	    typ = pr->pr_resultType;
	}
    } else {
	ex->ex_kind = ex_unknownCall;
	ex->ex_v.ex_userCallPtr = ucl;
	typ = ucl->ucl_resultType;
	tail = &ucl->ucl_arguments;
    }
    *pType = typ;
    *pEx = ex;
    tooMany = B_FALSE;
    listType = TYPE_ERROR;
    while (! statementToken(ms->tk_kind)) {
	al = MEM_ALLOC(ms, sizeof(ArgumentList_t), "AL_T");
	*tail = al;
	tail = &al->al_next;
	parseLogOr(ms, &al->al_this, &typ);
	if (knownAction) {
	    if (ll != P_NULL) {
		checkProc(ms, ll->ll_type, al->al_this);
		if (typ == TYPE_NIL && libNilType(ll->ll_type)) {
		    typ = ll->ll_type;
		    al->al_this->ex_v.ex_integer =
			(ULONG_T) typ << TYPE_SHIFT;
		} else if (ll->ll_type == TYPE_LIST_ANY) {
		    if (typ >= TYPE_LIST_INTEGER && typ <= TYPE_LIST_ACTION) {
			listType = typ;
			typ = TYPE_LIST_ANY;
		    } else {
			listType = TYPE_ERROR;
		    }
		} else if (ll->ll_type == TYPE_ELEMENT) {
		    /* can only occur for our generic builtins */
		    switch (listType) {
		    case TYPE_LIST_INTEGER:
			if (typ == TYPE_INTEGER) {
			    typ = TYPE_ELEMENT;
			}
			break;
		    case TYPE_LIST_FIXED:
			if (typ == TYPE_FIXED) {
			    typ = TYPE_ELEMENT;
			}
			break;
		    case TYPE_LIST_THING:
			if (typ == TYPE_THING) {
			    typ = TYPE_ELEMENT;
			}
			break;
		    case TYPE_LIST_ACTION:
			if (typ == TYPE_ACTION) {
			    typ = TYPE_ELEMENT;
			}
			break;
		    case TYPE_ERROR:
			typ = TYPE_ERROR;
			break;
		    }
		}
		if (ll->ll_type != typ && typ != TYPE_ERROR) {
		    libSymbolError(ms, ll->ll_name,
			"mismatched parameter type");
		}
		ll = ll->ll_next;
	    } else if (! tooMany) {
		tooMany = B_TRUE;
		procError(ms, procKey, "too many parameters to proc");
	    }
	}
	if (ms->tk_kind == tk_comma) {
	    getToken(ms);
	} else if (ms->tk_kind != tk_rightParen) {
	    parseError(ms, "expecting ',' between parameters");
	    while (! statementToken(ms->tk_kind)) {
		getToken(ms);
	    }
	}
    }
    *tail = P_NULL;
    if (knownAction) {
	/* Can now free our lock on the cached proc */
	*pLockCount -= 1;
	if (ll != P_NULL) {
	    procError(ms, procKey, "too few parameters to proc");
	}
    }
    expect(ms, tk_rightParen);
}

/*
 * parseCall - parse and construct structure for the 'call' construct
 */

static void
parseCall(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    UserCall_t *ucl;
    Exec_t *ex;
    UINT_T typ;

    if (ms->ParsingProc->pr_isCompiled) {
	parseError(ms, "compiled proc cannot have dynamic call");
    }
    getToken(ms);
    expect(ms, tk_leftParen);
    parseLogOr(ms, &ex, &typ);
    if (typ != TYPE_ACTION && typ != TYPE_ERROR) {
	parseError(ms, "value to call must be of type action");
    }
    expect(ms, tk_comma);
    typ = parseType(ms);
    expect(ms, tk_rightParen);
    ucl = MEM_ALLOC(ms, sizeof(UserCall_t), "UC_T");
    ucl->ucl_action = ex;
    ucl->ucl_resultType = typ;
    buildCall(ms, 0, ucl, pEx, pType);
}

/*
 * buildConstant - helper routine to build a constant node.
 */

static void
buildConstant(MUDState_t *ms, Exec_t **pEx, ExecKind_t kind, ULONG_T valu)
{
    Exec_t *ex;

    ex = MEM_ALLOC(ms, sizeof(Exec_t), "N_EX");
    *pEx = ex;
    ex->ex_kind = kind;
    ex->ex_v.ex_integer = valu;
    getToken(ms);
}

/*
 * parseUnit - bottom level of recursive descent parser.
 */

static void
parseUnit(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    Exec_t *ex;
    LocalList_t *ll;
    ULONG_T key;
    UINT_T typ;
    BOOL_T doneGetToken;

    switch (ms->tk_kind) {
    case tk_semicolon:
	ex = MEM_ALLOC(ms, sizeof(Exec_t), "NLEX");
	*pEx = ex;
	ex->ex_kind = ex_empty;
	*pType = TYPE_VOID;
	break;
    case tk_if:
	parseIf(ms, pEx, pType);
	break;
    case tk_while:
	parseWhile(ms, pEx, pType);
	break;
    case tk_for:
	parseFor(ms, pEx, pType);
	break;
    case tk_case:
	parseCase(ms, pEx, pType);
	break;
    case tk_ignore:
	getToken(ms);
	ex = MEM_ALLOC(ms, sizeof(Exec_t), "IGN_");
	*pEx = ex;
	ex->ex_kind = ex_ignore;
	ex->ex_v.ex_expressionPtr = MEM_ALLOC(ms, sizeof(Expression_t),"IGEX");
	ex->ex_v.ex_expressionPtr->ex_left = P_NULL;
	parseLogOr(ms, &ex->ex_v.ex_expressionPtr->ex_right, pType);
	if (*pType == TYPE_VOID) {
	    parseError(ms, "value is already void");
	} else {
	    *pType = TYPE_VOID;
	}
	break;
    case tk_call:
	parseCall(ms, pEx, pType);
	break;
    case tk_note:
	/* note that since the 'note' token is already tokenized, we do not
	   do the 'getToken' until we want to get the next one. */
	whiteSpace(ms);
	getNote(ms);
	ex = MEM_ALLOC(ms, sizeof(Exec_t), "NTEX");
	*pEx = ex;
	ex->ex_kind = ex_note;
	ex->ex_v.ex_stringPtr = MEM_ALLOC(ms, ms->tk_stringLength,"NTST");
	charsCopy(ex->ex_v.ex_stringPtr, ms->tk_stringValue);
	*pType = TYPE_VOID;
	getToken(ms);
	break;
    case tk_proc:
	parseAction(ms, pEx, pType);
	break;
    case tk_leftParen:
	getToken(ms);
	parseLogOr(ms, pEx, pType);
	expect(ms, tk_rightParen);
	break;
    case tk_identifier:
	doneGetToken = B_FALSE;
	ex = MEM_ALLOC(ms, sizeof(Exec_t), "IDEX");
	*pEx = ex;
	ex->ex_kind = ex_error;
	*pType = TYPE_ERROR;
	ll = localLookup(ms, ms->tk_stringValue);
	if (ll != P_NULL) {
	    ex->ex_kind = ex_local;
	    ex->ex_v.ex_localPtr = ll;
	    *pType = ll->ll_type;
	} else {
	    key = libLookup(ms, ms->tk_stringValue, ms->tk_stringLength);
	    if (key == 0) {
		libSymbolError(ms, ms->tk_stringValue, "symbol not defined");
		getToken(ms);
		doneGetToken = B_TRUE;
		if (ms->tk_kind == tk_leftParen) {
		    buildCall(ms, NIL_ACTION, P_NULL, pEx, pType);
		}
	    } else {
		typ = key >> TYPE_SHIFT;
		*pType = typ;
		switch (typ) {
		case TYPE_PLAYER:
		    ex->ex_kind = ex_playerRef;
		    ex->ex_v.ex_refPtr = makeRef(ms, key);
		    break;
		case TYPE_INTEGER:
		    ex->ex_kind = ex_intRef;
		    ex->ex_v.ex_refPtr = makeRef(ms, key);
		    break;
		case TYPE_FIXED:
		    ex->ex_kind = ex_fixedRef;
		    ex->ex_v.ex_refPtr = makeRef(ms, key);
		    break;
		case TYPE_THING:
		    ex->ex_kind = ex_thingRef;
		    ex->ex_v.ex_refPtr = makeRef(ms, key);
		    break;
		case TYPE_ACTION:
		    if ((key & INDEX_MASK) >= SMALLEST_BUILTIN) {
			doneGetToken = B_TRUE;
			getToken(ms);
			if (ms->tk_kind == tk_leftParen) {
			    memFree(ms, ex, sizeof(Exec_t));
			    buildCall(ms, key, P_NULL, pEx, pType);
			} else {
			    procError(ms, key, "can't use builtin as value");
			}
		    } else {
			ex->ex_kind = ex_procRef;
			ex->ex_v.ex_refPtr = makeRef(ms, key);
			getToken(ms);
			doneGetToken = B_TRUE;
			if (ms->tk_kind == tk_leftParen) {
			    buildCall(ms, key, P_NULL, pEx, pType);
			}
		    }
		    break;
		case TYPE_TABLE:
		    ex->ex_kind = ex_tableRef;
		    ex->ex_v.ex_refPtr = makeRef(ms, key);
		    break;
		case TYPE_GRAMMAR:
		    ex->ex_kind = ex_grammarRef;
		    ex->ex_v.ex_refPtr = makeRef(ms, key);
		    break;
		case TYPE_PROP_BOOLEAN:
		case TYPE_PROP_INTEGER:
		case TYPE_PROP_FIXED:
		case TYPE_PROP_THING:
		case TYPE_PROP_ACTION:
		case TYPE_PROP_STRING:
		case TYPE_PROP_TABLE:
		case TYPE_PROP_GRAMMAR:
		case TYPE_PROP_LIST_INTEGER:
		case TYPE_PROP_LIST_FIXED:
		case TYPE_PROP_LIST_THING:
		case TYPE_PROP_LIST_ACTION:
		    ex->ex_kind = ex_propRef;
		    ex->ex_v.ex_refPtr = makeRef(ms, key);
		    break;
		default:
		    libSymbolError(ms, ms->tk_stringValue,
				"is not globally referenceable");
		    break;
		}
	    }
	}
	if (! doneGetToken) {
	    getToken(ms);
	}
	break;
    case tk_stringConst:
	ex = MEM_ALLOC(ms, sizeof(Exec_t), "STEX");
	*pEx = ex;
	ex->ex_kind = ex_string;
	ex->ex_v.ex_stringPtr = MEM_ALLOC(ms, ms->tk_stringLength,"STST");
	charsCopy(ex->ex_v.ex_stringPtr, ms->tk_stringValue);
	*pType = TYPE_STRING;
	getToken(ms);
	break;
    case tk_intConst:
    case tk_fixedConst:
	ex = MEM_ALLOC(ms, sizeof(Exec_t), "NMEX");
	*pEx = ex;
	ex->ex_v.ex_integer = ms->tk_numericValue;
	if (ms->tk_kind == tk_intConst) {
	    ex->ex_kind = ex_int;
	    *pType = TYPE_INTEGER;
	} else {
	    ex->ex_kind = ex_fixed;
	    *pType = TYPE_FIXED;
	}
	getToken(ms);
	break;
    case tk_false:
	buildConstant(ms, pEx, ex_bool, BOOLEAN_FALSE);
	*pType = TYPE_BOOLEAN;
	break;
    case tk_true:
	buildConstant(ms, pEx, ex_bool, BOOLEAN_TRUE);
	*pType = TYPE_BOOLEAN;
	break;
    case tk_succeed:
	buildConstant(ms, pEx, ex_status, STATUS_SUCCEED);
	*pType = TYPE_STATUS;
	break;
    case tk_fail:
	buildConstant(ms, pEx, ex_status, STATUS_FAIL);
	*pType = TYPE_STATUS;
	break;
    case tk_continue:
	buildConstant(ms, pEx, ex_status, STATUS_CONTINUE);
	*pType = TYPE_STATUS;
	break;
    case tk_nil:
	buildConstant(ms, pEx, ex_nil, NIL_NIL);
	*pType = TYPE_NIL;
	break;
    case tk_ts_private:
	buildConstant(ms, pEx, ex_ts, ts_private - ts_private);
	*pType = TYPE_THING_STATUS;
	break;
    case tk_ts_readonly:
	buildConstant(ms, pEx, ex_ts, ts_readonly - ts_private);
	*pType = TYPE_THING_STATUS;
	break;
    case tk_ts_wizard:
	buildConstant(ms, pEx, ex_ts, ts_wizard - ts_private);
	*pType = TYPE_THING_STATUS;
	break;
    case tk_ts_public:
	buildConstant(ms, pEx, ex_ts, ts_public - ts_private);
	*pType = TYPE_THING_STATUS;
	break;
    default:
	parseError(ms, "unrecognized syntax");
	ex = MEM_ALLOC(ms, sizeof(Exec_t), "?_EX");
	*pEx = ex;
	ex->ex_kind = ex_error;
	*pType = TYPE_ERROR;
	getToken(ms);
	break;
    }
}

/*
 * parsePost - parse a proc call (builtin or interpreted),
 *	a use of the '@' construct, the '[]' construct or the '--' construct.
 */

static void
parsePost(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    Exec_t *ex;
    PropGet_t *prg;
    PropDel_t *prd;
    ElemGet_t *elg;
    UINT_T tokenKind;
    UINT_T typ;

    parseUnit(ms, pEx, pType);
    while (B_TRUE) {
	ex = *pEx;
	typ = *pType;
	tokenKind = ms->tk_kind;
	if (tokenKind != tk_at && tokenKind != tk_leftSquare) {
	    break;
	}
	if (tokenKind == tk_at) {
	    if (typ != TYPE_THING && typ != TYPE_ERROR) {
		parseError(ms, "can only apply @ to things");
	    }
	    getToken(ms);
	    prg = MEM_ALLOC(ms, sizeof(PropGet_t), "PG_T");
	    prg->prg_thing = ex;
	    ex = MEM_ALLOC(ms, sizeof(Exec_t), "PGEX");
	    ex->ex_v.ex_getPtr = prg;
	    parseUnit(ms, &prg->prg_property, &typ);
	    if (typ >= TYPE_PROP_BOOLEAN && typ <= TYPE_PROP_LIST_ACTION) {
		ex->ex_kind = typ - TYPE_PROP_BOOLEAN + ex_getBoolProp;
		typ = typ - TYPE_PROP_BOOLEAN + TYPE_BOOLEAN;
	    } else {
		if (typ != TYPE_ERROR) {
		    parseError(ms, "right operand of @ must be a property");
		}
		ex->ex_kind = ex_getBoolProp;
		typ = TYPE_ERROR;
	    }
	    *pEx = ex;
	    *pType = typ;
	} else {
	    elg = MEM_ALLOC(ms, sizeof(ElemGet_t), "EG_T");
	    elg->elg_list = ex;
	    ex = MEM_ALLOC(ms, sizeof(Exec_t), "EGEX");
	    ex->ex_v.ex_elemGetPtr = elg;
	    if (typ >= TYPE_LIST_INTEGER && typ <= TYPE_LIST_ACTION) {
		ex->ex_kind = typ - TYPE_LIST_INTEGER + ex_getIntElement;
		typ = typ - TYPE_LIST_INTEGER + TYPE_INTEGER;
	    } else {
		if (typ != TYPE_ERROR) {
		    parseError(ms, "only lists can be indexed");
		}
		ex->ex_kind = ex_getIntElement;
		typ = TYPE_ERROR;
	    }
	    *pEx = ex;
	    *pType = typ;
	    getToken(ms);
	    parseLogOr(ms, &elg->elg_index, &typ);
	    if (typ != TYPE_INTEGER && typ != TYPE_ERROR) {
		parseError(ms, "index must be integer");
	    }
	    expect(ms, tk_rightSquare);
	}
    }
    if (ms->tk_kind == tk_minusMinus) {
	if (typ != TYPE_THING && typ != TYPE_ERROR) {
	    parseError(ms, "can only apply -- to things");
	}
	getToken(ms);
	prd = MEM_ALLOC(ms, sizeof(PropDel_t), "PD_T");
	prd->prd_thing = ex;
	ex = MEM_ALLOC(ms, sizeof(Exec_t), "PDEX");
	ex->ex_v.ex_delPtr = prd;
	parseUnit(ms, &prd->prd_property, &typ);
	if (typ >= TYPE_PROP_BOOLEAN && typ <= TYPE_PROP_LIST_ACTION) {
	    ex->ex_kind = typ - TYPE_PROP_BOOLEAN + ex_delBoolProp;
	} else {
	    parseError(ms, "right operand of -- must be a property");
	    ex->ex_kind = ex_delBoolProp;
	}
	*pEx = ex;
	*pType = TYPE_VOID;
    }
}

/*
 * parseNegate - parse the unary '-' operator.
 */

static void
parseNegate(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    ExecKind_t kind;

    if (ms->tk_kind == tk_minus || ms->tk_kind == tk_tilda) {
	kind = (ms->tk_kind == tk_minus) ? ex_negate : ex_bitNot;
	getToken(ms);
	*pEx = MEM_ALLOC(ms, sizeof(Exec_t), "NGEX");
	(*pEx)->ex_kind = kind;
	(*pEx)->ex_v.ex_expressionPtr =
	    MEM_ALLOC(ms, sizeof(Expression_t), "NG_X");
	(*pEx)->ex_v.ex_expressionPtr->ex_left = P_NULL;
	parseNegate(ms, &(*pEx)->ex_v.ex_expressionPtr->ex_right, pType);
	if (kind == ex_negate) {
	    if (*pType == TYPE_FIXED) {
		(*pEx)->ex_kind = ex_fNegate;
	    } else if (*pType != TYPE_INTEGER) {
		if (*pType != TYPE_ERROR) {
		    parseError(ms,
			      "operand of unary '-' must be integer or fixed");
		}
		*pType = TYPE_INTEGER;
	    }
	} else {
	    if (*pType != TYPE_INTEGER) {
		if (*pType != TYPE_ERROR) {
		    parseError(ms, "operand of unary '~' must be integer");
		}
		*pType = TYPE_INTEGER;
	    }
	}
    } else {
	parsePost(ms, pEx, pType);
    }
}

/*
 * parseBitAndXorShift - parse the binary '&', '^', '<<' and '>>' operators.
 */

static void
parseBitAndXorShift(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    Expression_t *exp;
    UINT_T rightType;

    parseNegate(ms, pEx, pType);
    while (ms->tk_kind == tk_ampersand || ms->tk_kind == tk_xor ||
	ms->tk_kind == tk_shiftLeft || ms->tk_kind == tk_shiftRight)
    {
	exp = MEM_ALLOC(ms, sizeof(Expression_t), "ND_X");
	exp->ex_left = *pEx;
	*pEx = MEM_ALLOC(ms, sizeof(Exec_t), "NDEX");
	(*pEx)->ex_kind =
	    (ms->tk_kind == tk_ampersand) ?
		ex_bitAnd
	    : (ms->tk_kind == tk_xor) ?
		ex_bitXor
	    : (ms->tk_kind == tk_shiftLeft) ?
		ex_shiftLeft
	    :
		ex_shiftRight;
	(*pEx)->ex_v.ex_expressionPtr = exp;
	if (*pType != TYPE_INTEGER) {
	    if (*pType != TYPE_ERROR) {
		parseError(ms,
	       "left operand of binary '&', '^', '<<', '>>' must be integer");
	    }
	    *pType = TYPE_INTEGER;
	}
	getToken(ms);
	parseNegate(ms, &exp->ex_right, &rightType);
	if (rightType != TYPE_INTEGER && rightType != TYPE_ERROR) {
	    parseError(ms,
	      "right operand of binary '&', '^', '<<', '>>' must be integer");
	}
    }
}

/*
 * parseBitOr - parse the binary '|' operator.
 */

static void
parseBitOr(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    Expression_t *exp;
    UINT_T rightType;

    parseBitAndXorShift(ms, pEx, pType);
    while (ms->tk_kind == tk_stick) {
	exp = MEM_ALLOC(ms, sizeof(Expression_t), "OR_X");
	exp->ex_left = *pEx;
	*pEx = MEM_ALLOC(ms, sizeof(Exec_t), "OREX");
	(*pEx)->ex_kind = ex_bitOr;
	(*pEx)->ex_v.ex_expressionPtr = exp;
	if (*pType != TYPE_INTEGER) {
	    if (*pType != TYPE_ERROR) {
		parseError(ms, "left operand of binary '|' must be integer");
	    }
	    *pType = TYPE_INTEGER;
	}
	getToken(ms);
	parseBitAndXorShift(ms, &exp->ex_right, &rightType);
	if (rightType != TYPE_INTEGER && rightType != TYPE_ERROR) {
	    parseError(ms, "right operand of binary '|' must be integer");
	}
    }
}

/*
 * parseMulDivRem - parse the binary '*', '/' and '%' operators.
 */

static void
parseMulDivRem(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    Expression_t *exp;
    UINT_T leftType, rightType;
    ExecKind_t kind;

    parseBitOr(ms, pEx, pType);
    while (ms->tk_kind == tk_star || ms->tk_kind == tk_slash ||
	ms->tk_kind == tk_percent)
    {
	exp = MEM_ALLOC(ms, sizeof(Expression_t), "ML_X");
	exp->ex_left = *pEx;
	*pEx = MEM_ALLOC(ms, sizeof(Exec_t), "MLEX");
	kind =
	    (ms->tk_kind == tk_star) ?
		ex_times
	    : (ms->tk_kind == tk_slash) ?
		ex_divide
	    :
		ex_remainder;
	(*pEx)->ex_v.ex_expressionPtr = exp;
	leftType = *pType;
	if (kind == ex_remainder) {
	    if (leftType != TYPE_INTEGER) {
		if (leftType != TYPE_ERROR) {
		    parseError(ms,
			"left operand of binary '%' must be integer");
		}
		leftType = TYPE_INTEGER;
	    }
	} else {
	    if (leftType != TYPE_INTEGER && leftType != TYPE_FIXED) {
		if (leftType != TYPE_ERROR) {
		    parseError(ms,
		 "left operand of binary '*' or '/' must be integer or fixed");
		}
		leftType = TYPE_INTEGER;
	    }
	}
	getToken(ms);
	parseBitOr(ms, &exp->ex_right, &rightType);
	if (kind == ex_remainder) {
	    if (rightType != TYPE_INTEGER && rightType != TYPE_ERROR) {
		parseError(ms, "right operand of binary '%' must be integer");
	    }
	} else {
	    if (rightType != leftType && rightType != TYPE_ERROR) {
		parseError(ms,
			   "operands of binary '*' and '/' must be same type");
	    }
	    if (leftType == TYPE_FIXED) {
		kind = kind - ex_times + ex_fTimes;
	    }
	}
	(*pEx)->ex_kind = kind;
	*pType = leftType;
    }
}

/*
 * parseAddSub - parse the binary '+' and '-' operators.
 */

static void
parseAddSub(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    Expression_t *exp;
    ExecKind_t kind;
    UINT_T leftType, rightType, memType;
    BOOL_T typeError;

    parseMulDivRem(ms, pEx, pType);
    while (ms->tk_kind == tk_plus || ms->tk_kind == tk_minus) {
	exp = MEM_ALLOC(ms, sizeof(Expression_t), "AD_X");
	exp->ex_left = *pEx;
	*pEx = MEM_ALLOC(ms, sizeof(Exec_t), "ADEX");
	(*pEx)->ex_v.ex_expressionPtr = exp;
	leftType = *pType;
	kind = (ms->tk_kind == tk_plus) ? ex_plus : ex_minus;
	typeError = B_FALSE;
	if (kind == ex_plus) {
	    if (leftType != TYPE_INTEGER && leftType != TYPE_FIXED &&
		leftType != TYPE_STRING)
	    {
		if (leftType != TYPE_ERROR) {
		    parseError(ms,
		"left operand of binary '+' must be integer, fixed or string");
		}
		leftType = TYPE_INTEGER;
		typeError = B_TRUE;
	    }
	} else {
	    if (leftType != TYPE_INTEGER && leftType != TYPE_FIXED) {
		if (leftType != TYPE_ERROR) {
		    parseError(ms,
			"left operand of binary '-' must be integer or fixed");
		}
		leftType = TYPE_INTEGER;
		typeError = B_TRUE;
	    }
	}
	getToken(ms);
	parseMulDivRem(ms, &exp->ex_right, &memType);
	rightType = memType;
	if (kind == ex_plus) {
	    if (rightType != leftType && rightType != TYPE_ERROR) {
		if (typeError) {
		    if (rightType == TYPE_INTEGER || rightType == TYPE_FIXED ||
			rightType == TYPE_STRING)
		    {
			leftType = rightType;
		    }
		} else {
		    parseError(ms, "operands of binary '+' must be same type");
		}
	    }
	    if (leftType == TYPE_FIXED) {
		kind = ex_fPlus;
	    } else if (leftType == TYPE_STRING) {
		kind = ex_strPlus;
	    }
	} else {
	    if (rightType != leftType && rightType != TYPE_ERROR) {
		if (typeError) {
		    if (rightType == TYPE_INTEGER || rightType == TYPE_FIXED) {
			leftType = rightType;
		    }
		} else {
		    parseError(ms, "operands of binary '-' must be same type");
		}
	    }
	    if (leftType == TYPE_FIXED) {
		kind = ex_fMinus;
	    }
	}
	(*pEx)->ex_kind = kind;
	*pType = leftType;
    }
}

/*
 * parseComparison - parse the binary comparison operators.
 */

static void
parseComparison(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    Expression_t *exp;
    UINT_T tokenKind;
    ExecKind_t kind;
    UINT_T leftType, rightType, memType;

    parseAddSub(ms, pEx, pType);
    tokenKind = ms->tk_kind;
    if (tokenKind == tk_equal || tokenKind == tk_notEqual ||
	tokenKind == tk_less || tokenKind == tk_greater ||
	tokenKind == tk_lessOrEqual || tokenKind == tk_greaterOrEqual ||
	tokenKind == tk_equalEqual)
    {
	exp = MEM_ALLOC(ms, sizeof(Expression_t), "CM_X");
	exp->ex_left = *pEx;
	*pEx = MEM_ALLOC(ms, sizeof(Exec_t), "CMEX");
	(*pEx)->ex_v.ex_expressionPtr = exp;
	leftType = *pType;
	kind =
	    (tokenKind == tk_equal) ?
		ex_strEqual
	    : (tokenKind == tk_notEqual) ?
		ex_strNotEqual
	    : (tokenKind == tk_less) ?
		ex_strLess
	    : (tokenKind == tk_greater) ?
		ex_strGreater
	    : (tokenKind == tk_lessOrEqual) ?
		ex_strLessOrEqual
	    : (tokenKind == tk_greaterOrEqual) ?
		ex_strGreaterOrEqual
	    :
		ex_strEqualEqual;
	if (kind == ex_strEqualEqual) {
	    if (leftType != TYPE_STRING) {
		if (leftType != TYPE_ERROR) {
		    parseError(ms, "operands of '==' must be string");
		}
	    }
	} else if (kind == ex_strEqual || kind == ex_strNotEqual) {
	    if (leftType != TYPE_INTEGER && leftType != TYPE_FIXED &&
		leftType != TYPE_STATUS && leftType != TYPE_BOOLEAN &&
		leftType != TYPE_STRING && leftType != TYPE_THING_STATUS &&
		leftType != TYPE_NIL && ! libNilType(leftType))
	    {
		if (leftType != TYPE_ERROR) {
		    parseError(ms, "invalid operand type for '=' or '!='");
		    leftType = TYPE_ERROR;
		}
	    }
	} else {
	    if (leftType != TYPE_INTEGER && leftType != TYPE_FIXED &&
		leftType != TYPE_STRING)
	    {
		if (leftType != TYPE_ERROR) {
		    parseError(ms,
	       "operands of size comparison must be integer, fixed or string");
		    leftType = TYPE_ERROR;
		}
	    }
	}
	getToken(ms);
	parseAddSub(ms, &exp->ex_right, &memType);
	rightType = memType;
	if (leftType == TYPE_NIL && libNilType(rightType)) {
	    leftType = rightType;
	    exp->ex_left->ex_v.ex_integer = (ULONG_T) rightType << TYPE_SHIFT;
	} else if (rightType == TYPE_NIL && libNilType(leftType)) {
	    rightType = leftType;
	    exp->ex_right->ex_v.ex_integer = (ULONG_T) rightType << TYPE_SHIFT;
	}
	if (kind == ex_strEqualEqual) {
	    if (rightType != TYPE_STRING) {
		if (rightType != TYPE_ERROR) {
		    parseError(ms, "right operand of '==' must be string");
		}
	    }
	} else if (kind == ex_strEqual || kind == ex_strNotEqual) {
	    if (rightType != leftType && leftType != TYPE_ERROR &&
		rightType != TYPE_ERROR)
	    {
		parseError(ms, "operands of '=' or '~=' must be same type");
	    }
	} else {
	    if (leftType != rightType && rightType != TYPE_ERROR) {
		parseError(ms,"operands of size comparison must be same type");
	    }
	}
	if (kind != ex_strEqualEqual) {
	    if (leftType == TYPE_FIXED) {
		kind = kind - ex_strEqual + ex_fEqual;
	    } else if (leftType != TYPE_STRING) {
		/* int, bool, status, thing_status, pointers */
		kind = kind - ex_strEqual + ex_equal;
	    }
	}
	(*pEx)->ex_kind = kind;
	*pType = TYPE_BOOLEAN;
    }
}

/*
 * parseLogNot - parse the logical 'not' operator.
 */

static void
parseLogNot(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{

    if (ms->tk_kind == tk_not) {
	getToken(ms);
	*pEx = MEM_ALLOC(ms, sizeof(Exec_t), "NTEX");
	(*pEx)->ex_kind = ex_logNot;
	(*pEx)->ex_v.ex_expressionPtr =
	    MEM_ALLOC(ms, sizeof(Expression_t), "NT_X");
	(*pEx)->ex_v.ex_expressionPtr->ex_left = P_NULL;
	parseLogNot(ms, &(*pEx)->ex_v.ex_expressionPtr->ex_right, pType);
	if (*pType != TYPE_BOOLEAN) {
	    if (*pType != TYPE_ERROR) {
		parseError(ms, "operand of logical 'not' must be boolean");
	    }
	    *pType = TYPE_BOOLEAN;
	}
    } else {
	parseComparison(ms, pEx, pType);
    }
}

/*
 * parseLogAnd - parse the binary logical 'and' operator.
 */

static void
parseLogAnd(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    Expression_t *exp;
    UINT_T rightType;

    parseLogNot(ms, pEx, pType);
    while (ms->tk_kind == tk_and) {
	exp = MEM_ALLOC(ms, sizeof(Expression_t), "LN_X");
	exp->ex_left = *pEx;
	*pEx = MEM_ALLOC(ms, sizeof(Exec_t), "LNEX");
	(*pEx)->ex_kind = ex_logAnd;
	(*pEx)->ex_v.ex_expressionPtr = exp;
	if (*pType != TYPE_BOOLEAN) {
	    if (*pType != TYPE_ERROR) {
		parseError(ms,"left operand of logical 'and' must be boolean");
	    }
	    *pType = TYPE_BOOLEAN;
	}
	getToken(ms);
	parseLogNot(ms, &exp->ex_right, &rightType);
	if (rightType != TYPE_BOOLEAN && rightType != TYPE_ERROR) {
	    parseError(ms, "right operand of logical 'and' must be boolean");
	}
    }
}

/*
 * parseLogOr - parse the binary logical 'or' operator.
 */

static void
parseLogOr(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    Expression_t *exp;
    UINT_T rightType;

    parseLogAnd(ms, pEx, pType);
    while (ms->tk_kind == tk_or) {
	exp = MEM_ALLOC(ms, sizeof(Expression_t), "LO_X");
	exp->ex_left = *pEx;
	*pEx = MEM_ALLOC(ms, sizeof(Exec_t), "LOEX");
	(*pEx)->ex_kind = ex_logOr;
	(*pEx)->ex_v.ex_expressionPtr = exp;
	if (*pType != TYPE_BOOLEAN) {
	    if (*pType != TYPE_BOOLEAN) {
		parseError(ms, "left operand of logical 'or' must be boolean");
	    }
	    *pType = TYPE_BOOLEAN;
	}
	getToken(ms);
	parseLogAnd(ms, &exp->ex_right, &rightType);
	if (rightType != TYPE_BOOLEAN && rightType != TYPE_ERROR) {
	    parseError(ms, "right operand of logical 'or' must be boolean");
	}
    }
}

/*
 * parseAssignment - parse an assignment statement.
 */

static void
parseAssignment(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    Exec_t *ex;
    Assignment_t *as;
    PropAdd_t *pra;
    ElemPut_t *elp;
    UINT_T typ;
    ExecKind_t kind;
    UINT_T sourceType;

    parseLogOr(ms, pEx, pType);
    if (ms->tk_kind == tk_assign) {
	ex = *pEx;
	kind = ex->ex_kind;
	if (kind >= ex_getBoolProp && kind <= ex_getActionListProp) {
	    ex->ex_kind = kind - ex_getBoolProp + ex_addBoolProp;
	    if (*pType == TYPE_ERROR) {
		typ = TYPE_ERROR;
	    } else {
		typ = kind - ex_getBoolProp + TYPE_BOOLEAN;
	    }
	    getToken(ms);
	    pra = MEM_ALLOC(ms, sizeof(PropAdd_t), "PA_T");
	    pra->pra_thing = ex->ex_v.ex_getPtr->prg_thing;
	    pra->pra_property = ex->ex_v.ex_getPtr->prg_property;
	    memFree(ms, ex->ex_v.ex_getPtr, sizeof(PropGet_t));
	    ex->ex_v.ex_addPtr = pra;
	    parseLogOr(ms, &pra->pra_value, &sourceType);
	} else if (kind >= ex_getIntElement && kind <= ex_getActionElement) {
	    ex->ex_kind = kind - ex_getIntElement + ex_putIntElement;
	    if (*pType == TYPE_ERROR) {
		typ = TYPE_ERROR;
	    } else {
		typ = kind - ex_getIntElement + TYPE_INTEGER;
	    }
	    getToken(ms);
	    elp = MEM_ALLOC(ms, sizeof(ElemPut_t), "EP_T");
	    elp->elp_list = ex->ex_v.ex_elemGetPtr->elg_list;
	    elp->elp_index = ex->ex_v.ex_elemGetPtr->elg_index;
	    memFree(ms, ex->ex_v.ex_elemGetPtr, sizeof(ElemGet_t));
	    ex->ex_v.ex_elemPutPtr = elp;
	    parseLogOr(ms, &elp->elp_value, &sourceType);
	} else {
	    if (kind != ex_local && kind != ex_error) {
		parseError(ms, "can't assign to this");
	    }
	    typ = *pType;
	    ex = MEM_ALLOC(ms, sizeof(Exec_t), "ASEX");
	    ex->ex_kind = ex_assignment;
	    as = MEM_ALLOC(ms, sizeof(Assignment_t), "AS_T");
	    ex->ex_v.ex_assignmentPtr = as;
	    as->as_destination = *pEx;
	    *pEx = ex;
	    getToken(ms);
	    parseLogOr(ms, &as->as_source, &sourceType);
	    if (sourceType == TYPE_NIL && libNilType(typ)) {
		sourceType = typ;
		as->as_source->ex_v.ex_integer = (ULONG_T) typ << TYPE_SHIFT;
	    }
	    checkProc(ms, typ, as->as_source);
	}
	if (sourceType != typ && typ != TYPE_ERROR && sourceType != TYPE_ERROR)
	{
	    parseError(ms, "incompatible types in assignment");
	}
	*pType = TYPE_VOID;
    }
}

#if 0
Nobody needed this.
/*
 * parseExpr - top level interface to parse an expression.
 */

static void
parseExpr(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{

    parseAssignment(ms, pEx, pType);
}
#endif

/*
 * isStatementEnd - helper routine to check for something that can
 *	end a statement.
 */

static BOOL_T
isStatementEnd(UINT_T kind)
{

    return (BOOL_T) (kind == tk_eof || kind == tk_fi || kind == tk_elif ||
		     kind == tk_else || kind == tk_od || kind == tk_do ||
		     kind == tk_incase || kind == tk_default ||
		     kind == tk_esac || kind == tk_corp || kind == tk_dollar);
}

/*
 * parseSequence - parse a ';'-separated sequence of statements.
 */

static void
parseSequence(MUDState_t *ms, Exec_t **pEx, UINT_T *pType)
{
    StatementSequence_t *head;
    StatementSequence_t **tail;
    BOOL_T first, hadSemi, wasNote;

    tail = &head;
    hadSemi = B_FALSE;	/* be quiet, compiler! */
    first = B_TRUE;
    while (! isStatementEnd(ms->tk_kind)) {
	if (first) {
	    first = B_FALSE;
	} else if (! hadSemi) {
	    parseError(ms, "missing ';'");
	}
	*tail = MEM_ALLOC(ms, sizeof(StatementSequence_t), "SQ_T");
	parseAssignment(ms, &(*tail)->sts_this, pType);
	if ((*tail)->sts_this->ex_kind == ex_note) {
	    wasNote = B_TRUE;
	    hadSemi = B_TRUE;
	} else {
	    wasNote = B_FALSE;
	}
	tail = &(*tail)->sts_next;
	if (ms->tk_kind == tk_semicolon) {
	    if (*pType != TYPE_VOID && *pType != TYPE_ERROR) {
		parseError(ms, "use of expression as statement");
	    }
	    while (ms->tk_kind == tk_semicolon) {
		getToken(ms);
	    }
	    hadSemi = B_TRUE;
	    *pType = TYPE_VOID;
	} else if (! wasNote) {
	    hadSemi = B_FALSE;
	}
    }
    *tail = P_NULL;
    if (head == P_NULL) {
	*pEx = P_NULL;
	*pType = TYPE_VOID;
    } else {
	if (head->sts_next != P_NULL) {
	    *pEx = MEM_ALLOC(ms, sizeof(Exec_t), "SQEX");
	    (*pEx)->ex_kind = ex_sequence;
	    (*pEx)->ex_v.ex_sequencePtr = head;
	} else {
	    *pEx = head->sts_this;
	    memFree(ms, head, sizeof(StatementSequence_t));
	}
    }
}

/*
 * parseBody - top-level interface to parse a statement sequence (proc body).
 */

void
parseBody(MUDState_t *ms, Exec_t **pEx, UINT_T *pType, BOOL_T allowDir)
{

    parseSequence(ms, pEx, pType);
    if (*pEx != P_NULL && (*pEx)->ex_kind == ex_procRef && ! allowDir) {
	parseError(ms, "body of proc cannot be just a direct proc");
    }
}
