#include "MUD.h"
#include "types.h"
#include "exec.h"
#include "MUDInternal.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * procIO.c - code to convert procs to and from byte streams.
 */

/*
 * writeExec - lower level recursive proc writer.
 */

static void
writeExec(MUDState_t *ms, Exec_t *ex)
{

    if (ex == P_NULL) {
	writeByte(ms, ex_null - ex_null);
    } else {
	writeByte(ms, ex->ex_kind - ex_null);
	switch (ex->ex_kind) {
	case ex_byteCode:
	case ex_error:
	case ex_empty:
	    break;
	case ex_intRef:
	case ex_fixedRef:
	case ex_procRef:
	case ex_thingRef:
	case ex_grammarRef:
	case ex_playerRef:
	case ex_propRef:
	case ex_tableRef:
	    writeLong(ms, ex->ex_v.ex_refPtr->rl_key);
	    break;
	case ex_local:
	    writeWord(ms, ex->ex_v.ex_localPtr->ll_offset);
	    break;
	case ex_string:
	case ex_note:
	    writeString(ms, ex->ex_v.ex_stringPtr);
	    break;
	case ex_int:
	case ex_fixed:
	case ex_bool:
	case ex_status:
	case ex_nil:
	case ex_ts:
	    writeLong(ms, ex->ex_v.ex_integer);
	    break;
	case ex_if: {
	    If_t *ifPtr;
	    IfList_t *ifl;

	    ifPtr = ex->ex_v.ex_ifPtr;
	    writeExec(ms, ifPtr->if_else);
	    ifl = ifPtr->if_alternatives;
	    while (ifl != P_NULL) {
		writeExec(ms, ifl->ifl_condition);
		writeExec(ms, ifl->ifl_body);
		ifl = ifl->ifl_next;
	    }
	    writeByte(ms, ex_null - ex_null);
	    break;
	}
	case ex_while: {
	    While_t *wh;

	    wh = ex->ex_v.ex_whilePtr;
	    writeExec(ms, wh->wh_condition);
	    writeExec(ms, wh->wh_body);
	    break;
	}
	case ex_for: {
	    For_t *fo;

	    fo = ex->ex_v.ex_forPtr;
	    if (fo->fo_variable == P_NULL) {
		writeWord(ms, 1);
	    } else {
		writeWord(ms, fo->fo_variable->ll_offset);
	    }
	    writeExec(ms, fo->fo_start);
	    writeExec(ms, fo->fo_limit);
	    writeExec(ms, fo->fo_body);
	    break;
	}
	case ex_case: {
	    Case_t *ca;
	    CaseList_t *cal;
	    CaseIndexList_t *cil;

	    ca = ex->ex_v.ex_casePtr;
	    writeExec(ms, ca->ca_selector);
	    writeExec(ms, ca->ca_default);
	    cal = ca->ca_alternatives;
	    while (cal != P_NULL) {
		writeExec(ms, cal->cal_body);
		cil = cal->cal_indexes;
		while (cil != P_NULL) {
		    writeExec(ms, cil->cil_index);
		    cil = cil->cil_next;
		}
		writeByte(ms, ex_null - ex_null);
		cal = cal->cal_next;
	    }
	    writeByte(ms, ex_null - ex_null);
	    break;
	}
	case ex_ignore:
	case ex_ignoreS:
	case ex_negate:
	case ex_bitNot:
	case ex_fNegate:
	case ex_logNot:
	    writeExec(ms, ex->ex_v.ex_expressionPtr->ex_right);
	    break;
	case ex_builtinCall: {
	    BuiltinCall_t *bcl;
	    ArgumentList_t *al;

	    bcl = ex->ex_v.ex_builtinCallPtr;
	    writeLong(ms, bcl->bcl_key);
	    al = bcl->bcl_arguments;
	    while (al != P_NULL) {
		writeExec(ms, al->al_this);
		al = al->al_next;
	    }
	    writeByte(ms, ex_null - ex_null);
	    break;
	}
	case ex_knownCall:
	case ex_unknownCall: {
	    UserCall_t *ucl;
	    ArgumentList_t *al;

	    ucl = ex->ex_v.ex_userCallPtr;
	    writeExec(ms, ucl->ucl_action);
	    writeWord(ms, ucl->ucl_resultType);
	    al = ucl->ucl_arguments;
	    while (al != P_NULL) {
		writeExec(ms, al->al_this);
		al = al->al_next;
	    }
	    writeByte(ms, ex_null - ex_null);
	    break;
	}
	case ex_getBoolProp:
	case ex_getIntProp:
	case ex_getFixedProp:
	case ex_getThingProp:
	case ex_getActionProp:
	case ex_getStringProp:
	case ex_getTableProp:
	case ex_getGrammarProp:
	case ex_getIntListProp:
	case ex_getFixedListProp:
	case ex_getThingListProp:
	case ex_getActionListProp: {
	    PropGet_t *prg;

	    prg = ex->ex_v.ex_getPtr;
	    writeExec(ms, prg->prg_thing);
	    writeExec(ms, prg->prg_property);
	    break;
	}
	case ex_addBoolProp:
	case ex_addIntProp:
	case ex_addFixedProp:
	case ex_addThingProp:
	case ex_addActionProp:
	case ex_addStringProp:
	case ex_addTableProp:
	case ex_addGrammarProp:
	case ex_addIntListProp:
	case ex_addFixedListProp:
	case ex_addThingListProp:
	case ex_addActionListProp: {
	    PropAdd_t *pra;

	    pra = ex->ex_v.ex_addPtr;
	    writeExec(ms, pra->pra_thing);
	    writeExec(ms, pra->pra_property);
	    writeExec(ms, pra->pra_value);
	    break;
	}
	case ex_delBoolProp:
	case ex_delIntProp:
	case ex_delFixedProp:
	case ex_delThingProp:
	case ex_delActionProp:
	case ex_delStringProp:
	case ex_delTableProp:
	case ex_delGrammarProp:
	case ex_delIntListProp:
	case ex_delFixedListProp:
	case ex_delThingListProp:
	case ex_delActionListProp: {
	    PropDel_t *prd;

	    prd = ex->ex_v.ex_delPtr;
	    writeExec(ms, prd->prd_thing);
	    writeExec(ms, prd->prd_property);
	    break;
	}
	case ex_getIntElement:
	case ex_getFixedElement:
	case ex_getThingElement:
	case ex_getActionElement: {
	    ElemGet_t *elg;

	    elg = ex->ex_v.ex_elemGetPtr;
	    writeExec(ms, elg->elg_list);
	    writeExec(ms, elg->elg_index);
	    break;
	}
	case ex_putIntElement:
	case ex_putFixedElement:
	case ex_putThingElement:
	case ex_putActionElement: {
	    ElemPut_t *elp;

	    elp = ex->ex_v.ex_elemPutPtr;
	    writeExec(ms, elp->elp_list);
	    writeExec(ms, elp->elp_index);
	    writeExec(ms, elp->elp_value);
	    break;
	}
	case ex_shiftLeft:
	case ex_shiftRight:
	case ex_bitAnd:
	case ex_bitXor:
	case ex_bitOr:
	case ex_times:
	case ex_divide:
	case ex_remainder:
	case ex_plus:
	case ex_minus:
	case ex_equal:
	case ex_notEqual:
	case ex_less:
	case ex_greater:
	case ex_lessOrEqual:
	case ex_greaterOrEqual:
	case ex_fTimes:
	case ex_fDivide:
	case ex_fPlus:
	case ex_fMinus:
	case ex_fEqual:
	case ex_fNotEqual:
	case ex_fLess:
	case ex_fGreater:
	case ex_fLessOrEqual:
	case ex_fGreaterOrEqual:
	case ex_strPlus:
	case ex_strEqual:
	case ex_strNotEqual:
	case ex_strLess:
	case ex_strGreater:
	case ex_strLessOrEqual:
	case ex_strGreaterOrEqual:
	case ex_strEqualEqual:
	case ex_logAnd:
	case ex_logOr: {
	    Expression_t *exp;

	    exp = ex->ex_v.ex_expressionPtr;
	    writeExec(ms, exp->ex_left);
	    writeExec(ms, exp->ex_right);
	    break;
	}
	case ex_assignment: {
	    Assignment_t *as;

	    as = ex->ex_v.ex_assignmentPtr;
	    writeExec(ms, as->as_destination);
	    writeExec(ms, as->as_source);
	    break;
	}
	case ex_sequence: {
	    StatementSequence_t *sts;

	    sts = ex->ex_v.ex_sequencePtr;
	    while (sts != P_NULL) {
		writeExec(ms, sts->sts_this);
		sts = sts->sts_next;
	    }
	    writeByte(ms, ex_null - ex_null);
	    break;
	}
	default:
	    errorString(ms, "Invalid ex_kind (");
	    errorDecimal(ms, ex->ex_kind - ex_null);
	    errorString(ms, ") in writeExec\n");
	    libAbort(ms, B_FALSE);
	    break;
	}
    }
}

/*
 * writeLocalList - used to write parameters and local variables.
 */

static void
writeLocalList(MUDState_t *ms, LocalList_t *ll)
{

    while (ll != P_NULL) {
	/* Hack to avoid storing a 0 byte for the type, which can mess
	   us up in determining how many pars/vars there are. */
	writeByte(ms, (ll->ll_type == TYPE_ERROR ? 1 : ll->ll_type));
	writeWord(ms, ll->ll_offset);
	if (ll->ll_name == P_NULL) {
	    writeByte(ms, 0);
	} else {
	    writeString(ms, ll->ll_name);
	}
	ll = ll->ll_next;
    }
    writeByte(ms, 0);
}

/*
 * writeProc - top level of writing out a proc.
 */

void
writeProc(MUDState_t *ms, Proc_t *pr)
{
    RefList_t *rl;

    ms->WritePos = 0;
    /* Write this first, so the caching code in MUDServ can cheat and
       just poke it directly. */
    writeWord(ms, pr->pr_useCount);
    writeLocalList(ms, pr->pr_parameters);
    writeLocalList(ms, pr->pr_locals);
    rl = pr->pr_refs;
    while (rl != P_NULL) {
	writeLong(ms, rl->rl_key);
	rl = rl->rl_next;
    }
    writeLong(ms, 0);
    /* Note that this does not depend on pr_stripped - that is so that
       a proc marked as stripped at parse time will be sent to the
       server properly. */
    writeExec(ms, pr->pr_body);
    writeByte(ms, pr->pr_resultType);
    writeLong(ms, pr->pr_owner);
    writeByte(ms, pr->pr_localCount);
    writeByte(ms, pr->pr_utility - B_FALSE);
    writeByte(ms, pr->pr_wizard - B_FALSE);
    writeByte(ms, pr->pr_public - B_FALSE);
    writeByte(ms, pr->pr_isCompiled - B_FALSE);
    writeByte(ms, pr->pr_stripped - B_FALSE);
    writeWord(ms, pr->pr_byteCodeLen);
    if (pr->pr_stripped && pr->pr_byteCodeLen != 0) {
	writeBytes(ms, pr->pr_byteCode, pr->pr_byteCodeLen);
    }
}

/*
 * writeProcE - extern interface to writeProc.
 */

void
writeProcE(MUDState_t *ms, Proc_t *pr)
{
    writeProc(ms, pr);
}

/*
 * getLocalVar - find the local for this proc with the given offset.
 */

static LocalList_t *
getLocalVar(MUDState_t *ms, UINT_T offset)
{
    LocalList_t *ll;

    ll = ms->ReadingProc->pr_parameters;
    while (ll != P_NULL && ll->ll_offset != offset) {
	ll = ll->ll_next;
    }
    if (ll == P_NULL) {
	ll = ms->ReadingProc->pr_locals;
	while (ll != P_NULL && ll->ll_offset != offset) {
	    ll = ll->ll_next;
	}
	if (ll == P_NULL) {
	    errorString(ms, "Can't find local offset ");
	    errorHex(ms, offset);
	    errorString(ms, " in 'readExec'. Offsets are:\npars:");
	    ll = ms->ReadingProc->pr_parameters;
	    while (ll != P_NULL) {
		ms->ms_errorChar(' ');
		errorHex(ms, ll->ll_offset);
		ll = ll->ll_next;
	    }
	    errorString(ms, "  locs:");
	    ll = ms->ReadingProc->pr_locals;
	    while (ll != P_NULL) {
		ms->ms_errorChar(' ');
		errorHex(ms, ll->ll_offset);
		ll = ll->ll_next;
	    }
	    ms->ms_errorChar('\n');
	    libAbort(ms, B_FALSE);
	}
    }
    return ll;
}

#if 0
static char *ExecKindName[] = {
    "null",
    "byteCode",
    "error",
    "empty",
    "intRef",
    "fixedRef",
    "procRef",
    "thingRef",
    "grammarRef",
    "playerRef",
    "propRef",
    "tableRef",
    "local",
    "string",
    "int",
    "fixed",
    "bool",
    "status",
    "nil",
    "ts",
    "if",
    "while",
    "for",
    "case",
    "ignore",
    "ignoreS",
    "builtinCall",
    "knownCall",
    "unknownCall",
    "note",
    "getBoolProp",
    "getIntProp",
    "getFixedProp",
    "getThingProp",
    "getActionProp",
    "getStringProp",
    "getTableProp",
    "getGrammarProp",
    "getIntListProp",
    "getFixedListProp",
    "getThingListProp",
    "getActionListProp",
    "addBoolProp",
    "addIntProp",
    "addFixedProp",
    "addThingProp",
    "addActionProp",
    "addStringProp",
    "addTableProp",
    "addGrammarProp",
    "addIntListProp",
    "addFixedListProp",
    "addThingListProp",
    "addActionListProp",
    "delBoolProp",
    "delIntProp",
    "delFixedProp",
    "delThingProp",
    "delActionProp",
    "delStringProp",
    "delTableProp",
    "delGrammarProp",
    "delIntListProp",
    "delFixedListProp",
    "delThingListProp",
    "delActionListProp",
    "getIntElement",
    "getFixedElement",
    "getThingElement",
    "getActionElement",
    "putIntElement",
    "putFixedElement",
    "putThingElement",
    "putActionElement",
    "negate",
    "bitNot",
    "shiftLeft",
    "shiftRight",
    "bitAnd",
    "bitXor",
    "bitOr",
    "times",
    "divide",
    "remainder",
    "plus",
    "minus",
    "equal",
    "notEqual",
    "less",
    "greater",
    "lessOrEqual",
    "greaterOrEqual",
    "fNegate",
    "fTimes",
    "fDivide",
    "fPlus",
    "fMinus",
    "fEqual",
    "fNotEqual",
    "fLess",
    "fGreater",
    "fLessOrEqual",
    "fGreaterOrEqual",
    "strPlus",
    "strEqual",
    "strNotEqual",
    "strLess",
    "strGreater",
    "strLessOrEqual",
    "strGreaterOrEqual",
    "strEqualEqual",
    "logNot",
    "logAnd",
    "logOr",
    "assignment",
    "sequence"
};
#endif

/*
 * readExec - main recursive reading routine.
 */

static void *
readExec(MUDState_t *ms, Exec_t **pEx, void *p)
{
    Exec_t *ex;
    ExecKind_t kind;

    kind = *(BYTE_T *) p + ex_null;
    p = (BYTE_T *) p + 1;
    if (kind == ex_null) {
	*pEx = P_NULL;
    } else {
	ex = MEM_ALLOC(ms, sizeof(Exec_t), "_EX_");
	ex->ex_kind = kind;
	switch (kind) {
	case ex_byteCode:
	case ex_error:
	case ex_empty:
	    break;
	case ex_intRef:
	case ex_fixedRef:
	case ex_procRef:
	case ex_thingRef:
	case ex_grammarRef:
	case ex_playerRef:
	case ex_propRef:
	case ex_tableRef: {
	    RefList_t *rl;
	    ConstRefList_t *crl;
	    ULONG_T key;

	    key = readLong(&p);
	    rl = ms->ReadingProc->pr_refs;
	    while (rl != P_NULL && rl->rl_key != key) {
		rl = rl->rl_next;
	    }
	    if (rl == P_NULL) {
		errorString(ms, "can't find ref key 0x");
		errorHex(ms, key);
		errorString(ms, " in 'readExec'. Keys are:\n");
		rl = ms->ReadingProc->pr_refs;
		while (rl != P_NULL) {
		    errorHex(ms, rl->rl_key);
		    ms->ms_errorChar(' ');
		    rl = rl->rl_next;
		}
		ms->ms_errorChar('\n');
		libAbort(ms, B_FALSE);
	    }
	    ex->ex_v.ex_refPtr = rl;
	    if ((kind == ex_intRef || kind == ex_fixedRef) &&
		ms->ms_getIntConstValue != P_NULL)
	    {
		crl = MEM_ALLOC(ms, sizeof(ConstRefList_t), "CRL_");
		crl->crl_next = ms->ConstRefList;
		crl->crl_key = key;
		crl->crl_ptr = &ex->ex_intVal[0];
		ms->ConstRefList = crl;
	    }
	    break;
	}
	case ex_local:
	    ex->ex_v.ex_localPtr = getLocalVar(ms, readWord(&p));
	    break;
	case ex_string:
	case ex_note:
	    ex->ex_v.ex_stringPtr = charsDup(ms, (CHAR_T *) p);
	    p = (CHAR_T *) p + charsLen1((CHAR_T *) p);
	    break;
	case ex_int:
	case ex_fixed:
	case ex_bool:
	case ex_status:
	case ex_nil:
	case ex_ts:
	    ex->ex_v.ex_integer = readLong(&p);
	    break;
	case ex_if: {
	    If_t *ifPtr;
	    IfList_t *ifl;
	    IfList_t **iflTail;

	    ifPtr = MEM_ALLOC(ms, sizeof(If_t), "IF_T");
	    ex->ex_v.ex_ifPtr = ifPtr;
	    p = readExec(ms, &ifPtr->if_else, p);
	    iflTail = &ifPtr->if_alternatives;
	    while (*(BYTE_T *) p != ex_null - ex_null) {
		ifl = MEM_ALLOC(ms, sizeof(IfList_t), "IFLT");
		*iflTail = ifl;
		iflTail = &ifl->ifl_next;
		p = readExec(ms, &ifl->ifl_condition, p);
		p = readExec(ms, &ifl->ifl_body, p);
	    }
	    *iflTail = P_NULL;
	    p = (BYTE_T *) p + 1;
	    break;
	}
	case ex_while: {
	    While_t *wh;

	    wh = MEM_ALLOC(ms, sizeof(While_t), "WH_T");
	    ex->ex_v.ex_whilePtr = wh;
	    p = readExec(ms, &wh->wh_condition, p);
	    p = readExec(ms, &wh->wh_body, p);
	    break;
	}
	case ex_for: {
	    For_t *fo;
	    UINT_T offset;

	    fo = MEM_ALLOC(ms, sizeof(For_t), "FORT");
	    ex->ex_v.ex_forPtr = fo;
	    offset = readWord(&p);
	    if (offset == 1) {
		fo->fo_variable = P_NULL;
	    } else {
		fo->fo_variable = getLocalVar(ms, offset);
	    }
	    p = readExec(ms, &fo->fo_start, p);
	    p = readExec(ms, &fo->fo_limit, p);
	    p = readExec(ms, &fo->fo_body, p);
	    break;
	}
	case ex_case: {
	    Case_t *ca;
	    CaseList_t *cal;
	    CaseIndexList_t *cil;
	    CaseList_t **calTail;
	    CaseIndexList_t **cilTail;

	    ca = MEM_ALLOC(ms, sizeof(Case_t), "CASE");
	    ex->ex_v.ex_casePtr = ca;
	    p = readExec(ms, &ca->ca_selector, p);
	    p = readExec(ms, &ca->ca_default, p);
	    calTail = &ca->ca_alternatives;
	    while (*(BYTE_T *) p != ex_null - ex_null) {
		cal = MEM_ALLOC(ms, sizeof(CaseList_t), "CSLS");
		*calTail = cal;
		calTail = &cal->cal_next;
		p = readExec(ms, &cal->cal_body, p);
		cilTail = &cal->cal_indexes;
		while (*(BYTE_T *) p != ex_null - ex_null) {
		    cil = MEM_ALLOC(ms, sizeof(CaseIndexList_t), "CILT");
		    *cilTail = cil;
		    cilTail = &cil->cil_next;
		    p = readExec(ms, &cil->cil_index, p);
		}
		*cilTail = P_NULL;
		p = (BYTE_T *) p + 1;
	    }
	    *calTail = P_NULL;
	    p = (BYTE_T *) p + 1;
	    break;
	}
	case ex_ignore:
	case ex_ignoreS:
	case ex_negate:
	case ex_bitNot:
	case ex_fNegate:
	case ex_logNot: {
	    Expression_t *exp;

	    exp = MEM_ALLOC(ms, sizeof(Expression_t), "UNRY");
	    exp->ex_left = P_NULL;
	    ex->ex_v.ex_expressionPtr = exp;
	    p = readExec(ms, &exp->ex_right, p);
	    break;
	}
	case ex_builtinCall: {
	    BuiltinCall_t *bcl;
	    ArgumentList_t *al;
	    ArgumentList_t **alTail;

	    bcl = MEM_ALLOC(ms, sizeof(BuiltinCall_t), "BIC_");
	    ex->ex_v.ex_builtinCallPtr = bcl;
	    bcl->bcl_key = readLong(&p);
	    if (ms->ms_findBuiltin == P_NULL) {
		bcl->bcl_proc = P_NULL;
	    } else {
		bcl->bcl_proc = ms->ms_findBuiltin(bcl->bcl_key);
	    }
	    alTail = &bcl->bcl_arguments;
	    while (*(BYTE_T *) p != ex_null - ex_null) {
		al = MEM_ALLOC(ms, sizeof(ArgumentList_t), "BALT");
		*alTail = al;
		alTail = &al->al_next;
		p = readExec(ms, &al->al_this, p);
	    }
	    *alTail = P_NULL;
	    p = (BYTE_T *) p + 1;
	    break;
	}
	case ex_knownCall:
	case ex_unknownCall: {
	    UserCall_t *ucl;
	    ArgumentList_t *al;
	    ArgumentList_t **alTail;

	    ucl = MEM_ALLOC(ms, sizeof(UserCall_t), "UC_T");
	    ex->ex_v.ex_userCallPtr = ucl;
	    p = readExec(ms, &ucl->ucl_action, p);
	    ucl->ucl_resultType = readWord(&p);
	    alTail = &ucl->ucl_arguments;
	    while (*(BYTE_T *) p != ex_null - ex_null) {
		al = MEM_ALLOC(ms, sizeof(ArgumentList_t), "UCAL");
		*alTail = al;
		alTail = &al->al_next;
		p = readExec(ms, &al->al_this, p);
	    }
	    *alTail = P_NULL;
	    p = (BYTE_T *) p + 1;
	    break;
	}
	case ex_getBoolProp:
	case ex_getIntProp:
	case ex_getFixedProp:
	case ex_getThingProp:
	case ex_getActionProp:
	case ex_getStringProp:
	case ex_getTableProp:
	case ex_getGrammarProp:
	case ex_getIntListProp:
	case ex_getFixedListProp:
	case ex_getThingListProp:
	case ex_getActionListProp: {
	    PropGet_t *prg;

	    prg = MEM_ALLOC(ms, sizeof(PropGet_t), "PG_T");
	    ex->ex_v.ex_getPtr = prg;
	    p = readExec(ms, &prg->prg_thing, p);
	    p = readExec(ms, &prg->prg_property, p);
	    break;
	}
	case ex_addBoolProp:
	case ex_addIntProp:
	case ex_addFixedProp:
	case ex_addThingProp:
	case ex_addActionProp:
	case ex_addStringProp:
	case ex_addTableProp:
	case ex_addGrammarProp:
	case ex_addIntListProp:
	case ex_addFixedListProp:
	case ex_addThingListProp:
	case ex_addActionListProp: {
	    PropAdd_t *pra;

	    pra = MEM_ALLOC(ms, sizeof(PropAdd_t), "PA_T");
	    ex->ex_v.ex_addPtr = pra;
	    p = readExec(ms, &pra->pra_thing, p);
	    p = readExec(ms, &pra->pra_property, p);
	    p = readExec(ms, &pra->pra_value, p);
	    break;
	}
	case ex_delBoolProp:
	case ex_delIntProp:
	case ex_delFixedProp:
	case ex_delThingProp:
	case ex_delActionProp:
	case ex_delStringProp:
	case ex_delTableProp:
	case ex_delGrammarProp:
	case ex_delIntListProp:
	case ex_delFixedListProp:
	case ex_delThingListProp:
	case ex_delActionListProp: {
	    PropDel_t *prd;

	    prd = MEM_ALLOC(ms, sizeof(PropDel_t), "PD_T");
	    ex->ex_v.ex_delPtr = prd;
	    p = readExec(ms, &prd->prd_thing, p);
	    p = readExec(ms, &prd->prd_property, p);
	    break;
	}
	case ex_getIntElement:
	case ex_getFixedElement:
	case ex_getThingElement:
	case ex_getActionElement: {
	    ElemGet_t *elg;

	    elg = MEM_ALLOC(ms, sizeof(ElemGet_t), "EG_T");
	    ex->ex_v.ex_elemGetPtr = elg;
	    p = readExec(ms, &elg->elg_list, p);
	    p = readExec(ms, &elg->elg_index, p);
	    break;
	}
	case ex_putIntElement:
	case ex_putFixedElement:
	case ex_putThingElement:
	case ex_putActionElement: {
	    ElemPut_t *elp;

	    elp = MEM_ALLOC(ms, sizeof(ElemPut_t), "EP_T");
	    ex->ex_v.ex_elemPutPtr = elp;
	    p = readExec(ms, &elp->elp_list, p);
	    p = readExec(ms, &elp->elp_index, p);
	    p = readExec(ms, &elp->elp_value, p);
	    break;
	}
	case ex_shiftLeft:
	case ex_shiftRight:
	case ex_bitAnd:
	case ex_bitXor:
	case ex_bitOr:
	case ex_times:
	case ex_divide:
	case ex_remainder:
	case ex_plus:
	case ex_minus:
	case ex_equal:
	case ex_notEqual:
	case ex_less:
	case ex_greater:
	case ex_lessOrEqual:
	case ex_greaterOrEqual:
	case ex_fTimes:
	case ex_fDivide:
	case ex_fPlus:
	case ex_fMinus:
	case ex_fEqual:
	case ex_fNotEqual:
	case ex_fLess:
	case ex_fGreater:
	case ex_fLessOrEqual:
	case ex_fGreaterOrEqual:
	case ex_strPlus:
	case ex_strEqual:
	case ex_strNotEqual:
	case ex_strLess:
	case ex_strGreater:
	case ex_strLessOrEqual:
	case ex_strGreaterOrEqual:
	case ex_strEqualEqual:
	case ex_logAnd:
	case ex_logOr: {
	    Expression_t *exp;

	    exp = MEM_ALLOC(ms, sizeof(Expression_t), "BNRY");
	    ex->ex_v.ex_expressionPtr = exp;
	    p = readExec(ms, &exp->ex_left, p);
	    p = readExec(ms, &exp->ex_right, p);
	    break;
	}
	case ex_assignment: {
	    Assignment_t *as;

	    as = MEM_ALLOC(ms, sizeof(Assignment_t), "ASGN");
	    ex->ex_v.ex_assignmentPtr = as;
	    p = readExec(ms, &as->as_destination, p);
	    p = readExec(ms, &as->as_source, p);
	    break;
	}
	case ex_sequence: {
	    StatementSequence_t *sts;
	    StatementSequence_t **stsTail;

	    stsTail = &ex->ex_v.ex_sequencePtr;
	    while (*(BYTE_T *) p != ex_null - ex_null) {
		sts = MEM_ALLOC(ms, sizeof(StatementSequence_t), "SEQ_");
		*stsTail = sts;
		stsTail = &sts->sts_next;
		p = readExec(ms, &sts->sts_this, p);
	    }
	    *stsTail = P_NULL;
	    p = (BYTE_T *) p + 1;
	    break;
	}
	default:
	    errorString(ms, "Invalid ex_kind (");
	    errorDecimal(ms, kind - ex_null);
	    errorString(ms, ") in readExec\n");
	    libAbort(ms, B_FALSE);
	    break;
	}
	*pEx = ex;
    }
    return p;
}

/*
 * readLocalList - used for both parameter and local variable lists.
 */

static void *
readLocalList(MUDState_t *ms, LocalList_t **pll, void *p)
{
    LocalList_t *ll;
    BYTE_T b;

    while (B_TRUE) {
	b = *(BYTE_T *) p;
	if (b == 0) {
	    break;
	}
	ll = MEM_ALLOC(ms, sizeof(LocalList_t), "LL_T");
	p = (BYTE_T *) p + 1;
	ll->ll_type = (b == 1 ? TYPE_ERROR : b);
	ll->ll_offset = readWord(&p);
	if (*(BYTE_T *) p == 0) {
	    ll->ll_name = P_NULL;
	    p = (BYTE_T *) p + 1;
	} else {
	    ll->ll_name = charsDup(ms, (CHAR_T *) p);
	    p = (CHAR_T *) p + charsLen1((CHAR_T *) p);
	}
	*pll = ll;
	pll = &ll->ll_next;
    }
    p = (BYTE_T *) p + 1;
    *pll = P_NULL;
    return p;
}

/*
 * readProc - turn a byte stream into a proc tree, given the byte stream.
 */

Proc_t *
readProc(MUDState_t *ms, void *p)
{
    Proc_t *pr;
    RefList_t *rl;
    ConstRefList_t *crl;
    ULONG_T k;
    UINT_T len;

    pr = MEM_ALLOC(ms, sizeof(Proc_t), "PROC");
    ms->ReadingProc = pr;
    ms->ConstRefList = P_NULL;
    pr->pr_useCount = readWord(&p);
    p = readLocalList(ms, &pr->pr_parameters, p);
    p = readLocalList(ms, &pr->pr_locals, p);
    pr->pr_refs = P_NULL;
    while (B_TRUE) {
	k = readLong(&p);
	if (k == 0) {
	    break;
	}
	rl = MEM_ALLOC(ms, sizeof(RefList_t), "RL_T");
	rl->rl_next = pr->pr_refs;
	rl->rl_name = P_NULL;
	rl->rl_key = k;
	pr->pr_refs = rl;
    }
    p = readExec(ms, &pr->pr_body, p);
    pr->pr_compiled = P_NULL;
    pr->pr_resultType = *(BYTE_T *) p;
    p = (BYTE_T *) p + 1;
    pr->pr_owner = readLong(&p);
    pr->pr_localCount = *(BYTE_T *) p;
    p = (BYTE_T *) p + 1;
    pr->pr_utility = *(BYTE_T *) p + B_FALSE;
    p = (BYTE_T *) p + 1;
    pr->pr_wizard = *(BYTE_T *) p + B_FALSE;
    p = (BYTE_T *) p + 1;
    pr->pr_public = *(BYTE_T *) p + B_FALSE;
    p = (BYTE_T *) p + 1;
    pr->pr_isCompiled = *(BYTE_T *) p + B_FALSE;
    p = (BYTE_T *) p + 1;
    pr->pr_stripped = *(BYTE_T *) p + B_FALSE;
    p = (BYTE_T *) p + 1;
    len = readWord(&p);
    pr->pr_byteCodeLen = len;
    if (pr->pr_stripped && len != 0) {
	pr->pr_byteCode = MEM_ALLOC(ms, len, "BTCD");
	blockCopy(pr->pr_byteCode, p, len);
    } else {
	pr->pr_byteCode = P_NULL;
    }
    pr->pr_locked = B_FALSE;
    pr->pr_byteSwapped = B_FALSE;
    /* Do this *after* all references to the byte-array of the proc, since
       it indirectly calls 'ioRead' which could invalidate the pointer. */
    while (B_TRUE) {
	crl = ms->ConstRefList;
	if (crl == P_NULL) {
	    break;
	}
	ms->ConstRefList = crl->crl_next;
	k = ms->ms_getIntConstValue(crl->crl_key);
	p = crl->crl_ptr;
	*(BYTE_T *) p = k >> 16;
	p = (BYTE_T *) p + 1;
	*(BYTE_T *) p = k >> 8;
	p = (BYTE_T *) p + 1;
	*(BYTE_T *) p = k;
	memFree(ms, crl, sizeof(ConstRefList_t));
    }
    return pr;
}

/*
 * readProcE - external interface to readProc.
 */

Proc_t *
readProcE(MUDState_t *ms, void *p)
{
    return readProc(ms, p);
}

/*
 * freeExec - recursive routine to free an exec structure.
 */

void
freeExec(MUDState_t *ms, Exec_t *ex)
{

    if (ex != P_NULL) {
	switch (ex->ex_kind) {
	case ex_byteCode:
	case ex_error:
	case ex_empty:
	case ex_intRef:
	case ex_fixedRef:
	case ex_procRef:
	case ex_thingRef:
	case ex_grammarRef:
	case ex_playerRef:
	case ex_propRef:
	case ex_tableRef:
	case ex_local:
	case ex_int:
	case ex_fixed:
	case ex_bool:
	case ex_status:
	case ex_nil:
	case ex_ts:
	    break;
	case ex_string:
	case ex_note:
	    charsFree(ms, ex->ex_v.ex_stringPtr);
	    break;
	case ex_if: {
	    If_t *ifp;
	    IfList_t *ifl, *iflTemp;

	    ifl = ex->ex_v.ex_ifPtr->if_alternatives;
	    while (ifl != P_NULL) {
		iflTemp = ifl;
		ifl = ifl->ifl_next;
		freeExec(ms, iflTemp->ifl_condition);
		freeExec(ms, iflTemp->ifl_body);
		memFree(ms, iflTemp, sizeof(IfList_t));
	    }
	    ifp = ex->ex_v.ex_ifPtr;
	    freeExec(ms, ifp->if_else);
	    memFree(ms, ifp, sizeof(If_t));
	    break;
	}
	case ex_while: {
	    While_t *wh;

	    wh = ex->ex_v.ex_whilePtr;
	    freeExec(ms, wh->wh_condition);
	    freeExec(ms, wh->wh_body);
	    memFree(ms, wh, sizeof(While_t));
	    break;
	}
	case ex_for: {
	    For_t *fo;

	    fo = ex->ex_v.ex_forPtr;
	    freeExec(ms, fo->fo_start);
	    freeExec(ms, fo->fo_limit);
	    freeExec(ms, fo->fo_body);
	    memFree(ms, fo, sizeof(For_t));
	    break;
	}
	case ex_case: {
	    Case_t *ca;
	    CaseList_t *cal, *calTemp;
	    CaseIndexList_t *cil, *cilTemp;

	    cal = ex->ex_v.ex_casePtr->ca_alternatives;
	    while (cal != P_NULL) {
		cil = cal->cal_indexes;
		while (cil != P_NULL) {
		    freeExec(ms, cil->cil_index);
		    cilTemp = cil;
		    cil = cil->cil_next;
		    memFree(ms, cilTemp, sizeof(CaseIndexList_t));
		}
		freeExec(ms, cal->cal_body);
		calTemp = cal;
		cal = cal->cal_next;
		memFree(ms, calTemp, sizeof(CaseList_t));
	    }
	    ca = ex->ex_v.ex_casePtr;
	    freeExec(ms, ca->ca_selector);
	    freeExec(ms, ca->ca_default);
	    memFree(ms, ca, sizeof(Case_t));
	    break;
	}
	case ex_ignore:
	case ex_ignoreS:
	case ex_negate:
	case ex_bitNot:
	case ex_fNegate:
	case ex_logNot: {
	    Expression_t *exp;

	    exp = ex->ex_v.ex_expressionPtr;
	    freeExec(ms, exp->ex_right);
	    memFree(ms, exp, sizeof(Expression_t));
	    break;
	}
	case ex_builtinCall: {
	    BuiltinCall_t *bcl;
	    ArgumentList_t *al, *alTemp;

	    bcl = ex->ex_v.ex_builtinCallPtr;
	    al = bcl->bcl_arguments;
	    while (al != P_NULL) {
		alTemp = al;
		al = al->al_next;
		freeExec(ms, alTemp->al_this);
		memFree(ms, alTemp, sizeof(ArgumentList_t));
	    }
	    memFree(ms, bcl, sizeof(BuiltinCall_t));
	    break;
	}
	case ex_knownCall:
	case ex_unknownCall: {
	    UserCall_t *ucl;
	    ArgumentList_t *al, *alTemp;

	    ucl = ex->ex_v.ex_userCallPtr;
	    al = ucl->ucl_arguments;
	    while (al != P_NULL) {
		alTemp = al;
		al = al->al_next;
		freeExec(ms, alTemp->al_this);
		memFree(ms, alTemp, sizeof(ArgumentList_t));
	    }
	    freeExec(ms, ucl->ucl_action);
	    memFree(ms, ucl, sizeof(UserCall_t));
	    break;
	}
	case ex_getBoolProp:
	case ex_getIntProp:
	case ex_getFixedProp:
	case ex_getThingProp:
	case ex_getActionProp:
	case ex_getStringProp:
	case ex_getTableProp:
	case ex_getGrammarProp:
	case ex_getIntListProp:
	case ex_getFixedListProp:
	case ex_getThingListProp:
	case ex_getActionListProp: {
	    PropGet_t *prg;

	    prg = ex->ex_v.ex_getPtr;
	    freeExec(ms, prg->prg_thing);
	    freeExec(ms, prg->prg_property);
	    memFree(ms, prg, sizeof(PropGet_t));
	    break;
	}
	case ex_addBoolProp:
	case ex_addIntProp:
	case ex_addFixedProp:
	case ex_addThingProp:
	case ex_addActionProp:
	case ex_addStringProp:
	case ex_addTableProp:
	case ex_addGrammarProp:
	case ex_addIntListProp:
	case ex_addFixedListProp:
	case ex_addThingListProp:
	case ex_addActionListProp: {
	    PropAdd_t *pra;

	    pra = ex->ex_v.ex_addPtr;
	    freeExec(ms, pra->pra_thing);
	    freeExec(ms, pra->pra_property);
	    freeExec(ms, pra->pra_value);
	    memFree(ms, pra, sizeof(PropAdd_t));
	    break;
	}
	case ex_delBoolProp:
	case ex_delIntProp:
	case ex_delFixedProp:
	case ex_delThingProp:
	case ex_delActionProp:
	case ex_delStringProp:
	case ex_delTableProp:
	case ex_delGrammarProp:
	case ex_delIntListProp:
	case ex_delFixedListProp:
	case ex_delThingListProp:
	case ex_delActionListProp: {
	    PropDel_t *prd;

	    prd = ex->ex_v.ex_delPtr;
	    freeExec(ms, prd->prd_thing);
	    freeExec(ms, prd->prd_property);
	    memFree(ms, prd, sizeof(PropDel_t));
	    break;
	}
	case ex_getIntElement:
	case ex_getFixedElement:
	case ex_getThingElement:
	case ex_getActionElement: {
	    ElemGet_t *elg;

	    elg = ex->ex_v.ex_elemGetPtr;
	    freeExec(ms, elg->elg_list);
	    freeExec(ms, elg->elg_index);
	    memFree(ms, elg, sizeof(ElemGet_t));
	    break;
	}
	case ex_putIntElement:
	case ex_putFixedElement:
	case ex_putThingElement:
	case ex_putActionElement: {
	    ElemPut_t *elp;

	    elp = ex->ex_v.ex_elemPutPtr;
	    freeExec(ms, elp->elp_list);
	    freeExec(ms, elp->elp_index);
	    freeExec(ms, elp->elp_value);
	    memFree(ms, elp, sizeof(ElemPut_t));
	    break;
	}
	case ex_shiftLeft:
	case ex_shiftRight:
	case ex_bitAnd:
	case ex_bitXor:
	case ex_bitOr:
	case ex_times:
	case ex_divide:
	case ex_remainder:
	case ex_plus:
	case ex_minus:
	case ex_equal:
	case ex_notEqual:
	case ex_less:
	case ex_greater:
	case ex_lessOrEqual:
	case ex_greaterOrEqual:
	case ex_fTimes:
	case ex_fDivide:
	case ex_fPlus:
	case ex_fMinus:
	case ex_fEqual:
	case ex_fNotEqual:
	case ex_fLess:
	case ex_fGreater:
	case ex_fLessOrEqual:
	case ex_fGreaterOrEqual:
	case ex_strPlus:
	case ex_strEqual:
	case ex_strNotEqual:
	case ex_strLess:
	case ex_strGreater:
	case ex_strLessOrEqual:
	case ex_strGreaterOrEqual:
	case ex_strEqualEqual:
	case ex_logAnd:
	case ex_logOr: {
	    Expression_t *exp;

	    exp = ex->ex_v.ex_expressionPtr;
	    freeExec(ms, exp->ex_left);
	    freeExec(ms, exp->ex_right);
	    memFree(ms, exp, sizeof(Expression_t));
	    break;
	}
	case ex_assignment: {
	    Assignment_t *as;

	    as = ex->ex_v.ex_assignmentPtr;
	    freeExec(ms, as->as_destination);
	    freeExec(ms, as->as_source);
	    memFree(ms, as, sizeof(Assignment_t));
	    break;
	}
	case ex_sequence: {
	    StatementSequence_t *sts, *stsTemp;

	    sts = ex->ex_v.ex_sequencePtr;
	    while (sts != P_NULL) {
		stsTemp = sts;
		sts = sts->sts_next;
		freeExec(ms, stsTemp->sts_this);
		memFree(ms, stsTemp, sizeof(StatementSequence_t));
	    }
	    break;
	}
	default:
	    errorString(ms, "Invalid ex_kind (");
	    errorDecimal(ms, ex->ex_kind - ex_null);
	    errorString(ms, ") in freeExec\n");
	    libAbort(ms, B_FALSE);
	    break;
	}
	memFree(ms, ex, sizeof(Exec_t));
    }
}

/*
 * freeRefList - free a ref list. May involve decrementing reference counts.
 */

static void
freeRefList(MUDState_t *ms, RefList_t *rl, ULONG_T myKey, BOOL_T freeRefs)
{
    RefList_t *rlTemp;
    ULONG_T key;

    while (rl != P_NULL) {
	rlTemp = rl;
	rl = rl->rl_next;
	key = rlTemp->rl_key;
	memFree(ms, rlTemp, sizeof(RefList_t));
	if (freeRefs && key != myKey) {
	    libDecRef(ms, key);
	}
    }
}

/*
 * freeLocalList - part of freeProc
 */

static void
freeLocalList(MUDState_t *ms, LocalList_t *ll)
{
    LocalList_t *llTemp;

    while (ll != P_NULL) {
	llTemp = ll;
	ll = ll->ll_next;
	if (llTemp->ll_name != P_NULL) {
	    charsFree(ms, llTemp->ll_name);
	}
	memFree(ms, llTemp, sizeof(LocalList_t));
    }
}

/*
 * freeProc - free a full proc entry.
 */

void
freeProc(MUDState_t *ms, Proc_t *pr, ULONG_T myKey, BOOL_T freeRefs)
{

    freeLocalList(ms, pr->pr_locals);
    freeLocalList(ms, pr->pr_parameters);
    freeExec(ms, pr->pr_body);
    freeRefList(ms, pr->pr_refs, myKey, freeRefs);
    if (pr->pr_isCompiled && pr->pr_byteCode != P_NULL) {
	memFree(ms, pr->pr_byteCode, pr->pr_byteCodeLen);
    }
    memFree(ms, pr, sizeof(Proc_t));
}

/*
 * freeProcE - extern interface to freeProc
 */

void
freeProcE(MUDState_t *ms, Proc_t *pr, ULONG_T myKey, ULONG_T freeRefs)
{

    freeProc(ms, pr, myKey, freeRefs != BOOLEAN_FALSE);
}

/*
 * clearProc - partial free of a proc as a result of errors.
 */

void
clearProc(MUDState_t *ms, Proc_t *pr, ULONG_T myKey)
{

    freeLocalList(ms, pr->pr_locals);
    pr->pr_locals = P_NULL;
    if (myKey != NIL_ACTION) {
	freeRefList(ms, pr->pr_refs, myKey, B_TRUE);
	pr->pr_refs = P_NULL;
    }
    freeExec(ms, pr->pr_body);
    pr->pr_body = MEM_ALLOC(ms, sizeof(Exec_t), "EMPT");
    pr->pr_body->ex_kind = ex_error;
}

/*
 * clearProcE - extern interface to clearProc
 */

void
clearProcE(MUDState_t *ms, Proc_t *pr, ULONG_T myKey)
{

    clearProc(ms, pr, myKey);
}
