#include "MUD.h"
#include "types.h"
#include "MUDInternal.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * symbol.c - symbol table routines.
 */

/*
 * symCreate - create a new symbol table, in the pointed to structure.
 *	'shift' indicates the initial size of the table.
 */

void
symCreate(MUDState_t *ms, Table_t *tb, UINT_T shift)
{
    Entry_t *en;

    tb->tb_owner = SYSADMIN_KEY;
    tb->tb_useCount = 0;
    tb->tb_usedCount = 0;
    tb->tb_shift = shift;
    shift = 1 << shift;
    tb->tb_mask = shift - 1;
    tb->tb_entryCount = 0;	/* in case alloc call fails. */
    en = MEM_ALLOC(ms, (ULONG_T) shift * sizeof(Entry_t), "STAB");
    tb->tb_entryCount = shift;
    tb->tb_entries = en;
    tb->tb_dirty = B_TRUE;
    while (shift != 0) {
	shift -= 1;
	en->en_nameLen = EN_EMPTY;
	++en;
    }
}

/*
 * symCreateE - extern interface to symCreate
 */

void
symCreateE(MUDState_t *ms, Table_t *tb, ULONG_T shift)
{
    symCreate(ms, tb, shift);
}

/*
 * symDestroy - destroy the indicated table. This includes freeing the
 *	strings of the symbol names.
 */

void
symDestroy(MUDState_t *ms, Table_t *tb)
{
    UINT_T n;
    Entry_t *en;

    n = tb->tb_entryCount;
    if (n != 0) {
	en = tb->tb_entries;
	while (n != 0) {
	    n -= 1;
	    if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE &&
		en->en_name != P_NULL)
	    {
		memFree(ms, en->en_name, en->en_nameLen);
	    }
	    ++en;
	}
	memFree(ms, tb->tb_entries,
		(ULONG_T) tb->tb_entryCount * sizeof(Entry_t));
	tb->tb_entryCount = 0;
    }
}

/*
 * symDestroyE - external interface to symDestroy.
 */

void
symDestroyE(MUDState_t *ms, Table_t *tb)
{
    symDestroy(ms, tb);
}

/*
 * symSearch - internal routine to find a name, or the proper slot for the
 *	name, in a symbol table.
 */

static Entry_t *
symSearch(Table_t *tb, CHAR_T *name, UINT_T nameLen)
{
    Entry_t *en, *found, *enEnd, *start;
    CHAR_T *p, *q;
    ULONG_T hash;
    UINT_T len;
    BOOL_T first;

    p = name;
    len = nameLen;

    /* Use the first two and last two characters to form a hash key. */
    hash = *p++;
    if (len > 1) {
	hash <<= 4;
	hash |= *p;
	if (len > 2) {
	    hash <<= 2;
	    p += len - 3;
	    hash |= *p;
	    if (len > 3) {
		hash <<= 1;
		hash |= *--p;
	    }
	}
    }
    en = tb->tb_entries;
    enEnd = en + (tb->tb_entryCount - 1);
    en += (hash & tb->tb_mask);

    found = P_NULL;
    start = en;
    first = B_TRUE;
    while (B_TRUE) {
	if (en == start && ! first) {
	    /* scanned entire table - name not found - use the first of the
	       free slots we passed. */
	    en = found;
	    break;
	} else if (en->en_nameLen == EN_EMPTY) {
	    /* found an unused slot - name is not in table */
	    break;
	} else if (en->en_nameLen == EN_FREE) {
	    /* found a slot that used to be used - have to keep looking, but
	       remember this slot for entry if it is the first of them */
	    if (found == P_NULL) {
		found = en;
	    }
	} else if (en->en_hash != hash || en->en_nameLen != nameLen) {
	    /* match not possible */
	} else {
	    /* have to actually compare strings */
	    p = name;
	    q = en->en_name;
	    len = nameLen;
	    while (len != 0 && *p == *q) {
		len -= 1;
		++p;
		++q;
	    }
	    if (len == 0) {
		break;
	    }
	}
	if (en == enEnd) {
	    en = tb->tb_entries;
	} else {
	    ++en;
	}
	first = B_FALSE;
    }
    if (en->en_nameLen == EN_EMPTY || en->en_nameLen == EN_FREE) {
	if (found != P_NULL) {
	    en = found;
	}
	en->en_hash = hash;
    }
    return en;
}

/*
 * symTransfer - internal routine to copy all entries from one table to
 *	another, new, table, which is created here.
 */

static void
symTransfer(MUDState_t *ms, Table_t *oldTable, Table_t *newTable, UINT_T shift)
{
    Entry_t *en;
    UINT_T count;

    symCreate(ms, newTable, shift);
    newTable->tb_owner = oldTable->tb_owner;
    newTable->tb_useCount = oldTable->tb_useCount;
    newTable->tb_usedCount = oldTable->tb_usedCount;
    count = oldTable->tb_entryCount;
    en = oldTable->tb_entries;
    while (count != 0) {
	count -= 1;
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
	    *symSearch(newTable, en->en_name, en->en_nameLen) = *en;
	}
	++en;
    }
    newTable->tb_dirty = B_TRUE;
}

/*
 * symEnter - enter a new symbol into a table. Return a pointer to the
 *	slot for it. If the symbol is already present, return a pointer
 *	to the already existing slot.
 */

Entry_t *
symEnter(MUDState_t *ms, Table_t *tb, CHAR_T *name, UINT_T nameLen)
{
    Entry_t *en;
    Table_t newTable;

    if (tb->tb_usedCount >= tb->tb_entryCount * 4 / 5) {
	symTransfer(ms, tb, &newTable, tb->tb_shift + 1);
	memFree(ms, tb->tb_entries,
		(ULONG_T) tb->tb_entryCount * sizeof(Entry_t));
	*tb = newTable;
    }
    en = symSearch(tb, name, nameLen);
    if (en->en_nameLen == EN_EMPTY || en->en_nameLen == EN_FREE) {
	en->en_name = P_NULL;
	en->en_name = MEM_ALLOC(ms, nameLen, "NAME");
	en->en_nameLen = nameLen;
	charsCopy(en->en_name, name);
	tb->tb_usedCount += 1;
	tb->tb_dirty = B_TRUE;
    }
    return en;
}

/*
 * symEnterE - extern interface to symEnter.
 */

Entry_t *
symEnterE(MUDState_t *ms, Table_t *tb, CHAR_T *name, ULONG_T nameLen)
{

    return symEnter(ms, tb, name, nameLen);
}

/*
 * symLookup - look up a symbol in the table. Return a pointer to its
 *	slot, or 'P_NULL' if it is not found.
 */

Entry_t *
symLookup(Table_t *tb, CHAR_T *name, UINT_T nameLen)
{
    Entry_t *en;

    if (tb->tb_usedCount == 0 || tb->tb_entryCount == 0) {
	return P_NULL;
    } else {
	en = symSearch(tb, name, nameLen);
	if (en->en_nameLen == EN_EMPTY || en->en_nameLen == EN_FREE) {
	    en = P_NULL;
	}
	return en;
    }
}

/*
 * symLookupE - extern interface to symLookup.
 */

Entry_t *
symLookupE(Table_t *tb, CHAR_T *name, ULONG_T nameLen)
{

    return symLookup(tb, name, nameLen);
}

/*
 * symDelete - delete the given symbol from the table, if it is found.
 *	This includes freeing the name string.
 */

void
symDelete(MUDState_t *ms, Table_t *tb, CHAR_T *name, UINT_T nameLen)
{
    Entry_t *en;

    if (tb->tb_usedCount != 0 && tb->tb_entryCount != 0) {
	en = symSearch(tb, name, nameLen);
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
	    memFree(ms, en->en_name, en->en_nameLen);
	    en->en_nameLen = EN_FREE;
	    tb->tb_usedCount -= 1;
	    tb->tb_dirty = B_TRUE;
	}
    }
}

/*
 * symDeleteE - external interface to symDelete.
 */

void
symDeleteE(MUDState_t *ms, Table_t *tb, CHAR_T *name, ULONG_T nameLen)
{

    symDelete(ms, tb, name, nameLen);
}

/*
 * symKeyDelete - delete the given key from the table.
 */

void
symKeyDelete(MUDState_t *ms, Table_t *tb, ULONG_T key)
{
    Entry_t *en;
    UINT_T count;

    if (tb->tb_usedCount != 0 && tb->tb_entryCount != 0) {
	en = tb->tb_entries;
	count = tb->tb_entryCount;
	while (count != 0) {
	    if (en->en_nameLen != EN_FREE && en->en_nameLen != EN_EMPTY &&
		en->en_key == key)
	    {
		memFree(ms, en->en_name, en->en_nameLen);
		en->en_nameLen = EN_FREE;
		tb->tb_usedCount -= 1;
		tb->tb_dirty = B_TRUE;
		return;
	    } else {
		count -= 1;
		++en;
	    }
	}
    }
}

/*
 * symReadE - routine to do the actual reading of a symbol table from a
 *	linear byte-buffer. Return the updated byte-buffer pointer.
 */

void *
symReadE(MUDState_t *ms, Table_t *tb, void *b)
{
    CHAR_T *name;
    UINT_T nameLen, count;

    count = readWord(&b);
    symCreate(ms, tb, count);
    count = readWord(&b);
    tb->tb_owner = readLong(&b);
    tb->tb_useCount = readWord(&b);
    tb->tb_dirty = B_FALSE;
    while (count != 0) {
	count -= 1;
	nameLen = readWord(&b);
	name = (CHAR_T *) b;
	b = name + nameLen;
	symEnter(ms, tb, name, nameLen)->en_key = readLong(&b);
    }
    return b;
}

/*
 * symWriteE - routine to write a symbol table as a byte stream.
 */

void
symWriteE(MUDState_t *ms, Table_t *tb)
{
    Entry_t *en;
    UINT_T nameLen, count;

    count = tb->tb_usedCount * 12 + 10;
    if (ms->WriteBufferSize < count) {
	/* Write buffer will likely need to expand while we write this table,
	   but since it must disable the calling of MemReclaim, we do a quick
	   alloc/free here to try to do any flushing that might help us stay
	   within the specified memory limit. */
	en = MEM_ALLOC(ms, count, "TEMP");
	memFree(ms, en, count);
    }
    ms->WritePos = 0;
    writeWord(ms, tb->tb_shift);
    writeWord(ms, tb->tb_usedCount);
    writeLong(ms, tb->tb_owner);
    writeWord(ms, tb->tb_useCount);
    en = tb->tb_entries;
    count = tb->tb_entryCount;
    while (count != 0) {
	count -= 1;
	nameLen = en->en_nameLen;
	if (nameLen != EN_EMPTY && nameLen != EN_FREE && en->en_name != P_NULL)
	{
	    writeWord(ms, nameLen);
	    writeBytes(ms, en->en_name, nameLen);
	    writeLong(ms, en->en_key);
	}
	++en;
    }
    tb->tb_dirty = B_FALSE;
}
