#include "MUD.h"
#include "types.h"
#include "exec.h"
#include "bcCode.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1998 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * bcComp.c - code to compile a tree structure into byte-code.
 */

#define CHAIN_NIL	0
#define	CHAIN_HERE	0xffff

#define FETCH_UINT(p)	(((UINT_T) (*(p)) << 8) | *(p + 1))
#define STORE_UINT(p, v) { \
    *(p) = (BYTE_T) (v >> 8); \
    *(p + 1) = (BYTE_T) v; \
}

static void bcComp(Exec_t *ex);

UINT_T LocalSize, FalseChain, TrueChain;
BOOL_T HadError, WantNot;

/*
 * bcWriteByte - save a few bytes.
 */

static void
bcWriteByte(BYTE_T b)
{

    writeByteE(&MS, b);
}

/*
 * bcWriteWord - a handy utility.
 */

static void
bcWriteWord(UINT_T w)
{

    writeWordE(&MS, w);
}

/*
 * bcWriteLong - a handy utility.
 */

static void
bcWriteLong(ULONG_T l)
{

    writeLongE(&MS, l);
}

/*
 * bcWriteOp - a handy utility.
 */

static void
bcWriteOp(BcCode_t op)
{

    bcWriteByte(op - bc_error);
}

/*
 * bcGenLocalOp - generate an opcode that deals with a local variable.
 *	Based on the size of the resulting offset, and the 'full' flag,
 *	we can optimize the size. Assumption: opcode order.
 */

static void
bcGenLocalOp(BcCode_t op, UINT_T offset, BOOL_T full)
{

    if (full && offset <= 8) {
	if (offset == 0) {
	    bcWriteOp(op + 2);
	} else if (offset == 4) {
	    bcWriteOp(op + 3);
	} else {
	    bcWriteOp(op + 4);
	}
    } else if (offset <= 255) {
	bcWriteOp(op + 1);
	bcWriteByte(offset);
    } else {
	bcWriteOp(op);
	bcWriteWord(offset);
    }
}

/*
 * bcGenBranchOp - we are about to generate a branch. If that branch is
 *	conditional, and WantNot is set, flip the branch and clear
 *	WantNot, then generate the branch. Note: short branches not handled.
 */

static void
bcGenBranchOp(BcCode_t op)
{

    if (WantNot) {
	WantNot = B_FALSE;
	if (op == bc_bft) {
	    op = bc_bff;
	} else if (op == bc_bbt) {
	    op = bc_bbf;
	} else if (op == bc_bff) {
	    op = bc_bft;
	} else if (op == bc_bbf) {
	    op = bc_bbt;
	}
    }
    bcWriteOp(op);
}

/*
 * bcGetOffset - return the current offset in the instruction stream.
 */

static UINT_T
bcGetOffset(void)
{
    BYTE_T *ptr;
    UINT_T len;

    writeResultE(&MS, &ptr, &len);
    return len;
}

/*
 * bcGenBranch - raw generation of a branch instruction, returning the
 *	offset of its offset portion.
 */

static UINT_T
bcGenBranch(BcCode_t op, UINT_T offset)
{
    UINT_T hereOffset;

    bcGenBranchOp(op);
    hereOffset = bcGetOffset();
    bcWriteWord(offset);
    return hereOffset;
}

/*
 * bcBranchTo - generate a full branch to the indicated target.
 */

static void
bcBranchTo(BcCode_t op, UINT_T target)
{
    UINT_T offset;

    offset = bcGetOffset();
    if (offset + 1 > target) {
	/* backwards branch */
	op += 3;
	target = offset + 3 - target;
	if (target <= 255) {
	    /* can use a short backwards branch */
	    op = (op == bc_bbt) ? bc_bbts : bc_bbfs;
	    if (WantNot) {
		WantNot = B_FALSE;
		op = (op == bc_bbts) ? bc_bbfs : bc_bbts;
	    }
	    bcWriteOp(op);
	    bcWriteByte(target - 1);	/* adjust offset for short branch */
	} else {
	    bcGenBranchOp(op);
	    bcWriteWord(target);
	}
    } else {
	target = target - offset - 3;
	bcGenBranchOp(op);
	bcWriteWord(target);
    }
}

/*
 * bcFixChain - fix a chain of branches to go to the given offset.
 */

static void
bcFixChain(UINT_T chain, UINT_T target)
{
    BYTE_T *base, *p;
    INT_T displacement;
    UINT_T len;

    /* Need base here as well as len/offset. */
    writeResultE(&MS, &base, &len);
    if (target == CHAIN_HERE) {
	target = len;
    }
    while (chain != CHAIN_NIL) {
	p = base + chain;		/* point at disp word of branch */
	chain += 2;			/* code offset of end of branch */
	displacement = target - chain;	/* displacement to target from here */
	chain = FETCH_UINT(p);		/* where next branch disp is */
	if (displacement < 0) {
	    displacement = - displacement;
	    STORE_UINT(p, displacement);
	    *(p - 1) += 3;		/* change forward branch to backward */
	} else {
	    STORE_UINT(p, displacement);/* store final 16 bit disp */
	}
    }
}

/*
 * bcJoinChains - concatenate a pair of chains.
 */

static UINT_T
bcJoinChains(UINT_T ch1, UINT_T ch2)
{
    UINT_T ch;
    BYTE_T *base, *p;
    UINT_T len;

    if (ch1 == CHAIN_NIL) {
	return ch2;
    } else if (ch2 == CHAIN_NIL) {
	return ch1;
    } else {
	writeResultE(&MS, &base, &len);
	ch = ch1;
	do {
	    p = base + ch;
	    ch = FETCH_UINT(p);
	} while (ch != CHAIN_NIL);
	STORE_UINT(p, ch2);
	return ch1;
    }
}

/*
 * bcStackBool - a value on the stack is now needed. If we have a delayed
 *	boolean NOT for it, generate that. Similarly, if we have some
 *	outstanding branches, we need to target them here and cause the
 *	proper bool value to appear.
 */

static void
bcStackBool(void)
{

    if (TrueChain != CHAIN_NIL || FalseChain != CHAIN_NIL) {
	/* Icky expression involving 'and' and/or 'or' - have to arrange
	   for a final result to be on the stack. */
	FalseChain = bcGenBranch(bc_bff, FalseChain);
	bcFixChain(TrueChain, CHAIN_HERE);
	TrueChain = CHAIN_NIL;
	bcWriteOp(bc_psh1);
	bcBranchTo(bc_bf, bcGetOffset() + 4);
	bcFixChain(FalseChain, CHAIN_HERE);
	FalseChain = CHAIN_NIL;
	bcWriteOp(bc_pshZ);
    } else {
	/* Something already on top of the stack. Perhaps flip it. */
	if (WantNot) {
	    WantNot = B_FALSE;
	    bcWriteOp(bc_bNot);
	}
    }
}

/*
 * bcGenConst - push a constant, optimizing if it is zero, etc.
 */

static void
bcGenConst(ULONG_T l)
{

    if (l == 0) {
	bcWriteOp(bc_pshZ);
    } else if (l == 1) {
	bcWriteOp(bc_psh1);
    } else if ((l & 0x00ffffff) == 0) {
	bcWriteOp(bc_pshCH);
	bcWriteByte(l >> 24);
    } else if ((LONG_T) l <= 127 && (LONG_T) l >= -128) {
	bcWriteOp(bc_pshCL);
	bcWriteByte(l);
    } else {
	bcWriteOp(bc_pshC);
	bcWriteLong(l);
    }
}

/*
 * bcQuickString - if the passed Exec_t is a string value that we can
 *	easily access, then do so without copying it, and return 'true',
 *	else return 'false'.
 */

static BOOL_T
bcQuickString(Exec_t *ex)
{

    if (ex->ex_kind == ex_string) {
	bcWriteOp(bc_pshS);
	writeStringE(&MS, ex->ex_v.ex_stringPtr);
	return B_TRUE;
    } else if (ex->ex_kind == ex_local) {
	bcGenLocalOp(bc_pshL,
		     LocalSize-ex->ex_v.ex_localPtr->ll_offset-sizeof(ULONG_T),
		     B_TRUE);
	return B_TRUE;
    } else {
	return B_FALSE;
    }
}

/*
 * bcQuickStringBinary - generate a string binary, using quick if possible.
 *	Assumption: opcodes are: normal, 00, 01, 10
 */

static void
bcQuickStringBinary(Exec_t *ex, BcCode_t op)
{
    Expression_t *exp;

    exp = ex->ex_v.ex_expressionPtr;
    if (bcQuickString(exp->ex_left)) {
	if (bcQuickString(exp->ex_right)) {
	    bcWriteOp(op + 1);
	} else {
	    bcComp(exp->ex_right);
	    bcWriteOp(op + 2);
	}
    } else {
	bcComp(exp->ex_left);
	if (bcQuickString(exp->ex_right)) {
	    bcWriteOp(op + 3);
	} else {
	    bcComp(exp->ex_right);
	    bcWriteOp(op);
	}
    }
}

/*
 * bcGenBICall - emit code to do a byte-code call to a builtin.
 */

static void
bcGenBICall(BuiltinCall_t *bcl)
{
    Proc_t *pr;
    ArgumentList_t *al;
    BOOL_T done;

    pr = bcl->bcl_proc;
    al = bcl->bcl_arguments;
    done = B_FALSE;
    if (pr->pr_resultType != TYPE_VOID) {
	/* builtin has a a result */
	if (al == P_NULL) {
	    /* no parameters */
	    bcWriteOp(bc_bi01);
	    done = B_TRUE;
	} else {
	    if (pr->pr_parameters->ll_type == TYPE_STRING &&
		al->al_next != P_NULL && al->al_next->al_next != P_NULL &&
		al->al_next->al_next->al_next == P_NULL &&
		pr->pr_parameters->ll_next->ll_type != TYPE_STRING &&
		pr->pr_parameters->ll_next->ll_next->ll_type != TYPE_STRING
	   ) {
		/* builtin has 3 parameters - the first is a string, and
		   the other two are not. We mostly want to catch SubString. */
		if (bcQuickString(al->al_this)) {
		    bcComp(al->al_next->al_this);
		    bcComp(al->al_next->al_next->al_this);
		    bcWriteOp(bc_bi31);
		    done = B_TRUE;
		}
	    }
	}
    } else {
	/* builtin has no result */
	if (al != P_NULL && al->al_next == P_NULL) {
	    /* builtin has one parameter */
	    if (pr->pr_parameters->ll_type == TYPE_STRING) {
		/* that parameter is a string */
		if (bcQuickString(al->al_this)) {
		    done = B_TRUE;
		}
	    } else {
		bcComp(al->al_this);
		done = B_TRUE;
	    }
	}
	if (done) {
	    bcWriteOp(bc_bi10);
	}
    }
    if (! done) {
	while (al != P_NULL && ! HadError) {
	    bcComp(al->al_this);
	    bcStackBool();
	    al = al->al_next;
	}
	bcWriteOp(bc_bi);
    }
    bcWriteLong(bcl->bcl_key);
}

/*
 * bcGenStaticCall - emit code for a call to a static (known) proc.
 */

static void
bcGenStaticCall(UserCall_t *ucl)
{
    ArgumentList_t *al;
    Exec_t *ex;

    al = ucl->ucl_arguments;
    while (al != P_NULL && ! HadError) {
	bcComp(al->al_this);
	bcStackBool();
	al = al->al_next;
    }
    bcWriteOp(bc_call);
    ex = ucl->ucl_action;
    if (ex->ex_kind != ex_procRef) {
	runError("bad proc ref when compiling");
	bcWriteLong(NIL_ACTION);
    } else {
	bcWriteLong(ex->ex_v.ex_refPtr->rl_key);
    }
}

/*
 * bcGenBinary - generate code for a binary operator.
 */

static void
bcGenBinary(Exec_t *ex, BcCode_t opCode)
{
    Expression_t *exp;

    exp = ex->ex_v.ex_expressionPtr;
    bcComp(exp->ex_left);
    bcComp(exp->ex_right);
    bcWriteOp(opCode);
}

/*
 * bcGenIf - generate code for an 'if' construct
 */

static void
bcGenIf(If_t *ifp)
{
    IfList_t *ifl;
    UINT_T doneChain, altChain;
    UINT_T falseChainSave, trueChainSave;
    BOOL_T first;

    falseChainSave = FalseChain;
    trueChainSave = TrueChain;
    FalseChain = CHAIN_NIL;
    TrueChain = CHAIN_NIL;
    doneChain = CHAIN_NIL;
    altChain = CHAIN_NIL;
    first = B_TRUE;
    ifl = ifp->if_alternatives;
    while (ifl != P_NULL && ! HadError) {
	if (first) {
	    first = B_FALSE;
	} else {
	    doneChain = bcGenBranch(bc_bf, doneChain);
	    bcFixChain(altChain, CHAIN_HERE);
	}
	bcComp(ifl->ifl_condition);
	altChain = bcGenBranch(bc_bff, FalseChain);
	FalseChain = CHAIN_NIL;
	bcFixChain(TrueChain, CHAIN_HERE);
	TrueChain = CHAIN_NIL;
	bcComp(ifl->ifl_body);
	bcStackBool();
	ifl = ifl->ifl_next;
    }
    if (ifp->if_else != P_NULL) {
	doneChain = bcGenBranch(bc_bf, doneChain);
	bcFixChain(altChain, CHAIN_HERE);
	bcComp(ifp->if_else);
	bcStackBool();
    } else {
	bcFixChain(altChain, CHAIN_HERE);
    }
    bcFixChain(doneChain, CHAIN_HERE);
    FalseChain = falseChainSave;
    TrueChain = trueChainSave;
}

/*
 * bcGenWhile - generate code for a 'while' statement.
 */

static void
bcGenWhile(While_t *wh)
{
    UINT_T loopPos;

    loopPos = bcGetOffset();
    if (wh->wh_body != P_NULL && wh->wh_body->ex_kind != ex_empty) {
	loopPos = bcGenBranch(bc_bf, CHAIN_NIL);
	bcComp(wh->wh_body);
	bcFixChain(loopPos, CHAIN_HERE);
	loopPos = loopPos + 2;
    }
    bcComp(wh->wh_condition);
    bcBranchTo(bc_bft, loopPos);
    bcFixChain(TrueChain, loopPos);
    TrueChain = CHAIN_NIL;
    bcFixChain(FalseChain, CHAIN_HERE);
    FalseChain = CHAIN_NIL;
}

/*
 * bcGenFor - generate code for a 'for' statement.
 */

static void
bcGenFor(For_t *fo)
{
    UINT_T varOffset, loopPos;

    varOffset = LocalSize - fo->fo_variable->ll_offset - sizeof(ULONG_T);
    bcComp(fo->fo_limit);
    bcWriteOp(bc_dup);
    bcComp(fo->fo_start);
    bcWriteOp(bc_dup);
    bcGenLocalOp(bc_popL, varOffset, B_TRUE);
    loopPos = bcGenBranch(bc_bf, CHAIN_NIL);
    if (fo->fo_body != P_NULL) {
	bcComp(fo->fo_body);
    }
    bcWriteOp(bc_dup);
    bcGenLocalOp(bc_pshL, varOffset, B_TRUE);
    bcWriteOp(bc_psh1);
    bcWriteOp(bc_add);
    bcWriteOp(bc_dup);
    bcGenLocalOp(bc_popL, varOffset, B_TRUE);
    bcFixChain(loopPos, CHAIN_HERE);
    bcWriteOp(bc_iGE);
    bcBranchTo(bc_bft, loopPos + 2);
    bcWriteOp(bc_ign);
}

/*
 * bcGetIntValue - get the value of an ex_int or ex_intRef.
 */

static LONG_T
bcGetIntValue(Exec_t *ex)
{
    LONG_T n;

    if (ex->ex_kind == ex_int) {
	return ex->ex_v.ex_integer;
    } else {
	n =
	    ((ULONG_T) ex->ex_intVal[0] << 16) |
	    ((ULONG_T) ex->ex_intVal[1] << 8) |
	    (ULONG_T) ex->ex_intVal[2];
	if ((ex->ex_intVal[0] & 0x80) != 0) {
	    n |= 0xff000000;
	}
	return n;
    }
}

/*
 * bcGenCase - generate code for a 'case' construct.
 */

static void
bcGenCase(Case_t *ca)
{
    typedef struct indexList {
	struct indexList *il_next;
	LONG_T il_value;
    } indexList_t;
    indexList_t *il;
    indexList_t **pIl;
    CaseIndexList_t *cil;
    indexList_t *indexes;
    CaseList_t *cal;
    UINT_T thisPos;
    UINT_T caseCount, doneChain, defaultChain, patchPos, codeLen, offset;
    LONG_T index, caseMin, caseMax;
    BYTE_T *codePtr;

    bcComp(ca->ca_selector);

    caseCount = 0;
    caseMin = 0x7fffffff;
    caseMax = 0x80000000;
    indexes = P_NULL;
    cal = ca->ca_alternatives;
    while (cal != P_NULL) {
	cil = cal->cal_indexes;
	while (cil != P_NULL) {
	    caseCount += 1;
	    index = bcGetIntValue(cil->cil_index);
	    if (index < caseMin) {
		caseMin = index;
	    }
	    if (index > caseMax) {
		caseMax = index;
	    }
	    il = MEM_ALLOC_E(&MS, sizeof(indexList_t), "ixls");
	    il->il_value = index;
	    pIl = &indexes;
	    while (*pIl != P_NULL && (*pIl)->il_value < index) {
		pIl = &(*pIl)->il_next;
	    }
	    il->il_next = *pIl;
	    *pIl = il;
	    cil = cil->cil_next;
	}
	cal = cal->cal_next;
    }

    if (caseMax - caseMin > 1000 ||
	(ULONG_T) (caseMax - caseMin) > caseCount * 3)
    {
	/* Use the binary search form. */

	bcWriteOp(bc_caseS);
	bcWriteLong(caseMin);
	bcWriteLong(caseMax);
	bcWriteWord(0);
	bcWriteWord(caseCount);
	patchPos = bcGetOffset();
	il = indexes;
	while (il != P_NULL) {
	    bcWriteLong(il->il_value);
	    bcWriteWord(0);
	    il = il->il_next;
	}
	doneChain = CHAIN_NIL;
	cal = ca->ca_alternatives;
	while (cal != P_NULL) {
	    writeResultE(&MS, &codePtr, &codeLen);
	    offset = codeLen - patchPos;
	    cil = cal->cal_indexes;
	    while (cil != P_NULL) {
		index = bcGetIntValue(cil->cil_index);
		thisPos = patchPos;
		il = indexes;
		while (il != P_NULL && il->il_value != index) {
		    il = il->il_next;
		    thisPos += (sizeof(LONG_T) + sizeof(UINT_T));
		}
		thisPos += sizeof(LONG_T);
		STORE_UINT(codePtr + thisPos, offset);
		cil = cil->cil_next;
	    }
	    bcComp(cal->cal_body);
	    doneChain = bcGenBranch(bc_bf, doneChain);
	    cal = cal->cal_next;
	}
	writeResultE(&MS, &codePtr, &codeLen);
	offset = codeLen - patchPos;
	STORE_UINT(codePtr + patchPos - 4, offset);
	if (ca->ca_default != P_NULL) {
	    bcComp(ca->ca_default);
	}
	bcFixChain(doneChain, CHAIN_HERE);

    } else {
	/* Use the direct index form. */

	bcWriteOp(bc_caseI);
	bcWriteLong(caseMin);
	bcWriteLong(caseMax);
	bcWriteWord(0);
	patchPos = bcGetOffset();
	defaultChain = CHAIN_NIL;
	il = indexes;
	index = caseMin;
	while (index != caseMax + 1) {
	    while (index != il->il_value) {
		index += 1;
		bcWriteWord(defaultChain);
		defaultChain = bcGetOffset() - 2;
	    }
	    bcWriteWord(0);
	    index += 1;
	    il = il->il_next;
	}
	doneChain = CHAIN_NIL;
	cal = ca->ca_alternatives;
	while (cal != P_NULL) {
	    writeResultE(&MS, &codePtr, &codeLen);
	    offset = codeLen - patchPos;
	    cil = cal->cal_indexes;
	    while (cil != P_NULL) {
		index = bcGetIntValue(cil->cil_index);
		thisPos = patchPos + (index - caseMin) * sizeof(UINT_T);
		STORE_UINT(codePtr + thisPos, offset);
		cil = cil->cil_next;
	    }
	    bcComp(cal->cal_body);
	    doneChain = bcGenBranch(bc_bf, doneChain);
	    cal = cal->cal_next;
	}
	writeResultE(&MS, &codePtr, &codeLen);
	offset = codeLen - patchPos;
	STORE_UINT(codePtr + patchPos - 2, offset);
	while (defaultChain != CHAIN_NIL) {
	    caseCount = FETCH_UINT(codePtr + defaultChain);
	    STORE_UINT(codePtr + defaultChain, offset);
	    defaultChain = caseCount;
	}
	if (ca->ca_default != P_NULL) {
	    bcComp(ca->ca_default);
	}
	bcFixChain(doneChain, CHAIN_HERE);

    }

    pIl = &indexes;
    while (B_TRUE) {
	il = *pIl;
	if (il == P_NULL) {
	    break;
	}
	*pIl = il->il_next;
	memFreeE(&MS, il, sizeof(indexList_t));
    }
}

/*
 * bcComp - the main recursive byte-code compiler.
 */

static void
bcComp(Exec_t *ex)
{
    StatementSequence_t *sts;
    Expression_t *exp;
    Assignment_t *as;
    PropGet_t *prg;
    PropAdd_t *pra;
    PropDel_t *prd;
    ElemGet_t *elg;
    ElemPut_t *elp;
    LocalList_t *ll;
    ULONG_T ul;
    UINT_T falseChainSave, trueChainSave;
    ExecKind_t kind;

    kind = ex->ex_kind;
    switch (kind) {
    case ex_byteCode:
	runError("compiling bytecode?");
	HadError = B_TRUE;
	break;
    case ex_error:
	runError("compiling erroneous executable");
	HadError = B_TRUE;
    case ex_empty:
    case ex_note:
	break;
    case ex_intRef:
    case ex_fixedRef:
	ul =
	    ((ULONG_T) ex->ex_intVal[0] << 16) |
	    ((ULONG_T) ex->ex_intVal[1] << 8) |
	    (ULONG_T) ex->ex_intVal[2];
	if ((ex->ex_intVal[0] & 0x80) != 0) {
	    ul |= 0xff000000;
	}
	bcGenConst(ul);
	break;
    case ex_procRef:
    case ex_thingRef:
    case ex_grammarRef:
    case ex_playerRef:
    case ex_propRef:
    case ex_tableRef:
	bcWriteOp(bc_pshR);
	bcWriteLong(ex->ex_v.ex_refPtr->rl_key);
	break;
    case ex_local:
	ll = ex->ex_v.ex_localPtr;
	bcGenLocalOp(bc_pshL, LocalSize - ll->ll_offset - sizeof(ULONG_T),
		     B_TRUE);
	if (ll->ll_type == TYPE_STRING) {
	    bcWriteOp(bc_sDup);
	}
	break;
    case ex_string:
	bcWriteOp(bc_pshS);
	writeStringE(&MS, ex->ex_v.ex_stringPtr);
	bcWriteOp(bc_sDup);
	break;
    case ex_int:
    case ex_fixed:
    case ex_bool:
    case ex_status:
    case ex_nil:
    case ex_ts:
	bcGenConst(ex->ex_v.ex_integer);
	break;
    case ex_if:
	bcGenIf(ex->ex_v.ex_ifPtr);
	break;
    case ex_while:
	bcGenWhile(ex->ex_v.ex_whilePtr);
	break;
    case ex_for:
	bcGenFor(ex->ex_v.ex_forPtr);
	break;
    case ex_case:
	bcGenCase(ex->ex_v.ex_casePtr);
	break;
    case ex_ignore:
	bcComp(ex->ex_v.ex_expressionPtr->ex_right);
	bcStackBool();	/* must be there! */
	bcWriteOp(bc_ign);
	break;
    case ex_ignoreS:
	bcComp(ex->ex_v.ex_expressionPtr->ex_right);
	bcStackBool();
	bcWriteOp(bc_sFree);
	break;
    case ex_builtinCall:
	bcGenBICall(ex->ex_v.ex_builtinCallPtr);
	break;
    case ex_knownCall:
	bcGenStaticCall(ex->ex_v.ex_userCallPtr);
	break;
    case ex_unknownCall:
	runError("unknown call encountered during compilation");
	break;
    case ex_getBoolProp:
    case ex_getIntProp:
    case ex_getFixedProp:
    case ex_getThingProp:
    case ex_getActionProp:
    case ex_getStringProp:
    case ex_getTableProp:
    case ex_getGrammarProp:
    case ex_getIntListProp:
    case ex_getFixedListProp:
    case ex_getThingListProp:
    case ex_getActionListProp:
	prg = ex->ex_v.ex_getPtr;
	bcComp(prg->prg_thing);
	bcComp(prg->prg_property);
	bcWriteOp(bc_getProp);
	bcWriteWord(kind - ex_getBoolProp + TYPE_BOOLEAN);
	break;
    case ex_addBoolProp:
    case ex_addIntProp:
    case ex_addFixedProp:
    case ex_addThingProp:
    case ex_addActionProp:
    case ex_addStringProp:
    case ex_addTableProp:
    case ex_addGrammarProp:
    case ex_addIntListProp:
    case ex_addFixedListProp:
    case ex_addThingListProp:
    case ex_addActionListProp:
	pra = ex->ex_v.ex_addPtr;
	bcComp(pra->pra_thing);
	bcComp(pra->pra_property);
	if (kind == ex_addStringProp && bcQuickString(pra->pra_value)) {
	    bcWriteOp(bc_addPrpS);
	} else {
	    bcComp(pra->pra_value);
	    bcStackBool();
	    bcWriteOp(bc_addProp);
	    bcWriteWord(kind - ex_addBoolProp + TYPE_BOOLEAN);
	}
	break;
    case ex_delBoolProp:
    case ex_delIntProp:
    case ex_delFixedProp:
    case ex_delThingProp:
    case ex_delActionProp:
    case ex_delStringProp:
    case ex_delTableProp:
    case ex_delGrammarProp:
    case ex_delIntListProp:
    case ex_delFixedListProp:
    case ex_delThingListProp:
    case ex_delActionListProp:
	prd = ex->ex_v.ex_delPtr;
	bcComp(prd->prd_thing);
	bcComp(prd->prd_property);
	bcWriteOp(bc_delProp);
	bcWriteWord(kind - ex_delBoolProp + TYPE_BOOLEAN);
	break;
    case ex_getIntElement:
    case ex_getFixedElement:
    case ex_getThingElement:
    case ex_getActionElement:
	elg = ex->ex_v.ex_elemGetPtr;
	bcComp(elg->elg_list);
	bcComp(elg->elg_index);
	bcWriteOp(kind - ex_getIntElement + bc_getIEl);
	break;
    case ex_putIntElement:
    case ex_putFixedElement:
    case ex_putThingElement:
    case ex_putActionElement:
	elp = ex->ex_v.ex_elemPutPtr;
	bcComp(elp->elp_list);
	bcComp(elp->elp_index);
	bcComp(elp->elp_value);
	bcWriteOp(kind - ex_putIntElement + bc_putIEl);
	break;
    case ex_negate:
    case ex_fNegate:
	ex = ex->ex_v.ex_expressionPtr->ex_right;
	if (ex->ex_kind == ex_int) {
	    bcGenConst(- ex->ex_v.ex_integer);
	} else {
	    bcComp(ex);
	    bcWriteOp(bc_neg);
	}
	break;
    case ex_bitNot:
	bcComp(ex->ex_v.ex_expressionPtr->ex_right);
	bcWriteOp(bc_not);
	break;
    case ex_shiftLeft:
	bcGenBinary(ex, bc_shl);
	break;
    case ex_shiftRight:
	bcGenBinary(ex, bc_shr);
	break;
    case ex_bitAnd:
	bcGenBinary(ex, bc_and);
	break;
    case ex_bitXor:
	bcGenBinary(ex, bc_xor);
	break;
    case ex_bitOr:
	bcGenBinary(ex, bc_ior);
	break;
    case ex_times:
	bcGenBinary(ex, bc_iMul);
	break;
    case ex_divide:
	bcGenBinary(ex, bc_iDiv);
	break;
    case ex_remainder:
	bcGenBinary(ex, bc_iRem);
	break;
    case ex_plus: {
	Exec_t *right;

	right = ex->ex_v.ex_expressionPtr->ex_right;
	if (right->ex_kind == ex_int && right->ex_v.ex_integer == 1) {
	    bcComp(ex->ex_v.ex_expressionPtr->ex_left);
	    bcWriteOp(bc_add1);
	} else {
	    bcGenBinary(ex, bc_add);
	}
	break;
    }
    case ex_fPlus:
	bcGenBinary(ex, bc_add);
	break;
    case ex_minus: {
	Exec_t *right;

	right = ex->ex_v.ex_expressionPtr->ex_right;
	if (right->ex_kind == ex_int && right->ex_v.ex_integer == 1) {
	    bcComp(ex->ex_v.ex_expressionPtr->ex_left);
	    bcWriteOp(bc_sub1);
	} else {
	    bcGenBinary(ex, bc_sub);
	}
	break;
    }
    case ex_fMinus:
	bcGenBinary(ex, bc_sub);
	break;
    case ex_equal:
    case ex_fEqual:
	bcGenBinary(ex, bc_iEQ);
	break;
    case ex_notEqual:
    case ex_fNotEqual:
	bcGenBinary(ex, bc_iNE);
	break;
    case ex_less:
    case ex_fLess:
	bcGenBinary(ex, bc_iLT);
	break;
    case ex_greater:
    case ex_fGreater:
	bcGenBinary(ex, bc_iGT);
	break;
    case ex_lessOrEqual:
    case ex_fLessOrEqual:
	bcGenBinary(ex, bc_iLE);
	break;
    case ex_greaterOrEqual:
    case ex_fGreaterOrEqual:
	bcGenBinary(ex, bc_iGE);
	break;
    case ex_fTimes:
	bcGenBinary(ex, bc_fMul);
	break;
    case ex_fDivide:
	bcGenBinary(ex, bc_fDiv);
	break;
    case ex_strPlus:
	bcGenBinary(ex, bc_sCat);
	break;
    case ex_strEqual:
	bcQuickStringBinary(ex, bc_sEQ);
	break;
    case ex_strNotEqual:
	bcQuickStringBinary(ex, bc_sNE);
	break;
    case ex_strLess:
    case ex_strGreater:
    case ex_strLessOrEqual:
    case ex_strGreaterOrEqual:
	bcGenBinary(ex, kind - ex_strLess + bc_sLT);
	break;
    case ex_strEqualEqual:
	bcQuickStringBinary(ex, bc_sEE);
	break;
    case ex_logNot:
	bcComp(ex->ex_v.ex_expressionPtr->ex_right);
	trueChainSave = TrueChain;
	TrueChain = FalseChain;
	FalseChain = trueChainSave;
	WantNot = ! WantNot;
	break;
    case ex_logAnd:
	/* Note the convention here for and/or: they leave the last part
	   of themselves as testable on the stack, but have not generated
	   a conditional branch. This is true whether the subexpression
	   below here is a simple variable, a procedure call result, or
	   and sub and/or. Note also that the assignments to TrueChain and
	   FalseChain that this code does are the only bottom-level ones
	   that generate new things for the chains - other places just
	   save and restore them when nesting something. */
#if 1
	if (TrueChain != CHAIN_NIL) {
	    runError("had TrueChain in logAnd");
	    HadError = B_TRUE;
	}
	if (FalseChain != CHAIN_NIL) {
	    runError("had FalseChain in logAnd");
	    HadError = B_TRUE;
	}
	exp = ex->ex_v.ex_expressionPtr;
	bcComp(exp->ex_left);
	falseChainSave = bcGenBranch(bc_bff, FalseChain);
	FalseChain = CHAIN_NIL;
	bcFixChain(TrueChain, CHAIN_HERE);
	TrueChain = CHAIN_NIL;
	bcComp(exp->ex_right);
	FalseChain = bcJoinChains(FalseChain, falseChainSave);
#else
	falseChainSave = FalseChain;
	FalseChain = CHAIN_NIL;
	trueChainSave = TrueChain;
	TrueChain = CHAIN_NIL;
	exp = ex->ex_v.ex_expressionPtr;
	bcComp(exp->ex_left);
	falseChainSave = bcGenBranch(bc_bff, falseChainSave);
	falseChainSave = bcJoinChains(FalseChain, falseChainSave);
	FalseChain = CHAIN_NIL;
	bcFixChain(TrueChain, CHAIN_HERE);
	TrueChain = CHAIN_NIL;
	bcComp(exp->ex_right);
	FalseChain = bcJoinChains(FalseChain, falseChainSave);
	TrueChain = bcJoinChains(TrueChain, trueChainSave);
#endif
	break;
    case ex_logOr:
#if 1
	if (TrueChain != CHAIN_NIL) {
	    runError("had TrueChain in logAnd");
	    HadError = B_TRUE;
	}
	if (FalseChain != CHAIN_NIL) {
	    runError("had FalseChain in logAnd");
	    HadError = B_TRUE;
	}
	exp = ex->ex_v.ex_expressionPtr;
	bcComp(exp->ex_left);
	trueChainSave = bcGenBranch(bc_bft, TrueChain);
	TrueChain = CHAIN_NIL;
	bcFixChain(FalseChain, CHAIN_HERE);
	FalseChain = CHAIN_NIL;
	bcComp(exp->ex_right);
	TrueChain = bcJoinChains(TrueChain, trueChainSave);
#else
	falseChainSave = FalseChain;
	FalseChain = CHAIN_NIL;
	trueChainSave = TrueChain;
	TrueChain = CHAIN_NIL;
	exp = ex->ex_v.ex_expressionPtr;
	bcComp(exp->ex_left);
	trueChainSave = bcGenBranch(bc_bft, trueChainSave);
	trueChainSave = bcJoinChains(TrueChain, trueChainSave);
	TrueChain = CHAIN_NIL;
	bcFixChain(FalseChain, CHAIN_HERE);
	FalseChain = CHAIN_NIL;
	bcComp(exp->ex_right);
	FalseChain = bcJoinChains(FalseChain, falseChainSave);
	TrueChain = bcJoinChains(TrueChain, trueChainSave);
#endif
	break;
    case ex_assignment:
	as = ex->ex_v.ex_assignmentPtr;
	bcComp(as->as_source);
	bcStackBool();
	if (as->as_destination->ex_kind == ex_local) {
	    ll = as->as_destination->ex_v.ex_localPtr;
	    bcGenLocalOp(
		(ll->ll_type == TYPE_STRING) ? bc_popS : bc_popL,
		LocalSize - ll->ll_offset - sizeof(ULONG_T),
		(BOOL_T) (ll->ll_type != TYPE_STRING));
	} else {
	    runError("illegal destination in assignment at compile time!");
	    HadError = B_TRUE;
	}
	break;
    case ex_sequence:
	sts = ex->ex_v.ex_sequencePtr;
	while (sts != P_NULL && ! HadError) {
	    bcComp(sts->sts_this);
	    sts = sts->sts_next;
	}
	break;
    default:
	runError("compile time unknown ExecKind_t");
	HadError = B_TRUE;
	break;
    }
}

/*
 * bcEmitFrees - emit bc_freeL instructions to free parameter/local strings.
 */

static void
bcEmitFrees(LocalList_t *ll)
{

    while (ll != P_NULL) {
	if (ll->ll_type == TYPE_STRING) {
	    bcWriteOp(bc_freeL);
	    bcWriteWord(LocalSize - ll->ll_offset - sizeof(ULONG_T));
	}
	ll = ll->ll_next;
    }
}

/*
 * bcSyncProc - synchronize a modified proc to the database.
 */

static void
bcSyncProc(Proc_t *pr, ULONG_T key)
{
    BYTE_T *p;
    UINT_T len;

    writeProcE(&MS, pr);
    writeResultE(&MS, &p, &len);
    ioReplace(key, p, len);
}

/*
 * Compile - the builtin to compile to byte-code.
 */

void
Compile(ULONG_T key)
{
    LocalList_t *ll;
    UINT_T parSize, localCount;
    Proc_t *pr;
    UINT_T *pLockCount;
    BYTE_T *p;
    UINT_T len;

    TRACE("Compile", key);
    if (key == NIL_ACTION) {
	runError("nil action to Compile");
    } else {
	pr = findCachedProc(key, &pLockCount);
	if (pr->pr_isCompiled) {
	    runError("proc is already compiled to byte-code");
	} else {
	    HadError = B_FALSE;
	    WantNot = B_FALSE;
	    LocalSize = pr->pr_localCount;
	    writeResetE(&MS);

	    /* function prolog */
	    ll = pr->pr_parameters;
	    parSize = 0;
	    while (ll != P_NULL) {
		parSize += sizeof(QValue_t);
		ll = ll->ll_next;
	    }
	    if (pr->pr_localCount == parSize) {
		localCount = 0;
	    } else {
		/* space already left for fp and pc */
		localCount = pr->pr_localCount - parSize - 2 * sizeof(ULONG_T);
	    }
	    if (localCount != 0) {
		bcWriteOp(bc_spDec);
		bcWriteWord(localCount);
	    }
	    ll = pr->pr_locals;
	    while (ll != P_NULL) {
		if (ll->ll_type == TYPE_STRING) {
		    bcWriteOp(bc_initL);
		    bcWriteWord(LocalSize - ll->ll_offset - sizeof(ULONG_T));
		}
		ll = ll->ll_next;
	    }

	    /* function body */
	    if (pr->pr_body != P_NULL) {
		bcComp(pr->pr_body);
		bcStackBool();
	    }

	    /* function epilog */
	    bcEmitFrees(pr->pr_parameters);
	    bcEmitFrees(pr->pr_locals);
	    bcWriteOp((pr->pr_resultType == TYPE_VOID) ? bc_ret : bc_ret1);
	    bcWriteWord(localCount);
	    bcWriteWord(parSize);

	    if (! HadError) {
		writeResultE(&MS, &p, &len);
		pr->pr_byteCode = MEM_ALLOC_E(&MS, len, "BT-C");
		copy(pr->pr_byteCode, p, len);
		pr->pr_byteCodeLen = len;
		pr->pr_isCompiled = B_TRUE;
		bcSyncProc(pr, key);
	    }
	}
	*pLockCount -= 1;
    }
}

void
UnCompile(ULONG_T key)
{
    Proc_t *pr;

    TRACE("UnCompile", key);
    if (key == NIL_ACTION) {
	runError("nil action to UnCompile");
    } else {
	pr = findCachedProcNoLock(key);
	if (! pr->pr_isCompiled) {
	    runError("proc had no byte-code to UnCompile");
	} else if (pr->pr_locked) {
	    runError("proc is byte-code locked and cannot be UnCompiled");
	} else {
	    memFreeE(&MS, pr->pr_byteCode, pr->pr_byteCodeLen);
	    pr->pr_byteCode = P_NULL;
	    pr->pr_byteCodeLen = 0;
	    pr->pr_isCompiled = B_FALSE;
	    bcSyncProc(pr, key);
	}
    }
}

/*
 * StripBody - builtin to remove the non-byte-code body from a proc. After
 *	this has been done, the proc cannot be edited or recompiled.
 */

void
StripBody(ULONG_T key)
{
    LocalList_t *ll;
    Proc_t *pr;
    UINT_T *pLockCount;
    BYTE_T *oldByteCode;

    TRACE("StripBody", key);
    if (key == NIL_ACTION) {
	runError("nil action to StripBody");
    } else {
	pr = findCachedProc(key, &pLockCount);
	if (pr->pr_stripped) {
	    runError("action is already stripped");
	} else if (! pr->pr_isCompiled) {
	    runError("action has no byte-code - cannot strip");
	} else {
	    /* Tricky point here. If the byte-code has been run, then it
	       may have had some optimizations done to it, where keys are
	       replaced by addresses. We do not want to store that version
	       in the database. So, if the code has been run, we need to
	       remove and regenerate a fresh copy of the code, before
	       stripping the body and writing the result to the database.
	       However, if some other compiled code points at this code,
	       we cannot move it. So, we save it, regenerate afresh, strip
	       the body, save to DB, discard new byte-code and re-insert
	       the old byte-code pointer. */
	    oldByteCode = P_NULL;	/* keep gcc happy */
	    if (pr->pr_locked) {
		oldByteCode = pr->pr_byteCode;
		pr->pr_isCompiled = B_FALSE;
		Compile(key);
	    }
	    /* Pass NIL_ACTION - want to keep the reflist. */
	    clearProcE(&MS, pr, NIL_ACTION);
	    pr->pr_body->ex_kind = ex_byteCode;
	    pr->pr_stripped = B_TRUE;
	    ll = pr->pr_parameters;
	    while (ll != P_NULL) {
		freeString(ll->ll_name);
		ll->ll_name = P_NULL;
		ll = ll->ll_next;
	    }
	    bcSyncProc(pr, key);
	    if (pr->pr_locked) {
		memFreeE(&MS, pr->pr_byteCode, pr->pr_byteCodeLen);
		pr->pr_byteCode = oldByteCode;
	    }
	}
	*pLockCount -= 1;
    }
}
