#include "MUD.h"
#include "types.h"
#include "exec.h"
#include "bcCode.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1998 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * bcDis.c - code to disassemble byte-code.
 */

typedef enum {
    bct_noargs,
    bct_local,
    bct_localS,
    bct_local0,
    bct_local4,
    bct_local8,
    bct_loByte,
    bct_hiByte,
    bct_word,
    bct_twoWord,
    bct_string,
    bct_const,
    bct_key,
    bct_addr,
    bct_type,
    bct_branchf,
    bct_branchb,
    bct_branchbs,
    bct_caseS,
    bct_caseI
} BcType_t;

typedef struct {
    char *bco_name;
    BcType_t bco_type;
} BcOp_t;

static BcOp_t OpTable[] = {
    {"error  "	, bct_noargs},
    {"ret    "	, bct_twoWord},
    {"ret1   "	, bct_twoWord},
    {"spDec  "	, bct_word},
    {"bi     "	, bct_key},
    {"bi01   "	, bct_key},
    {"BI01   "	, bct_addr},
    {"bi10   "	, bct_key},
    {"BI10   "	, bct_addr},
    {"bi31   "	, bct_key},
    {"BI31   "	, bct_addr},
    {"call   "	, bct_key},
    {"JSR    "	, bct_addr},
    {"CALL   "	, bct_addr},
    {"pshL   "	, bct_local},
    {"pshLs  "	, bct_localS},
    {"pshL0  "	, bct_local0},
    {"pshL4  "	, bct_local4},
    {"pshL8  "	, bct_local8},
    {"popL   "	, bct_local},
    {"popLs  "	, bct_localS},
    {"popL0  "	, bct_local0},
    {"popL4  "	, bct_local4},
    {"popL8  "	, bct_local8},
    {"popS   "	, bct_local},
    {"popSs  "	, bct_localS},
    {"initL  "	, bct_local},
    {"freeL  "	, bct_local},
    {"pshC   "	, bct_const},
    {"pshZ   "	, bct_noargs},
    {"pshCL  "	, bct_loByte},
    {"pshCH  "	, bct_hiByte},
    {"psh1   "	, bct_noargs},
    {"pshS   "	, bct_string},
    {"pshR   "	, bct_key},
    {"ign    "	, bct_noargs},
    {"dup    "	, bct_noargs},
    {"add    "	, bct_noargs},
    {"add1   "	, bct_noargs},
    {"sub    "	, bct_noargs},
    {"sub1   "	, bct_noargs},
    {"iMul   "	, bct_noargs},
    {"iDiv   "	, bct_noargs},
    {"iRem   "	, bct_noargs},
    {"shl    "	, bct_noargs},
    {"shr    "	, bct_noargs},
    {"neg    "	, bct_noargs},
    {"not    "	, bct_noargs},
    {"and    "	, bct_noargs},
    {"ior    "	, bct_noargs},
    {"xor    "	, bct_noargs},
    {"fMul   "	, bct_noargs},
    {"fDiv   "	, bct_noargs},
    {"sFree  "	, bct_noargs},
    {"sDup   "	, bct_noargs},
    {"sCat   "	, bct_noargs},
    {"iLT    "	, bct_noargs},
    {"iLE    "	, bct_noargs},
    {"iEQ    "	, bct_noargs},
    {"iNE    "	, bct_noargs},
    {"iGE    "	, bct_noargs},
    {"iGT    "	, bct_noargs},
    {"sEQ    "	, bct_noargs},
    {"sEQ00  "	, bct_noargs},
    {"sEQ01  "	, bct_noargs},
    {"sEQ10  "	, bct_noargs},
    {"sNE    "	, bct_noargs},
    {"sNE00  "	, bct_noargs},
    {"sNE01  "	, bct_noargs},
    {"sNE10  "	, bct_noargs},
    {"sLT    "	, bct_noargs},
    {"sGT    "	, bct_noargs},
    {"sLE    "	, bct_noargs},
    {"sGE    "	, bct_noargs},
    {"sEE    "	, bct_noargs},
    {"sEE00  "	, bct_noargs},
    {"sEE01  "	, bct_noargs},
    {"sEE10  "	, bct_noargs},
    {"bNot   "	, bct_noargs},
    {"bf     "	, bct_branchf},
    {"bft    "	, bct_branchf},
    {"bff    "	, bct_branchf},
    {"bb     "	, bct_branchb},
    {"bbt    "	, bct_branchb},
    {"bbf    "	, bct_branchb},
    {"bbts   "	, bct_branchbs},
    {"bbfs   "	, bct_branchbs},
    {"getProp"	, bct_type},
    {"addProp"	, bct_type},
    {"addPrpS"	, bct_noargs},
    {"delProp"	, bct_type},
    {"getIEl "	, bct_noargs},
    {"getFEl "	, bct_noargs},
    {"getTEl "	, bct_noargs},
    {"getAEl "	, bct_noargs},
    {"putIEl "	, bct_noargs},
    {"putFEl "	, bct_noargs},
    {"putTEl "	, bct_noargs},
    {"putAEl "	, bct_noargs},
    {"caseI  "	, bct_caseI},
    {"caseS  "	, bct_caseS}
};

static void
fmtHexWord(UINT_T n)
{
    char *p;
    UINT_T cnt;
    BYTE_T b;
    char buff[5];

    p = &buff[4];
    *p = '\0';
    for (cnt = 0; cnt != 4; cnt += 1) {
	b = (BYTE_T) n & 0xf;
	n >>= 4;
	*--p = (b > 9) ? (b - 10 + 'a') : (b + '0');
    }
    fmtString(p);
}

/*
 * showLocal - display a local variable name, or its offset.
 */

static void
showLocal(Proc_t *pr, UINT_T offset)
{
    LocalList_t *ll;

    offset = pr->pr_localCount - offset - sizeof(ULONG_T);
    ll = pr->pr_parameters;
    while (ll != P_NULL && ll->ll_offset != offset) {
	ll = ll->ll_next;
    }
    if (ll == P_NULL) {
	ll = pr->pr_locals;
	while (ll != P_NULL && ll->ll_offset != offset) {
	    ll = ll->ll_next;
	}
    }
    if (ll == P_NULL || ll->ll_name == P_NULL) {
	fmtInteger(pr->pr_localCount - offset - sizeof(ULONG_T));
    } else {
	fmtString(ll->ll_name);
    }
}

/*
 * doDis - the real work is done here.
 */

static void
doDis(Proc_t *pr)
{
    BYTE_T *p;
    BcOp_t *bco;
    UINT_T len, count;
    BYTE_T op;
    ULONG_T ul;
    UINT_T offset, ui;
    LONG_T min, max, curr;
    UINT_T saveCount;
    BOOL_T byteSwapped;

#define FETCH_AND_INC_BYTE() { \
    ui = *p++; \
}

#if UNALIGNED_OK

#define FETCH_AND_INC_UINT() { \
    if (byteSwapped) { \
	ui = *(UINT_T *) p; \
	p += 2; \
    } else { \
	ui = *p++; \
	ui <<= 8; \
	ui |= *p++; \
    } \
}

#define FETCH_AND_INC_ULONG() { \
    if (byteSwapped) { \
	ul = *(ULONG_T *) p; \
	p += 4; \
    } else { \
	ul = *p++; \
	ul <<= 8; \
	ul |= *p++; \
	ul <<= 8; \
	ul |= *p++; \
	ul <<= 8; \
	ul |= *p++; \
    } \
}

#else

#define FETCH_AND_INC_UINT() { \
    ui = *p++; \
    ui <<= 8; \
    ui |= *p++; \
}
#define FETCH_AND_INC_ULONG() { \
    ul = *p++; \
    ul <<= 8; \
    ul |= *p++; \
    ul <<= 8; \
    ul |= *p++; \
    ul <<= 8; \
    ul |= *p++; \
}

#endif

#define BUMP(n) { \
    len -= (n); \
    offset += (n); \
}

    p = pr->pr_byteCode;
    len = pr->pr_byteCodeLen;
    byteSwapped = pr->pr_byteSwapped;
    offset = 0;
    while (len != 0) {
	fmtHexWord(offset);
	fmtString(":   ");
	op = *p++;
	BUMP(1);
	if (op > bc_caseS) {
	    runError("Invalid byte-code opcode in disassembly\n");
	    return;
	}
	bco = &OpTable[op];
	fmtString(bco->bco_name);
	if (bco->bco_type != bct_noargs) {
	    fmtString(" ");
	    switch (bco->bco_type) {
	    case bct_noargs:
		break;
	    case bct_word:
		FETCH_AND_INC_UINT();
		fmtInteger(ui);
		BUMP(2);
		break;
	    case bct_local:
		FETCH_AND_INC_UINT();
		showLocal(pr, ui);
		BUMP(2);
		break;
	    case bct_localS:
		FETCH_AND_INC_BYTE();
		showLocal(pr, ui);
		BUMP(1);
		break;
	    case bct_local0:
		showLocal(pr, 0);
		break;
	    case bct_local4:
		showLocal(pr, 4);
		break;
	    case bct_local8:
		showLocal(pr, 8);
		break;
	    case bct_loByte:
		FETCH_AND_INC_BYTE();
		fmtString("0x");
		fmtHex((ULONG_T) (signed char) ui);
		BUMP(1);
		break;
	    case bct_hiByte:
		FETCH_AND_INC_BYTE();
		fmtString("0x");
		fmtHex((ULONG_T) ui << 24);
		BUMP(1);
		break;
	    case bct_twoWord:
		FETCH_AND_INC_UINT();
		fmtChar('L');
		fmtInteger(ui);
		fmtString(",P");
		FETCH_AND_INC_UINT();
		fmtInteger(ui);
		BUMP(4);
		break;
	    case bct_string:
		fmtStringConst((char *) p);
		while (*p++ != 0) {
		    BUMP(1);
		}
		BUMP(1);
		break;
	    case bct_const:
		fmtString("0x");
		FETCH_AND_INC_ULONG();
		fmtHex(ul);
		BUMP(4);
		break;
	    case bct_key:
		FETCH_AND_INC_ULONG();
		fmtAnyName(ul, ul >> TYPE_SHIFT);
		BUMP(4);
		break;
	    case bct_addr:
		FETCH_AND_INC_ULONG();
		fmtString("0x");
		fmtHex(ul);
		BUMP(4);
		break;
	    case bct_type:
		FETCH_AND_INC_UINT();
		fmtType(ui);
		BUMP(2);
		break;
	    case bct_branchf:
		FETCH_AND_INC_UINT();
		BUMP(2);
		fmtHexWord(offset + ui);
		break;
	    case bct_branchb:
		FETCH_AND_INC_UINT();
		BUMP(2);
		fmtHexWord(offset - ui);
		break;
	    case bct_branchbs:
		FETCH_AND_INC_BYTE();
		BUMP(1);
		fmtHexWord(offset - ui);
		break;
	    case bct_caseS:
		FETCH_AND_INC_ULONG();
		fmtInteger(ul);
		fmtCheck("->");
		FETCH_AND_INC_ULONG();
		fmtInteger(ul);
		fmtCheck(" default ");
		FETCH_AND_INC_UINT();
		BUMP(12);
		fmtHexWord(offset + ui);
		fmtCheck(" count ");
		FETCH_AND_INC_UINT();
		count = ui;
		fmtInteger(count);
		saveCount = count;
		while (count != 0) {
		    fmtLine();
		    fmtCheck("                ");
		    FETCH_AND_INC_ULONG();
		    fmtInteger(ul);
		    fmtCheck("=>");
		    FETCH_AND_INC_UINT();
		    fmtHexWord(offset + ui);
		    count -= 1;
		}
		BUMP(saveCount * 6);
		break;
	    case bct_caseI:
		FETCH_AND_INC_ULONG();
		fmtInteger(ul);
		min = ul;
		fmtCheck("->");
		FETCH_AND_INC_ULONG();
		fmtInteger(ul);
		max = ul;
		fmtCheck(" default ");
		FETCH_AND_INC_UINT();
		BUMP(10);
		fmtHexWord(offset + ui);
		curr = min;
		while (B_TRUE) {
		    fmtLine();
		    fmtCheck("                [");
		    fmtInteger(curr);
		    fmtCheck("]=>");
		    FETCH_AND_INC_UINT();
		    fmtHexWord(offset + ui);
		    if (curr == max) {
			break;
		    }
		    curr += 1;
		}
		BUMP((max - min + 1) * 2);
		break;
	    }
	}
	fmtLine();
    }
#undef FETCH_AND_INC_ULONG
#undef FETCH_AND_INC_UINT
#undef FETCH_AND_INC_BYTE
}

/*
 * Disassemble - the builtin entry point.
 */

void
Disassemble(ULONG_T key)
{
    Proc_t *pr;

    TRACE("Disassemble", key);
    if (key == NIL_ACTION) {
	runError("nil action to Disassemble");
    } else {
	pr = findCachedProcNoLock(key);
	if (! pr->pr_isCompiled) {
	    runError("action has no byte-code - cannot disassemble");
	} else if (pr->pr_byteCode == P_NULL) {
	    runError("action has nil byte-code pointer!");
	} else if (pr->pr_owner != EffectivePlayer && ! pr->pr_public &&
	    (EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY))
	{
	    runError("action is not public - cannot disassemble");
	} else {
	    fmtStart();
	    doDis(pr);
	}
    }
}

#if UNALIGNED_OK && MUD_ENDIAN == MUD_LITTLE_ENDIAN

/* This is in this source file so that it can make use of the same opcode
   description stuff that the disasembler uses. */

/*
 * bcByteSwap - on a machine that allows unaligned accesses, and which is
 *	a little-endian machine, byteswap all 2 and 4 byte numbers in the
 *	given byte-code sequence.
 */

void
bcByteSwap(BYTE_T *p, UINT_T len)
{
    BcOp_t *bco;
    ULONG_T ul;
    LONG_T min, max;
    UINT_T ui, count;
    BYTE_T op;

#define SWAP_AND_INC_UINT() { \
    ui = *p++; \
    ui <<= 8; \
    ui |= *p++; \
    *(UINT_T *) (p - 2) = ui; \
}
#define SWAP_AND_INC_ULONG() { \
    ul = *p++; \
    ul <<= 8; \
    ul |= *p++; \
    ul <<= 8; \
    ul |= *p++; \
    ul <<= 8; \
    ul |= *p++; \
    *(ULONG_T *) (p - 4) = ul; \
}

    while (len != 0) {
	op = *p++;
	len -= 1;
	if (op > bc_caseS) {
	    runError("Invalid byte-code opcode in byteswap");
	    return;
	}
	bco = &OpTable[op];
	switch (bco->bco_type) {
	case bct_noargs:
	case bct_local0:
	case bct_local4:
	case bct_local8:
	    break;
	case bct_localS:
	case bct_loByte:
	case bct_hiByte:
	case bct_branchbs:
	    ++p;
	    len -= 1;
	    break;
	case bct_word:
	case bct_local:
	case bct_type:
	case bct_branchf:
	case bct_branchb:
	    SWAP_AND_INC_UINT();
	    len -= 2;
	    break;
	case bct_twoWord:
	    SWAP_AND_INC_UINT();
	    SWAP_AND_INC_UINT();
	    len -= 4;
	    break;
	case bct_const:
	case bct_key:
	case bct_addr:
	    SWAP_AND_INC_ULONG();
	    len -= 4;
	    break;
	case bct_string:
	    while (*p++ != 0) {
		len -= 1;
	    }
	    len -= 1;
	    break;
	case bct_caseS:
	    SWAP_AND_INC_ULONG();
	    SWAP_AND_INC_ULONG();
	    SWAP_AND_INC_UINT();
	    SWAP_AND_INC_UINT();
	    count = *(UINT_T *) (p - 2);
	    len -= 12;
	    len -= count * 6;
	    while (count != 0) {
		SWAP_AND_INC_ULONG();
		SWAP_AND_INC_UINT();
		count -= 1;
	    }
	    break;
	case bct_caseI:
	    SWAP_AND_INC_ULONG();
	    min = *(LONG_T *) (p - 4);
	    SWAP_AND_INC_ULONG();
	    max = *(LONG_T *) (p - 4);
	    SWAP_AND_INC_UINT();
	    len -= 10;
	    len -= (max - min + 1) * 2;
	    while (B_TRUE) {
		SWAP_AND_INC_UINT();
		if (min == max) {
		    break;
		}
		min += 1;
	    }
	    break;
	default:
	    runError("Bad kind case in bytecode byteswap");
	    break;
	}
    }
#undef SWAP_AND_INC_UINT
#undef SWAP_AND_INC_ULONG
}

#endif
