#include "MUD.h"
#include "types.h"
#include "exec.h"
#include "bcCode.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1998 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * bcRun.c - code to execute byte-code.
 */

#define STACK_SIZE	8192

#if UNALIGNED_OK

#define FETCH_ULONG_PC()	(*(ULONG_T *) pc)

#define STORE_ULONG_PC(v)	*(ULONG_T *) pc = v;

#else

#define FETCH_ULONG_PC() \
    (((ULONG_T) *pc) << 24) | \
    (((ULONG_T) *(pc + 1)) << 16) | \
    (((ULONG_T) *(pc + 2)) << 8) | \
    ((ULONG_T) *(pc + 3))

#define STORE_ULONG_PC(v) { \
    *pc = (BYTE_T) ((v) >> 24); \
    *(pc + 1) = (BYTE_T) ((v) >> 16); \
    *(pc + 2) = (BYTE_T) ((v) >> 8); \
    *(pc + 3) = (BYTE_T) (v); \
}

#endif

/* Stack grows downwards, which works best for most architectures. At
   least for the 68K it definitely does. This means that the parameters/
   locals are in reverse memory order from what they are with the
   tree interpreter. The traceback code has to know about that. */
/* Stack pointer points to bottom used slot. So, to push something, it is
   pre-decrement and to pop something it is post-increment. */
/* FP (and the local copy, fp) isn't modified very often. The global FP
   is updated whenever it is. Note that on a recursive call into bytecode,
   PC, SP and FP are saved/restored/initialized.
   PC needs to be valid for a traceback, so it must be updated from a
   local copy whenever bcRun returns (since there is a run-time check after
   it returns), or before any call which can or does trigger a traceback.
   SP isn't needed for tracebacks, but must be updated in any situation
   where a recursive call to bcRun (via doBcRun) can happen. */

static Proc_t *CurrentProc;
static ULONG_T *Stack, *StackTop, *SP, *FP;
static BYTE_T *PC;

/*
 * bcInit - initialization done at start of server run.
 */

void
bcInit(void)
{

    Stack = MEM_ALLOC_E(&MS, STACK_SIZE, "STAK");
    StackTop = (ULONG_T *) ((BYTE_T *) Stack + STACK_SIZE);
    SP = StackTop;
    FP = SP;
    CurrentProc = P_NULL;
}

/*
 * bcTerm - cleanup done at end of server run.
 */

void
bcTerm(BOOL_T freeMemory)
{

    if (freeMemory) {
	memFreeE(&MS, Stack, STACK_SIZE);
    }
}

/*
 * bcTraceBack - do one step of a traceback through bytecode calls
 */

void
bcTraceBack(TraceBack_t *tb)
{
    Proc_t *pr;
    LocalList_t *ll;
    ULONG_T *fp;
    char *p;
    BYTE_T *pc;
    UINT_T localCount, parCount, offset, i;
    BYTE_T digit;
    ULONG_T procKey;
    char buff[5];

    fp = FP;
    pc = PC;
    while (B_TRUE) {
	pr = findProcByPC(pc, &procKey);
	if (pr == P_NULL) {
	    fmtString("Cannot find action with byte-code for PC - "
		      "traceback aborted.");
	    fmtLine();
	    break;
	}
	offset = pc - pr->pr_byteCode;
	p = &buff[4];
	*p = '\0';
	for (i = 0; i != 4; i += 1) {
	    digit = offset & 0xf;
	    *--p = (digit >= 10) ? (digit - 10 + 'a') : (digit + '0');
	    offset >>= 4;
	}
	fmtCheck(p);
	fmtChar('+');
	traceName(procKey, "ACTION");
	fmtChar('(');
	localCount = pr->pr_localCount - sizeof(ULONG_T);
	ll = pr->pr_parameters;
	parCount = 0;
	while (ll != P_NULL) {
	    tbShowValue((QValue_t *) ((BYTE_T *) fp + localCount -
				      ll->ll_offset), ll);
	    parCount = parCount + sizeof(QValue_t);
	    ll = ll->ll_next;
	    if (ll != P_NULL) {
		fmtCheck(",");
		fmtCheck(" ");
	    }
	}
	fmtChar(')');
	fmtLine();
	fp = (ULONG_T *) ((BYTE_T *) fp + localCount);
	pc = *(BYTE_T **) ((BYTE_T *) fp - parCount - sizeof(ULONG_T));
	fp = *(ULONG_T **) ((BYTE_T *) fp - parCount);
	if (fp == tb->tb_FP) {
	    break;
	}
    }
    /* Ready for any next step of traceback */
    PC = tb->tb_PC;
    SP = tb->tb_SP;
    FP = tb->tb_FP;
}

/*
 * builtinCall - call to the given native-code builtin function. This
 *	code is responsible for freeing of string parameters. It must also
 *	pop the arguments from the stack, and leave any result pushed
 *	onto the stack.
 */

static ULONG_T *
builtinCall(ULONG_T key, ULONG_T *sp)
{
    Proc_t *pr;
    LocalList_t *ll;
    Comp_t comp;
    LONG_T result;
    UINT_T offset;
    TraceBack_t tb;
    BOOL_T isVoid;

    SP = sp;	/* needed in case the builtin calls more byte-code */
    pr = findBuiltIn(key);
    tb.tb_byteCode = B_TRUE;
    tb.tb_next = RunTraceBack;
    tb.tb_proc = pr;
    tb.tb_key = key;
    tb.tb_PC = PC;
    tb.tb_SP = sp;
    tb.tb_FP = FP;
    RunTraceBack = &tb;
    isVoid = (BOOL_T) (pr->pr_resultType == TYPE_VOID);
    comp = pr->pr_compiled;
    switch (pr->pr_localCount / sizeof(LONG_T)) {
    case 0:
	result = (*(LONG_T (*)(void)) comp)();
	break;
    case 1:
	result = (*(LONG_T (*)(LONG_T)) comp)(*sp);
	break;
    case 2:
	result = (*(LONG_T (*)(LONG_T, LONG_T)) comp)(*(sp + 1), *sp);
	break;
    case 3:
	result =
	    (*(LONG_T (*)(LONG_T, LONG_T, LONG_T)) comp)
		(*(sp + 2), *(sp + 1), *sp);
	break;
    case 4:
	result =
	    (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T)) comp)
		(*(sp + 3), *(sp +  2), *(sp + 1), *sp);
	break;
    case 5:
	result =
	    (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T, LONG_T)) comp)
		(*(sp + 4), *(sp + 3), *(sp + 2), *(sp + 1), *sp);
	break;
    case 6:
	result =
	    (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T, LONG_T, LONG_T))comp)
		(*(sp + 5), *(sp + 4), *(sp + 3), *(sp + 2), *(sp + 1), *sp);
	break;
    case 7:
	result =
	    (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T, LONG_T, LONG_T,
			LONG_T)) comp)
		(*(sp + 6), *(sp + 5), *(sp + 4), *(sp + 3), *(sp + 2),
		 *(sp + 1), *sp);
	break;
    case 8:
	result =
	    (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T, LONG_T, LONG_T,
			LONG_T, LONG_T)) comp)
		(*(sp + 7), *(sp + 6), *(sp + 5), *(sp + 4), *(sp + 3),
		 *(sp + 2), *(sp + 1), *sp);
	break;
    case 9:
	result =
	    (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T, LONG_T, LONG_T,
			LONG_T, LONG_T, LONG_T)) comp)
		(*(sp + 8), *(sp + 7), *(sp + 6), *(sp + 5), *(sp + 4),
		 *(sp + 3), *(sp + 2), *(sp + 1), *sp);
	break;
    case 10:
	result =
	    (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T, LONG_T, LONG_T,
			LONG_T, LONG_T, LONG_T, LONG_T)) comp)
		(*(sp + 9), *(sp + 8), *(sp + 7), *(sp + 6), *(sp + 5),
		 *(sp + 4), *(sp + 3), *(sp + 2), *(sp + 1), *sp);
	break;
    default:
	procError(key, "too many pars to compiled action");
	result = 0;
	break;
    }
    ll = pr->pr_parameters;
    offset = pr->pr_localCount;
    while (ll != P_NULL) {
	offset -= sizeof(QValue_t);
	if (ll->ll_type == TYPE_STRING) {
	    freeString(*(char **) ((BYTE_T *) sp + offset));
	}
	ll = ll->ll_next;
    }
    sp = (ULONG_T *) ((BYTE_T *) sp + pr->pr_localCount);
    if (! isVoid) {
	*--sp = result;
    }
    RunTraceBack = tb.tb_next;
    return sp;
}

/*
 * doStaticCall - this routine is called when a directly specified routine
 *	is to be called, but that routine is not compiled into byte-code.
 *	Our responsibilities are very similar to those of builtinCall.
 */

static ULONG_T *
doStaticCall(ULONG_T key, ULONG_T *sp)
{
    LocalList_t *ll;
    QValue_t *sl;
    Proc_t *pr;
    UINT_T *pLockCount;
    QValue_t *newLocals, *localsSave;
    ULONG_T localCount;
    UINT_T parCount;
    TraceBack_t tb;
    ULONG_T oldPlayer;
    PlayerStatus_t oldStatus;
    QValue_t quickLocals[QUICK_LOCAL_COUNT];
    QValue_t res;
    UINT_T resultType, resType;
    BOOL_T hadError;

    SP = sp;	/* needed in case the routine calls more byte-code */
    pr = findCachedProc(key, &pLockCount);

    parCount = 0;
    ll = pr->pr_parameters;
    while (ll != P_NULL) {
	ll = ll->ll_next;
	parCount += sizeof(QValue_t);
    }

    hadError = B_FALSE;
    resultType = pr->pr_resultType;
    RunCallDepth += 1;
    if (RunCallDepth > MAX_CALL_DEPTH) {
	runError("action call depth limit exceeded");
	hadError = B_TRUE;
    } else if (pr->pr_wizard && EffectiveStatus != ps_wizard) {
	procError(key, "apprentice cannot call this action");
	hadError = B_TRUE;
    } else {

	localCount = pr->pr_localCount;
	if (localCount != BC_STACK_EXTRA) {
	    if (localCount <= QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		newLocals = &quickLocals[0];
	    } else {
		newLocals = MEM_ALLOC_E(&MS, localCount, "lclu");
	    }
	} else {
	    if (pr->pr_parameters != P_NULL) {
		abort("got parameters with 0 localCount!");
	    }
	    if (pr->pr_locals != P_NULL) {
		abort("got locals with 0 localCount!");
	    }
	    newLocals = &quickLocals[0];    /* keep gcc quiet */
	}
	checkAbort();

	if (! RunExiting) {
	    sl = newLocals;

	    /* Copy parameters from stack to locals array in forward order */
	    sp = (ULONG_T *) ((BYTE_T *) sp + parCount);
	    ll = pr->pr_parameters;
	    while (ll != P_NULL) {
		sl->qv_integer = *--sp;
		++sl;
		ll = ll->ll_next;
	    }

	    /* Initialize locals */
	    sl = (QValue_t *) ((BYTE_T *) sl + BC_STACK_EXTRA);
	    ll = pr->pr_locals;
	    while (ll != P_NULL) {
		sl->qv_integer = nilValue(ll->ll_type);
		++sl;
		ll = ll->ll_next;
	    }

	    /* Security stuff */
	    oldPlayer = EffectivePlayer;
	    oldStatus = EffectiveStatus;
	    if (! pr->pr_utility || EffectivePlayer == SYSADMIN_KEY ||
		(EffectiveStatus == ps_wizard &&
		 pr->pr_ownerStatus == ps_apprentice))
	    {
		/* See runUserProc in exec.c */
		EffectivePlayer = pr->pr_owner;
		EffectiveStatus = pr->pr_ownerStatus;
	    }

	    if (pr->pr_isCompiled) {
		/* It is byte-code - cycle around and call again. */
		doBcRun(key, pr, newLocals, &res, parCount / sizeof(ULONG_T));
	    } else {
		/* Initialize our new traceback entry */
		tb.tb_byteCode = B_FALSE;
		tb.tb_next = RunTraceBack;
		tb.tb_proc = pr;
		tb.tb_key = key;
		tb.tb_locals = newLocals;

		/* Call on non-byte-code interpreter */
		localsSave = RunLocals;
		RunLocals = newLocals;
		RunTraceBack = &tb;
		runExec(pr->pr_body, &res, &resType);
		RunTraceBack = tb.tb_next;
		RunLocals = localsSave;
	    }

	    /* ALWAYS restore - body could have SetEffectiveToNone */
	    EffectivePlayer = oldPlayer;
	    EffectiveStatus = oldStatus;

	    if (RunExiting || ! pr->pr_isCompiled) {
		/* Free string parameters and locals */
		ll = pr->pr_parameters;
		sl = newLocals;
		while (ll != P_NULL) {
		    if (ll->ll_type == TYPE_STRING) {
#if CHECKS_PRESENT
			if (sl->qv_string == P_NULL) {
			    abort("nil string in runUserProc #2");
			}
#endif
			freeString(sl->qv_string);
		    }
		    ll = ll->ll_next;
		    ++sl;
		}

		sl = (QValue_t *) ((BYTE_T *) sl + BC_STACK_EXTRA);
		ll = pr->pr_locals;
		while (ll != P_NULL) {
		    if (ll->ll_type == TYPE_STRING) {
#if CHECKS_PRESENT
			if (sl->qv_string == P_NULL) {
			    abort("nil string in runUserProc #3");
			}
#endif
			freeString(sl->qv_string);
		    }
		    ll = ll->ll_next;
		    ++sl;
		}
	    }
	} else {
	    hadError = B_TRUE;
	}

	if (localCount != BC_STACK_EXTRA) {
	    if (localCount > QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		memFreeE(&MS, newLocals, localCount);
	    }
	}
    }

    /* Pop parameters, and push result value */
    sp = (ULONG_T *) ((BYTE_T *) sp + parCount);
    if (resultType != TYPE_VOID) {
	if (hadError) {
	    res.qv_integer = nilValue(resultType);
	}
	*--sp = res.qv_integer;
    }
    *pLockCount -= 1;
    RunCallDepth -= 1;
    /* Return updated sp to main byte-code interpreter */
    return sp;
}

/*
 * fixCall - this is the first use of a static call (a direct call by
 *	name to a fixed routine) for some compiled code. This routine will
 *	not execute the call, but will patch the byte-code to be one of
 *	JSR or CALL, depending on whether or not the called function is
 *	compiled into byte-code. For JSR, the long-word operand (key) is
 *	replaced by the current address of the byte-code for the called
 *	routine.
 */

static void
fixCall(BYTE_T *pc)
{
    Proc_t *pr;
    ULONG_T ul;
    UINT_T *pLockCount;

    ul = FETCH_ULONG_PC();
    pr = findCachedProc(ul, &pLockCount);
    if (! pr->pr_isCompiled ||
	(! pr->pr_utility && pr->pr_owner != CurrentProc->pr_owner))
    {
	*--pc = bc_CALL - bc_error;
    } else {
	pr->pr_locked = B_TRUE;		/* can no longer do certain things */
#if UNALIGNED_OK && MUD_ENDIAN == MUD_LITTLE_ENDIAN
	if (! pr->pr_byteSwapped) {
	    bcByteSwap(pr->pr_byteCode, pr->pr_byteCodeLen);
	    pr->pr_byteSwapped = B_TRUE;
	}
#endif
	STORE_ULONG_PC((ULONG_T) pr->pr_byteCode);
	*--pc = bc_JSR - bc_error;
    }
    *pLockCount -= 1;
}

/*
 * fixBIXX - this is a call to a builtin function, which we can replace with
 *	a faster variant in which the builtin has been looked up, and its
 *	key replaced with the actual address of the native code for the
 *	builtin. Note that *no* string handling is done on any parameters.
 *	A result is possible, since the opcode is replaced by just adding
 *	one to it, hence several variants are possible.
 */

static void
fixBIXX(BYTE_T *pc)
{
    Proc_t *pr;

    pr = findBuiltIn(FETCH_ULONG_PC());
    STORE_ULONG_PC((ULONG_T) pr->pr_compiled);
    *(pc - 1) += 1;	/* add one to the opcode */
}

/*
 * bcRun - central core of interpreter.
 *	Whenever we do something that can trigger a call to runError,
 *	we should have the up-to-date values of PC and FP saved globally,
 *	so that a traceback works. Similarly for tests of 'BcAbort', since
 *	that can be set by execution time-outs.
 *
 *	A single call to bcRun can do more than one level of byte-code
 *	function calls, if all of the functions are byte-code. Also, calls
 *	out of here, to non-byte-code user functions or to builtins, can
 *	result in recursive calls back to here.
 */

static void
bcRun(QValue_t *pRes)
{
    BYTE_T *pc;
    ULONG_T *sp, *fp;
#ifdef GCC
    static void *(codeTable[]) = {
	&&label_bc_error,
	&&label_bc_ret,
	&&label_bc_ret1,
	&&label_bc_spDec,
	&&label_bc_bi,
	&&label_bc_bi01,
	&&label_bc_BI01,
	&&label_bc_bi10,
	&&label_bc_BI10,
	&&label_bc_bi31,
	&&label_bc_BI31,
	&&label_bc_call,
	&&label_bc_JSR,
	&&label_bc_CALL,
	&&label_bc_pshL,
	&&label_bc_pshLs,
	&&label_bc_pshL0,
	&&label_bc_pshL4,
	&&label_bc_pshL8,
	&&label_bc_popL,
	&&label_bc_popLs,
	&&label_bc_popL0,
	&&label_bc_popL4,
	&&label_bc_popL8,
	&&label_bc_popS,
	&&label_bc_popSs,
	&&label_bc_initL,
	&&label_bc_freeL,
	&&label_bc_pshC,
	&&label_bc_pshZ,
	&&label_bc_pshCL,
	&&label_bc_pshCH,
	&&label_bc_psh1,
	&&label_bc_pshS,
	&&label_bc_pshR,
	&&label_bc_ign,
	&&label_bc_dup,
	&&label_bc_add,
	&&label_bc_add1,
	&&label_bc_sub,
	&&label_bc_sub1,
	&&label_bc_iMul,
	&&label_bc_iDiv,
	&&label_bc_iRem,
	&&label_bc_shl,
	&&label_bc_shr,
	&&label_bc_neg,
	&&label_bc_not,
	&&label_bc_and,
	&&label_bc_ior,
	&&label_bc_xor,
	&&label_bc_fMul,
	&&label_bc_fDiv,
	&&label_bc_sFree,
	&&label_bc_sDup,
	&&label_bc_sCat,
	&&label_bc_iLT,
	&&label_bc_iLE,
	&&label_bc_iEQ,
	&&label_bc_iNE,
	&&label_bc_iGE,
	&&label_bc_iGT,
	&&label_bc_sEQ,
	&&label_bc_sEQ00,
	&&label_bc_sEQ01,
	&&label_bc_sEQ10,
	&&label_bc_sNE,
	&&label_bc_sNE00,
	&&label_bc_sNE01,
	&&label_bc_sNE10,
	&&label_bc_sLT,
	&&label_bc_sGT,
	&&label_bc_sLE,
	&&label_bc_sGE,
	&&label_bc_sEE,
	&&label_bc_sEE00,
	&&label_bc_sEE01,
	&&label_bc_sEE10,
	&&label_bc_bNot,
	&&label_bc_bf,
	&&label_bc_bft,
	&&label_bc_bff,
	&&label_bc_bb,
	&&label_bc_bbt,
	&&label_bc_bbf,
	&&label_bc_bbts,
	&&label_bc_bbfs,
	&&label_bc_getProp,
	&&label_bc_addProp,
	&&label_bc_addPrpS,
	&&label_bc_delProp,
	&&label_bc_getIEl,
	&&label_bc_getFEl,
	&&label_bc_getTEl,
	&&label_bc_getAEl,
	&&label_bc_putIEl,
	&&label_bc_putFEl,
	&&label_bc_putTEl,
	&&label_bc_putAEl,
	&&label_bc_caseI,
	&&label_bc_caseS
    };
#define CASE(x)	label_ ## x:
#define BREAK	goto *(codeTable[*pc++])
#else
    BcCode_t op;	/* make it enum to try to avoid switch range check */
#define CASE(x) case x:
#define BREAK	break
#endif

#define FETCH_AND_INC_BYTE() { \
    ui = *pc++; \
}

#if UNALIGNED_OK

#define FETCH_AND_INC_UINT() { \
    ui = *(UINT_T *) pc; \
    pc += 2; \
}

#define FETCH_AND_INC_ULONG() { \
    ul = *(ULONG_T *) pc; \
    pc += 4; \
}

#else

#define FETCH_AND_INC_UINT() { \
    ui = *pc++; \
    ui <<= 8; \
    ui |= *pc++; \
}
#define FETCH_AND_INC_ULONG() { \
    ul = *pc++; \
    ul <<= 8; \
    ul |= *pc++; \
    ul <<= 8; \
    ul |= *pc++; \
    ul <<= 8; \
    ul |= *pc++; \
}

#endif

    pc = PC;
    sp = SP;
    fp = FP;
#ifdef GCC
    BREAK;
#else
    while (B_TRUE) {
	op = *pc++ + bc_error;
	switch (op) {
#endif
	CASE(bc_error)
	    PC = pc;
	    runError("error 1 in byte-code execution");
	    return;
	CASE(bc_ret) {
	    ULONG_T ul;
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    sp = (ULONG_T *) ((BYTE_T *) sp + ui);  /* pop locals */
	    ul = *sp++;				/* pop return addr */
	    fp = (ULONG_T *) *sp++;		/* pop caller frame pointer */
	    FETCH_AND_INC_UINT();
	    sp = (ULONG_T *) ((BYTE_T *) sp + ui);  /* pop parameters */
	    pc = (BYTE_T *) ul;			/* return */
	    FP = fp;
	    if (sp == StackTop || BcAbort) {
		/* Have returned from an outer call - return from bcRun */
		PC = pc;
		return;
	    }
	    BREAK;
	}
	CASE(bc_ret1) {
	    ULONG_T ul, ul2;
	    UINT_T ui;

	    ul2 = *sp++;			/* pop return value */
	    FETCH_AND_INC_UINT();
	    sp = (ULONG_T *) ((BYTE_T *) sp + ui);  /* pop locals */
	    ul = *sp++;				/* pop return addr */
	    fp = (ULONG_T *) *sp++;		/* pop caller frame pointer */
	    FETCH_AND_INC_UINT();
	    sp = (ULONG_T *) ((BYTE_T *) sp + ui);  /* pop parameters */
	    pc = (BYTE_T *) ul;			/* return address */
	    FP = fp;
	    if (sp == StackTop) {
		/* Have returned from an outer call - return from bcRun */
		pRes->qv_integer = ul2;		/* return value */
		return;
	    }
	    *--sp = ul2;			/* push return value */
	    if (BcAbort) {
		PC = pc;
		return;
	    }
	    BREAK;
	}
	CASE(bc_spDec) {
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    sp = (ULONG_T *) ((BYTE_T *) sp - ui);  /* make space for locals */
	    fp = sp;
	    FP = fp;
	    if ((BYTE_T *) sp - 40 < (BYTE_T *) Stack) {
		PC = pc;
		runError("stack overflow in byte-code");
		return;
	    }
	    if (BcAbort) {
		PC = pc;
		return;
	    }
	    BREAK;
	}
	CASE(bc_bi) {
	    ULONG_T ul;

	    FETCH_AND_INC_ULONG();
	    PC = pc;
	    sp = builtinCall(ul, sp);
	    if (BcAbort) {
		return;
	    }
	    BREAK;
	}
	CASE(bc_bi01)		/* builtin with no args, 1 result */
	    fixBIXX(pc);	/* replace with BI01 and do that instruction */
	    /* fallthrough */
	CASE(bc_BI01) {
	    ULONG_T ul;

	    FETCH_AND_INC_ULONG();
	    PC = pc;
	    --sp;
	    SP = sp;		/* save to SP *before* calling builtin */
	    *sp = (*(ULONG_T (*)(void)) ul)();
	    if (BcAbort) {
		return;
	    }
	    BREAK;
	}
	CASE(bc_bi10)		/* builtin with 1 arg, no result */
	    fixBIXX(pc);	/* replace with BI10 and do that instruction */
	    /* fallthrough */
	CASE(bc_BI10) {
	    ULONG_T ul;

	    SP = sp;
	    FETCH_AND_INC_ULONG();
	    PC = pc;
	    (*(void (*)(ULONG_T)) ul)(*sp);
	    ++sp;
	    if (BcAbort) {
		return;
	    }
	    BREAK;
	}
	CASE(bc_bi31)		/* builtin with 3 args, 1 result */
	    fixBIXX(pc);	/* replace with BI31 & do that instruction */
	    /* fallthrough */
	CASE(bc_BI31) {
	    ULONG_T ul, ul2, ul3;

	    FETCH_AND_INC_ULONG();
	    ul2 = *sp++;
	    ul3 = *sp++;
	    PC = pc;
	    SP = sp;
	    *sp = (*(ULONG_T (*)(ULONG_T, ULONG_T, ULONG_T))ul)(*sp, ul3, ul2);
	    if (BcAbort) {
		return;
	    }
	    BREAK;
	}
	CASE(bc_call)		/* call to another user routine */
	    fixCall(pc);	/* replace with JSR or CALL & redo this */
	    pc -= 1;		/* continue with the new instruction */
	    BREAK;
	CASE(bc_JSR) {		/* optimized user-routine call */
	    ULONG_T ul;

	    FETCH_AND_INC_ULONG();
	    *--sp = (ULONG_T) fp;	/* save our frame pointer */
	    *--sp = (ULONG_T) pc;	/* save return address */
	    pc = (BYTE_T *) ul;		/* branch to subroutine */
	    fp = sp;		/* in case subroutine has no spDec */
	    FP = fp;
	    BREAK;
	}
	CASE(bc_CALL) {		/* unoptimized user-routine call */
	    ULONG_T ul;

	    FETCH_AND_INC_ULONG();
	    PC = pc;
	    sp = doStaticCall(ul, sp);
	    if (BcAbort) {
		return;
	    }
	    BREAK;
	}
	CASE(bc_pshL) {		/* generic push of local/parameter */
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    *--sp = *(ULONG_T *) ((BYTE_T *) fp + ui);
	    BREAK;
	}
	CASE(bc_pshLs) {	/* push local/parameter - 1 byte offset */
	    UINT_T ui;

	    FETCH_AND_INC_BYTE();
	    *--sp = *(ULONG_T *) ((BYTE_T *) fp + ui);
	    BREAK;
	}
	CASE(bc_pshL0)		/* push local at offset 0 */
	    *--sp = *fp;
	    BREAK;
	CASE(bc_pshL4)		/* push local at offset 4 */
	    *--sp = *(fp + 1);
	    BREAK;
	CASE(bc_pshL8)		/* push local/parameter at offset 8 */
	    *--sp = *(fp + 2);
	    BREAK;
	CASE(bc_popL) {		/* generic pop into local/parameter */
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    *(ULONG_T *) ((BYTE_T *) fp + ui) = *sp++;
	    BREAK;
	}
	CASE(bc_popLs) {	/* pop into local/parameter - 1 byte offset */
	    UINT_T ui;

	    FETCH_AND_INC_BYTE();
	    *(ULONG_T *) ((BYTE_T *) fp + ui) = *sp++;
	    BREAK;
	}
	CASE(bc_popL0)
	    *fp = *sp++;
	    BREAK;
	CASE(bc_popL4)
	    *(fp + 1) = *sp++;
	    BREAK;
	CASE(bc_popL8)
	    *(fp + 2) = *sp++;
	    BREAK;
	CASE(bc_popS) {		/* pop into string local/parameter */
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    freeString(*(char **) ((BYTE_T *) fp + ui));
	    *(ULONG_T *) ((BYTE_T *) fp + ui) = *sp++;
	    BREAK;
	}
	CASE(bc_popSs) {	/* short pop into string local/parameter */
	    UINT_T ui;

	    FETCH_AND_INC_BYTE();
	    freeString(*(char **) ((BYTE_T *) fp + ui));
	    *(ULONG_T *) ((BYTE_T *) fp + ui) = *sp++;
	    BREAK;
	}
	CASE(bc_initL) {	/* initialize string local */
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    *(ULONG_T *) ((BYTE_T *) fp + ui) = (ULONG_T) stringCopy("");
	    BREAK;
	}
	CASE(bc_freeL) {	/* free string local */
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    freeString(*(char **) ((BYTE_T *) fp + ui));
	    BREAK;
	}
	CASE(bc_pshC)		/* push 32 bit constant */
	CASE(bc_pshR) {		/* push 32 bit reference */
	    ULONG_T ul;

	    FETCH_AND_INC_ULONG();
	    *--sp = ul;
	    BREAK;
	}
	CASE(bc_pshZ)		/* push 32 bit zero */
	    *--sp = 0;
	    BREAK;
	CASE(bc_pshCL)		/* push sign extended 8 bit value */
	    /* Want sign extension of a byte */
	    *--sp = (ULONG_T) (signed char) *pc++;
	    BREAK;
	CASE(bc_pshCH)		/* push typed nil */
	    *--sp = (ULONG_T) *pc++ << 24;
	    BREAK;
	CASE(bc_psh1)		/* push 32 bit constant 1 */
	    *--sp = 1;
	    BREAK;
	CASE(bc_pshS)		/* push ptr to inline string constant */
	    *--sp = (ULONG_T) pc;
	    while (*pc++ != 0) ;
	    BREAK;
	CASE(bc_ign)		/* pop top-of-stack */
	    ++sp;
	    BREAK;
	CASE(bc_dup) {		/* duplicate top-of-stack */
	    ULONG_T ul;

	    ul = *sp;
	    *--sp = ul;
	    BREAK;
	}
	CASE(bc_add) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp += ul;
	    BREAK;
	}
	CASE(bc_add1) {
	    ++*sp;
	    BREAK;
	}
	CASE(bc_sub) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp -= ul;
	    BREAK;
	}
	CASE(bc_sub1) {
	    --*sp;
	    BREAK;
	}
	CASE(bc_iMul) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp = (ULONG_T) ((LONG_T) *sp * (LONG_T) ul);
	    BREAK;
	}
	CASE(bc_iDiv) {
	    ULONG_T ul;

	    ul = *sp++;
	    if (ul == 0) {
		PC = pc;
		runError("run-time divide by zero");
		return;
	    }
	    *sp = (ULONG_T) ((LONG_T) *sp / (LONG_T) ul);
	    BREAK;
	}
	CASE(bc_iRem) {
	    ULONG_T ul;

	    ul = *sp++;
	    if (ul == 0) {
		PC = pc;
		runError("run-time remainder by zero");
		return;
	    }
	    *sp = (ULONG_T) ((LONG_T) *sp % (LONG_T) ul);
	    BREAK;
	}
	CASE(bc_shl) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp <<= ul;
	    BREAK;
	}
	CASE(bc_shr) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp >>= ul;
	    BREAK;
	}
	CASE(bc_neg)
	    *sp = (ULONG_T) (- (LONG_T) *sp);
	    BREAK;
	CASE(bc_not)
	    *sp = ~ *sp;
	    BREAK;
	CASE(bc_and) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp &= ul;
	    BREAK;
	}
	CASE(bc_ior) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp |= ul;
	    BREAK;
	}
	CASE(bc_xor) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp ^= ul;
	    BREAK;
	}
	CASE(bc_fMul) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp = mulFixed(*sp, ul);
	    BREAK;
	}
	CASE(bc_fDiv) {
	    ULONG_T ul;

	    ul = *sp++;
	    if (ul == 0) {
		PC = pc;
		runError("run-time divide by zero");
		return;
	    }
	    *sp = divFixed(*sp, ul);
	    BREAK;
	}
	CASE(bc_sFree)		/* free top-of-stack string */
	    freeString((char *) *sp++);
	    BREAK;
	CASE(bc_sDup)		/* duplicate top-of-stack string */
	    *sp = (ULONG_T) stringCopy((char *) *sp);
	    BREAK;
	CASE(bc_sCat) {
	    ULONG_T ul;

	    ul = *sp++;
	    PC = pc;
	    *sp = (ULONG_T) concatenate((char *) *sp, (char *) ul);
	    BREAK;
	}
	CASE(bc_iLT) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp = ((LONG_T) *sp < (LONG_T) ul) ? 1 : 0;
	    BREAK;
	}
	CASE(bc_iLE) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp = ((LONG_T) *sp <= (LONG_T) ul) ? 1 : 0;
	    BREAK;
	}
	CASE(bc_iEQ) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp = (*sp == ul) ? 1 : 0;
	    BREAK;
	}
	CASE(bc_iNE) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp = (*sp != ul) ? 1 : 0;
	    BREAK;
	}
	CASE(bc_iGE) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp = ((LONG_T) *sp >= (LONG_T) ul) ? 1 : 0;
	    BREAK;
	}
	CASE(bc_iGT) {
	    ULONG_T ul;

	    ul = *sp++;
	    *sp = ((LONG_T) *sp > (LONG_T) ul) ? 1 : 0;
	    BREAK;
	}
	CASE(bc_sEQ) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) == 0) ? 1 : 0;
	    freeString((char *) ul2);
	    freeString((char *) ul);
	    BREAK;
	}
	CASE(bc_sEQ00) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) == 0) ? 1 : 0;
	    BREAK;
	}
	CASE(bc_sEQ01) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) == 0) ? 1 : 0;
	    freeString((char *) ul);
	    BREAK;
	}
	CASE(bc_sEQ10) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) == 0) ? 1 : 0;
	    freeString((char *) ul2);
	    BREAK;
	}
	CASE(bc_sNE) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) != 0) ? 1 : 0;
	    freeString((char *) ul2);
	    freeString((char *) ul);
	    BREAK;
	}
	CASE(bc_sNE00) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) != 0) ? 1 : 0;
	    BREAK;
	}
	CASE(bc_sNE01) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) != 0) ? 1 : 0;
	    freeString((char *) ul);
	    BREAK;
	}
	CASE(bc_sNE10) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) != 0) ? 1 : 0;
	    freeString((char *) ul2);
	    BREAK;
	}
	CASE(bc_sLT) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) < 0) ? 1 : 0;
	    freeString((char *) ul2);
	    freeString((char *) ul);
	    BREAK;
	}
	CASE(bc_sGT) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) > 0) ? 1 : 0;
	    freeString((char *) ul2);
	    freeString((char *) ul);
	    BREAK;
	}
	CASE(bc_sLE) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) <= 0) ? 1 : 0;
	    freeString((char *) ul2);
	    freeString((char *) ul);
	    BREAK;
	}
	CASE(bc_sGE) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmp((char *) ul2, (char *) ul) >= 0) ? 1 : 0;
	    freeString((char *) ul2);
	    freeString((char *) ul);
	    BREAK;
	}
	CASE(bc_sEE) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmpLC((char *) ul2, (char *) ul) == 0) ? 1 : 0;
	    freeString((char *) ul2);
	    freeString((char *) ul);
	    BREAK;
	}
	CASE(bc_sEE00) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmpLC((char *) ul2, (char *) ul) == 0) ? 1 : 0;
	    BREAK;
	}
	CASE(bc_sEE01) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmpLC((char *) ul2, (char *) ul) == 0) ? 1 : 0;
	    freeString((char *) ul);
	    BREAK;
	}
	CASE(bc_sEE10) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    *sp = (strCmpLC((char *) ul2, (char *) ul) == 0) ? 1 : 0;
	    freeString((char *) ul2);
	    BREAK;
	}
	CASE(bc_bNot)
	    *sp = (*sp == 0) ? 1 : 0;
	    BREAK;
	CASE(bc_bf) {		/* branch forward */
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    pc += ui;
	    BREAK;
	}
	CASE(bc_bft) {		/* branch forward if true */
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    if (*sp++ != 0) {
		pc += ui;
	    }
	    BREAK;
	}
	CASE(bc_bff) {		/* branch forward if false */
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    if (*sp++ == 0) {
		pc += ui;
	    }
	    BREAK;
	}
	CASE(bc_bb) {		/* branch backward */
	    UINT_T ui;

	    if (BcAbort) {
		PC = pc;
		return;
	    }
	    FETCH_AND_INC_UINT();
	    pc -= ui;
	    BREAK;
	}
	CASE(bc_bbt) {		/* branch backward if true */
	    UINT_T ui;

	    if (BcAbort) {
		PC = pc;
		return;
	    }
	    FETCH_AND_INC_UINT();
	    if (*sp++ != 0) {
		pc -= ui;
	    }
	    BREAK;
	}
	CASE(bc_bbf) {		/* branch backward if false */
	    UINT_T ui;

	    if (BcAbort) {
		PC = pc;
		return;
	    }
	    FETCH_AND_INC_UINT();
	    if (*sp++ == 0) {
		pc -= ui;
	    }
	    BREAK;
	}
	CASE(bc_bbts) {		/* branch backward on true, short */
	    UINT_T ui;

	    FETCH_AND_INC_BYTE();
	    if (*sp++ != 0) {
		pc -= ui;
		if (BcAbort) {
		    PC = pc;
		    return;
		}
	    }
	    BREAK;
	}
	CASE(bc_bbfs) {		/* branch backward on false, short */
	    UINT_T ui;

	    FETCH_AND_INC_BYTE();
	    if (*sp++ == 0) {
		pc -= ui;
		if (BcAbort) {
		    PC = pc;
		    return;
		}
	    }
	    BREAK;
	}
	CASE(bc_getProp) {	/* get property */
	    ULONG_T ul;
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    ul = *sp++;
	    PC = pc;
	    *sp = getProp(*sp, ul, ui);
	    BREAK;
	}
	CASE(bc_addProp) {	/* put property generic */
	    ULONG_T ul, ul2, ul3;
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    ul = *sp++;
	    ul2 = *sp++;
	    ul3 = *sp++;
	    PC = pc;
	    addProp(ul3, ul2, ul, ui);
	    if (ui == TYPE_STRING) {
		freeString((char *) ul);
	    }
	    BREAK;
	}
	CASE(bc_addPrpS) {	/* put uncopied string property */
	    ULONG_T ul, ul2, ul3;

	    ul = *sp++;
	    ul2 = *sp++;
	    ul3 = *sp++;
	    PC = pc;
	    addProp(ul3, ul2, ul, TYPE_STRING);
	    BREAK;
	}
	CASE(bc_delProp) {	/* delete property */
	    ULONG_T ul, ul2;
	    UINT_T ui;

	    FETCH_AND_INC_UINT();
	    ul = *sp++;
	    ul2 = *sp++;
	    PC = pc;
	    delProp(ul2, ul, ui);
	    BREAK;
	}
	CASE(bc_getIEl) {	/* get int list element */
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    PC = pc;
	    *sp = GetIntElement(ul2, ul);
	    BREAK;
	}
	CASE(bc_getFEl) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    PC = pc;
	    *sp = GetFixedElement(ul2, ul);
	    BREAK;
	}
	CASE(bc_getTEl) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    PC = pc;
	    *sp = GetThingElement(ul2, ul);
	    BREAK;
	}
	CASE(bc_getAEl) {
	    ULONG_T ul, ul2;

	    ul = *sp++;
	    ul2 = *sp;
	    PC = pc;
	    *sp = GetActionElement(ul2, ul);
	    BREAK;
	}
	CASE(bc_putIEl) {	/* put int list element */
	    ULONG_T ul, ul2, ul3;

	    ul = *sp++;
	    ul2 = *sp++;
	    ul3 = *sp++;
	    PC = pc;
	    PutIntElement(ul3, ul2, ul);
	    BREAK;
	}
	CASE(bc_putFEl) {
	    ULONG_T ul, ul2, ul3;

	    ul = *sp++;
	    ul2 = *sp++;
	    ul3 = *sp++;
	    PC = pc;
	    PutFixedElement(ul3, ul2, ul);
	    BREAK;
	}
	CASE(bc_putTEl) {
	    ULONG_T ul, ul2, ul3;

	    ul = *sp++;
	    ul2 = *sp++;
	    ul3 = *sp++;
	    PC = pc;
	    PutThingElement(ul3, ul2, ul);
	    BREAK;
	}
	CASE(bc_putAEl) {
	    ULONG_T ul, ul2, ul3;

	    ul = *sp++;
	    ul2 = *sp++;
	    ul3 = *sp++;
	    PC = pc;
	    PutActionElement(ul3, ul2, ul);
	    BREAK;
	}
	CASE(bc_caseI) {		/* direct indexed case */
	    ULONG_T ul, ul2, ul3;
	    UINT_T ui;

	    ul2 = *sp++;		/* ul2 is the index value */
	    FETCH_AND_INC_ULONG();
	    ul3 = ul;			/* ul3 is the minimum value */
	    FETCH_AND_INC_ULONG();	/* ul is the maximum value */
	    FETCH_AND_INC_UINT();	/* ui is the offset to the default */
	    if ((LONG_T) ul2 < (LONG_T) ul3 || (LONG_T) ul2 > (LONG_T) ul) {
		/* index out of range - go to default */
		pc += ui;
	    } else {
		/* in-range - calculate position of offset, and fetch it */
		ul2 -= ul3;
		ul2 *= 2;
		ul = (ULONG_T) pc;	/* save a copy */
		pc += ul2;		/* index into table of offsets */
		FETCH_AND_INC_UINT();	/* get ui = offset to desired case */
		pc = (BYTE_T *) ul + ui;/* branch to desired alternative */
	    }
	    BREAK;
	}
	CASE(bc_caseS) {		/* binary search case */
	    BYTE_T *pcSave;
	    ULONG_T ul;
	    LONG_T index, lo, hi;
	    UINT_T ui, defOffset;

	    if (BcAbort) {
		PC = pc;
		return;
	    }
	    index = *sp++;		/* index is the index value */
	    FETCH_AND_INC_ULONG();
	    lo = ul;			/* lo is the minimum value */
	    FETCH_AND_INC_ULONG();
	    hi = ul;			/* hi is the maximum value */
	    FETCH_AND_INC_UINT();
	    defOffset = ui;		/* defOffset is offset to default */
	    FETCH_AND_INC_UINT();	/* ui is the count of cases in table */
	    if ((LONG_T) index < (LONG_T) lo || (LONG_T) index > (LONG_T) hi) {
		/* index out of range - go to default */
		pc += defOffset;
	    } else {
		/* have to do a binary search in the table. At this point,
		   'index' is the index we are searching for and 'pc' points
		   at the table. Each table entry consists of a long-word
		   index value, followed by an offset (from the *current*
		   PC) to that alternative. */
		pcSave = pc;
		lo = 0;
		hi = ui;
		while (lo + 1 < hi) {
		    LONG_T mid;

		    mid = (lo + hi) / 2;
		    pc = pcSave + mid * (sizeof(LONG_T) + sizeof(UINT_T));
		    FETCH_AND_INC_ULONG();
		    if (index < (LONG_T) ul) {
			hi = mid;
		    } else {
			lo = mid;
		    }
		}
		pc = pcSave + lo * (sizeof(LONG_T) + sizeof(UINT_T));
		FETCH_AND_INC_ULONG();
		if (index == ul) {
		    /* Found it - get the corresponding offset */
		    FETCH_AND_INC_UINT();
		    pc = pcSave + ui;
		} else {
		    /* Not found - use the default */
		    pc = pcSave + defOffset;
		}
	    }
	    BREAK;
	    }
#ifndef GCC
	}
    }
#endif
#undef FETCH_AND_INC_ULONG
#undef FETCH_AND_INC_UINT
#undef FETCH_AND_INC_BYTE
#undef BREAK
#undef CASE
}

/*
 * doBcRun - the actual interface to bcRun.
 */

void
doBcRun(ULONG_T key, Proc_t *pr, QValue_t *pArgs,
	QValue_t *pRes, UINT_T argCount)
{
    ULONG_T *sp;
    Proc_t *saveCurrentProc;
    ULONG_T *saveStackTop;
    TraceBack_t tb;

#if UNALIGNED_OK && MUD_ENDIAN == MUD_LITTLE_ENDIAN
    if (! pr->pr_byteSwapped) {
	bcByteSwap(pr->pr_byteCode, pr->pr_byteCodeLen);
	pr->pr_byteSwapped = B_TRUE;
    }
#endif

    sp = SP;
    /* StackTop lets bcRun know when to return */
    saveStackTop = StackTop;
    StackTop = sp;
    tb.tb_byteCode = B_TRUE;
    tb.tb_next = RunTraceBack;
    tb.tb_proc = pr;
    tb.tb_key = key;
    tb.tb_PC = PC;
    tb.tb_SP = sp;
    tb.tb_FP = FP;
    RunTraceBack = &tb;
    saveCurrentProc = CurrentProc;
    CurrentProc = pr;
    while (argCount != 0) {
	*--sp = pArgs->qv_integer;
	++pArgs;
	argCount -= 1;
    }
    PC = pr->pr_byteCode;
    *--sp = (ULONG_T) FP;
    *--sp = (ULONG_T) PC;
    SP = sp;
    FP = sp;				/* in case no locals, hence no spDec */

    bcRun(pRes);
    checkAbort();			/* catch run-time exceeded */

    CurrentProc = saveCurrentProc;
    FP = tb.tb_FP;
    SP = tb.tb_SP;
    PC = tb.tb_PC;
    RunTraceBack = tb.tb_next;
    StackTop = saveStackTop;
}
