#include "MUD.h"
#include "types.h"
#include "exec.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * caches.c - stuff to handle proc, table and grammar caches.
 */

#define DEBUG		B_FALSE
#define CHECK		B_FALSE
#define CHECK_HARD	B_FALSE

#define PROC_CACHE_KEEP 	10
#define PROC_TABLE_SIZE 	1024
#define TABLE_FLUSH_POINT	50
#define REF_CACHE_SIZE		1024

typedef struct ProcCache {
    struct ProcCache *pc_next;
    struct ProcCache *pc_prev;
    Proc_t *pc_proc;
    ULONG_T pc_key;
    UINT_T pc_originalUseCount;
    UINT_T pc_lockCount;
} ProcCache_t;

typedef struct GrammarCache {
    struct GrammarCache *gc_next;
    struct GrammarCache *gc_prev;
    ULONG_T gc_key;
    Grammar_t gc_grammar;
} GrammarCache_t;

typedef struct TableCache {
    struct TableCache *tc_next;
    struct TableCache *tc_prev;
    ULONG_T tc_key;
    Table_t tc_table;
} TableCache_t;

typedef struct {
    ULONG_T rc_key;
    UINT_T rc_storedRefCount;
    UINT_T rc_currentRefCount;
} RefCache_t;

static UINT_T DummyLockCount;

static ProcCache_t *ProcCacheHead, *ProcCacheTail;
static ULONG_T ProcCacheCount;
static GrammarCache_t *GrammarCache;
static TableCache_t *TableCache;

static ProcCache_t *(ProcTable[PROC_TABLE_SIZE]);

static RefCache_t RefCache[REF_CACHE_SIZE];

#if 0

/*
 * dumpCache - debugging routine to show what is in the proc cache.
 */

void
dumpCache(CHAR_T *where)
{
    ProcCache_t *pc, *prev;
    CHAR_T *name;

    putString("At ");
    putString(where);
    putString(" procCache:\n");
    prev = P_NULL;
    pc = ProcCacheHead;
    while (pc != P_NULL) {
	putHex(pc->pc_key);
	if (pc->pc_prev != prev) {
	    putChar('*');
	}
	putChar(':');
	putHex((ULONG_T) pc->pc_proc);
	putChar(':');
	name = findAnyName(pc->pc_key);
	if (name == P_NULL) {
	    putString("???");
	} else {
	    putString(name);
	}
	newLine();
	prev = pc;
	pc = pc->pc_next;
    }
    newLine();
}

#endif

/*
 * freeProcCache - free up the entire proc cache.
 */

void
freeProcCache(BOOL_T freeMemory, BOOL_T doWrites)
{
    ProcCache_t *pc, *pcTemp;
    Proc_t *pr;
    BYTE_T *bytes;
    UINT_T len;

    TRACE("freeProcCache", freeMemory - B_FALSE);
    pc = ProcCacheHead;
    while (pc != P_NULL) {
	pcTemp = pc;
	pc = pc->pc_next;
	pr = pcTemp->pc_proc;
	if (pr->pr_useCount != pcTemp->pc_originalUseCount && doWrites) {
	    /* The 'writeProcE' call can require memory allocation, but
	       we have removed this entry from the accessible list. */
	    writeProcE(&MS, pr);
	    writeResultE(&MS, &bytes, &len);
	    ioReplace(pcTemp->pc_key, bytes, len);
	    pcTemp->pc_originalUseCount = pr->pr_useCount;
	}
	if (freeMemory) {
	    freeProcE(&MS, pr, NIL_ACTION, BOOLEAN_FALSE);
	    memFreeE(&MS, pcTemp, sizeof(ProcCache_t));
	}
    }
}

/*
 * flushProcCache - flush all of the dirty entries in the proc cache.
 *	This is called by 'Flush' and when we want to reclaim memory.
 */

void
flushProcCache(void)
{

    freeProcCache(B_FALSE, B_TRUE);
}

/*
 * addProcToCache1 - internal routine to add an entry to the proc cache.
 *	A pointer to the proc cache entry is returned.
 */

static ProcCache_t *
addProcToCache1(Proc_t *pr, ULONG_T key)
{
    ProcCache_t *pc;
    Player_t *pl;

    TRACE("addProcToCache1", key);
#if DEBUG
    logXNL("addProcToCache1 ", key);
#endif

    /* Note: if a player's status is changed, any of his procs in the
       cache have their pr_ownerStatus changed by 'updateProcsInCache'.
       Note that this ioRead of the player requires that the player
       structure not go away until no procs point at it anymore. This
       is not actually done - the Player_t and its table and thing do
       not *ever* go away! See DestroyCharacter. */
    if (pr->pr_owner == SYSADMIN_KEY) {
	pr->pr_ownerStatus = ps_wizard;
    } else {
	pl = ioRead(pr->pr_owner);
	pr->pr_ownerStatus = pl->pl_status;
    }

    pc = MEM_ALLOC_E(&MS, sizeof(ProcCache_t), "pche");
    pc->pc_next = ProcCacheHead;
    pc->pc_prev = P_NULL;
    pc->pc_proc = pr;
    pc->pc_key = key;
    pc->pc_originalUseCount = pr->pr_useCount;
    pc->pc_lockCount = 0;
    if (ProcCacheHead != P_NULL) {
	ProcCacheHead->pc_prev = pc;
    } else {
	ProcCacheTail = pc;
    }
    ProcCacheHead = pc;
    ProcCacheCount += 1;
    key %= PROC_TABLE_SIZE;
    ProcTable[key] = pc;
    return pc;
}

/*
 * addProcToCache - public entry to add a proc to the cache.
 */

void
addProcToCache(Proc_t *pr, ULONG_T key)
{

    (void) addProcToCache1(pr, key);
}

/*
 * unlinkProc - unlink an entry from the proc cache, also removing it
 *	from the hash table if (needed.
 */

static void
unlinkProc(ProcCache_t *pc, BOOL_T freeRefs)
{
    Proc_t *pr;
    ProcCache_t *prev, *next;
    ULONG_T key;

    key = pc->pc_key;
    pr = pc->pc_proc;
    /* The freeProcE call can recursively do a decRef on this proc,
       resulting in it getting into the cache again. So, do the freeProcE
       call BEFORE removing it from the cache. Also, increment its usecount
       first, so that any such recursive visit will not try to remove
       it again. */
    if (freeRefs) {
	pr->pr_useCount += 1;
    }
    freeProcE(&MS, pr, key, freeRefs ? BOOLEAN_TRUE : BOOLEAN_FALSE);
    key %= PROC_TABLE_SIZE;
    if (ProcTable[key] == pc) {
	ProcTable[key] = P_NULL;
    }
    prev = pc->pc_prev;
    next = pc->pc_next;
    if (prev != P_NULL) {
	prev->pc_next = next;
    } else {
	ProcCacheHead = next;
    }
    if (next != P_NULL) {
	next->pc_prev = prev;
    } else {
	ProcCacheTail = prev;
    }
    ProcCacheCount -= 1;
    memFreeE(&MS, pc, sizeof(ProcCache_t));
}

/*
 * reclaimMem - try to recover some memory by removing something from the
 *	proc cache. Return 'B_TRUE' if can. Note that we cannot use any
 *	of the IO code here, since that could mess up a pointer to a
 *	cached proc bytestream that we are in the middle of building
 *	a tree for. That in turn allocates memory, which can get us here.
 *
 *  May 20, 2001. In getting this stuff ready for source release, I was
 *  pretty much ready when I decided to try running with constrained
 *  memory (-m1000000). This worked on the Amiga, so should have worked
 *  for this version. However, at the end of the run I had a memory
 *  leak of a BTCD memory chunk. That tag is only used in ../Library/procIO.c
 *  when bytecode for a stripped function is being read from the database.
 *  In the standard scenario as shipped, there are no compiled functions,
 *  let alone stripped ones, so that should be impossible. Experimentation
 *  showed that it was the "RoomGetLocation" scenario function, whose body
 *  consists of just a call to builtin "Here". That routine is not compiled,
 *  and looking with gdb showed more than just the "pr_stripped" flag
 *  was messed up. Disabling all memory reclaiming makes the problem go
 *  away (which makes sense, since if you don't limit the amount of memory
 *  to use, there will be no memory reclaiming). So, instead of spending
 *  a long time (our Insure license at work expired!) trying to find this,
 *  I've verified that things work with memory recovery enabled, and just
 *  made this routine only do the setNeedRecover, rather than trying to
 *  free any procs right here. Kludge!
 */

BOOL_T
reclaimMem(BOOL_T panicing)
{
#if 1
    setNeedRecover();
#else
    ProcCache_t *pc;
    ULONG_T count, limit;

    /* find one to flush */
    limit = panicing ? 1 : PROC_CACHE_KEEP;
    count = ProcCacheCount;
    pc = ProcCacheTail;
    while (pc != P_NULL && count > limit &&
	(pc->pc_lockCount != 0 || pc->pc_proc->pr_locked ||
	 pc->pc_proc->pr_useCount != pc->pc_originalUseCount))
    {
	count -= 1;
	pc = pc->pc_prev;
    }
    if (pc != P_NULL && count > limit) {
	/* do not remove one that is close to the head. */
	TRACE("reclaimMem removing", pc->pc_key);
#if DEBUG
	logXNL("reclaimMem removing", pc->pc_key);
#endif
	/* the 'B_FALSE' is important - we do NOT want to decrement
	   refs to anything this proc references */
	unlinkProc(pc, B_FALSE);
	return B_TRUE;
    }
    if (count == ProcCacheCount) {
	setNeedRecover();
    }
#endif
    return B_FALSE;
}

/*
 * recoverMem - called from the allocator to restore memory use because
 *	memRecoverE was called. memRecoverE is only called from the main
 *	processing loop of MUDServ, so it is safe to flush things to
 *	the IO cache at this point. Return 'B_TRUE' if we were able to
 *	free something.
 */

BOOL_T
recoverMem(void)
{
    ProcCache_t *pc;
    BYTE_T *bytes;

    pc = ProcCacheTail;
    while (pc != P_NULL && (pc->pc_lockCount != 0 || pc->pc_proc->pr_locked)) {
	pc = pc->pc_prev;
    }
    if (pc != P_NULL) {
	TRACE("reclaimMem [flushing and] removing", pc->pc_key);
#if DEBUG
	logXNL("reclaimMem [flushing and] removing", pc->pc_key);
#endif
	if (pc->pc_proc->pr_useCount != pc->pc_originalUseCount) {
	    bytes = ioRead(pc->pc_key);
	    *bytes = pc->pc_proc->pr_useCount >> 8;	/* big endian */
	    *(bytes + 1) = pc->pc_proc->pr_useCount;
	    ioDirty(pc->pc_key);
	}
	/* the 'B_FALSE' is important - we do NOT want to decrement
	   refs to anything this proc references */
	unlinkProc(pc, B_FALSE);
	return B_TRUE;
    }
    return B_FALSE;
}

/*
 * findProcCacheEntry - find the entry in the proc cache for the indicated
 *	proc. If its not there - add it.
 */

static ProcCache_t *
findProcCacheEntry(ULONG_T key)
{
    ProcCache_t *pc;
    Proc_t *pr;

    TRACE("findProcCacheEntry", key);
    pc = ProcTable[key % PROC_TABLE_SIZE];
    if (pc == P_NULL || pc->pc_key != key) {
	pc = ProcCacheHead;
	while (pc != P_NULL && pc->pc_key != key) {
	    pc = pc->pc_next;
	}
	ProcTable[key % PROC_TABLE_SIZE] = pc;
    }
    if (pc != P_NULL) {
	if (pc->pc_prev != P_NULL) {
	    /* maintain the list LRU - move this one to the front */
	    pc->pc_prev->pc_next = pc->pc_next;
	    if (pc->pc_next != P_NULL) {
		pc->pc_next->pc_prev = pc->pc_prev;
	    } else {
		ProcCacheTail = pc->pc_prev;
	    }
	    pc->pc_prev = P_NULL;
	    pc->pc_next = ProcCacheHead;
	    ProcCacheHead->pc_prev = pc;
	    ProcCacheHead = pc;
	}
    } else {
	/* Note that this readProcE call can affect the proc cache, but
	   that is OK here. */
	pr = readProcE(&MS, ioRead(key));
	pc = addProcToCache1(pr, key);
if (pr->pr_isCompiled) {
    myAbort("compiled code!\n");
}
	/* Warning: body could be nil, since what is in the DB initially
	   for a proc being defined, is just the header for it. */
	if (pr->pr_isCompiled && pr->pr_body != P_NULL &&
	    pr->pr_body->ex_kind != ex_byteCode)
	{
	    pr->pr_isCompiled = B_FALSE;
	    Compile(key);
	    if (pr->pr_stripped) {
		pr->pr_stripped = B_FALSE;
		StripBody(key);
	    }
	}
    }
    return pc;
}

/*
 * findCachedProc - find the given proc in the cache. If it is not there,
 *	read it in and add it. External interface used when interpreting.
 *	Note: this interface automatically increments the lock count on
 *	the entry. It returns two values - the Proc_t and a pointer to
 *	the lock count.
 */

Proc_t *
findCachedProc(ULONG_T key, UINT_T **pLockCount)
{
    ProcCache_t *pc;

    TRACE("findCachedProc", key);
    if ((key & INDEX_MASK) >= SMALLEST_BUILTIN) {
	*pLockCount = &DummyLockCount;
	return findBuiltIn(key);
    }
    pc = findProcCacheEntry(key);
    pc->pc_lockCount += 1;
    *pLockCount = &pc->pc_lockCount;
    return pc->pc_proc;
}

/*
 * findCachedProcNoLock - variant that does not increment the lock count.
 *	This is only a valid alternative when no storage allocations or
 *	other proc cache calls can happen while the proc pointer is needed.
 */

Proc_t *
findCachedProcNoLock(ULONG_T key)
{

    TRACE("findCachedProcNoLock", key);
    if ((key & INDEX_MASK) >= SMALLEST_BUILTIN) {
	return findBuiltIn(key);
    }
    return findProcCacheEntry(key)->pc_proc;
}

/*
 * removeProc - if the given key is in the proc cache, remove it.
 *	Unconditionally make its references go away. Return its own ref count.
 */

BOOL_T
removeProc(ULONG_T key, UINT_T *pUseCount)
{
    ProcCache_t *pc;

    TRACE("removeProc", key);
#if DEBUG
    logXNL("removeProc", key);
#endif
    pc = findProcCacheEntry(key);
    if (pUseCount != P_NULL) {
	*pUseCount = pc->pc_proc->pr_useCount;
    }
    /* The only place this routine is called from is in processing a
       rt_replaceProc transaction. In that situation, the current
       incarnation of the proc is being replaced with a new one, hence
       all references done by this, the old version, should be removed.
       A new call is now from 'ZapAction', where the same results are
       wanted. */
    if (pc->pc_proc->pr_locked) {
	return B_FALSE;
    } else {
	unlinkProc(pc, B_TRUE);
	return B_TRUE;
    }
}

/*
 * updateProcsInCache - someone has just been made a full wizard - update
 *	the 'ownerStatus' field of any cached procs of that player.
 */

void
updateProcsInCache(ULONG_T ownerKey, PlayerStatus_t newStatus)
{
    ProcCache_t *pc;
    Proc_t *pr;

    TRACE("updateProcsInCache", ownerKey);
    pc = ProcCacheHead;
    while (pc != P_NULL) {
	pr = pc->pc_proc;
	if (pr->pr_owner == ownerKey) {
	    pr->pr_ownerStatus = newStatus;
	}
	pc = pc->pc_next;
    }
}

/*
 * incProcRef - someone has created a new ref to a proc - increment that
 *	proc's ref count. The proc is found via the proc cache.
 */

void
incProcRef(ULONG_T key)
{
    ProcCache_t *pc;
    Proc_t *pr;

    TRACE("incProcRef", key);
#if DEBUG
    logXNL("incProcRef", key);
#endif
    if (key != NIL_ACTION && (key & INDEX_MASK) < SMALLEST_BUILTIN) {
	pc = findProcCacheEntry(key);
	pr = pc->pc_proc;
	if (pr->pr_useCount != 0xffff) {
	    pr->pr_useCount += 1;
	}
    }
}

/*
 * decProcRef - someone has destroyed a ref to a proc - decrement that
 *	proc's ref count. The proc is found via the proc cache.
 */

void
decProcRef(ULONG_T key)
{
    ProcCache_t *pc;
    Proc_t *pr;

    TRACE("decProcRef", key);
#if DEBUG
    logXNL("decProcRef", key);
#endif
    if (key != NIL_ACTION && (key & INDEX_MASK) < SMALLEST_BUILTIN) {
	pc = findProcCacheEntry(key);
	pr = pc->pc_proc;
	if (pr->pr_useCount != 0xffff) {
#if CHECK
	    if (pr->pr_useCount == 0) {
		myAbort("proc useCount is 0 in decProcRef");
	    }
#endif
	    pr->pr_useCount -= 1;
	    if (pr->pr_useCount == 0) {
		/* No-one references the proc - nuke it fully. */
		unlinkProc(pc, B_TRUE);
		ioDelete(key);
		deleteFromClients(key);
	    }
	}
    }
}

/*
 * decProcRefAndNil - if one of the 'SetXXXAction' routines is freeing up
 *	the last use of an action, then that action is no longer valid,
 *	and its value should not be returned. We want to return NIL_ACTION
 *	instead. This routine arranges for that.
 */

void
decProcRefAndNil(ULONG_T *pKey)
{
    ULONG_T key;

    key = *pKey;
    if (key != NIL_ACTION) {
	if (findCachedProcNoLock(key)->pr_useCount == 1) {
	    *pKey = NIL_ACTION;
	}
	decProcRef(key);
    }
}

/*
 * findProcByPC - using a BYTE_T * PC value, try to find a proc such that
 *	it is in the cache, has byte-code, and that PC value is within
 *	the byte-code of the proc.
 */

Proc_t *
findProcByPC(BYTE_T *pcVal, ULONG_T *pKey)
{
    ProcCache_t *pc;
    Proc_t *pr;

    pc = ProcCacheHead;
    while (pc != P_NULL) {
	pr = pc->pc_proc;
	if (pr->pr_isCompiled && pcVal >= pr->pr_byteCode &&
	    pcVal < pr->pr_byteCode + pr->pr_byteCodeLen)
	{
	    *pKey = pc->pc_key;
	    return pr;
	}
	pc = pc->pc_next;
    }
    return P_NULL;
}

/*
 * unlinkGrammar - unlink a grammar cache entry.
 */

static void
unlinkGrammar(GrammarCache_t *gc)
{
    GrammarCache_t *prev, *next;

    prev = gc->gc_prev;
    next = gc->gc_next;
    if (prev != P_NULL) {
	prev->gc_next = next;
    } else {
	GrammarCache = next;
    }
    if (next != P_NULL) {
	next->gc_prev = prev;
    }
    memFreeE(&MS, gc, sizeof(GrammarCache_t));
}

/*
 * nukeGrammar - make a grammar's contents and representation go away,
 *	but do not delete the grammar's database entry.
 */

static void
nukeGrammar(GrammarCache_t *gc)
{

    /* Temporarily increase the usecount, so that we cannot recurse into
       it and try to remove it while we are already doing so. (E.g.
       because of a top level ZapGrammar call.) */
    gc->gc_grammar.g_words.tb_useCount += 1;
    releaseGrammar(&gc->gc_grammar);
    freeGrammar(&gc->gc_grammar);
    unlinkGrammar(gc);
}

/*
 * findGrammarCacheEntry - find a grammar in the grammar cache.
 */

static GrammarCache_t *
findGrammarCacheEntry(ULONG_T key)
{
    GrammarCache_t *gc, *gct;

    TRACE("findGrammarCacheEntry", key);
    gc = GrammarCache;
    while (gc != P_NULL && gc->gc_key != key) {
	gc = gc->gc_next;
    }
    if (gc == P_NULL) {
	gc = MEM_ALLOC_E(&MS, sizeof(GrammarCache_t), "gche");
	gct = GrammarCache;
	gc->gc_next = gct;
	gc->gc_prev = P_NULL;
	gc->gc_key = key;
	if (gct != P_NULL) {
	    gct->gc_prev = gc;
	}
	GrammarCache = gc;
	readGrammar(&gc->gc_grammar, key);
	/* Grammars are big, so use ioWrite to make sure that the entry
	   for it in the database cache is not dirty, so that it can be
	   removed if needed. */
	ioWrite(key);
    }
    return gc;
}

/*
 * findCachedGrammar - find a grammar in the cache. Read it in if needed.
 */

Grammar_t *
findCachedGrammar(ULONG_T key)
{

    if (key == NIL_GRAMMAR) {
	runError("nil grammar encountered at run time");
	return &GrammarCache->gc_grammar;
    }
    return &findGrammarCacheEntry(key)->gc_grammar;
}

/*
 * ZapGrammar - the builtin.
 */

void
ZapGrammar(ULONG_T key)
{
    GrammarCache_t *gc;
    Grammar_t g;

    if (key == NIL_GRAMMAR) {
	runError("nil grammar to ZapGrammar");
    } else {
	gc = findGrammarCacheEntry(key);
	if (gc->gc_grammar.g_words.tb_owner != EffectivePlayer &&
	    EffectivePlayer != SYSADMIN_KEY)
	{
	    runError("you don't own that grammar");
	} else {
	    /* A subset of CreateGrammar, to replace it by an empty one.
	       This is not needed/wanted when called from ZapTable, but
	       what the heck. */
	    symCreateE(&MS, &g.g_words, 1);
	    g.g_words.tb_owner = gc->gc_grammar.g_words.tb_owner;
	    g.g_words.tb_useCount = gc->gc_grammar.g_words.tb_useCount;
	    g.g_wordList = P_NULL;
	    g.g_nextCode = 1;
	    nukeGrammar(gc);
	    writeGrammar(&g, key);
	    symDestroyE(&MS, &g.g_words);
	}
    }
}

/*
 * unlinkTable - unlink a table cache entry.
 */

static void
unlinkTable(TableCache_t *tc)
{
    TableCache_t *prev, *next;

    prev = tc->tc_prev;
    next = tc->tc_next;
    if (prev != P_NULL) {
	prev->tc_next = next;
    } else {
	TableCache = next;
    }
    if (next != P_NULL) {
	next->tc_prev = prev;
    }
    memFreeE(&MS, tc, sizeof(TableCache_t));
}

/*
 * releaseTable - a table is being destroyed - decrement references.
 */

static void
releaseTable(Table_t *tb)
{
    Entry_t *en;
    UINT_T count;

    count = tb->tb_entryCount;
    en = tb->tb_entries;
    while (count != 0) {
	count -= 1;
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
	    decRef(en->en_key);
	}
	++en;
    }
}

/*
 * nukeTable - do everything needed to make a table go away. Note that
 *	this is different from nukeGrammar, in that it *does* do the
 *	ioDelete.
 */

static void
nukeTable(TableCache_t *tc)
{

    releaseTable(&tc->tc_table);
    ioDelete(tc->tc_key);
    symDestroyE(&MS, &tc->tc_table);
    unlinkTable(tc);
}

/*
 * findTableCacheEntry - find a table in the table cache.
 */

static TableCache_t *
findTableCacheEntry(ULONG_T key)
{
    TableCache_t *tc, *tct;

    TRACE("findTableCacheEntry", key);
    tc = TableCache;
    while (tc != P_NULL && tc->tc_key != key) {
	tc = tc->tc_next;
    }
    if (tc == P_NULL) {
	tc = MEM_ALLOC_E(&MS, sizeof(TableCache_t), "tche");
	tct = TableCache;
	tc->tc_next = tct;
	tc->tc_prev = P_NULL;
	tc->tc_key = key;
	if (tct != P_NULL) {
	    tct->tc_prev = tc;
	}
	TableCache = tc;
	(void) symReadE(&MS, &tc->tc_table, ioRead(key));
	ioWrite(key);
    }
    return tc;
}

/*
 * findCachedTable - find a table in the cache. Read it in if needed.
 */

Table_t *
findCachedTable(ULONG_T key)
{

    if (key == NIL_TABLE) {
	runError("nil table encountered at run time");
	return &TableCache->tc_table;
    }
    return &findTableCacheEntry(key)->tc_table;
}

/*
 * flushCachedTable - write out the indicated table, and remove it from
 *	our cache of tables.
 */

void
flushCachedTable(ULONG_T key)
{
    TableCache_t *tc;
    BYTE_T *bytes;
    UINT_T len;

    TRACE("flushCachedTable", key);
    /* Have to be a bit careful - some wizard could even have some
       character's private table in use! */
    if (notInUse(key)) {
	tc = TableCache;
	while (tc != P_NULL && tc->tc_key != key) {
	    tc = tc->tc_next;
	}
	if (tc != P_NULL) {
	    if (tc->tc_table.tb_useCount == 0) {
		releaseTable(&tc->tc_table);
		ioDelete(key);
	    } else if (tc->tc_table.tb_dirty) {
		symWriteE(&MS, &tc->tc_table);
		writeResultE(&MS, &bytes, &len);
		ioReplace(tc->tc_key, bytes, len);
		if (tc->tc_table.tb_usedCount > TABLE_FLUSH_POINT) {
		    ioWrite(tc->tc_key);
		}
	    }
	    symDestroyE(&MS, &tc->tc_table);
	    unlinkTable(tc);
	}
    }
}

/*
 * NOTE: we deliberately only put things and properties into the ref cache!
 */

#if CHECK_HARD

/*
 * checkRefCache - make sure all is OK in the ref cache.
 */

static void
checkRefCache(CHAR_T *where)
{
    RefCache_t *rc;
    ULONG_T key;
    UINT_T i, typ;

    rc = &RefCache[0];
    for (i = 0; i != REF_CACHE_SIZE; i += 1) {
	key = rc->rc_key;
	if (key != 0) {
	    typ = key >> TYPE_SHIFT;
	    if (typ != TYPE_THING && typ < TYPE_PROP_BOOLEAN ||
		typ > TYPE_PROP_LIST_ACTION)
	    {
		setLogging();
		putString("bad key ");
		putHex(key);
		putString(" in ref cache\n");
		myAbort("checkRefCache");
	    }
	    if (rc->rc_currentRefCount == 0) {
		setLogging();
		putString("current ref count is 0 for key ");
		putHex(key);
		putString(" in ref cache\n");
		myAbort("checkRefCache");
	    }
	}
	++rc;
    }
}

#endif

/*
 * updateRefCount - make the real database entity have the correct ref count.
 */

static void
updateRefCount(RefCache_t *rc)
{
    ULONG_T key;
    Thing_t *th;
    Property_t *prp;

    key = rc->rc_key;
    TRACE("updateRefCount", key);
#if CHECK
    if (rc < &RefCache[0] || rc >= &RefCache[REF_CACHE_SIZE]) {
	myAbort("rc is out of range in updateRefCount");
    }
    if (rc->rc_currentRefCount == 0) {
	myAbort("currentRefCount is 0 in updateRefCount");
    }
    /* Cannot check storedRefCount - it can be 0 if we are doing the
       first incRef on a newly created entity. */
#endif
#if CHECK_HARD
    checkRefCache("updateRefCount");
#endif
    if ((key >> TYPE_SHIFT) == TYPE_THING) {
	th = ioRead(key);
	th->th_useCount = rc->rc_currentRefCount;
	rc->rc_storedRefCount = rc->rc_currentRefCount;
    } else {
	prp = ioRead(key);
	prp->prp_useCount = rc->rc_currentRefCount;
	rc->rc_storedRefCount = rc->rc_currentRefCount;
    }
    ioDirty(key);
    /* Used to set the refcache key to 0 here, thus clearing the entry. I
       don't see any reason to do that. */
}

/*
 * incRef - general reference incrementer.
 */

void
incRef(ULONG_T key)
{
    Thing_t *th;
    Integer_t *intValue;
    GrammarCache_t *gc;
    Property_t *prp;
    TableCache_t *tc;
    RefCache_t *rc;
    UINT_T count;

    TRACE("incRef", key);
#if DEBUG
    logXNL("incRef", key);
#endif
#if CHECK_HARD
    checkRefCache("incRef");
#endif
    if ((key & INDEX_MASK) != 0) {
	switch (key >> TYPE_SHIFT) {
	case TYPE_INTEGER:
	case TYPE_FIXED:
	    intValue = ioRead(key);
	    count = intValue->int_useCount;
	    if (count != 0xffff) {
		count += 1;
		intValue->int_useCount = count;
		ioDirty(key);
	    }
	    break;
	case TYPE_THING:
	    rc = &RefCache[key % REF_CACHE_SIZE];
	    if (rc->rc_key == key) {
		count = rc->rc_currentRefCount;
		if (count != 0xffff) {
#if CHECK
		    if (count == 0) {
			myAbort("thing useCount is 0 in incRef");
		    }
#endif
		    count += 1;
		    rc->rc_currentRefCount = count;
		    /* Optimization here, for props, and in decRef: do not
		       remove it from the ref cache even if its ref counts
		       become the same again - that saves having to put
		       it right back in the ref cache later! */
		}
	    } else {
		th = ioRead(key);
		count = th->th_useCount;
		if (rc->rc_key != 0x00000000 &&
		    rc->rc_currentRefCount != rc->rc_storedRefCount)
		{
		    updateRefCount(rc);
		}
		rc->rc_key = key;
		rc->rc_storedRefCount = count;
		if (count != 0xffff) {
		    count += 1;
		}
		rc->rc_currentRefCount = count;
	    }
	    break;
	case TYPE_ACTION:
	    if ((key & INDEX_MASK) < SMALLEST_BUILTIN) {
		incProcRef(key);
	    }
	    break;
	case TYPE_TABLE:
	    tc = findTableCacheEntry(key);
	    count = tc->tc_table.tb_useCount;
	    if (count != 0xffff) {
		count += 1;
		tc->tc_table.tb_useCount = count;
		tc->tc_table.tb_dirty = B_TRUE;
	    }
	    break;
	case TYPE_GRAMMAR:
	    gc = findGrammarCacheEntry(key);
	    count = gc->gc_grammar.g_words.tb_useCount;
	    if (count != 0xffff) {
		count += 1;
		gc->gc_grammar.g_words.tb_useCount = count;
		gc->gc_grammar.g_words.tb_dirty = B_TRUE;
	    }
	    break;
	case TYPE_PROP_BOOLEAN:
	case TYPE_PROP_INTEGER:
	case TYPE_PROP_FIXED:
	case TYPE_PROP_THING:
	case TYPE_PROP_ACTION:
	case TYPE_PROP_STRING:
	case TYPE_PROP_TABLE:
	case TYPE_PROP_GRAMMAR:
	case TYPE_PROP_LIST_INTEGER:
	case TYPE_PROP_LIST_FIXED:
	case TYPE_PROP_LIST_THING:
	case TYPE_PROP_LIST_ACTION:
	    rc = &RefCache[key % REF_CACHE_SIZE];
	    if (rc->rc_key == key) {
		count = rc->rc_currentRefCount;
		if (count != 0xffff) {
#if CHECK
		    if (count == 0) {
			myAbort("prop useCount is 0 in incRef");
		    }
#endif
		    count += 1;
		    rc->rc_currentRefCount = count;
		}
	    } else {
		prp = ioRead(key);
		count = prp->prp_useCount;
		if (rc->rc_key != 0x00000000 &&
		    rc->rc_currentRefCount != rc->rc_storedRefCount)
		{
		    updateRefCount(rc);
		}
		rc->rc_key = key;
		rc->rc_storedRefCount = count;
		if (count != 0xffff) {
		    count += 1;
		}
		rc->rc_currentRefCount = count;
	    }
	    break;
	}
    }
}

/*
 * decRef - general reference decrementer.
 */

void
decRef(ULONG_T key)
{
    Thing_t *th;
    Integer_t *intValue;
    GrammarCache_t *gc;
    TableCache_t *tc;
    RefCache_t *rc;
    Property_t *prp;
    UINT_T count;

    TRACE("decRef", key);
#if DEBUG
    logXNL("decRef", key);
#endif
#if CHECK_HARD
    checkRefCache("decRef");
#endif
    if ((key & INDEX_MASK) != 0) {
	switch (key >> TYPE_SHIFT) {
	case TYPE_INTEGER:
	case TYPE_FIXED:
	    intValue = ioRead(key);
	    count = intValue->int_useCount;
	    if (count != 0xffff) {
#if CHECK
		if (count == 0) {
		    myAbort("INT/FIXED useCount is 0 in decRef");
		}
#endif
		count -= 1;
		intValue->int_useCount = count;
		if (count == 0) {
		    ioDelete(key);
		}
	    }
	    break;
	case TYPE_THING:
	    rc = &RefCache[key % REF_CACHE_SIZE];
	    if (rc->rc_key == key) {
		count = rc->rc_currentRefCount;
		if (count != 0xffff) {
#if CHECK
		    if (count == 0) {
			myAbort("thing useCount is 0 in decRef #1");
		    }
#endif
		    count -= 1;
		    rc->rc_currentRefCount = count;
		    if (count == 0) {
			th = ioRead(key);
			/* Do this *before* the recursive decRef! */
			rc->rc_key = 0x00000000;
			if (th->th_parent != NIL_THING) {
			    decRef(th->th_parent);
			}
			nukeThing(key);
			ioDelete(key);
			ThingsDestroyed += 1;
		    }
		}
	    } else {
		th = ioRead(key);
		count = th->th_useCount;
		if (count == 1) {
		    if (th->th_parent != NIL_THING) {
			decRef(th->th_parent);
		    }
		    nukeThing(key);
		    ioDelete(key);
		    ThingsDestroyed += 1;
		} else {
#if CHECK
		    if (count == 0) {
			myAbort("thing useCount is 0 in decRef #2");
		    }
#endif
		    if (rc->rc_key != 0x00000000 &&
			rc->rc_currentRefCount != rc->rc_storedRefCount)
		    {
			updateRefCount(rc);
		    }
		    rc->rc_key = key;
		    rc->rc_storedRefCount = count;
		    if (count != 0xffff) {
			count -= 1;
		    }
		    rc->rc_currentRefCount = count;
		}
	    }
	    break;
	case TYPE_ACTION:
	    if ((key & INDEX_MASK) < SMALLEST_BUILTIN) {
		decProcRef(key);
	    }
	    break;
	case TYPE_TABLE:
	    tc = findTableCacheEntry(key);
	    count = tc->tc_table.tb_useCount;
	    if (count != 0xffff) {
#if CHECK
		if (count == 0) {
		    myAbort("table useCount is 0 in decRef");
		}
#endif
		count -= 1;
		if (count == 0 && notInUse(key)) {
		    nukeTable(tc);
		} else {
		    tc->tc_table.tb_useCount = count;
		    tc->tc_table.tb_dirty = B_TRUE;
		}
	    }
	    break;
	case TYPE_GRAMMAR:
	    gc = findGrammarCacheEntry(key);
	    count = gc->gc_grammar.g_words.tb_useCount;
	    if (count != 0xffff) {
#if CHECK
		if (count == 0) {
		    myAbort("grammar useCount is 0 in decRef");
		}
#endif
		count -= 1;
		if (count == 0) {
		    nukeGrammar(gc);
		    ioDelete(key);
		} else {
		    gc->gc_grammar.g_words.tb_useCount = count;
		    gc->gc_grammar.g_words.tb_dirty = B_TRUE;
		}
	    }
	    break;
	case TYPE_PROP_BOOLEAN:
	case TYPE_PROP_INTEGER:
	case TYPE_PROP_FIXED:
	case TYPE_PROP_THING:
	case TYPE_PROP_ACTION:
	case TYPE_PROP_STRING:
	case TYPE_PROP_TABLE:
	case TYPE_PROP_GRAMMAR:
	case TYPE_PROP_LIST_INTEGER:
	case TYPE_PROP_LIST_FIXED:
	case TYPE_PROP_LIST_THING:
	case TYPE_PROP_LIST_ACTION:
	    rc = &RefCache[key % REF_CACHE_SIZE];
	    if (rc->rc_key == key) {
		count = rc->rc_currentRefCount;
		if (count != 0xffff) {
#if CHECK
		    if (count == 0) {
			myAbort("prop useCount is 0 in decRef #1");
		    }
#endif
		    count -= 1;
		    rc->rc_currentRefCount = count;
		    if (count == 0) {
			ioDelete(key);
			rc->rc_key = 0x00000000;
		    }
		}
	    } else {
		prp = ioRead(key);
		count = prp->prp_useCount;
		if (count == 1) {
		    ioDelete(key);
		} else {
#if CHECK
		    if (count == 0) {
			myAbort("prop useCount is 0 in decRef #2");
		    }
#endif
		    if (rc->rc_key != 0x00000000 &&
			rc->rc_currentRefCount != rc->rc_storedRefCount)
		    {
			updateRefCount(rc);
		    }
		    rc->rc_key = key;
		    rc->rc_storedRefCount = count;
		    if (count != 0xffff) {
			count -= 1;
		    }
		    rc->rc_currentRefCount = count;
		}
	    }
	    break;
	}
    }
}

/*
 * cacheInit - initialization of stuff here.
 */

void
cacheInit(void)
{
    int i;

    ProcCacheHead = P_NULL;
    ProcCacheTail = P_NULL;
    ProcCacheCount = 0;
    for (i = 0; i != PROC_TABLE_SIZE; i += 1) {
	ProcTable[i] = P_NULL;
    }
    /* we can now do this safely, since the proc cache is set up */
    MS.ms_ReclaimMem = reclaimMem;
    MS.ms_RecoverMem = recoverMem;

    GrammarCache = P_NULL;
    TableCache = P_NULL;
    for (i = 0; i != REF_CACHE_SIZE; i += 1) {
	RefCache[i].rc_key = 0x00000000;
    }
}

/*
 * freeGrammarCache - flush out the grammar cache.
 */

static void
freeGrammarCache(BOOL_T freeMemory, BOOL_T doWrites)
{
    GrammarCache_t *gc, *gct;
    ULONG_T key;

    TRACE("freeGrammarCache",
	(ULONG_T) ((freeMemory - B_FALSE) << 8) + (doWrites - B_FALSE));
    gc = GrammarCache;
    while (gc != P_NULL) {
	gct = gc;
	gc = gc->gc_next;
	if (gct->gc_grammar.g_words.tb_useCount == 0) {
	    key = gct->gc_key;
	    nukeGrammar(gct);
	    ioDelete(key);
	} else {
	    if (doWrites && gct->gc_grammar.g_words.tb_dirty) {
		writeGrammar(&gct->gc_grammar, gct->gc_key);
		if (! freeMemory) {
		    /* we want the grammar valid again! */
		    freeGrammar(&gct->gc_grammar);
		    readGrammar(&gct->gc_grammar, gct->gc_key);
		}
		if (gct->gc_grammar.g_words.tb_usedCount > TABLE_FLUSH_POINT) {
		    ioWrite(gct->gc_key);
		}
	    }
	    if (freeMemory) {
		freeGrammar(&gct->gc_grammar);
		memFreeE(&MS, gct, sizeof(GrammarCache_t));
	    }
	}
    }
}

/*
 * freeTableCache - flush out the table cache.
 */

static void
freeTableCache(BOOL_T freeMemory, BOOL_T doWrites)
{
    TableCache_t *tc, *tct;
    BYTE_T *bytes;
    UINT_T len;

    TRACE("freeTableCache",
	((ULONG_T) (freeMemory - B_FALSE) << 8) + (doWrites - B_FALSE));
    tc = TableCache;
    while (tc != P_NULL) {
	tct = tc;
	tc = tc->tc_next;
	if (tct->tc_table.tb_useCount == 0 && notInUse(tct->tc_key)) {
	    releaseTable(&tct->tc_table);
	    symDestroyE(&MS, &tct->tc_table);
	    ioDelete(tct->tc_key);
	    unlinkTable(tct);
	} else {
	    if (doWrites && tct->tc_table.tb_dirty) {
		symWriteE(&MS, &tct->tc_table);
		writeResultE(&MS, &bytes, &len);
		ioReplace(tct->tc_key, bytes, len);
		if (tct->tc_table.tb_usedCount > TABLE_FLUSH_POINT) {
		    ioWrite(tct->tc_key);
		}
	    }
	    if (freeMemory) {
		symDestroyE(&MS, &tct->tc_table);
		memFreeE(&MS, tct, sizeof(TableCache_t));
	    }
	}
    }
}

/*
 * flushRefCache - flush out the ref cache.
 */

void
flushRefCache(void)
{
    int i;
    RefCache_t *rc;

    TRACE("flushRefCache", 0);
    rc = &RefCache[0];
    for (i = 0; i < REF_CACHE_SIZE; i += 1) {
	if (rc->rc_key != 0x00000000 &&
	    rc->rc_currentRefCount != rc->rc_storedRefCount)
	{
	    updateRefCount(rc);
	}
	++rc;
    }
}

/*
 * flushRefKey - allow someone to force the ref count on the disk to be
 *	up to date.
 */

void
flushRefKey(ULONG_T key)
{
    RefCache_t *rc;

    TRACE("flushRefKey", key);
    rc = &RefCache[key % REF_CACHE_SIZE];
    if (rc->rc_key == key && rc->rc_currentRefCount != rc->rc_storedRefCount) {
	updateRefCount(rc);
    }
}

/*
 * flushCaches - single call to do all of the flushing here.
 */

void
flushCaches(void)
{

    freeGrammarCache(B_FALSE, B_TRUE);
    freeTableCache(B_FALSE, B_TRUE);
    freeProcCache(B_FALSE, B_TRUE);
    flushRefCache();
}

/*
 * cacheTerm - termination of stuff here.
 */

void
cacheTerm(BOOL_T freeMemory, BOOL_T doWrites)
{
    ProcCache_t *pc;

    pc = ProcCacheHead;
    while (pc != P_NULL) {
	if (pc->pc_lockCount != 0) {
	    logXNL("locked in cache ", pc->pc_key);
	} else if (pc->pc_proc->pr_locked) {
	    logXNL("bytecode locked in cache ", pc->pc_key);
	}
	pc = pc->pc_next;
    }

    freeGrammarCache(freeMemory, doWrites);
    freeTableCache(freeMemory, doWrites);
    freeProcCache(freeMemory, doWrites);
    if (doWrites) {
	flushRefCache();
    }
    /* just in case someone tries to use this */
    MS.ms_ReclaimMem = P_NULL;
}

/*
 * clearRefs - clear refs inside cached procs to the indicated key - this
 *	is because it is being deleted from a symbol table and the name
 *	pointer may no longer be valid.
 *
 *	January 26, 1993: I do not see why it is necessary to ever do this.
 *	rl_name fields are kept ONLY during the print of a single function
 *	body, and are then immediately nil-ed.
 *
 *	July 3, 1994: It is possible to delete a proc that is still
 *	referenced by other procs. The proc itself stays around, but the
 *	symbol table entry for it is deleted. That may invalidate the
 *	reflist pointer, so we DO have to do this.
 */

void
clearRefs(ULONG_T key)
{
    ProcCache_t *pc;
    RefList_t *rl;

    TRACE("clearRefs", key);
    pc = ProcCacheHead;
    while (pc != P_NULL) {
	rl = pc->pc_proc->pr_refs;
	while (rl != P_NULL) {
	    if (rl->rl_key == key) {
		rl->rl_name = P_NULL;
	    }
	    rl = rl->rl_next;
	}
	pc = pc->pc_next;
    }
}
