#include "MUD.h"
#include "types.h"
#include "exec.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * exec.c - the interpreter.
 */

/* note for future consideration: instead of allocating 'stackframes' for
   parameters and locals during execution, can just allocate a big 'stack
   page' on startup and have a 'stack pointer' into it move around during
   execution. This would be faster, but would require a tradeoff between
   using space for the 'stack' and limiting the call depth/size. */

void runExec(Exec_t *ex, QValue_t *pRes, UINT_T *pType);

/*
 * runAbort - asynchronous request to abort execution.
 */

void
runAbort(void)
{

    RunAborting = B_TRUE;
    BcAbort = B_TRUE;
}

/*
 * checkAbort - at strategic times, call this to see if we should exit.
 */

void
checkAbort(void)
{

    if (RunAborting) {
	RunAborting = B_FALSE;
	if (! RunExiting) {
	    runError("execution timeout or abort");
	    RunExiting = B_TRUE;
	}
    }
}

/*
 * traceProcName - utility for traceBack.
 */

void
traceName(ULONG_T key, char *kind)
{
    CHAR_T *name;

    if ((key & INDEX_MASK) == 0) {
	fmtValue(key >> TYPE_SHIFT, key);
    } else {
	name = findAnyName(key);
	if (name == P_NULL) {
	    fmtCheck("<");
	    fmtCheck(kind);
	    fmtCheck(":");
	    fmtHex(key);
	    fmtCheck(">");
	} else {
	    fmtCheck(name);
	}
    }
}

/*
 * tbShowValue - show a parameter value during traceback.
 */

void
tbShowValue(QValue_t *p, LocalList_t *ll)
{
    ULONG_T key;

    key = p->qv_integer;
    switch (ll->ll_type) {
    case TYPE_PLAYER:
	traceName(key, "character");
	break;
    case TYPE_THING:
	traceName(key, "thing");
	break;
    case TYPE_ACTION:
	/* avoid the long pretty-print here */
	traceName(key, "action");
	break;
    case TYPE_TABLE:
	traceName(key, "table");
	break;
    case TYPE_GRAMMAR:
	traceName(key, "grammar");
	break;
    case TYPE_PROP_BOOLEAN:
    case TYPE_PROP_INTEGER:
    case TYPE_PROP_FIXED:
    case TYPE_PROP_THING:
    case TYPE_PROP_ACTION:
    case TYPE_PROP_STRING:
    case TYPE_PROP_TABLE:
    case TYPE_PROP_GRAMMAR:
    case TYPE_PROP_LIST_INTEGER:
    case TYPE_PROP_LIST_FIXED:
    case TYPE_PROP_LIST_THING:
    case TYPE_PROP_LIST_ACTION:
	traceName(key, "property");
	break;
    case TYPE_ELEMENT:
	--p;
	switch ((ULONG_T) p->qv_integer >> TYPE_SHIFT) {
	case TYPE_LIST_INTEGER:
	    fmtValue(TYPE_INTEGER, key);
	    break;
	case TYPE_LIST_FIXED:
	    fmtValue(TYPE_FIXED, key);
	    break;
	case TYPE_LIST_THING:
	    traceName(key, "thing");
	    break;
	case TYPE_LIST_ACTION:
	    traceName(key, "action");
	    break;
	}
	break;
    default:
	fmtValue(ll->ll_type, key);
	break;
    }
}

/*
 * execTraceBack - one level of traceback for Exec_t execution.
 */

static void
execTraceBack(TraceBack_t *tb)
{
    QValue_t *p;
    LocalList_t *ll;

    traceName(tb->tb_key, "ACTION");
    fmtChar('(');
    p = tb->tb_locals;
    if (p != P_NULL) {
	ll = tb->tb_proc->pr_parameters;
	while (ll != P_NULL) {
	    tbShowValue(p, ll);
	    ++p;
	    ll = ll->ll_next;
	    if (ll != P_NULL) {
		fmtCheck(",");
		fmtCheck(" ");
	    }
	}
    }
    fmtChar(')');
    fmtLine();
}

/*
 * traceBack - produce a call traceback from the point of error.
 */

void
traceBack(void)
{
    TraceBack_t *tb;

#if DEBUG_PRESENT
    if (WantDebug) {
	mess("traceBack\n");
    }
#endif
    tb = RunTraceBack;
    if (tb == P_NULL) {
	putString("Error occurred at top level.\n");
    } else {
	putString("Traceback:\n");
	fmtStart();
	while (tb != P_NULL) {
	    if (tb->tb_byteCode) {
		bcTraceBack(tb);
	    } else {
		execTraceBack(tb);
	    }
	    tb = tb->tb_next;
	}
    }
}

/*
 * runError - report an error at run-time and stop interpretation.
 */

void
runError(CHAR_T *message)
{

    setToError(B_TRUE);
    putError(message);
    RunExiting = B_TRUE;
    BcAbort = B_TRUE;
    traceBack();
    setToError(B_FALSE);
}

/*
 * runTypeError - a run-time type mismatch of some kind.
 */

static void
runTypeError(CHAR_T *name, CHAR_T *message, UINT_T got, UINT_T wanted)
{
    CHAR_T gotName[30], wantedName[30];

    getTypeName(&gotName, got);
    getTypeName(&wantedName, wanted);
    setToError(B_TRUE);
    putString("*** ");
    if (name != P_NULL) {
	putString(name);
	putString(": ");
    }
    putString(message);
    putString(" (");
    putString(&gotName[0]);
    putString(" not ");
    putString(&wantedName[0]);
    putString(") ***\n");
    RunExiting = B_TRUE;
    traceBack();
    setToError(B_FALSE);
}

/*
 * symbolError - a run-time error involving a symbol.
 */

void
symbolError(CHAR_T *name, CHAR_T *message)
{

    setToError(B_TRUE);
    putString("*** ");
    putString(name);
    putString(": ");
    putString(message);
    putString(" ***\n");
    RunExiting = B_TRUE;
    traceBack();
    setToError(B_FALSE);
}

/*
 * procError - an error involving a proc.
 */

void
procError(ULONG_T key, CHAR_T *message)
{
    CHAR_T *name;

    name = findAnyName(key);
    if (name == P_NULL) {
	name = "<ACTION>";
    }
    symbolError(name, message);
}

/*
 * nilType - is the given type compatible with 'nil'?
 */

BOOL_T
nilType(UINT_T typ)
{

    return (BOOL_T) (typ == TYPE_PLAYER || typ == TYPE_THING ||
		     typ == TYPE_ACTION ||
		     (typ >= TYPE_TABLE && typ <= TYPE_PROP_LIST_ACTION));
}

#if DEBUG_PRESENT

static void
debugName(CHAR_T *what, ULONG_T key)
{
    CHAR_T *p;

    p = findAnyName(key);
    mess(what);
    messChar(' ');
    mess(p == P_NULL ? (CHAR_T *) "???" : p);
    messChar('\n');
}

#endif

/*
 * runBuiltIn - an interpreted proc call to a builtin proc.
 */

static void
runBuiltIn(BuiltinCall_t *bcl, QValue_t *pRes, UINT_T *pType)
{
    Proc_t *pr;
    Comp_t comp;
    ArgumentList_t *al;
    LocalList_t *ll;
    QValue_t *sl;
    Exec_t *ex;
    BOOL_T *pCs;
    QValue_t parameters[10];
    BOOL_T copySkipped[10];
    ULONG_T result;
    TraceBack_t tb;
    UINT_T count1, offset;

#if DEBUG_PRESENT
    if (WantDebug) {
	debugName("runBuiltIn", bcl->bcl_key);
    }
#endif
    pr = bcl->bcl_proc;
    al = bcl->bcl_arguments;
    ll = pr->pr_parameters;
    sl = &parameters[0];
    pCs = &copySkipped[0];
    count1 = 0;
    result = 0;	/* keep gcc happy */
    while (al != P_NULL && ! RunExiting) {
	/* Special optimization - do not copy string constants or string
	   local variables that we use as parameters to builtin functions.
	   This was most noticeable when building the tiled castle. Note
	   that we cannot do this optimization when calling user procs,
	   since they can change the string they are passed. That is fine
	   as LONG_T as they leave a valid string there. For builtins, the
	   string pointer is passed by value, so they cannot change it. */
	ex = al->al_this;
	if (ex->ex_kind == ex_string) {
	    sl->qv_string = ex->ex_v.ex_stringPtr;
	    *pType = TYPE_STRING;
	    *pCs = B_TRUE;
	} else if (ex->ex_kind == ex_local &&
	    ex->ex_v.ex_localPtr->ll_type == TYPE_STRING)
	{
	    offset = ex->ex_v.ex_localPtr->ll_offset;
	    sl->qv_string = (RunLocals + offset / sizeof(QValue_t))->qv_string;
	    *pType = TYPE_STRING;
	    *pCs = B_TRUE;
	} else {
	    runExec(ex, sl, pType);
	    *pCs = B_FALSE;
	}
	if (! RunExiting) {
	    if (*pType == TYPE_NIL && nilType(ll->ll_type)) {
		sl->qv_integer = (ULONG_T) ll->ll_type << TYPE_SHIFT;
	    }
#if CHECKS_PRESENT
	    if (ll->ll_type == TYPE_STRING && sl->qv_string == P_NULL) {
		myAbort("nil string in runBuiltIn #1");
	    }
#endif
	    ll = ll->ll_next;
	    al = al->al_next;
	    ++sl;
	    ++pCs;
	    count1 += 1;
	}
    }

    if (! RunExiting) {
	tb.tb_byteCode = B_FALSE;
	tb.tb_next = RunTraceBack;
	tb.tb_proc = pr;
	tb.tb_key = bcl->bcl_key;
	sl = &parameters[0];
	tb.tb_locals = sl;
	RunTraceBack = &tb;
	comp = pr->pr_compiled;

	switch (count1) {
	case 0:
	    result = comp();
	    break;
	case 1:
	    result = (*(LONG_T (*)(LONG_T)) comp)(sl->qv_integer);
	    break;
	case 2:
	    result = (*(LONG_T (*)(LONG_T, LONG_T)) comp)
			(sl->qv_integer, (sl + 1)->qv_integer);
	    break;
	case 3:
	    result = (*(LONG_T (*)(LONG_T, LONG_T, LONG_T)) comp)
			(sl->qv_integer, (sl + 1)->qv_integer,
			 (sl + 2)->qv_integer);
	    break;
	case 4:
	    result = (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T)) comp)
			(sl->qv_integer, (sl + 1)->qv_integer,
			 (sl + 2)->qv_integer, (sl + 3)->qv_integer);
	    break;
	case 5:
	    result = (*(LONG_T (*)(LONG_T, LONG_T, LONG_T,
				   LONG_T, LONG_T))comp)
			(sl->qv_integer, (sl + 1)->qv_integer,
			 (sl + 2)->qv_integer, (sl + 3)->qv_integer,
			 (sl + 4)->qv_integer);
	    break;
	case 6:
	    result = (*(LONG_T (*)(LONG_T, LONG_T, LONG_T,
				   LONG_T, LONG_T, LONG_T)) comp)
			(sl->qv_integer, (sl + 1)->qv_integer,
			 (sl + 2)->qv_integer, (sl + 3)->qv_integer,
			 (sl + 4)->qv_integer, (sl + 5)->qv_integer);
	    break;
	case 7:
	    result = (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T,
				   LONG_T, LONG_T, LONG_T)) comp)
			(sl->qv_integer, (sl + 1)->qv_integer,
			 (sl + 2)->qv_integer, (sl + 3)->qv_integer,
			 (sl + 4)->qv_integer, (sl + 5)->qv_integer,
			 (sl + 6)->qv_integer);
	    break;
	case 8:
	    result = (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T,
				   LONG_T, LONG_T, LONG_T, LONG_T)) comp)
			(sl->qv_integer, (sl + 1)->qv_integer,
			 (sl + 2)->qv_integer, (sl + 3)->qv_integer,
			 (sl + 4)->qv_integer, (sl + 5)->qv_integer,
			 (sl + 6)->qv_integer, (sl + 7)->qv_integer);
	    break;
	case 9:
	    result = (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T, LONG_T,
				   LONG_T, LONG_T, LONG_T, LONG_T)) comp)
			(sl->qv_integer, (sl + 1)->qv_integer,
			 (sl + 2)->qv_integer, (sl + 3)->qv_integer,
			 (sl + 4)->qv_integer, (sl + 5)->qv_integer,
			 (sl + 6)->qv_integer, (sl + 7)->qv_integer,
			 (sl + 8)->qv_integer);
	    break;
	case 10:
	    result = (*(LONG_T (*)(LONG_T, LONG_T, LONG_T, LONG_T, LONG_T,
				   LONG_T, LONG_T, LONG_T, LONG_T,LONG_T))comp)
			(sl->qv_integer, (sl + 1)->qv_integer,
			 (sl + 2)->qv_integer, (sl + 3)->qv_integer,
			 (sl + 4)->qv_integer, (sl + 5)->qv_integer,
			 (sl + 6)->qv_integer, (sl + 7)->qv_integer,
			 (sl + 8)->qv_integer, (sl + 9)->qv_integer);
	    break;
	default:
	    /* Having run off the end of the two size 10 arrays! */
	    procError(bcl->bcl_key, "too many pars to compiled proc");
	    break;
	}

	RunTraceBack = tb.tb_next;
    }

    ll = pr->pr_parameters;
    sl = &parameters[0];
    pCs = &copySkipped[0];
    while (ll != P_NULL && count1 != 0) {
	if (ll->ll_type == TYPE_STRING) {
#if CHECKS_PRESENT
	    if (sl->qv_string == P_NULL) {
		myAbort("nil string in runBuiltIn #2");
	    }
#endif
	    if (! *pCs) {
		freeString(sl->qv_string);
	    }
	}
	ll = ll->ll_next;
	++sl;
	++pCs;
	count1 -= 1;
    }

    pRes->qv_integer = result;
    *pType = pr->pr_resultType;
    if (RunExiting) {
	*pType = bcl->bcl_proc->pr_resultType;
	pRes->qv_integer = nilValue(*pType);
    }
}

/*
 * runUserProc - an interpreted proc call to a user defined proc.
 */

static void
runUserProc(UserCall_t *ucl, QValue_t *pRes, UINT_T *pType)
{
    LocalList_t *ll;
    QValue_t *sl, *newLocals, *localsSave;
    Proc_t *pr;
    ArgumentList_t *al;
    UINT_T *pLockCount;
    ULONG_T localCount, procKey, oldPlayer;
    UINT_T count1, resultType, parType;
    TraceBack_t tb;
    PlayerStatus_t oldStatus;
    QValue_t quickLocals[QUICK_LOCAL_COUNT];
    BOOL_T hadError;

    resultType = ucl->ucl_resultType;
    RunCallDepth += 1;
    if (RunCallDepth > MAX_CALL_DEPTH) {
	runError("proc call depth limit exceeded");
	procKey = NIL_ACTION;
	hadError = B_TRUE;
    } else {
	hadError = B_FALSE;
	runExec(ucl->ucl_action, pRes, pType);
	if (*pType == TYPE_ERROR) {
	    /* should not happen at run-time, but let's be paranoid */
	    procKey = NIL_ACTION;
	    hadError = B_TRUE;
	} else {
	    procKey = pRes->qv_integer;
	    if (procKey == NIL_ACTION) {
		runError("attempt to call nil proc");
		hadError = B_TRUE;
	    }
	}
    }
    if (procKey != NIL_ACTION) {
#if DEBUG_PRESENT
	if (WantDebug) {
	    debugName("runUserProc", procKey);
	}
#endif
	pr = findCachedProc(procKey, &pLockCount);
	if (pr->pr_resultType != resultType) {
	    runTypeError(P_NULL, "dynamic proc has wrong result type",
		pr->pr_resultType, resultType);
	    hadError = B_TRUE;
	} else if (pr->pr_wizard && EffectiveStatus != ps_wizard) {
	    procError(procKey, "apprentice cannot call this proc");
	    hadError = B_TRUE;
	} else {

	    localCount = pr->pr_localCount;
	    if (localCount != BC_STACK_EXTRA) {
		if (localCount <= QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		    newLocals = &quickLocals[0];
		} else {
		    newLocals = MEM_ALLOC_E(&MS, localCount, "lclu");
		}
	    } else {
		if (pr->pr_parameters != P_NULL) {
		    myAbort("got parameters with 0 localCount!");
		}
		if (pr->pr_locals != P_NULL) {
		    myAbort("got locals with 0 localCount!");
		}
		newLocals = P_NULL;	/* there are none */
	    }
	    sl = newLocals;

	    /* count1 is maintained as the count of the parameters that have
	       been initialized in the local area. Only that many are cleared
	       out afterwards. */
	    count1 = 0;
	    ll = pr->pr_parameters;
	    al = ucl->ucl_arguments;
	    while (ll != P_NULL && al != P_NULL && ! RunExiting) {
		runExec(al->al_this, sl, &parType);
		if (! RunExiting) {
		    if (parType == TYPE_NIL && nilType(ll->ll_type)) {
			parType = ll->ll_type;
			sl->qv_integer = (ULONG_T) parType << TYPE_SHIFT;
		    }
#if CHECKS_PRESENT
		    if (ll->ll_type == TYPE_STRING && sl->qv_string == P_NULL){
			myAbort("nil string in runUserProc #1");
		    }
#endif
		    if (ll->ll_type != parType) {
			if (parType != TYPE_ERROR) {
			    runTypeError(ll->ll_name,
				"run-time argument type mismatch",
				parType, ll->ll_type);
			}
			RunExiting = B_TRUE;
			if (parType == TYPE_STRING) {
			    freeString(sl->qv_string);
			}
		    } else {
			ll = ll->ll_next;
			al = al->al_next;
			++sl;
			count1 += 1;
		    }
		}
	    }

	    if (! RunExiting) {
		if (ll != P_NULL) {
		    procError(procKey, "run-time not enough arguments");
		} else if (al != P_NULL) {
		    procError(procKey, "run-time too many arguments");
		}
	    }
	    checkAbort();

	    if (! RunExiting) {
		oldPlayer = EffectivePlayer;
		oldStatus = EffectiveStatus;
		if (! pr->pr_utility || EffectivePlayer == SYSADMIN_KEY ||
		    (EffectiveStatus == ps_wizard &&
			pr->pr_ownerStatus == ps_apprentice))
		{
		    /* The second part of the above test is to prevent some
		       security problems. Without the test, consider this:
			a wizard puts a checker onto a room exit
			the checker is marked utility
			SysAdmin walks through the room, thus executing the
			    wizard's code, with EffectivePlayer == SysAdmin
			that wizard's code can then do, e.g. Execute!
		       The third part of the test is for a similar situation
		       which would allow an apprentice to execute as wizard.
		    */
		    EffectivePlayer = pr->pr_owner;
		    EffectiveStatus = pr->pr_ownerStatus;
		}
		if (pr->pr_isCompiled) {
		    /* newLocals can be nil, but then count1 is 0 */
		    doBcRun(procKey, pr, newLocals, pRes, count1);
		    *pType = pr->pr_resultType;
		} else {
		    tb.tb_byteCode = B_FALSE;
		    tb.tb_next = RunTraceBack;
		    tb.tb_proc = pr;
		    tb.tb_key = procKey;
		    tb.tb_locals = newLocals;
		    RunTraceBack = &tb;
		    ll = pr->pr_locals;
		    /* only need to add one, so that the locals are cleared
		       out after the body is executed */
		    count1 += 1;
		    sl = (QValue_t *) ((BYTE_T *) sl + BC_STACK_EXTRA);
		    while (ll != P_NULL) {
			sl->qv_integer = nilValue(ll->ll_type);
			++sl;
			ll = ll->ll_next;
		    }
		    localsSave = RunLocals;
		    RunLocals = newLocals;

		    runExec(pr->pr_body, pRes, pType);

		    if (*pType == TYPE_NIL && nilType(resultType)) {
			*pType = resultType;
			pRes->qv_integer = (ULONG_T) resultType << TYPE_SHIFT;
		    }
		    if (*pType != resultType) {
			if (*pType != TYPE_ERROR) {
			    /* this isn't possible, but check anyway */
			    procError(procKey,"run-time result type mismatch");
			}
			RunExiting = B_TRUE;
		    }
		    RunLocals = localsSave;
		    RunTraceBack = tb.tb_next;
		}
		/* ALWAYS restore - body could have SetEffectiveToNone */
		EffectivePlayer = oldPlayer;
		EffectiveStatus = oldStatus;
	    }

	    if (RunExiting || ! pr->pr_isCompiled) {
		/* byte-code procs will already have cleaned up. */
		ll = pr->pr_parameters;
		sl = newLocals;
		while (ll != P_NULL && count1 != 0) {
		    if (ll->ll_type == TYPE_STRING) {
#if CHECKS_PRESENT
			if (sl->qv_string == P_NULL) {
			    myAbort("nil string in runUserProc #2");
			}
#endif
			freeString(sl->qv_string);
		    }
		    ll = ll->ll_next;
		    ++sl;
		    count1 -= 1;
		}

		if (count1 != 0) {
		    sl = (QValue_t *) ((BYTE_T *) sl + BC_STACK_EXTRA);
		    ll = pr->pr_locals;
		    while (ll != P_NULL) {
			if (ll->ll_type == TYPE_STRING) {
#if CHECKS_PRESENT
			    if (sl->qv_string == P_NULL) {
				myAbort("nil string in runUserProc #3");
			    }
#endif
			    freeString(sl->qv_string);
			}
			ll = ll->ll_next;
			++sl;
		    }
		}
	    }

	    if (localCount > QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		memFreeE(&MS, newLocals, localCount);
	    }
	    if (RunExiting) {
		hadError = B_TRUE;
	    }
	}

	*pLockCount -= 1;

    }
    if (hadError) {
	pRes->qv_integer = nilValue(resultType);
    }
    *pType = resultType;
    RunCallDepth -= 1;
}

/*
 * doProc0 - allows the parser to call 0-parameter actions.
 */

ULONG_T
doProc0(ULONG_T key)
{
    Proc_t *pr;
    LocalList_t *ll;
    QValue_t *sl, *localsSave, *newLocals;
    UINT_T *pLockCount;
    QValue_t quickLocals[QUICK_LOCAL_COUNT];
    ULONG_T localCount, oldPlayer;
    PlayerStatus_t oldStatus;
    UINT_T resType;
    QValue_t res;
    TraceBack_t tb;

#if DEBUG_PRESENT
    if (WantDebug) {
	debugName("doProc0", key);
    }
#endif
    RunCallDepth += 1;
    if (RunCallDepth > MAX_CALL_DEPTH) {
	runError("proc call depth limit exceeded");
	pr = P_NULL;	/* keep gcc happy */
    } else {
	pr = findCachedProc(key, &pLockCount);
	tb.tb_byteCode = B_FALSE;
	tb.tb_next = RunTraceBack;
	tb.tb_proc = pr;
	tb.tb_key = key;
	tb.tb_locals = P_NULL;
	if (pr->pr_compiled != P_NULL) {
	    RunTraceBack = &tb;
	    res.qv_integer = pr->pr_compiled();
	    RunTraceBack = tb.tb_next;
	} else {
	    oldPlayer = EffectivePlayer;
	    oldStatus = EffectiveStatus;
	    if (! pr->pr_utility || EffectivePlayer == SYSADMIN_KEY ||
		(EffectiveStatus == ps_wizard &&
		    pr->pr_ownerStatus == ps_apprentice))
	    {
		EffectivePlayer = pr->pr_owner;
		EffectiveStatus = pr->pr_ownerStatus;
	    }
	    localCount = pr->pr_localCount;
	    if (localCount != BC_STACK_EXTRA) {
		if (localCount <= QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		    newLocals = &quickLocals[0];
		} else {
		    newLocals = MEM_ALLOC_E(&MS, localCount, "lcl0");
		}
		sl = newLocals;
		ll = pr->pr_locals;
		sl += BC_STACK_EXTRA / sizeof(QValue_t);
		while (ll != P_NULL) {
		    sl->qv_integer = nilValue(ll->ll_type);
		    ll = ll->ll_next;
		    ++sl;
		}
	    } else {
		/* Keep gcc happy */
		newLocals = P_NULL;
	    }
	    if (pr->pr_isCompiled) {
		doBcRun(key, pr, newLocals, &res, 0);
	    } else {
		RunTraceBack = &tb;
		localsSave = RunLocals;
		RunLocals = newLocals;
		runExec(pr->pr_body, &res, &resType);
		RunLocals = localsSave;
		RunTraceBack = tb.tb_next;
	    }
	    if (localCount != BC_STACK_EXTRA) {
		if (RunExiting || ! pr->pr_isCompiled) {
		    ll = pr->pr_locals;
		    sl = newLocals + BC_STACK_EXTRA / sizeof(QValue_t);
		    while (ll != P_NULL) {
			if (ll->ll_type == TYPE_STRING) {
			    if (sl->qv_string != P_NULL) {
				freeString(sl->qv_string);
			    }
			}
			ll = ll->ll_next;
			++sl;
		    }
		}
		if (localCount > QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		    memFreeE(&MS, newLocals, localCount);
		}
	    }
	    EffectivePlayer = oldPlayer;
	    EffectiveStatus = oldStatus;
	}
	*pLockCount -= 1;
    }
    if (RunExiting) {
	res.qv_integer = nilValue(pr->pr_resultType);
    }
    RunCallDepth -= 1;
    return (ULONG_T) res.qv_integer;
}

/*
 * doProc0Top - allow top-level calls to 0-parameter actions.
 */

ULONG_T
doProc0Top(ULONG_T key)
{
    ULONG_T res;

    RunCallDepth = 0;
    RunTraceBack = P_NULL;
    RunExiting = B_FALSE;
    RunAborting = B_FALSE;
    BcAbort = B_FALSE;
    startTimeout();
    res = doProc0(key);
    stopTimeout();
    return res;
}

/*
 * doProc1 - allows the parser to call 1-parameter actions.
 */

ULONG_T
doProc1(ULONG_T key, void *par1)
{
    Proc_t *pr;
    LocalList_t *ll;
    QValue_t *sl, *localsSave, *newLocals;
    UINT_T *pLockCount;
    QValue_t quickLocals[QUICK_LOCAL_COUNT];
    ULONG_T localCount, oldPlayer;
    PlayerStatus_t oldStatus;
    UINT_T resType;
    QValue_t res;
    TraceBack_t tb;

#if DEBUG_PRESENT
    if (WantDebug) {
	debugName("doProc1", key);
    }
#endif
    RunCallDepth += 1;
    if (RunCallDepth > MAX_CALL_DEPTH) {
	runError("proc call depth limit exceeded");
	pr = P_NULL;	/* keep gcc happy */
    } else {
	pr = findCachedProc(key, &pLockCount);
	tb.tb_byteCode = B_FALSE;
	tb.tb_next = RunTraceBack;
	tb.tb_proc = pr;
	tb.tb_key = key;
	if (pr->pr_compiled != P_NULL) {
	    RunTraceBack = &tb;
	    tb.tb_locals = (QValue_t *) &par1;
	    res.qv_integer = (*(ULONG_T (*)(void *)) pr->pr_compiled)(par1);
	    if (pr->pr_parameters->ll_type == TYPE_STRING) {
		freeString(par1);
	    }
	    RunTraceBack = tb.tb_next;
	} else {
	    oldPlayer = EffectivePlayer;
	    oldStatus = EffectiveStatus;
	    if (! pr->pr_utility || EffectivePlayer == SYSADMIN_KEY ||
		(EffectiveStatus == ps_wizard &&
		    pr->pr_ownerStatus == ps_apprentice))
	    {
		EffectivePlayer = pr->pr_owner;
		EffectiveStatus = pr->pr_ownerStatus;
	    }
	    localCount = pr->pr_localCount;
	    if (localCount <= QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		newLocals = &quickLocals[0];
	    } else {
		newLocals = MEM_ALLOC_E(&MS, localCount, "lcl1");
	    }
	    sl = newLocals;
	    ll = pr->pr_locals;
	    tb.tb_locals = sl;
	    sl->qv_string = par1;
	    sl += BC_STACK_EXTRA / sizeof(QValue_t);
	    while (ll != P_NULL) {
		++sl;
		sl->qv_integer = nilValue(ll->ll_type);
		ll = ll->ll_next;
	    }
	    if (pr->pr_isCompiled) {
		doBcRun(key, pr, newLocals, &res, 1);
	    } else {
		RunTraceBack = &tb;
		localsSave = RunLocals;
		RunLocals = newLocals;
		runExec(pr->pr_body, &res, &resType);
		RunLocals = localsSave;
		RunTraceBack = tb.tb_next;
	    }
	    if (RunExiting || ! pr->pr_isCompiled) {
		sl = newLocals;
		if (pr->pr_parameters->ll_type == TYPE_STRING) {
		    freeString(sl->qv_string);
		}
		sl += BC_STACK_EXTRA / sizeof(QValue_t);
		ll = pr->pr_locals;
		while (ll != P_NULL) {
		    ++sl;
		    if (ll->ll_type == TYPE_STRING) {
			if (sl->qv_string != P_NULL) {
			    freeString(sl->qv_string);
			}
		    }
		    ll = ll->ll_next;
		}
	    }
	    if (localCount > QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		memFreeE(&MS, newLocals, localCount);
	    }
	    EffectivePlayer = oldPlayer;
	    EffectiveStatus = oldStatus;
	}
	*pLockCount -= 1;
    }
    if (RunExiting) {
	res.qv_integer = nilValue(pr->pr_resultType);
    }
    RunCallDepth -= 1;
    return (ULONG_T) res.qv_integer;
}

/*
 * doProc1Top - allow top-level calls to 1-parameter actions.
 */

ULONG_T
doProc1Top(ULONG_T key, void *par1)
{
    ULONG_T res;

    RunCallDepth = 0;
    RunTraceBack = P_NULL;
    RunExiting = B_FALSE;
    RunAborting = B_FALSE;
    BcAbort = B_FALSE;
    startTimeout();
    res = doProc1(key, par1);
    stopTimeout();
    return res;
}

/*
 * doProc2 - allows the parser to call 2-parameter actions.
 */

ULONG_T
doProc2(ULONG_T key, void *par1, void *par2)
{
    Proc_t *pr;
    LocalList_t *ll;
    QValue_t *sl, *localsSave, *newLocals;
    UINT_T *pLockCount;
    QValue_t quickLocals[QUICK_LOCAL_COUNT];
    ULONG_T localCount, oldPlayer;
    PlayerStatus_t oldStatus;
    UINT_T resType;
    QValue_t res;
    TraceBack_t tb;
    QValue_t parArray[2];

#if DEBUG_PRESENT
    if (WantDebug) {
	debugName("doProc2", key);
    }
#endif
    RunCallDepth += 1;
    if (RunCallDepth > MAX_CALL_DEPTH) {
	runError("proc call depth limit exceeded");
	pr = P_NULL;	/* keep gcc happy */
    } else {
	pr = findCachedProc(key, &pLockCount);
	tb.tb_byteCode = B_FALSE;
	tb.tb_next = RunTraceBack;
	tb.tb_proc = pr;
	tb.tb_key = key;
	if (pr->pr_compiled != P_NULL) {
	    RunTraceBack = &tb;
	    parArray[0].qv_string = par1;
	    parArray[1].qv_string = par2;
	    tb.tb_locals = &parArray[0];
	    res.qv_integer =
		(*(ULONG_T (*)(void *, void *)) pr->pr_compiled)(par1, par2);
	    ll = pr->pr_parameters;
	    if (ll->ll_type == TYPE_STRING) {
		freeString(par1);
	    }
	    ll = ll->ll_next;
	    if (ll->ll_type == TYPE_STRING) {
		freeString(par2);
	    }
	    RunTraceBack = tb.tb_next;
	} else {
	    oldPlayer = EffectivePlayer;
	    oldStatus = EffectiveStatus;
	    if (! pr->pr_utility || EffectivePlayer == SYSADMIN_KEY ||
		(EffectiveStatus == ps_wizard &&
		    pr->pr_ownerStatus == ps_apprentice))
	    {
		EffectivePlayer = pr->pr_owner;
		EffectiveStatus = pr->pr_ownerStatus;
	    }
	    localCount = pr->pr_localCount;
	    if (localCount <= QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		newLocals = &quickLocals[0];
	    } else {
		newLocals = MEM_ALLOC_E(&MS, localCount, "lcl2");
	    }
	    ll = pr->pr_locals;
	    sl = newLocals;
	    tb.tb_locals = sl;
	    sl->qv_string = par1;
	    ++sl;
	    sl->qv_string = par2;
	    sl += BC_STACK_EXTRA / sizeof(QValue_t);
	    while (ll != P_NULL) {
		++sl;
		sl->qv_integer = nilValue(ll->ll_type);
		ll = ll->ll_next;
	    }
	    if (pr->pr_isCompiled) {
		doBcRun(key, pr, newLocals, &res, 2);
	    } else {
		RunTraceBack = &tb;
		localsSave = RunLocals;
		RunLocals = newLocals;
		runExec(pr->pr_body, &res, &resType);
		RunLocals = localsSave;
		RunTraceBack = tb.tb_next;
	    }
	    if (RunExiting || ! pr->pr_isCompiled) {
		sl = newLocals;
		ll = pr->pr_parameters;
		if (ll->ll_type == TYPE_STRING) {
		    freeString(sl->qv_string);
		}
		++sl;
		ll = ll->ll_next;
		if (ll->ll_type == TYPE_STRING) {
		    freeString(sl->qv_string);
		}
		sl += BC_STACK_EXTRA / sizeof(QValue_t);
		ll = pr->pr_locals;
		while (ll != P_NULL) {
		    ++sl;
		    if (ll->ll_type == TYPE_STRING) {
			if (sl->qv_string != P_NULL) {
			    freeString(sl->qv_string);
			}
		    }
		    ll = ll->ll_next;
		}
	    }
	    if (localCount > QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		memFreeE(&MS, newLocals, localCount);
	    }
	    EffectivePlayer = oldPlayer;
	    EffectiveStatus = oldStatus;
	}
	*pLockCount -= 1;
    }
    if (RunExiting) {
	res.qv_integer = nilValue(pr->pr_resultType);
    }
    RunCallDepth -= 1;
    return (ULONG_T) res.qv_integer;
}

/*
 * doProc2Top - allow top-level calls to 2-parameter actions.
 */

ULONG_T
doProc2Top(ULONG_T key, void *par1, void *par2)
{
    ULONG_T res;

    RunCallDepth = 0;
    RunTraceBack = P_NULL;
    RunExiting = B_FALSE;
    RunAborting = B_FALSE;
    BcAbort = B_FALSE;
    startTimeout();
    res = doProc2(key, par1, par2);
    stopTimeout();
    return res;
}

/*
 * doProc3 - allows the parser to call 3-parameter actions.
 */

ULONG_T
doProc3(ULONG_T key, void *par1, void *par2, void *par3)
{
    Proc_t *pr;
    LocalList_t *ll;
    QValue_t *sl, *localsSave, *newLocals;
    UINT_T *pLockCount;
    QValue_t quickLocals[QUICK_LOCAL_COUNT];
    ULONG_T localCount, oldPlayer;
    PlayerStatus_t oldStatus;
    UINT_T resType;
    QValue_t res;
    TraceBack_t tb;
    QValue_t parArray[3];

#if DEBUG_PRESENT
    if (WantDebug) {
	debugName("doProc3", key);
    }
#endif
    RunCallDepth += 1;
    if (RunCallDepth > MAX_CALL_DEPTH) {
	runError("proc call depth limit exceeded");
	pr = P_NULL;	/* keep gcc happy */
    } else {
	pr = findCachedProc(key, &pLockCount);
	tb.tb_byteCode = B_FALSE;
	tb.tb_next = RunTraceBack;
	tb.tb_proc = pr;
	tb.tb_key = key;
	if (pr->pr_compiled != P_NULL) {
	    RunTraceBack = &tb;
	    parArray[0].qv_string = par1;
	    parArray[1].qv_string = par2;
	    parArray[2].qv_string = par3;
	    tb.tb_locals = &parArray[0];
	    res.qv_integer =
		(*(ULONG_T (*)(void *, void *, void *)) pr->pr_compiled)
		    (par1, par2, par3);
	    ll = pr->pr_parameters;
	    if (ll->ll_type == TYPE_STRING) {
		freeString(par1);
	    }
	    ll = ll->ll_next;
	    if (ll->ll_type == TYPE_STRING) {
		freeString(par2);
	    }
	    ll = ll->ll_next;
	    if (ll->ll_type == TYPE_STRING) {
		freeString(par3);
	    }
	    RunTraceBack = tb.tb_next;
	} else {
	    oldPlayer = EffectivePlayer;
	    oldStatus = EffectiveStatus;
	    if (! pr->pr_utility || EffectivePlayer == SYSADMIN_KEY ||
		(EffectiveStatus == ps_wizard &&
		    pr->pr_ownerStatus == ps_apprentice))
	    {
		EffectivePlayer = pr->pr_owner;
		EffectiveStatus = pr->pr_ownerStatus;
	    }
	    localCount = pr->pr_localCount;
	    if (localCount <= QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		newLocals = &quickLocals[0];
	    } else {
		newLocals = MEM_ALLOC_E(&MS, localCount, "lcl3");
	    }
	    ll = pr->pr_locals;
	    sl = newLocals;
	    tb.tb_locals = sl;
	    sl->qv_string = par1;
	    ++sl;
	    sl->qv_string = par2;
	    ++sl;
	    sl->qv_string = par3;
	    sl += BC_STACK_EXTRA / sizeof(QValue_t);
	    while (ll != P_NULL) {
		++sl;
		sl->qv_integer = nilValue(ll->ll_type);
		ll = ll->ll_next;
	    }
	    if (pr->pr_isCompiled) {
		doBcRun(key, pr, newLocals, &res, 3);
	    } else {
		RunTraceBack = &tb;
		localsSave = RunLocals;
		RunLocals = newLocals;
		runExec(pr->pr_body, &res, &resType);
		RunLocals = localsSave;
		RunTraceBack = tb.tb_next;
	    }
	    if (RunExiting || ! pr->pr_isCompiled) {
		sl = newLocals;
		ll = pr->pr_parameters;
		if (ll->ll_type == TYPE_STRING) {
		    freeString(sl->qv_string);
		}
		++sl;
		ll = ll->ll_next;
		if (ll->ll_type == TYPE_STRING) {
		    freeString(sl->qv_string);
		}
		++sl;
		ll = ll->ll_next;
		if (ll->ll_type == TYPE_STRING) {
		    freeString(sl->qv_string);
		}
		sl += BC_STACK_EXTRA / sizeof(QValue_t);
		ll = pr->pr_locals;
		while (ll != P_NULL) {
		    ++sl;
		    if (ll->ll_type == TYPE_STRING) {
			if (sl->qv_string != P_NULL) {
			    freeString(sl->qv_string);
			}
		    }
		    ll = ll->ll_next;
		}
	    }
	    if (localCount > QUICK_LOCAL_COUNT * sizeof(QValue_t)) {
		memFreeE(&MS, newLocals, localCount);
	    }
	    EffectivePlayer = oldPlayer;
	    EffectiveStatus = oldStatus;
	}
	*pLockCount -= 1;
    }
    if (RunExiting) {
	res.qv_integer = nilValue(pr->pr_resultType);
    }
    RunCallDepth -= 1;
    return (ULONG_T) res.qv_integer;
}

/*
 * doProc3Top - allow top-level calls to 3-parameter actions.
 */

ULONG_T
doProc3Top(ULONG_T key, void *par1, void *par2, void *par3)
{
    ULONG_T res;

    RunCallDepth = 0;
    RunTraceBack = P_NULL;
    RunExiting = B_FALSE;
    RunAborting = B_FALSE;
    BcAbort = B_FALSE;
    startTimeout();
    res = doProc3(key, par1, par2, par3);
    stopTimeout();
    return res;
}

/*
 * fetchString - check for and do string-copy avoidance.
 */

static BOOL_T
fetchString(Exec_t *ex, QValue_t *pRes)
{
    UINT_T typ, offset;

    if (ex->ex_kind == ex_string) {
	pRes->qv_string = ex->ex_v.ex_stringPtr;
	return B_FALSE;
    } else if (ex->ex_kind == ex_local) {
	offset = ex->ex_v.ex_localPtr->ll_offset;
	pRes->qv_string = (RunLocals + offset / sizeof(QValue_t))->qv_string;
	return B_FALSE;
    } else {
	/* string is copied inside this call */
	runExec(ex, pRes, &typ);
	return B_TRUE;
    }
}

/*
 * mulFixed - multiply a pair of fixed's.
 */

LONG_T
mulFixed(LONG_T a, LONG_T b)
{
    ULONG_T ua, ub, lowa, lowb;
    BOOL_T resNeg;

    resNeg = B_FALSE;
    if (a < 0) {
	a = -a;
	resNeg = B_TRUE;
    }
    if (b < 0) {
	b = -b;
	resNeg = ! resNeg;
    }
    lowa = a & MAX_FRAC;
    lowb = b & MAX_FRAC;
    ua = (ULONG_T) a;
    ub = (ULONG_T) b;
    ua = ua >> FRAC_BITS;
    ub = ub >> FRAC_BITS;
    a = ((ua * ub) << FRAC_BITS) +
	(ua * lowb) + (ub * lowa) +
	((lowa * lowb) >> FRAC_BITS);
    if (resNeg) {
	a = -a;
    }
    return a;
}

LONG_T
divFixed(LONG_T left, LONG_T right)
{
    long long int l, r;

    l = (long long int) left;
    r = (long long int) right;
    return (LONG_T) ((l << 16) / r);
}

void
runExec(Exec_t *ex, QValue_t *pRes, UINT_T *pType)
{
    Expression_t *exp;
    QValue_t leftValue, rightValue;
    int res;

    /* we do not check RunExiting here since we want the correct result type
       from any call of 'runExec' - we do not check it in higher level
       places (like the string operators). */
    if (ex == P_NULL) {
	*pType = TYPE_VOID;
    } else {
	switch (ex->ex_kind) {
	case ex_error:
	    pRes->qv_integer = 1;
	    *pType = TYPE_ERROR;
	    runError("executing an erroneous executable");
	    break;
	case ex_empty:
	    *pType = TYPE_VOID;
	    break;
	case ex_intRef:
	    pRes->qv_integer =
		((ULONG_T) ex->ex_intVal[0] << 16) |
		((ULONG_T) ex->ex_intVal[1] << 8) |
		(ULONG_T) ex->ex_intVal[2];
	    if ((ex->ex_intVal[0] & 0x80) != 0) {
		pRes->qv_integer |= 0xff000000;
	    }
	    *pType = TYPE_INTEGER;
	    break;
	case ex_fixedRef:
	    pRes->qv_integer =
		((ULONG_T) ex->ex_intVal[0] << 16) |
		((ULONG_T) ex->ex_intVal[1] << 8) |
		(ULONG_T) ex->ex_intVal[2];
	    if ((ex->ex_intVal[0] & 0x80) != 0) {
		pRes->qv_integer |= 0xff000000;
	    }
	    *pType = TYPE_FIXED;
	    break;
	case ex_procRef:
	    pRes->qv_integer = ex->ex_v.ex_refPtr->rl_key;
	    *pType = TYPE_ACTION;
	    break;
	case ex_thingRef:
	    pRes->qv_integer = ex->ex_v.ex_refPtr->rl_key;
	    *pType = TYPE_THING;
	    break;
	case ex_grammarRef:
	    pRes->qv_integer = ex->ex_v.ex_refPtr->rl_key;
	    *pType = TYPE_GRAMMAR;
	    break;
	case ex_playerRef:
	    pRes->qv_integer = ex->ex_v.ex_refPtr->rl_key;
	    *pType = TYPE_PLAYER;
	    break;
	case ex_propRef:
	    pRes->qv_integer = ex->ex_v.ex_refPtr->rl_key;
	    *pType = (ULONG_T) pRes->qv_integer >> TYPE_SHIFT;
	    break;
	case ex_tableRef:
	    pRes->qv_integer = ex->ex_v.ex_refPtr->rl_key;
	    *pType = TYPE_TABLE;
	    break;
	case ex_local: {
	    LocalList_t *ll;

	    ll = ex->ex_v.ex_localPtr;
	    pRes->qv_integer =
		(RunLocals + ll->ll_offset / sizeof(QValue_t))->qv_integer;
	    *pType = ll->ll_type;
	    if (*pType == TYPE_STRING) {
		pRes->qv_string = stringCopy(pRes->qv_string);
	    }
	    break;
	}
	case ex_string:
	    pRes->qv_string = stringCopy(ex->ex_v.ex_stringPtr);
	    *pType = TYPE_STRING;
	    break;
	case ex_int:
	    pRes->qv_integer = ex->ex_v.ex_integer;
	    *pType = TYPE_INTEGER;
	    break;
	case ex_fixed:
	    pRes->qv_integer = ex->ex_v.ex_integer;
	    *pType = TYPE_FIXED;
	    break;
	case ex_bool:
	    pRes->qv_integer = ex->ex_v.ex_integer;
	    *pType = TYPE_BOOLEAN;
	    break;
	case ex_status:
	    pRes->qv_integer = ex->ex_v.ex_integer;
	    *pType = TYPE_STATUS;
	    break;
	case ex_nil:
	    pRes->qv_integer = ex->ex_v.ex_integer;
	    *pType = TYPE_NIL;
	    break;
	case ex_ts:
	    pRes->qv_integer = ex->ex_v.ex_integer;
	    *pType = TYPE_THING_STATUS;
	    break;
	case ex_if: {
	    IfList_t *ifl;

	    ifl = ex->ex_v.ex_ifPtr->if_alternatives;
	    while (ifl != P_NULL) {
		runExec(ifl->ifl_condition, &leftValue, pType);
		if (leftValue.qv_integer != BOOLEAN_FALSE) {
		    break;
		}
		ifl = ifl->ifl_next;
	    }
	    if (ifl != P_NULL) {
		runExec(ifl->ifl_body, pRes, pType);
	    } else {
		if (ex->ex_v.ex_ifPtr->if_else == P_NULL) {
		    *pType = TYPE_VOID;
		} else {
		    runExec(ex->ex_v.ex_ifPtr->if_else, pRes, pType);
		}
	    }
	    break;
	}
	case ex_while: {
	    While_t *wh;

	    wh = ex->ex_v.ex_whilePtr;
	    while (B_TRUE) {
		runExec(wh->wh_condition, &leftValue, pType);
		if (leftValue.qv_integer == BOOLEAN_FALSE || RunExiting) {
		    break;
		}
		runExec(wh->wh_body, pRes, pType);
		checkAbort();
	    }
	    *pType = TYPE_VOID;
	    break;
	}
	case ex_for: {
	    For_t *fo;
	    LocalList_t *ll;

	    fo = ex->ex_v.ex_forPtr;
	    ll = fo->fo_variable;
	    if (ll == P_NULL) {
		runError("executing an erroneous 'for'");
	    } else {
		pRes = RunLocals + ll->ll_offset / sizeof(QValue_t);
		runExec(fo->fo_start, pRes, pType);
		runExec(fo->fo_limit, &rightValue, pType);
		while (pRes->qv_integer <= rightValue.qv_integer &&
		       ! RunExiting)
		{
		    runExec(fo->fo_body, &leftValue, pType);
		    checkAbort();
		    pRes->qv_integer += 1;
		}
	    }
	    *pType = TYPE_VOID;
	    break;
	}
	case ex_case: {
	    Case_t *ca;
	    CaseList_t *cal;
	    CaseIndexList_t *cil;
	    Exec_t *exLocal;
	    ULONG_T n;

	    ca = ex->ex_v.ex_casePtr;
	    runExec(ca->ca_selector, &leftValue, pType);
	    cal = ca->ca_alternatives;
	    while (cal != P_NULL) {
		cil = cal->cal_indexes;
		while (cil != P_NULL) {
		    exLocal = cil->cil_index;
		    if (exLocal->ex_kind == ex_int) {
			n = exLocal->ex_v.ex_integer;
		    } else {
			n = ((ULONG_T) exLocal->ex_intVal[0] << 16) |
			    ((ULONG_T) exLocal->ex_intVal[1] << 8) |
			    (ULONG_T) exLocal->ex_intVal[2];
			if ((exLocal->ex_intVal[0] & 0x80) != 0) {
			    n |= 0xff000000;
			}
		    }
		    if (n == leftValue.qv_integer) {
			break;
		    }
		    cil = cil->cil_next;
		}
		if (cil != P_NULL) {
		    break;
		}
		cal = cal->cal_next;
	    }
	    if (cal != P_NULL) {
		runExec(cal->cal_body, pRes, pType);
	    } else {
		exLocal = ex->ex_v.ex_casePtr->ca_default;
		if (exLocal == P_NULL) {
		    *pType = TYPE_VOID;
		} else {
		    runExec(exLocal, pRes, pType);
		}
	    }
	    break;
	}
	case ex_ignore:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_right, &rightValue, pType);
	    if (*pType == TYPE_STRING) {
		freeString(rightValue.qv_string);
	    }
	    pRes->qv_integer = 0;
	    *pType = TYPE_VOID;
	    break;
	case ex_builtinCall:
	    /* much less checking needed here */
	    runBuiltIn(ex->ex_v.ex_builtinCallPtr, pRes, pType);
	    break;
	case ex_knownCall:
	case ex_unknownCall:
	    runUserProc(ex->ex_v.ex_userCallPtr, pRes, pType);
	    break;
	case ex_note:
	    /* do nothing for a note */
	    *pType = TYPE_VOID;
	    break;
	case ex_getBoolProp:
	case ex_getIntProp:
	case ex_getFixedProp:
	case ex_getThingProp:
	case ex_getActionProp:
	case ex_getStringProp:
	case ex_getTableProp:
	case ex_getGrammarProp:
	case ex_getIntListProp:
	case ex_getFixedListProp:
	case ex_getThingListProp:
	case ex_getActionListProp: {
	    PropGet_t *prg;

	    prg = ex->ex_v.ex_getPtr;
	    runExec(prg->prg_thing, pRes, pType);
	    runExec(prg->prg_property, &rightValue, pType);
	    *pType = ex->ex_kind - ex_getBoolProp + TYPE_BOOLEAN;
	    pRes->qv_integer =
		getProp(pRes->qv_integer, rightValue.qv_integer, *pType);
	    break;
	}
	case ex_addBoolProp:
	case ex_addIntProp:
	case ex_addFixedProp:
	case ex_addThingProp:
	case ex_addActionProp:
	case ex_addStringProp:
	case ex_addTableProp:
	case ex_addGrammarProp:
	case ex_addIntListProp:
	case ex_addFixedListProp:
	case ex_addThingListProp:
	case ex_addActionListProp: {
	    PropAdd_t *pra;
	    BOOL_T rightCopied;

	    pra = ex->ex_v.ex_addPtr;
	    runExec(pra->pra_thing, pRes, pType);
	    runExec(pra->pra_property, &leftValue, pType);
	    if (ex->ex_kind == ex_addStringProp) {
		rightCopied = fetchString(pra->pra_value, &rightValue);
	    } else {
		runExec(pra->pra_value, &rightValue, pType);
		rightCopied = B_FALSE;
	    }
	    addProp(pRes->qv_integer, leftValue.qv_integer,
		    rightValue.qv_integer,
		    ex->ex_kind - ex_addBoolProp + TYPE_BOOLEAN);
	    if (rightCopied) {
		freeString(rightValue.qv_string);
	    }
	    *pType = TYPE_VOID;
	    break;
	}
	case ex_delBoolProp:
	case ex_delIntProp:
	case ex_delFixedProp:
	case ex_delThingProp:
	case ex_delActionProp:
	case ex_delStringProp:
	case ex_delTableProp:
	case ex_delGrammarProp:
	case ex_delIntListProp:
	case ex_delFixedListProp:
	case ex_delThingListProp:
	case ex_delActionListProp: {
	    PropDel_t *prd;

	    prd = ex->ex_v.ex_delPtr;
	    runExec(prd->prd_thing, pRes, pType);
	    runExec(prd->prd_property, &rightValue, pType);
	    delProp(pRes->qv_integer, rightValue.qv_integer,
		    ex->ex_kind - ex_delBoolProp + TYPE_BOOLEAN);
	    *pType = TYPE_VOID;
	    break;
	}
	case ex_getIntElement:
	case ex_getFixedElement:
	case ex_getThingElement:
	case ex_getActionElement: {
	    ElemGet_t *elg;

	    elg = ex->ex_v.ex_elemGetPtr;
	    runExec(elg->elg_list, pRes, pType);
	    runExec(elg->elg_index, &rightValue, pType);
	    switch (ex->ex_kind) {
	    case ex_getIntElement:
		pRes->qv_integer =
		    GetIntElement(pRes->qv_integer, rightValue.qv_integer);
		*pType = TYPE_INTEGER;
		break;
	    case ex_getFixedElement:
		pRes->qv_integer =
		    GetFixedElement(pRes->qv_integer, rightValue.qv_integer);
		*pType = TYPE_FIXED;
		break;
	    case ex_getThingElement:
		pRes->qv_integer =
		    GetThingElement(pRes->qv_integer, rightValue.qv_integer);
		*pType = TYPE_THING;
		break;
	    case ex_getActionElement:
		pRes->qv_integer =
		    GetActionElement(pRes->qv_integer, rightValue.qv_integer);
		*pType = TYPE_ACTION;
		break;
	    }
	    break;
	}
	case ex_putIntElement:
	case ex_putFixedElement:
	case ex_putThingElement:
	case ex_putActionElement: {
	    ElemPut_t *elp;

	    elp = ex->ex_v.ex_elemPutPtr;
	    runExec(elp->elp_list, pRes, pType);
	    runExec(elp->elp_index, &leftValue, pType);
	    runExec(elp->elp_value, &rightValue, pType);
	    switch (ex->ex_kind) {
	    case ex_putIntElement:
		PutIntElement(pRes->qv_integer, leftValue.qv_integer,
			      rightValue.qv_integer);
		break;
	    case ex_putFixedElement:
		PutFixedElement(pRes->qv_integer, leftValue.qv_integer,
				rightValue.qv_integer);
		break;
	    case ex_putThingElement:
		PutThingElement(pRes->qv_integer, leftValue.qv_integer,
				rightValue.qv_integer);
		break;
	    case ex_putActionElement:
		PutActionElement(pRes->qv_integer, leftValue.qv_integer,
				 rightValue.qv_integer);
		break;
	    }
	    *pType = TYPE_VOID;
	    break;
	}
	case ex_negate:
	case ex_fNegate:
	    runExec(ex->ex_v.ex_expressionPtr->ex_right, &rightValue, pType);
	    pRes->qv_integer = - rightValue.qv_integer;
	    break;
	case ex_bitNot:
	    runExec(ex->ex_v.ex_expressionPtr->ex_right, &rightValue, pType);
	    pRes->qv_integer = ~ rightValue.qv_integer;
	    break;
	case ex_shiftLeft:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    pRes->qv_integer =
		(ULONG_T) leftValue.qv_integer << rightValue.qv_integer;
	    break;
	case ex_shiftRight:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    pRes->qv_integer =
		(ULONG_T) leftValue.qv_integer >> rightValue.qv_integer;
	    break;
	case ex_bitAnd:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    pRes->qv_integer = leftValue.qv_integer & rightValue.qv_integer;
	    break;
	case ex_bitXor:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    pRes->qv_integer = leftValue.qv_integer ^ rightValue.qv_integer;
	    break;
	case ex_bitOr:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    pRes->qv_integer = leftValue.qv_integer | rightValue.qv_integer;
	    break;
	case ex_times:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    pRes->qv_integer = leftValue.qv_integer * rightValue.qv_integer;
	    break;
	case ex_divide:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    if (rightValue.qv_integer == 0) {
		runError("run-time divide by zero");
		pRes->qv_integer = 1;
	    } else {
		pRes->qv_integer =
		    leftValue.qv_integer / rightValue.qv_integer;
	    }
	    break;
	case ex_remainder:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    if (rightValue.qv_integer == 0) {
		runError("run-time remainder by zero");
		pRes->qv_integer = 1;
	    } else {
		pRes->qv_integer =
		    leftValue.qv_integer % rightValue.qv_integer;
	    }
	    break;
	case ex_plus:
	case ex_fPlus:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    pRes->qv_integer = leftValue.qv_integer + rightValue.qv_integer;
	    break;
	case ex_minus:
	case ex_fMinus:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    pRes->qv_integer = leftValue.qv_integer - rightValue.qv_integer;
	    break;
	case ex_equal:
	case ex_notEqual:
	case ex_less:
	case ex_greater:
	case ex_lessOrEqual:
	case ex_greaterOrEqual:
	case ex_fEqual:
	case ex_fNotEqual:
	case ex_fLess:
	case ex_fGreater:
	case ex_fLessOrEqual:
	case ex_fGreaterOrEqual:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    switch (ex->ex_kind) {
	    case ex_equal:
	    case ex_fEqual:
		if (leftValue.qv_integer == rightValue.qv_integer) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    case ex_notEqual:
	    case ex_fNotEqual:
		if (leftValue.qv_integer != rightValue.qv_integer) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    case ex_less:
	    case ex_fLess:
		if (leftValue.qv_integer < rightValue.qv_integer) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    case ex_greater:
	    case ex_fGreater:
		if (leftValue.qv_integer > rightValue.qv_integer) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    case ex_lessOrEqual:
	    case ex_fLessOrEqual:
		if (leftValue.qv_integer <= rightValue.qv_integer) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    case ex_greaterOrEqual:
	    case ex_fGreaterOrEqual:
		if (leftValue.qv_integer >= rightValue.qv_integer) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    }
	    *pType = TYPE_BOOLEAN;
	    break;
	case ex_fTimes:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    pRes->qv_integer =
		mulFixed(leftValue.qv_integer, rightValue.qv_integer);
	    break;
	case ex_fDivide:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    if (rightValue.qv_integer == 0) {
		runError("run-time divide by zero");
		pRes->qv_integer = 1 << FRAC_BITS;
	    } else {
		pRes->qv_integer =
		    divFixed(leftValue.qv_integer, rightValue.qv_integer);
	    }
	    break;
	case ex_strPlus:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    runExec(exp->ex_right, &rightValue, pType);
	    pRes->qv_string =
		concatenate(leftValue.qv_string, rightValue.qv_string);
	    break;
	case ex_strEqual:
	case ex_strNotEqual:
	case ex_strLess:
	case ex_strGreater:
	case ex_strLessOrEqual:
	case ex_strGreaterOrEqual: {
	    BOOL_T leftCopied, rightCopied;

	    exp = ex->ex_v.ex_expressionPtr;
	    leftCopied = fetchString(exp->ex_left, &leftValue);
	    rightCopied = fetchString(exp->ex_right, &rightValue);
	    res = strCmp(leftValue.qv_string, rightValue.qv_string);
	    switch (ex->ex_kind) {
	    case ex_strEqual:
		if (res == 0) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    case ex_strNotEqual:
		if (strCmp(leftValue.qv_string,rightValue.qv_string) != 0) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    case ex_strLess:
		if (strCmp(leftValue.qv_string, rightValue.qv_string) == -1) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    case ex_strGreater:
		if (strCmp(leftValue.qv_string, rightValue.qv_string) == +1) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    case ex_strLessOrEqual:
		if (strCmp(leftValue.qv_string,rightValue.qv_string) != +1) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    case ex_strGreaterOrEqual:
		if (strCmp(leftValue.qv_string, rightValue.qv_string) != -1) {
		    pRes->qv_integer = BOOLEAN_TRUE;
		} else {
		    pRes->qv_integer = BOOLEAN_FALSE;
		}
		break;
	    }
	    if (leftCopied) {
		freeString(leftValue.qv_string);
	    }
	    if (rightCopied) {
		freeString(rightValue.qv_string);
	    }
	    *pType = TYPE_BOOLEAN;
	    break;
	}
	case ex_strEqualEqual: {
	    BOOL_T leftCopied, rightCopied;

	    exp = ex->ex_v.ex_expressionPtr;
	    leftCopied = fetchString(exp->ex_left, &leftValue);
	    rightCopied = fetchString(exp->ex_right, &rightValue);
	    if (strCmpLC(leftValue.qv_string, rightValue.qv_string) == 0) {
		pRes->qv_integer = BOOLEAN_TRUE;
	    } else {
		pRes->qv_integer = BOOLEAN_FALSE;
	    }
	    if (leftCopied) {
		freeString(leftValue.qv_string);
	    }
	    if (rightCopied) {
		freeString(rightValue.qv_string);
	    }
	    *pType = TYPE_BOOLEAN;
	    break;
	}
	case ex_logNot:
	    runExec(ex->ex_v.ex_expressionPtr->ex_right, &rightValue, pType);
	    pRes->qv_integer = (rightValue.qv_integer == 0) ? 1 : 0;
	    break;
	case ex_logAnd:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    if (leftValue.qv_integer == BOOLEAN_FALSE) {
		pRes->qv_integer = BOOLEAN_FALSE;
	    } else {
		runExec(exp->ex_right, pRes, pType);
	    }
	    break;
	case ex_logOr:
	    exp = ex->ex_v.ex_expressionPtr;
	    runExec(exp->ex_left, &leftValue, pType);
	    if (leftValue.qv_integer != BOOLEAN_FALSE) {
		pRes->qv_integer = BOOLEAN_TRUE;
	    } else {
		runExec(exp->ex_right, pRes, pType);
	    }
	    break;
	case ex_assignment: {
	    Assignment_t *as;
	    LocalList_t *ll;

	    as = ex->ex_v.ex_assignmentPtr;
	    runExec(as->as_source, &rightValue, pType);
	    if (as->as_destination->ex_kind == ex_local) {
		ll = as->as_destination->ex_v.ex_localPtr;
		pRes = RunLocals + ll->ll_offset / sizeof(QValue_t);
	    } else {
		runError("illegal destination in assignment at run-time!");
	    }
	    if (! RunExiting) {
		if (*pType == TYPE_STRING) {
		    freeString(pRes->qv_string);
		    pRes->qv_string = rightValue.qv_string;
		} else {
		    pRes->qv_integer = rightValue.qv_integer;
		}
	    }
	    *pType = TYPE_VOID;
	    break;
	}
	case ex_sequence: {
	    StatementSequence_t *sts;

	    sts = ex->ex_v.ex_sequencePtr;
	    while (sts != P_NULL && ! RunExiting) {
		runExec(sts->sts_this, pRes, pType);
		sts = sts->sts_next;
	    }
	    break;
	}
	default:
	    runError("run-time unknown ExecKind_t");
	    *pType = TYPE_ERROR;
	}
    }
}

void
runExpr(Exec_t *ex, QValue_t *pRes, UINT_T *pType)
{

#if DEBUG_PRESENT
    if (WantDebug) {
	mess("runExpr\n");
    }
#endif
    RunCallDepth = 0;
    RunExiting = B_FALSE;
    RunAborting = B_FALSE;
    RunTraceBack = P_NULL;
    BcAbort = B_FALSE;
    startTimeout();
    runExec(ex, pRes, pType);
    stopTimeout();
    if (RunExiting) {
	*pType = TYPE_ERROR;
    }
#if DEBUG_PRESENT
    if (WantDebug) {
	mess("runExpr done\n");
    }
#endif
}
