#include <sys/stat.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * fileIO.c - database cache and I/O handling.
 */

#define DO_CHECK_KEYS		B_TRUE

#define CACHE_KEEP		5	/* number head slots to NEVER reuse */
#define DISK_BLOCK_SIZE 	512	/* disk block size ** POWER OF 2 ** */
#define DISK_BUFFER_SIZE	(512*8) /* size of single disk buffer */
#define FREE_TABLE_SIZE 	64	/* size of table of free file slots */

#define INDEX_MIN		1000	/* minimum in-mem index table size */
#define FILE_LENGTH_SLOP	32	/* trim off excess more than this */
#define FILE_LENGTH_MIN 	8	/* minimum size of file slot */
#define SORT_SHIFT		7
#define SORT_BUCKETS		(1<<SORT_SHIFT)/* number buckets for ioFlush */
#define FREE_TABLE_INTERVAL	100	/* # frees before merge and retable */

#define FLAGS_OFFSET		(6*4)	/* offset DungeonFlags in index file */

/* Convention: if slot is free, then ca_usedLength == CA_FREE and
   ca_slotLength does not have CA_DIRTY set. */
#define CA_DIRTY 	0x80000000	/* set in ca_slotLength */
#define CA_FREE		0x80000000	/* set in ca_usedLength */
#define IX_FREE		0x80000000	/* set in ix_offset if slot is free */

typedef struct {
    UINT_T ix_slotLength;	/* full length of slot */
    UINT_T ix_usedLength;	/* space used by this entry */
    ULONG_T ix_offset;		/* byte offset in file */
} Index_t;

typedef struct Cache {
    struct Cache *ca_next;	/* next in LRU or free list */
    struct Cache *ca_prev;	/* prev in LRU list */
    ULONG_T ca_key;		/* key of item stored in this cache slot */
    ULONG_T ca_slotLength;	/* total length of this slot */
    ULONG_T ca_usedLength;	/* length used by the entry in this slot */
    /* Note: ca_slotLength includes the sizeof(Cache_t), but
       ca_usedLength does not. Also, all slotLength values are rounded
       up to the next 4 bytes. */
} Cache_t;

static int DataFd;			/* fd for use with main data file */

static ULONG_T CacheSize;		/* size of contiguous memory region */
static BYTE_T *CachePtr;		/* ptr to contiguous memory region */
static Cache_t *CacheLRUHead;		/* most recently used cache entry */
static Cache_t *CacheLRUTail;		/* least recently used cache entry */
static Cache_t *CacheFreeHead;		/* head of free list in cache */
static Cache_t *LastRefCa;		/* along with LastRefKey, LastRefPtr */

static ULONG_T IndexSize;		/* allocated size of main index */
static ULONG_T IndexCurrent;		/* entries used in the index */
static Index_t *IndexPtr;		/* ptr to actual file index */
static Cache_t **IndexCa;		/* ix pointers to Cache_t's */
static ULONG_T DataSize;		/* current size of data file */
static ULONG_T WrittenSize;	/* amount actually written to the file */
static ULONG_T CurrentPosition; /* current seek position in data file */
static ULONG_T IndexSlotFreeHead;	/* head of unused index slots */
static ULONG_T FreeTable[FREE_TABLE_SIZE];	/* table of free file slots */
static UINT_T FreeTableNext;	/* count of used entries in FreeTable */
static BOOL_T FreedSinceTabled; 	/* table could be out of date */
static BYTE_T FreedSinceTabledCount;	/* how long since we did it */

static BYTE_T DiskBuffer[DISK_BUFFER_SIZE];	/* the single disk buffer */
static int DiskBufferMax;	/* number of bytes currently in disk buffer */
static ULONG_T DiskBufferOffset;	/* offset in file of start of buffer */
static BOOL_T DiskBufferDirty;	/* data in buffer needs to be written */

static ULONG_T PublicSymbolTableKey;	/* key rep'ing public symbol table */
static ULONG_T EntryAction;		/* done on player startup */
static ULONG_T CreationPassword;	/* string for the creation password */
static ULONG_T DungeonFlags;		/* flag bits for the whole dungeon */
static ULONG_T MachineList;		/* list of machines in dungeon */
static ULONG_T AbortTimeout;		/* timeout for interpretation */

static ULONG_T LastKey; 		/* used when aborting */

static ULONG_T IndexByteSize;		/* save recalc */

static ULONG_T
    CreateCount,
    DeleteCount,
    ReadCount,
    ReplaceCount,
    ShrinkCount,
    ExpandCount,
    DirtyCount,

    FReadCount,
    FWriteCount,
    FSeekCount;

static int LogFd;
static char LastLogChar;		/* check for new lines */
static BOOL_T Aborting;

static BOOL_T Flushing; 		/* so we don't do it recursively */

/*
 * log - write a simple string to the log file.
 */

void
log(char *message)
{
#define DATE_LEN	19
    char dateBuff[DATE_LEN + 1];
    char *p;

    if (message != P_NULL) {
	if (LogFd == -1) {
	    LogFd = open(LOG_NAME, O_WRONLY | O_APPEND);
	    if (LogFd == -1) {
		LogFd = open(LOG_NAME, O_WRONLY | O_CREAT, 0600);
	    } else {
		(void) lseek(LogFd, 0, SEEK_END);
	    }
	}
	if (LogFd != -1) {
	    if (LastLogChar == '\n') {
		ConvTimeShort(GetCurrentTime(), &dateBuff[0]);
		dateBuff[DATE_LEN] = ' ';
		(void) write(LogFd, &dateBuff[0], DATE_LEN + 1);
	    }
	    p = message;
	    while (*p != '\0') {
		++p;
	    }
	    (void) write(LogFd, message, p - message);
	    close(LogFd);
	    LogFd = -1;
	    LastLogChar = *--p;
	}
    }
#undef DATE_LEN
}

/*
 * logX - write a 32 bit hex value and two strings as a log message.
 */

void
logX(char *m1, ULONG_T n, char *m2)
{
    char *p;
    char buffer[300];
    char hexBuff[9];
    int i;
    unsigned char b;

    if (m1 == P_NULL) {
	p = &buffer[0];
    } else {
	p = strTail(&buffer[0], m1);
    }
    m1 = &hexBuff[8];
    *m1 = '\0';
    for (i = 0; i < 8; i += 1) {
	--m1;
	b = n & 0xf;
	n >>= 4;
	if (b >= 10) {
	    *m1 = b - 10 + 'a';
	} else {
	    *m1 = b + '0';
	}
    }
    p = strTail(p, m1);
    if (m2 != P_NULL) {
	p = strTail(p, m2);
    }
    *p = '\0';
    log(&buffer[0]);
}

/*
 * logXNL - stub to supply a 3rd argument of newline.
 */

void
logXNL(char *m1, ULONG_T n)
{

    logX(m1, n, "\n");
}

/*
 * logN - write a 32 bit decimal value and two strings as a log message.
 */

void
logN(char *m1, ULONG_T n, char *m2)
{
    char *p;
    char buffer[300];
    char decBuff[11];

    if (m1 == P_NULL) {
	p = &buffer[0];
    } else {
	p = strTail(&buffer[0], m1);
    }
    m1 = &decBuff[10];
    *m1 = '\0';
    do {
	*--m1 = n % 10 + '0';
	n /= 10;
    } while (n != 0);
    p = strTail(p, m1);
    if (m2 != P_NULL) {
	p = strTail(p, m2);
    }
    *p = '\0';
    log(&buffer[0]);
}

/*
 * log2NL - log a two-part message, then add a newline.
 */

void
log2NL(char *s1, char *s2)
{
    char buffer[300];

    *strTail(strTail(strTail(&buffer[0], s1), s2), "\n") = '\0';
    log(&buffer[0]);
}

/*
 * localAbortTail - end part of localAbort and badKey.
 */

static void
localAbortTail(char *where)
{

    putString(where);
    newLine();
    myAbortNoWrite();
}

/*
 * localAbort - small utility to abort here.
 */

static void
localAbort(void)
{

    if (! Aborting) {
	Aborting = B_TRUE;
	if (DataFd != -1) {
	    close(DataFd);
	    DataFd = -1;
	}
	setLogging();
	putString("last key = ");
	putHex(LastKey);
	newLine();
	localAbortTail("*** FileIO abort ***");
    }
}

/*
 * ioShutDown - shut down processing here. Log one final message.
 *	Note: in abort situations, this routine is called twice, so
 *	it has to protect itself.
 */

void
ioShutDown(char *message)
{

    if (DataFd != -1) {
	close(DataFd);
	DataFd = -1;
    }
    if (IndexCa != P_NULL) {
	free(IndexCa);
	IndexCa = P_NULL;
    }
    if (IndexPtr != P_NULL) {
	free(IndexPtr);
	IndexPtr = P_NULL;
    }
    if (CachePtr != P_NULL) {
	free(CachePtr);
	CachePtr = P_NULL;
    }
    log(message);
    if (LogFd != -1) {
	close(LogFd);
	LogFd = -1;
    }
}

/*
 * getLong - attempt to read a ULONG_T from the Index file.
 */

static BOOL_T
getLong(ULONG_T *lp, char *what)
{
    char line[80];

    if (read(DataFd, lp, sizeof(ULONG_T)) != sizeof(ULONG_T)) {
	*strTail(strTail(strTail(&line[0], "can't read "), what),
		 " from index file\n") = '\0';
	ioShutDown(&line[0]);
	return B_FALSE;
    } else {
	*lp = ntohl(*lp);
	return B_TRUE;
    }
}

/*
 * mergeFreeFileSlots - merge free file slots.
 *	Note that the free table (FreeTableNext, etc.) is not valid after
 *	running this routine. 'buildFreeTable' should be called to set it up.
 *	This also can invalidate the free index slot list, so
 *	'buildFreeIndexList' also needs to be called.
 */

static void
mergeFreeFileSlots(void)
{
    Index_t *ix, *ix2;
    ULONG_T offset, oldSize, oldCount;
    UINT_T len;
    BOOL_T found;

    TRACE("mergeFreeFileSlots", IndexCurrent);
    oldSize = DataSize;
    oldCount = IndexCurrent;
    ix = IndexPtr;
    offset = oldCount;
    offset -= 1;
    ix += offset;
    found = B_TRUE;
    while (found &&
	   ((ix->ix_offset & IX_FREE) != 0 || ix->ix_slotLength == 0))
    {
	if (ix->ix_slotLength == 0) {
	    /* unused slot - just delete it */
	    IndexCurrent -= 1;
	    --ix;
	    TRACE("mergeFreeFileSlots 1 IC", IndexCurrent);
	} else {
	    found = B_FALSE;
	    /* the last slot is free - can we just erase it or merge it? */
	    offset = ix->ix_offset & ~IX_FREE;
	    len = ix->ix_slotLength;
	    if (offset == DataSize - len) {
		/* just delete it */
		DataSize -= len;
		IndexCurrent -= 1;
		TRACE("mergeFreeFileSlots 2 IC", IndexCurrent);
		--ix;
		found = B_TRUE;
	    } else {
		/* Search for free entry that is adjacent to it in the file.
		   Note that unused slots do not have IX_FREE set, so we will
		   not try to merge with them. */
		ix2 = ix;
		--ix2;
		while (! found && ix2 != IndexPtr) {
		    if (ix2->ix_offset ==
			((offset - ix2->ix_slotLength) | IX_FREE))
		    {
			/* this one is just before the last one - merge */
			ix2->ix_slotLength += len;
			IndexCurrent -= 1;
			TRACE("mergeFreeFileSlots 3 IC", IndexCurrent);
			--ix;
			found = B_TRUE;
		    } else if (ix2->ix_offset == ((offset + len) | IX_FREE)) {
			/* this one is just after the last one - merge */
			ix2->ix_offset = offset | IX_FREE;
			ix2->ix_slotLength += len;
			IndexCurrent -= 1;
			TRACE("mergeFreeFileSlots 4 IC", IndexCurrent);
			--ix;
			found = B_TRUE;
		    }
		    /* do this is all cases */
		    --ix2;
		}
	    }
	}
    }
    if (DataSize != oldSize || IndexCurrent != oldCount) {
	logX("DataSize ", oldSize, "");
	logX("=>", DataSize, "");
	logX(" IndexCurrent ", oldCount, "");
	logX("=>", IndexCurrent, "\n");
    }
}

/*
 * buildFreeTable - set up FreeTable with free files slots.
 */

static void
buildFreeTable(void)
{
    Index_t *ix;
    ULONG_T *ft;
    ULONG_T indexCount, index;
    UINT_T pos;

    TRACE("buildFreeTable", 0);

    indexCount = IndexCurrent;
    index = 1;
    pos = 0;
    ix = IndexPtr;
    ++ix;
    ft = &FreeTable[0];
    while (pos != FREE_TABLE_SIZE && index != indexCount) {
	if ((ix->ix_offset & IX_FREE) != 0) {
	    /* Note that this will not catch slots that are on the list of
	       free slot headers, since their offset fields do not have
	       IX_FREE set. */
	    *ft++ = index;
	    pos += 1;
	}
	index += 1;
	++ix;
    }
    if (index == indexCount) {
	FreedSinceTabled = B_FALSE;
    } else {
	FreedSinceTabled = B_TRUE;
    }
    FreeTableNext = pos;
    FreedSinceTabledCount = 0;
}

/*
 * buildFreeIndexList - build the list of index slots that are unused (have
 *	slotLength == 0).
 */

static void
buildFreeIndexList(void)
{
    Index_t *ix;
    ULONG_T slot, prev;

    TRACE("buildFreeIndexList", 0);
    prev = 0;
    ix = IndexPtr;
    for (slot = 1; slot < IndexCurrent; slot += 1) {
	++ix;
	if (ix->ix_slotLength == 0) {
	    ix->ix_offset = prev;
	    prev = slot;
	}
    }
    IndexSlotFreeHead = prev;
}

/*
 * buildFreeStuff - call all of the routines to build free lists, etc.
 */

static void
buildFreeStuff(void)
{

    mergeFreeFileSlots();
    buildFreeTable();
    buildFreeIndexList();
}

/*
 * ioInitialize - initialize the I/O structures. Return 'B_FALSE' if fail.
 */

BOOL_T
ioInitialize(char *path, ULONG_T cacheSize)
{
    Cache_t *ca;
    ULONG_T byteSize, version;
    int i;
    struct stat statBuf;

    TRACE("ioInitialize", cacheSize);
    DataFd = -1;
    LogFd = -1;
    LastLogChar = '\n'; 		/* want date on first line */
    IndexPtr = P_NULL;
    IndexCa = P_NULL;
    Flushing = B_FALSE;
    Aborting = B_FALSE;
    LastRefKey = 0;
    LastRefPtr = NULL;
    LastRefCa = NULL;

    CacheSize = cacheSize;
    CachePtr = malloc(cacheSize);
    if (CachePtr == P_NULL) {
	log("can't allocate memory for cache\n");
	if (LogFd != -1) {
	    close(LogFd);
	    LogFd = -1;
	}
	return B_FALSE;
    }

    DataFd = open(INDEX_NAME, O_RDONLY);
    if (DataFd == -1) {
	if (Verbose) {
	    mess(
"MUDServ could not open the MUD.index file. This means that there is no\n"
"AmigaMUD database in the directory it was looking in. This is either the\n"
"current directory when MUDServ is started, or the directory specified on\n"
"the MUDServ command line.\n\n"
	    );
	}
	ioShutDown("can't open index file on given path\n");
	return B_FALSE;
    }
    if (! getLong(&version, "database version") ||
	! getLong(&IndexCurrent, "index size") ||
	! getLong(&DataSize, "data size") ||
	! getLong(&PublicSymbolTableKey, "public symbol table key") ||
	! getLong(&EntryAction, "entry action") ||
	! getLong(&CreationPassword, "creation password") ||
	! getLong(&DungeonFlags, "dungeon flags") ||
	! getLong(&MachineList, "machine list") ||
	! getLong(&AbortTimeout, "abort timeout"))
    {
	/* ioShutDown already done. */
	return B_FALSE;
    }
    if (version > (VERSION1 - '0') * 10 + (VERSION2 - '0')) {
	ioShutDown("database has newer version than server\n");
	return B_FALSE;
    }
    if (version < 10 && (VERSION1 - '0') * 10 + (VERSION2 - '0') >= 10) {
	ioShutDown("database has pre-V1.0 format\n");
	return B_FALSE;
    }
    if ((DungeonFlags & FLAG_DATABASE_STABLE) == 0) {
	ioShutDown("database was not properly saved\n");
	return B_FALSE;
    }
    WrittenSize = DataSize;
    if (IndexCurrent < INDEX_MIN) {
	IndexSize = INDEX_MIN;
    } else {
	IndexSize = IndexCurrent + INDEX_MIN;
    }
    IndexByteSize = IndexSize * sizeof(Index_t);
    IndexPtr = malloc(IndexByteSize);
    if (IndexPtr == P_NULL) {
	ioShutDown("can't allocate memory for main index\n");
	return B_FALSE;
    }
    IndexCa = malloc(IndexSize * sizeof(Cache_t *));
    if (IndexCa == P_NULL) {
	ioShutDown("can't allocate memory for index cache pointers\n");
	return B_FALSE;
    }
    byteSize = IndexCurrent * sizeof(Index_t);
    if (read(DataFd, IndexPtr, byteSize) != byteSize) {
	ioShutDown("can't read main index\n");
	return B_FALSE;
    }
    close(DataFd);
    DataFd = -1;

    /* It is possible for DataSize to be greater than the actual size of
       the data file if the last created entry was deleted before the
       database cache was flushed. */
    if (stat(DATA_NAME, &statBuf) != 0) {
	ioShutDown("can't stat data file to get current size\n");
	return B_FALSE;
    }
    if (WrittenSize > statBuf.st_size) {
	WrittenSize = statBuf.st_size;
    }

    DataFd = open(DATA_NAME, O_RDWR);
    if (DataFd == -1) {
	ioShutDown("can't open data file on given path\n");
	return B_FALSE;
    }
    CurrentPosition = 0;

    ca = (Cache_t *) CachePtr;
    ca->ca_next = P_NULL;
    ca->ca_prev = P_NULL;
    ca->ca_slotLength = CacheSize;
    ca->ca_usedLength = CA_FREE;
    CacheFreeHead = ca;
    CacheLRUHead = P_NULL;
    CacheLRUTail = P_NULL;

    buildFreeStuff();
    for (i = 0; i != IndexSize; i += 1) {
	IndexCa[i] = P_NULL;
    }

    DiskBufferMax = 0;
    DiskBufferOffset = 0;
    DiskBufferDirty = B_FALSE;

    CreateCount = 0;
    DeleteCount = 0;
    ReadCount = 0;
    ReplaceCount = 0;
    ShrinkCount = 0;
    ExpandCount = 0;
    DirtyCount = 0;

    FReadCount = 0;
    FWriteCount = 0;
    FSeekCount = 0;

    return B_TRUE;
}

#if DO_CHECK_KEYS

/*
 * badKey - use of an invalid key.
 */

static void
badKey(ULONG_T key, char *where)
{

    setLogging();
    putString("*** bad key ");
    putHex(key);
    if ((key & INDEX_MASK) >= IndexCurrent) {
	putString(" (beyond max ");
	putHex(IndexCurrent);
    } else {
	putString(" (already freed");
    }
    putString(") on ");
    localAbortTail(where);
}

#endif

#if MUD_ENDIAN == MUD_LITTLE_ENDIAN

/* When we read things from disk into the cache, we need to convert
   the endianness of all multi-byte numeric fields. This is because
   the database is in network byte order, but manipulations of the
   structures stored in it are done in host byte order. */

static void
swapBytes(void *p, ULONG_T key, BOOL_T toHost)
{
    switch (key >> TYPE_SHIFT) {
    case TYPE_PLAYER: {
	Player_t *pl;

	pl = (Player_t *) p;
	pl->pl_key = ntohl(pl->pl_key);
	pl->pl_symbolTableKey = ntohl(pl->pl_symbolTableKey);
	pl->pl_sponsor = ntohl(pl->pl_sponsor);
	pl->pl_location = ntohl(pl->pl_location);
	pl->pl_inputAction = ntohl(pl->pl_inputAction);
	pl->pl_rawKeyAction = ntohl(pl->pl_rawKeyAction);
	pl->pl_mouseDownAction = ntohl(pl->pl_mouseDownAction);
	pl->pl_buttonAction = ntohl(pl->pl_buttonAction);
	pl->pl_promptKey = ntohl(pl->pl_promptKey);
	pl->pl_lastOn= ntohl(pl->pl_lastOn);
	pl->pl_idleAction = ntohl(pl->pl_idleAction);
	pl->pl_activeAction = ntohl(pl->pl_activeAction);
	pl->pl_effectDoneAction = ntohl(pl->pl_effectDoneAction);
	pl->pl_textWidth = ntohs(pl->pl_textWidth);
	pl->pl_textHeight = ntohs(pl->pl_textHeight);
	pl->pl_flags = ntohs(pl->pl_textHeight);
	break;
    }
    case TYPE_INTEGER:
    case TYPE_FIXED:
	*(ULONG_T *) p = ntohl(*(ULONG_T *) p);
	break;
    case TYPE_THING: {
	Thing_t *th;
	Attr_t *at;
	BYTE_T i;

	th = (Thing_t *) p;
	th->th_parent = ntohl(th->th_parent);
	th->th_owner = ntohl(th->th_owner);
	th->th_useCount = ntohs(th->th_useCount);
	at = (Attr_t *) (th + 1);
	for (i = 0; i < th->th_propCount; i += 1) {
	    at->at_prop = ntohl(at->at_prop);
	    at->at_value = ntohl(at->at_value);
	    ++at;
	}
	break;
    }
    case TYPE_ACTION:	/* handled by the byteIO code */
    case TYPE_STRING:	/* nothing to do - no unicode */
    case TYPE_TABLE:	/* handled by byteIO and symRead/symWrite */
    case TYPE_GRAMMAR:	/* ditto */
	break;
    case TYPE_LIST_INTEGER:
    case TYPE_LIST_FIXED:
    case TYPE_LIST_THING:
    case TYPE_LIST_ACTION: {
	ItemList_t *l;
	ULONG_T *ulp;
	UINT_T count, i;

	l = (ItemList_t *) p;
	l->l_owner = ntohl(l->l_owner);
	l->l_useCount = ntohs(l->l_useCount);
	count = 0;	/* keep silly gcc happy */
	if (! toHost) {
	    count = l->l_elementCount;
	}
	l->l_elementCount = ntohs(l->l_elementCount);
	if (toHost) {
	    count = l->l_elementCount;
	}
	ulp = (ULONG_T *) (l + 1);
	for (i = 0; i < count; i += 1) {
	    *ulp = ntohl(*ulp);
	    ++ulp;
	}
	break;
    }
    case TYPE_PROP_BOOLEAN:
    case TYPE_PROP_INTEGER:
    case TYPE_PROP_FIXED:
    case TYPE_PROP_THING:
    case TYPE_PROP_ACTION:
    case TYPE_PROP_STRING:
    case TYPE_PROP_TABLE:
    case TYPE_PROP_GRAMMAR:
    case TYPE_PROP_LIST_INTEGER:
    case TYPE_PROP_LIST_FIXED:
    case TYPE_PROP_LIST_THING:
    case TYPE_PROP_LIST_ACTION: {
	Property_t *prp;

	prp = (Property_t *) p;
	prp->prp_owner = ntohl(prp->prp_owner);
	prp->prp_useCount = ntohs(prp->prp_useCount);
	break;
    }
    case TYPE_MACHINE: {
	Machine_t *mch;

	mch = (Machine_t *) p;
	mch->mch_owner = ntohl(mch->mch_owner);
	mch->mch_key = ntohl(mch->mch_key);
	mch->mch_location = ntohl(mch->mch_location);
	mch->mch_sayProc = ntohl(mch->mch_sayProc);
	mch->mch_whisperMeProc = ntohl(mch->mch_whisperMeProc);
	mch->mch_whisperOtherProc = ntohl(mch->mch_whisperOtherProc);
	mch->mch_pageProc = ntohl(mch->mch_pageProc);
	mch->mch_poseProc = ntohl(mch->mch_poseProc);
	mch->mch_otherProc = ntohl(mch->mch_otherProc);
	mch->mch_idleProc = ntohl(mch->mch_idleProc);
	mch->mch_activeProc = ntohl(mch->mch_activeProc);
	mch->mch_useCount = ntohs(mch->mch_useCount);
    }
    case W_VERBTAIL: {
	ULONG_T *ulp;

	ulp = (ULONG_T *) p;
	*ulp = ntohl(*ulp);
	++ulp;
	*ulp = ntohl(*ulp);
	break;
    }
    case W_VERB: {
	ULONG_T *ulp;
	ULONG_T altCount, i;

	ulp = (ULONG_T *) p;
	*ulp = ntohl(*ulp);
	++ulp;
	altCount = 0;	/* keep silly gcc happy */
	if (! toHost) {
	    altCount = *ulp;
	}
	*ulp = ntohl(*ulp);
	if (toHost) {
	    altCount = *ulp;
	}
	++ulp;
	for (i = 0; i < altCount; i += 1) {
	    *ulp = ntohl(*ulp);
	    ++ulp;
	    *ulp = ntohl(*ulp);
	    ++ulp;
	}
	break;
    }
    default:
	setLogging();
	putString("Bad key ");
	putHex(key);
	localAbortTail(" to swapBytes");
    }
}

#endif

/*
 * myWrite - allow for retries.
 */

static LONG_T
myWrite(int fd, void *buffer, ULONG_T len, BOOL_T mark)
{
    ULONG_T valu;
    int gotLen, indexFd;
    char errBuff[200];

    if (mark && (DungeonFlags & FLAG_DATABASE_STABLE) != 0) {
	/* mark database as now non-stable */
	DungeonFlags &= ~ FLAG_DATABASE_STABLE;
	indexFd = open(INDEX_NAME, O_WRONLY);
	if (indexFd == -1) {
	    log("can't open index file\n");
	    localAbort();
	}
	FSeekCount += 1;
	if (lseek(indexFd, FLAGS_OFFSET, SEEK_SET) == -1) {
	    log("Fail to seek to flags\n");
	    close(indexFd);
	    localAbort();
	}
	FWriteCount += 1;
	valu = htonl(DungeonFlags);
	if (write(indexFd, &valu, sizeof(ULONG_T)) != sizeof(ULONG_T)) {
	    log("Fail to write flags\n");
	    close(indexFd);
	    localAbort();
	}
	close(indexFd);
    }

    while (B_TRUE) {
	FWriteCount += 1;
	gotLen = write(fd, buffer, len);
	if (gotLen == len) {
	    break;
	}
	if (gotLen != -1) {
	    errno = ENOSPC;
	}
	strcpy(&errBuff[0], "MUDServ: write failed - ");
	if (errno > sys_nerr) {
	    strcat(&errBuff[0], "unknown error");
	} else {
	    strcat(&errBuff[0], sys_errlist[errno]);
	}

	if (! checkForRetry(&errBuff[0])) {
	    gotLen = -1;
	    break;
	}
	if (gotLen != -1) {
	    buffer = (BYTE_T *) buffer + gotLen;
	    len -= gotLen;
	}
    }
    return gotLen;
}

/*
 * ioFlushBuffer - flush the dirty disk buffer to the data file.
 */

static void
ioFlushBuffer(void)
{
    ULONG_T offset, max, count;

    TRACE("ioFlushBuffer", 0);
    offset = DiskBufferOffset;
    max = DiskBufferMax;
    if (offset > WrittenSize) {
	/* The file offset of the stuff at the beginning of the disk buffer
	   is beyond the length of the file. Have to write some stuff out
	   there to expand the file. Write either the amount of the gap, or
	   a bufferfull, whichever is greater. */
	/* Well, I thought that this might be THE bug, but it did not turn out
	   to be so. Sigh. I believe that 'WrittenSize', not 'DataSize', is
	   the correct thing to compare against, however. */
	if (CurrentPosition != WrittenSize) {
	    FSeekCount += 1;
	    if (lseek(DataFd, 0, SEEK_END) == -1) {
		log("Fail to seek to end to expand\n");
		localAbort();
	    }
	}
	count = offset - WrittenSize;
	/* count is the gap size */
	if (count > DISK_BUFFER_SIZE) {
	    /* Need to write more than a bufferful - use the complete cache
	       buffer as a handy pointer to some junk to write. I believe it
	       is not possible to need more than the cache size in a single
	       write here, since we could not store that much in the cache! */
	    if (myWrite(DataFd, CachePtr, count, B_TRUE) != count) {
		logN("Fail to write ", count, " bytes to expand\n");
		localAbort();
	    }
	    WrittenSize = offset;
	} else {
	    if (myWrite(DataFd, &DiskBuffer[0], DISK_BUFFER_SIZE, B_TRUE) !=
		DISK_BUFFER_SIZE)
	    {
		log("Fail to write block to expand\n");
		localAbort();
	    }
	    WrittenSize += DISK_BUFFER_SIZE;
	}
	CurrentPosition = WrittenSize;
    }
    if (CurrentPosition != offset) {
	FSeekCount += 1;
	if (lseek(DataFd, offset, SEEK_SET) == -1) {
	    logX("Fail to seek to ", offset, " to flush buffer\n");
	    localAbort();
	}
    }
    if (myWrite(DataFd, &DiskBuffer[0], max, B_TRUE) != max) {
	logN("Fail to write ", max, " bytes to flush buffer\n");
	localAbort();
    }
    offset += max;
    CurrentPosition = offset;
    if (offset > WrittenSize) {
	WrittenSize = offset;
    }
    DiskBufferDirty = B_FALSE;
}

/*
 * ioGetBytes - read some bytes from the data file - abort if fail.
 */

static void
ioGetBytes(void *buffer, ULONG_T offset, int count)
{
    BYTE_T *p, *q;
    int buffCount, saveCount;

    TRACE("ioGetBytes", offset);
    p = buffer;
    if (offset >= DiskBufferOffset &&
	offset < DiskBufferOffset + DiskBufferMax)
    {
	/* beginning of required bytes are in the buffer */
	buffCount = offset - DiskBufferOffset;
	q = &DiskBuffer[buffCount];
	buffCount = DiskBufferMax - buffCount;
	while (buffCount != 0 && count != 0) {
	    *p++ = *q++;
	    count -= 1;
	    buffCount -= 1;
	    offset += 1;
	}
    }
    if (count > DISK_BUFFER_SIZE) {
	/* the amount we still need is bigger than a bufferful - just read it
	   directly */
	saveCount = getTimeout();
	if (CurrentPosition != offset) {
	    FSeekCount += 1;
	    if (lseek(DataFd, offset, SEEK_SET) == -1) {
		logX("Fail to seek to ", offset, " to read direct\n");
		localAbort();
	    }
	}
	FReadCount += 1;
	if (read(DataFd, p, count) != count) {
	    logN("Fail to read ", count, " bytes direct\n");
	    localAbort();
	}
	CurrentPosition = offset + count;
	restoreTimeout(saveCount);
    } else if (count != 0) {
	/* we are going to read into the disk buffer */
	saveCount = getTimeout();
	if (DiskBufferDirty) {
	    /* flush the buffer if it is dirty */
	    ioFlushBuffer();
	}
	DiskBufferOffset = offset;
	if (CurrentPosition != offset) {
	    FSeekCount += 1;
	    if (lseek(DataFd, offset, SEEK_SET) == -1) {
		logX("Fail to seek to ", offset, " to fill buffer\n");
		localAbort();
	    }
	}
	/* calculate the next block boundary past what we need */
	buffCount =
	    (offset + count + (DISK_BLOCK_SIZE - 1)) & ~(DISK_BLOCK_SIZE - 1);
	/* calculate the size of read that will get us there */
	buffCount -= offset;
	if (buffCount > DISK_BUFFER_SIZE) {
	    /* would not fit in DiskBuffer - read only what we need to read */
	    buffCount = count;
	}
	FReadCount += 1;
	DiskBufferMax = read(DataFd, &DiskBuffer[0], buffCount);
	if (DiskBufferMax < count) {
	    log("Fail to read enough into buffer\n");
	    logN("DataSize ", DataSize, P_NULL);
	    logN(" WrittenSize ", WrittenSize, "\n");
	    localAbort();
	}
	offset += DiskBufferMax;
	CurrentPosition = offset;
	q = &DiskBuffer[0];
	while (count != 0) {
	    *p++ = *q++;
	    count -= 1;
	}
	restoreTimeout(saveCount);
    }
}

/*
 * ioPutBytes - write some bytes to the data file. Abort if fail.
 */

static void
ioPutBytes(void *buffer, ULONG_T offset, ULONG_T count)
{
    BYTE_T *p, *q;
    ULONG_T buffCount, saveCount;
    BOOL_T writtenHere;

    TRACE("ioPutBytes", offset);
    p = buffer;
    writtenHere = B_FALSE;
    if (DiskBufferDirty && offset >= DiskBufferOffset &&
	offset <= DiskBufferOffset + DiskBufferMax)
    {
	/* Close-by stuff is already in the buffer. Note that we DO want the
	   second test to be '<=', since the '=' happens if the chunk we
	   want to write now is exactly after the chunk in the buffer, and it
	   is very important to do that case properly. */
	writtenHere = B_TRUE;
	buffCount = offset - DiskBufferOffset;
	q = &DiskBuffer[buffCount];
	if (buffCount + count > DISK_BUFFER_SIZE) {
	    DiskBufferMax = DISK_BUFFER_SIZE;
	} else {
	    DiskBufferMax = buffCount + count;
	}
	buffCount = DISK_BUFFER_SIZE - buffCount;
	while (buffCount != 0 && count != 0) {
	    *q++ = *p++;
	    count -= 1;
	    buffCount -= 1;
	    offset += 1;
	}
	/* no need to set DiskBufferDirty, since that is true here */
    }
    if (count > DISK_BUFFER_SIZE) {
	/* The amount we still have is bigger than a bufferful - just write it
	   directly. If the buffer contains part of this chunk, we want to
	   write it first, to maintain linear writing to disk. */
	saveCount = getTimeout();
	if (writtenHere) {
	    /* flush the buffer */
	    ioFlushBuffer();
	}
	if (offset > WrittenSize) {
	    /* have to write to expand the file */
	    if (CurrentPosition != DataSize) {
		FSeekCount += 1;
		if (lseek(DataFd, 0, SEEK_END) == -1) {
		    log("Fail to seek to end to expand\n");
		    localAbort();
		}
	    }
	    buffCount = offset - WrittenSize;
	    /* buffCount is the size of the gap to write */
	    if (buffCount > DISK_BUFFER_SIZE) {
		/* just use CachePtr as a convenient, large-enough, place to
		   write from */
		if (myWrite(DataFd, CachePtr, buffCount, B_TRUE) != buffCount){
		    logN("Fail to write ", buffCount, " bytes to expand\n");
		    localAbort();
		}
		WrittenSize = offset;
	    } else {
		if (myWrite(DataFd, &DiskBuffer[0], DISK_BUFFER_SIZE,B_TRUE) !=
		    DISK_BUFFER_SIZE)
		{
		    log("Fail to write block to expand\n");
		    localAbort();
		}
		WrittenSize += DISK_BUFFER_SIZE;
	    }
	    CurrentPosition = WrittenSize;
	}
	if (CurrentPosition != offset) {
	    FSeekCount += 1;
	    if (lseek(DataFd, offset, SEEK_SET) == -1) {
		logX("Fail to seek to ", offset, " to write direct\n");
		localAbort();
	    }
	}
	if (myWrite(DataFd, p, count, B_TRUE) != count) {
	    logN("Fail to write ", count, " bytes direct\n");
	    localAbort();
	}
	offset += count;
	CurrentPosition = offset;
	if (offset > WrittenSize) {
	    WrittenSize = offset;
	}
	restoreTimeout(saveCount);
    } else if (count != 0) {
	/* we are going to write into the disk buffer */
	if (DiskBufferDirty) {
	    /* flush the buffer if it is dirty */
	    saveCount = getTimeout();
	    ioFlushBuffer();
	    restoreTimeout(saveCount);
	}
	/* question: should the buffer be filled from disk? */
	DiskBufferDirty = B_TRUE;
	DiskBufferOffset = offset;
	DiskBufferMax = count;
	q = &DiskBuffer[0];
	while (count != 0) {
	    *q++ = *p++;
	    count -= 1;
	}
    }
}

/*
 * ioFlushBucket - part of ioFlush.
 *	NOTE: be very careful of any changes to the flushing algorithm. The
 *	code that allocates and expands entries (findFileSlot, ioExpand and
 *	ioReplace) does NOT write the new entries to the end of the file, thus
 *	trying to seek to them will fail until they have been written here.
 *	Thus, it is important that any flush be a total flush, and that the
 *	flush code expand the file if it needs to. The current code also
 *	writes the file in increasing offset, which is as desired.
 */

static void
ioFlushBucket(Cache_t *start)
{
    Cache_t *ca, *minCa;
    Index_t *ix;
    ULONG_T minOffset, offset;

    TRACE("ioFlushBucket", start);
    while (B_TRUE) {
	ca = start;
	if (ca == P_NULL) {
	    break;
	}
	minCa = P_NULL;
	minOffset = 0xffffffff;
	while (ca != P_NULL) {
	    if ((ca->ca_slotLength & CA_DIRTY) != 0) {
		ix = IndexPtr + (ca->ca_key & INDEX_MASK);
		offset = ix->ix_offset;
		if (offset < minOffset) {
		    minOffset = offset;
		    minCa = ca;
		}
	    }
	    ca = ca->ca_next;
	}
	if (minCa == P_NULL) {
	    start = P_NULL;
	} else {
	    minCa->ca_slotLength &= ~CA_DIRTY;
	    ix = IndexPtr + (minCa->ca_key & INDEX_MASK);
	    /* Note: we really only need to write ix_usedLength bytes, but
	       if we write ix_slotLength, then we are MUCH more likely to
	       have the file position pointer at the right place for the
	       next write when doing 'ioFlush', and that will get rid of
	       a lot of Seek calls. */
#if MUD_ENDIAN == MUD_LITTLE_ENDIAN
	    swapBytes(minCa + 1, minCa->ca_key, B_FALSE);
#endif
	    ioPutBytes(minCa + 1, minOffset, ix->ix_slotLength);
#if MUD_ENDIAN == MUD_LITTLE_ENDIAN
	    swapBytes(minCa + 1, minCa->ca_key, B_TRUE);
#endif
	}
    }
}

/*
 * saveLong - local utility to write a LONG_T to the index file.
 */

static void
saveLong(int fd, ULONG_T valu, char *what)
{
    char message[100];

    valu = htonl(valu);
    if (myWrite(fd, &valu, sizeof(ULONG_T), B_FALSE) != sizeof(ULONG_T)) {
	close(fd);
	*strTail(strTail(strTail(&message[0], "can't write "), what),
		 " to index file\n") = '\0';
	log(&message[0]);
	localAbort();
    }
}

/*
 * ioFlush0 - flush all I/O structures to disk. Not a destructive flush.
 */

static void
ioFlush0(BOOL_T doIndex)
{
    Cache_t *(radixTable[SORT_BUCKETS]);
    int fd;
    Cache_t *ca, *next;
    Index_t *ix;
    ULONG_T offset, saveCount;
    UINT_T i, shift;

    if (Verbose) {
	mess("Flushing database to disk...\n");
    }
    TRACE("ioFlush", doIndex);

    offset = SORT_BUCKETS;
    shift = 0;
    while (offset < DataSize) {
	offset *= 2;
	shift += 1;
    }

    saveCount = getTimeout();

    Flushing = B_TRUE;

    /* initialize the buckets to all empty */

    for (i = 0; i < SORT_BUCKETS; i += 1) {
	radixTable[i] = P_NULL;
    }

    /* portion the dirty entries into the buckets based on their file offset */

    ca = CacheLRUHead;
    while (ca != P_NULL) {
	next = ca->ca_next;
	if ((ca->ca_slotLength & CA_DIRTY) != 0) {
	    ix = IndexPtr + (ca->ca_key & INDEX_MASK);
	    offset = ix->ix_offset;
	    if ((offset & IX_FREE) != 0) {
		logX("object ", ca->ca_key, " is dirty, but free in index\n");
		localAbort();
	    }
	    offset >>= shift;
	    ca->ca_next = radixTable[offset];
	    radixTable[offset] = ca;
	}
	ca = next;
    }

    /* go flush each bucket (an n**2 operation) */

    for (i = 0; i < SORT_BUCKETS; i += 1) {
	ca = radixTable[i];
	if (ca != P_NULL) {
	    ioFlushBucket(ca);
	}
    }
    if (DiskBufferDirty) {
	ioFlushBuffer();
    }

    /* now rebuild the LRU forward pointers based on the remaining
       backwards pointers */

    ca = CacheLRUTail;
    next = P_NULL;
    while (ca != P_NULL) {
	ca->ca_next = next;
	next = ca;
	ca = ca->ca_prev;
    }
    if (CacheLRUHead != next) {
	log("CacheLRUHead != next in ioFlush\n");
	localAbort();
    }

    /* now write out the index data */

    if (doIndex) {
	(void) unlink(INDEX_NAME);
	fd = open(INDEX_NAME, O_WRONLY | O_CREAT, 0600);
	if (fd == -1) {
	    log("can't open/create new index file\n");
	    localAbort();
	}
	saveLong(fd, (VERSION1 - '0') * 10 + (VERSION2 - '0'),
		 "server version");
	saveLong(fd, IndexCurrent, "current index size");
	saveLong(fd, DataSize, "current data size");
	saveLong(fd, PublicSymbolTableKey, "public symbol table key");
	saveLong(fd, EntryAction, "entry action");
	saveLong(fd, CreationPassword, "creation password");
	saveLong(fd, DungeonFlags, "dungeon flags");
	saveLong(fd, MachineList, "machine list");
	saveLong(fd, AbortTimeout, "abort timeout");
	offset = IndexCurrent * sizeof(Index_t);
	if (myWrite(fd, IndexPtr, offset, B_FALSE) != offset) {
	    close(fd);
	    log("can't write index to index file\n");
	    localAbort();
	}
	if ((DungeonFlags & FLAG_DATABASE_STABLE) == 0) {
	    /* mark database as now stable */
	    DungeonFlags |= FLAG_DATABASE_STABLE;
	    if (lseek(fd, FLAGS_OFFSET, SEEK_SET) == -1) {
		close(fd);
		log("fail to seek to flags\n");
		localAbort();
	    }
	    offset = htonl(DungeonFlags);
	    if (write(fd, &offset, sizeof(ULONG_T)) != sizeof(ULONG_T)) {
		close(fd);
		log("fail to rewrite flags\n");
		localAbort();
	    }
	}
	close(fd);
    }

    Flushing = B_FALSE;

    restoreTimeout(saveCount);
    if (Verbose) {
	mess("...done.\n");
    }
}

/*
 * ioCacheFree - free a cache slot.
 */

static void
ioCacheFree(Cache_t *ca)
{
    Cache_t *this, *next, *prev;
    BYTE_T *cacheEnd;
#if CLEAR_ON_FREE
    ULONG_T *lp;
#endif

    TRACE("ioCacheFree", ca);

    ca->ca_slotLength &= ~CA_DIRTY;
    ca->ca_usedLength = CA_FREE;
    IndexCa[ca->ca_key & INDEX_MASK] = P_NULL;

#if CLEAR_ON_FREE
    ca->ca_key = 0;
    lp = (ULONG_T *) (ca + 1);
    offset2 = ca->ca_slotLength / sizeof(ULONG_T);
    while (offset2 != 0) {
	offset2 -= 1;
	*lp++ = 0;
    }
#endif

    /* first, remove the piece from the in-use cache chain */

    prev = ca->ca_prev;
    next = ca->ca_next;
    if (prev != P_NULL) {
	prev->ca_next = next;
    } else {
	CacheLRUHead = next;
    }
    if (next != P_NULL) {
	next->ca_prev = prev;
    } else {
	CacheLRUTail = prev;
    }

    /* Merge this piece with any free ones after it in the cache */

    cacheEnd = CachePtr + CacheSize - sizeof(Cache_t);
    while (B_TRUE) {
	this = (Cache_t *) ((char *) ca + ca->ca_slotLength);
	if ((BYTE_T *) this > cacheEnd) {
	    break;
	}
	if (this->ca_usedLength != CA_FREE) {
	    break;
	}
	ca->ca_slotLength += this->ca_slotLength;
	prev = this->ca_prev;
	next = this->ca_next;
	if (prev == P_NULL) {
	    CacheFreeHead = next;
	} else {
	    prev->ca_next = next;
	}
	if (next != P_NULL) {
	    next->ca_prev = prev;
	}
    }

    /* link the merged piece into the free list */

    next = CacheFreeHead;
    ca->ca_next = next;
    ca->ca_prev = P_NULL;
    if (next != P_NULL) {
	next->ca_prev = ca;
    }
    CacheFreeHead = ca;

    /* do this for good measure */
    ca->ca_key = 0;
}

/*
 * mergeFreeCacheSlots - run through the entire cache, merging all free
 *	cache slots that are adjacent.
 */

static void
mergeFreeCacheSlots(void)
{
    Cache_t *ca, *this, *prev, *next;
    BYTE_T *cacheEnd;

    cacheEnd = CachePtr + CacheSize - sizeof(Cache_t);
    ca = (Cache_t *) CachePtr;
    while ((BYTE_T *) ca <= cacheEnd) {
	/* Search for a free slot */
	while ((BYTE_T *) ca <= cacheEnd) {
	    if (ca->ca_usedLength == CA_FREE) {
		break;
	    }
	    ca = (Cache_t *) ((BYTE_T *) ca + (ca->ca_slotLength & ~CA_DIRTY));
	}
	if ((BYTE_T *) ca > cacheEnd) {
	    break;
	}
	this = ca;
	/* Got a free slot - merge it with all free ones after it. */
	while (B_TRUE) {
	    this = (Cache_t *) ((BYTE_T *) this + this->ca_slotLength);
	    if ((BYTE_T *) this > cacheEnd) {
		ca = this;
		break;
	    }
	    if (this->ca_usedLength != CA_FREE) {
		ca = this;
		break;
	    }
	    ca->ca_slotLength += this->ca_slotLength;
	    prev = this->ca_prev;
	    next = this->ca_next;
	    if (prev == P_NULL) {
		CacheFreeHead = next;
	    } else {
		prev->ca_next = next;
	    }
	    if (next != P_NULL) {
		next->ca_prev = prev;
	    }
	}
    }
}

/*
 * ioFlush - external interface to ioFlush0. Also merge cache slots.
 */

void
ioFlush(BOOL_T doIndex)
{

    ioFlush0(doIndex);
    mergeFreeCacheSlots();
}

/*
 * findCacheSlot - find or create space in the cache for an entry of the
 *	given size.
 *	Note: in normal situations this is not too hard. What do we do,
 *	however, if we cannot find space? Well, we have to flush the whole
 *	cache and go through freeing slots until we have one big enough
 *	for what we need. Note that we NEVER try to re-use a slot which is
 *	within CACHE_KEEP of the LRU head, since that might still be in
 *	use by server code (e.g. especially addProp).
 */

static Cache_t *
findCacheSlot(UINT_T lenP)
{
    Cache_t *ca, *prev, *next;
    Index_t *ix;
    ULONG_T len, otherLen;

    len = lenP;
    TRACE("findCacheSlot", len);

    /* round size up so that we get long-word alignment */
    len = (len + 3) & (-4);
    len += sizeof(Cache_t);

    /* First, simple search in the free list */
    ca = CacheFreeHead;
    while (ca != P_NULL && ca->ca_slotLength < len) {
	ca = ca->ca_next;
    }
    if (ca == P_NULL) {
	/* No luck. Merge free slots and try again. */
	mergeFreeCacheSlots();
	ca = CacheFreeHead;
	while (ca != P_NULL && ca->ca_slotLength < len) {
	    ca = ca->ca_next;
	}
	if (ca == P_NULL) {
	    /* Still no luck. Have to re-use an in-use slot. Make sure we
	       do not pick one too close to CacheLRUHead. */

	    otherLen = CACHE_KEEP;
	    next = CacheLRUHead;
	    while (otherLen != 0 && next != P_NULL) {
		otherLen -= 1;
		next = next->ca_next;
	    }
	    if (otherLen != 0) {
		logN("Too few cache entries to make one for len ", len, "\n");
		localAbort();
	    }

	    ca = CacheLRUTail;
	    while (B_TRUE) {
		if ((ca->ca_slotLength & ~CA_DIRTY) >= len || ca == next) {
		    break;
		}
		ca = ca->ca_prev;
	    }
	    if (ca == next) {
		/* No in-use slot is big enough. Time to get drastic. Flush
		   everything to disk, then run through and free, starting at
		   the LRU end of the chain. Have already done the
		   mergeFreeCacheSlots, so no need to do it again here.
		   However, lets request a compaction of the cache. */
		LastRefKey = 0;
		requestCacheCompact();
		ioFlush0(B_FALSE);
		ca = CacheLRUTail;
		while (ca != next) {
		    prev = ca->ca_prev;
		    ioCacheFree(ca);
		    ca = CacheFreeHead;
		    if (ca->ca_slotLength >= len) {
			/* Created one big enough! */
			break;
		    }
		    ca = prev;
		}
		if (ca == next) {
		    /* Could not find a slot. Give up in disgust. */
		    logN("Cannot create cache slot len ", len, " for item\n");
		    localAbort();
		}
	    } else {
		/* Found a single slot that is dirty, but big enough for
		   us to use. Flush it and use it. */
		if (ca->ca_key == LastRefKey) {
		    LastRefKey = 0;
		}
		ix = IndexPtr + (ca->ca_key & INDEX_MASK);
#if MUD_ENDIAN == MUD_LITTLE_ENDIAN
		swapBytes(ca + 1, ca->ca_key, B_FALSE);
#endif
		ioPutBytes(ca + 1, ix->ix_offset, ix->ix_usedLength);
#if MUD_ENDIAN == MUD_LITTLE_ENDIAN
		swapBytes(ca + 1, ca->ca_key, B_TRUE);
#endif
		ioCacheFree(ca);
		ca = CacheFreeHead;
	    }
	}
    }

    /* At this point we have found a slot big enough. It is now on the
       free list, so take it off of the free list. */
    prev = ca->ca_prev;
    next = ca->ca_next;
    if (prev == P_NULL) {
	CacheFreeHead = next;
    } else {
	prev->ca_next = next;
    }
    if (next != P_NULL) {
	next->ca_prev = prev;
    }

    /* mark what is ACTUALLY used */
    ca->ca_usedLength = lenP;

    /* Note: 'len' already includes sizeof(Cache_t) */
    otherLen = ca->ca_slotLength - len;
    if (otherLen >= sizeof(Cache_t)) {
	/* Put the tail of this slot onto the free list */
	ca->ca_slotLength = len;
	next = (Cache_t *) ((BYTE_T *) ca + len);
	next->ca_slotLength = otherLen;
	next->ca_usedLength = CA_FREE;
	next->ca_prev = P_NULL;
	prev = CacheFreeHead;
	next->ca_next = prev;
	if (prev != NULL) {
	    prev->ca_prev = next;
	}
	CacheFreeHead = next;
    }

    /* Put the slot on the front of the LRU chain. */
    next = CacheLRUHead;
    ca->ca_prev = P_NULL;
    ca->ca_next = next;
    if (next != P_NULL) {
	next->ca_prev = ca;
    }
    CacheLRUHead = ca;
    if (CacheLRUTail == P_NULL) {
	CacheLRUTail = ca;
    }
    return ca;
}

/*
 * expandIndexTable - attempt to expand the in-memory index table.
 */

static void
expandIndexTable(void)
{
    Index_t *newTable;
    Cache_t **newCa;
    ULONG_T newSize, newByteSize, i;

    TRACE("expandIndexTable", 0);
    newSize = IndexSize + INDEX_MIN;
    logN("expandIndexTable ", newSize, "\n");
    if ((newSize & TYPE_MASK) != 0) {
	log("database overflow - too many entries!!!\n");
	localAbort();
    }
    newByteSize = newSize * sizeof(Index_t);
    newTable = realloc(IndexPtr, newByteSize);
    if (newTable == P_NULL) {
	log("expansion failed\n");
	localAbort();
    }
    newCa = realloc(IndexCa, newSize * sizeof(Cache_t *));
    if (newCa == P_NULL) {
	log("expansion failed\n");
	localAbort();
    }
    for (i = IndexSize; i != newSize; i += 1) {
	newCa[i] = P_NULL;
    }
    IndexPtr = newTable;
    IndexSize = newSize;
    IndexByteSize = newByteSize;
    IndexCa = newCa;
}

/*
 * freeFileSlot - free the file slot pointed at.
 *	Assumption: the passed slot is not currently marked as free.
 */

static void
freeFileSlot(ULONG_T element)
{
    Index_t *ix, *ix2;
    ULONG_T *fp;
    ULONG_T offset;
    UINT_T len, i;
    BOOL_T found;

    TRACE("freeFileSlot", element);

    ix = IndexPtr + element;
    offset = ix->ix_offset;
    len = ix->ix_slotLength;
    found = B_FALSE;
    if (offset + len == DataSize) {
	/* We can just shrink the file. */
	DataSize -= len;
	if (element + 1 == IndexCurrent) {
	    IndexCurrent = element;
	    TRACE("freeFileSlot 1 IC", element);
	} else {
	    ix->ix_slotLength = 0;
	    ix->ix_offset = IndexSlotFreeHead;
	    IndexSlotFreeHead = element;
	}
	found = B_TRUE;
    } else {
	/* search through FreeTable for an entry that is adjacent to it, so
	   that we can merge */
	i = FreeTableNext;
	fp = &FreeTable[0];
	while (! found && i != 0) {
	    i -= 1;
	    ix2 = IndexPtr + *fp;
	    if (ix2->ix_offset == ((offset - ix2->ix_slotLength) | IX_FREE)) {
		/* this one is just before the freeing one - merge */
		ix2->ix_slotLength += len;
		if (element + 1 == IndexCurrent) {
		    IndexCurrent = element;
		    TRACE("freeFileSlot 2 IC", element);
		} else if (*fp + 1 == IndexCurrent) {
		    *fp = element;
		    ix->ix_offset = ix2->ix_offset;
		    ix->ix_slotLength = ix2->ix_slotLength;
		    IndexCurrent -= 1;
		    TRACE("freeFileSlot 3 IC", IndexCurrent);
		} else {
		    ix->ix_slotLength = 0;
		    ix->ix_offset = IndexSlotFreeHead;
		    IndexSlotFreeHead = element;
		}
		found = B_TRUE;
	    } else if (ix2->ix_offset == ((offset + len) | IX_FREE)) {
		/* this one is just after the freeing one - merge */
		ix2->ix_offset = offset | IX_FREE;
		ix2->ix_slotLength += len;
		if (element + 1 == IndexCurrent) {
		    IndexCurrent = element;
		    TRACE("freeFileSlot 4 IC", element);
		} else if (*fp + 1 == IndexCurrent) {
		    *fp = element;
		    ix->ix_offset = ix2->ix_offset;
		    ix->ix_slotLength = ix2->ix_slotLength;
		    IndexCurrent -= 1;
		    TRACE("freeFileSlot 5 IC", IndexCurrent);
		} else {
		    ix->ix_slotLength = 0;
		    ix->ix_offset = IndexSlotFreeHead;
		    IndexSlotFreeHead = element;
		}
		found = B_TRUE;
	    } else {
		++fp;
	    }
	}
    }

    if (! found) {
	if (FreeTableNext == FREE_TABLE_SIZE) {
	    FreedSinceTabledCount += 1;
	    if (FreedSinceTabledCount == FREE_TABLE_INTERVAL) {
		FreedSinceTabledCount = 0;
		buildFreeStuff();
	    }
	    /* We want this in the free table, since there are good chances
	       that we may be able to merge with it soon */
	    if (FreeTableNext == FREE_TABLE_SIZE) {
		TRACE("freeFileSlot punted", FreeTable[FREE_TABLE_SIZE - 1]);
		FreedSinceTabled = B_TRUE;
		FreeTable[FREE_TABLE_SIZE - 1] = element;
	    } else {
		FreeTable[FreeTableNext] = element;
		FreeTableNext += 1;
	    }
	} else {
	    FreeTable[FreeTableNext] = element;
	    FreeTableNext += 1;
	}
	/* Do NOT do this before the above call to 'buildFreeStuff', since
	   that could result in this one being in the table twice! */
	ix->ix_offset |= IX_FREE;
    }
}

/*
 * createFreeFileSlot - someone is carving off a piece. Create a new free
 *	file slot with the given offset and length.
 */

static void
createFreeFileSlot(ULONG_T offset, UINT_T len)
{
    Index_t *ix;
    ULONG_T freeSlot;

    TRACE("createFreeFileSlot", offset);
    TRACE("len is", len);
    freeSlot = IndexSlotFreeHead;
    if (freeSlot == 0) {
	freeSlot = IndexCurrent;
	if (freeSlot == IndexSize) {
	    expandIndexTable();
	}
	ix = IndexPtr + freeSlot;
	IndexCurrent += 1;
	TRACE("createFreeFileSlot 2 IC", IndexCurrent);
    } else {
	ix = IndexPtr + freeSlot;
	IndexSlotFreeHead = ix->ix_offset;
    }
    ix->ix_slotLength = len;
    ix->ix_usedLength = len;
    ix->ix_offset = offset;
    freeFileSlot(freeSlot);
    /* CANNOT do a checkIndex on ix/freeSlot here, since we may have merged
       that slot out of existence, and it might have been the last one in
       the index, in which case freeSlot now == IndexCurrent! */
}

/*
 * scanFreeTable - helper routine for 'findFileSlot' - try to find a file
 *	slot in 'FreeTable' that is the smallest size >= the passed length
 *	and is at the lowest file offset.
 *	Note: it is assumed that FreeTableNext != 0.
 */

static ULONG_T
scanFreeTable(UINT_T len)
{
    Index_t *base, *ix;
    ULONG_T *tp;
    ULONG_T offset, index, minOffset;
    UINT_T thisLen, minLen, i, minPos;

    minPos = 0xffff;
    minLen = 0xffff;
    minOffset = 0xffffffff;
    base = IndexPtr;
    tp = &FreeTable[0];
    for (i = 0; i < FreeTableNext; i += 1) {
	index = *tp++;
	ix = base + index;
	thisLen = ix->ix_slotLength;
	TRACE("index", index);
	TRACE("has len", thisLen);
	if (thisLen >= len) {
	    offset = ix->ix_offset & ~IX_FREE;
	    if (thisLen < minLen) {
		minLen = thisLen;
		minOffset = offset;
		minPos = i;
	    } else if (thisLen == minLen) {
		if (offset < minOffset) {
		    minOffset = offset;
		    minPos = i;
		}
	    }
	}
    }

    if (minPos == 0xffff) {
	TRACE("scanFreeTable - none found", FreeTableNext);
	index = 0;
    } else {
	FreeTableNext -= 1;
	index = FreeTable[minPos];
	/* replace the used one with the last one */
	FreeTable[minPos] = FreeTable[FreeTableNext];
    }
    return index;
}

/*
 * findFileSlot - find a slot in the data file for an item of the given length.
 *	Return the index (partial key) of the found slot.
 */

static ULONG_T
findFileSlot(UINT_T origLen)
{
    Index_t *ix;
    ULONG_T freeIndex, freeOffset, foundIndex;
    UINT_T len, newLen, minLen;

    len = origLen;
    TRACE("findFileSlot", len);
    if (len < FILE_LENGTH_MIN) {
	len = FILE_LENGTH_MIN;
    }

    /* It seems odd to do this when we do not need to. So, I will #if it
       out, and duplicate the free table scan below after a call. That should
       reduce CPU time used when lots of creations/deletions are happening. */
#if 0
    if (FreeTableNext != FREE_TABLE_SIZE && FreedSinceTabled) {
	buildFreeTable();
    }
#endif
    freeIndex = 0;
    ix = P_NULL;    /* shut up gcc */
    minLen = FreeTableNext;
    if (minLen != 0 || FreedSinceTabled) {
	/* Whole new strategy. Search the whole table, and accept the slot
	   with the smallest length that is big enough, and in the case of
	   duplicates (hopefully happens), use the one with the smallest
	   file offset. */
	if (minLen != 0) {
	    freeIndex = scanFreeTable(len);
	}
	if (freeIndex == 0) {
	    buildFreeTable();
	    if (FreeTableNext != 0) {
		freeIndex = scanFreeTable(len);
	    }
	}
	if (freeIndex != 0) {
	    /* Found a file slot big enough in the FreeTable. We will use it.*/
	    ix = IndexPtr + freeIndex;
	} else if (FreedSinceTabled) {
	    /* Note the use of 'FreedSinceTabled'. We only do this lengthy
	       linear search if we know there is a free slot that is not in
	       the FreeTable. */
	    freeIndex = 1;
	    ix = IndexPtr + 1;
	    foundIndex = 0;
	    minLen = 0xffff;
	    while (freeIndex != IndexCurrent &&
		(ix->ix_slotLength != len || (ix->ix_offset & IX_FREE) == 0))
	    {
		if (ix->ix_slotLength > len && ix->ix_slotLength < minLen &&
		    (ix->ix_offset & IX_FREE) != 0)
		{
		    foundIndex = freeIndex;
		    minLen = ix->ix_slotLength;
		}
		freeIndex += 1;
		++ix;
	    }
	    if (freeIndex == IndexCurrent) {
		/* no free slot of exactly right size */
		if (foundIndex != 0) {
		    /* found one that was larger - use it */
		    freeIndex = foundIndex;
		    ix = IndexPtr + freeIndex;
		} else {
		    /* nothing at all found */
		    freeIndex = 0;
		}
	    }
	} else {
	    freeIndex = 0;
	}
    }

    if (freeIndex == 0) {
	/* note - we never actually check slot 0 - that's fine by me */
	/* no suitable free slot - must allocate a new one on the end */
	freeIndex = IndexSlotFreeHead;
	if (freeIndex == 0) {
	    freeIndex = IndexCurrent;
	    if (freeIndex == IndexSize) {
		/* no more room in index table - try to expand it */
		expandIndexTable();
		/* must not use old 'ix' after this call */
	    }
	    ix = IndexPtr + freeIndex;
	    IndexCurrent += 1;
	    TRACE("findFileSlot IC", IndexCurrent);
	} else {
	    ix = IndexPtr + freeIndex;
	    IndexSlotFreeHead = ix->ix_offset;
	}
	ix->ix_usedLength = origLen;
	ix->ix_slotLength = len;
	ix->ix_offset = DataSize;
	/* we do NOT in fact have to write these new pieces out! Because of
	   the way we flush in increasing file order, and when we do flush,
	   we flush everything that is dirty, these will always get written
	   out correctly before we ever try to read them. */
	DataSize += len;
	TRACE("findFileSlot offset", ix->ix_offset);
    } else {
	/* A free slot was found - use it. Note that all cases where we can
	   get here will have 'ix' pointing at the slot. */
	ix->ix_offset &= ~IX_FREE;
	ix->ix_usedLength = origLen;
	newLen = ix->ix_slotLength - len;
	TRACE("findFileSlot offset", ix->ix_offset);
	if (newLen >= FILE_LENGTH_SLOP) {
	    /* split off the tail of the file slot and mark it as free */
	    ix->ix_slotLength = len;
	    freeOffset = len + ix->ix_offset;
	    createFreeFileSlot(freeOffset, newLen);
	    /* Note that ix may no longer be valid */
	}
    }
    return freeIndex;
}

/*
 * ioCreate - create a new item of the given length. Note that the cache
 *	entry is marked as dirty, so ioDirty need not be called, unless
 *	further changes are made.
 */

ULONG_T
ioCreate(UINT_T typ, void *p, UINT_T len)
{
    ULONG_T index;
    Cache_t *ca;

    CreateCount += 1;
    TRACE("ioCreate", typ);

    ca = findCacheSlot(len);
    index = findFileSlot(len);
    ca->ca_key = ((ULONG_T) typ << TYPE_SHIFT) | index;
    copy(ca + 1, p, len);
    ca->ca_slotLength |= CA_DIRTY;
    TRACE("ioCreated", ca->ca_key);
    IndexCa[index] = ca;
    return ca->ca_key;
}

/*
 * ioRead - read the given entry. It might be in the cache. If not, it must
 *	be read from disk and put into the cache. Return an absolute (and
 *	VERY volatile) pointer to the data in the cache.
 */

void *
ioRead(ULONG_T key)
{
    ULONG_T index, offset;
    Cache_t *ca, *prev, *next;
    Index_t *ix;

    if (key == LastRefKey) {
	return LastRefPtr;
    }

    LastRefKey = key;
    ReadCount += 1;
    LastKey = key;
    TRACE("ioRead", key);
    TRACE("IC", IndexCurrent);

    index = key & INDEX_MASK;
#if DO_CHECK_KEYS
    if (index == 0 || index >= IndexCurrent) {
	badKey(key, "ioRead");
    }
#endif
    ca = IndexCa[index];
    if (ca != P_NULL) {
	/* found the key already in the cache */
	prev = ca->ca_prev;
	if (prev != P_NULL) {
	    /* maintain LRU - move found entry to the front of the list */
	    next = ca->ca_next;
	    prev->ca_next = next;
	    if (next != P_NULL) {
		next->ca_prev = prev;
	    } else {
		CacheLRUTail = prev;
	    }
	    prev = CacheLRUHead;
	    prev->ca_prev = ca;
	    ca->ca_prev = P_NULL;
	    ca->ca_next = prev;
	    CacheLRUHead = ca;
	}
	TRACE("ioRead[return 1]", ca + 1);

	LastRefCa = ca;
	LastRefPtr = ca + 1;
	return ca + 1;
    }

    /* Not in cache - have to read it from the data file. First, find or
       create space for it in the cache. */

    ix = IndexPtr + index;
    offset = ix->ix_offset;
#if DO_CHECK_KEYS
    if ((offset & IX_FREE) != 0x0) {
	badKey(key, "ioRead");
    }
#endif
    ca = findCacheSlot(ix->ix_usedLength);
    ca->ca_key = key;
    IndexCa[index] = ca;

    /* now actually read the item into the cache slot */

    ioGetBytes(ca + 1, offset, ix->ix_usedLength);
#if MUD_ENDIAN == MUD_LITTLE_ENDIAN
    swapBytes(ca + 1, key, B_TRUE);
#endif

    TRACE("ioRead[main return]", ca + 1);
    LastRefCa = ca;
    LastRefPtr = ca + 1;
    return ca + 1;
}

/*
 * ioDirty - mark the cache entry dirty for this guy.
 */

void
ioDirty(ULONG_T key)
{
    ULONG_T index;
    Cache_t *ca, *prev;

    if (key == LastRefKey) {
	LastRefCa->ca_slotLength |= CA_DIRTY;
	return;
    }

    DirtyCount += 1;
    TRACE("ioDirty", key);

    index = key & INDEX_MASK;
    ca = IndexCa[index];
#if DO_CHECK_KEYS
    if (index > IndexCurrent) {
	badKey(key, "ioDirty");
    }
    if (ca == P_NULL) {
	logX("Key ", key, " not in cache for ioDirty\n");
	localAbort();
    }
#endif

    ca->ca_slotLength |= CA_DIRTY;
    prev = ca->ca_prev;
    if (prev != P_NULL) {
	/* maintain LRU - move entry to the front of the list */
	prev->ca_next = ca->ca_next;
	if (ca->ca_next != P_NULL) {
	    ca->ca_next->ca_prev = prev;
	} else {
	    CacheLRUTail = prev;
	}
	prev = CacheLRUHead;
	prev->ca_prev = ca;
	ca->ca_next = prev;
	ca->ca_prev = P_NULL;
	CacheLRUHead = ca;
    }
}

/*
 * ioExpand - this guy needs to expand. Find a new slot, and return the
 *	new cache pointer.
 */

void *
ioExpand(ULONG_T key, UINT_T newLen)
{
    Cache_t *ca, *prev, *this, *next;
    BYTE_T *cacheEnd;
    Index_t *ix, *ix2;
    ULONG_T *ft;
    void *resultPtr;
    ULONG_T index, thisLen, otherLen, freeIndex, newOffset;
    UINT_T x, i;

    if (key == LastRefKey) {
	ca = LastRefCa;
	if ((ca->ca_slotLength & ~CA_DIRTY) >= sizeof(Cache_t) + newLen) {
	    ix = IndexPtr + (key & INDEX_MASK);
	    if (ix->ix_slotLength >= newLen) {
		ix->ix_usedLength = newLen;
		ca->ca_usedLength = newLen;
		ca->ca_slotLength |= CA_DIRTY;
		return ca + 1;
	    }
	}
    }

    ExpandCount += 1;
    LastKey = key;
    TRACE("ioExpand", key);

    index = key & INDEX_MASK;
    ca = IndexCa[index];
#if DO_CHECK_KEYS
    if (index > IndexCurrent) {
	badKey(key, "ioExpand");
    }
    if (ca == P_NULL) {
	logX("Key ", key, " not in cache for 'ioExpand'\n");
	localAbort();
    }
#endif

    /* We unconditionally do the LRU stuff. If we can reuse the old slot,
       then we want to do it. If we can't then we want to do it anyway, so
       that the assumption about 'findCacheSlot' not invalidating our
       current pointer is much more likely to be true. */
    prev = ca->ca_prev;
    if (prev != P_NULL) {
	/* maintain LRU - move entry to the front of the list */
	next = ca->ca_next;
	prev->ca_next = next;
	if (next != P_NULL) {
	    next->ca_prev = prev;
	} else {
	    CacheLRUTail = prev;
	}
	prev = CacheLRUHead;
	prev->ca_prev = ca;
	ca->ca_next = prev;
	ca->ca_prev = P_NULL;
	CacheLRUHead = ca;
    }

    /* is the cache slot still big enough? If not, try to expand it. */

    ca->ca_slotLength &= ~CA_DIRTY;	/* Gets re-flagged later */
    if (ca->ca_slotLength < sizeof(Cache_t) + newLen) {
	cacheEnd = CachePtr + CacheSize - sizeof(Cache_t);
	while (B_TRUE) {
	    this = (Cache_t *) ((char *) ca + ca->ca_slotLength);
	    if ((BYTE_T *) this > cacheEnd) {
		break;
	    }
	    if (this->ca_usedLength != CA_FREE) {
		break;
	    }
	    ca->ca_slotLength += this->ca_slotLength;
	    prev = this->ca_prev;
	    next = this->ca_next;
	    if (prev == P_NULL) {
		CacheFreeHead = next;
	    } else {
		prev->ca_next = next;
	    }
	    if (next != P_NULL) {
		next->ca_prev = prev;
	    }
	}
	/* May have made it *really* big. Check for carve off. */
	thisLen = (newLen + 3) & (-4);
	thisLen += sizeof(Cache_t);
	if (ca->ca_slotLength >= thisLen + sizeof(Cache_t)) {
	    /* Put the tail of this slot onto the free list */
	    otherLen = ca->ca_slotLength - thisLen;
	    ca->ca_slotLength = thisLen;
	    this = (Cache_t *) ((BYTE_T *) ca + thisLen);
	    this->ca_slotLength = otherLen;
	    this->ca_usedLength = CA_FREE;
	    this->ca_prev = P_NULL;
	    prev = CacheFreeHead;
	    this->ca_next = prev;
	    if (prev != NULL) {
		prev->ca_prev = this;
	    }
	    CacheFreeHead = this;
	}
    }

    /* is the cache slot big enough yet? If not, go get another one. */

    if ((ca->ca_slotLength & ~CA_DIRTY) < sizeof(Cache_t) + newLen) {
	/* need to find a new cache slot for it */
	this = findCacheSlot(newLen);
	/* assumption: findCacheSlot does not invalidate ca */
	resultPtr = this + 1;
	copy(resultPtr, ca + 1, ca->ca_usedLength);
	this->ca_key = key;
	this->ca_usedLength = newLen;
	this->ca_slotLength |= CA_DIRTY;
	ioCacheFree(ca);
	IndexCa[index] = this;	/* AFTER the ioCacheFree call */
	if (key == LastRefKey) {
	    LastRefCa = this;
	    LastRefPtr = resultPtr;
	}
	TRACE("ioExpand got new cache slot", this);
    } else {
	/* using old slot */
	ca->ca_usedLength = newLen;
	resultPtr = ca + 1;
	ca->ca_slotLength |= CA_DIRTY;
	TRACE("ioExpand using old slot", ca);
    }

    /* now make sure the file slot is big enough */

    ix = IndexPtr + index;
    x = ix->ix_slotLength;
    if (x < newLen) {
	if (ix->ix_offset + x == DataSize) {
	    /* file slot is last in file - just expand it */
	    DataSize += newLen - x;
	    ix->ix_slotLength = newLen;
	    ix->ix_usedLength = newLen;
	    /* again, we don't actually have to write to expand the file */
	    TRACE("ioExpand grew file to", DataSize);
	} else {
	    /* see if the file slot after this one is in FreeTable */
	    i = FreeTableNext;
	    ft = &FreeTable[0];
	    freeIndex = 0;	/* shut up gcc */
	    while (i != 0) {
		i -= 1;
		freeIndex = *ft;
		ix2 = IndexPtr + freeIndex;
		if (((ix->ix_offset + x) | IX_FREE) == ix2->ix_offset) {
		    x += ix2->ix_slotLength;
		    freeIndex = *ft;
		    if (freeIndex + 1 == IndexCurrent) {
			/* it was last in index - just forget about it */
			IndexCurrent = freeIndex;
			TRACE("ioExpand merged & shrank to IC", freeIndex);
		    } else {
			TRACE("ioExpand merged away", freeIndex);
			ix2->ix_slotLength = 0;
			ix2->ix_offset = IndexSlotFreeHead;
			IndexSlotFreeHead = freeIndex;
		    }
		    FreeTableNext -= 1;
		    *ft = FreeTable[FreeTableNext];
		    /* want to check the one we just moved here, if any left */
		} else {
		    ++ft;
		}
	    }
	    if (x >= newLen) {
		ix->ix_usedLength = newLen;
		/* we have extended the slot enough to use it */
		if (x - newLen < FILE_LENGTH_SLOP) {
		    /* just use the whole, merged slot */
		    ix->ix_slotLength = x;
		    TRACE("ioExpand swallowing", freeIndex);
		} else {
		    /* carve off what we need and leave the rest */
		    ix->ix_slotLength = newLen;
		    createFreeFileSlot(ix->ix_offset + newLen, x - newLen);
		}
	    } else {
		/* find a new file slot and swap the contents of the two */
		freeIndex = findFileSlot(newLen);
		/* ix may not be valid after this call! */
		ix = IndexPtr + index;
		ix2 = IndexPtr + freeIndex;
		newOffset = ix->ix_offset;
		ix->ix_offset = ix2->ix_offset;
		ix2->ix_offset = newOffset;
		x = ix->ix_slotLength;
		ix->ix_slotLength = ix2->ix_slotLength;
		ix2->ix_slotLength = x;
		ix2->ix_usedLength = x;
		ix->ix_usedLength = newLen;
		freeFileSlot(freeIndex);
		TRACE("ioExpand replaced at", ix - IndexPtr);
	    }
	}
    } else {
	ix->ix_usedLength = newLen;
    }

    return resultPtr;
}

/*
 * ioShrink - this guy is shrinking - see if we should chop off part.
 */

void
ioShrink(ULONG_T key, UINT_T newLen)
{
    Cache_t *ca, *prev, *ca2;
    Index_t *ix;
    ULONG_T index;
    UINT_T thisLen, otherLen;

    if (key == LastRefKey) {
	ca = LastRefCa;
	ca->ca_usedLength = newLen;
	ca->ca_slotLength |= CA_DIRTY;
	(IndexPtr + (key & INDEX_MASK))->ix_usedLength = newLen;
	return;
    }

    ShrinkCount += 1;
    LastKey = key;
    TRACE("ioShrink", key);

    index = key & INDEX_MASK;
    ca = IndexCa[index];
#if DO_CHECK_KEYS
    if (index > IndexCurrent) {
	badKey(key, "ioShrink");
    }
    if (ca == P_NULL) {
	logX("Key ", key, " not in cache for 'ioShrink'\n");
	localAbort();
    }
#endif

    thisLen = ca->ca_slotLength & ~CA_DIRTY;
    if (thisLen >= newLen + sizeof(Cache_t) && thisLen >= newLen * 2) {
	/* old slot is overlarge, and it would be wasteful to keep
	   using it - find a new one and use that instead. */
	ca2 = findCacheSlot(newLen);
	ca2->ca_key = key;
	copy(ca2 + 1, ca + 1, newLen);
	ca2->ca_slotLength |= CA_DIRTY;
	ioCacheFree(ca);
	IndexCa[index] = ca2;	/* AFTER the ioCacheFree */
    } else {
	/* do LRU stuff */
	prev = ca->ca_prev;
	if (prev != P_NULL) {
	    /* maintain LRU - move entry to the front of the list */
	    prev->ca_next = ca->ca_next;
	    if (ca->ca_next != P_NULL) {
		ca->ca_next->ca_prev = prev;
	    } else {
		CacheLRUTail = prev;
	    }
	    prev = CacheLRUHead;
	    prev->ca_prev = ca;
	    ca->ca_next = prev;
	    ca->ca_prev = P_NULL;
	    CacheLRUHead = ca;
	}

	ca->ca_usedLength = newLen;
	ca->ca_slotLength |= CA_DIRTY;
    }

    ix = IndexPtr + index;
    ix->ix_usedLength = newLen;
    if (newLen < FILE_LENGTH_MIN) {
	newLen = FILE_LENGTH_MIN;
    }
    otherLen = ix->ix_slotLength - newLen;
    if (otherLen >= FILE_LENGTH_SLOP) {
	/* chop off the tail of the old file slot and free it */
	ix->ix_slotLength = newLen;
	createFreeFileSlot(ix->ix_offset + newLen, otherLen);
    }
}

/*
 * ioReplace - supply a replacement for an entry. The new one could be
 *	larger or smaller, or the same length.
 */

void
ioReplace(ULONG_T key, void *p, UINT_T newLen)
{
    Cache_t *ca, *prev;
    Index_t *ix, *ix2;
    ULONG_T index, len, freeIndex, newOffset;
    UINT_T otherLen;

    ReplaceCount += 1;
    LastKey = key;
    TRACE("ioReplace", key);

    index = key & INDEX_MASK;
    ca = IndexCa[index];
#if DO_CHECK_KEYS
    if (index > IndexCurrent) {
	badKey(key, "ioReplace");
    }
#endif
    if (ca == P_NULL) {
	ca = findCacheSlot(newLen);
	ca->ca_key = key;
    } else {
	len = ca->ca_slotLength & ~CA_DIRTY;
	if (len < sizeof(Cache_t) + newLen) {
	    ioCacheFree(ca);
	    ca = findCacheSlot(newLen);
	    ca->ca_key = key;
	} else if (len >= newLen + sizeof(Cache_t) && len >= newLen * 2) {
	    /* old slot is overlarge, and it would be wasteful to keep
	       using it - free it, find a new one, and use that. */
	    ioCacheFree(ca);
	    ca = findCacheSlot(newLen);
	    ca->ca_key = key;
	} else {
	    /* still using the old slot - do LRU stuff */
	    prev = ca->ca_prev;
	    if (prev != P_NULL) {
		/* maintain LRU - move entry to the front of the list */
		prev->ca_next = ca->ca_next;
		if (ca->ca_next != P_NULL) {
		    ca->ca_next->ca_prev = prev;
		} else {
		    CacheLRUTail = prev;
		}
		prev = CacheLRUHead;
		prev->ca_prev = ca;
		ca->ca_next = prev;
		ca->ca_prev = P_NULL;
		CacheLRUHead = ca;
	    }
	}
    }
    copy(ca + 1, p, newLen);
    ca->ca_slotLength |= CA_DIRTY;
    IndexCa[index] = ca;
    if (key == LastRefKey) {
	LastRefCa = ca;
	LastRefPtr = ca + 1;
    }

    ix = IndexPtr + index;
    newOffset = ix->ix_offset;
    if (ix->ix_slotLength >= newLen) {
	/* Old file slot is still big enough. Should we split it? */
	ix->ix_usedLength = newLen;
	if (newLen < FILE_LENGTH_MIN) {
	    newLen = FILE_LENGTH_MIN;
	}
	otherLen = ix->ix_slotLength;
	otherLen -= newLen;
	if (otherLen >= FILE_LENGTH_SLOP) {
	    /* chop off the tail of the old file slot and free it */
	    ix->ix_slotLength = newLen;
	    newOffset += newLen;
	    createFreeFileSlot(newOffset, otherLen);
	}
    } else {
	if (newOffset + ix->ix_slotLength == DataSize) {
	    /* file slot is last in file - just expand it */
	    DataSize += newLen - ix->ix_slotLength;
	    ix->ix_slotLength = newLen;
	    ix->ix_usedLength = newLen;
	    /* once again, we don't have to write to expand the file here */
	} else {
	    /* Old file slot is not big enough - must replace it. */
	    freeIndex = findFileSlot(newLen);
	    /* ix may not be valid after this call! */
	    ix = IndexPtr + index;
	    ix2 = IndexPtr + freeIndex;
	    newOffset = ix2->ix_offset;
	    ix2->ix_offset = ix->ix_offset;
	    ix->ix_offset = newOffset;
	    otherLen = ix2->ix_slotLength;
	    ix2->ix_slotLength = ix->ix_slotLength;
	    ix->ix_slotLength = otherLen;
	    ix2->ix_usedLength = ix->ix_usedLength;
	    ix->ix_usedLength = newLen;
	    freeFileSlot(freeIndex);
	}
    }
}

/*
 * ioDelete - free up an entry in the data file.
 */

void
ioDelete(ULONG_T key)
{
    Cache_t *ca;
    ULONG_T index;

    DeleteCount += 1;
    LastKey = key;
    TRACE("ioDelete", key);

    index = key & INDEX_MASK;
#if DO_CHECK_KEYS
    if (index > IndexCurrent) {
	badKey(key, "ioDelete");
    }
#endif
    freeFileSlot(index);

    /* have to remove any reference from the cache */
    ca = IndexCa[index];
    if (ca != P_NULL) {
	ioCacheFree(ca);
    }
    if (key == LastRefKey) {
	LastRefKey = 0;
    }
}

/*
 * ioWrite - flush the given value from the cache to disk.
 *	The item does stay in the cache, however.
 */

void
ioWrite(ULONG_T key)
{
    ULONG_T index;
    Cache_t *ca;
    Index_t *ix;

    LastKey = key;
    TRACE("ioWrite", key);

    index = key & INDEX_MASK;
#if DO_CHECK_KEYS
    if (index > IndexCurrent) {
	badKey(key, "ioWrite");
    }
#endif
    ca = IndexCa[index];
    if (ca != P_NULL && (ca->ca_slotLength & CA_DIRTY) != 0) {
	ca->ca_slotLength &= ~CA_DIRTY;
	ix = IndexPtr + index;
#if MUD_ENDIAN == MUD_LITTLE_ENDIAN
	swapBytes(ca + 1, key, B_FALSE);
#endif
	ioPutBytes(ca + 1, ix->ix_offset, ix->ix_usedLength);
#if MUD_ENDIAN == MUD_LITTLE_ENDIAN
	swapBytes(ca + 1, key, B_TRUE);
#endif
    }
}

/*
 * ioCacheCompact - no mud code is running right now, and we very
 *	recently could not find a cache slot. So, go through and
 *	compact the cache. Expensive!
 */

void
ioCacheCompact(void)
{
    Cache_t *caOld, *caNew, *prev;
    BYTE_T *caEnd;
    ULONG_T usedLen, oldSlotLen, newSlotLen, cleanOldSlotLen;

    if (Verbose) {
	mess("Compacting DB cache...\n");
    }
    LastRefKey = 0;
    caOld = (Cache_t *) CachePtr;
    caNew = caOld;
    caEnd = CachePtr + CacheSize - sizeof(Cache_t);
    CacheLRUHead = caNew;
    prev = P_NULL;
    while ((BYTE_T *) caOld <= caEnd) {
	usedLen = caOld->ca_usedLength;
	if (usedLen == CA_FREE) {
	    /* This slot is free - just skip it. */
	    caOld = (Cache_t *) ((BYTE_T *) caOld + caOld->ca_slotLength);
	} else {
	    /* This slot is in use. Copy data down. Be careful to do
	       the fields in increasing address order, in case they
	       overlap. If they do overlap, it can only be by multiples
	       of 4 bytes. */
	    caNew->ca_key = caOld->ca_key;
	    IndexCa[caNew->ca_key & INDEX_MASK] = caNew;
	    caNew->ca_prev = prev;
	    if (prev != P_NULL) {
		prev->ca_next = caNew;
	    }
	    prev = caNew;
	    newSlotLen = ((usedLen + 3) & (-4)) + sizeof(Cache_t);
	    oldSlotLen = caOld->ca_slotLength;
	    cleanOldSlotLen = oldSlotLen & ~CA_DIRTY;
	    caNew->ca_slotLength = newSlotLen;
	    if ((oldSlotLen & CA_DIRTY) != 0) {
		caNew->ca_slotLength |= CA_DIRTY;
	    }
	    caNew->ca_usedLength = usedLen;
	    copy(caNew + 1, caOld + 1, usedLen);
	    caOld = (Cache_t *) ((BYTE_T *) caOld + cleanOldSlotLen);
	    caNew = (Cache_t *) ((BYTE_T *) caNew + newSlotLen);
	}
    }

    /* We have totally trashed the LRU list. However, we have been linking
       the new blocks, so just finish off the tail end. */
    if (prev != P_NULL) {
	prev->ca_next = P_NULL;
    }
    CacheLRUTail = prev;

    /* Put the rest of the cache space into one large free block. */
    CacheFreeHead = caNew;
    caNew->ca_prev = P_NULL;
    caNew->ca_next = P_NULL;
    caNew->ca_slotLength = caEnd + sizeof(Cache_t) - (BYTE_T *) caNew;
    caNew->ca_usedLength = CA_FREE;

    if (Verbose) {
	mess("...done.\n");
    }
}

/*
 * ioTerminate - shut down the I/O structures, including flushing everything.
 */

void
ioTerminate(void)
{
    /* Grrr. This wouldn't be necessary if C had proper string escapes! */
    static char message[] = {
	'M', 'U', 'D', ' ', 's', 'e', 'r', 'v', 'e', 'r', ' ',
	'V', VERSION1, '.', VERSION2, ' ',
	'C', 'o', 'p', 'y', 'r', 'i', 'g', 'h', 't', ' ',
	'1', '9', YEAR1, YEAR2, ' ', 'b', 'y', ' ',
	'C', 'h', 'r', 'i', 's', ' ', 'G', 'r', 'a', 'y', ' ',
	's', 'h', 'u', 't', ' ', 'd', 'o', 'w', 'n', '\n', '\0'
    };

    TRACE("ioTerminate", 0);
    if (! Flushing) {
	ioFlush0(B_TRUE);
    }

    logN("cre ", CreateCount, "");
    logN(" del ", DeleteCount, "");
    logN(" rd ", ReadCount, "");
    logN(" rep ", ReplaceCount, "");
    logN(" shr ", ShrinkCount, "");
    logN(" exp ", ExpandCount, "");
    logN(" dir ", DirtyCount, "\n");
    logN("frd ", FReadCount, "");
    logN(" fwt ", FWriteCount, "");
    logN(" fsk ", FSeekCount, "\n");

    if (Verbose) {
	mess("MUDServ database flushed and closed.\n");
    }
    ioShutDown(&message[0]);
}

/*
 * ioDisplay - print what we know about a given possible key
 */

void
ioDisplay(ULONG_T key)
{
    Cache_t *ca;
    Index_t *ix;
    ULONG_T *p;
    ULONG_T index;

    index = key & INDEX_MASK;
    ca = IndexCa[index];
    if (ca != P_NULL) {
	putString("in cache ");
	key = ca->ca_slotLength;
	if ((key & CA_DIRTY) != 0) {
	    key &= ~CA_DIRTY;
	    putString("dirty ");
	} else {
	    putString("not dirty ");
	}
	newLine();
	p = (ULONG_T *) (ca + 1);
	for (key = 0; key < 8; key += 1) {
	    putHex(*p++);
	    putChar(' ');
	}
    } else {
	putString("not in cache ");
	ix = IndexPtr + index;
	if ((ix->ix_offset & IX_FREE) != 0) {
	    putString("FREED ");
	} else {
	    putString("slot length ");
	    putDecimal(ix->ix_slotLength);
	    putString(" used length ");
	    putDecimal(ix->ix_usedLength);
	    putString(" offset ");
	    putHex(key);
	}
    }
    putChar('\n');
}

/*
 * getPublicSymbolTableKey - return the key to the public symbol table.
 */

ULONG_T
getPublicSymbolTableKey(void)
{

    return PublicSymbolTableKey;
}

/*
 * getEntryAction - return the entry action.
 */

ULONG_T
getEntryAction(void)
{

    return EntryAction;
}

/*
 * setEntryAction - set in a new entry action (already checked).
 */

void
setEntryAction(ULONG_T key)
{

    EntryAction = key;
}

/*
 * getCreationPassword - return the key for the creation password string.
 */

ULONG_T
getCreationPassword(void)
{

    return CreationPassword;
}

/*
 * getDungeonFlags
 */

ULONG_T
getDungeonFlags(void)
{

    return DungeonFlags;
}

/*
 * setDungeonFlags
 */

void
setDungeonFlags(ULONG_T newFlags)
{

    DungeonFlags = newFlags;
}

/*
 * getMachineList
 */

ULONG_T
getMachineList(void)
{

    return MachineList;
}

/*
 * getAbortTimeout
 */

ULONG_T
getAbortTimeout(void)
{

    return AbortTimeout;
}

void
setAbortTimeout(ULONG_T newValue)
{

    AbortTimeout = newValue;
}

/*
 * ShowDBStatus - SysAdmin-callable function that shows a bunch of useful
 *	info about the status of the database and cache.
 */

typedef struct CountElement {
    struct CountElement *ce_next;
    ULONG_T ce_size1, ce_size2;
    UINT_T ce_count;
} CountElement_t;

static void
freeCountElements(CountElement_t *ce)
{
    CountElement_t *ceTemp;

    while (ce != P_NULL) {
	ceTemp = ce;
	ce = ce->ce_next;
	free(ceTemp);
    }
}

static void
insertCountElement(CountElement_t **cep, ULONG_T size1, ULONG_T size2)
{
    CountElement_t *ce;

    while (B_TRUE) {
	ce = *cep;
	if (ce == P_NULL) {
	    break;
	}
	if (ce->ce_size1 > size1) {
	    break;
	}
	if (ce->ce_size1 == size1) {
	    if (ce->ce_size2 > size2) {
		break;
	    }
	    if (ce->ce_size2 == size2) {
		ce->ce_count += 1;
		return;
	    }
	}
	cep = &ce->ce_next;
    }
    ce = malloc(sizeof(CountElement_t));
    ce->ce_next = *cep;
    ce->ce_size1 = size1;
    ce->ce_size2 = size2;
    ce->ce_count = 1;
    *cep = ce;
}

static void
printCountElements(char *header, CountElement_t *ce, BOOL_T showBoth)
{
    ULONG_T total1, total2, count;

    putString(header);
    putChar(':');
    total1 = 0;
    total2 = 0;
    count = 0;
    while (ce != P_NULL) {
	count += ce->ce_count;
	total1 += ce->ce_size1 * ce->ce_count;
	putChar(' ');
	putDecimal(ce->ce_size1);
	if (showBoth) {
	    total2 += ce->ce_size2 * ce->ce_count;
	    putChar(':');
	    putDecimal(ce->ce_size2);
	}
	putChar('#');
	putDecimal(ce->ce_count);
	ce = ce->ce_next;
    }
    if (showBoth) {
	putString(" total1 ");
	putDecimal(total1);
	putString(" total2 ");
	putDecimal(total2);
    } else {
	putString(" total ");
	putDecimal(total1);
    }
    putString(" count ");
    putDecimal(count);
    putChar('\n');
}

static void
countFreeCacheSlots(void)
{
    CountElement_t *ceList;
    Cache_t *ca;

    ceList = P_NULL;
    ca = CacheFreeHead;
    while (ca != P_NULL) {
	insertCountElement(&ceList, ca->ca_slotLength, 0);
	ca = ca->ca_next;
    }
    printCountElements("Free cache slots", ceList, B_FALSE);
    freeCountElements(ceList);
}

static void
countUsedCacheSlots(void)
{
    CountElement_t *ceList;
    Cache_t *ca;

    ceList = P_NULL;
    ca = CacheLRUHead;
    while (ca != P_NULL) {
	insertCountElement(&ceList, (ca->ca_slotLength & ~CA_DIRTY),
			   ca->ca_usedLength);
	ca = ca->ca_next;
    }
    printCountElements("Used cache slots", ceList, B_TRUE);
    freeCountElements(ceList);
}

static void
showFileSlots(void)
{
    ULONG_T i, totalFree, totalUsed, totalActive, slotsFree;
    Index_t *ix;
    CountElement_t *ceFree, *ceUsed;

    ceFree = P_NULL;
    ceUsed = P_NULL;
    totalFree = 0;
    totalUsed = 0;
    totalActive = 0;
    slotsFree = 0;
    for (i = 0, ix = IndexPtr; i != IndexSize; i += 1, ++ix) {
	if (ix->ix_slotLength != 0) {
	    if ((ix->ix_offset & IX_FREE) != 0) {
		insertCountElement(&ceFree, ix->ix_slotLength, 0);
		totalFree += ix->ix_slotLength;
	    } else {
		insertCountElement(&ceUsed, ix->ix_slotLength,
				   ix->ix_usedLength);
		totalUsed += ix->ix_slotLength;
		totalActive += ix->ix_usedLength;
	    }
	} else {
	    slotsFree += 1;
	}
    }
    printCountElements("Free file slots", ceFree, B_FALSE);
    printCountElements("Used file slots", ceUsed, B_TRUE);
    freeCountElements(ceFree);
    freeCountElements(ceUsed);
    putString("free index slots ");
    putDecimal(slotsFree);
    putString("  total free ");
    putDecimal(totalFree);
    putString("  total used ");
    putDecimal(totalUsed);
    putString("  total active ");
    putDecimal(totalActive);
    putString("  ");
    putDecimal(totalFree + totalUsed);
    putChar('/');
    putDecimal(DataSize);
    putChar('\n');
}

void
ioShowDBStatus(void)
{

    putString("cache size ");
    putDecimal(CacheSize);
    putString(" IndexCurrent ");
    putDecimal(IndexCurrent);
    putString("  IndexSize ");
    putDecimal(IndexSize);
    putString("  DataSize ");
    putDecimal(DataSize);
    putChar('\n');

    countFreeCacheSlots();
    countUsedCacheSlots();
    showFileSlots();
}
