#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * listprop.c - routines to support all the builtin list and property procs.
 */

#define IOREAD(key) (((key) == LastRefKey) ? LastRefPtr : ioRead(key))

/*
 * NOTE: freeList can do a lot of cache activities, so it is unwise to
 *	assume anything stays in the cache through a freeList call.
 */

void
freeList(ULONG_T list, UINT_T typ)
{
    ItemList_t *l;
    ULONG_T *p;
    UINT_T i, count;

    TRACE("freeList", list);
    l = IOREAD(list);
    l->l_useCount -= 1;
    if (l->l_useCount != 0) {
	ioDirty(list);
	return;
    }

    if (typ == TYPE_LIST_THING || typ == TYPE_LIST_ACTION) {
	count = l->l_elementCount;
	for (i = 0; i < count; i += 1) {
	    p = (ULONG_T *) (l + 1) + i;
	    if (typ == TYPE_LIST_THING) {
		decRef(*p);
	    } else {
		decProcRef(*p);
	    }
	    /* Refetch it each time. Because of the decRef or
	       decProcRef calls, it is a little too dangerous to assume
	       it will stay in the cache */
	    l = IOREAD(list);
	}
    }
    ioDelete(list);
}

static ULONG_T
createProp(UINT_T typ)
{
    Property_t prp;

    TRACE("createProp", typ);
    prp.prp_owner = EffectivePlayer;
    prp.prp_useCount = 0;
    return ioCreate(typ, &prp, sizeof(Property_t));
}

ULONG_T
CreateBoolProp(void)
{

    return createProp(TYPE_PROP_BOOLEAN);
}

ULONG_T
CreateIntProp(void)
{

    return createProp(TYPE_PROP_INTEGER);
}

ULONG_T
CreateFixedProp(void)
{

    return createProp(TYPE_PROP_FIXED);
}

ULONG_T
CreateStringProp(void)
{

    return createProp(TYPE_PROP_STRING);
}

ULONG_T
CreateThingProp(void)
{

    return createProp(TYPE_PROP_THING);
}

ULONG_T
CreateActionProp(void)
{

    return createProp(TYPE_PROP_ACTION);
}

ULONG_T
CreateTableProp(void)
{

    return createProp(TYPE_PROP_TABLE);
}

ULONG_T
CreateGrammarProp(void)
{

    return createProp(TYPE_PROP_GRAMMAR);
}

ULONG_T
CreateIntListProp(void)
{

    return createProp(TYPE_PROP_LIST_INTEGER);
}

ULONG_T
CreateFixedListProp(void)
{

    return createProp(TYPE_PROP_LIST_FIXED);
}

ULONG_T
CreateThingListProp(void)
{

    return createProp(TYPE_PROP_LIST_THING);
}

ULONG_T
CreateActionListProp(void)
{

    return createProp(TYPE_PROP_LIST_ACTION);
}

ULONG_T
CreateThing(ULONG_T parent)
{
    Thing_t th;

    TRACE("CreateThing", 0);
    ThingsCreated += 1;
    if (parent != NIL_THING) {
	incRef(parent);
    }
    th.th_parent = parent;
    th.th_owner = EffectivePlayer;
    th.th_useCount = 0;
    th.th_propCount = 0;
    th.th_status = ts_private;
    return ioCreate(TYPE_THING, &th, sizeof(Thing_t));
}

void
SetParent(ULONG_T key, ULONG_T newParent)
{
    Thing_t *th;
    ULONG_T k;

    if (key == NIL_THING) {
	runError("nil thing passed to SetParent");
	return;
    }

    if (newParent != NIL_THING) {
	k = newParent;
	do {
	    if (k == key) {
		runError("attempt to create loop in SetParent");
		return;
	    }
	    th = IOREAD(k);
	    k = th->th_parent;
	} while (k != NIL_THING);
	incRef(newParent);
    }
    th = IOREAD(key);
    k = th->th_parent;
    th->th_parent = newParent;
    ioDirty(key);
    if (k != NIL_THING) {
	decRef(k);
    }
}

ULONG_T
Parent(ULONG_T key)
{

    if (key == NIL_THING) {
	runError("nil thing passed to Parent");
	return NIL_THING;
    }

    return ((Thing_t *) IOREAD(key))->th_parent;
}

ULONG_T
IsAncestor(ULONG_T myKey, ULONG_T parentKey)
{

    if (myKey == NIL_THING || parentKey == NIL_THING) {
	runError("nil thing passed to HasParent");
	return BOOLEAN_FALSE;
    }

    do {
	myKey = ((Thing_t *) IOREAD(myKey))->th_parent;
    } while (myKey != parentKey && myKey != NIL_THING);
    return ((myKey == parentKey) ? BOOLEAN_TRUE : BOOLEAN_FALSE);
}

void
SetThingStatus(ULONG_T key, ULONG_T status)
{
    Thing_t *th;

    if (key == NIL_THING) {
	runError("nil thing passed to SetThingStatus");
	return;
    }

    th = IOREAD(key);
    if (th->th_owner == EffectivePlayer ||
	(EffectivePlayer == SYSADMIN_KEY && RealPlayer == SYSADMIN_KEY))
    {
	th->th_status = status + ts_private;
	ioDirty(key);
    }
}

ULONG_T
GetThingStatus(ULONG_T key)
{

    if (key == NIL_THING) {
	runError("nil thing passed to GetThingStatus");
	return ts_private - ts_private;
    }

    return (ULONG_T) (((Thing_t *) IOREAD(key))->th_status - ts_private);
}

ULONG_T
Mine(ULONG_T key)
{

    TRACE("Mine", key);
    if (key == NIL_THING) {
	runError("nil thing passed to Mine");
	return BOOLEAN_FALSE;
    }

    return ((((Thing_t *) IOREAD(key))->th_owner == EffectivePlayer) ?
	BOOLEAN_TRUE : BOOLEAN_FALSE);
}

ULONG_T
Owner(ULONG_T key)
{

    TRACE("Owner", key);
    if (key == NIL_THING) {
	runError("nil thing passed to Owner");
	return NIL_PLAYER;
    }

    return ((Thing_t *) IOREAD(key))->th_owner;
}

void
addProp(ULONG_T thing, ULONG_T property, ULONG_T valu, UINT_T typ)
{
    Thing_t *th;
    Attr_t *at;
    CHAR_T *cp;
    ItemList_t *l;
    ULONG_T oldValu;
    BYTE_T count, oldCount;

    TRACE("addProp", thing);
    if (typ >= TYPE_THING && typ <= TYPE_LIST_ACTION &&
	(valu & INDEX_MASK) == 0)
    {
	runError("attempt to add a nil value for a property");
	return;
    }

    if ((property & INDEX_MASK) == 0) {
	runError("attempt to add a nil property");
	return;
    }

    if (thing == NIL_THING) {
	runError("attempt to add a property to nil");
	return;
    }

    th = IOREAD(thing);
    if (th->th_owner != EffectivePlayer && th->th_status != ts_public &&
	(th->th_status != ts_wizard || EffectiveStatus != ps_wizard) &&
	(EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY))
    {
	if (th->th_status != ts_public) {
	    runError("thing is not writeable by others");
	} else {
	    runError("thing is not writeable by apprentices");
	}
	return;
    }

    count = th->th_propCount;
    oldCount = count;
    at = (Attr_t *) (th + 1);
    while (count != 0 && at->at_prop != property) {
	count -= 1;
	++at;
    }
    if (count != 0) {
	/* We are just updating an existing value. */
	if (typ == TYPE_STRING) {
	    cp = (CHAR_T *) valu;
	    ioReplace(at->at_value, cp, strLen1(cp));
	} else {
	    oldValu = at->at_value;
	    if (valu != oldValu) {
		at->at_value = valu;
		ioDirty(thing);
		if (typ == TYPE_ACTION) {
		    /* Do the incProcRef first */
		    incProcRef(valu);
		    decProcRef(oldValu);
		} else if (typ == TYPE_THING || typ == TYPE_TABLE ||
			   typ == TYPE_GRAMMAR)
		{
		    /* *MUST* do the incRef first, now that a decRef
		       can recurse through pointed-at things - since if
		       there is a linked list of things, headed by the
		       old value, doing the decRef first will nuke the
		       entire list! */
		    incRef(valu);
		    decRef(oldValu);
		} else if (typ >= TYPE_LIST_INTEGER &&
			   typ <= TYPE_LIST_ACTION)
		{
		    l = IOREAD(valu);	    /* fetch the new list value */
		    l->l_useCount += 1;     /* now has one more use */
		    ioDirty(valu);	    /* mark new as dirty in cache */
		    freeList(oldValu, typ); /* one less ref to old list */
		}
	    }
	}
    } else if (oldCount == 0xff) {
	runError("attempt to add a property to a full thing");
    } else {
	/* We are adding a whole new property to the thing. */
	incRef(property);
	if (typ == TYPE_STRING) {
	    cp = (CHAR_T *) valu;
	    valu = ioCreate(TYPE_STRING, cp, strLen1(cp));
	}
	count = oldCount + 1;
	th = ioExpand(thing,
		      (ULONG_T) count * sizeof(Attr_t) + sizeof(Thing_t));
	th->th_propCount = count;
	/* Old 'at' pointer is likely now invalid - recompute it. */
	at = (Attr_t *) (th + 1) + oldCount;
	at->at_prop = property;
	at->at_value = valu;
	/* already marked as dirty by ioExpand */
	if (typ == TYPE_THING || typ == TYPE_TABLE || typ == TYPE_GRAMMAR) {
	    incRef(valu);
	} else if (typ == TYPE_ACTION) {
	    incProcRef(valu);
	} else if (typ >= TYPE_LIST_INTEGER && typ <= TYPE_LIST_ACTION) {
	    l = IOREAD(valu);
	    l->l_useCount += 1;
	    ioDirty(valu);
	}
    }
}

void
delProp(ULONG_T thing, ULONG_T property, UINT_T typ)
{
    Thing_t *th;
    Attr_t *at;
    BYTE_T count, n;

    TRACE("delProp", thing);
    if (thing == NIL_THING) {
	runError("attempt to delete a property from nil");
	return;
    }

    th = IOREAD(thing);
    if (th->th_owner != EffectivePlayer && th->th_status != ts_public &&
	(th->th_status == ts_wizard || EffectiveStatus != ps_wizard) &&
	(EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY))
    {
	if (th->th_status != ts_public) {
	    runError("thing is not writeable by others");
	} else {
	    runError("thing is not writeable by apprentices");
	}
	return;
    }

    n = th->th_propCount;
    at = (Attr_t *) (th + 1);
    count = 0;
    while (count != n && at->at_prop != property) {
	count += 1;
	++at;
    }
    if (count == n) {
	/* it is NOT an error to not find it */
	return;
    }

    decRef(property);
    property = at->at_value;		/* what was deleted */
    count += 1;
    if (count != n) {
	copy(at, at + 1, (ULONG_T) (n - count) * sizeof(Attr_t));
    }
    n -= 1;
    th->th_propCount = n;
    ioShrink(thing, (ULONG_T) n * sizeof(Attr_t) + sizeof(Thing_t));
    /* Note that it is important to do all of the above before
       doing any of these decRef's. This is because we don't
       want anything happening here recursing through structures
       and finding the old, unupdated form of the thing which
       might look like it still contains the property. This is
       especially the case if something ends up getting deleted. */
    if (typ == TYPE_STRING) {
	ioDelete(property);
    } else if (typ == TYPE_ACTION) {
	decProcRef(property);
    } else if (typ == TYPE_THING || typ == TYPE_TABLE || typ == TYPE_GRAMMAR) {
	decRef(property);
    } else if (typ >= TYPE_LIST_INTEGER && typ <= TYPE_LIST_ACTION) {
	freeList(property, typ);
    }
}

ULONG_T
getProp(ULONG_T thing, ULONG_T property, UINT_T typ)
{
    Thing_t *th;
    BYTE_T count;
    Attr_t *at;

    TRACE("getProp", thing);
    if ((property & INDEX_MASK) == 0) {
	runError("attempt to retrieve a nil property");
	return (ULONG_T) TYPE_ERROR << TYPE_SHIFT;
    }

    if (thing == NIL_THING) {
	runError("attempt to retrieve a property from nil");
	return nilValue(typ);
    }

    do {
	th = IOREAD(thing);
	if (th->th_owner != EffectivePlayer && th->th_status == ts_private &&
	    (EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY))
	{
	    runError("thing is not readable by others");
	    return nilValue(typ);
	}

	thing = th->th_parent;
	count = th->th_propCount;
	at = (Attr_t *) (th + 1);
	while (count != 0 && at->at_prop != property) {
	    count -= 1;
	    ++at;
	}
	if (count != 0) {
	    property = at->at_value;
	    if (typ == TYPE_STRING) {
		return (ULONG_T) stringCopy(IOREAD(property));
	    }
	    return property;
	}
    } while (thing != NIL_THING);

    /* this is NOT an error */
    return nilValue(typ);
}

ULONG_T
CreateIntList(void)
{
    ItemList_t l;

    TRACE("CreateIntList", 0);
    l.l_owner = EffectivePlayer;
    l.l_elementCount = 0;
    l.l_useCount = 0;
    return ioCreate(TYPE_LIST_INTEGER, &l, sizeof(ItemList_t));
}

ULONG_T
CreateFixedList(void)
{
    ItemList_t l;

    TRACE("CreateFixedList", 0);
    l.l_owner = EffectivePlayer;
    l.l_elementCount = 0;
    l.l_useCount = 0;
    return ioCreate(TYPE_LIST_FIXED, &l, sizeof(ItemList_t));
}

ULONG_T
CreateThingList(void)
{
    ItemList_t l;

    TRACE("CreateThingList", 0);
    l.l_owner = EffectivePlayer;
    l.l_elementCount = 0;
    l.l_useCount = 0;
    return ioCreate(TYPE_LIST_THING, &l, sizeof(ItemList_t));
}

ULONG_T
CreateActionList(void)
{
    ItemList_t l;

    TRACE("CreateActionList", 0);
    l.l_owner = EffectivePlayer;
    l.l_elementCount = 0;
    l.l_useCount = 0;
    return ioCreate(TYPE_LIST_ACTION, &l, sizeof(ItemList_t));
}

ULONG_T
CreateIntArray(ULONG_T size)
{
    ULONG_T *p;
    UINT_T n;
    ItemList_t l;
    ULONG_T key;

    TRACE("CreateIntArray", 0);
    if (size > 0xffff) {
	runError("size too big for CreateIntArray");
	return NIL_LIST_INTEGER;
    }

    l.l_owner = EffectivePlayer;
    l.l_elementCount = size;
    l.l_useCount = 0;
    key = ioCreate(TYPE_LIST_INTEGER, &l, sizeof(ItemList_t));
    p = ioExpand(key, size * sizeof(ULONG_T) + sizeof(ItemList_t));
    p = (ULONG_T *) ((ItemList_t *) p + 1);
    n = size;
    while (n != 0) {
	n -= 1;
	*p++ = 0;
    }
    return key;
}

ULONG_T
Count(ULONG_T list)
{

    if ((list & INDEX_MASK) == 0) {
	runError("attempt to Count nil");
	return 0;
    }

    return ((ItemList_t *) IOREAD(list))->l_elementCount;
}

static void
addHeadInt(ULONG_T list, ULONG_T valu)
{
    ItemList_t *l;
    ULONG_T *a, *b;
    UINT_T n;

    l = IOREAD(list);
    n = l->l_elementCount;
    if (n == 0xffff) {
	runError("attempt to AddHead to a full list");
	return;
    }

    n += 1;
    l = ioExpand(list, (ULONG_T) n * sizeof(ULONG_T) + sizeof(ItemList_t));
    l->l_elementCount = n;
    a = (ULONG_T *) (l + 1) + n;
    b = a - 1;
    while (B_TRUE) {
	n -= 1;
	if (n == 0) {
	    break;
	}
	*--a = *--b;
    }
    *b = valu;
}

void
AddHead(ULONG_T list, ULONG_T valu)
{

    TRACE("AddHead", list);
    if ((list & INDEX_MASK) != 0) {
	switch (list >> TYPE_SHIFT) {
	case TYPE_LIST_INTEGER:
	case TYPE_LIST_FIXED:
	    addHeadInt(list, valu);
	    break;
	case TYPE_LIST_THING:
	    if (valu == NIL_THING) {
		runError("attempt to AddHead nil to a list");
	    } else {
		addHeadInt(list, valu);
		incRef(valu);
	    }
	    break;
	case TYPE_LIST_ACTION:
	    if (valu == NIL_ACTION) {
		runError("attempt to AddHead nil to a list");
	    } else {
		addHeadInt(list, valu);
		incProcRef(valu);
	    }
	    break;
	default:
	    runError("invalid list passed to AddHead");
	    break;
	}
    } else {
	runError("attempt to AddHead to nil list");
    }
}

static ULONG_T
remHeadOne(ULONG_T list)
{
    ItemList_t *l;
    ULONG_T *a, *b;
    ULONG_T key;
    UINT_T n, t;

    l = IOREAD(list);
    n = l->l_elementCount;
    if (n == 0) {
	runError("attempt to RemHead from an empty list");
	return 0;
    }

    n -= 1;
    l->l_elementCount = n;
    a = (ULONG_T *) (l + 1);
    key = *a;
    if (n != 0) {
	b = a + 1;
	t = n;
	while (t != 0) {
	    t -= 1;
	    *a++ = *b++;
	}
    }
    ioShrink(list, n * sizeof(ULONG_T) + sizeof(ItemList_t));
    return key;
}

void
RemHead(ULONG_T list)
{
    ULONG_T key;

    TRACE("RemHead", list);
    if ((list & INDEX_MASK) == 0) {
	runError("attempt to RemHead from nil list");
	return;
    }

    switch (list >> TYPE_SHIFT) {
    case TYPE_LIST_INTEGER:
    case TYPE_LIST_FIXED:
	(void) remHeadOne(list);
	break;
    case TYPE_LIST_THING:
	key = remHeadOne(list);
	if (key != 0) {
	    decRef(key);
	}
	break;
    case TYPE_LIST_ACTION:
	key = remHeadOne(list);
	if (key != 0) {
	    decProcRef(key);
	}
	break;
    }
}

static void
addTailInt(ULONG_T list, ULONG_T valu)
{
    ItemList_t *l;
    UINT_T len;

    l = IOREAD(list);
    len = l->l_elementCount;
    if (len == 0xffff) {
	runError("attempt to AddTail to a full list");
	return;
    }

    l = ioExpand(list, (len + 1) * sizeof(ULONG_T) + sizeof(ItemList_t));
    l->l_elementCount = len + 1;
    *((ULONG_T *) (l + 1) + len) = valu;
}

void
xAddTail(ULONG_T list, ULONG_T valu)
{

    TRACE("AddTail", list);
    if ((list & INDEX_MASK) == 0) {
	runError("attempt to AddTail to a nil list");
	return;
    }

    switch (list >> TYPE_SHIFT) {
    case TYPE_LIST_INTEGER:
    case TYPE_LIST_FIXED:
	addTailInt(list, valu);
	break;
    case TYPE_LIST_THING:
	if (valu == NIL_THING) {
	    runError("attempt to AddTail nil thing to a list");
	} else {
	    addTailInt(list, valu);
	    incRef(valu);
	}
	break;
    case TYPE_LIST_ACTION:
	if (valu == NIL_ACTION) {
	    runError("attempt to AddTail nil action to a list");
	} else {
	    addTailInt(list, valu);
	    incProcRef(valu);
	}
	break;
    }
}

static ULONG_T
remTailOne(ULONG_T list)
{
    ItemList_t *l;
    ULONG_T key;
    UINT_T len;

    l = IOREAD(list);
    len = l->l_elementCount;
    if (len == 0) {
	runError("attempt to RemTail from an empty list");
	return 0;
    }

    len -= 1;
    l->l_elementCount = len;
    key = *((ULONG_T *) (l + 1) + len);
    ioShrink(list, (ULONG_T) len * sizeof(ULONG_T) + sizeof(ItemList_t));
    return key;
}

void
RemTail(ULONG_T list)
{
    ULONG_T key;

    TRACE("RemTail", list);
    if ((list & INDEX_MASK) == 0) {
	runError("attempt to RemTail from nil list");
    }

    switch (list >> TYPE_SHIFT) {
    case TYPE_LIST_INTEGER:
    case TYPE_LIST_FIXED:
	(void) remTailOne(list);
	break;
    case TYPE_LIST_THING:
	key = remTailOne(list);
	if (key != 0) {
	    decRef(key);
	}
	break;
    case TYPE_LIST_ACTION:
	key = remTailOne(list);
	if (key != 0) {
	    decProcRef(key);
	}
	break;
    }
}

ULONG_T
GetIntElement(ULONG_T list, ULONG_T which)
{
    ItemList_t *l;

    TRACE("GetIntElement", list);
    if (list == NIL_LIST_INTEGER) {
	runError("attempt to get integer element from nil");
	return 0;
    }

    l = IOREAD(list);
    if (l->l_elementCount <= which) {
	runError("attempt to get integer element beyond size of list");
	return 0;
    }

    return *((ULONG_T *) (l + 1) + which);
}

ULONG_T
GetFixedElement(ULONG_T list, ULONG_T which)
{
    ItemList_t *l;

    TRACE("GetFixedElement", list);
    if (list == NIL_LIST_FIXED) {
	runError("attempt to get fixed element from nil");
	return 0;
    }

    l = IOREAD(list);
    if (l->l_elementCount <= which) {
	runError("attempt to get fixed element beyond size of list");
	return 0;
    }

    return *((ULONG_T *) (l + 1) + which);
}

ULONG_T
GetThingElement(ULONG_T list, ULONG_T which)
{
    ItemList_t *l;

    TRACE("GetThingElement", list);
    if (list == NIL_LIST_THING) {
	runError("attempt to get thing element from nil");
	return NIL_THING;
    }

    l = IOREAD(list);
    if (l->l_elementCount <= which) {
	runError("attempt to get thing element beyond size of list");
	return NIL_THING;
    }

    return *((ULONG_T *) (l + 1) + which);
}

ULONG_T
GetActionElement(ULONG_T list, ULONG_T which)
{
    ItemList_t *l;

    TRACE("GetActionElement", list);
    if (list == NIL_LIST_ACTION) {
	runError("attempt to get action element from nil");
	return NIL_ACTION;
    }

    l = IOREAD(list);
    if (l->l_elementCount <= which) {
	runError("attempt to get action element beyond size of list");
	return NIL_ACTION;
    }

    return *((ULONG_T *) (l + 1) + which);
}

void
PutIntElement(ULONG_T list, ULONG_T which, ULONG_T valu)
{
    ItemList_t *l;

    TRACE("PutIntElement", list);
    if ((list & INDEX_MASK) == 0) {
	runError("attempt to put integer element to nil list");
	return;
    }

    l = IOREAD(list);
    if (l->l_elementCount <= which) {
	runError("attempt to put integer element beyond size of list");
	return;
    }

    *((ULONG_T *) (l + 1) + which) = valu;
    ioDirty(list);
}

void
PutFixedElement(ULONG_T list, ULONG_T which, ULONG_T valu)
{
    ItemList_t *l;

    TRACE("PutFixedElement", list);
    if ((list & INDEX_MASK) == 0) {
	runError("attempt to put fixed element to nil list");
	return;
    }

    l = IOREAD(list);
    if (l->l_elementCount <= which) {
	runError("attempt to put fixed element beyond size of list");
	return;
    }

    *((ULONG_T *) (l + 1) + which) = valu;
    ioDirty(list);
}

void
PutThingElement(ULONG_T list, ULONG_T which, ULONG_T valu)
{
    ItemList_t *l;
    ULONG_T *p;
    ULONG_T oldKey;

    TRACE("PutThingElement", list);
    if (list == NIL_LIST_THING) {
	runError("attempt to put thing element to nil list");
	return;
    }

    if (valu == NIL_THING) {
	runError("attempt to put thing element nil to a list");
	return;
    }

    l = IOREAD(list);
    if (l->l_elementCount <= which) {
	runError("attempt to put thing element beyond size of list");
	return;
    }

    p = (ULONG_T *) (l + 1) + which;
    oldKey = *p;
    if (oldKey != valu) {
	*p = valu;
	ioDirty(list);
	/* Do the incRef first! */
	incRef(valu);
	decRef(oldKey);
    }
}

void
PutActionElement(ULONG_T list, ULONG_T which, ULONG_T valu)
{
    ItemList_t *l;
    ULONG_T *p;
    ULONG_T oldKey;

    TRACE("PutActionElement", list);
    if (list == NIL_LIST_ACTION) {
	runError("attempt to put action element to nil list");
	return;
    }

    if (valu == NIL_THING) {
	runError("attempt to put action element nil to list");
	return;
    }

    l = IOREAD(list);
    if (l->l_elementCount <= which) {
	runError("attempt to put action element beyond size of list");
	return;
    }

    p = (ULONG_T *) (l + 1) + which;
    oldKey = *p;
    if (oldKey != valu) {
	*p = valu;
	ioDirty(list);
	incProcRef(valu);
	decProcRef(oldKey);
    }
}

LONG_T
FindElement(ULONG_T list, ULONG_T valu)
{
    ItemList_t *l;
    ULONG_T *p;
    UINT_T count, n;

    TRACE("FindElement", list);
    if ((list & INDEX_MASK) != 0) {
	l = IOREAD(list);
	count = l->l_elementCount;
	p = (ULONG_T *) (l + 1);
	n = 0;
	while (n != count) {
	    if (*p == valu) {
		return n;
	    }
	    ++p;
	    n += 1;
	}
    } else {
	runError("nil list passed to FindElement");
    }
    return -1;
}

static BOOL_T
delOneElement(ULONG_T list, ULONG_T valu)
{
    ItemList_t *l;
    ULONG_T *p, *q;
    UINT_T count, newCount;

    l = IOREAD(list);
    count = l->l_elementCount;
    if (count == 0) {
	return B_FALSE;
    }

    newCount = count;
    p = (ULONG_T *) (l + 1);
    while (*p != valu) {
	count -= 1;
	if (count == 0) {
	    /* this is not an error */
	    return B_FALSE;
	}
	++p;
    }
    newCount -= 1;
    l->l_elementCount = newCount;
    q = p + 1;
    while (B_TRUE) {
	count -= 1;
	if (count == 0) {
	    break;
	}
	*p++ = *q++;
    }
    ioShrink(list, (ULONG_T) newCount * sizeof(ULONG_T) + sizeof(ItemList_t));
    return B_TRUE;
}

void
DelElement(ULONG_T list, ULONG_T valu)
{

    TRACE("DelElement", list);
    if ((list & INDEX_MASK) == 0) {
	runError("attempt to DelElement from nil list");
	return;
    }

    switch (list >> TYPE_SHIFT) {
    case TYPE_LIST_INTEGER:
    case TYPE_LIST_FIXED:
	(void) delOneElement(list, valu);
	break;
    case TYPE_LIST_THING:
	if (delOneElement(list, valu)) {
	    decRef(valu);
	}
	break;
    case TYPE_LIST_ACTION:
	if (delOneElement(list, valu)) {
	    decProcRef(valu);
	}
	break;
    }
}

/*
 * nukeThing - the actual work of going through and clearing out the
 *	properties of a thing.
 */

void
nukeThing(ULONG_T key)
{
    Thing_t *th;
    Attr_t *at;
    ULONG_T valu;
    UINT_T count, i, typ;

    th = IOREAD(key);
    count = th->th_propCount;
    th->th_propCount = 0;
    ioDirty(key);
    /* Use 1 upto count in case count is 0. */
    for (i = 1; i <= count; i += 1) {
	/* refetch it each time to make sure it stays in the cache */
	th = IOREAD(key);
	at = (Attr_t *) (th + 1) + (i - 1);
	valu = at->at_value;
	typ = at->at_prop >> TYPE_SHIFT;
	decRef(at->at_prop);
	switch (typ) {
	case TYPE_PROP_THING:
	case TYPE_PROP_TABLE:
	case TYPE_PROP_GRAMMAR:
	    decRef(valu);
	    break;
	case TYPE_PROP_ACTION:
	    decProcRef(valu);
	    break;
	case TYPE_PROP_STRING:
	    ioDelete(valu);
	    break;
	case TYPE_PROP_LIST_INTEGER:
	case TYPE_PROP_LIST_FIXED:
	case TYPE_PROP_LIST_THING:
	case TYPE_PROP_LIST_ACTION:
	    freeList(valu, typ - TYPE_PROP_LIST_INTEGER + TYPE_LIST_INTEGER);
	    break;
	}
    }
}

/*
 * ClearThing - nuke a thing, by removing all of its properties.
 *
 *	BIG NOTE: in this and most of the other routines that fiddle with
 *	things that have been brought into the cache, we make no attempt to
 *	make sure that they stay at the same place in the cache as we go
 *	around fiddling with them. THE CACHE HAD BETTER BE BIG ENOUGH!
 *
 *	User note: when calling ClearThing, make sure that the thing still
 *	exists when you call this routine, since a thing with no properties
 *	will automatically be destroyed when the last reference to it goes away
 *
 *	Security note: This routine violates the normal security mechanisms,
 *	since it allows the deletion of all properties from the thing,
 *	whether that is normally and individually allowed or not!
 */

void
ClearThing(ULONG_T key)
{
    Thing_t *th;

    TRACE("ClearThing", key);
    if (key == NIL_THING) {
	runError("nil thing passed to ClearThing");
	return;
    }

    th = IOREAD(key);
    if (th->th_owner != EffectivePlayer && th->th_owner != RealPlayer &&
	EffectivePlayer != SYSADMIN_KEY)
    {
	runError("attempt to ClearThing someone else's thing");
	return;
    }

    /* make sure we do not recurse around through it */
    incRef(key);
    nukeThing(key);
    ioShrink(key, sizeof(Thing_t));
    /* This call can also destroy it if there were no references
       when 'ClearThing' was called. */
    decRef(key);
}

#if 0

void
CheckList(ULONG_T list, CHAR_T *where)
{
    ItemList_t *l;
    ULONG_T *p;
    UINT_T count, n;

    TRACE("CheckList", list);
    if ((list & INDEX_MASK) == 0) {
	runError("nil list passed to CheckList");
	return;
    }

    l = IOREAD(list);
    count = l->l_elementCount;
    p = (ULONG_T *) (l + 1);
    n = 0;
    while (n != count) {
	if ((*p & TYPE_MASK) != TYPE_THING << TYPE_SHIFT) {
	    log(where);
	    logX(": non-thing found in list ", list, " value ");
	    logX("", *p, "\n");
	    abortNoWrite();
	}
	++p;
	n += 1;
    }
}

#endif
