#include <unistd.h>

#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <ctype.h>

#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDReq.h"
#define EXTERN
#include "MUDServ.h"
#include "ClientStuff.h"
#include "telnetFlags.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * main1.c - initial startup stuff, exit handling, error handling, etc.
 */

#define DEFAULT_PATH	""		/* default to current directory */

#define DEFAULT_CACHE_SIZE	500000		/* default FileIO cache size */
#define DEFAULT_MEMORY_LIMIT	1000000000	/* def allocated mem limit */
#define MINIMUM_CACHE_SIZE	32768
#define MINIMUM_MEMORY_LIMIT	50000
#define DEFAULT_REQUEST_LIMIT	100
#define DEFAULT_BINARY_PORT	16667
#define OUTPUT_BUFFER_SIZE	128

#define ERROR_TO_MESS		B_FALSE

static UINT_T ErrorPos, MessagePos;
static BOOL_T AskForRetry;
static char ErrorBuffer[OUTPUT_BUFFER_SIZE],
    MessageBuffer[OUTPUT_BUFFER_SIZE];

/*
 * errorFlush - flush the error buffer to the log file.
 */

static void
errorFlush(void)
{

    if (ErrorPos != 0) {
	ErrorBuffer[ErrorPos] = '\0';
	log(&ErrorBuffer[0]);
	ErrorPos = 0;
    }
}

/*
 * errorChar - add a single character to the error output stream.
 */

void
errorChar(char ch)
{

#if ERROR_TO_MESS
    messChar(ch);
#else
    ErrorBuffer[ErrorPos] = ch;
    ErrorPos += 1;
    if (ch == '\n' || ErrorPos == OUTPUT_BUFFER_SIZE - 1) {
	errorFlush();
    }
#endif
}

/*
 * messChar - bottom level of debugging/complaint output code.
 *
 */

void
messChar(char ch)
{

    MessageBuffer[MessagePos] = ch;
    MessagePos += 1;
    if (MessagePos == OUTPUT_BUFFER_SIZE - 1) {
	MessageBuffer[MessagePos] = '\n';
	MessagePos = OUTPUT_BUFFER_SIZE;
	ch = '\n';
    }
    if (ch == '\n') {
#ifdef AMIGA
	/* *MUCH* to my surprise, the SAS C library version of 'write'
	   does not have any pre-opened units! */
	(void) fwrite(&MessageBuffer[0], 1, MessagePos, stdout);
#else
	(void) write(1, &MessageBuffer[0], MessagePos);
#endif
	MessagePos = 0;
    }
}

/*
 * mess - simple routine to write a message to the operator.
 */

void
mess(char *p)
{

    while (*p != '\0') {
	messChar(*p++);
    }
}

/*
 * mHex - write a value in hex. (only during debugging)
 */

void
mHex(ULONG_T n)
{
    char *p;
    UINT_T nibble;
    int i;
    char buff[9];

    p = &buff[8];
    *p = '\0';
    for (i = 0; i != 8; i += 1) {
	nibble = n & 0xf;
	*--p = (nibble >= 10) ? (nibble - 10 + 'a') : (nibble + '0');
	n >>= 4;
    }
    mess(p);
}

/*
 * mNum - print a decimal number.
 */

void
mNum(ULONG_T n)
{
    char buffer[11];
    char *p;

    p = &buffer[10];
    *p = '\0';
    do {
	*--p = n % 10 + '0';
	n /= 10;
    } while (n != 0);
    mess(p);
}

/*
 * checkForRetry - called from various places that would normally abort,
 *	to see if the user can intervene to help out.
 *	Return 'B_TRUE' if user says to try again.
 */

BOOL_T
checkForRetry(char *message)
{
    /* Not much we can do in a generic situation, so just return B_FALSE.
       Custom implementations can put up a requestor or something. */

    return B_FALSE;
}

/*
 * finalExit - the final steps in exitting.
 */

void
finalExit(void)
{
    /* Nothing left to do here, generically. On specific implementations,
       might want to release some resources, perhaps close a message
       window, etc. */

    exit(1);
}

/*
 * getNumber - parse a number from the passed string. No error checking.
 */

static ULONG_T
getNumber(char **pStr)
{
    char *str;
    ULONG_T n;

    str = *pStr;
    n = 0;
    while (isdigit(*str)) {
	n = n * 10 + (*str - '0');
	++str;
    }
    *pStr = str;
    return n;
}

/*
 * startup2 - next steps in startup.
 */

static void
startup2(void)
{
    char *p;

    p = &Path[0];
    if (*p != '\0') {
	while (*p != '\0') {
	    ++p;
	}
	--p;
	if (*p != ':' && *p != '/') {
	    *++p = '/';
	    *++p = '\0';
	}
    }
    if (Verbose) {
	mess("Using path '");
	mess(&Path[0]);
	mess("'.\n");
    }
    runServer();
}

/*
 * startup - continue with startup.
 */

static void
startup(int argc, char **argv)
{
    char *par;
    ULONG_T n;
    char ch;
    BOOL_T hadError;
    int port, flags;

    utilInit();
    TRACE("startup", 0);

    strCpy(&Path[0], DEFAULT_PATH);
    ForceWizardMode = B_FALSE;
#if DEBUG_PRESENT
    WantDebug = B_FALSE;
#endif
    CacheSize = DEFAULT_CACHE_SIZE;
    MemoryLimit = DEFAULT_MEMORY_LIMIT;
    RequestLimit = DEFAULT_REQUEST_LIMIT;
    BinaryPort = DEFAULT_BINARY_PORT;
    AskForRetry = B_TRUE;
    NoMachines = B_FALSE;
    ForceNoDelay = B_FALSE;
    Verbose = B_TRUE;

    ErrorPos = 0;
    messageInit();

    hadError = B_FALSE;
    while (--argc > 0) {
	par = *++argv;
	if (*par == '-') {
	    ++par;
	    while (B_TRUE) {
		ch = *par++;
		if (ch == '\0') {
		    break;
		}
		switch (ch) {
		case 'W':
		    ForceWizardMode = B_TRUE;
		    break;
		case 'Q':
		    Verbose = B_FALSE;
		    break;
		case 'D':
		    if (DEBUG_PRESENT) {
			WantDebug = B_TRUE;
		    }
		    break;
		case 'N':
		    AskForRetry = B_FALSE;
		    break;
		case 'c':
		    n = getNumber(&par);
		    if (n < MINIMUM_CACHE_SIZE) {
			n = MINIMUM_CACHE_SIZE;
		    }
		    CacheSize = n;
		    break;
		case 'm':
		    n = getNumber(&par);
		    if (n < MINIMUM_MEMORY_LIMIT) {
			n = MINIMUM_MEMORY_LIMIT;
		    }
		    MemoryLimit = n;
		    break;
		case 'r':
		    RequestLimit = getNumber(&par);
		    break;
		case 'b':
		    BinaryPort = getNumber(&par);
		    break;
		case 't':
		    port = getNumber(&par);
		    if (*par == ':') {
			++par;
			if (*par == 't') {
			    flags = DEFAULT_TELNET_FLAGS;
			    ++par;
			} else if (*par == 's') {
			    flags = SIMPLE_TELNET_FLAGS;
			    ++par;
			} else if (*par == 'l') {
			    flags = LPMUD_TELNET_FLAGS;
			    ++par;
			} else {
			    flags = getNumber(&par);
			}
		    } else {
			flags = DEFAULT_TELNET_FLAGS;
		    }
		    if (*par != '\0') {
			mess("Invalid telnet port/flag\n");
			mess("Use -t# or -t#:t or -t#:s or -t#:l or t#:#\n");
			hadError = B_TRUE;
		    } else {
			addTelnetPort(port, flags);
		    }
		    break;
		case 'M':
		    NoMachines = B_TRUE;
		    break;
		case 'F':
		    ForceNoDelay = B_TRUE;
		    break;
		default:
		    hadError = B_TRUE;
		    break;
		}
	    }
	} else {
	    strCpy(&Path[0], par);
	}
    }
    if (hadError) {
#if DEBUG_PRESENT
	mess("Use is: run MUDServ [-WQDNcmrbtMF] [path]\n");
#else
	mess("Use is: run MUDServ [-WQNcmrbtMF] [path]\n");
#endif
    } else {
	startup2();
    }
}

/*
 * main - entry point for MUD server.
 */

int
main(int argc, char **argv)
{

    AlreadyAborting = B_FALSE;
    startup(argc, argv);
    return 0;
}
