#include <ctype.h>

#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * match.c - routines to do parsing-related matching stuff.
 */

#define TAIL_MAX	100

static CHAR_T TailBuffer[TAIL_MAX + 1];
static UINT_T TailTotal, TailPos;

static ULONG_T FoundThing;

/*
 * matchInit - initialize match processing.
 */

void
matchInit(void)
{

    TailTotal = 0;
    TailPos = 0;
    FoundThing = NIL_THING;
}

/*
 * stuffTail - called from parser to stuff a VerbTail tail.
 */

CHAR_T *
stuffTail(CHAR_T *p)
{
    CHAR_T *q;
    CHAR_T ch;
    UINT_T len;

    q = &TailBuffer[0];
    len = 0;
    while (B_TRUE) {
	ch = *p;
	if (sentEnd(ch)) {
	    break;
	}
	if (ch == '\"') {
	    /* swallow a quoted string */
	    if (len != TAIL_MAX) {
		*q++ = ch;
		len += 1;
	    }
	    ++p;
	    while (B_TRUE) {
		ch = *p;
		if (ch == '\"' || ch == '\0') {
		    break;
		}
		if (len != TAIL_MAX) {
		    *q++ = ch;
		    len += 1;
		}
		++p;
	    }
	    if (ch != '\0') {
		if (len != TAIL_MAX) {
		    *q++ = ch;
		    len += 1;
		}
		++p;
	    }
	} else {
	    if (len != TAIL_MAX) {
		*q++ = ch;
		len += 1;
	    }
	    ++p;
	}
    }
    TailPos = 0;
    TailTotal = len;
    return p;
}

/*
 * GetTail - return the current contents of TailBuffer as a string.
 */

CHAR_T *
GetTail(void)
{
    CHAR_T *p, *q, *st;
    UINT_T len;

    len = TailTotal - TailPos + 1;
    p = MEM_ALLOC_E(&MS, len, "tail");
    st = p;
    q = &TailBuffer[TailPos];
    while (B_TRUE) {
	len -= 1;
	if (len == 0) {
	    break;
	}
	*p++ = *q++;
    }
    *p = '\0';
    return st;
}

/*
 * GetWord - pull a word out of TailBuffer, and move up TailPos accordingly.
 */

CHAR_T *
GetWord(void)
{
    CHAR_T *p, *q, *st;
    UINT_T total, pos;

    TRACE("GetWord", 0);
    total = TailTotal;
    pos = TailPos;
    q = &TailBuffer[pos];
    if (pos != total && *q == '\"') {
	/* a quoted string - return the whole thing */
	pos += 1;
	++q;
	p = q;
	while (pos != total && *p != '\"') {
	    pos += 1;
	    ++p;
	}
	pos -= TailPos; 	/* includes the quote, so also the \0 */
	if (*p != '\"') {
	    pos += 1;
	}
	p = MEM_ALLOC_E(&MS, pos, "gwd1");
	st = p;
	pos = TailPos + 1;	/* skip the opening quote */
	while (pos != total && *q != '\"') {
	    pos += 1;
	    *p++ = *q++;
	}
	if (pos != total) {
	    pos += 1;
	    ++q;
	}
    } else {
	while (pos != total && *q != ' ') {
	    pos += 1;
	    ++q;
	}
	pos -= TailPos;
	p = MEM_ALLOC_E(&MS, pos + 1, "gwd2");
	st = p;
	pos = TailPos;
	q = &TailBuffer[pos];
	while (pos != total && *q != ' ') {
	    pos += 1;
	    *p++ = *q++;
	}
    }
    *p = '\0';
    while (pos != total && *q == ' ') {
	pos += 1;
	++q;
    }
    TailPos = pos;
    return st;
}

/*
 * SetTail - allow a caller to set a string into TailBuffer.
 */

void
SetTail(CHAR_T *p)
{
    CHAR_T *q;
    UINT_T total;

    TRACE("SetTail", p);
    while (*p == ' ') {
	++p;
    }
    total = 0;
    q = &TailBuffer[0];
    while (*p != '\0' && total != TAIL_MAX) {
	*q++ = *p++;
	total += 1;
    }
    TailPos = 0;
    TailTotal = total;
}

/*
 * sayWhisper - common code for SetSay and SetWhisperMe.
 */

static CHAR_T *
sayWhisper(CHAR_T *what, CHAR_T *separator)
{
    CHAR_T *p, *q;
    UINT_T len, total;
    BOOL_T ok;

    ok = B_FALSE;
    p = what;
    if (*p != '\0' && *p != ' ') {
	len = 1;
	while (*p != ' ' && *p != '\0') {
	    ++p;
	    len += 1;
	}
	if (*p != '\0') {
	    ++p;
	    if (*p != ' ' && *p != '\0') {
		q = separator;
		while (*p == *q && *q != '\0') {
		    ++p;
		    ++q;
		}
		if (*q == '\0' && *p != '\0' && *p != ' ') {
		    q = &TailBuffer[0];
		    total = 0;
		    while (*p != '\0' && total != TAIL_MAX) {
			*q++ = *p++;
			total += 1;
		    }
		    TailTotal = total;
		    TailPos = 0;
		    p = MEM_ALLOC_E(&MS, len, "wsp1");
		    q = what;
		    what = p;
		    while (*q != ' ') {
			*p++ = *q++;
		    }
		    *p = '\0';
		    ok = B_TRUE;
		}
	    }
	}
    }
    if (ok) {
	return what;
    } else {
	return nullString();
    }
}

/*
 * SetSay - break apart a say string, using TailBuffer.
 */

CHAR_T *
SetSay(CHAR_T *what)
{

    TRACE("SetSay", what);
    return sayWhisper(what, "says: ");
}

/*
 * SetWhisperMe - break apart a whisper string, using TailBuffer.
 *	Expected form: X whispers: Y
 *	string X is returned, Y is left in TailBuffer.
 */

CHAR_T *
SetWhisperMe(CHAR_T *what)
{

    TRACE("SetWhisperMe", what);
    return sayWhisper(what, "whispers: ");
}

/*
 * SetWhisperOther - break apart a whisper string, using TailBuffer.
 *	Expected form:	X whispers to Y: Z
 *	string X is returned, and Y Z is left in TailBuffer.
 */

CHAR_T *
SetWhisperOther(CHAR_T *what)
{
    CHAR_T *p, *q;
    UINT_T len, total;
    BOOL_T ok;

    TRACE("SetWhisperOther", what);
    ok = B_FALSE;
    p = what;
    if (*p != '\0' && *p != ' ') {
	len = 1;
	while (*p != ' ' && *p != '\0') {
	    ++p;
	    len += 1;
	}
	if (*p != '\0') {
	    ++p;
	    if (*p != ' ' && *p != '\0') {
		q = "whispers to ";
		while (*p == *q && *q != '\0') {
		    ++p;
		    ++q;
		}
		if (*q == '\0' && *p != '\0' && *p != ' ' && *p != '\"') {
		    q = p;
		    while (*q != ' ' && *q != '\0') {
			++q;
		    }
		    --q;
		    if (*q == ':') {
			++q;
			if (*q == ' ') {
			    while (*q == ' ') {
				++q;
			    }
			    if (*q != '\0') {
				q = &TailBuffer[0];
				total = 0;
				while (*p != '\0' && total != TAIL_MAX &&
				    *p != ':')
				{
				    *q++ = *p++;
				    total += 1;
				}
				/* do NOT copy in the : after the target */
				++p;
				while (*p != '\0' && total != TAIL_MAX) {
				    *q++ = *p++;
				    total += 1;
				}
				TailTotal = total;
				TailPos = 0;
				p = MEM_ALLOC_E(&MS, len, "wsp2");
				q = what;
				what = p;
				while (*q != ' ') {
				    *p++ = *q++;
				}
				*p = '\0';
				ok = B_TRUE;
			    }
			}
		    }
		}
	    }
	}
    }
    if (ok) {
	return what;
    } else {
	return nullString();
    }
}

/*
 * HasAdjective - return B_TRUE if (the given standard-form name string has
 *	the specified adjective in it.
 */

LONG_T
HasAdjective(CHAR_T *name, CHAR_T *adjective)
{
    CHAR_T *adj;
    CHAR_T ch1, ch2;

    /* note: assume no alternate forms */
    while (*name != ';' && *name != '\0') {
	++name;
    }
    if (*name == ';') {
	++name;
	while (*name != '\0') {
	    adj = adjective;
	    while (B_TRUE) {
		ch1 = *name;
		if (isupper(ch1)) {
		    ch1 = tolower(ch1);
		}
		ch2 = *adj;
		if (isupper(ch2)) {
		    ch2 = tolower(ch2);
		}
		if (ch1 != ch2 || ch1 == ',' || ch1 == '\0') {
		    break;
		}
		++name;
		++adj;
	    }
	    if (ch2 == '\0' && (ch1 == ',' || ch1 == '\0')) {
		return BOOLEAN_TRUE;
	    }
	    while (*name != ',' && *name != '\0') {
		++name;
	    }
	    if (*name == ',') {
		++name;
	    }
	}
    }
    return BOOLEAN_FALSE;
}

/*
 * MatchName - match two standard-form name strings. Return the index of the
 *	portion of the first argument that the second matches, else -1.
 *	Syntax of the parsed form:
 *		noun[;adjective{,adjective}]
 *	Syntax of the stored form:
 *		<alternative> := noun{,noun}[;adjective{,adjective}]
 *		<stored> := <alternative>{.alternative}
 */

LONG_T
MatchName(CHAR_T *stored, CHAR_T *saveParsed)
{
    CHAR_T *parsed, *p, *q;
    CHAR_T ch1, ch2, chT;
    ULONG_T which;
    BOOL_T matched;

    TRACE("MatchName", saveParsed);

    if (*saveParsed == '\0') {
	return -1;
    }

    which = 0;

    /* outer loop - tries the parsed form against each alternative of the
       stored form. */

    while (*stored != '\0') {

	/* first, check the noun parts - loop through the various synonyms
	   in the stored form. */

	while (B_TRUE) {
	    parsed = saveParsed;
	    while (B_TRUE) {
		ch1 = *stored;
		if (isupper(ch1)) {
		    ch1 = tolower(ch1);
		}
		ch2 = *parsed;
		if (isupper(ch2)) {
		    ch2 = tolower(ch2);
		}
		if (ch1 != ch2 || ch2 == ';' || ch2 == '\0') {
		    break;
		}
		++stored;
		++parsed;
	    }

	    /* special hack - ignore an extra 's' on either (won't be both!) */
	    chT = *(stored + 1);
	    if (ch1 == 's' && (ch2 == '\0' || ch2 == ';') &&
		(chT == ';' || chT == '\0' || chT == ',' || chT == '.'))
	    {
		++stored;
		ch1 = *stored;
		if (isupper(ch1)) {
		    ch1 = tolower(ch1);
		}
	    }
	    chT = *(parsed + 1);
	    if (ch2 == 's' && (chT == '\0' || chT == ';') &&
		(ch1 == ';' || ch1 == '\0' || ch1 == ',' || ch1 == '.'))
	    {
		++parsed;
		ch2 = *parsed;
		if (isupper(ch2)) {
		    ch2 = tolower(ch2);
		}
	    }

	    if ((ch2 == '\0' || (ch2 == ';' && *(parsed + 1) == '\0')) &&
		(ch1 == ';' || ch1 == '\0' || ch1 == ',' || ch1 == '.'))
	    {
		/* end of parsed form, and end of a noun synonym in the
		   stored form - we have a total match.
		   NOTE: this implies that if we get into the adjective
		   matching code below, then adjectives DO exist in the
		   parsed form. */
		return (LONG_T) which;
	    }

	    if (ch2 == ';' &&
		(ch1 == ';' || ch1 == '\0' || ch1 == ',' || ch1 == '.'))
	    {
		/* we made it to the end of both nouns, so they have matched,
		   we do NOT continue this inner noun-synonym loop */
		matched = B_TRUE;
		break;
	    } else {
		/* not at the end of something - move to the end of the stored
		   form and see if there are any noun synonyms to try */
		matched = B_FALSE;
		while (ch1 != ';' && ch1 != '\0' && ch1 != ',' && ch1 != '.')
		{
		    ++stored;
		    ch1 = *stored;
		}
		if (ch1 == '\0') {
		    /* ran out of everything in the stored form - fail */
		    return -1;
		}
		/* loop here only if have another noun synonym */
		if (ch1 != ',') {
		    break;
		}
	    }
	    /* skip the ',' */
	    ++stored;
	}

	/* we have either matched, or come to the end of the noun synonyms.
	   If we have matched, skip any remaining noun synonyms, and then
	   go try the adjectives. */

	if (matched) {
	    /* the nouns matched - skip the extra stuff */
	    ++parsed;		/* skip ';' */
	    while (ch1 != ';' && ch1 != '\0' && ch1 != '.') {
		++stored;
		ch1 = *stored;
	    }
	    if (ch1 == ';') {
		/* get to the adjectives in the stored form */
		++stored;
	    }

	    /* now, loop through the parsed form checking each adjective
	       against the ones in the stored form. */

	    while (B_TRUE) {
		/* loop through the adjectives in the stored form */
		q = stored;
		matched = B_FALSE;
		while (B_TRUE) {
		    p = parsed;
		    while (B_TRUE) {
			ch1 = *q;
			if (isupper(ch1)) {
			    ch1 = tolower(ch1);
			}
			ch2 = *p;
			if (isupper(ch2)) {
			    ch2 = tolower(ch2);
			}
			if (ch2 != ch1 || ch2 == '\0' || ch2 == ',') {
			    break;
			}
			++p;
			++q;
		    }
		    if ((ch2 == '\0' || ch2 == ',') &&
			(ch1 == '\0' || ch1 == ',' || ch1 == '.'))
		    {
			/* adjective match - exit this inner loop */
			matched = B_TRUE;
			break;
		    } else {
			/* no match of this parsed adjective against this
			   stored one. Go look for more stored adjectives
			   to try. */
			while (ch1 != '\0' && ch1 != ',' && ch1 != '.') {
			    ++q;
			    ch1 = *q;
			}
			/* we can try more stored forms if they exist */
			if (ch1 != ',') {
			    break;
			}
		    }
		    /* skip the ',' in the stored form */
		    ++q;
		}
		/* continue with the parsed form only if (we matched */
		if (matched) {
		    if (ch2 == '\0') {
			/* end of all parsed adjectives - return success */
			return (LONG_T) which;
		    }
		    /* keep looping through parsed adjectives */
		} else {
		    /* fail on this parsed adjective - have to exit the
		       loop and try another entire alternative. Skip the
		       '.' in stored form (if not at end) */
		    stored = q;
		    if (ch1 == '.') {
			++stored;
		    }
		    break;
		}
		/* skip the ',' in the parsed form */
		parsed = p + 1;
	    }
	    /* we need to try another entire alternative, since we did not
	       get an adjective match on this one. We could just return
	       failure here, on the assumption that no noun form will appear
	       in more than one entire alternative, but what the heck.
	       Actually, allowing the same noun in more than one alternative
	       is quite useful - see 'ColourMatch'. */

	} else {

	    /* we came to the end of a group of noun synonyms without finding
	       one that matched the parsed noun. Try to find an entire
	       alternative form in the stored form. */

	    while (ch1 != '\0' && ch1 != '.') {
		++stored;
		ch1 = *stored;
	    }
	    /* continue with entire alternatives if there are any */
	    if (ch1 == '.') {
		++stored;
	    }
	    /* otherwise we will exit the outer loop and return failure */
	}
	which += 1;
    }

    /* gone all through the stored string without a match - failed */
    return -1;
}

/*
 * SelectName - select the indicated name alternative.
 */

CHAR_T *
SelectName(CHAR_T *names, ULONG_T which)
{
    CHAR_T *q, *result;

    while (*names != '\0' && which != 0) {
	while (*names != '.' && *names != '\0') {
	    ++names;
	}
	if (*names == '.') {
	    ++names;
	}
	which -= 1;
    }
    q = names;
    which = 0;
    while (*q != '.' && *q != '\0') {
	++q;
	which += 1;
    }
    q = MEM_ALLOC_E(&MS, which + 1, "seln");
    result = q;
    while (*names != '.' && *names != '\0') {
	*q++ = *names++;
    }
    *q = '\0';
    return result;
}

/*
 * SelectWord - select one of a comma-separated list of words.
 */

CHAR_T *
SelectWord(CHAR_T *words, ULONG_T which)
{
    CHAR_T *q, *result;

    while (*words != '\0' && *words != '.' && which != 0) {
	while (*words != ',' && *words != '.' && *words != '\0') {
	    ++words;
	}
	if (*words == ',') {
	    ++words;
	}
	which -= 1;
    }
    q = words;
    which = 0;
    while (*q != ',' && *q != '.' && *q != '\0') {
	++q;
	which += 1;
    }
    q = MEM_ALLOC_E(&MS, which + 1, "selw");
    result = q;
    while (*words != ',' && *words != '.' && *words != '\0') {
	*q++ = *words++;
	words = words + sizeof(char);
    }
    *q = '\0';
    return result;
}

/*
 * FormatName - format a standard-form name into a pretty string form.
 */

CHAR_T *
FormatName(CHAR_T *name)
{
#define BUFF_LEN	256
    CHAR_T buffer[BUFF_LEN];
    CHAR_T countBuffer[8];
    CHAR_T *p, *q, *countPos;
    UINT_T len, count;

    TRACE("FormatName", name);
    p = name;

    count = 0;
    countPos = 0;	/* shut up gcc */
    if (isdigit(*p)) {
	q = p;
	while (isdigit(*p)) {
	    count *= 10;
	    count += (*p - '0');
	    ++p;
	}
	if (*p == ':') {
	    ++p;
	    countPos = &countBuffer[5];
	    *countPos = '\0';
	    do {
		*--countPos = count % 10 + '0';
		count /= 10;
	    } while (count != 0);
	    count = &countBuffer[7] - countPos;
	    name = p;
	} else {
	    count = 0;
	    p = q;
	}
    }

    len = 0;
    q = &buffer[0];
    while (*p != '\0' && *p != ';' && *p != '.') {
	++p;
    }
    if (*p == ';') {
	++p;
	while (*p != '\0' && *p != '.') {
	    while (*p != '\0' && *p != ',' && *p != '.') {
		if (len < BUFF_LEN) {
		    len += 1;
		    *q++ = *p;
		}
		++p;
	    }
	    if (len < BUFF_LEN) {
		len += 1;
		*q++ = ' ';
	    }
	    if (*p == ',') {
		++p;
	    }
	}
    }
    p = name;
    while (*p != '\0' && *p != ';' && *p != ',' && *p != '.') {
	if (len < BUFF_LEN) {
	    len += 1;
	    *q++ = *p;
	}
	++p;
    }
    p = MEM_ALLOC_E(&MS, len + count + 1, "fmtn");
    name = p;
    q = &buffer[0];
    while (len != 0) {
	len -= 1;
	*p++ = *q++;
    }
    if (count != 0) {
	*p++ = ' ';
	*p++ = '#';
	while (*countPos != '\0') {
	    *p++ = *countPos++;
	}
    }
    *p = '\0';
    return name;
#undef BUFF_LEN
}

/*
 * xFindName - try to find a matching name in a given thing list.
 */

ULONG_T
xFindName(ULONG_T originalThingList, ULONG_T nameProp, CHAR_T *name)
{
    Thing_t *th;
    Attr_t *at;
    ItemList_t *l;
    ULONG_T thingCount, thingParent;
    UINT_T n;
    BYTE_T propCount;
    CHAR_T ch;
    CHAR_T *saveName;
    ULONG_T thingList, thingLimit, thingChild, thingFound, prevChild;
    UINT_T wantCount, matchCount;
    BOOL_T countGiven, matchHere;

    thingList = originalThingList;
    TRACE("xFindName", thingList);
    FoundThing = NIL_THING;
    if (thingList == NIL_LIST_THING) {
	runError("nil list to FindName");
	return STATUS_FAIL;
    }
    if (nameProp == NIL_PROP_STRING) {
	runError("nil prop to FindName");
	return STATUS_FAIL;
    }
    countGiven = B_FALSE;
    wantCount = 0;	/* shut up gcc */
    saveName = name;

    ch = *name;
    if (isdigit(ch)) {
	n = 0;
	while (B_TRUE) {
	    ch = *name;
	    if (! isdigit(ch)) {
		break;
	    }
	    n *= 10;
	    n += (ch - '0');
	    ++name;
	}
	if (n != 0 && ch == ':') {
	    countGiven = B_TRUE;
	    wantCount = n;
	    ++name;
	} else {
	    name = saveName;
	}
    }

    prevChild = NIL_THING;	/* Hush up, compiler! */

    matchCount = 0;
    l = ioRead(thingList);
    thingLimit = l->l_elementCount;
    thingCount = 0;
    while (thingCount != thingLimit) {
	l = (ItemList_t *) ioRead(originalThingList) + 1;
	thingFound = *((ULONG_T *) l + thingCount);
	thingCount += 1;
	thingChild = thingFound;
	matchHere = B_FALSE;
	while (B_TRUE) {
	    th = ioRead(thingFound);
	    thingParent = th->th_parent;
	    propCount = th->th_propCount;
	    at = (Attr_t *) (th + 1);
	    while (propCount != 0 && at->at_prop != nameProp) {
		propCount -= 1;
		++at;
	    }
	    if (propCount != 0) {
		if (MatchName(ioRead(at->at_value), name) != -1) {
		    matchCount += 1;
		    if (countGiven) {
			if (matchCount == wantCount) {
			    FoundThing = thingChild;
			    return STATUS_SUCCEED;
			}
		    } else {
			if (matchCount > 1) {
			    FoundThing = prevChild;
			    return STATUS_CONTINUE;
			}
			prevChild = thingChild;
		    }
		    matchHere = B_TRUE;
		}
	    }
	    if (matchHere || thingParent == NIL_THING) {
		break;
	    }
	    thingFound = thingParent;
	}
    }
    if (matchCount == 1 && ! countGiven) {
	FoundThing = prevChild;
	return STATUS_SUCCEED;
    }
    return STATUS_FAIL;
}

/*
 * setResult - set the thing to be found.
 */

void
setResult(ULONG_T theThing)
{

    FoundThing = theThing;
}

/*
 * FindResult - return the found thing.
 */

ULONG_T
FindResult(void)
{
    ULONG_T res;

    res = FoundThing;
    FoundThing = NIL_THING;
    return res;
}

/*
 * Strip - utility to unquote and unescape a string.
 */

CHAR_T *
Strip(CHAR_T *st)
{
    CHAR_T *p, *res;
    UINT_T len;
    BOOL_T hadQuote;

    p = &TailBuffer[0];
    len = 0;
    if (*st == '\"') {
	hadQuote = B_TRUE;
	++st;
    } else {
	hadQuote = B_FALSE;
    }
    while (*st != '\0') {
	if (*st == '\\') {
	    ++st;
	    if (len != TAIL_MAX) {
		len += 1;
		switch (*st) {
		case 'n':
		    *p++ = '\n';
		    ++p;
		    break;
		case 't':
		    *p++ = '\t';
		    ++p;
		    break;
		default:
		    *p++ = *st;
		}
	    }
	} else {
	    if (*st != '\"' || ! hadQuote || *(st + 1) != '\0') {
		if (len != TAIL_MAX) {
		    len += 1;
		    *p++ = *st;
		}
	    }
	}
	++st;
    }
    st = MEM_ALLOC_E(&MS, len + 1, "strp");
    res = st;
    p = &TailBuffer[0];
    while (len != 0) {
	len -= 1;
	*st++ = *p++;
    }
    *st = '\0';
    return res;
}
