#include "MUD.h"
#include "types.h"
#include "Request.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"
#include "ClientStuff.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

#define INDENT_ESCAPE	'\001'

/*
 * messages.c - code for sending messages to various clients.
 */

static Request_t *RequestPtr;
static Client_t *ClientPtr;
static CHAR_T *MessagePtr;
static UINT_T AllocLength, LengthLeft;

static ULONG_T SpecialAgent1, SpecialAgent2, SpecialLocation, EditProcKey;
static UINT_T MessagePos, WordPos, TargetIndent;
static CHAR_T Message[TEXT_LENGTH];
static CHAR_T WordBuffer[TEXT_COLUMNS];
static BOOL_T ToLog, ToError, ToOtherAgents, ToSpecificClient, ToAllClients,
    ToAllAgentsBut,
    HadNonWizard, EditProcActive, DoingProcToString, HadOverflow;
static CHAR_T LastChar;

/*
 * messageInit - entry to allow our initialization here.
 */

void
messageInit(void)
{

    ToLog = B_FALSE;
    ToError = B_FALSE;
    ToOtherAgents = B_FALSE;
    ToSpecificClient = B_FALSE;
    ToAllClients = B_FALSE;
    ToAllAgentsBut = B_FALSE;
    HadNonWizard = B_FALSE;
    EditProcActive = B_FALSE;
    EditProcKey = NIL_ACTION;
    DoingProcToString = B_FALSE;
    HadOverflow = B_FALSE;
    MessagePos = 0;
    WordPos = 0;
    CurrentIndent = 0;
    TargetIndent = 0;
    LastChar = '\n';
    Prefix[0] = '\0';
    PrefixLength = 0;
    LengthLeft = 0;
    RequestPtr = P_NULL;
}

/*
 * sendText - external interface to quickly send text to a client.
 */

void
sendText(Client_t *cl, CHAR_T *message, UINT_T len)
{
    Request_t *rq;

    if (cl->cl_player != NIL_PLAYER && cl->cl_socket != -1) {
	rq = allocRequestE(&MS, len);
	rq->rq_type = rt_message;
	copy(&rq->rq_u.ru_text[0], message, len);
	if (cl->cl_telnet == P_NULL) {
	    rq->rq_key = cl->cl_defaultTextComponent;
	    putMessageCarefully(cl, rq);
	} else {
	    telnetPutText(cl, rq, (BOOL_T) (cl != ActiveClient));
	}
    }
}

/*
 * sendIt - send the currently built message.
 */

static void
sendIt(void)
{
    Client_t *cl;
    Request_t *rq;

    cl = ClientPtr;
    rq = RequestPtr;
    if (rq == P_NULL) {
	myAbort("RequestPtr NULL in sendIt");
    }
    if (EditProcActive && ! cl->cl_editing) {
	/* Some kind of error on the client end has sent back an
	   rt_editProcDone, which has cleared the cl_editing flag - don't
	   bother sending any more edit data over. */
	freeRequestE(&MS, rq);
    } else {
	rq->rq_usedLen = AllocLength - LengthLeft;
	if (cl->cl_telnet == P_NULL) {
	    rq->rq_key = cl->cl_defaultTextComponent;
	    putMessageCarefully(cl, rq);
	} else {
	    telnetPutText(cl, rq, (BOOL_T) (cl != ActiveClient));
	}
	RequestPtr = P_NULL;
    }
}

/*
 * newRequest - allocate a new request for a message.
 */

static void
newRequest(UINT_T len)
{
    Request_t *rq;

    if (RequestPtr != P_NULL) {
	myAbort("RequestPtr not NULL in newRequest");
    }
    rq = allocRequestE(&MS, len);
    AllocLength = rq->rq_availLen;
    if (EditProcActive) {
	rq->rq_type = rt_editProc;
	rq->rq_key = EditProcKey;
	rq->rq_flag = B_FALSE;
    } else {
	rq->rq_type = rt_message;
    }
    RequestPtr = rq;
    LengthLeft = rq->rq_availLen;
    MessagePtr = &rq->rq_u.ru_text[0];
}

/*
 * sendChar - inner routine to add a single character to the output message.
 *	The many levels of output used here are quite annoying, but I don't
 *	see any easy alternatives, except perhaps having the clients do all
 *	of the formatting stuff.
 */

static void
sendChar(CHAR_T ch)
{

    if (LengthLeft == 0) {
	sendIt();
	newRequest(TEXT_COLUMNS);
    }
    *MessagePtr++ = ch;
    LengthLeft -= 1;
}

/*
 * sendMessage - each chunk of message that goes to a client must be
 *	properly formatted within that client's output width. That means
 *	that the '@'s inserted for apprentice code must be done differently
 *	for each client. This is also why we force a '\n' when the output
 *	mode is switching - else we could easily get confused about when
 *	we need to put in '@'s.
 */

static void
sendMessage(Client_t *cl, CHAR_T *message, UINT_T len)
{
    CHAR_T *q, *wordEnd;
    UINT_T width, i, column, currentIndent;
    CHAR_T ch;
    BOOL_T noAts;

    if (cl->cl_deleted || cl->cl_player == NIL_PLAYER || cl->cl_socket == -1) {
	return;
    }

    ClientPtr = cl;
    noAts = cl->cl_noAts;
    currentIndent = CurrentIndent;
    /* allow for newlines and '@'s we may need to insert */
    i = len / 10;
    /* i is now our estimate of the number of lines */
    i = len + i * (currentIndent + PrefixLength + 2) + 5;
    /* i is now our estimate of the message size needed */
    newRequest(i);
    column = cl->cl_textColumn;
    width = getTextWidth(cl);
    while (len != 0) {
	ch = *message;
	if (ch == INDENT_ESCAPE) {
	    ++message;
	    ch = *message++;
	    currentIndent = ch - 'a';
	    ++message;
	    len = len - 3;
	} else if (ch == '\n' || ch == ' ' || ch == '\t') {
	    if (ch == '\n') {
		++message;
		len -= 1;
		sendChar(ch);
		column = 0;
	    }
	    if (column == 0 && len != 0) {
		if (HadNonWizard && ! noAts) {
		    sendChar('@');
		    column = 1;
		}
		q = &Prefix[0];
		while (*q != '\0') {
		    sendChar(*q++);
		}
		column += PrefixLength;
		i = currentIndent;
		column += i;
		while (i != 0) {
		    i -= 1;
		    sendChar(' ');
		}
	    }
	    /* preserve any indentation/spacing */
	    q = message;
	    while (B_TRUE) {
		ch = *message;
		if (len == 0 || (ch != ' ' && ch != '\t')) {
		    break;
		}
		++message;
		len -= 1;
		if (ch == ' ') {
		    column += 1;
		} else {
		    column = (column + 8) & (-8);
		}
	    }
	    if (column > width) {
		/* the whitespace took us past the end of the line -
		   replace it all with a single newline */
		sendChar('\n');
		column = 0;
	    } else {
		/* actually add the whitespace */
		while (q != message) {
		    sendChar(*q++);
		}
	    }
	} else {
	    /* non-whitespace - find the end of the word */
	    q = message;
	    while (B_TRUE) {
		ch = *q;
		if (len == 0 ||
		    ch == ' ' || ch == '\n' || ch == '\t' ||
		    ch == INDENT_ESCAPE)
		{
		    break;
		}
		++q;
		len -= 1;
	    }
	    i = q - message;
	    wordEnd = q;
	    /* end of word - did it fit? */
	    if (column + i > width && column != 0) {
		/* did not fit - have to wrap if not on new line */
		sendChar('\n');
		column = 0;
	    }
	    if (column == 0) {
		if (HadNonWizard && ! noAts) {
		    sendChar('@');
		    column += 1;
		}
		q = &Prefix[0];
		while (*q != '\0') {
		    sendChar(*q++);
		}
		column += PrefixLength;
		i = currentIndent;
		column += i;
		while (i != 0) {
		    i -= 1;
		    sendChar(' ');
		}
	    }
	    /* add in the word - split it up if it is too long */
	    while (message != wordEnd) {
		if (column == width) {
		    sendChar('\n');
		    column = 0;
		    if (HadNonWizard && ! noAts) {
			sendChar('@');
			column = 1;
		    }
		    q = &Prefix[0];
		    while (*q != '\0') {
			sendChar(*q++);
		    }
		    column += PrefixLength;
		    i = currentIndent;
		    column += i;
		    while (i != 0) {
			i -= 1;
			sendChar(' ');
		    }
		}
		sendChar(*message++);
		column += 1;
	    }
	}
    }
    if (AllocLength - LengthLeft == 0) {
	log("nothing to send in sendMessage");
	freeRequestE(&MS, RequestPtr);
	RequestPtr = P_NULL;
    } else {
	sendIt();
    }
    CurrentIndent = currentIndent;
    cl->cl_textColumn = column;
}

/*
 * messagePut - flush out any message by sending it to the proper clients.
 *	Note: the reason for this rather indirect method of doing the output
 *	is so that multi-request outputs will work for all of the modes.
 *	Return the new value of 'MessagePos'.
 */

static UINT_T
messagePut(void)
{
    Client_t *sender, *cl, *clFirst;
    CHAR_T *message, *p;
    ULONG_T here, save1, save2, save3;
    UINT_T len;

    len = MessagePos;
    if (len != 0) {
	sender = ActiveClient;
	message = &Message[0];
	/* always will be room */
	Message[len] = '\0';
	if (ToLog || (ToError && sender == P_NULL)) {
	    log(message);
	    len = 0;
	} else if (ToOtherAgents) {
	    /* Since this is so common, I've optimized it a bit. I don't
	       use 'findClientsAt' if the active agent is a client. */
	    if (sender != P_NULL) {
		here = sender->cl_location;
		if (here != NIL_THING) {
		    cl = sender->cl_nextLoc;
		    while (cl != sender) {
			sendMessage(cl, message, len);
			cl = cl->cl_nextLoc;
		    }
		}
	    } else {
		here = ActiveMachine->ml_location;
		cl = findClientsAt(here);
		if (cl != P_NULL) {
		    clFirst = cl;
		    do {
			if (cl != sender) {
			    sendMessage(cl, message, len);
			}
			cl = cl->cl_nextLoc;
		    } while (cl != clFirst);
		}
	    }
	    MessagePos = 0;    /* prevent recursion problem */
	    ToOtherAgents = B_FALSE;
	    sendToMachines(here, NIL_THING, NIL_THING, message);
	    ToOtherAgents = B_TRUE;
	    len = 0;
	} else if (ToAllAgentsBut) {
	    cl = findClientsAt(SpecialLocation);
	    if (cl != P_NULL) {
		clFirst = cl;
		do {
		    if (cl->cl_thing != SpecialAgent1 &&
			cl->cl_thing != SpecialAgent2)
		    {
			sendMessage(cl, message, len);
		    }
		    cl = cl->cl_nextLoc;
		} while (cl != clFirst);
	    }
	    MessagePos = 0;    /* prevent recursion problem */
	    save1 = SpecialLocation;
	    save2 = SpecialAgent1;
	    save3 = SpecialAgent2;
	    ToAllAgentsBut = B_FALSE;
	    sendToMachines(SpecialLocation, SpecialAgent1, SpecialAgent2,
			   message);
	    /* machine action could have done outputs */
	    ToAllAgentsBut = B_TRUE;
	    SpecialLocation = save1;
	    SpecialAgent1 = save2;
	    SpecialAgent2 = save3;
	    len = 0;
	} else if (ToSpecificClient) {
	    cl = getClientByThing(SpecialAgent1);
	    if (cl != P_NULL) {
		sendMessage(cl, message, len);
	    }
	    len = 0;
	} else if (ToAllClients) {
	    cl = Clients;
	    while (cl != P_NULL) {
		sendMessage(cl, message, len);
		cl = cl->cl_next;
	    }
	    len = 0;
	} else if (sender != P_NULL && sender->cl_player != NIL_PLAYER) {
	    if (EditProcActive) {
		/* want to avoid breaking lines up */
		p = message + len;
		while (*--p != '\n') {
		    if (p == message) {
			myAbort("found no newline in edit proc send");
		    }
		}
		++p;
		MessagePos = p - message;
		sendMessage(sender, message, MessagePos);
		len -= MessagePos;
		copy(message, p, len);
	    } else {
		sendMessage(sender, message, len);
		len = 0;
	    }
	} else {
	    len = 0;
	}
    }
    HadNonWizard = B_FALSE;
    MessagePos = len;
    return len;
}

void
setEditProcActive(BOOL_T state, ULONG_T procKey)
{
    Request_t *rq;

    if (! state) {
	messageFlush();
	rq = allocRequestE(&MS, 0);
	rq->rq_type = rt_editProc;
	rq->rq_key = procKey;
	rq->rq_flag = B_TRUE;
	putMessageCarefully(ActiveClient, rq);
    }
    EditProcActive = state;
    EditProcKey = procKey;
}

/*
 * putChar - bottom-most level of normal output.
 */

void
putChar(CHAR_T ch)
{
    CHAR_T *mp, *wp;
    UINT_T i, pos;

    pos = MessagePos;
    if (DoingProcToString) {
	/* all 'fmtXXX' output comes through putChar */
	if (MessagePos == TEXT_LENGTH) {
	    if (! HadOverflow) {
		runError("result too long in ProcToString");
		HadOverflow = B_TRUE;
	    }
	} else {
	    Message[pos] = ch;
	    pos += 1;
	    MessagePos = pos;
	}
	return;
    }

    i = WordPos;

    /* this can cause some extra messages by splitting text up, but it will
       also reduce the number of unwanted prepended '@'s */
    if (pos != 0 && i == 0 && Message[pos - 1] == '\n') {
	if (EffectiveStatus == ps_wizard) {
	    if (HadNonWizard) {
		pos = messagePut();
	    }
	} else {
	    if (! HadNonWizard) {
		pos = messagePut();
	    }
	}
    }

    if (ch == ' ' || ch == '\n' || ch == '\t' || ch == INDENT_ESCAPE) {
	if (i != 0) {
	    /* This needs to be >=, since it is '} else if' below */
	    if (pos + i >= TEXT_LENGTH - 1) {
		pos = messagePut();
	    }
	    mp = &Message[pos];
	    wp = &WordBuffer[0];
	    pos += i;
	    while (i != 0) {
		i -= 1;
		*mp++ = *wp++;
	    }
	} else if (pos == TEXT_LENGTH - 1) {
	    pos = messagePut();
	}
	Message[pos] = ch;
	pos += 1;
    } else {
	if (i == TEXT_COLUMNS) {
	    /* word buffer is full */
	    if (pos > TEXT_LENGTH - TEXT_COLUMNS - 1) {
		/* no room in 'Message' for the word */
		pos = messagePut();
	    }
	    mp = &Message[pos];
	    wp = &WordBuffer[0];
	    pos += i;
	    while (i != 0) {
		i -= 1;
		*mp++ = *wp++;
	    }
	}
	WordBuffer[i] = ch;
	i += 1;
    }
    MessagePos = pos;
    WordPos = i;
    LastChar = ch;
    if (EffectiveStatus != ps_wizard) {
	HadNonWizard = B_TRUE;
    }
}

/*
 * putString - send a whole string to the message buffer.
 */

void
putString(CHAR_T *message)
{
    CHAR_T *p, *q;
    UINT_T len, i;
    CHAR_T ch;
    BOOL_T gotWhite;

    if (*message != '\0') {
	if (MessagePos != 0 && Message[MessagePos - 1] == '\n') {
	    if (EffectiveStatus == ps_wizard) {
		if (HadNonWizard) {
		    (void) messagePut();
		}
	    } else {
		if (! HadNonWizard) {
		    (void) messagePut();
		}
	    }
	}

	/* We try for an optimization - avoid calling 'putChar' for each
	   character in the string if we can just copy the string straight
	   into the output buffer. */

	/* find the end of the string */
	p = message;
	while (*p != '\0') {
	    ++p;
	}
	/* scan backwards for the last whitespace character */
	do {
	    ch = *--p;
	} while (p != message && ch != ' ' && ch != '\n' && ch != '\t');
	gotWhite = B_FALSE;
	/* go to the end of the whitespace */
	if (p != message) {
	    gotWhite = B_TRUE;
	    ++p;
	}
	len = p - message;
	i = MessagePos;
	if (gotWhite && i + WordPos + len < TEXT_LENGTH) {
	    /* No problems with a word boundary, and this piece will fit
	       into the current buffer - just copy it there directly. */
	    if (EffectiveStatus != ps_wizard) {
		HadNonWizard = B_TRUE;
	    }
	    q = &Message[i];
	    i += len;
	    MessagePos = i;
	    i = WordPos;
	    if (i != 0) {
		MessagePos += i;
		p = &WordBuffer[0];
		while (i != 0) {
		    i -= 1;
		    *q++ = *p++;
		}
		WordPos = 0;
		/* sometimes this is important! */
		/* NO WAY! p is not used again, LastChar is overwritten */
#if 0
		LastChar = *--p;
#endif
	    }
	    while (len != 0) {
		len -= 1;
		*q++ = *message++;
	    }
	    /* have to remember to do this! */
	    LastChar = *(message - 1);
	}
	/* handle whatever is left the hard way */
	while (*message != '\0') {
	    putChar(*message++);
	}
    }
}

/*
 * messageFlush - flush out any existing message or portion thereof.
 */

void
messageFlush(void)
{

    TRACE("messageFlush", ActiveClient);
    if (LastChar != '\n') {
	putChar('\n');
    }
    (void) messagePut();
}

/*
 * setSpecific - external routine to allow building up and sending a
 *	message to a specific client. Usage: call first with a non-NULL
 *	client pointer. Do output. Then call with a NULL client pointer.
 */

void
setSpecific(ULONG_T agent)
{

    messageFlush();
    if (agent == NIL_THING) {
	ToSpecificClient = B_FALSE;
    } else {
	ToSpecificClient = B_TRUE;
	SpecialAgent1 = agent;
    }
}

/*
 * setOthers - external routine to allow selection of 'ToOtherAgents' mode.
 *	Used for things like 'Say'. Those routines do their own call
 *	to 'messageFlush'.
 */

void
setOthers(BOOL_T flag)
{

    ToOtherAgents = flag;
}

/*
 * setLogging - used when aborting - messages are logged.
 */

void
setLogging(void)
{

    if (! ToLog) {
	messageFlush();
	TRACE("*** setLogging - we must be aborting here ***", 0);
	ToLog = B_TRUE;
    }
}

/*
 * setToError - control whether we are outputting error stuff or not.
 *	This has a very specific use inside exec.c - it will only output
 *	to the log file if there is no active client. More general error
 *	logging prior to abort simply uses 'setLogging'.
 */

void
setToError(BOOL_T flag)
{

    messageFlush();
    ToError = flag;
}

/*
 * TextWidth - allow the player to change the display width.
 */

LONG_T
TextWidth(LONG_T newWidth)
{
    Client_t *cl;
    Player_t *pl;
    UINT_T oldWidth;

    cl = ActiveClient;
    if (cl == P_NULL) {
	runError("machines cannot TextWidth");
	return 0;
    } else if ((newWidth < 32 && newWidth != 0) || newWidth > MAX_COLUMNS) {
	runError("new display width out of range");
	return 0;
    } else {
	messageFlush();
	oldWidth = getTextWidth(cl);
	if (newWidth != 0) {
	    cl->cl_textWidth = newWidth;
	    pl = ioRead(cl->cl_player);
	    pl->pl_textWidth = newWidth;
	    ioDirty(cl->cl_player);
	}
	return oldWidth;
    }
}

/*
 * TextHeight - allow the player to change the display height.
 */

LONG_T
TextHeight(LONG_T newHeight)
{
    Client_t *cl;
    Player_t *pl;
    UINT_T oldHeight;

    cl = ActiveClient;
    if (cl == P_NULL) {
	runError("machines cannot TextHeight");
	return 0;
    } else if (newHeight < 10 && newHeight != 0) {
	runError("new display height too small");
	return 0;
    } else {
	messageFlush();
	oldHeight = getTextHeight(cl);
	if (newHeight != 0) {
	    cl->cl_textHeight = newHeight;
	    pl = ioRead(cl->cl_player);
	    pl->pl_textHeight = newHeight;
	    ioDirty(cl->cl_player);
	}
	return oldHeight;
    }
}

/*
 * SetIndent - allow setting of left-margin indent amount.
 */

void
SetIndent(ULONG_T n)
{
    CHAR_T *p, *q;
    UINT_T pos, i;

    if (n < 0 || n > 20) {
	runError("SetIndent value must be 0 - 20");
    } else if (n != TargetIndent) {
	i = WordPos;
	pos = MessagePos;
	if (pos + i + 3 >= TEXT_LENGTH) {
	    pos = messagePut();
	}
	if (i != 0) {
	    p = &Message[pos];
	    q = &WordBuffer[0];
	    pos += i;
	    MessagePos = pos;
	    WordPos = 0;
	    while (i != 0) {
		i -= 1;
		*p++ = *q++;
	    }
	}
	putChar(INDENT_ESCAPE);
	putChar(n + 'a');
	putChar(INDENT_ESCAPE);
	TargetIndent = n;
    }
}

/*
 * GetIndent - return the current indent value.
 */

ULONG_T
GetIndent(void)
{

    return TargetIndent;
}

/*
 * SetPrefix - set the prefix.
 */

void
SetPrefix(CHAR_T *newPrefix)
{
    UINT_T len;

    len = strLen1(newPrefix) - 1;
    if (len > PREFIX_MAX) {
	runError("SetPrefix string must be <= 20 characters");
    } else {
	strCpy(&Prefix[0], newPrefix);
	PrefixLength = len;
    }
}

/*
 * NPrint - same as 'Print' (which maps to 'putString'), except that it will
 *	not print out any newlines in the string, and will print one newline
 *	at the end of the string. It also forces the current value of
 *	EffectiveStatus to be ps_wizard. This is used from verbs routines
 *	(such as 'parseInput' in util.m) which want to print out a string
 *	retrieved from an arbitrary place, such that there is no possibility
 *	of the string doing spoofing, and also that there will be no '@'
 *	printed at the beginning of the line.
 */

void
NPrint(CHAR_T *message)
{
    CHAR_T *p;
    PlayerStatus_t saveStatus;

    (void) messagePut();
    saveStatus = EffectiveStatus;
    EffectiveStatus = ps_wizard;
    p = message;
    while (*p != '\0' && *p != '\n') {
	++p;
    }
    if (*p == '\n') {
	*p = '\0';
	putString(message);
	*p = '\n';
    } else {
	putString(message);
    }
    putChar('\n');
    (void) messagePut();
    EffectiveStatus = saveStatus;
}

/*
 * APrint - send the given string to ALL connected clients.
 */

void
APrint(CHAR_T *message)
{

    TRACE("APrint", message);
    if (EffectivePlayer != SYSADMIN_KEY) {
	runError("only SysAdmin can use APrint");
    } else {
	messageFlush();
	ToAllClients = B_TRUE;
	putString(message);
	messageFlush();
	ToAllClients = B_FALSE;
    }
}

/*
 * OPrint - send string to all agents in same location as the current
 *	client (player or machine), except the sending client.
 */

void
OPrint(CHAR_T *st)
{

    TRACE("OPrint", st);
    if (ActiveClient != P_NULL || ActiveMachine != P_NULL) {
	messageFlush();
	ToOtherAgents = B_TRUE;
	putString(st);
	messageFlush();
	ToOtherAgents = B_FALSE;
    }
}

/*
 * SPrint - send the message to the client who is the specified thing.
 */

void
SPrint(ULONG_T agent, CHAR_T *st)
{

    TRACE("SPrint", agent);
    if (agent == NIL_THING) {
	runError("nil agent passed to SPrint");
    } else {
	messageFlush();
	ToSpecificClient = B_TRUE;
	SpecialAgent1 = agent;
	putString(st);
	messageFlush();
	ToSpecificClient = B_FALSE;
    }
}

/*
 * ABPrint - send the message to all clients in the given location, except
 *	the two mentioned.
 */

void
ABPrint(ULONG_T location, ULONG_T agent1, ULONG_T agent2, CHAR_T *st)
{

    TRACE("ABPrint", location);
    if (location == NIL_THING) {
	runError("nil location passed to ABPrint");
    } else {
	messageFlush();
	ToAllAgentsBut = B_TRUE;
	SpecialLocation = location;
	SpecialAgent1 = agent1;
	SpecialAgent2 = agent2;
	putString(st);
	messageFlush();
	ToAllAgentsBut = B_FALSE;
    }
}

/*
 * toggleProcToString - set the mode which says we are formatting an action
 *	into a string buffer. On the second call, we return the resulting
 *	string.
 */

CHAR_T *
toggleProcToString(void)
{
    CHAR_T *p;

    if (DoingProcToString) {
	DoingProcToString = B_FALSE;
	if (HadOverflow) {
	    MessagePos = 0;
	    return P_NULL;
	} else {
	    p = MEM_ALLOC_E(&MS, MessagePos + 1, "acst");
	    copy(p, &Message[0], MessagePos);
	    *(p + MessagePos) = '\0';
	    MessagePos = 0;
	    return p;
	}
    } else {
	messageFlush();
	DoingProcToString = B_TRUE;
	HadOverflow = B_FALSE;
	return P_NULL;
    }
}
