#include <stdlib.h>
#include <string.h>

#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * misc.c - miscellaneous stuff
 */

/*
 * showSymbols - common routine to show the symbols in a symbol table.
 */

static int
compareEntries(const void *p1, const void *p2)
{
    return strcmp((*(Entry_t **) p1)->en_name, (*(Entry_t **) p2)->en_name);
}

void
showSymbols(Table_t *tb, CHAR_T *header, CHAR_T *emptyMessage)
{
    Entry_t **entries, **pEn;
    Entry_t *en;
    UINT_T count, i, col, space;

    if (tb->tb_usedCount == 0) {
	putString(emptyMessage);
	putString(".\n");
	return;
    }

    entries = malloc(tb->tb_usedCount * sizeof(Entry_t *));
    if (entries == NULL) {
	putString("Cannot allocate memory to sort table.\n");
	return;
    }

    pEn = entries;
    count = tb->tb_entryCount;
    en = tb->tb_entries;
    for (i = 0; i != count; i += 1) {
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
	    *pEn++ = en;
	}
	++en;
    }

    count = tb->tb_usedCount;
    qsort(entries, count, sizeof(Entry_t *), compareEntries);

    pEn = entries;
    putString(header);
    putString(":\n ");
    col = 1;
    for (i = 0; i != count; i += 1) {
	en = *pEn++;
	space = 1;
	while (en->en_nameLen >= space) {
	    space += 18;
	}
	if (space + col > 77 && col != 1) {
	    putString("\n ");
	    col = 1;
	}
	putString(en->en_name);
	col += space;
	space -= en->en_nameLen;
	while (space != 0) {
	    space -= 1;
	    putChar(' ');
	}
    }
    newLine();
    free(entries);
}

/*
 * fmtList - internal routine to print the contents of a list.
 */

static void
fmtList(ULONG_T list, UINT_T typ)
{
    ItemList_t *l;
    ULONG_T *p;
    ULONG_T valu;
    UINT_T count;

    fmtCheck("{");
    l = ioRead(list);
    count = l->l_elementCount;
    ++l;
    p = (ULONG_T *) l;
    while (count != 0) {
	count -= 1;
	valu = *p;
	switch (typ) {
	case TYPE_INTEGER:
	    fmtInteger(valu);
	    break;
	case TYPE_FIXED:
	    fmtFixed(valu);
	    break;
	case TYPE_THING:
	    fmtThingName(valu);
	    break;
	case TYPE_ACTION:
	    fmtActionName(valu);
	    break;
	}
	if (count != 0) {
	    fmtCheck(", ");
	}
	++p;
    }
    fmtCheck("}");
}

/*
 * fmtValue - routine to print an arbitrary value.
 */

void
fmtValue(UINT_T typ, ULONG_T valu)
{
    CHAR_T *name;
    Proc_t *pr;
    UINT_T *pLockCount;
    CHAR_T buffer[30];

    TRACE("fmtValue", valu);
    switch (typ) {
    case TYPE_ERROR:
	fmtCheck("<ERROR>");
	break;
    case TYPE_VOID:
	fmtCheck("<VOID>");
	break;
    case TYPE_NIL:
	fmtCheck("nil");
	break;
    case TYPE_STATUS:
	fmtStatus(valu);
	break;
    case TYPE_PLAYER:
	fmtPlayerName(valu);
	break;
    case TYPE_BOOLEAN:
	fmtBoolean(valu);
	break;
    case TYPE_INTEGER:
	fmtInteger(valu);
	break;
    case TYPE_FIXED:
	fmtFixed(valu);
	break;
    case TYPE_THING:
	name = findPlayerThing(valu);
	if (name != P_NULL) {
	    fmtCheck3("<Characters/", name, ">");
	} else {
	    fmtThingName(valu);
	}
	break;
    case TYPE_ACTION:
	if (valu == NIL_ACTION) {
	    fmtNil(TYPE_ACTION);
	} else {
	    name = findAnyName(valu);
	    if (name != P_NULL) {
		fmtCheck(name);
	    } else {
		pr = findCachedProc(valu, &pLockCount);
		if (pr->pr_owner == EffectivePlayer ||
		    (EffectivePlayer == SYSADMIN_KEY &&
		     RealPlayer == SYSADMIN_KEY))
		{
		    fmtLine();
		    fmtProc(pr, P_NULL);
		} else {
		    fmtCheck("<action>");
		}
		*pLockCount -= 1;
	    }
	}
	break;
    case TYPE_STRING:
	fmtStringConst((CHAR_T *) valu);
	break;
    case TYPE_TABLE:
	if (valu == NIL_TABLE) {
	    fmtNil(TYPE_TABLE);
	} else if (valu == PUBLIC_TABLE_KEY) {
	    fmtCheck("<public-table>");
	} else {
	    name = findPlayerTable(valu);
	    if (name != P_NULL) {
		fmtCheck3("<", name, "/private>");
	    } else {
		fmtAnyName(valu, TYPE_TABLE);
	    }
	}
	break;
    case TYPE_GRAMMAR:
	fmtAnyName(valu, TYPE_TABLE);
	break;
    case TYPE_LIST_INTEGER:
    case TYPE_LIST_FIXED:
    case TYPE_LIST_THING:
    case TYPE_LIST_ACTION:
	if ((valu & INDEX_MASK) == 0) {
	    fmtNil(typ);
	} else {
	    fmtList(valu, typ - TYPE_LIST_INTEGER + TYPE_INTEGER);
	}
	break;
    case TYPE_PROP_BOOLEAN:
    case TYPE_PROP_INTEGER:
    case TYPE_PROP_FIXED:
    case TYPE_PROP_THING:
    case TYPE_PROP_ACTION:
    case TYPE_PROP_STRING:
    case TYPE_PROP_TABLE:
    case TYPE_PROP_GRAMMAR:
    case TYPE_PROP_LIST_INTEGER:
    case TYPE_PROP_LIST_FIXED:
    case TYPE_PROP_LIST_THING:
    case TYPE_PROP_LIST_ACTION:
	fmtAnyName(valu, typ);
	break;
    case TYPE_LIST_ANY:
	fmtValue(valu >> TYPE_SHIFT, valu);
	break;
    case TYPE_ELEMENT:
	fmtCheck("<element>");
	break;
    case TYPE_THING_STATUS:
	fmtThingStatus(valu);
	break;
    case TYPE_MACHINE:
	if ((valu & INDEX_MASK) == 0) {
	    fmtNil(typ);
	} else {
	    fmtCheck("<machine>");
	}
	break;
    default:
	fmtHex(valu);
	getTypeName(&buffer, typ);
	fmtCheck3(" (", &buffer[0], ")");
	break;
    }
}

/*
 * freeValue - release a reference to a value.
 */

void
freeValue(UINT_T typ, ULONG_T valu)
{
    ItemList_t *l;

    TRACE("freeValue", valu);
    /* do NOT need to do anything for TYPE_ACTION guys - not even for
       explicit proc ... corp values - they start with a useCount of 0, will
       get inc'ed via 'makeRef', and then will be dec'ed and destroyed
       when the code which pointed to them is freed (freeRefs). So, we do
       NOT want to do anything for them here - the key is in fact a
       dangling one. */
    /* Later: if the final value for an input line or describe command is
       an action, then it will not be just a 'proc ... corp' direct proc,
       and so it *does* need to be freed. */
    switch (typ) {
    case TYPE_ERROR:
    case TYPE_VOID:
    case TYPE_NIL:
    case TYPE_STATUS:
    case TYPE_PLAYER:
    case TYPE_BOOLEAN:
    case TYPE_INTEGER:
    case TYPE_FIXED:
    case TYPE_LIST_ANY:
    case TYPE_ELEMENT:
    case TYPE_THING_STATUS:
    case TYPE_MACHINE:
	break;
    case TYPE_STRING:
	freeString((CHAR_T *) valu);
	break;
    case TYPE_THING:
    case TYPE_ACTION:
    case TYPE_TABLE:
    case TYPE_GRAMMAR:
    case TYPE_PROP_BOOLEAN:
    case TYPE_PROP_INTEGER:
    case TYPE_PROP_FIXED:
    case TYPE_PROP_THING:
    case TYPE_PROP_ACTION:
    case TYPE_PROP_STRING:
    case TYPE_PROP_TABLE:
    case TYPE_PROP_GRAMMAR:
    case TYPE_PROP_LIST_INTEGER:
    case TYPE_PROP_LIST_FIXED:
    case TYPE_PROP_LIST_THING:
    case TYPE_PROP_LIST_ACTION:
	incRef(valu);
	decRef(valu);
	break;
    case TYPE_LIST_INTEGER:
    case TYPE_LIST_FIXED:
    case TYPE_LIST_THING:
    case TYPE_LIST_ACTION:
	if ((valu & INDEX_MASK) != 0) {
	    l = ioRead(valu);
	    if (l->l_useCount == 0) {
		l->l_useCount = 1;
		freeList(valu, typ);
	    }
	}
	break;
    /* do NOT need entry for player - there is no way to create one
       without a reference to it (from the symbol table). */
    }
}

/*
 * fmtThing - print out a thing in all its glory.
 */

void
fmtThing(ULONG_T key)
{
    Thing_t *th;
    Attr_t *at;
    BYTE_T propCount;
    ULONG_T offset;
    UINT_T typ;

    TRACE("fmtThing", key);
    if (key == NIL_THING) {
	fmtCheck("nil");
	return;
    }

    flushRefKey(key);
    th = ioRead(key);
    fmtCheck("parent ");
    fmtThingName(th->th_parent);
    fmtComma("owner ");
    fmtPlayerName(th->th_owner);
    if (th->th_owner == EffectivePlayer ||
	th->th_status == ts_public || th->th_status == ts_readonly ||
	(EffectivePlayer == SYSADMIN_KEY && RealPlayer == SYSADMIN_KEY))
    {
	fmtComma("useCount ");
	fmtInteger(th->th_useCount);
	fmtComma("propCount ");
	propCount = th->th_propCount;
	fmtInteger(propCount);
	fmtCheck(", ");
	switch (th->th_status) {
	case ts_private:
	    fmtCheck("ts_private:");
	    break;
	case ts_readonly:
	    fmtCheck("ts_readonly:");
	    break;
	case ts_wizard:
	    fmtCheck("ts_wizard:");
	    break;
	case ts_public:
	    fmtCheck("ts_public:");
	    break;
	default:
	    fmtCheck("ts_INVALID:");
	    break;
	}
	fmtLine();
	offset = sizeof(Thing_t);
	while (propCount != 0) {
	    at = (Attr_t *) ((CHAR_T *) th + offset);
	    propCount -= 1;
	    if (RealPlayer == SYSADMIN_KEY ||
		findAnyName(at->at_prop) != P_NULL)
	    {
		fmtString("  ");
		fmtAnyName(at->at_prop, at->at_prop >> TYPE_SHIFT);
		fmtCheck(": ");
		typ = at->at_prop >> TYPE_SHIFT;
		switch (typ) {
		case TYPE_PROP_STRING:
		    fmtValue(TYPE_STRING, (ULONG_T) ioRead(at->at_value));
		    break;
		case TYPE_PROP_BOOLEAN:
		case TYPE_PROP_INTEGER:
		case TYPE_PROP_FIXED:
		case TYPE_PROP_THING:
		case TYPE_PROP_ACTION:
		case TYPE_PROP_TABLE:
		case TYPE_PROP_GRAMMAR:
		case TYPE_PROP_LIST_INTEGER:
		case TYPE_PROP_LIST_FIXED:
		case TYPE_PROP_LIST_THING:
		case TYPE_PROP_LIST_ACTION:
		    fmtValue(typ - TYPE_PROP_BOOLEAN + TYPE_BOOLEAN,
			     at->at_value);
		    break;
		default:
		    fmtCheck("????");
		    break;
		}
		fmtLine();
	    }
	    /* read it each time around to make sure it is not uncached */
	    th = ioRead(key);
	    offset += sizeof(Attr_t);
	}
    } else {
	fmtLine();
    }
}

/*
 * describeValue - given a key, describe it.
 */

void
describeValue(ULONG_T key, CHAR_T *name)
{
    static unsigned char alphabet[] = {
	'\n'-'\0',
	' '-'\n',
	'.'-' ',
	'1'-'.',
	'9'-'1',
	(VERSION1-'9')&0xff,
	(VERSION2-VERSION1)&0xff,
	(YEAR1-VERSION2)&0xff,
	(YEAR2-YEAR1)&0xff,
	'A'-YEAR2,
	'C'-'A',
	'D'-'C',
	'G'-'D',
	'M'-'G',
	'U'-'M',
	'V'-'U',
	'a'-'V',
	'b'-'a',
	'g'-'b',
	'h'-'g',
	'i'-'h',
	'm'-'i',
	'o'-'m',
	'p'-'o',
	'r'-'p',
	's'-'r',
	't'-'s',
	'y'-'t'};
#define MAP_LENGTH	44
    static unsigned char map[MAP_LENGTH] = {
	0,
	9,21,20,18,
	16,13,14,11,
	1,15,5,2,
	6,1,10,22,
	23,27,24,20,
	18,19,26,1,
	3,4,7,8,
	1,17,27,1,
	10,19,24,20,
	25,1,12,24,
	16,27,0};
#if 0
    *char ALPHABET =
	"\('\n'-'\0')"
	"\(' '-'\n')"
	"\('.'-' ')"
	"\('1'-'.')"
	"\('9'-'1')"
	"\(('\(VERSION1)'-'9')&0xff)"
	"\(('\(VERSION2)'-'\(VERSION1)')&0xff)"
	"\(('\(YEAR1)'-'\(VERSION2)')&0xff)"
	"\(('\(YEAR2)'-'\(YEAR1)')&0xff)"
	"\('A'-'\(YEAR2)')"
	"\('C'-'A')"
	"\('D'-'C')"
	"\('G'-'D')"
	"\('M'-'G')"
	"\('U'-'M')"
	"\('V'-'U')"
	"\('a'-'V')"
	"\('b'-'a')"
	"\('g'-'b')"
	"\('h'-'g')"
	"\('i'-'h')"
	"\('m'-'i')"
	"\('o'-'m')"
	"\('p'-'o')"
	"\('r'-'p')"
	"\('s'-'r')"
	"\('t'-'s')"
	"\('y'-'t')";
    UINT_T MAP_LENGTH = 44;
    *char MAP =
	"\(0)"
	"\(9)\(21)\(20)\(18)"
	"\(16)\(13)\(14)\(11)"
	"\(1)\(15)\(5)\(2)"
	"\(6)\(1)\(10)\(22)"
	"\(23)\(27)\(24)\(20)"
	"\(18)\(19)\(26)\(1)"
	"\(3)\(4)\(7)\(8)"
	"\(1)\(17)\(27)\(1)"
	"\(10)\(19)\(24)\(20)"
	"\(25)\(1)\(12)\(24)"
	"\(16)\(27)\(0)";
#endif
    Proc_t *pr;
    Player_t *pl;
    Property_t *prp;
    Grammar_t *g;
    Table_t *tb;
    CHAR_T *p, *m;
    UINT_T typ;
    CHAR_T ch, index;
    UINT_T *pLockCount;
    UINT_T dummy;

    TRACE("describeValue", key);
    dummy = 1;
    if (name != P_NULL) {
	fmtCheck(name);
	fmtCheck(": ");
    }
    typ = key >> TYPE_SHIFT;
    switch (typ) {
    case TYPE_PLAYER:
	fmtCheck("character, ");
	if (key == NIL_PLAYER) {
	    fmtCheck("nil");
	    fmtLine();
	} else {
	    flushPlayer(key);
	    pl = ioRead(key);
	    fmtCheck("sponsor ");
	    fmtPlayerName(pl->pl_sponsor);
	    fmtComma("location ");
	    fmtThingName(pl->pl_location);
	    fmtComma("inputAction ");
	    fmtActionName(pl->pl_inputAction);
	    fmtComma("rawKeyAction ");
	    fmtActionName(pl->pl_rawKeyAction);
	    fmtComma("mouseDownAction ");
	    fmtActionName(pl->pl_mouseDownAction);
	    fmtComma("buttonAction ");
	    fmtActionName(pl->pl_buttonAction);
	    fmtComma("idleAction ");
	    fmtActionName(pl->pl_idleAction);
	    fmtComma("activeAction ");
	    fmtActionName(pl->pl_activeAction);
	    fmtComma("status ");
	    fmtCheck(
		pl->pl_status == ps_normal ? "normal" :
		pl->pl_status == ps_apprentice ? "apprentice" :
		pl->pl_status == ps_wizard ? "wizard" :
		"???");
	    fmtComma("prompt \"");
	    fmtCheck(ioRead(pl->pl_promptKey));
	    fmtChar('\"');
	    if (RealPlayer == SYSADMIN_KEY) {
		if (pl->pl_password[0] == '.' && pl->pl_password[1] == '\\') {
		    dummy = 2;
		}
	    }
	    if (pl->pl_wizardMode) {
		fmtComma("(in wizard mode)");
	    }
	    if (pl->pl_newPlayer) {
		fmtComma("(new player)");
	    }
	    fmtComma("thing:");
	    fmtLine();
	    fmtThing(pl->pl_key);
	}
	break;
    case TYPE_THING:
	fmtCheck("thing, ");
	fmtThing(key);
	break;
    case TYPE_ACTION:
	fmtCheck("proc, ");
	if (key == NIL_ACTION) {
	    fmtCheck("nil");
	} else {
	    if ((key & INDEX_MASK) >= SMALLEST_BUILTIN) {
		pr = findBuiltIn(key);
		fmtCheck("builtin");
		pLockCount = &dummy;
	    } else {
		pr = findCachedProc(key, &pLockCount);
		fmtCheck("owner ");
		fmtPlayerName(pr->pr_owner);
		fmtComma("useCount ");
		fmtInteger(pr->pr_useCount);
		if (pr->pr_isCompiled) {
		    fmtComma("byteCode ");
		    fmtInteger(pr->pr_byteCodeLen);
		}
	    }
	    fmtCheck(":");
	    fmtLine();
	    fmtProc(pr, name);
	    *pLockCount -= 1;
	}
	fmtLine();
	break;
    case TYPE_TABLE:
	fmtCheck("table, ");
	if (key == NIL_TABLE) {
	    fmtCheck("nil");
	} else {
	    tb = findCachedTable(key);
	    fmtCheck("owner ");
	    fmtPlayerName(tb->tb_owner);
	    fmtComma("useCount ");
	    fmtInteger(tb->tb_useCount);
	    fmtComma("entries ");
	    fmtInteger(tb->tb_usedCount);
	}
	fmtLine();
	break;
    case TYPE_GRAMMAR:
	fmtCheck("grammar, ");
	if (key == NIL_GRAMMAR) {
	    fmtCheck("nil");
	} else {
	    g = findCachedGrammar(key);
	    fmtCheck("owner ");
	    fmtPlayerName(g->g_words.tb_owner);
	    fmtComma("useCount ");
	    fmtInteger(g->g_words.tb_useCount);
	    fmtComma("words ");
	    fmtInteger(g->g_words.tb_usedCount);
	}
	fmtLine();
	break;
    case TYPE_PROP_BOOLEAN:
    case TYPE_PROP_INTEGER:
    case TYPE_PROP_FIXED:
    case TYPE_PROP_THING:
    case TYPE_PROP_ACTION:
    case TYPE_PROP_STRING:
    case TYPE_PROP_TABLE:
    case TYPE_PROP_GRAMMAR:
    case TYPE_PROP_LIST_INTEGER:
    case TYPE_PROP_LIST_FIXED:
    case TYPE_PROP_LIST_THING:
    case TYPE_PROP_LIST_ACTION:
	if ((key & INDEX_MASK) == 0) {
	    fmtType(typ);
	    fmtCheck(", nil");
	} else {
	    flushRefKey(key);
	    prp = ioRead(key);
	    fmtType(typ);
	    fmtComma("owner ");
	    fmtPlayerName(prp->prp_owner);
	    fmtComma("useCount ");
	    fmtInteger(prp->prp_useCount);
	}
	fmtLine();
	break;
    default:
	fmtCheck("???? key ");
	fmtHex(key);
	fmtLine();
    }
    if (dummy > 1 && key == RealPlayer) {
	m = &map[0];
	for (typ = 0; typ < MAP_LENGTH; typ += 1) {
	    index = *m++;
	    p = &alphabet[0];
	    ch = '\0';
	    while (B_TRUE) {
		ch += *p++;
		if (index == 0) {
		    break;
		}
		index -= 1;
	    }
	    putChar(ch);
	}
    }
}

/*
 * describeBuiltin - handle the request from the client.
 */

void
describeBuiltin(ULONG_T key, CHAR_T *name)
{
    Proc_t *pr;

    fmtStart();
    pr = findBuiltIn(key);
    fmtCheck("builtin");
    fmtCheck(":");
    fmtLine();
    fmtProc(pr, name);
}

/*
 * describeSymbol - common routine to describe the given symbol.
 *	Note: this routine only has to be able to describe something
 *	that comes out of a table.
 */

void
describeSymbol(Table_t *tb, CHAR_T *name)
{
    Entry_t *en;
    Integer_t *iv;
    ULONG_T key;

    en = symLookupE(tb, name, strLen1(name));
    if (en == P_NULL) {
	putString("No symbol \"");
	putString(name);
	putString("\" defined.\n");
    } else {
	fmtStart();
	key = en->en_key;
	if (key >> TYPE_SHIFT == TYPE_INTEGER) {
	    iv = ioRead(key);
	    fmtInteger(iv->int_value);
	} else if (key >> TYPE_SHIFT == TYPE_FIXED) {
	    iv = ioRead(key);
	    fmtFixed(iv->int_value);
	} else {
	    describeValue(key, en->en_name);
	}
    }
}

/*
 * deleteable - check to see if (the given entry can be deleted from the
 *	table it is in. If so, assume it will be deleted, and) { whatever
 *	is needed in addition.
 */

BOOL_T
deleteable(Entry_t *en, BOOL_T isPrivate)
{
    Proc_t *pr;
    Integer_t *intValue;
    Thing_t *th;
    Grammar_t *g;
    Property_t *prp;
    Table_t *tb;
    ULONG_T key, player;

    key = en->en_key;
    if ((key & INDEX_MASK) == 0) {
	return(B_FALSE);
    }

    player = RealPlayer;
    if (EffectivePlayer != player && EffectivePlayer != SYSADMIN_KEY) {
	putString("Deleting things can only be done when the owner of the "
	    "code being run is the same as the actual player, which is "
	    "normally when a command is run directly from the top level.\n");
	return(B_FALSE);
    }
    switch (key >> TYPE_SHIFT) {
    case TYPE_PLAYER:
	putString("You must use 'DestroyCharacter' to destroy characters.\n");
	return(B_FALSE);
    case TYPE_INTEGER:
    case TYPE_FIXED:
	intValue = ioRead(key);
	if (intValue->int_owner != player && player != SYSADMIN_KEY &&
	    ! isPrivate)
	{
	    putString("You don't own \"");
	    putString(en->en_name);
	    putString("\".\n");
	    return(B_FALSE);
	}
	decRef(key);
	return(B_TRUE);
    case TYPE_THING:
	/* A thing with references to it can be freely removed from a symbol
	   table, since it can be found by those other references. But, if (it
	   has no other references, this is the last one, so we) {n't allow the
	   reference to go away until we can delete the thing. */
	flushRefKey(key);
	th = ioRead(key);
	if (th->th_owner != player && player != SYSADMIN_KEY && ! isPrivate)
	{
	    putString("You don't own \"");
	    putString(en->en_name);
	    putString("\".\n");
	    return(B_FALSE);
	}
	if (th->th_useCount == 1 && th->th_propCount != 0) {
	    putString("Thing \"");
	    putString(en->en_name);
	    putString("\" still has some properties.\n");
	    return(B_FALSE);
	}
	decRef(key);
	return(B_TRUE);
    case TYPE_ACTION:
	/* Removing a proc from a symbol table is normally fine, but we
	   don't want to allow anyone other than the owner to delete a
	   symbol from the public symbol table. */
	if ((key & INDEX_MASK) >= SMALLEST_BUILTIN) {
	    putString("You can't delete a builtin proc.\n");
	    return(B_FALSE);
	}
	pr = findCachedProcNoLock(key);
	if (pr->pr_owner != player && player != SYSADMIN_KEY && ! isPrivate) {
	    putString("You don't own \"");
	    putString(en->en_name);
	    putString("\".\n");
	    return(B_FALSE);
	}
	decProcRef(key);
	return(B_TRUE);
    case TYPE_TABLE:
	tb = findCachedTable(key);
	if (tb->tb_owner != player && player != SYSADMIN_KEY && ! isPrivate) {
	    putString("You don't own \"");
	    putString(en->en_name);
	    putString("\".\n");
	    return(B_FALSE);
	}
	if (tb->tb_useCount == 1 && tb->tb_usedCount != 0) {
	    putString("Table \"");
	    putString(en->en_name);
	    putString("\" still has entries.\n");
	    return(B_FALSE);
	}
	decRef(key);
	return(B_TRUE);
    case TYPE_GRAMMAR:
	g = findCachedGrammar(key);
	if (g->g_words.tb_owner != player && player != SYSADMIN_KEY &&
	    ! isPrivate)
	{
	    putString("You don't own \"");
	    putString(en->en_name);
	    putString("\".\n");
	    return(B_FALSE);
	}
	decRef(key);
	return(B_TRUE);
    case TYPE_PROP_BOOLEAN:
    case TYPE_PROP_INTEGER:
    case TYPE_PROP_FIXED:
    case TYPE_PROP_THING:
    case TYPE_PROP_ACTION:
    case TYPE_PROP_STRING:
    case TYPE_PROP_TABLE:
    case TYPE_PROP_GRAMMAR:
    case TYPE_PROP_LIST_INTEGER:
    case TYPE_PROP_LIST_FIXED:
    case TYPE_PROP_LIST_THING:
    case TYPE_PROP_LIST_ACTION:
	flushRefKey(key);
	prp = ioRead(key);
	if (prp->prp_owner != player && player != SYSADMIN_KEY && ! isPrivate){
	    putString("You don't own \"");
	    putString(en->en_name);
	    putString("\".\n");
	    return(B_FALSE);
	}
	/* Since there is no way to reference a property other than by its
	   name from a symbol table, we cannot allow it to go away if it has
	   other references. This is a problem if we make a synonym for it
	   in this or another symbol table. Oh well. */
	if (prp->prp_useCount != 1) {
	    putString("\"");
	    putString(en->en_name);
	    putString("\" has other references.\n");
	    return(B_FALSE);
	}
	decRef(key);
	return(B_TRUE);
    default:
	putString("?????\n");
    }
    return(B_FALSE);
}
