#include "MUD.h"
#include "types.h"
#include "exec.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"
#include "ClientStuff.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * miscFuncs.c - some miscellaneous builtins and their support routines.
 */

#define DATE_LEN	24

typedef struct Path {
    struct Path *p_next;
    CHAR_T *p_string;
} Path_t;

/*
 * GiveThing - give the indicated thing to the indicated character. This is
 *	needed to allow the "owner" of the thing to destroy it.
 */

void
GiveThing(ULONG_T thKey, ULONG_T plKey)
{
    Thing_t *th;

    TRACE("GiveThing", thKey);
    if (thKey == NIL_THING) {
	runError("nil thing to GiveThing");
    } else if (plKey == NIL_PLAYER) {
	runError("nil character to GiveThing");
    } else {
	th = ioRead(thKey);
	if (th->th_owner != EffectivePlayer && EffectivePlayer != SYSADMIN_KEY)
	{
	    runError("you do not own the thing (GiveThing)");
	} else {
	    th->th_owner = plKey;
	    ioDirty(thKey);
	}
    }
}

/*
 * ShowTable - builtin to show a table.
 */

void
ShowTable(ULONG_T key)
{

    if (key == NIL_TABLE) {
	runError("nil table to ShowTable");
    } else {
	showSymbols(findCachedTable(key),
		    "Symbols in table", "Table has no entries");
    }
}

/*
 * DescribeSymbol - interface for the builtin.
 */

void
DescribeSymbol(ULONG_T key, CHAR_T *name)
{

    if (key == NIL_TABLE) {
	runError("nil table passed to DescribeSymbol");
    } else {
	describeSymbol(findCachedTable(key), name);
    }
}

/*
 * Flush - the builtin - flush everything to disk.
 */

void
Flush(void)
{

    flushClients();
    flushCaches();
    ioFlush(B_TRUE);
}

/*
 * ShowClients - print info about all currently connect clients.
 *	Return the count of them found.
 */

ULONG_T
ShowClients(ULONG_T flag)
{
    CHAR_T dateBuffer[DATE_LEN + 1];
    Client_t *cl;
    ULONG_T clientCount, saveIndent;
    BOOL_T wantLong;

    TRACE("ShowClients", flag);
    wantLong = (flag != BOOLEAN_FALSE);
    ConvTime(GetCurrentTime(), &dateBuffer[0]);
    dateBuffer[DATE_LEN] = '\0';
    putString("Active clients as of ");
    putString(&dateBuffer[0]);
    putChar(':');
    if (wantLong) {
	newLine();
	saveIndent = 0;	/* stupid gcc! */
    } else {
	saveIndent = GetIndent();
	SetIndent(saveIndent + 2);
    }
    clientCount = 0;
    cl = Clients;
    while (cl != P_NULL) {
	if (cl->cl_player != NIL_PLAYER) {
	    clientCount += 1;
	    if (! wantLong) {
		putChar(' ');
	    }
	    putString(cl->cl_name);
	    if (wantLong) {
		putString(": on at ");
		ConvTime(cl->cl_onAt, &dateBuffer[0]);
		putString(&dateBuffer[0]);
		if (cl->cl_wizardMode) {
		    putString(" (wizarding)");
		}
		newLine();
	    }
	}
	cl = cl->cl_next;
    }
    if (! wantLong) {
	SetIndent(saveIndent);
	newLine();
    }
    return clientCount;
}

/*
 * fmt60 - helper routine for doShowCharacter.
 */

static void
fmt60(ULONG_T n)
{
    fmtChar(':');
    fmtChar(n / 10 + '0');
    fmtChar(n % 10 + '0');
}

/*
 * doShowCharacter - innards of ShowCharacter and ShowCharacters(true)
 */

static void
doShowCharacter(ULONG_T key)
{
    CHAR_T dateBuffer[DATE_LEN + 1];
    ULONG_T n, seconds, minutes;
    Player_t *pl;
    Client_t *cl;

    flushPlayer(key);
    cl = getClientByPlayer(key);
    pl = ioRead(key);
    if (cl == P_NULL) {
	fmtString(" last on ");
	ConvTime(pl->pl_lastOn, &dateBuffer[0]);
	fmtString(&dateBuffer[0]);
    } else {
	fmtString(" on at ");
	ConvTime(cl->cl_onAt, &dateBuffer[0]);
	fmtString(&dateBuffer[0]);
    }
    fmtChar(' ');
    switch (pl->pl_status) {
    case ps_normal:
	fmtString("normal");
	break;
    case ps_apprentice:
	fmtString("apprentice to ");
	fmtPlayerName(pl->pl_sponsor);
	break;
    case ps_wizard:
	fmtString("wizard sponsored by ");
	fmtPlayerName(pl->pl_sponsor);
	break;
    }
    if (pl->pl_newPlayer) {
	fmtString(" (new character)");
    } else {
	fmtString(" (");
	n = pl->pl_logonCount;
	if (cl != P_NULL) {
	    n += 1;
	}
	fmtInteger(n);
	fmtString(" logons, ");
	n = pl->pl_totalSeconds;
	if (cl != P_NULL) {
	    n += (GetCurrentTime() - cl->cl_onAt);
	}
	seconds = n % 60;
	n /= 60;
	minutes = n % 60;
	n /= 60;
	fmtInteger(n);
	fmt60(minutes);
	fmt60(seconds);
	fmtString(" total time)");
    }
    fmtLine();
}

/*
 * ShowCharacters - print info about all currently existing characters.
 *	Return the count of them found.
 */

ULONG_T
ShowCharacters(ULONG_T flag)
{
    CHAR_T dateBuffer[DATE_LEN + 1];
    Entry_t *en;
    UINT_T count;
    ULONG_T playerCount, saveIndent;
    BOOL_T wantLong;

    TRACE("ShowCharacters", flag);
    fmtStart();
    wantLong = (flag != BOOLEAN_FALSE);
    ConvTime(GetCurrentTime(), &dateBuffer[0]);
    dateBuffer[DATE_LEN] = '\0';
    fmtString("Existing characters as of ");
    fmtString(&dateBuffer[0]);
    fmtChar(':');
    if (wantLong) {
	fmtLine();
	saveIndent = 0;	/* shut up gcc */
    } else {
	saveIndent = GetIndent();
	SetIndent(saveIndent + 2);
    }
    playerCount = 0;
    en = PlayerTable->tl_this->tb_entries;
    count = PlayerTable->tl_this->tb_entryCount;
    while (count != 0) {
	count -= 1;
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE &&
	    en->en_key >> TYPE_SHIFT == TYPE_PLAYER)
	{
	    playerCount += 1;
	    if (! wantLong) {
		fmtChar(' ');
	    }
	    fmtString(en->en_name);
	    if (wantLong) {
		doShowCharacter(en->en_key);
	    }
	}
	++en;
    }
    if (! wantLong) {
	SetIndent(saveIndent);
	fmtLine();
    }
    return playerCount;
}

/*
 * ShowCharacter - show long form for the given character.
 */

void
ShowCharacter(ULONG_T key)
{

    TRACE("ShowCharacter", key);
    fmtStart();
    fmtPlayerName(key);
    doShowCharacter(key);
}

/*
 * CreateTable - builtin to create a new table
 */

ULONG_T
CreateTable(void)
{
    Table_t tb;
    BYTE_T *bytes;
    UINT_T len;

    symCreateE(&MS, &tb, 4);
    tb.tb_owner = EffectivePlayer;
    symWriteE(&MS, &tb);
    symDestroyE(&MS, &tb);
    writeResultE(&MS, &bytes, &len);
    return ioCreate(TYPE_TABLE, bytes, len);
}

/*
 * DefineThing - allow a program to define a private name for a thing.
 */

ULONG_T
DefineThing(ULONG_T table, CHAR_T *name, ULONG_T key)
{

    TRACE("DefineThing", key);
    return defineXXX(table, name, "Thing", key, NIL_THING);
}

/*
 * DefineAction - allow a program to define a private name for an action.
 */

ULONG_T
DefineAction(ULONG_T table, CHAR_T *name, ULONG_T key)
{

    TRACE("DefineAction", key);
    return defineXXX(table, name, "Action", key, NIL_ACTION);
}

/*
 * DefineFlag - allow a program to define a private name for a bool prop.
 */

ULONG_T
DefineFlag(ULONG_T table, CHAR_T *name, ULONG_T key)
{

    TRACE("DefineFlag", (ULONG_T) name);
    return defineXXX(table, name, "Flag", key, NIL_PROP_BOOLEAN);
}

/*
 * DefineCounter - allow a program to define a private name for an int prop.
 */

ULONG_T
DefineCounter(ULONG_T table, CHAR_T *name, ULONG_T key)
{

    TRACE("DefineCounter", (ULONG_T) name);
    return defineXXX(table, name, "Counter", key, NIL_PROP_INTEGER);
}

/*
 * DefineString - allow a program to define a private name for a string prop.
 */

ULONG_T
DefineString(ULONG_T table, CHAR_T *name, ULONG_T key)
{

    TRACE("DefineString", (ULONG_T) name);
    return defineXXX(table, name, "String", key, NIL_PROP_STRING);
}

/*
 * DefineTable - allow a program to define a private name for a new table.
 */

ULONG_T
DefineTable(ULONG_T table, CHAR_T *name, ULONG_T key)
{

    TRACE("DefineTable", (ULONG_T) name);
    return defineXXX(table, name, "Table", key, NIL_TABLE);
}

/*
 * LookupThing - look for a thing symbol in the active tables.
 *	Return nil if the symbol is not defined or is not a thing.
 */

ULONG_T
LookupThing(ULONG_T table, CHAR_T *name)
{

    return lookupXXX(name, "Thing", TYPE_THING, table, NIL_THING);
}

/*
 * LookupAction - look for an action symbol in the active tables.
 *	Return nil if the symbol is not defined or is not an action.
 */

ULONG_T
LookupAction(ULONG_T table, CHAR_T *name)
{

    return lookupXXX(name, "Action", TYPE_ACTION, table, NIL_ACTION);
}

/*
 * LookupFlag - look for a flag symbol in the active tables.
 *	Return nil if the symbol is not defined or is not a bool prop.
 */

ULONG_T
LookupFlag(ULONG_T table, CHAR_T *name)
{

    return lookupXXX(name, "Flag", TYPE_PROP_BOOLEAN, table, NIL_PROP_BOOLEAN);
}

/*
 * LookupCounter - look for a counter symbol in the active tables.
 *	Return nil if the symbol is not defined or is not an int prop.
 */

ULONG_T
LookupCounter(ULONG_T table, CHAR_T *name)
{

    return lookupXXX(name, "Counter",TYPE_PROP_INTEGER,table,NIL_PROP_INTEGER);
}

/*
 * LookupString - look for a string symbol in the active tables.
 *	Return nil if the symbol is not defined or is not a string prop.
 */

ULONG_T
LookupString(ULONG_T table, CHAR_T *name)
{

    return lookupXXX(name, "String", TYPE_PROP_STRING, table, NIL_PROP_STRING);
}

/*
 * LookupTable - look for a table symbol in the active tables.
 *	Return nil if the symbol is not defined or is not a table.
 */

ULONG_T
LookupTable(ULONG_T table, CHAR_T *name)
{

    return lookupXXX(name, "Table", TYPE_TABLE, table, NIL_TABLE);
}

/*
 * FindXXXSymbol - see if a given XXX is defined in the given table.
 *	If so, return the string it is entered under.
 */

CHAR_T *
FindXXXSymbol(ULONG_T table, ULONG_T key)
{
    Client_t *cl;
    TableList_t *tl;
    NameMap_t *nm;
    CHAR_T *p;
    BOOL_T doUse;

    TRACE("FindXXXSymbol", key);
    /* need a pointer to the namemap, so have to do this the hard way */
    doUse = B_FALSE;
    tl = P_NULL;    /* shut up gcc */
    if (table == NIL_TABLE) {
	p = findAnyName(key);
    } else {
	if (table == BuiltinTable->tl_key) {
	    nm = BuiltinTable->tl_nameMap;
	} else if (table == PlayerTable->tl_key) {
	    nm = PlayerTable->tl_nameMap;
	} else if (table == PublicSymbolTableKey) {
	    nm = PublicNameMap;
	} else {
	    cl = ActiveClient;
	    if (cl != P_NULL && table == cl->cl_privateTableKey) {
		nm = cl->cl_privateMap;
	    } else {
		tl = addGlobalUse(findCachedTable(table), key);
		nm = tl->tl_nameMap;
		doUse = B_TRUE;
	    }
	}
	p = findName(nm, key);
	if (doUse) {
	    removeGlobalUse(tl);
	}
    }
    if (p == P_NULL) {
	p = nullString();
    } else {
	p = stringCopy(p);
    }
    return p;
}

/*
 * PrintAction - allow examining an action explicitly.
 */

void
PrintAction(ULONG_T key)
{

    TRACE("PrintAction", key);
    if (key == NIL_ACTION) {
	runError("nil action passed to PrintAction");
    } else {
	fmtStart();
	fmtProc(findCachedProcNoLock(key), findAnyName(key));
	fmtLine();
    }
}

static void
printPath(Path_t *p)
{

    if (p->p_next != P_NULL) {
	printPath(p->p_next);
	fmtChar('/');
    }
    fmtString(p->p_string);
}
	
static BOOL_T
scanTable(Table_t *tb, ULONG_T key, Path_t *p, BOOL_T oneOnly)
{
    Entry_t *en;
    UINT_T count;
    Path_t path;
    BOOL_T foundOne;

    foundOne = B_FALSE;
    path.p_next = p;
    count = tb->tb_entryCount;
    en = tb->tb_entries;
    while (count != 0 && (! foundOne || ! oneOnly)) {
	count -= 1;
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
	    path.p_string = en->en_name;
	    if (en->en_key == key) {
		if (oneOnly) {
		    fmtChar('<');
		}
		printPath(&path);
		if (oneOnly) {
		    fmtChar('>');
		} else {
		    newLine();
		}
		foundOne = B_TRUE;
	    } else if ((en->en_key & TYPE_MASK) == (TYPE_TABLE << TYPE_SHIFT)){
		if (scanTable(findCachedTable(en->en_key),key,&path,oneOnly)) {
		    foundOne = B_TRUE;
		}
	    }
	}
	++en;
    }
    return foundOne;
}

BOOL_T
doFindKey(ULONG_T key, BOOL_T oneOnly)
{
    Table_t *tb;
    Entry_t *en;
    Player_t *pl;
    UINT_T count;
    Path_t path;
    BOOL_T foundOne;

    foundOne = scanTable(PublicSymbolTable, key, P_NULL, oneOnly);
    path.p_next = P_NULL;
    tb = PlayerTable->tl_this;
    count = tb->tb_entryCount;
    en = tb->tb_entries;
    while (count != 0 && (! foundOne || ! oneOnly)) {
	count -= 1;
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
	    path.p_string = en->en_name;
	    pl = ioRead(en->en_key);
	    if (scanTable(findCachedTable(pl->pl_symbolTableKey),
			  key, &path, oneOnly))
	    {
		foundOne = B_TRUE;
	    }
	}
	++en;
    }
    return foundOne;
}

/*
 * ScanTable - call a function for each entry in a table.
 */

void
ScanTable(ULONG_T table, ULONG_T action)
{
    Table_t *tb;
    Entry_t *en;
    UINT_T count;

    if (table == NIL_TABLE) {
	runError("nil table to ScanTable");
    } else if (action == NIL_ACTION) {
	runError("nil action to ScanTable");
    } else if (checkStringAction(action, "ScanTable")) {
	tb = findCachedTable(table);
	count = tb->tb_entryCount;
	en = tb->tb_entries;
	while (count != 0) {
	    count -= 1;
	    if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
		(void) doProc1(action, stringCopy(en->en_name));
	    }
	    ++en;
	}
    }
}

CHAR_T *
findPlayerThing(ULONG_T key)
{
    Table_t *tb;
    Entry_t *en;
    Player_t *pl;
    UINT_T count;

    tb = PlayerTable->tl_this;
    count = tb->tb_entryCount;
    en = tb->tb_entries;
    while (count != 0) {
	count -= 1;
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
	    pl = ioRead(en->en_key);
	    if (pl->pl_key == key) {
		return en->en_name;
	    }
	}
	++en;
    }
    return P_NULL;
}

CHAR_T *
findPlayerTable(ULONG_T key)
{
    Table_t *tb;
    Entry_t *en;
    Player_t *pl;
    UINT_T count;

    tb = PlayerTable->tl_this;
    count = tb->tb_entryCount;
    en = tb->tb_entries;
    while (count != 0) {
	count -= 1;
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
	    pl = ioRead(en->en_key);
	    if (pl->pl_symbolTableKey == key) {
		return en->en_name;
	    }
	}
	++en;
    }
    return P_NULL;
}

/*
 * SymbolType - the builtin.
 */

CHAR_T *
SymbolType(ULONG_T theTable, CHAR_T *name)
{
    Entry_t *en;
    CHAR_T buffer[30];

    if (theTable == NIL_TABLE) {
	return nullString();
    } else {
	en = symLookupE(findCachedTable(theTable), name, strLen1(name));
	if (en == P_NULL) {
	    return nullString();
	} else {
	    getTypeName(&buffer, en->en_key >> TYPE_SHIFT);
	    return stringCopy(&buffer[0]);
	}
    }
}

/*
 * ZapAction - the builtin.
 */

void
ZapAction(ULONG_T key)
{
    Proc_t *pr;
    LocalList_t *ll;
    UINT_T localCount;
    Proc_t newPr;
    Exec_t body;
    BYTE_T *bytes;
    UINT_T len;

    if (key == NIL_ACTION) {
	runError("nil action to ZapAction");
    } else if ((key & INDEX_MASK) >= SMALLEST_BUILTIN) {
	runError("can't Zap a builtin");
    } else {
	pr = findCachedProcNoLock(key);
	if (pr->pr_owner != EffectivePlayer && EffectivePlayer != SYSADMIN_KEY)
	{
	    runError("you don't own that proc");
	} else if (pr->pr_locked) {
	    runError("proc is byte-code locked and cannot be zapped");
	} else {
	    newPr = *pr;
	    ll = pr->pr_parameters;
	    newPr.pr_parameters = ll;
	    pr->pr_parameters = P_NULL;
	    localCount = BC_STACK_EXTRA;
	    while (ll != P_NULL) {
		localCount = localCount + sizeof(QValue_t);
		ll = ll->ll_next;
	    }
	    (void) removeProc(key, P_NULL);
	    newPr.pr_locals = P_NULL;
	    newPr.pr_refs = P_NULL;
	    newPr.pr_body = &body;
	    if (nilType(newPr.pr_resultType)) {
		body.ex_kind = ex_nil;
		body.ex_v.ex_integer = NIL_NIL;
	    } else {
		body.ex_v.ex_integer = 0;
		switch (newPr.pr_resultType) {
		case TYPE_VOID:
		    newPr.pr_body = P_NULL;
		    break;
		case TYPE_STATUS:
		    body.ex_kind = ex_status;
		    body.ex_v.ex_integer = STATUS_SUCCEED;
		    break;
		case TYPE_BOOLEAN:
		    body.ex_kind = ex_bool;
		    break;
		case TYPE_INTEGER:
		    body.ex_kind = ex_int;
		    break;
		case TYPE_FIXED:
		    body.ex_kind = ex_fixed;
		    break;
		case TYPE_STRING:
		    body.ex_kind = ex_string;
		    body.ex_v.ex_stringPtr = nullString();
		    break;
		case TYPE_THING_STATUS:
		    body.ex_kind = ex_ts;
		    break;
		}
	    }
	    newPr.pr_localCount = localCount;
	    newPr.pr_byteCode = P_NULL;	/* released by 'removeProc' */
	    newPr.pr_byteCodeLen = 0;
	    newPr.pr_isCompiled = B_FALSE;
	    newPr.pr_stripped = B_FALSE;
	    writeProcE(&MS, &newPr);
	    writeResultE(&MS, &bytes, &len);
	    ioReplace(key, bytes, len);
	    if (newPr.pr_resultType == TYPE_STRING) {
		freeString(body.ex_v.ex_stringPtr);
	    }
	    while (B_TRUE) {
		ll = newPr.pr_parameters;
		if (ll == P_NULL) {
		    break;
		}
		newPr.pr_parameters = ll->ll_next;
		freeString(ll->ll_name);
		memFreeE(&MS, ll, sizeof(LocalList_t));
	    }
	}
    }
}

/*
 * ZapTable - the builtin.
 */

void
ZapTable(ULONG_T key)
{
    Table_t *tb;
    Entry_t *en;
    TableList_t *tl;
    Thing_t *th;
    Proc_t *pr;
    Table_t *tb2;
    Grammar_t *g;
    ULONG_T effectivePlayer, pKey;
    UINT_T count, saveCount;
    BOOL_T isCharacters, isPublic, doNotDelete;

    if (key == NIL_TABLE) {
	runError("nil table to ZapTable");
    } else if (key == BuiltinTable->tl_key) {
	runError("can't ZapTable Builtin");
    } else {
	tb = findCachedTable(key);
	if (tb->tb_owner != EffectivePlayer && EffectivePlayer != SYSADMIN_KEY)
	{
	    runError("you don't own that table");
	} else {
	    isCharacters = (key == PlayerTable->tl_key);
	    isPublic = (key == PublicSymbolTableKey);
	    effectivePlayer = EffectivePlayer;
	    tl = TableList;
	    while (tl != P_NULL && tl->tl_this != tb) {
		tl = tl->tl_next;
	    }
	    if (tl != P_NULL && ! isCharacters) {
		/* Avoid dangling references! Because of the way namemaps
		   work, this isn't necessary, but it is cleaner. */
		emptyNameMap(tl->tl_nameMap);
	    }
	    /* Do this so that if we by chance get into here recursively,
	       the recursive call will not do anything. */
	    saveCount = tb->tb_entryCount;
	    tb->tb_entryCount = 0;
	    tb->tb_dirty = B_TRUE;
	    count = saveCount;
	    en = tb->tb_entries;
	    while (count != 0) {
		if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
		    doNotDelete = isCharacters;
		    key = en->en_key;
		    switch (key >> TYPE_SHIFT) {
		    case TYPE_PLAYER:
			if (effectivePlayer == SYSADMIN_KEY) {
			    putString("Zapping character ");
			    putString(en->en_name);
			    newLine();
			    ZapCharacter(key);
			}
			break;
		    case TYPE_THING:
			th = ioRead(key);
			if (th->th_owner == effectivePlayer ||
			    effectivePlayer == SYSADMIN_KEY)
			{
			    putString("Zapping thing ");
			    putString(en->en_name);
			    newLine();
			    pKey = th->th_parent;
			    if (pKey != NIL_THING) {
				th->th_parent = NIL_THING;
				ioDirty(key);
				decRef(pKey);
			    }
			    /* Do this *after* the above pKey stuff, so that
			       we do not care if the thing is still in the
			       cache. */
			    nukeThing(key);
			}
			break;
		    case TYPE_ACTION:
			pr = findCachedProcNoLock(key);
			if (pr->pr_owner == effectivePlayer ||
			    effectivePlayer == SYSADMIN_KEY)
			{
			    putString("Zapping action ");
			    putString(en->en_name);
			    newLine();
			    ZapAction(key);
			}
			break;
		    case TYPE_TABLE:
			if (key == BuiltinTable->tl_key ||
			    key == PlayerTable->tl_key)
			{
			    /* Cannot delete those two from the public
			       table, but can delete them from other
			       tables. */
			    doNotDelete = isPublic;
			} else {
			    tb2 = findCachedTable(key);
			    if (tb2->tb_owner == effectivePlayer ||
				effectivePlayer == SYSADMIN_KEY)
			    {
				putString("Zapping table ");
				putString(en->en_name);
				newLine();
				ZapTable(key);
			    }
			}
			break;
		    case TYPE_GRAMMAR:
			g = findCachedGrammar(key);
			if (g->g_words.tb_owner == effectivePlayer ||
			    effectivePlayer == SYSADMIN_KEY)
			{
			    putString("Zapping grammar ");
			    putString(en->en_name);
			    newLine();
			    ZapGrammar(key);
			}
			break;
		    default:
			if (key == NAME_KEY || key == ICON_KEY) {
			    doNotDelete = B_TRUE;
			}
			break;
		    }
		    if (! doNotDelete) {
			decRef(key);
			memFreeE(&MS, en->en_name, en->en_nameLen);
			en->en_nameLen = EN_EMPTY;
			/* Doing the following call for each symbol in the
			   table is expensive! */
			clearRefs(key);
			/* Doing the following call for each symbol in the
			   table is extremely expensive - it requires a message
			   to each active client!!!!! */
			flushSymbol(key);
			tb->tb_usedCount -= 1;
		    }
		}
		count -= 1;
		++en;
	    }
	    tb->tb_entryCount = saveCount;
	    tb->tb_dirty = B_TRUE;
	}
    }
}
