#include <ctype.h>

#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * parse.c - routines to do the "natural language parsing".
 */

/*
 * A grammar consists of a symbol table and a wordlist. The entries in the
 * symbol table, during execution, point directly at the corresponding
 * wordlist structure. In a wordlist, the 'code' field is the unique
 * numeric code for this word, if the word is not a synonym. If it is a
 * synonym, then it is the code of the word it is a synonym of. The
 * 'string' field is used for synonyms, so that when a word is referenced
 * via a synonym, the proper original word can be substituted. When a
 * grammar is stored on disk, each word is represented only by its key,
 * which is one of:
 *	W_WORD << TYPE_SHIFT | code
 *	W_SYNONYM << TYPE_SHIFT | code-of-other-word
 *	W_VERBTAIL << TYPE_SHIFT | index-of-my-entry:
 *		my-code, key-of-action-proc
 *	W_VERB << TYPE_SHIFT | index-of-my_entry:
 *		my-code, alt-count,
 *		    {key-of-action-proc, (code-of-separator, pad, verb-type)}*
 */

typedef struct ObjectList {
    struct ObjectList *ol_next;
    CHAR_T *ol_this;
} ObjectList_t;

#define VERB_LEN	50
static CHAR_T VerbBuffer[VERB_LEN];
static CHAR_T *NamePointer, *IndirPointer;
static CHAR_T LastPunctuation;

/*
 * parseInit - needed just to set up VerbBuffer.
 */

void
parseInit(void)
{

#if W_VERB <= TYPE_LAST
%%% W_VERB <= TYPE_LAST - too many types
#endif

    VerbBuffer[0] = '\0';
    NamePointer = "";
    IndirPointer = "";
    LastPunctuation = ' ';
}

/*
 * Verb - builtin to return the latest verb parsed.
 */

CHAR_T *
Verb(void)
{

    return stringCopy(&VerbBuffer[0]);
}

/*
 * ItName - builtin to return the string of the latest direct object used.
 */

CHAR_T *
ItName(void)
{

    return stringCopy(NamePointer);
}

/*
 * WhoName - builtin to return the string of the latest indirect object used.
 */

CHAR_T *
WhoName(void)
{

    return stringCopy(IndirPointer);
}

/*
 * Punctuation - builtin to return the terminating character.
 */

CHAR_T *
Punctuation(void)
{
    CHAR_T buffer[2];

    buffer[0] = LastPunctuation;
    buffer[1] = '\0';
    return stringCopy(&buffer[0]);
}

/*
 * writeGrammar - write a grammar to the data file.
 *	Note: the in-memory grammar is not usable for parsing after this.
 */

void
writeGrammar(Grammar_t *g, ULONG_T key)
{
    Entry_t *en;
    WordList_t *wl;
    Alternative_t *alt;
    ULONG_T *pair, *pairs;
    UINT_T count, altCount;
    BYTE_T *ptr;
    UINT_T len;

    TRACE("writeGrammar", key);
    count = g->g_words.tb_entryCount;
    en = g->g_words.tb_entries;
    while (count != 0) {
	count -= 1;
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
	    wl = (WordList_t *) en->en_key;
	    if (wl->wl_key == (ULONG_T) W_VERB << TYPE_SHIFT) {
		/* must be a new verb */
		altCount = 0;
		alt = wl->wl_.wl_alternatives;
		while (alt != P_NULL) {
		    altCount += 1;
		    alt = alt->alt_next;
		}
		len = altCount * (2 * sizeof(ULONG_T)) + 2 * sizeof(ULONG_T);
		pair = MEM_ALLOC_E(&MS, len, "gmtp");
		pairs = pair;
		*pair++ = ((WordList_t *) en->en_key)->wl_code;
		*pair++ = altCount;
		alt = ((WordList_t *) en->en_key)->wl_.wl_alternatives;
		while (alt != P_NULL) {
		    *pair++ = alt->alt_action;
		    *pair++ =
			((ULONG_T) alt->alt_separator << 16) |
			(alt->alt_type - vt_noObj);
		    alt = alt->alt_next;
		}
		en->en_key = ioCreate(W_VERB, pairs, len);
		memFreeE(&MS, pairs, len);
	    } else {
		/* convert from internal (en_key is ptr to wl) to external
		   (en_key is key of db entry for word) form */
		en->en_key = wl->wl_key;
	    }
	}
	++en;
    }
    /* note: the symWriteE call MUST be first, because it is the one that
       clears the WritePos value */
    symWriteE(&MS, &g->g_words);
    writeLongE(&MS, g->g_nextCode);
    writeResultE(&MS, &ptr, &len);
    ioReplace(key, ptr, len);
}

/*
 * readGrammar - read a grammar from the data file.
 */

void
readGrammar(Grammar_t *g, ULONG_T key)
{
    WordList_t *wl, *wl2;
    Entry_t *en;
    ULONG_T (*fileStuff)[3];
    ULONG_T *pair;
    Alternative_t *alt;
    void *b;
    ULONG_T synCode;
    UINT_T count, altCount;

    TRACE("readGrammar", key);
    b = ioRead(key);
    b = symReadE(&MS, &g->g_words, b);
    g->g_nextCode = readLongE(&b);
    g->g_wordList = P_NULL;
    count = g->g_words.tb_entryCount;
    en = g->g_words.tb_entries;
    while (count != 0) {
	count -= 1;
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
	    wl = MEM_ALLOC_E(&MS, sizeof(WordList_t), "wdls");
	    wl->wl_next = g->g_wordList;
	    g->g_wordList = wl;
	    wl->wl_word = en->en_name;
	    wl->wl_key = en->en_key;
	    wl->wl_refCount = 0;
	    en->en_key = (ULONG_T) wl;
	    switch (wl->wl_key >> TYPE_SHIFT) {
	    case W_WORD:
		wl->wl_code = wl->wl_key & INDEX_MASK;
		break;
	    case W_SYNONYM:
		wl->wl_code = 0;
		break;
	    case W_VERBTAIL:
		fileStuff = ioRead(wl->wl_key);
		wl->wl_code = (*fileStuff)[0];
		wl->wl_.wl_action = (*fileStuff)[1];
		break;
	    case W_VERB:
		fileStuff = ioRead(wl->wl_key);
		wl->wl_code = (*fileStuff)[0];
		wl->wl_.wl_alternatives = P_NULL;
		altCount = (*fileStuff)[1];
		pair = &(*fileStuff)[2];
		while (altCount != 0) {
		    altCount -= 1;
		    alt = MEM_ALLOC_E(&MS, sizeof(Alternative_t), "alt_");
		    alt->alt_next = wl->wl_.wl_alternatives;
		    alt->alt_action = *pair++;
		    alt->alt_separator = *pair >> 16;
		    alt->alt_type = ((BYTE_T) *pair & 0xff) + vt_noObj;
		    ++pair;
		    wl->wl_.wl_alternatives = alt;
		}
		break;
	    default:
		myAbort("invalid word kind in readGrammar");
		break;
	    }
	}
	++en;
    }
    wl = g->g_wordList;
    while (wl != P_NULL) {
	if (wl->wl_key >> TYPE_SHIFT == W_SYNONYM) {
	    synCode = wl->wl_key & INDEX_MASK;
	    wl2 = g->g_wordList;
	    while (wl2 != P_NULL && wl2->wl_code != synCode) {
		wl2 = wl2->wl_next;
	    }
	    if (wl2 == P_NULL) {
		myAbort("Can't find synonym in wordlist.");
	    }
	    wl->wl_.wl_synonym = wl2;
	    wl2->wl_refCount += 1;
	}
	wl = wl->wl_next;
    }
}

/*
 * CreateGrammar - allow programmer to create a new grammar.
 */

ULONG_T
CreateGrammar(void)
{
    Grammar_t g;
    ULONG_T key;

    symCreateE(&MS, &g.g_words, 1);
    g.g_words.tb_owner = EffectivePlayer;
    g.g_wordList = P_NULL;
    g.g_nextCode = 1;
    /* magic number is size of stored grammar */
    key = ioCreate(TYPE_GRAMMAR, &g, 14);
    writeGrammar(&g, key);
    symDestroyE(&MS, &g.g_words);
    return key;
}

/*
 * freeGrammar - free the space associated with a grammar.
 */

void
freeGrammar(Grammar_t *g)
{
    WordList_t *wl;
    Alternative_t *alt;

    while (B_TRUE) {
	wl = g->g_wordList;
	if (wl == P_NULL) {
	    break;
	}
	g->g_wordList = wl->wl_next;
	if (wl->wl_key >> TYPE_SHIFT == W_VERB) {
	    while (B_TRUE) {
		alt = wl->wl_.wl_alternatives;
		if (alt == P_NULL) {
		    break;
		}
		wl->wl_.wl_alternatives = alt->alt_next;
		memFreeE(&MS, alt, sizeof(Alternative_t));
	    }
	}
	memFreeE(&MS, wl, sizeof(WordList_t));
    }
    symDestroyE(&MS, &g->g_words);
}

/*
 * releaseGrammar - free up all of the database entries related to a grammar.
 */

void
releaseGrammar(Grammar_t *g)
{
    WordList_t *wl;
    Alternative_t *alt;

    wl = g->g_wordList;
    while (wl != P_NULL) {
	switch (wl->wl_key >> TYPE_SHIFT) {
	case W_VERB:
	    putString("Zapping verb ");
	    putString(wl->wl_word);
	    newLine();
	    alt = wl->wl_.wl_alternatives;
	    while (alt != P_NULL) {
		decProcRef(alt->alt_action);
		alt = alt->alt_next;
	    }
	    if ((wl->wl_key & INDEX_MASK) != 0) {
		/* has an on-disk representation */
		ioDelete(wl->wl_key);
	    }
	    break;
	case W_VERBTAIL:
	    putString("Zapping VerbTail ");
	    putString(wl->wl_word);
	    newLine();
	    decProcRef(wl->wl_.wl_action);
	    ioDelete(wl->wl_key);
	    break;
	default:
	    putString("Zapping word/synonym ");
	    putString(wl->wl_word);
	    newLine();
	    break;
	}
	wl = wl->wl_next;
    }
}

/*
 * enterWord - utility to check/enter a word into the dictionary.
 *	BEWARE: this routine may cause the symbol table for the grammar
 *	to be expanded, which will invalidate any pointers into it.
 */

static WordList_t *
enterWord(Grammar_t *g, CHAR_T *wrd, ULONG_T key, BOOL_T isVerb)
{
    CHAR_T *p;
    Entry_t *en;
    WordList_t *wl;
    UINT_T len;

    TRACE("enterWord", key);
    if (g->g_words.tb_owner != EffectivePlayer &&
	(EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY))
    {
	runError("you don't own that grammar");
	return P_NULL;
    }
    p = wrd;
    len = 1;
    while (*p != '\0') {
	if (isupper(*p)) {
	    *p = tolower(*p);
	}
	++p;
	len += 1;
    }
    en = symLookupE(&g->g_words, wrd, len);
    if (en != P_NULL) {
	wl = (WordList_t *) en->en_key;
	if (! isVerb || wl->wl_key >> TYPE_SHIFT != W_VERB) {
	    symbolError(wrd, "already defined in grammar");
	    return P_NULL;
	}
	if ((wl->wl_key & INDEX_MASK) != 0) {
	    /* There is now one more alternative for this verb. That will
	       require a larger in-database record for them. So, delete the
	       current one, and set the key to indicate that there is no
	       current entry, so that 'writeGrammar' will make a new one. */
	    ioDelete(wl->wl_key);
	    wl->wl_key = (ULONG_T) W_VERB << TYPE_SHIFT;
	}
	g->g_words.tb_dirty = B_TRUE;
    } else {
	en = symEnterE(&MS, &g->g_words, wrd, len);
	wl = MEM_ALLOC_E(&MS, sizeof(WordList_t), "wdls");
	wl->wl_next = g->g_wordList;
	g->g_wordList = wl;
	wl->wl_word = en->en_name;
	wl->wl_code = g->g_nextCode;
	wl->wl_.wl_alternatives = P_NULL;      /* catches all union variants */
	wl->wl_key = key;
	wl->wl_refCount = 0;
	g->g_nextCode += 1;
	en->en_key = (ULONG_T) wl;
    }
    return wl;
}

/*
 * Word - add an individual word to a grammar. Return its code.
 */

ULONG_T
Word(ULONG_T key, CHAR_T *wrd)
{
    Grammar_t *g;

    g = findCachedGrammar(key);
    if (enterWord(g, wrd,
	((ULONG_T) W_WORD << TYPE_SHIFT) | g->g_nextCode, B_FALSE) != P_NULL)
    {
	return g->g_nextCode - 1;
    } else {
	return 0;
    }
}

/*
 * Synonym - define one word to be a synonym of another.
 */

void
Synonym(ULONG_T key, CHAR_T *oldWord, CHAR_T *newWord)
{
    Grammar_t *g;
    CHAR_T *p;
    Entry_t *en;
    WordList_t *wl;
    UINT_T len;

    p = oldWord;
    while (*p != '\0') {
	if (isupper(*p)) {
	    *p = tolower(*p);
	}
	++p;
    }
    len = strLen1(oldWord);
    g = findCachedGrammar(key);
    en = symLookupE(&g->g_words, oldWord, len);
    if (en == P_NULL) {
	symbolError(oldWord, "is not defined in grammar");
    } else {
	wl = (WordList_t *) en->en_key;
	if ((wl->wl_key >> TYPE_SHIFT) == W_SYNONYM) {
	    /* this works for W_VERB, too */
	    symbolError(oldWord, "can't make synonym of synonym");
	} else {
	    wl =
		enterWord(g, newWord,
			  ((ULONG_T) W_SYNONYM << TYPE_SHIFT) | wl->wl_code,
			  B_FALSE);
	    if (wl != P_NULL) {
		/* the call to 'enterWord' may have expanded the symbol
		   table of the grammar, thus invalidating 'en', so we go
		   get it again. */
		en = symLookupE(&g->g_words, oldWord, len);
		wl->wl_.wl_synonym = (WordList_t *) en->en_key;
		wl->wl_code = 0;
		g->g_nextCode -= 1;
		wl = (WordList_t *) en->en_key;
		wl->wl_refCount += 1;
	    }
	}
    }
}

/*
 * VerbTail - add a simple tail-string verb to a grammar.
 */

void
VerbTail(ULONG_T key, CHAR_T *wrd, ULONG_T action)
{
    Proc_t *pr;
    WordList_t *wl;
    ULONG_T fileStuff[2];

    if (action == NIL_ACTION) {
	runError("nil action given to VerbTail");
    } else {
	pr = findCachedProcNoLock(action);
	if (pr->pr_parameters != P_NULL || pr->pr_resultType != TYPE_BOOLEAN) {
	    runError(
		"VerbTail proc should have 0 parameters and bool result");
	} else {
	    wl = enterWord(findCachedGrammar(key), wrd, 0, B_FALSE);
	    if (wl != P_NULL) {
		fileStuff[0] = wl->wl_code;
		fileStuff[1] = action;
		wl->wl_key =
		    ioCreate(W_VERBTAIL, &fileStuff[0], 2 * sizeof(ULONG_T));
		wl->wl_.wl_action = action;
		incProcRef(action);
	    }
	}
    }
}

/*
 * addAlternative - add an alternative to a verb.
 */

static void
addAlternative(ULONG_T key, CHAR_T *wrd, ULONG_T sepCode, ULONG_T action,
	       VerbType_t vt)
{
    WordList_t *wl;
    Alternative_t *alt;

    wl = enterWord(findCachedGrammar(key), wrd,
		   (ULONG_T) W_VERB << TYPE_SHIFT, B_TRUE);
    if (wl != P_NULL) {
	alt = MEM_ALLOC_E(&MS, sizeof(Alternative_t), "alt_");
	alt->alt_next = wl->wl_.wl_alternatives;
	alt->alt_action = action;
	alt->alt_separator = sepCode;
	alt->alt_type = vt;
	wl->wl_.wl_alternatives = alt;
	incProcRef(action);
    }
}

/*
 * Verb0 - add a simple verb alternative to a grammar.
 */

void
Verb0(ULONG_T key, CHAR_T *wrd, ULONG_T sepCode, ULONG_T action)
{
    Proc_t *pr;

    if (action == NIL_ACTION) {
	runError("nil action given to Verb0");
    } else {
	pr = findCachedProcNoLock(action);
	if (pr->pr_parameters != P_NULL || pr->pr_resultType != TYPE_BOOLEAN) {
	    runError("Verb0 proc should have 0 parameters and bool result");
	} else {
	    addAlternative(key, wrd, sepCode, action, vt_noObj);
	}
    }
}

/*
 * Verb1 - add a verb with a direct object to a grammar.
 */

void
Verb1(ULONG_T key, CHAR_T *wrd, ULONG_T sepCode, ULONG_T action)
{
    Proc_t *pr;
    LocalList_t *ll;

    if (action == NIL_ACTION) {
	runError("nil action given to Verb1");
    } else {
	pr = findCachedProcNoLock(action);
	ll = pr->pr_parameters;
	if (pr->pr_resultType != TYPE_BOOLEAN || ll == P_NULL ||
	    ll->ll_type != TYPE_STRING || ll->ll_next != P_NULL)
	{
	    runError(
		"Verb1 proc should have 1 string parameter and bool result");
	} else {
	    addAlternative(key, wrd, sepCode, action, vt_dirObj);
	}
    }
}

/*
 * Verb2 - add a verb with direct and indirect objects to a grammar.
 */

void
Verb2(ULONG_T key, CHAR_T *wrd, ULONG_T sepCode, ULONG_T action)
{
    Proc_t *pr;
    LocalList_t *ll;

    if (action == NIL_ACTION) {
	runError("nil action given to Verb2");
    } else if (sepCode == 0) {
	runError("separator required for Verb2");
    } else {
	pr = findCachedProcNoLock(action);
	ll = pr->pr_parameters;
	if (ll == P_NULL || ll->ll_type != TYPE_STRING) {
	    ll = P_NULL;
	} else {
	    ll = ll->ll_next;
	    if (ll == P_NULL || ll->ll_type != TYPE_STRING ||
		ll->ll_next != P_NULL)
	    {
		ll = P_NULL;
	    }
	}
	if (ll == P_NULL || pr->pr_resultType != TYPE_BOOLEAN) {
	    runError(
	       "Verb2 proc should have 2 string parameters and bool result");
	} else {
	    addAlternative(key, wrd, sepCode, action, vt_twoObj);
	}
    }
}

/*
 * FindWord - look up a word in a grammer. Return its code.
 *	NOTE: if the word is a synonym, the code returned is 0, so most
 *	uses will want to use 'FindAnyWord' instead.
 */

ULONG_T
FindWord(ULONG_T key, CHAR_T *str)
{
    CHAR_T *p;
    Entry_t *en;

    p = str;
    while (*p != '\0') {
	if (isupper(*p)) {
	    *p = tolower(*p);
	}
	++p;
    }
    en = symLookupE(&findCachedGrammar(key)->g_words, str, strLen1(str));
    if (en == P_NULL) {
	return 0;
    } else {
	return ((WordList_t *) en->en_key)->wl_code;
    }
}

/*
 * FindAnyWord - look up a word in a grammer. Return its code.
 *	This variant will skip synonyms, so it can tell if any given word
 *	is in the grammar.
 */

ULONG_T
FindAnyWord(ULONG_T key, CHAR_T *str)
{
    CHAR_T *p;
    Entry_t *en;
    WordList_t *wl;

    p = str;
    while (*p != '\0') {
	if (isupper(*p)) {
	    *p = tolower(*p);
	}
	++p;
    }
    en = symLookupE(&findCachedGrammar(key)->g_words, str, strLen1(str));
    if (en == P_NULL) {
	return 0;
    } else {
	wl = (WordList_t *) en->en_key;
	if ((wl->wl_key >> TYPE_SHIFT) == W_SYNONYM) {
	    wl = wl->wl_.wl_synonym;
	}
	return wl->wl_code;
    }
}

/*
 * ShowWord - look up a word in a grammar. Display it.
 */

void
ShowWord(ULONG_T key, CHAR_T *str)
{
    Grammar_t *g;
    CHAR_T *p;
    WordList_t *wl;
    Entry_t *en;
    Alternative_t *alt;
    ULONG_T sepCode;

    g = findCachedGrammar(key);
    p = str;
    while (*p != '\0') {
	if (isupper(*p)) {
	    *p = tolower(*p);
	}
	++p;
    }
    en = symLookupE(&g->g_words, str, strLen1(str));
    putChar('\"');
    putString(str);
    putChar('\"');
    if (en == P_NULL) {
	putString(": not in grammar\n");
    } else {
	wl = (WordList_t *) en->en_key;
	if (wl->wl_refCount != 0) {
	    putString(" (");
	    putDecimal(wl->wl_refCount);
	    putString(" synonyms)");
	}
	putString(": ");
	switch (wl->wl_key >> TYPE_SHIFT) {
	case W_WORD:
	    putString("code ");
	    putDecimal(wl->wl_code);
	    break;
	case W_SYNONYM:
	    putString("synonym of \"");
	    putString(wl->wl_.wl_synonym->wl_word);
	    putChar('\"');
	    break;
	case W_VERBTAIL:
	    putString("code ");
	    putDecimal(wl->wl_code);
	    putString(" string tail verb => ");
	    fmtStart();
	    fmtActionName(wl->wl_.wl_action);
	    break;
	case W_VERB:
	    putString("code ");
	    putDecimal(wl->wl_code);
	    putString(" verb with alternatives:");
	    alt = wl->wl_.wl_alternatives;
	    while (alt != P_NULL) {
		putString("\n    ");
		switch (alt->alt_type) {
		case vt_noObj:
		    putString("no object");
		    break;
		case vt_dirObj:
		    putString("direct object");
		    break;
		case vt_twoObj:
		    putString("direct and indirect objects");
		    break;
		default:
		    myAbort("invalid verb kind in ShowWord");
		    break;
		}
		putString(", separator ");
		sepCode = alt->alt_separator;
		if (sepCode == 0) {
		    putString("<NONE>");
		} else {
		    wl = g->g_wordList;
		    while (wl != P_NULL && wl->wl_code != sepCode) {
			wl = wl->wl_next;
		    }
		    if (wl == P_NULL) {
			putString("????");
		    } else {
			putChar('\"');
			putString(wl->wl_word);
			putChar('\"');
		    }
		}
		putString(" => ");
		fmtStart();
		fmtActionName(alt->alt_action);
		alt = alt->alt_next;
	    }
	    break;
	default:
	    myAbort("invalid word kind in ShowWord");
	    break;
	}
	newLine();
    }
}

/*
 * ShowWords - show all of the words in the grammar.
 */

void
ShowWords(ULONG_T key)
{

    showSymbols(&findCachedGrammar(key)->g_words,
		"Words in grammar", "No words in grammar yet");
}

/*
 * DeleteWord - remove a word from a grammar.
 */

void
DeleteWord(ULONG_T key, CHAR_T *wrd)
{
    Grammar_t *g;
    CHAR_T *p;
    Entry_t *en;
    WordList_t *wl;
    Alternative_t *alt;
    WordList_t **wlp;
    UINT_T len;

    g = findCachedGrammar(key);
    if (g->g_words.tb_owner != EffectivePlayer &&
	(EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY))
    {
	runError("you don't own that grammar");
    } else {
	p = wrd;
	len = 1;
	while (*p != '\0') {
	    if (isupper(*p)) {
		*p = tolower(*p);
	    }
	    ++p;
	    len += 1;
	}
	en = symLookupE(&g->g_words, wrd, len);
	if (en == P_NULL) {
	    symbolError(wrd, "not defined in grammar");
	} else {
	    wl = (WordList_t *) en->en_key;
	    if (wl->wl_refCount != 0) {
		symbolError(wrd, "has existing synonyms - not deleted");
	    } else {
		symDeleteE(&MS, &g->g_words, wrd, len);
		switch (wl->wl_key >> TYPE_SHIFT) {
		case W_SYNONYM:
		    wl->wl_.wl_synonym->wl_refCount -= 1;
		    break;
		case W_VERBTAIL:
		    decProcRef(wl->wl_.wl_action);
		    ioDelete(wl->wl_key);
		    break;
		case W_VERB:
		    while (B_TRUE) {
			alt = wl->wl_.wl_alternatives;
			if (alt == P_NULL) {
			    break;
			}
			wl->wl_.wl_alternatives = alt->alt_next;
			decProcRef(alt->alt_action);
			memFreeE(&MS, alt, sizeof(Alternative_t));
		    }
		    ioDelete(wl->wl_key);
		    break;
		}
		wlp = &g->g_wordList;
		while (*wlp != wl) {
		    if (*wlp == P_NULL) {
			myAbort("word not on wordlist in 'DeleteWord'");
		    }
		    wlp = &(*wlp)->wl_next;
		}
		*wlp = wl->wl_next;
		memFreeE(&MS, wl, sizeof(WordList_t));
	    }
	}
    }
}

static BOOL_T
phraseEnd(CHAR_T ch)
{
    return (BOOL_T) (ch == '\0' || ch == ' ' || ch == ',' ||
		     ch == '.' || ch == ';' || ch == '!');
}

BOOL_T
sentEnd(CHAR_T ch)
{
    return (BOOL_T) (ch == '\0' || ch == '.' || ch == ';' || ch == '!');
}

/*
 * getNounPhrase - helper routine for 'Parse'. Strip a noun phrase off of
 *	the input sentence. The noun phrase can be empty. Return 'B_FALSE' if
 *	something goes wrong.
 */

static BOOL_T
getNounPhrase(Grammar_t *g, CHAR_T **pSentence, CHAR_T **pPhrase,
	      Alternative_t *alternatives)
{
#define PHRASE_LEN	100
    CHAR_T *p, *q, *r;
    Entry_t *en;
    WordList_t *wl;
    Alternative_t *alt;
    CHAR_T *savePtr;
    UINT_T len, count, saveLen;
    CHAR_T adjectives[PHRASE_LEN];
    CHAR_T countBuffer[6];
    CHAR_T ch;
    PlayerStatus_t saveStatus;

    TRACE("getNounPhrase", 0);

    /* first case - end of input - no noun phrase given */

    p = *pSentence;
    ch = *p;
    if (phraseEnd(ch)) {
	/* no noun phrase - could be OK */
	*pPhrase = nullString();
	return B_TRUE;
    }

    /* second case - a quoted string */

    q = p;
    ch = *q++;
    len = 0;
    if (ch == '\"') {
	/* a quoted string - swallow the whole thing */
	p = &adjectives[0];
	while (B_TRUE) {
	    ch = *q;
	    if (ch == '\"' || ch == '\0') {
		break;
	    }
	    if (len < PHRASE_LEN - 1) {
		len += 1;
		*p++ = ch;
	    }
	    ++q;
	}
	if (ch != '\0') {
	    ++q;
	}
	while (*q == ' ') {
	    ++q;
	}
	*pSentence = q;
	*p = '\0';
	len += 1;
	p = MEM_ALLOC_E(&MS, len, "nph1");
	*pPhrase = p;
	q = &adjectives[0];
	do {
	    ch = *q++;
	    *p++ = ch;
	} while (ch != '\0');
	return B_TRUE;
    }

    /* third case - a normal noun phrase - first skip any 'the', 'a' or 'an' */

    if (ch == 't' || ch == 'T') {
	ch = *q++;
	if (ch == 'h' || ch == 'H') {
	    ch = *q++;
	    if (ch == 'e' || ch == 'E') {
		ch = *q;
		if (phraseEnd(ch)) {
		    /* got a 'the' */
		    p = q;
		}
	    }
	}
    } else if (ch == 'a' || ch == 'A') {
	ch = *q;
	if (phraseEnd(ch)) {
	    /* got a 'a' */
	    p = q;
	} else if (ch == 'n' || ch == 'N') {
	    ++q;
	    ch = *q;
	    if (phraseEnd(ch)) {
		/* got a 'an' */
		p = q;
	    }
	}
    } else if (ch == 'l' || ch == 'L') {
	ch = *q++;
	if (ch == 'a' || ch == 'A' || ch == 'e' || ch == 'E' ||
	    ch == 'o' || ch == 'O')
	{
	    ch = *q;
	    if (phraseEnd(ch)) {
		/* got a 'la' or 'le' */
		p = q;
	    } else if (ch == 's' || ch == 'S') {
		ch = *q;
		if (phraseEnd(ch)) {
		    /* got a 'les' or 'los' (or 'las'!) */
		    p = q;
		}
	    }
	}
    }

    /* Run through words until end of input, 'and' or ',', or we hit a
       separator word. Put the words into 'adjectives', separated by commas */

    count = 0;
    r = &adjectives[0];
    while (B_TRUE) {
	while (B_TRUE) {
	    /* skip whitespace between words and after last one */
	    while (*p == ' ') {
		++p;
	    }
	    /* allow for a #<n> count specifier in the noun phrase */
	    q = p;
	    if (*p != '#') {
		break;
	    }
	    saveLen = count;
	    count = 0;
	    ++p;
	    while (isdigit(*p)) {
		count *= 10;
		count += (*p - '0');
		++p;
	    }
	    if (count == 0 || ! phraseEnd(*p)) {
		/* not valid - treat it as a normal word */
		p = q;
		count = saveLen;
		break;
	    }
	    /* Got a valid #<n> - continue */
	}

	/* Note that this means that 'adjectives' will start with a ','. */
	if (len < PHRASE_LEN - 1) {
	    len += 1;
	    *r++ = ',';
	}
	if (phraseEnd(*p)) {
	    /* no more adjectives */
	    break;
	}

	q = p;
	savePtr = r;
	saveLen = len;
	while (B_TRUE) {
	    ch = *q++;
	    if (phraseEnd(ch)) {
		break;
	    }
	    if (isupper(ch)) {
		ch = tolower(ch);
	    }
	    if (len < PHRASE_LEN - 1) {
		len += 1;
		*r++ = ch;
	    }
	}
	*r = '\0';
	len += 1;

	/* OK, got a potential adjective - is it a separator word? */

	if (g != P_NULL) {
	    en = symLookupE(&g->g_words, savePtr, len - saveLen);
	} else {
	    en = P_NULL;
	}
	if (en != P_NULL) {
	    wl = (WordList_t *) en->en_key;
	    if ((wl->wl_key >> TYPE_SHIFT) == W_SYNONYM) {
		wl = wl->wl_.wl_synonym;
	    }
	    len = wl->wl_code;
	    alt = alternatives;
	    while (alt != P_NULL && alt->alt_separator != len) {
		alt = alt->alt_next;
	    }
	} else {
	    alt = P_NULL;
	}
	r = savePtr;
	len = saveLen;
	if (alt != P_NULL) {
	    /* end while on separator word */
	    break;
	}

	/* OK, its not marked as a separator word in the grammar. */

	q = p;
	ch = *q++;
	if (ch == 'a' || ch == 'A') {
	    ch = *q++;
	    if (ch == 'n' || ch == 'N') {
		ch = *q++;
		if ((ch == 'd' || ch == 'D') && *q == ' ') {
		    /* end while on 'and' */
		    break;
		}
	    }
	}

	/* OK, not 'and' - copy it in as an adjective. (Don't use previous
	   copy, since we lower-cased that.) */

	while (B_TRUE) {
	    ch = *p++;
	    if (phraseEnd(ch)) {
		break;
	    }
	    if (len < PHRASE_LEN - 1) {
		len += 1;
		*r++ = ch;
	    }
	}
	--p;
    }
    /* We have consumed part of the input "sentence". */
    *pSentence = p;

    if (count != 0) {
	len += 1;			/* the ':' */
	q = &countBuffer[5];
	*q = '\0';
	do {
	    *--q = count % 10 + '0';
	    len += 1;
	    count /= 10;
	} while (count != 0);
	count = 1;			/* use as flag */
    }

    *--r = '\0';			/* overwrite trailing ',' */
    if (adjectives[0] == '\0') {

	/* no real words at all given */

	saveStatus = EffectiveStatus;
	EffectiveStatus = ps_wizard;
	putString("I need a noun in that sentence.\n");
	EffectiveStatus = saveStatus;
	return B_FALSE;
    }

    /* allocate a string for the noun phrase, and copy it in, taking the
       last word in 'adjectives' as the noun, and separating it from the
       rest by a ';' instead of the ','. If there is any count, put it at
       the start, with a ':' after it. */

    p = MEM_ALLOC_E(&MS, len, "nph2");
    *pPhrase = p;

    /* First comes any count, then a ':'. */
    if (count != 0) {
	while (*q != '\0') {
	    *p++ = *q++;
	}
	*p++ = ':';
    }

    /* skip back over the last adjective - it will be the noun */

    --r;
    while (*--r != ',') {
    }
    *r++ = '\0';			/* terminate the adjectives */

    /* copy in the noun */

    while (B_TRUE) {
	ch = *r++;
	if (ch == '\0') {
	    break;
	}
	*p++ = ch;
    }

    r = &adjectives[0];
    ch = *r++;
    if (ch != '\0') {

	/* if any adjectives, add ';' and copy in the adjectives */

	*p++ = ';';
	while (B_TRUE) {
	    ch = *r++;
	    if (ch == '\0') {
		break;
	    }
	    *p++ = ch;
	}
    }
    *p = '\0';

    return B_TRUE;
#undef PHRASE_LEN
}

/*
 * GetNounPhrase - external interface to allow caller to pull a noun phrase
 *	out of a passed string.
 */

CHAR_T *
GetNounPhrase(ULONG_T key, CHAR_T *st, ULONG_T sepCode)
{
    extern void SetTail(CHAR_T *str);
    CHAR_T *np, *sp;
    Alternative_t alt;

    alt.alt_next = P_NULL;
    alt.alt_separator = sepCode;
    sp = st;
    if (getNounPhrase(findCachedGrammar(key), &sp, &np, &alt)) {
	SetTail(sp);
	return np;
    } else {
	return nullString();
    }
}

/*
 * skipSeparator - check for and skip any separator found.
 */

static Alternative_t *
skipSeparator(Grammar_t *g, Alternative_t *alt, CHAR_T **sentence)
{
#define SEP_LEN 	20
    CHAR_T sepBuffer[SEP_LEN];
    CHAR_T *p, *q;
    Entry_t *en;
    WordList_t *wl;
    UINT_T len;
    CHAR_T ch;

    /* pick off the first word */

    p = *sentence;
    q = &sepBuffer[0];
    len = 0;
    while (B_TRUE) {
	ch = *p++;
	if (phraseEnd(ch)) {
	    break;
	}
	if (isupper(ch)) {
	    ch = tolower(ch);
	}
	if (len < SEP_LEN - 1) {
	    len += 1;
	    *q++ = ch;
	}
    }
    *q = '\0';
    len += 1;

    /* if it is one of our separators, skip it and the whitespace after it */

    en = symLookupE(&g->g_words, &sepBuffer[0], len);
    if (en != P_NULL) {
	wl = (WordList_t *) en->en_key;
	if ((wl->wl_key >> TYPE_SHIFT) == W_SYNONYM) {
	    wl = wl->wl_.wl_synonym;
	}
	len = wl->wl_code;
	while (alt != P_NULL && alt->alt_separator != len) {
	    alt = alt->alt_next;
	}
	if (alt != P_NULL && len != 0) {
	    /* found a separator word - skip it and any trailing spaces */
	    --p;
	    while (*p == ' ') {
		++p;
	    }
	    *sentence = p;
	}
    } else {
	alt = P_NULL;
    }
    return alt;
#undef SEP_LEN
}

/*
 * putSeparator - put out the separator if we have one.
 */

static void
putSeparator(Grammar_t *g, UINT_T sepCode)
{
    WordList_t *wl;

    if (sepCode != 0) {
	wl = g->g_wordList;
	while (wl != P_NULL && wl->wl_code != sepCode) {
	    wl = wl->wl_next;
	}
	if (wl != P_NULL) {
	    putChar(' ');
	    putString(wl->wl_word);
	}
    }
}

/*
 * freeObjects - free up a list of object strings.
 */

static void
freeObjects(ObjectList_t *objects)
{
    ObjectList_t *ol;

    while (B_TRUE) {
	ol = objects;
	if (ol == P_NULL) {
	    break;
	}
	objects = ol->ol_next;
	freeString(ol->ol_this);
	memFreeE(&MS, ol, sizeof(ObjectList_t));
    }
}

/*
 * complain - utility to complain about syntax problems.
 */

static void
complain(Grammar_t *g, CHAR_T *prev, CHAR_T *verb, CHAR_T *post,
	 Alternative_t *alt, WordList_t *syn)
{
    PlayerStatus_t saveStatus;

    /* avoid the '@' */
    saveStatus = EffectiveStatus;
    EffectiveStatus = ps_wizard;
    if (prev != P_NULL) {
	putString(prev);
    } else {
	if (islower(*verb)) {
	    *verb = toupper(*verb);
	}
    }
    putChar('\"');
    putString(verb);
    if (alt != P_NULL && syn == P_NULL) {
	putSeparator(g, alt->alt_separator);
    }
    putChar('\"');
    if (syn != P_NULL) {
	putString(" (\"");
	putString(syn->wl_word);
	if (alt != P_NULL) {
	    putSeparator(g, alt->alt_separator);
	}
	putString("\")");
    }
    if (post != P_NULL) {
	putString(post);
    } else {
	newLine();
    }
    EffectiveStatus = saveStatus;
}

/*
 * complain2 - complain about use of a two-object verb.
 */

static void
complain2(Grammar_t *g, CHAR_T *verb, CHAR_T *obj, CHAR_T *indir,
	  Alternative_t *alt, WordList_t *syn)
{
    PlayerStatus_t saveStatus;

    saveStatus = EffectiveStatus;
    EffectiveStatus = ps_wizard;
    if (islower(*verb)) {
	*verb = toupper(*verb);
    }
    putChar('\"');
    putString(verb);
    putChar('\"');
    if (syn != P_NULL) {
	putString(" (\"");
	putString(syn->wl_word);
	putString("\")");
    }
    putChar(' ');
    if (obj != P_NULL) {
	putString(obj);
    } else {
	putString("who/what");
    }
    putSeparator(g, alt->alt_separator);
    putChar(' ');
    if (indir != P_NULL) {
	putString(indir);
    } else {
	putString("who/what");
    }
    putString("?\n");
    EffectiveStatus = saveStatus;
}

/*
 * parseOne - parse the given sentence according to a grammar. This
 *	involves only identifying the verb and any noun phrases. The action
 *	associated with the verb is then called to handle the noun phrases.
 *	'parseOne' does one command from the input string.
 */

static CHAR_T *
parseOne(Grammar_t *g, CHAR_T *sentence)
{
    CHAR_T *p, *q, *dirObj, *indirObj;
    Entry_t *en;
    WordList_t *wl;
    Alternative_t *alt1, *alternatives, *alt2;
    ObjectList_t **objectTail;
    ObjectList_t *objects, *ol;
    ULONG_T key;
    UINT_T len, typ;
    CHAR_T ch;
    BOOL_T hadAlt1, found;

    TRACE("parseOne", sentence);

    /* skip leading spaces, and return if input is empty */

    p = sentence;
    while (*p == ' ') {
	++p;
    }
    if (*p == '\0') {
	return P_NULL;
    }
    SetIt(NIL_THING);

    /* Pull off the verb into 'VerbBuffer'. It is either a single non-
       alphabetic character, or a word starting with an alphabetic, ending
       with whitespace. */

    q = &VerbBuffer[0];
    ch = *p;
    /* The hex codes are for ISO Latin-1 characters with accents. */
    if (isalpha(ch) ||
	(ch >= 0xc0 + '\0' && ch <= 0xd6 + '\0') ||
	(ch >= 0xd8 + '\0' && ch <= 0xf6 + '\0') ||
	(ch >= 0xf8 + '\0' && ch <= 0xff + '\0'))
    {
	len = 0;
	while (! phraseEnd(ch)) {
	    if (isupper(ch)) {
		ch = tolower(ch);
	    }
	    if (len < VERB_LEN - 1) {
		len += 1;
		*q++ = ch;
	    }
	    ++p;
	    ch = *p;
	}
    } else {
	*q++ = ch;
	++p;
	len = 1;
    }
    *q = '\0';
    len += 1;

    /* does the grammar know this word? */

    en = symLookupE(&g->g_words, &VerbBuffer[0], len);
    if (en == P_NULL) {
	complain(g, "I don't know the word ", &VerbBuffer[0], ".\n",
		 P_NULL, P_NULL);
	return P_NULL;
    }

    /* check for it being a synonym of something else */

    wl = (WordList_t *) en->en_key;
    typ = wl->wl_key >> TYPE_SHIFT;
    key = wl->wl_.wl_action;
    alternatives = wl->wl_.wl_alternatives;
    if (typ == W_SYNONYM) {
	wl = wl->wl_.wl_synonym;
	typ = wl->wl_key >> TYPE_SHIFT;
	key = wl->wl_.wl_action;
	alternatives = wl->wl_.wl_alternatives;
    } else {
	wl = P_NULL;
    }

    /* skip whitespace after the verb */

    while (*p == ' ') {
	++p;
    }
    sentence = p;
    LastPunctuation = ' ';

    /* do the appropriate thing for the verb type (if it IS a verb!) */

    switch (typ) {

    case W_VERBTAIL:

	/* a verb which just swallows the rest of the sentence for
	   its own use */

	p = stuffTail(p);
	ch = *p;
	if (ch != '\0') {
	    LastPunctuation = ch;
	}

	/* Beware magic! This is calling the interpreted routine that the
	   scenario has associated with this verb. If that routine returns
	   'true', then all is well and we return non-P_NULL to allow
	   handling of more command in the input string. */

	if (doProc0(key) != BOOLEAN_FALSE) {
	    return p;
	}
	break;

    case W_VERB:

	/* a verb with several alternatives */

	/* check for a separator right away */

	alt1 = skipSeparator(g, alternatives, &sentence);
	hadAlt1 = (BOOL_T) (alt1 != P_NULL);

	/* Loop through any direct objects given, and make a list of them.
	   This loop will exit at the end of the input, or when we hit a
	   separator for this verb. */

	objects = P_NULL;
	objectTail = &objects;
	while (B_TRUE) {
	    if (objects != P_NULL) {
		alt2 = skipSeparator(g, alternatives, &sentence);
	    } else {
		alt2 = P_NULL;
	    }
	    ch = *sentence;
	    if (alt2 != P_NULL || sentEnd(ch)) {
		break;
	    }
	    if (! getNounPhrase(g, &sentence, &dirObj, alternatives)) {
		/* Something wrong in the noun phrase. We would have already
		   complained. */
		freeObjects(objects);
		return P_NULL;
	    }
	    ol = MEM_ALLOC_E(&MS, sizeof(ObjectList_t), "ob_l");
	    ol->ol_next = P_NULL;
	    ol->ol_this = dirObj;
	    *objectTail = ol;
	    objectTail = &ol->ol_next;
	    /* skip spaces after a noun phrase and before an 'and' or ',' */
	    p = sentence;
	    while (*p == ' ') {
		++p;
	    }
	    /* skip any ',' or 'and' that is here */
	    ch = *p++;
	    if (ch == ',') {
		/* comma - skip it and continue */
		sentence = p;
	    } else if (ch == 'a' || ch == 'A') {
		ch = *p++;
		if (ch == 'n' || ch == 'N') {
		    ch = *p++;
		    if ((ch == 'd' || ch == 'D') && *p == ' ') {
			/* 'and' - skip it and continue */
			sentence = p;
		    }
		}
	    }
	    /* skip spaces ready for next noun phrase or end of sentence */
	    p = sentence;
	    while (*p == ' ') {
		++p;
	    }
	    sentence = p;
	}

	if (alt1 != P_NULL && alt2 != P_NULL) {
	    /* two separators were given */
	    complain(g, "I don't understand your use of ", &VerbBuffer[0],
		     ".\n", P_NULL, wl);
	    freeObjects(objects);
	    return P_NULL;
	}

	if (alt1 == P_NULL && alt2 == P_NULL) {
	    /* no separator found - find a reasonable default */
	    if (objects == P_NULL) {
		/* try for a no-object verb, then a direct object verb */
		alt1 = alternatives;
		while (alt1 != P_NULL && (alt1->alt_type != vt_noObj ||
					  alt1->alt_separator != 0))
		{
		    alt1 = alt1->alt_next;
		}
		if (alt1 == P_NULL) {
		    alt1 = alternatives;
		    while (alt1 != P_NULL && (alt1->alt_type != vt_dirObj ||
					      alt1->alt_separator != 0))
		    {
			alt1 = alt1->alt_next;
		    }
		    if (alt1 == P_NULL) {
			alt1 = alternatives;
		    }
		}
	    } else {
		/* try: 1) direct object with no separator; 2) direct object
		   with separator; 3) no-object verb; 4) whatever we get */
		alt1 = alternatives;
		while (alt1 != P_NULL &&
		    (alt1->alt_type != vt_dirObj || alt1->alt_separator != 0))
		{
		    alt1 = alt1->alt_next;
		}
		if (alt1 == P_NULL) {
		    alt1 = alternatives;
		    while (alt1 != P_NULL && alt1->alt_type != vt_dirObj) {
			alt1 = alt1->alt_next;
		    }
		    if (alt1 == P_NULL) {
			alt1 = alternatives;
			while (alt1 != P_NULL && alt1->alt_type != vt_noObj) {
			    alt1 = alt1->alt_next;
			}
			if (alt1 == P_NULL) {
			    alt1 = alternatives;
			}
		    }
		}
	    }
	} else if (alt1 == P_NULL) {
	    alt1 = alt2;
	}

	if (alt1->alt_type == vt_twoObj) {
	    /* two-object verb - get the indirect object */
	    if (! getNounPhrase(g, &sentence, &indirObj, P_NULL)) {
		freeObjects(objects);
		return P_NULL;
	    }
	}

	if (! sentEnd(*sentence)) {
	    complain(g, "Unexpected words after ", &VerbBuffer[0],
		     " command.\n", alt1, wl);
	    freeObjects(objects);
	    return P_NULL;
	}
	
	if (alt1->alt_type == vt_noObj && objects != P_NULL) {
	    /* Our alternative wants no objects, but we were given some.
	       Try to find a better alternative. */
	    alt2 = alt1->alt_next;
	    found = B_FALSE;
	    while (! found && alt2 != P_NULL) {
		if (alt2->alt_separator == alt1->alt_separator &&
		    alt2->alt_type == vt_dirObj)
		{
		    found = B_TRUE;
		} else {
		    alt2 = alt2->alt_next;
		}
	    }
	    if (found) {
		alt1 = alt2;
	    }
	} else if (alt1->alt_type == vt_dirObj && objects == P_NULL) {
	    /* Our alternative wants an object, but none were given.
	       Try to find a better alternative. */
	    alt2 = alt1->alt_next;
	    found = B_FALSE;
	    while (! found && alt2 != P_NULL) {
		if (alt2->alt_separator == alt1->alt_separator &&
		    alt2->alt_type == vt_noObj)
		{
		    found = B_TRUE;
		} else {
		    alt2 = alt2->alt_next;
		}
	    }
	    if (found) {
		alt1 = alt2;
	    }
	}

	ch = *sentence;
	if (ch != '\0') {
	    LastPunctuation = ch;
	}
	switch (alt1->alt_type) {

	case vt_noObj:

	    if (objects == P_NULL) {
		if (doProc0(alt1->alt_action) != BOOLEAN_FALSE) {
		    return sentence;
		}
	    } else {
		complain(g, P_NULL, &VerbBuffer[0],
			 " doesn't take an object.\n", alt1, wl);
	    }
	    break;

	case vt_dirObj:

	    if (objects == P_NULL) {
		/* Allow direct-object verbs to be used with no object, so
		   that the user code can provide its own error message. */
		if (doProc1(alt1->alt_action, nullString()) != BOOLEAN_FALSE) {
		    return sentence;
		}
	    } else {
		ol = objects;
		while (ol != P_NULL) {
		    NamePointer = ol->ol_this;
		    if (doProc1(alt1->alt_action, stringCopy(ol->ol_this)) ==
			BOOLEAN_FALSE)
		    {
			break;
		    }
		    ol = ol->ol_next;
		}
		NamePointer = "";
		if (ol == P_NULL) {
		    /* successfully did all of them */
		    freeObjects(objects);
		    return sentence;
		}
	    }
	    break;

	case vt_twoObj:

	    if (*indirObj == '\0' && objects == P_NULL) {
		complain2(g, &VerbBuffer[0], P_NULL, P_NULL, alt1, wl);
	    } else if (*indirObj == '\0') {
		p = FormatName(objects->ol_this);
		if (hadAlt1) {
		    /* In this situation, the user entered something
		       like:   <verb> <sep> <np>
		       but the code that found this alternative as the
		       only one has re-arranged things a bit. So, we
		       have this special test to try to get the right
		       error comment. A kludge. */
		    complain2(g, &VerbBuffer[0], P_NULL, p, alt1, wl);
		} else {
		    complain2(g, &VerbBuffer[0], p, P_NULL, alt1, wl);
		}
		freeString(p);
	    } else if (objects == P_NULL) {
		p = FormatName(indirObj);
		complain2(g, &VerbBuffer[0], P_NULL, p, alt1, wl);
		freeString(p);
	    } else {
		IndirPointer = indirObj;
		ol = objects;
		while (ol != P_NULL) {
		    NamePointer = ol->ol_this;
		    if (doProc2(alt1->alt_action, stringCopy(ol->ol_this),
				stringCopy(indirObj)) == BOOLEAN_FALSE)
		    {
			break;
		    }
		    ol = ol->ol_next;
		}
		NamePointer = "";
		IndirPointer = "";
		if (ol == P_NULL) {
		    /* successfully did all of them */
		    freeString(indirObj);
		    freeObjects(objects);
		    return sentence;
		}
	    }
	    freeString(indirObj);
	    break;

	}
	freeObjects(objects);

	break;
    default:

	/* not even a verb! */

	complain(g, "I don't understand a command starting with ",
		 &VerbBuffer[0], ".\n", P_NULL, wl);
	break;

    }

    return P_NULL;
}

/*
 * Parse - top level parser - call 'parseOne' repeatedly for multiple
 *	commands within the given string.
 */

ULONG_T
Parse(ULONG_T key, CHAR_T *sentence)
{
    Grammar_t *g;
    ULONG_T sentenceCount;
    CHAR_T ch;

    TRACE("Parse", sentence);
    /* There is another call to this inside 'parseOne'. We do it this
       way so that there isn't an extra one *after* handling the last
       command in the input string. */
    SetIt(NIL_THING);
    g = findCachedGrammar(key);
    sentenceCount = 0;
    do {
	sentence = parseOne(g, sentence);
	if (sentence == P_NULL) {
	    break;
	}
	sentenceCount += 1;
	while (B_TRUE) {
	    ch = *sentence;
	    if (sentEnd(ch)) {
		break;
	    }
	    ++sentence;
	}
	if (ch != '\0') {
	    ch = *++sentence;
	}
    } while (ch != '\0');
    return sentenceCount;
}
