#include <time.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * util.c - some random utility routines.
 */

#define FILE_BUFFER_SIZE	512

#define TRACE_SIZE		100

#define DEBUG_TRACE		B_FALSE

typedef enum {
    fm_read,
    fm_write,
    fm_update
} FileMode_t;

typedef struct FileList {
    struct FileList *fl_next;
    CHAR_T fl_buffer[FILE_BUFFER_SIZE];
    UINT_T fl_buffPos;
    UINT_T fl_buffMax;
    int fl_fd;
    FileMode_t fl_mode;
} FileList_t;

static char *(TraceWhere[TRACE_SIZE]);
static ULONG_T TraceValue[TRACE_SIZE];
static ULONG_T TraceFront, TraceBack;

static FileList_t *OpenFiles;

/*
 * utilInit - initialize the stuff here.
 */

void
utilInit(void)
{

#if TRACING
    TraceFront = 0;
    TraceBack = 0;
#endif
    OpenFiles = P_NULL;
}

/*
 * utilTerm - termination code for the stuff here.
 */

void
utilTerm(void)
{
    FileList_t *fl;

    while (B_TRUE) {
	fl = OpenFiles;
	if (fl == P_NULL) {
	    break;
	}
	OpenFiles = fl->fl_next;
	(void) close(fl->fl_fd);
	memFreeE(&MS, fl, sizeof(FileList_t));
    }
}

#if TRACING

/*
 * trace - add an entry to the trace buffer.
 */

void
trace(char *where, ULONG_T valu)
{

#if DEBUG_TRACE
    mess(where);
    mess(": ");
    mHex(valu);
    messChar('\n');
#endif
    TraceWhere[TraceFront] = where;
    TraceValue[TraceFront] = valu;
    TraceFront = (TraceFront == TRACE_SIZE - 1) ? 0 : TraceFront + 1;
    if (TraceFront == TraceBack) {
	TraceBack = (TraceBack == TRACE_SIZE - 1) ? 0 : TraceBack + 1;
    }
}

/*
 * traceDump - dump out the trace buffer.
 */

void
traceDump(void)
{
    ULONG_T tf;

    putString(">>>> Trace buffer (latest first):\n");
    tf = TraceFront;
    while (B_TRUE) {
	if (tf == 0) {
	    tf = TRACE_SIZE - 1;
	} else {
	    tf -= 1;
	}
	if (tf == TraceBack) {
	    break;
	}
	putString(TraceWhere[tf]);
	putString(": ");
	putHex(TraceValue[tf]);
	newLine();
    }
    putString("<<<<\n");
}

#endif

/*
 * copy - utility to copy a block of data.
 */

void copy(void *pa, void *qa, ULONG_T n)
{
    BYTE_T *p, *q;

    p = (BYTE_T *) pa;
    q = (BYTE_T *) qa;
    while (n != 0) {
	n -= 1;
	*p++ = *q++;
    }
}

/*
 * strTail - append the second string to the end of the first, returning
 *	the final tail pointer.
 */

CHAR_T *
strTail(CHAR_T *p, CHAR_T *q)
{

    while (*q != '\0') {
	*p++ = *q++;
    }
    return p;
}

/*
 * strCmp - compare two strings.
 */

int
strCmp(CHAR_T *left, CHAR_T *right)
{

#if CHECKS_PRESENT
    if (left == P_NULL) {
	myAbort("P_NULL left to strCmp");
    }
    if (right == P_NULL) {
	myAbort("P_NULL right to strCmp");
    }
#endif
    while (*left == *right) {
	if (*left == '\0') {
	    return 0;
	}
	++left;
	++right;
    }
    if (*left < *right) {
	return -1;
    } else {
	return 1;
    }
}

/*
 * strCmpLC - compare two strings, ignoring case.
 */

int
strCmpLC(CHAR_T *left, CHAR_T *right)
{
    CHAR_T chl, chr;

#if CHECKS_PRESENT
    if (left == P_NULL) {
	myAbort("P_NULL left to strCmpLC");
    }
    if (right == P_NULL) {
	myAbort("P_NULL right to strCmpLC");
    }
#endif
    while (B_TRUE) {
	chl = *left++;
	if (isupper(chl)) {
	    chl = tolower(chl);
	}
	chr = *right++;
	if (isupper(chr)) {
	    chr = tolower(chr);
	}
	if (chl != chr) {
	    break;
	}
	if (chl == '\0') {
	    return 0;
	}
    }
    if (chl < chr) {
	return -1;
    } else {
	return 1;
    }
}

/*
 * strCpy - copy one string into another.
 */

void
strCpy(CHAR_T *dest, CHAR_T *src)
{

#if CHECKS_PRESENT
    if (dest == P_NULL) {
	myAbort("P_NULL dest to strCpy");
    }
    if (src == P_NULL) {
	myAbort("P_NULL src to strCpy");
    }
#endif
    while (B_TRUE) {
	*dest = *src;
	if (*src == '\0') {
	    break;
	}
	++dest;
	++src;
    }
}

/*
 * strNCpy - copy one string into another, but limiting the length.
 */

void
strNCpy(CHAR_T *dest, CHAR_T *src, UINT_T len)
{

#if CHECKS_PRESENT
    if (dest == P_NULL) {
	myAbort("P_NULL dest to strNCpy");
    }
    if (src == P_NULL) {
	myAbort("P_NULL src to strNCpy");
    }
    if (len < 2) {
	myAbort("small len to strNCpy");
    }
#endif
    len -= 1;	/* len *includes* the terminating NUL */
    while (len != 0 && *src != '\0') {
	*dest++ = *src++;
	len -= 1;
    }
    *dest = '\0';
}

/*
 * nullString - allocate and return an empty string.
 */

CHAR_T *
nullString(void)
{
    CHAR_T *p;

    p = MEM_ALLOC_E(&MS, 1, "nlst");
    *p = '\0';
    return p;
}

/*
 * concatenate - join two strings and return the newly allocated result.
 */

CHAR_T *
concatenate(CHAR_T *left, CHAR_T *right)
{
    CHAR_T *p, *q, *st;
    ULONG_T l;

#if CHECKS_PRESENT
    if (left == P_NULL) {
	myAbort("P_NULL left to concatenate");
    }
    if (right == P_NULL) {
	myAbort("P_NULL right to concatenate");
    }
#endif
    p = left;
    q = p;
    l = 1;
    while (*p != '\0') {
	++p;
	l += 1;
    }
    p = right;
    while (*p != '\0') {
	++p;
	l += 1;
    }
    if (l >= 4096) {
	runError("result string too long in string concatenation");
	freeString(right);
	return left;
    }
    p = MEM_ALLOC_E(&MS, l, "conc");
    st = p;
    while (*q != '\0') {
	*p++ = *q++;
    }
    q = right;
    while (B_TRUE) {
	*p++ = *q;
	if (*q == '\0') {
	    break;
	}
	++q;
    }
    freeString(left);
    freeString(right);
    return st;
}

/*
 * P_NULLValue - return a P_NULL of the appropriate type.
 */

ULONG_T
nilValue(UINT_T typ)
{

    switch (typ) {
    case TYPE_STATUS:
	return STATUS_FAIL;
    case TYPE_BOOLEAN:
	return BOOLEAN_FALSE;
    case TYPE_INTEGER:
    case TYPE_FIXED:
	return 0;
    case TYPE_STRING:
	return (ULONG_T) nullString();
    case TYPE_PLAYER:
    case TYPE_THING:
    case TYPE_ACTION:
    case TYPE_TABLE:
    case TYPE_GRAMMAR:
    case TYPE_LIST_INTEGER:
    case TYPE_LIST_FIXED:
    case TYPE_LIST_THING:
    case TYPE_LIST_ACTION:
    case TYPE_PROP_BOOLEAN:
    case TYPE_PROP_INTEGER:
    case TYPE_PROP_FIXED:
    case TYPE_PROP_THING:
    case TYPE_PROP_ACTION:
    case TYPE_PROP_STRING:
    case TYPE_PROP_TABLE:
    case TYPE_PROP_GRAMMAR:
    case TYPE_PROP_LIST_INTEGER:
    case TYPE_PROP_LIST_FIXED:
    case TYPE_PROP_LIST_THING:
    case TYPE_PROP_LIST_ACTION:
	return (ULONG_T) typ << TYPE_SHIFT;
    default:
	return 0;
    }
}

/*
 * freeString - free a string value.
 */

void
freeString(CHAR_T *st)
{
    CHAR_T *p;

#if CHECKS_PRESENT
    if (st == P_NULL) {
	myAbort("P_NULL string to freeString");
    }
#endif
    p = st;
    while (*p != '\0') {
	++p;
    }
    ++p;
    memFreeE(&MS, st, p - st);
}

/*
 * stringCopy - allocate and return a copy of a string.
 */

CHAR_T *
stringCopy(CHAR_T *st)
{
    CHAR_T *p, *q;

#if CHECKS_PRESENT
    if (st == P_NULL) {
	myAbort("P_NULL string to stringCopy");
    }
#endif
    p = st;
    while (*p != '\0') {
	++p;
    }
    ++p;
    q = MEM_ALLOC_E(&MS, p - st, "stcp");
    p = st;
    st = q;
    while (*p != '\0') {
	*q++ = *p++;
    }
    *q = '\0';
    return st;
}

/*
 * strLen1 - return the length in bytes of a string, INCLUDING the trailing \0.
 */

UINT_T
strLen1(CHAR_T *str)
{
    CHAR_T *p;

#if CHECKS_PRESENT
    if (str == P_NULL) {
	myAbort("P_NULL string to strLen1");
    }
#endif
    p = str;
    while (*p != '\0') {
	++p;
    }
    ++p;
    return (UINT_T) (p - str);
}

/*
 * newLine - add a newline to the message buffer.
 */

void
newLine(void)
{

    putChar('\n');
}

/*
 * putError - put a message with *** around it.
 */

void
putError(CHAR_T *m)
{

    putString("*** ");
    putString(m);
    putString(" ***\n");
}

/*
 * putStrError - put an error line with a string.
 */

void
putStrError(CHAR_T *s1, CHAR_T *s2, CHAR_T *s3)
{

    putString("*** ");
    putString(s1);
    putString(s2);
    putString(s3);
    putString(" ***\n");
}

/*
 * putDecimal - output a positive decimal value.
 */

void
putDecimal(ULONG_T n)
{
    CHAR_T buffer[11];
    CHAR_T *p;

    p = &buffer[10];
    *p = '\0';
    while (B_TRUE) {
	*--p = n % 10 + '0';
	n /= 10;
	if (n == 0) {
	    break;
	}
    }
    putString(p);
}

/*
 * putSigned - output a signed decimal value.
 */

void
putSigned(LONG_T n)
{
    CHAR_T buffer[12];
    CHAR_T *p;
    BOOL_T isNeg;

    if (n < 0) {
	isNeg = B_TRUE;
    } else {
	isNeg = B_FALSE;
	n = -n;
    }
    p = &buffer[11];
    *p = '\0';
    while (B_TRUE) {
	*--p = - (n % 10) + '0';
	n /= 10;
	if (n == 0) {
	    break;
	}
    }
    if (isNeg) {
	*--p = '-';
    }
    putString(p);
}

/*
 * putHex - output a value in hex. (used for debugging mostly)
 */

void
putHex(ULONG_T n)
{
    CHAR_T *p;
    int i;
    BYTE_T b;
    CHAR_T buff[9];

    p = &buff[8];
    *p = '\0';
    for (i = 0; i != 8; i += 1) {
	b = n & 0xf;
	n >>= 4;
	*--p = (b > 9) ? (b - 10 + 'a') : (b + '0');
    }
    putString(p);
}

/*
 * symErrNilActPar - little utility.
 */

static void
symErrNilActPar(CHAR_T *function)
{

    symbolError(function, "nil action not valid as parameter");
}

/*
 * checkAction - verify that the given action takes no parameters.
 */

BOOL_T
checkAction(ULONG_T key, CHAR_T *function)
{
    Proc_t *pr;

    if (key == NIL_ACTION) {
	symErrNilActPar(function);
	return B_FALSE;
    }
    pr = findCachedProcNoLock(key);
    if (pr->pr_parameters != P_NULL || pr->pr_resultType != TYPE_VOID) {
	symbolError(function, "proc should have no parameters and no result");
	return B_FALSE;
    }
    return B_TRUE;
}

/*
 * checkBoolAction - verify action with one bool parameter.
 */

BOOL_T
checkBoolAction(ULONG_T key, CHAR_T *function)
{
    Proc_t *pr;
    LocalList_t *ll;

    if (key == NIL_ACTION) {
	symErrNilActPar(function);
	return B_FALSE;
    }
    pr = findCachedProcNoLock(key);
    ll = pr->pr_parameters;
    if ((ll == P_NULL && pr->pr_resultType != TYPE_VOID &&
	 ll->ll_next != P_NULL) || ll->ll_type != TYPE_BOOLEAN)
    {
	symbolError(function,
	    "proc should have 1 bool parameter and no result");
	return B_FALSE;
    }
    return B_TRUE;
}

/*
 * checkIntAction - verify that the given action takes one integer parameter.
 */

BOOL_T
checkIntAction(ULONG_T key, CHAR_T *function)
{
    Proc_t *pr;
    LocalList_t *ll;

    if (key == NIL_ACTION) {
	symErrNilActPar(function);
	return B_FALSE;
    }
    pr = findCachedProcNoLock(key);
    ll = pr->pr_parameters;
    if (ll == P_NULL || pr->pr_resultType != TYPE_VOID ||
	ll->ll_next != P_NULL || ll->ll_type != TYPE_INTEGER)
    {
	symbolError(function,
	    "proc should have 1 integer parameter and no result");
	return B_FALSE;
    }
    return B_TRUE;
}

/*
 * checkInt2Action - verify that the given action takes 2 integer parameters.
 */

BOOL_T
checkInt2Action(ULONG_T key, CHAR_T *function)
{
    Proc_t *pr;
    LocalList_t *ll;
    BOOL_T ok;

    if (key == NIL_ACTION) {
	symErrNilActPar(function);
	return B_FALSE;
    }
    pr = findCachedProcNoLock(key);
    ll = pr->pr_parameters;
    ok = B_FALSE;
    if (ll != P_NULL && pr->pr_resultType == TYPE_VOID &&
	ll->ll_type == TYPE_INTEGER)
    {
	ll = ll->ll_next;
	if (ll != P_NULL && ll->ll_type == TYPE_INTEGER &&
	    ll->ll_next == P_NULL)
	{
	    ok = B_TRUE;
	}
    }
    if (ok) {
	return B_TRUE;
    }
    symbolError(function,
	"proc should have 2 integer parameters and no result");
    return B_FALSE;
}

/*
 * checkInt3Action - verify that the given action takes 3 integer parameters.
 */

BOOL_T
checkInt3Action(ULONG_T key, CHAR_T *function)
{
    Proc_t *pr;
    LocalList_t *ll;
    BOOL_T ok;

    if (key == NIL_ACTION) {
	symErrNilActPar(function);
	return B_FALSE;
    }
    pr = findCachedProcNoLock(key);
    ll = pr->pr_parameters;
    ok = B_FALSE;
    if (ll != P_NULL && pr->pr_resultType == TYPE_VOID &&
	ll->ll_type == TYPE_INTEGER)
    {
	ll = ll->ll_next;
	if (ll != P_NULL && ll->ll_type == TYPE_INTEGER) {
	    ll = ll->ll_next;
	    if (ll != P_NULL && ll->ll_type == TYPE_INTEGER &&
		ll->ll_next == P_NULL)
	    {
		ok = B_TRUE;
	    }
	}
    }
    if (ok) {
	return B_TRUE;
    }
    symbolError(function,
	"proc should have 3 integer parameters and no result");
    return B_FALSE;
}

/*
 * checkStringAction - verify that the given action takes one string parameter.
 */

BOOL_T
checkStringAction(ULONG_T key, CHAR_T *function)
{
    Proc_t *pr;
    LocalList_t *ll;

    if (key == NIL_ACTION) {
	symErrNilActPar(function);
	return B_FALSE;
    }
    pr = findCachedProcNoLock(key);
    ll = pr->pr_parameters;
    if (ll == P_NULL || pr->pr_resultType != TYPE_VOID ||
	ll->ll_next != P_NULL || ll->ll_type != TYPE_STRING)
    {
	symbolError(function,
	    "proc should have 1 string parameter and no result");
	return B_FALSE;
    }
    return B_TRUE;
}

/*
 * checkThingAction - verify that the given action takes one thing parameter.
 */

BOOL_T
checkThingAction(ULONG_T key, CHAR_T *function)
{
    Proc_t *pr;
    LocalList_t *ll;

    if (key == NIL_ACTION) {
	symErrNilActPar(function);
	return B_FALSE;
    }
    pr = findCachedProcNoLock(key);
    ll = pr->pr_parameters;
    if (ll == P_NULL || pr->pr_resultType != TYPE_VOID ||
	ll->ll_next != P_NULL || ll->ll_type != TYPE_THING)
    {
	symbolError(function,
	    "proc should have 1 thing parameter and no result");
	return B_FALSE;
    }
    return B_TRUE;
}

/*
 * checkStringBoolAction - verify action with string and bool parameters.
 */

BOOL_T
checkStringBoolAction(ULONG_T key, CHAR_T *function)
{
    Proc_t *pr;
    LocalList_t *ll;
    BOOL_T ok;

    if (key == NIL_ACTION) {
	symErrNilActPar(function);
	return B_FALSE;
    }
    ok = B_FALSE;
    pr = findCachedProcNoLock(key);
    ll = pr->pr_parameters;
    if (ll != P_NULL && pr->pr_resultType == TYPE_VOID &&
	ll->ll_type == TYPE_STRING)
    {
	ll = ll->ll_next;
	if (ll != P_NULL && ll->ll_type == TYPE_BOOLEAN &&
	    ll->ll_next == P_NULL)
	{
	    ok = B_TRUE;
	}
    }
    if (! ok) {
	symbolError(function,
	    "proc should have string parm, bool parm and no result");
    }
    return ok;
}

/*
 * sysAdminFiles - little utility.
 */

static void
sysAdminFiles(void)
{

    runError("only SysAdmin can use files");
}

/*
 * makeFl - make a new fl structure.
 */

static void
makeFl(int fd, FileMode_t mode)
{
    FileList_t *fl;

    fl = MEM_ALLOC_E(&MS, sizeof(FileList_t), "fl_t");
    fl->fl_next = OpenFiles;
    fl->fl_buffPos = 0;
    fl->fl_buffMax = 0;
    fl->fl_fd = fd;
    fl->fl_mode = mode;
    OpenFiles = fl;
}

/*
 * FileOpenForRead - builtin to allow opening of files.
 */

ULONG_T
FileOpenForRead(CHAR_T *fileName)
{
    int fd;

    if (EffectivePlayer != SYSADMIN_KEY) {
	sysAdminFiles();
	return 0;
    }
    fd = open(fileName, O_RDONLY);
    if (fd < 0) {
	return 0;
    }
    makeFl(fd, fm_read);
    return (ULONG_T) fd;
}

/*
 * FileOpenForWrite - builtin to allow opening of files.
 */

ULONG_T
FileOpenForWrite(CHAR_T *fileName)
{
    int fd;

    if (EffectivePlayer != SYSADMIN_KEY) {
	sysAdminFiles();
	return 0;
    }
    fd = open(fileName, O_RDWR);
    if (fd < 0) {
	return 0;
    }
    makeFl(fd, fm_write);
    return (ULONG_T) fd;
}

/*
 * FileOpenForUpdate - builtin to allow opening of files.
 */

ULONG_T
FileOpenForUpdate(CHAR_T *fileName)
{
    int fd;

    if (EffectivePlayer != SYSADMIN_KEY) {
	sysAdminFiles();
	return 0;
    }
    fd = open(fileName, O_RDWR);
    if (fd < 0) {
	return 0;
    }
    makeFl(fd, fm_update);
    return (ULONG_T) fd;
}

/*
 * FileClose - builtin to allow closing a file.
 */

void
FileClose(ULONG_T id)
{
    FileList_t *fl;
    FileList_t **flp;

    if (EffectivePlayer != SYSADMIN_KEY) {
	sysAdminFiles();
    } else {
	flp = &OpenFiles;
	while (B_TRUE) {
	    fl = *flp;
	    if (fl == P_NULL || fl->fl_fd == id) {
		break;
	    }
	    flp = &fl->fl_next;
	}
	if (fl == P_NULL) {
	    runError("invalid file id to FileClose");
	} else {
	    if (fl->fl_mode == fm_write && fl->fl_buffPos != 0) {
		(void) write(id, &fl->fl_buffer[0], fl->fl_buffPos);
	    }
	    (void) close(id);
	    *flp = fl->fl_next;
	    memFreeE(&MS, fl, sizeof(FileList_t));
	}
    }
}

/*
 * FileRead - return the next 'line' from the given file.
 */

CHAR_T *
FileRead(ULONG_T id)
{
#define MAX_LINE_LEN	512
    CHAR_T lineBuffer[MAX_LINE_LEN + 1];
    FileList_t *fl;
    CHAR_T *p;
    UINT_T pos;
    CHAR_T ch;
    int len;

    if (EffectivePlayer != SYSADMIN_KEY) {
	sysAdminFiles();
	return nullString();
    }
    fl = OpenFiles;
    while (fl != P_NULL && fl->fl_fd != id) {
	fl = fl->fl_next;
    }
    if (fl == P_NULL) {
	runError("invalid file id to FileRead");
	return nullString();
    }
    if (fl->fl_mode == fm_write) {
	runError("file was opened for writing");
	return nullString();
    }
    if (fl->fl_mode == fm_update) {
	runError("file was opened for update");
	return nullString();
    }
    pos = 0;
    p = &lineBuffer[0];
    while (B_TRUE) {
	if (fl->fl_buffPos == fl->fl_buffMax) {
	    len = read(id, &fl->fl_buffer[0], FILE_BUFFER_SIZE);
	    if (len <= 0) {
		return(nullString());
	    }
	    fl->fl_buffPos = 1;
	    fl->fl_buffMax = len;
	    ch = fl->fl_buffer[0];
	} else {
	    ch = fl->fl_buffer[fl->fl_buffPos];
	    fl->fl_buffPos = fl->fl_buffPos + 1;
	}
	if (ch == '\n') {
	    break;
	}
	if (pos != MAX_LINE_LEN) {
	    *p++ = ch;
	    pos += 1;
	}
    }
    if (p == &lineBuffer[0]) {
	/* convert an empty line to a line of one space, to allow
	   it to be distinguished from the empty string returned for
	   end of file. */
	*p++ = ' ';
    }
    *p = '\0';
    return stringCopy(&lineBuffer[0]);
#undef MAX_LINE_LEN
}

/*
 * FileWrite - write a string to a file.
 */

void
FileWrite(ULONG_T id, CHAR_T *p)
{
    FileList_t *fl;
    CHAR_T *q;
    UINT_T pos;

    if (EffectivePlayer != SYSADMIN_KEY) {
	sysAdminFiles();
    } else {
	fl = OpenFiles;
	while (fl != P_NULL && fl->fl_fd != id) {
	    fl = fl->fl_next;
	}
	if (fl == P_NULL) {
	    runError("invalid file id to FileWrite");
	} else if (fl->fl_mode == fm_read) {
	    runError("file was opened for reading");
	} else {
	    if (fl->fl_mode == fm_update) {
		(void) write(id, p, strLen1(p) - 1);
	    } else {
		pos = fl->fl_buffPos;
		q = &fl->fl_buffer[pos];
		while (*p != '\0') {
		    if (pos == FILE_BUFFER_SIZE) {
			(void) write(id, &fl->fl_buffer[0], FILE_BUFFER_SIZE);
			pos = 0;
			q = &fl->fl_buffer[0];
		    }
		    *q++ = *p++;
		    pos -= 1;
		}
		fl->fl_buffPos = pos;
	    }
	}
    }
}

/*
 * FileReadBinary - read some binary data from the given file.
 */

ULONG_T
FileReadBinary(ULONG_T id, ULONG_T list)
{
    FileList_t *fl;
    ItemList_t *l;

    if (EffectivePlayer != SYSADMIN_KEY) {
	sysAdminFiles();
	return 0;
    }
    if (list == NIL_LIST_INTEGER) {
	runError("P_NULL list to FileReadBinary");
	return 0;
    }
    fl = OpenFiles;
    while (fl != P_NULL && fl->fl_fd != id) {
	fl = fl->fl_next;
    }
    if (fl == P_NULL) {
	runError("invalid file id to FileReadBinary");
	return 0;
    }
    if (fl->fl_mode != fm_update) {
	runError("file was not opened for update");
	return 0;
    }
    l = ioRead(list);
    return (ULONG_T) read(id, l + 1, l->l_elementCount * sizeof(ULONG_T));
}

/*
 * FileWriteBinary - write some binary data to a file.
 */

void
FileWriteBinary(ULONG_T id, ULONG_T list)
{
    FileList_t *fl;
    ItemList_t *l;

    if (EffectivePlayer != SYSADMIN_KEY) {
	sysAdminFiles();
    } else if (list == NIL_LIST_INTEGER) {
	runError("P_NULL list to FileWriteBinary");
    } else {
	fl = OpenFiles;
	while (fl != P_NULL && fl->fl_fd != id) {
	    fl = fl->fl_next;
	}
	if (fl == P_NULL) {
	    runError("invalid file id to FileWriteBinary");
	} else if (fl->fl_mode != fm_update) {
	    runError("file was not opened for update");
	} else {
	    l = ioRead(list);
	    (void) write(id, l + 1, l->l_elementCount * sizeof(ULONG_T));
	}
    }
}

/*
 * FileSeek - seek in a binary file.
 */

void
FileSeek(ULONG_T id, ULONG_T pos)
{
    FileList_t *fl;

    if (EffectivePlayer != SYSADMIN_KEY) {
	sysAdminFiles();
    } else {
	fl = OpenFiles;
	while (fl != P_NULL && fl->fl_fd != id) {
	    fl = fl->fl_next;
	}
	if (fl == P_NULL) {
	    runError("invalid file id to FileSeek");
	} else if (fl->fl_mode != fm_update) {
	    runError("file was not opened for update");
	} else {
	    (void) lseek(id, pos, SEEK_SET);
	}
    }
}

/*
 * myExecute - builtin to allow execution of a dos command.
 */

void
myExecute(CHAR_T *command)
{

    if (EffectivePlayer != SYSADMIN_KEY) {
	runError("only SysAdmin can use Execute");
    } else {
	/* Danger Will Robinson!! Only allowed for SysAdmin code. */
	(void) system(command);
    }
}

/*
 * Trim - trim leading and trailing blanks off the string and return a copy.
 */

CHAR_T *
Trim(CHAR_T *s)
{
    CHAR_T *p, *q;
    ULONG_T len;

#if (CHECKS_PRESENT)
    if (s == P_NULL) {
	myAbort("P_NULL string to Trim");
    }
#endif
    p = s;
    while (*p == ' ') {
	++p;
    }
    if (*p == '\0') {
	return nullString();
    }
    len = 1;
    while (*p != '\0') {
	++p;
	len += 1;
    }
    while (B_TRUE) {
	--p;
	if (*p != ' ') {
	    break;
	}
	len -= 1;
    }
    q = MEM_ALLOC_E(&MS, len, "trim");
    len -= 1;
    p = q;
    while (len != 0) {
	len -= 1;
	*p++ = *s++;
    }
    *p = '\0';
    return q;
}

/*
 * hasBadRefs - find out if the given proc has references not resolved
 *	in any of the in-use tables.
 */

BOOL_T
hasBadRefs(Proc_t *pr)
{
    RefList_t *rl;

    rl = pr->pr_refs;
    while (rl != P_NULL) {
	if (findAnyName(rl->rl_key) == P_NULL) {
	    return(B_TRUE);
	}
	rl = rl->rl_next;
    }
    return B_FALSE;
}

/* Compatiblity stuff added when converting to C. */

ULONG_T
GetCurrentTime(void)
{
    time_t timeval;

    (void) time(&timeval);
    return (ULONG_T) timeval;
}

#if 0

/*
 * Grr. Why would anyone want spaces for the high digits in times?
 * Note that I stomp all over 'ctime's internal buffer.
 */

static char *
myCTime(time_t *t)
{
    char *p;

    p = ctime(t);
    if (p[11] == ' ') {
	p[11] = '0';
    }
    if (p[14] == ' ') {
	p[14] = '0';
    }
    if (p[17] == ' ') {
	p[18] = '0';
    }
    return p;
}

#endif

void
ConvTimeShort(ULONG_T seconds, CHAR_T *buffer)
{
    CHAR_T *p;
    static CHAR_T *(months[12]) = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
    };
    int i;

    p = ctime((time_t *) &seconds);
    strNCpy(buffer, p + 20, 5);
    *(buffer + 4) = '-';
    for (i = 0; i != 12; i += 1) {
	if (strncmp(p + 4, months[i], 3) == 0) {
	    break;
	}
    }
    if (i == 12) {
	*(buffer + 5) = '?';
	*(buffer + 6) = '?';
    } else {
	i += 1;
	*(buffer + 5) = i / 10 + '0';
	*(buffer + 6) = i % 10 + '0';
    }
    *(buffer + 7) = '-';
    strNCpy(buffer + 8, p + 8, 12);
    if (buffer[8] == ' ') {
	buffer[8] = '0';
    }
}

void
ConvTime(ULONG_T seconds, CHAR_T *buffer)
{

    strNCpy(buffer, ctime((time_t *) &seconds), 25);
}
