#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <fcntl.h>
#include <netdb.h>
#include <errno.h>
#include <unistd.h>
#include <termios.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "MUD.h"
#include "Request.h"
#include "MUDLib.h"
#include "Effects.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * SMUD.c - simple non-GUI MUD client.
 */

#define DEBUG_ON		B_FALSE
#define DEBUG_MESSAGES		B_FALSE
#define DEBUG_INPUT		B_FALSE

#define INPUT_BUFFER_SIZE	512
#define DEFAULT_TIMEOUT 	(60 * 60)
#define DEFAULT_PORT		16667
#define DEFAULT_HOST		"localhost"
#define INITIAL_OUTPUT_BUFFER	1024

#define MAX_OUTPUT_LINE 	256
#define MAX_PROMPT		80

/* These are in /Include/types.h, but we do not include that. */
#define BOOLEAN_FALSE		0
#define BOOLEAN_TRUE		1

#if DEBUG_ON
#define DEBUG(x)    printf x
#else
#define DEBUG(x)
#endif

#if DEBUG_MESSAGES
#define DEBUG_M(x)  printf x
#else
#define DEBUG_M(x)
#endif

#if DEBUG_INPUT
#define DEBUG_I(x)  printf x
#else
#define DEBUG_I(x)
#endif

/* put MS first, so that it is likely to be on a long-word boundary - this
   will speed uses of long-word fields within it up a bit on 32 bit CPUs */

static MUDState_t MS;

static char *OutputBuffer;
static ULONG_T OutputBufferLength;

static char InputBuffer[INPUT_BUFFER_SIZE + 1], EditBuffer[INPUT_BUFFER_SIZE];
static ULONG_T InputMax, InputPos, EditMax, EditPos;

static char *ForcePlayerName;
static char *ForcePassword;

static ULONG_T TimeOut;
static ULONG_T LastInputTime;
static UINT_T PromptLen, EditMode;
static int Socket, EditFd;
static CHAR_T PromptBuffer[MAX_PROMPT + 1 + 1];
static CHAR_T EraseBuffer[MAX_PROMPT + 2];
static char EditFileName[L_tmpnam];

static BOOL_T ShuttingDown, EchoOn, Interactive, EditError, EditDone,
    EditReading, EditIsProc, IsRemote, EditFirst, EditFailing,
    PromptShown, NeedPrompt, TimedOut, NLConvert, NoCRs;

/*
 * myPerror - like perror, but add some '::' stuff.
 */

static void
myPerror(const char *m)
{

    fprintf(stderr, "::%s: %s::\n", m, strerror(errno));
}

/*
 * editFileComplain - complaint involving the edit file.
 */

static void
editFileComplain(char *m1, char *m3)
{

    if (! EditFailing) {
	fprintf(stderr, "::%s%s%s::\n", m1, &EditFileName[0],
		(m3 == P_NULL ? "" : m3));
	EditFailing = B_TRUE;
    }
}

/*
 * complain1 - a complaint with :: around it.
 */

static void
complain1(char *m)
{

    fprintf(stderr, "::%s::\n", m);
}

/*
 * editComplain - duh.
 */

static void
editComplain(char *m)
{

    if (! EditFailing) {
	complain1(m);
	EditFailing = B_TRUE;
    }
}

/*
 * errorChar - needed by library.
 */

static void
errorChar(char ch)
{

    if (PromptShown && Interactive) {
	DEBUG(("Erasing prompt\n"));
	PromptShown = B_FALSE;
	NeedPrompt = B_TRUE;
	if (! NoCRs) {
	    (void) write(2, &EraseBuffer[0], PromptLen + 2);
	}
    }
    fputc(ch, stderr);
    if (ch == '\n' && Interactive && NeedPrompt) {
	DEBUG(("Redrawing prompt\n"));
	if (NoCRs) {
	    (void) write(2, &PromptBuffer[1], PromptLen);
	} else {
	    (void) write(2, &PromptBuffer[0], PromptLen + 1);
	}
	PromptShown = B_TRUE;
	NeedPrompt = B_FALSE;
    }
}

/*
 * errorChar1 - interface which does nothing after a control-C.
 */

static void
errorChar1(CHAR_T ch)
{

    if (! ShuttingDown) {
	errorChar((char) ch);
    }
}

/*
 * smudAbort - something has gone wrong in our communications with the
 *	MUD server. We shut down. This routine must not return.
 */

static void setEcho(BOOL_T echoOn);

static void
smudAbort(void)
{

    if (OutputBuffer != P_NULL) {
	free(OutputBuffer);
    }
    shutdown(Socket, 2);
    close(Socket);
    (void) unlink(&EditFileName[0]);
    if (! EchoOn) {
	setEcho(B_TRUE);
    }
    exit(1);
}

/*
 * setEcho - turn echoing on/off.
 */

static void
setEcho(BOOL_T echoOn)
{
    struct termios ti;

    EchoOn = echoOn;
    if (! NoCRs) {
	if (tcgetattr(1, &ti) != 0) {
	    myPerror("tcgetattr to control echo");
	    smudAbort();
	}
    }
    if (echoOn) {
	ti.c_lflag |= ECHO;
    } else {
	ti.c_lflag &= ~ECHO;
    }
    if (! NoCRs) {
	if (tcsetattr(1, TCSANOW, &ti) != 0) {
	    myPerror("tcsetattr to control echo");
	    smudAbort();
	}
    }
}

/*
 * getVar - get the value of an environment.
 */

static BOOL_T
getVar(char *varName, char *buffer, ULONG_T buffLen)
{
    char *p;

    p = getenv(varName);
    if (p == P_NULL) {
	return B_FALSE;
    }
    strncpy(buffer, p, buffLen - 1);
    strcat(buffer, "\0");
    return B_TRUE;
}

/*
 * editWrite - try to write something to the edit file.
 */

static void
editWrite(void *buf, ULONG_T len)
{

    if (! EditError) {
	if (write(EditFd, buf, len) != len) {
	    EditError = B_TRUE;
	}
    }
}

/*
 * editNL - write a newline to the edit file.
 */

static void
editNL(void)
{

    editWrite("\n", 1);
}

/*
 * editCleanup - clean up an edit.
 */

static void
editCleanup(void)
{

    DEBUG(("::editCleanup::\n"));
    if (EditIsProc) {
	editProcCancel(&MS);
    } else {
	editStringDone(&MS, "", 0, BOOLEAN_FALSE);
    }
    EditFailing = B_FALSE;
    EditFirst = B_TRUE;
    EditDone = B_FALSE;
}

/*
 * doEditTail - second half of editing.
 */

static void
doEditTail(void)
{
    struct stat st;
    char *buf, *backPos, *allocBuf;
    ULONG_T bLen, len, column, allocLen, extraLen;
    unsigned char ch;

    EditDone = B_FALSE;
    if (EditIsProc) {
	DEBUG(("::doEditTail - isProc::\n"));
	EditFd = open(&EditFileName[0], O_RDONLY);
	if (EditFd < 0) {
	    /* eek! the file has gone away */
	    editFileComplain("Can't reopen ", " after edit");
	    editCleanup();
	    return;
	}
	EditReading = B_TRUE;
	EditMax = 0;
	EditPos = 0;
	editProcStart(&MS);
	editCleanup();
    } else {
	if (stat(&EditFileName[0], &st) != 0) {
	    editFileComplain("Can't stat ", " after edit");
	    (void) unlink(&EditFileName[0]);
	    editCleanup();
	    return;
	}
	bLen = st.st_size;
	DEBUG(("::doEditTail - not isProc, size %ld::\n", bLen));
	extraLen = bLen / 10;
	if (extraLen < 50) {
	    extraLen = 50;
	}
	allocLen = bLen + extraLen;
	allocBuf = malloc(allocLen);
	if (allocBuf == P_NULL) {
	    editComplain("Can't allocate string buffer after edit");
	    (void) unlink(&EditFileName[0]);
	    editCleanup();
	    return;
	}
	EditFd = open(&EditFileName[0], O_RDONLY);
	if (EditFd < 0) {
	    /* eek! the file has gone away */
	    editFileComplain("Can't reopen ", " after edit");
	    free(allocBuf);
	    (void) unlink(&EditFileName[0]);
	    editCleanup();
	    return;
	}
	buf = allocBuf + extraLen;
	if (read(EditFd, buf, bLen) != bLen) {
	    editFileComplain("Can't read ", " after edit");
	    free(allocBuf);
	    (void) close(EditFd);
	    (void) unlink(&EditFileName[0]);
	    editCleanup();
	    return;
	}
	close(EditFd);
	(void) unlink(&EditFileName[0]);
	backPos = allocBuf;
	len = 0;
	column = 0;
	while (bLen != 0) {
	    bLen -= 1;
	    ch = *buf;
	    if (ch >= ' ') {
		if (ch == '\\' && bLen != 0 && EditMode == EDIT_COOKED) {
		    bLen -= 1;
		    ++buf;
		    if (*buf == 'n') {
			*backPos = '\n';
		    } else if (*buf == 't') {
			*backPos = '\t';
		    } else {
			*backPos = *buf;
		    }
		    extraLen += 1;
		} else {
		    *backPos = ch;
		}
		++backPos;
		len += 1;
		column += 1;
	    } else if (ch == '\n') {
		if (EditMode != EDIT_COOKED) {
		    *backPos++ = ch;
		    len += 1;
		    column = 0;
		} else {
		    if (len != 0 && bLen != 0) {
			ch = *(backPos - 1);
			if (ch != ' ' && ch != '\n') {
			    *backPos++ = ' ';
			    len += 1;
			    column += 1;
			} else {
			    extraLen += 1;
			}
		    } else {
			extraLen += 1;
		    }
		}
	    } else if (ch == '\t') {
		/* An editor can put tabs in the file to save file space on
		   a sequence of blanks. We have trouble here properly
		   expanding them again. We do what we can without a lot of
		   hassle, and at least will always put in at least one
		   space. I am sure someone will complain! */
		extraLen += 1;
		do {
		    *backPos++ = ' ';
		    len += 1;
		    extraLen += 1;
		    column += 1;
		} while ((column & 7) != 0 && extraLen != 0);
	    }
	    ++buf;
	}
	/* note that this is a call to the mud.library from inside a callout
	   from that library! */
	editStringDone(&MS, allocBuf, len, BOOLEAN_TRUE);
	free(allocBuf);
    }
}

/*
 * execEditor - try to run command to execute the user's editor.
 */

static BOOL_T
execEditor(void)
{
#define VAR_BUF_LENGTH		100
#define MAX_ARGS		20
    ULONG_T len;
    char *args[MAX_ARGS];
    char *p;
    pid_t pid;
    char varBuffer[VAR_BUF_LENGTH];

    len = strlen(&EditFileName[0]);
    if (getVar("MUDEDITOR", &varBuffer[0], VAR_BUF_LENGTH - len - 2)) {
	strcat(&varBuffer[0], " ");
    } else if (getVar("EDITOR", &varBuffer[9], VAR_BUF_LENGTH - len - 10)) {
	strcpy(&varBuffer[0], "xterm -e");
	varBuffer[8] = ' ';
	strcat(&varBuffer[0], " ");
    } else {
	strcpy(&varBuffer[0], "xterm -e vi ");
    }
    strcat(&varBuffer[0], &EditFileName[0]);
    len = 0;
    p = &varBuffer[0];
    while (*p != '\0') {
	if (len == MAX_ARGS - 1) {
	    editComplain("Too many args in editor command");
	    (void) unlink(&EditFileName[0]);
	    return(B_FALSE);
	}
	args[len] = p;
	len += 1;
	while (*p != ' ' && *p != '\0') {
	    ++p;
	}
	if (*p == ' ') {
	    *p = '\0';
	    ++p;
	}
    }
    args[len] = P_NULL;
    pid = fork();
    if (pid < 0) {
	editComplain("Can't fork editor");
	(void) unlink(&EditFileName[0]);
	return(B_FALSE);
    }
    if (pid == 0) {
	/* Child - execute the editor command */
	DEBUG(("::child ready to exec::\n"));
	(void) close(0);
	(void) close(1);
	(void) close(2);
	(void) close(Socket);
	(void) execvp(args[0], &args[0]);
	editComplain("Can't execvp editor");
	exit(1);
    }
    /* Parent - continue with other stuff. Get SIGCHLD when edit done. */
    return(B_TRUE);
#undef MAX_ARGS
#undef VAR_BUF_LENGTH
}

/*
 * doEdit - allow the player to edit a proc or string.
 */
		
static BOOL_T
doEdit(CHAR_T *buf, ULONG_T bLen, UINT_T mode,
       BOOL_T isProc, BOOL_T isLast, BOOL_T deleteFirst, CHAR_T *prompt)
{
    const ULONG_T MAX_COLUMN = 77;
    char *backPos;
    ULONG_T len, column, shrunkMax;
    char ch;
    BOOL_T wasFirst, needSpace;

    DEBUG(("::doEdit len %ld mode %d isProc %c isLast %c del %c %s::\n",
	   bLen, mode, isProc ? 'T' : 'F', isLast ? 'T' : 'F',
	   deleteFirst ? 'T' : 'F', prompt));
    if (EditReading) {
	return(B_FALSE);
    }
    wasFirst = EditFirst;
    if (wasFirst) {
	EditFirst = B_FALSE;
	EditFailing = B_FALSE;
    }

    if (bLen != 0) {
	EditMode = mode;
	EditIsProc = isProc;
	if (deleteFirst && wasFirst) {
	    DEBUG(("::unlinking %s::\n", &EditFileName[0]));
	    (void) unlink(&EditFileName[0]);
	    EditFd = -1;
	} else {
	    EditFd = open(&EditFileName[0], O_WRONLY);
	}
	if (EditFd < 0) {
	    EditFd = open(&EditFileName[0], O_WRONLY | O_CREAT, 0600);
	    if (EditFd < 0) {
		/* oh-oh - can't write */
		editFileComplain("Can't create ", P_NULL);
		return(B_FALSE);
	    }
	}

	EditError = B_FALSE;
	if (isProc || mode == EDIT_CODE) {
	    (void) lseek(EditFd, 0, SEEK_END);
	    editWrite(buf, bLen);
	} else {
	    column = 0;
	    needSpace = B_FALSE;
	    while (bLen != 0) {
		ch = *buf;
		if (ch == '\n' || ch == '\t' || ch == '\\') {
		    /* a special character in the string - expand it */
		    if (needSpace) {
			needSpace = B_FALSE;
			editWrite(" ", 1);
			column += 1;
		    }
		    if (column > MAX_COLUMN - 2) {
			editNL();
			column = 0;
		    }
		    editWrite(
			(ch == '\n') ? "\\n" :
			(ch == '\t') ? "\\t" :
			"\\\\",
			2);
		    ++buf;
		    bLen -= 1;
		    column += 2;
		} else if (ch == ' ') {
		    backPos = buf;
		    len = 0;
		    while (bLen != 0 && *buf == ' ') {
			len += 1;
			++buf;
			bLen -= 1;
		    }
		    if (column != 0 &&
			(column + len > MAX_COLUMN ||
			 (column + len == MAX_COLUMN && bLen != 0)))
		    {
			editNL();
			column = 0;
			if (len == 1) {
			    /* chuck a single separating space */
			    len = 0;
			}
		    }
		    if (len != 0 && len <= MAX_COLUMN) {
			if (len == 1) {
			    needSpace = B_TRUE;
			} else {
			    editWrite(backPos, len);
			    column += len;
			}
		    }
		} else {
		    /* something that is part of a 'word' */
		    shrunkMax = MAX_COLUMN;
		    if (needSpace) {
			shrunkMax -= 1;
		    }
		    backPos = buf;
		    len = 0;
		    while (B_TRUE) {
			if (column + len > shrunkMax && column != 0) {
			    shrunkMax = MAX_COLUMN;
			    needSpace = B_FALSE;
			    editNL();
			    column = 0;
			}
			if (len > shrunkMax) {
			    shrunkMax = MAX_COLUMN;
			    needSpace = B_FALSE;
			    editWrite(backPos, len - 1);
			    editNL();
			    column = 0;
			    backPos = buf - 1;
			    len = 1;
			}
			if (bLen == 0 || *buf == ' ' || *buf == '\n' ||
			    *buf == '\t' || *buf == '\\')
			{
			    break;
			}
			len += 1;
			++buf;
			bLen -= 1;
		    }
		    if (len != 0) {
			if (needSpace) {
			    needSpace = B_FALSE;
			    editWrite(" ", 1);
			    column += 1;
			}
			editWrite(backPos, len);
			column += len;
		    }
		}
	    }
	    editNL();
	}

	(void) close(EditFd);
	if (EditError) {
	    editFileComplain("Error writing to ", P_NULL);
	    (void) unlink(&EditFileName[0]);
	}
    }
    if (EditFailing) {
	if (isLast) {
	    EditFirst = B_TRUE;
	    EditFailing = B_FALSE;
	}
	return(B_FALSE);
    }
    if (! isLast) {
	return(B_TRUE);
    }

    return execEditor();
}

/*
 * doPrompt - refresh the prompt if needed.
 */

static void
doPrompt(void)
{

    DEBUG(("::doPrompt %c %c::\n", Interactive ? 'T' : 'F',
	   NeedPrompt ? 'T' : 'F'));
    if (Interactive && NeedPrompt) {
	NeedPrompt = B_FALSE;
	if (NoCRs) {
	    (void) write(1, &PromptBuffer[1], PromptLen);
	} else {
	    (void) write(1, &PromptBuffer[0], PromptLen + 1);
	}
	PromptShown = B_TRUE;
    }
}

/*
 * putText - output some text from the server.
 */

static void
putText(CHAR_T *buffer, UINT_T len)
{
    CHAR_T *p;
    ULONG_T totalLen;
    UINT_T l;
    CHAR_T ch;

    DEBUG(("putText, len %u\n", len));
    if (len != 0) {
	totalLen = len;
	if (PromptShown) {
	    totalLen += PromptLen + 2;
	}
	if (NLConvert) {
	    p = buffer;
	    l = len;
	    while (l != 0) {
		l -= 1;
		if (*p == '\n') {
		    totalLen += 1;
		}
		++p;
	    }
	}
	if (totalLen > OutputBufferLength) {
	    free(OutputBuffer);
	    OutputBufferLength = totalLen;
	    OutputBuffer = malloc(totalLen);
	    if (OutputBuffer == P_NULL) {
		fprintf(stderr, "::Cannot grow output buffer!::\n");
		MUDAbort(&MS);
		smudAbort();
	    }
	}
	totalLen = 0;
	if (PromptShown) {
	    if (! NoCRs) {
		totalLen = PromptLen + 2;
		memcpy(OutputBuffer, &EraseBuffer[0], totalLen);
	    }
	    PromptShown = B_FALSE;
	    DEBUG(("putText erasing prompt\n"));
	}
	if (NLConvert) {
	    ch = ' ';	/* keep gcc happy */
	    l = len;
	    p = OutputBuffer + totalLen;
	    while (l != 0) {
		l -= 1;
		ch = *buffer++;
		if (ch == '\n') {
		    *p++ = '\r';
		    totalLen += 1;
		}
		*p++ = ch;
	    }
	} else {
	    memcpy(OutputBuffer + totalLen, buffer, len);
	    ch = *(buffer + len - 1);
	}
	totalLen += len;
	(void) write(1, OutputBuffer, totalLen);
	NeedPrompt = ((ch == '\n') ? B_TRUE : B_FALSE);
	doPrompt();
    }
}

/*
 * setPrompt - set the prompt for input lines.
 */

static void
setPrompt(CHAR_T *newPrompt)
{
    UINT_T i;
    CHAR_T *p;

    DEBUG(("::setPrompt '%s'::\n", newPrompt));
    if (PromptShown) {
	PromptShown = B_FALSE;
	NeedPrompt = B_TRUE;
	if (! NoCRs) {
	    (void) write(1, &EraseBuffer[0], PromptLen + 2);
	}
    }
    /* Outputting a \r before the prompt is a way to avoid some problems
       (described below) in which we get two prompts out in a row. */
    i = strlen((char *) newPrompt);
    PromptLen = i;
    memcpy(&PromptBuffer[1], newPrompt, i);
    /* The sequence \r spaces \r is a safer way to erase the prompt than
       a sequence of space-backspace-space, since the latter can end up
       messing up the PREVIOUS line if it comes out at an inopportune
       time. */
    p = &EraseBuffer[1];
    while (i != 0) {
	*p++ = ' ';
	i -= 1;
    }
    *p = '\r';
}

/*
 * serverSend - send a message to the server. This blocks.
 */

static void
serverSend(Request_t *rq)
{
    unsigned int len;
    RequestType_t rt;

    DEBUG_M(("::serverSend %p type %d used %d avail %d::\n",
	     rq, rq->rq_type, rq->rq_usedLen, rq->rq_availLen));
    len = REQ_LEN + rq->rq_usedLen;
    rt = rq->rq_type;
    rq->rq_usedLen = htons(rq->rq_usedLen);
    rq->rq_key = htonl(rq->rq_key);
    rq->rq_uint = htons(rq->rq_uint);
    if (rt == rt_symbolEnter || rt == rt_symbolDelete ||
	rt == rt_regionSelect || rt == rt_effectDone)
    {
	rq->rq_u.ru_otherKey = htonl(rq->rq_u.ru_otherKey);
    }
    if (send(Socket, &rq->rq_usedLen, len, 0) != len) {
	myPerror("send to server");
	MUDAbort(&MS);
	smudAbort();
    }
    freeRequestE(&MS, rq);
}

/*
 * checkTimeOut - has too long gone by with no activity?
 */

static void
checkTimeOut(void)
{
    ULONG_T now;

    (void) time(&now);
    if (now - LastInputTime > TimeOut) {
	DEBUG(("::now = %lx, LastInputTime = %lx, TimeOut = %lx::\n",
	       now, LastInputTime, TimeOut));
	fprintf(stderr, "::No activity timeout - shutting down::\n");
	TimedOut = B_TRUE;
	MUDShutDown(&MS);
    }
}

/*
 * handleCNTLC - got a CNTL-C signal - handle it.
 */

static void
handleCNTLC(int sig)
{

    if (ShuttingDown) {
	fprintf(stderr, "::Second CNTL-C - aborting::\n");
	MUDAbort(&MS);
	smudAbort();
    } else {
	MUDShutDown(&MS);
	ShuttingDown = B_TRUE;
    }
}

/*
 * doRecv - blocking routine to receive a server message.
 */

static void
doRecv(void *buff, unsigned int len)
{
    BYTE_T *p;
    unsigned int lenSoFar;
    int thisLen;

    p = buff;
    lenSoFar = 0;
    while (lenSoFar < len) {
	thisLen = recv(Socket, p, len - lenSoFar, 0);
	if (thisLen <= 0) {
	    if (! ShuttingDown) {
		if (thisLen < 0) {
		    myPerror("recv on server socket");
		}
		MUDAbort(&MS);
		smudAbort();
	    }
	    break;
	}
	if (thisLen > len - lenSoFar) {
	    fprintf(stderr, "::Received too much on socket?::\n");
	    MUDAbort(&MS);
	    smudAbort();
	}
	p += thisLen;
	lenSoFar += thisLen;
    }
}

/*
 * readFromSocket - read a message from the server socket.
 */

static Request_t *
readFromSocket(void)
{
    Request_t *rq;
    UINT_T len;
    Request_t rqBuf;
    RequestType_t rt;

    DEBUG_M(("::readFromSocket::\n"));
    doRecv(&rqBuf.rq_usedLen, REQ_LEN);
    len = ntohs(rqBuf.rq_usedLen);
    rq = allocRequestE(&MS, len);
    DEBUG_M(("::got request, used %d avail %d::\n",
	     rq->rq_usedLen, rq->rq_availLen));
    rt = rqBuf.rq_type;
    rq->rq_key = ntohl(rqBuf.rq_key);
    rq->rq_type = rt;
    rq->rq_flag = rqBuf.rq_flag;
    rq->rq_uint = ntohs(rqBuf.rq_uint);
    if (len != 0) {
	doRecv(&rq->rq_u.ru_bytes[0], len);
	if (rt == rt_symbolEnter || rt == rt_symbolDelete ||
	    rt == rt_regionSelect || rt == rt_effectDone)
	{
	    rq->rq_u.ru_otherKey = ntohl(rq->rq_u.ru_otherKey);
	}
    }
    return rq;
}

/*
 * waitForStuff - generic routine to wait for something to happen, and
 *	and to handle the exceptional stuff. Return any received request.
 */

static Request_t *
waitForStuff(BOOL_T *pHadInput, BOOL_T noWait)
{
    fd_set readFds, exceptFds;
    struct timeval tv;
    Request_t *rq;
    int res;

    DEBUG_M(("::waitForStuff %c::\n", (pHadInput == P_NULL) ? 'N' : 'I'));
    rq = P_NULL;
    FD_ZERO(&readFds);
    FD_SET(Socket, &readFds);
    if (pHadInput != P_NULL) {
	*pHadInput = B_FALSE;
	FD_SET(0, &readFds);
    }
    exceptFds = readFds;
    tv.tv_sec = (noWait ? 0 : 10);
    tv.tv_usec = 0;
    res = select(Socket + 1, &readFds, P_NULL, &exceptFds, &tv);
    if (res < 0) {
	DEBUG(("::select res < 0::\n"));
	if (errno != EINTR) {
	    myPerror("select");
	    MUDAbort(&MS);
	    smudAbort();
	}
    } else if (res != 0) {
	if (FD_ISSET(Socket, &exceptFds) || FD_ISSET(0, &exceptFds)) {
	    fprintf(stderr, "::exception signalled on select::\n");
	    MUDAbort(&MS);
	    smudAbort();
	} else {
	    if (FD_ISSET(Socket, &readFds) && ! ShuttingDown) {
		rq = readFromSocket();
		DEBUG_M(("::Got rq type %d\n", rq->rq_type));
	    }
	    if (pHadInput != P_NULL && FD_ISSET(0, &readFds)) {
		DEBUG(("::user input available::\n"));
		*pHadInput = B_TRUE;
	    }
	}
    }
    if (EditDone) {
	doEditTail();
    }
    return rq;
}

#if 0
    static char *RequestNames[] = {
	"first",
	"killClient",
	"nukeClient",
	"message",
	"enterWizardMode",
	"leaveWizardMode",
	"setContinue",
	"askPassword",
	"setPrompt",
	"deleteProc",
	"effects",
	"editString",
	"editProc",
	"getString",
	"flushSymbol",
	"defineEffect",
	"queryFile",
	"newName",
	"shutDown",
	"flush",
	"syncShutDown",
	"startClient",
	"stopClient",
	"beginClient",
	"endClient",
	"askPlayer",
	"createPlayer",
	"lookup",
	"findName",
	"readProc",
	"writeProcNew",
	"runProc",
	"runClient",
	"log",
	"inputLine",
	"setSync",
	"rawKey",
	"regionSelect",
	"buttonHit",
	"flushEffect",
	"textResize",
	"incRef",
	"decRef",
	"symbolEnter",
	"symbolDelete",
	"useTable",
	"unuseTable",
	"describeBuiltin",
	"describe",
	"replyPassword",
	"graphicsFlip",
	"voiceFlip",
	"soundFlip",
	"musicFlip",
	"doEditProc",
	"editStringDone",
	"replaceProc",
	"editProcDone",
	"getStringDone",
	"effectDone",
	"queryFileDone",
	"illegal"
    };
#endif

/*
 * serverRequest - send the current request to the server and wait for
 *	a reply. Handle other requests coming back.
 */

static Request_t *
serverRequest(Request_t *rq)
{
    RequestType_t rt;
    BOOL_T gotReply;

    DEBUG_M(("::serverRequest %p type %d used %d avail %d::\n",
	     rq, rq->rq_type, rq->rq_usedLen, rq->rq_availLen));
    rt = rq->rq_type;
    gotReply = B_FALSE;
    serverSend(rq);
    rq = P_NULL;	/* In case loop not done at all */
    while (! gotReply && ! ShuttingDown) {
	rq = waitForStuff(P_NULL, B_FALSE);
	if (rq != P_NULL) {
	    if (rq->rq_type == rt) {
		gotReply = B_TRUE;
	    } else {
		serverMessage(&MS, rq);
	    }
	}
    }
    checkTimeOut();
    /* Cannot return a null rq. */
    if (rq == P_NULL) {
	DEBUG_M(("::serverRequest got no reply - aborting::\n"));
	MUDAbort(&MS);
	smudAbort();
    }
    DEBUG_M(("::serverRequest returning rq %p::\n", rq));
    return rq;
}

/*
 * editInput - read some input from an editor session.
 */

static void
editInput(void)
{
#define LINE_LENGTH	256
    char line[LINE_LENGTH + 1];
    LONG_T len;
    UINT_T pos;
    char ch;

    DEBUG(("::editInput::\n"));
    pos = 0;
    while (B_TRUE) {
	if (EditPos == EditMax) {
	    len = read(EditFd, &EditBuffer[0], INPUT_BUFFER_SIZE);
	    if (len <= 0) {
		if (pos != 0) {
		    line[pos] = '\0';
		    DEBUG_I(("::editInput: >%s<::\n", &line[0]));
		    inputLine(&MS, &line[0], pos);
		}
		DEBUG(("::editInput got EOF - edit done::\n"));
		EditReading = B_FALSE;
		(void) close(EditFd);
		editProcDone(&MS);
		EditFirst = B_TRUE;
		EditFailing = B_FALSE;
		break;
	    }
	    EditMax = len;
	    EditPos = 0;
	} else {
	    ch = EditBuffer[EditPos];
	    EditPos += 1;
	    if (ch == '\n') {
		line[pos] = '\0';
		DEBUG_I(("::editInput: >%s<::\n", &line[0]));
		inputLine(&MS, &line[0], pos);
		break;
	    }
	    if (pos < LINE_LENGTH) {
		line[pos] = ch;
		pos += 1;
	    }
	}
    }
#undef LINE_LENGTH
}

/*
 * fileInput - do some non-interactive reading of input.
 */

static void
fileInput(void)
{
#define LINE_LENGTH	256
    char line[LINE_LENGTH + 1];
    LONG_T len;
    UINT_T pos;
    char ch;

    DEBUG(("::fileInput::\n"));
    PromptShown = B_FALSE;
    NeedPrompt = B_FALSE;
    pos = 0;
    while (B_TRUE) {
	if (InputPos == InputMax) {
	    len = read(0, &InputBuffer[0], INPUT_BUFFER_SIZE);
	    if (len <= 0) {
		DEBUG(("::fileInput got EOF::\n"));
		if (pos != 0) {
		    line[pos] = '\0';
		    DEBUG_I(("::fileInput: >%s<::\n", &line[0]));
		    inputLine(&MS, &line[0], pos);
		}
		MUDShutDown(&MS);
		ShuttingDown = B_TRUE;
		break;
	    }
	    DEBUG(("::fileInput got %ld::\n", len));
	    InputMax = len;
	    InputPos = 0;
	}
	ch = InputBuffer[InputPos];
	InputPos += 1;
	if (ch == '\n') {
	    line[pos] = '\0';
	    DEBUG_I(("::fileInput >%s<::\n", &line[0]));
	    inputLine(&MS, &line[0], pos);
	    break;
	}
	if (pos < LINE_LENGTH) {
	    line[pos] = ch;
	    pos += 1;
	}
    }
#undef LINE_LENGTH
}

/*
 * handleInput - something detected coming from the player.
 */

static void
handleInput(void)
{
    LONG_T len;

    DEBUG(("::handleInput::\n"));
    len = read(0, &InputBuffer[0], INPUT_BUFFER_SIZE);
    if (len <= 0) {
	/* error or end of file */
	(void) write(1, "\n", 1);
	MUDShutDown(&MS);
	ShuttingDown = B_TRUE;
    } else {
	if (! EchoOn) {
	    (void) write(1, "\n", 1);
	}
	PromptShown = B_FALSE;
	(void) time(&LastInputTime);
	if (InputBuffer[len - 1] == '\n') {
	    len -= 1;
	}
	InputBuffer[len] = '\0';
	DEBUG_I(("::handleInput >%s<::\n", &InputBuffer[0]));
	inputLine(&MS, &InputBuffer[0], len);
	NeedPrompt = B_TRUE;
    }
}

/*
 * handleAll - MUD library is idle - wait for something to happen.
 */

static void
handleAll(BOOL_T wantInput)
{
    Request_t *rq;
    BOOL_T first, hadOne, gotInput;

    DEBUG(("::handleAll %c::\n", wantInput ? 'T' : 'F'));
    /* First, handle anything coming from the server. */
    hadOne = B_FALSE;
    while (! ShuttingDown && ! TimedOut) {
	rq = waitForStuff(P_NULL, B_TRUE);
	if (rq == P_NULL) {
	    break;
	}
	hadOne = B_TRUE;
	serverMessage(&MS, rq);
    }
    if (wantInput) {
	if (Interactive) {
	    doPrompt();
	}
	first = B_TRUE;
	while (first && ! ShuttingDown && ! TimedOut) {
	    rq = waitForStuff(&gotInput, EditReading);
	    if (rq != P_NULL) {
		serverMessage(&MS, rq);
	    } else if (! gotInput) {
		checkTimeOut();
	    }
	    if (EditReading) {
		editInput();
		first = B_FALSE;
	    } else if (Interactive) {
		if (gotInput) {
		    handleInput();
		    first = B_FALSE;
		}
	    } else {
		fileInput();
		first = B_FALSE;
	    }
	}
    } else {
	if (! hadOne) {
	    if (Interactive) {
		doPrompt();
	    }
	    rq = waitForStuff(P_NULL, B_FALSE);
	    if (rq != P_NULL) {
		serverMessage(&MS, rq);
	    }
	}
    }
}

/*
 * checkMessages - MUD library is busy, but taking a while. This is a
 *	kind of poll for server stuff that needs handling.
 */

static void
checkMessages(void)
{
    Request_t *rq;

    DEBUG_M(("::checkMessages::\n"));
    /* Do not want to do any 'fileInput' stuff here! */
    while (B_TRUE) {
	rq = waitForStuff(P_NULL, B_TRUE);
	if (rq == P_NULL) {
	    break;
	}
	serverMessage(&MS, rq);
    }
}

/*
 * getEffects - tell the library we do not do effects
 */

static void
getEffects(EffectsInfo_t *ei)
{

    ei->ei_graphicsType[0] = '\0';
    ei->ei_graphicsRows = 0;
    ei->ei_graphicsCols = 0;
    ei->ei_graphicsColours = 0;
    ei->ei_textWidth = 0;
    ei->ei_textHeight = 0;
    ei->ei_version = 0;
    ei->ei_canEdit = ! IsRemote;
    ei->ei_canGetString = B_FALSE;
    ei->ei_canQueryFile = B_FALSE;
    ei->ei_canWizard = B_TRUE;
    ei->ei_graphicsOn = B_FALSE;
    ei->ei_graphicsPalette = B_FALSE;
    ei->ei_voiceOn = B_FALSE;
    ei->ei_soundOn = B_FALSE;
    ei->ei_musicOn = B_FALSE;
}

/*
 * doEffects - do not do effects even if the server tries.
 */

static void
doEffects(BYTE_T *buf, UINT_T len)
{
}

/*
 * getString - do not try to get a string even if the server tries.
 */

static BOOL_T
getString(CHAR_T *prompt, CHAR_T *initial, CHAR_T **pRes)
{
    return B_FALSE;
}

/*
 * shutDownAndAbort - an abort call from the library - shutdown cleanly.
 */

static void
shutDownAndAbort(BOOL_T isServerNuke)
{

    DEBUG(("::shutDownAndAbort::\n"));
    smudAbort();
}

/*
 * Some stubs to convert parameter conventions.
 */

static void *
myAllocMem(ULONG_T len)
{
    return malloc(len);
}

static void
myFreeMem(void *ptr, ULONG_T len)
{
    free(ptr);
}

static void
handleSigChld(int signum)
{
    EditDone = B_TRUE;
}

static void
mySignal(int signum, void (*handler)(int))
{
    struct sigaction sa;
    int flags;

    flags = SA_RESTART;
    if (signum == SIGCHLD) {
	flags |= SA_NOCLDSTOP;
    }
    sa.sa_handler = handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = flags;
    if (sigaction(signum, &sa, P_NULL) < 0) {
	myPerror("sigaction");
	smudAbort();
    }
}

/*
 * startup2 - third level of running.
 */

static void
startup2(UINT_T MUDflags)
{

    ShuttingDown = B_FALSE;
    EchoOn = B_TRUE;
    PromptBuffer[0] = '\r';
    EraseBuffer[0] = '\r';
    EraseBuffer[1] = '\r';
    PromptLen = 0;
    InputMax = 0;
    InputPos = 0;
    EditDone = B_FALSE;
    EditReading = B_FALSE;
    EditFirst = B_TRUE;
    EditFailing = B_FALSE;
    PromptShown = B_FALSE;
    NeedPrompt = B_TRUE;
    TimedOut = B_FALSE;
    Interactive = (isatty(0) != 0);

    MS.ms_AllocMem = myAllocMem;
    MS.ms_FreeMem = myFreeMem;
    MS.ms_ReclaimMem = P_NULL;
    MS.ms_RetryMem = P_NULL;
    MS.ms_Open = (int (*)(const char *, int, ...)) open;
    MS.ms_Close = close;
    MS.ms_Read = read;

    MS.ms_abort = shutDownAndAbort;
    MS.ms_errorChar = errorChar;
    MS.ms_outputChar = errorChar1;

    MS.ms_incRef = P_NULL;
    MS.ms_decRef = P_NULL;
    MS.ms_lookup = P_NULL;
    MS.ms_findAnyName = P_NULL;
    MS.ms_findProc = P_NULL;
    MS.ms_writeProcNew = P_NULL;
    MS.ms_findBuiltin = P_NULL;
    MS.ms_getIntConstValue = P_NULL;

    MS.ms_putText = putText;
    MS.ms_setPrompt = setPrompt;
    MS.ms_setEcho = setEcho;
    MS.ms_serverSend = serverSend;
    MS.ms_serverRequest = serverRequest;
    MS.ms_handleAll = handleAll;
    MS.ms_checkMessages = checkMessages;
    MS.ms_getEffects = getEffects;
    MS.ms_doEffects = doEffects;
    MS.ms_doEdit = doEdit;
    MS.ms_editErrorStart = P_NULL;
    MS.ms_getString = getString;
    MS.ms_defineEffect = P_NULL;
    MS.ms_flushEffect = P_NULL;

    if (MUDInitialize(&MS, MUDflags, 10)) {
	(void) time(&LastInputTime);
	mySignal(SIGINT, handleCNTLC);
	mySignal(SIGTERM, handleCNTLC);
	mySignal(SIGCHLD, handleSigChld);
	MUDMain(&MS, ForcePlayerName, ForcePassword);
	if (PromptShown && ! NoCRs) {
	    (void) write(1, &EraseBuffer[0], PromptLen + 2);
	}
	ShuttingDown = B_TRUE;	/* for 'doRecv' */
	MUDTerminate(&MS);
	if (! EchoOn) {
	    setEcho(B_TRUE);
	}
    } else {
	complain1("mud library initialization failed");
    }
}

/*
 * startup - second level of running.
 */

static void
startup(char *hostName, short int port, UINT_T flags, BOOL_T forceNoDelay)
{
    struct sockaddr_in addr;
    struct hostent *hp;
    int on;
    BOOL_T ok;

    if (tmpnam(&EditFileName[0]) == P_NULL) {
	fprintf(stderr, "::Can't create unique edit file name::\n");
	return;
    }
    OutputBufferLength = INITIAL_OUTPUT_BUFFER;
    OutputBuffer = malloc(OutputBufferLength);
    if (OutputBuffer != P_NULL) {
	hp = gethostbyname(hostName);
	if (hp == P_NULL) {
	    fprintf(stderr, "::Can't find host '%s': %s::\n",
		    hostName, strerror(errno));
	} else {
	    Socket = socket(AF_INET, SOCK_STREAM, 0);
	    if (Socket < 0) {
		myPerror("socket");
	    } else {
		ok = B_TRUE;
		if (forceNoDelay || strcmp(hostName, "localhost") == 0) {
		    on = 1;
		    if (setsockopt(Socket, IPPROTO_TCP, TCP_NODELAY,
				   (char *) &on, sizeof(int)) != 0)
		    {
			myPerror("setsockopt");
			ok = B_FALSE;
		    }
		}
		if (ok) {
		    memset(&addr, 0, sizeof(struct sockaddr_in));
#if 0
		    addr.sin_len = sizeof(struct sockaddr_in);
#endif
		    addr.sin_family = AF_INET;
		    addr.sin_port = htons(port);
		    addr.sin_addr.s_addr = *(ULONG_T *) hp->h_addr;
		    if (connect(Socket, (struct sockaddr *) &addr,
				sizeof(struct sockaddr_in)) < 0)
		    {
			fprintf(stderr,
			  "::Can't connect to server at '%s', port %d: %s::\n",
				hostName, port, strerror(errno));
		    } else {
			startup2(flags);
		    }
		    shutdown(Socket, 2);
		    close(Socket);
		    (void) unlink(&EditFileName[0]);
		}
	    }
	}
	free(OutputBuffer);
    } else {
	fprintf(stderr, "::Can't allocate output buffer::\n");
    }
}

/*
 * main - start everything up.
 */

int
main(int argc, char **argv)
{
    char *par;
    char *hostName;
    short int port;
    char ch;
    UINT_T flags;
    BOOL_T hadError, seenHost, seenPort, forceNoDelay;

    hadError = B_FALSE;
    seenHost = B_FALSE;
    seenPort = B_FALSE;
    forceNoDelay = B_FALSE;
    IsRemote = B_FALSE;
    NLConvert = B_FALSE;
    NoCRs = B_FALSE;
    TimeOut = DEFAULT_TIMEOUT;
    hostName = DEFAULT_HOST;
    port = DEFAULT_PORT;
    flags = FLAG_CACHE_PROCS | FLAG_CACHE_SYMBOLS;
    ForcePlayerName = P_NULL;
    ForcePassword = P_NULL;
    argc -= 1;
    ++argv;
    while (argc != 0) {
	argc -= 1;
	par = *argv++;
	if (*par == '-') {
	    ++par;
	    while (B_TRUE) {
		ch = *par;
		if (ch == '\0') {
		    break;
		}
		++par;
		switch (ch) {
		case 'n':
		    NLConvert = B_TRUE;
		    break;
		case 'c':
		    NoCRs = B_TRUE;
		    break;
		case 'r':
		    IsRemote = B_TRUE;
		    break;
		case 't':
		    TimeOut = 0;
		    while (B_TRUE) {
			ch = *par;
			if (ch < '0' || ch > '9') {
			    break;
			}
			++par;
			TimeOut *= 10;
			TimeOut += (ch - '0');
		    }
		    if (*par != '\0') {
			fprintf(stderr, "Invalid number for -t\n");
			hadError = B_TRUE;
			while (*par != '\0') {
			    ++par;
			}
		    }
		    break;
		case 'f':
		    flags = 0x0;
		    while (B_TRUE) {
			ch = *par;
			if (ch < '0' || ch > '9') {
			    break;
			}
			++par;
			flags *= 10;
			flags += (ch - '0');
		    }
		    if (*par != '\0') {
			fprintf(stderr, "Invalid number for -f\n");
			hadError = B_TRUE;
			while (*par != '\0') {
			    ++par;
			}
		    }
		    break;
		case 'F':
		    forceNoDelay = B_TRUE;
		    break;
		case 'N':
		    ForcePlayerName = par;
		    par = "";
		    break;
		case 'W':
		    ForcePassword = par;
		    par = "";
		    break;
		default:
		    hadError = B_TRUE;
		    break;
		}
	    }
	} else {
	    if (seenHost) {
		if (seenPort) {
		    hadError = B_TRUE;
		} else {
		    seenPort = B_TRUE;
		    port = 0;
		    while (B_TRUE) {
			ch = *par;
			if (ch < '0' || ch > '9') {
			    break;
			}
			++par;
			port *= 10;
			port += (ch - '0');
		    }
		    if (*par != '\0') {
			fprintf(stderr, "Invalid number for -p\n");
			hadError = B_TRUE;
			while (*par != '\0') {
			    ++par;
			}
		    }
		}
	    } else {
		seenHost = B_TRUE;
		hostName = par;
	    }
	}
    }
    if (hadError) {
	fprintf(stderr,
	   "Use is: SMUD [-n -c -r -t# -f# -F -Nname -Wword] [host [port]]\n");
	return 1;
    } else {
	startup(hostName, port, flags, forceNoDelay);
	return 0;
    }
}
